/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altdrv.h,v $
 * Revision 1.1.5.88  1999/01/30  02:32:13  fhwang
 * 	add ALT_SNAP define.
 * 	[1999/01/30  00:58:09  fhwang]
 *
 * Revision 1.1.5.87  1999/01/07  23:09:12  ted
 * 	Add locks for debugger enabled NT driver
 * 	[1999/01/07  23:03:32  ted]
 * 
 * Revision 1.1.5.86  1998/12/23  16:21:31  leonid
 * 	MiniportReset changed from synchronous to asynchronous operation to fix 1c_resetfailures.
 * 	[1998/12/23  16:20:37  leonid]
 * 
 * Revision 1.1.5.85  1998/11/19  00:15:56  dfoy
 * 	Fixed FILL_IN_SEND_BD for VLAN support.
 * 	[1998/11/18  23:58:59  dfoy]
 * 
 * Revision 1.1.5.84  1998/11/16  23:35:04  ted
 * 	Add registry support for the READ_MAX_DMA value
 * 	[1998/11/16  23:34:39  ted]
 * 
 * Revision 1.1.5.83  1998/11/09  19:33:44  ted
 * 	support jumbos on x86 Solaris (test extended recv BD)
 * 	[1998/11/09  19:33:17  ted]
 * 
 * Revision 1.1.5.82  1998/11/04  21:35:40  ted
 * 	Fix send ring full backout by updating producer anyway
 * 	[1998/11/04  21:35:31  ted]
 * 
 * Revision 1.1.5.81  1998/11/04  17:42:35  hayes
 * 	change conditional locks from ALT_DESERIALIZED to USE_INTERNAL_QUEUE
 * 	[1998/11/04  17:38:58  hayes]
 * 
 * Revision 1.1.5.80  1998/11/03  23:25:20  ted
 * 	data structures for new NT internal queue, altReapSendRing recursion and checksum offload failure stats
 * 	[1998/11/03  22:08:11  ted]
 * 
 * Revision 1.1.5.79  1998/10/13  03:33:31  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:02  ted]
 * 
 * Revision 1.1.5.78  1998/09/09  01:01:22  hayes
 * 	fixed struct copy in FILL_IN_SEND_BD() for host based send rings
 * 	[1998/09/09  01:00:07  hayes]
 * 
 * Revision 1.1.5.77  1998/09/03  22:16:47  hayes
 * 	add bugfixes from nt_r1_9_5
 * 	[1998/09/03  22:13:14  hayes]
 * 
 * Revision 1.1.5.76  1998/09/03  20:04:55  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:04  ted]
 * 
 * Revision 1.1.5.75  1998/09/02  21:00:46  ted
 * 	Upgrade Tigon 2 nics to new better interfaces
 * 	[1998/09/02  20:57:45  ted]
 * 
 * Revision 1.1.5.74  1998/08/27  18:41:04  hayes
 * 	fix merge for solaris code... :-(
 * 	[1998/08/27  18:40:53  hayes]
 * 
 * Revision 1.1.5.73  1998/08/27  16:57:22  hayes
 * 	fix send ring window on backup
 * 	[1998/08/27  16:56:54  hayes]
 * 
 * Revision 1.1.5.72  1998/08/20  23:56:52  hayes
 * 	fix REAP_FROM_SEND
 * 	[1998/08/20  23:56:18  hayes]
 * 
 * Revision 1.1.5.70  1998/08/12  01:57:58  ted
 * 	bci by hand because bmerge ran out of open file descriptors, sigh
 * 	[1998/08/10  22:47:52  ted]
 * 
 * 	First shot at real deserialization
 * 	[1998/08/06  18:10:54  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:01:52  ted]
 * 
 * Revision 1.1.5.69  1998/08/06  20:11:24  hayes
 * 	more minDma to common from NDIS only
 * 	[1998/08/06  20:11:12  hayes]
 * 
 * Revision 1.1.5.68  1998/08/03  22:06:26  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:09  hayes]
 * 
 * Revision 1.1.5.67  1998/08/03  17:11:19  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:33  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:48  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:20  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:42  hayes]
 * 
 * Revision 1.1.5.69  1998/08/06  20:11:24  hayes
 * 	more minDma to common from NDIS only
 * 	[1998/08/06  20:11:12  hayes]
 * 
 * Revision 1.1.5.68  1998/08/03  22:06:26  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:09  hayes]
 * 
 * Revision 1.1.5.67  1998/08/03  17:11:19  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:33  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:48  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:20  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:42  hayes]
 * 
 * Revision 1.1.5.69  1998/08/06  20:11:24  hayes
 * 	more minDma to common from NDIS only
 * 	[1998/08/06  20:11:12  hayes]
 * 
 * Revision 1.1.5.68  1998/08/03  22:06:26  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:09  hayes]
 * 
 * Revision 1.1.5.67  1998/08/03  17:11:19  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:33  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:48  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:20  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:42  hayes]
 * 
 * Revision 1.1.5.66  1998/07/14  18:34:49  hayes
 * 	cleanup locks
 * 	[1998/07/14  18:14:49  hayes]
 * 
 * Revision 1.1.5.65  1998/07/14  02:32:35  ted
 * 	shutdown fixes, deserialized send queue
 * 	[1998/07/14  02:12:33  ted]
 * 
 * Revision 1.1.5.64  1998/07/08  22:26:24  hayes
 * 	add 450gx fix
 * 	[1998/07/08  22:24:45  hayes]
 * 
 * Revision 1.1.5.63  1998/07/02  05:30:41  andy
 * 	Collect more cksum stats
 * 	[1998/07/02  05:21:46  andy]
 * 
 * Revision 1.1.5.62  1998/06/18  17:35:23  hayes
 * 	fix merge for Solaris
 * 	[1998/06/18  17:35:13  hayes]
 * 
 * Revision 1.1.5.61  1998/06/18  00:06:32  hayes
 * 	add caching fixes
 * 	[1998/06/17  23:57:56  hayes]
 * 
 * Revision 1.1.5.60  1998/06/16  16:45:44  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:42:56  ted]
 * 
 * Revision 1.1.5.59  1998/06/12  23:47:11  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:07  ted]
 * 
 * 	snapshot of merge with new Registry handling
 * 	[1998/06/10  22:15:12  ted]
 * 
 * Revision 1.1.5.58  1998/06/08  23:55:27  ted
 * 	Generify RingDequeueMsgs
 * 	[1998/06/04  23:40:13  ted]
 * 
 * 	Snapshot NT generification phase 1
 * 	[1998/06/04  19:52:15  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:11  ted]
 * 
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:08:35  ted]
 * 
 * 	snapshot 12.0 NT jumbo support
 * 	[1998/05/20  20:29:19  ted]
 * 
 * Revision 1.1.5.57  1998/05/20  05:38:47  andy
 * 	collect detail cksum stats
 * 	[1998/05/20  02:56:26  andy]
 * 
 * Revision 1.1.5.56  1998/05/14  02:56:26  andy
 * 	checksum debug logic
 * 	[1998/05/14  02:44:49  andy]
 * 
 * Revision 1.1.5.55  1998/05/01  05:35:08  hayes
 * 	alpha NT works with transparent shmem accesses
 * 	[1998/05/01  05:34:38  hayes]
 * 
 * 	checkpoint new ALPHA NT initialization and memory handling
 * 	[1998/04/29  23:32:16  hayes]
 * 
 * Revision 1.1.5.54  1998/04/24  01:40:34  hayes
 * 	remove obsolete vars
 * 	[1998/04/24  01:35:27  hayes]
 * 
 * Revision 1.1.5.53  1998/04/17  03:08:21  ted
 * 	Add check for fragment receive bailout
 * 	[1998/04/17  03:07:28  ted]
 * 
 * Revision 1.1.5.52  1998/04/14  21:42:35  ted
 * 	stuff for VLAN assist and new 12.0 tuning parms
 * 	[1998/04/14  21:25:15  ted]
 * 
 * Revision 1.1.5.51  1998/04/07  06:35:51  ted
 * 	bring ODI driver fixes from 1.8 release forward to ge_latest
 * 	[1998/04/07  06:34:10  ted]
 * 
 * Revision 1.1.5.50  1998/04/02  06:03:08  ted
 * 	Add checksum debugging support
 * 	[1998/04/02  05:55:39  ted]
 * 
 * Revision 1.1.5.49  1998/03/11  22:00:00  hayes
 * 	fix solaris sheme changes
 * 	[1998/03/11  21:59:39  hayes]
 * 
 * Revision 1.1.5.48  1998/03/09  21:36:56  venkat
 * 	introduced shared memory access MACROs to let individual platforms define
 * 	them to be system specific
 * 	[1998/03/09  21:24:44  venkat]
 * 
 * Revision 1.1.5.47  1998/03/03  22:23:44  ted
 * 	Implement faster dual homing code
 * 	[1998/03/03  22:22:46  ted]
 * 
 * Revision 1.1.5.46  1998/02/04  02:56:49  ted
 * 	Add command line option support for flow control to ODI
 * 	[1998/02/04  02:52:23  ted]
 * 
 * Revision 1.1.5.45  1998/01/24  04:06:49  venkat
 * 	moved SEND_RING_PTR to alttypes.h, new ENQ_SEND_RING and SET_NIC_MEM_WINDOW
 * 	[1998/01/24  04:01:41  venkat]
 * 
 * Revision 1.1.5.44  1998/01/09  19:05:36  ted
 * 	put stats info back into Tigon 2 send buffer descriptors
 * 	[1998/01/09  19:04:41  ted]
 * 
 * Revision 1.1.5.43  1998/01/05  01:31:54  ted
 * 	additional stat, removal of 'int' keyword from common structure tg_nic
 * 	[1998/01/05  01:03:11  ted]
 * 
 * Revision 1.1.5.42  1997/12/31  02:59:41  venkat
 * 	cleaned up code related to FDR mode. TIGON2 and TIGON1 are both working now.
 * 	[1997/12/31  02:59:05  venkat]
 * 
 * Revision 1.1.5.41  1997/12/23  19:49:00  ted
 * 	add multicast mode indicator to flag bits
 * 	[1997/12/23  19:28:51  ted]
 * 
 * Revision 1.1.5.40  1997/12/18  23:21:25  venkat
 * 	merged from ge_0_37_latest
 * 	[1997/12/18  23:14:27  venkat]
 * 
 * Revision 1.1.5.39  1997/12/15  20:03:05  ted
 * 	New fields and flags for "soft" reset reinitialization
 * 	Fix TX stats generation for non-copydata code
 * 	[1997/12/15  19:47:30  ted]
 * 
 * Revision 1.1.5.38  1997/12/03  01:22:56  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:15  ted]
 * 
 * Revision 1.1.5.37  1997/12/01  20:29:39  ted
 * 	Fix bugs 590 and 591
 * 	[1997/12/01  20:28:58  ted]
 * 
 * Revision 1.1.5.36  1997/11/27  02:53:17  ted
 * 	Get Suspend/Resume to not crash
 * 	[1997/11/27  02:52:49  ted]
 * 
 * Revision 1.1.5.35  1997/11/19  19:23:51  ted
 * 	Support Tigon and Tigon2 in the same driver
 * 	[1997/11/19  19:20:48  ted]
 * 
 * Revision 1.1.5.34  1997/11/19  00:13:59  ted
 * 	merged from release 2.0
 * 	[1997/11/19  00:11:07  ted]
 * 
 * Revision 1.1.5.33  1997/11/14  19:17:11  hayes
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:09  hayes]
 * 
 * Revision 1.1.5.32  1997/10/28  17:11:28  hayes
 * 	New bits for supporting cleanup after firmware didn't start
 * 	[1997/10/28  01:59:05  hayes]
 * 
 * Revision 1.1.5.31  1997/10/21  00:12:56  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  22:59:58  ted]
 * 
 * Revision 1.1.33.1  1997/10/16  17:30:35  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:24:29  ted]
 * 
 * Revision 1.1.5.30  1997/09/16  03:11:20  ted
 * 	Change s to spl for better coding style
 * 	[1997/09/16  03:00:17  ted]
 * 
 * Revision 1.1.5.29  1997/08/23  17:40:26  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:38:40  skapur]
 * 
 * Revision 1.1.5.28  1997/07/29  01:13:46  ted
 * 	Merge by hand due to bsubmit core dump, sigh
 * 	[1997/07/29  01:13:05  ted]
 * 
 * Revision 1.1.5.27  1997/07/28  22:40:23  ted
 * 	Merge
 * 	[1997/07/28  22:37:16  ted]
 * 
 * 	Solaris dual homing support
 * 	[1997/07/28  21:39:34  ted]
 * 
 * Revision 1.1.5.26  1997/07/27  22:03:49  ted
 * 	Fix merge bug
 * 	[1997/07/27  22:00:52  ted]
 * 
 * Revision 1.1.5.25  1997/07/26  00:28:47  ted
 * 	Fix merge errors
 * 	[1997/07/26  00:28:13  ted]
 * 
 * Revision 1.1.5.24  1997/07/25  22:59:35  ted
 * 	Merged with changes from 1.1.5.23
 * 	[1997/07/25  22:57:49  ted]
 * 
 * 	Merge fixes from ge_0_22 and add link negotiation support
 * 	[1997/07/25  22:33:05  ted]
 * 
 * 	Architectural performance change
 * 	[1997/07/24  21:45:46  ted]
 * 
 * 	Support flow control for STREAMS
 * 	[1997/07/19  20:37:33  ted]
 * 
 * 	New receive mutexes for finer mutex control
 * 	[1997/07/07  22:55:52  ted]
 * 
 * Revision 1.1.5.23  1997/07/09  00:24:23  akhan
 * 	Merged code from ge_0_24_latest branch into the ge_latest.
 * 	Added SNOOZE in ENQ_COMMAND_RING
 * 	[1997/07/09  00:15:45  akhan]
 * 
 * Revision 1.1.5.22  1997/07/08  00:55:25  suresh
 * 	Change occurences of "#if define(ALT_RECV_EXT)" to
 * 	"#if ALT_RECV_EXT" in altdrv.h. Adjust altdrv.c accordingly.
 * 	[1997/07/08  00:54:54  suresh]
 * 
 * Revision 1.1.5.21  1997/07/05  21:17:40  ted
 * 	Fix naming issues as reported in bug 317
 * 	[1997/07/05  21:15:00  ted]
 * 
 * Revision 1.1.5.20  1997/06/17  21:29:17  ted
 * 	fix critical section bug
 * 	[1997/06/17  21:29:08  ted]
 * 
 * Revision 1.1.5.19  1997/06/17  00:36:44  ted
 * 	Merged with changes from 1.1.5.18
 * 	[1997/06/17  00:36:15  ted]
 * 
 * 	Data structures for improved performance on overflow logic
 * 	[1997/06/12  18:08:19  ted]
 * 
 * 	tighten up driver code for better performance
 * 	[1997/06/06  22:38:26  ted]
 * 
 * Revision 1.1.5.18  1997/06/11  02:44:59  akhan
 * 	Added NT specific variables related to send queuing.
 * 	[1997/06/11  02:44:48  akhan]
 * 
 * Revision 1.1.5.17  1997/05/03  05:00:26  ted
 * 	Add Intel flash support and fix x86_solaris compile problem
 * 	[1997/04/27  21:31:50  ted]
 * 
 * Revision 1.1.5.16  1997/04/26  01:08:54  akhan
 * 	NDIS and ODI changes.
 * 	[1997/04/26  01:04:36  akhan]
 * 
 * Revision 1.1.5.15  1997/04/25  18:49:00  ted
 * 	Support single driver for Sbus and PCI on Solaris
 * 	[1997/04/25  18:44:26  ted]
 * 
 * Revision 1.1.5.14  1997/04/12  00:11:24  ted
 * 	Merged with changes from 1.1.5.13
 * 	[1997/04/12  00:10:48  ted]
 * 
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:07:31  ted]
 * 
 * 	Working SBUS code prior to MP optimization
 * 	[1997/04/06  19:47:50  ted]
 * 
 * Revision 1.1.5.13  1997/04/05  00:39:30  suresh
 * 	missing #endif
 * 	[1997/04/05  00:39:17  suresh]
 * 
 * Revision 1.1.5.12  1997/04/04  23:14:06  suresh
 * 	AIX related merges
 * 	[1997/04/04  23:13:22  suresh]
 * 
 * Revision 1.1.5.11  1997/04/04  22:38:23  ted
 * 	Merged with changes from 1.1.5.10
 * 	[1997/04/04  22:37:50  ted]
 * 
 * 	Data structures for receive overflow ring.
 * 	[1997/04/04  22:14:09  ted]
 * 
 * Revision 1.1.5.10  1997/04/04  06:22:26  akhan
 * 	added xmit_acked member in the instance structure.
 * 	[1997/04/04  05:44:43  akhan]
 * 
 * 	Code merge for NT
 * 	[1997/04/02  18:16:05  akhan]
 * 
 * Revision 1.1.5.9  1997/03/28  01:39:44  ted
 * 	Major merge of NDIS and ODI with mainstream Unix ports
 * 	[1997/03/28  01:32:32  ted]
 * 
 * Revision 1.1.5.8  1997/03/24  08:34:37  ted
 * 	Merged with changes from 1.1.5.7
 * 	[1997/03/24  08:33:49  ted]
 * 
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:28  ted]
 * 
 * Revision 1.1.5.7  1997/03/19  06:19:41  suresh
 * 	AIX related changes to tg_nic
 * 	[1997/03/19  06:19:26  suresh]
 * 
 * Revision 1.1.5.6  1997/03/17  08:47:57  skapur
 * 	Add second recv ring structure and watchdog flag.
 * 	[1997/03/17  08:36:33  skapur]
 * 
 * Revision 1.1.5.5  1997/03/02  18:36:51  ted
 * 	redundancy support and a few bug fixes
 * 	[1997/03/02  18:25:47  ted]
 * 
 * Revision 1.1.5.4  1997/02/21  02:06:41  ted
 * 	Prepare data structures and code structures to support redundant operation
 * 	[1997/02/21  01:56:45  ted]
 * 
 * Revision 1.1.5.3  1997/02/15  04:41:16  ted
 * 	Change ALT_LOCKS_SUPPORTED from a #if defined to a #if check
 * 	[1997/02/15  03:05:07  ted]
 * 
 * Revision 1.1.5.2  1997/02/12  00:28:06  ted
 * 	Implement ndd support using code borrowed from Sun
 * 	[1997/02/12  00:27:38  ted]
 * 
 * Revision 1.1.5.1  1997/02/07  21:58:02  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:47:46  ted]
 * 
 * Revision 1.1.2.2  1996/11/24  04:02:37  ted
 * 	Removal of unused structure
 * 	[1996/11/24  04:00:56  ted]
 * 
 * Revision 1.1.2.1  1996/11/21  18:01:34  ted
 * 	Moved from ../unix/common and merged in WINNT changes
 * 	[1996/11/21  17:42:06  ted]
 * 
 * Revision 1.6.2.2  1996/11/20  20:04:43  ted
 * 	Add new structure level for gen_com.
 * 	New genComm structure type (tg_shmem_t).
 * 	[1996/11/20  19:56:12  ted]
 * 
 * Revision 1.6.2.1  1996/11/01  18:25:32  ted
 * 	change rr_ to tg_ and add proper copyright
 * 	[1996/10/30  22:58:32  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altdrv.h
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/nic/host/common/altdrv.h,v $
 * $Revision: 1.1.5.88 $ $Locker:  $
 * $Date: 1999/01/30 02:32:13 $ $Author: fhwang $ $State: Exp $
 */

#ifndef _ALTDRV_H
#define _ALTDRV_H

/* macro for incrementing the ring pointer */
#define INCRING(i, size) (((i) + 1) & ((size) -1))
#define DECRING(i, size) (((i) - 1) & ((size) -1))

/* illegal tag value recognized as "don't insert vlan tag" by 
 * altQueueSendMsg() */
#define NO_VLAN_TAG 0xffffffff

/* local data context */
/* TG only information
 */
struct tg_nic {
    char eyeCatch[8];		/* eye catcher for dumps */
    char name[80];		/* space for OEM name */
    U32 dmaPrepsActive;		/* DMA PREPS outstanding */
    U32 flags;			/* flag bits, below */
    U32 init_flags;		/* flag bits, below */
    bool watchdog_ok;		/* don't need to touch rest of flags... */
    bool watchdog_first;	/* don't need to touch rest of flags... */
#if USE_CKSUM_OFFLOAD
    U32 csumFlags;		/* flag bits for checksum offload, see below */
#endif /* USE_CKSUM_OFFLOAD */
    struct altinst *altp;	/* pointer to stack's instance struct */
    struct altinst *initAltp;	/* pointer to initializing instance */
    caddr_t nicDump;		/* ptr to allocated mem for nic memory dump */
    caddr_t nicRegs;		/* ptr to allocated mem for nic hw reg dump */
    U32 tigonRev;		/* Tigon revision of this piece of hardware */
#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE flagMu;		/* flag bits mutex */
#endif /* ALT_LOCKS_SUPPORTED */

    /* interrupt information */
#if defined(ALT_SYSV)
    ddi_iblock_cookie_t iblock_cookie; /* interrupt block cookie */
    ddi_idevice_cookie_t idev_cookie; /* interrupt cookie */
#endif /* defined(ALT_SYSV) */
#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE intrMu;		/* keep rupts out of each other */
    LOCK_TYPE memWindowMu;	/* keep the memory window happy */
#if REAP_FROM_SEND
    LOCK_TYPE sendReapMu;	/* protect reaping send bds from send pkt */
    bool inSendReap;		/* keep track for recursion on thread */
#endif /* REAP_FROM_SEND */
#if ALT_DESERIALIZE
    LOCK_TYPE infopQueueMu;	/* for the infop free queue */
#endif /* ALT_DESERIALIZE */
#if USE_INTERNAL_QUEUE
    LOCK_TYPE sendQueueMu;	/* have to do our own packet queuing */
#endif /* USE_INTERNAL_QUEUE */
#endif /* ALT_LOCKS_SUPPORTED */

    /* hardware specific structures */
#if defined(ALT_SYSV)
    IO_BASE_TYPE configBase;	/* base addr for PCI card's config space */
#else /* defined(ALT_SYSV) */
#define configBase base		/* for compatibility */
#endif /* defined(ALT_SYSV) */
    IO_BASE_TYPE base;		/* base addr for PCI card's shared mem */
    tg_shmem_t *genComm;	/* ptr to Gen Comm in shr mem */
    U32 flash_type;		/* set to appropriate flash type if present */
#if defined(ALT_SYSV)
    ddi_acc_handle_t configHandle; /* handle for pci config space accesses */
    ddi_acc_handle_t handle;	/* handle for pci bus accesses */
#if defined(__sun)
    struct ddi_dma_attr dmaAttr; /* DMA attributes */
#if FAST_DMA
    ddi_dma_lim_t dmaLims;	/* DMA limits (for dvma_xxxx xfers) */
#endif /* FAST_DMA */
#endif /* defined(__sun) */
#endif /* defined(ALT_SYSV) */

#if defined(WINNT)
#if USE_INTERNAL_QUEUE
    NDIS_PACKET *pktQueue;	/* internal queue for deserialized driver */
    NDIS_PACKET *pktTail;	/* end of queue for adding new guys */
#endif /* USE_INTERNAL_QUEUE */
    U32 ruptLevel;		/* interrupt level of board */
    NDIS_PHYSICAL_ADDRESS unmappedBase; /* base reg from resource list */
    U32 memLength;		/* len for card shared memory (best be 16k) */
    NDIS_MINIPORT_INTERRUPT ruptStruct;	/* opaque rupt structure */
    NDIS_MINIPORT_TIMER rslTimer; /* opaque timer structure */
    NDIS_MINIPORT_TIMER linkTimer; /* opaque timer structure */
    U32         linkTimeout;       /* link negotiation timeout */
    NDIS_HANDLE handle;	/* handle for pci bus accesses */
    NDIS_HANDLE configHandle;	/* handle for config time pci bus accesses */
    NDIS_HANDLE rcv_packet_pool_handle;	/* packet pool for recv bufs */
    NDIS_HANDLE recv_flush_buffer_pool_handle; /* handle for recv buf pool */
    NDIS_HANDLE jumbo_rcv_packet_pool_handle; /* pkt pool for jumbo bufs */
    NDIS_HANDLE recv_flush_jumbo_buffer_pool_handle; /* handle for jumbo pool */
#if USE_MINI_RING
    NDIS_HANDLE mini_rcv_packet_pool_handle; /* pkt pool for mini bufs */
    NDIS_HANDLE recv_flush_mini_buffer_pool_handle; /* handle for mini pool */
#endif /* USE_MINI_RING */
    NDIS_HANDLE send_flush_buffer_pool_handle;
    NDIS_HANDLE send_flush_fly_buffer_pool_handle;

    /* pkt array for returning multiple recv frames */
    RECV_PKT *tmpPacketArr[RECV_RETURN_RING_ENTRIES]; 
   
    U32 ruptTimeout;
    U32 outEarly;
    U32 slot;
    U8 latTimer;
#endif /* defined(WINNT) */
    U32 minDma;

#if defined(__NETWARE_386__) || defined(WINNT)
    U32 send_bd_coal;           /* holder for configured send bd coal count */
    U32 send_bd_ticks;          /* holder for configured send bd ticks count */
    U32 recv_bd_coal;           /* holder for configured recv bd coal count */
    U32 recv_bd_ticks;		/* holder for configured recv bd ticks cnt */
    U32 stat_ticks;             /* holder for configured stat ticks count */
    U32 nic_tracing;            /* holder for configured nic tracing value */
    U32 pci_read_max;           /* holder for configured PCI read max value */
#endif /* defined(__NETWARE_386__)  || defined(WINNT) */

#if defined(__NETWARE_386__)
    void *handle;
#define configHandle handle
    char string1[132];		/* print string from rupt handler */
    char string2[132];		/* print string from rupt handler */
#endif /* defined(__NETWARE_386__) */

#if defined(__sun)
    /* special TTP page for SBus diagnostics */
    caddr_t ttpAlloc;		/* pointer to where ttpPage is allocated */
    caddr_t ttpPage;		/* pointer to ttpPage (aligned) */
    struct alt_dma_info ttpPageDmaInfo;
#endif /* defined(__sun) */

    /* statistics sort of stuff up front so we can find it */
    U32 inpkts;			/* # pkts in */
    U32 outpkts;		/* # pkts out */
    U32 outDrops;		/* # pkts dropped on way out */
    U32 outDropsDmaResources;	/* # pkts dropped due to no DMA resources */
    U32 outDropsNoSpace;	/* # pkts dropped due to no space in ring */
    int flowControls;		/* number of times enter flow control */
#if defined(WINNT)
    U32 reapedFromSend;		/* # of times we were full and reaped */
#endif /* defined(WINNT) */
    U32 xmit_acked;		/* # pkts acked by NIC */
    U32 cmdFull;		/* # times we waited on the command ring */
    S32 bufsInStdRecvRing;	/* # bufs in standard recv ring */
    S32 bufsInJumboRecvRing;	/* # bufs in jumbo recv ring */
#if USE_MINI_RING
    S32 bufsInMiniRecvRing;	/* # bufs in mini recv ring */
#endif /* USE_MINI_RING */
    S32 bufsOnSendRing;		/* # bufs on the send ring */
    S32 bufsLowWater;		/* low water mark of bufsInRecvRing */
    U32 dataRtnPkts;		/* # of packets returned via RecvDataRtn */
#if USE_CKSUM_OFFLOAD
    U32 badIpCksums;		/* received bad ip checksum (calc by fw) */
    U32 badTcpCksums;		/* received bad tcp checksum (calc by fw) */
    U32 badUdpCksums;		/* received bad udp checksum (calc by fw) */
#endif /* USE_CKSUM_OFFLOAD */

#if defined(__sun)
    int overFlowRecvBufs;	/* number of outstanding overflow bufs */
    int totOverFlowRecvBufs;	/* total number of overflow bufs */
    int spillBufs;		/* number of current bufs in the spill q */
    int totSpillBufs;		/* total number of bufs put in spill q */
    int numFreeOverFlowBufs;	/* keep some overflow bufs around for reuse */
#endif /* defined(__sun) */

#if defined(WINNT)
    U32 pendPkts;		/* # of pending from NdisMIndicateReceive.. */
    U32 requeuedPkts;		/* # of success from NdisMIndicateReceive.. */
    U32 lowResources;		/* # of times we handed up STATUS_RESOURCES */
    U32 normRecv;		/* # of times we handed up STATUS_SUCCESS */
#endif /* defined(WINNT) */

#if !ALT_RECV_EXT
    U32 noRecvResources;	/* count of on fly replacing recv ring bufs */
#endif /* !ALT_RECV_EXT */

    U32 recvStdCoal;		/* receive buf reinsertion coallescing ctr */
    U32 recvJumboCoal;		/* recv jumbo buf reinsertion coallesce ctr */
#if USE_MINI_RING
    U32 recvMiniCoal;		/* recv mini buf reinsertion coallesce ctr */
#endif /* USE_MINI_RING */

    struct tg_gen_info *genInfo; /* ptr to general info page */
    caddr_t genInfoAlloc;	/* ptr to where data is allocated */
    struct alt_dma_info genInfoDmaInfo;

    struct tg_stats *genInfoStats2; /* ptr to 2nd stats area for refresh */
    struct alt_dma_info genInfoStats2DmaInfo;

    /* for keeping track of oversize frames so we can bleat on the console */
    U32 lastTooLongs;
    U32 lastCount;
#define MAX_LAST_COUNTS 30	/* nominally 3 seconds */

    /* this is for all the ring pointers, the eventProd, the sendCons, and the
     * recvCons */
    caddr_t ringPointersAlloc;
    struct alt_dma_info ringPointersDmaInfo;

    U16 eventCons;		/* shadow consumer for speed */
    U32 *eventProdPtr;		/* pointer to event producer */

    struct tg_event (*eventArr)[EVENT_RING_ENTRIES]; /* ptr to event ring */
    caddr_t eventArrAlloc;	/* ptr to where data is allocated */
    struct alt_dma_info eventArrDmaInfo;

    struct tg_command (*cmdArr)[COMMAND_RING_ENTRIES]; /* ptr to cmd ring */
    U32 commandProd;		/* shadow producer for speed */
    U32 commandCons;		/* shadow consumer for speed */
#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE cmdMutex;		/* command mutex */
#endif /* ALT_LOCKS_SUPPORTED */

    struct send_bd (*sendRing)[SEND_RING_ENTRIES]; /* ptr to send ring */
#if HOST_SEND_RING
    struct alt_dma_info sendRingDmaInfo;
    caddr_t sendRingAlloc;
#endif /* HOST_SEND_RING */
    struct alt_dma_info smsgs[SEND_RING_ENTRIES]; /* keep track of send MSGs */
#if ALT_DESERIALIZE
    struct alt_dma_info *smsgPtrs[SEND_RING_ENTRIES]; /* indirect to infops */
#if defined(WINNT)
    SLIST_HEADER freeSmsgs;
#else /* defined(WINNT) */
    struct alt_dma_info *freeSmsgs; /* q of available infops */
#endif /* defined(WINNT) */
    U32 smsgsFree;		/* counter */
#endif /* ALT_DESERIALIZE */
#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE sendMutex;	/* send ring mutex */
#endif /* ALT_LOCKS_SUPPORTED */
    U16 send_producer;		/* send producer index */
    U16 send_consumer;		/* local value so we know where we've been */
    U32 *sendConsPtr;		/* pointer to send consumer */
    U32 sendBufCount;		/* number of bufs actually allocated */
    caddr_t sendData;		/* pointer to the send buffer data addr */
#if USE_DMA_MEMALLOC
    struct alt_dma_info sendDataDmaInfo; /* infop for sendData alloc */
#endif /* USE_DMA_MEMALLOC */
    U32 hybridLen;		/* cross over point for copy vs. map */
#if defined(__sun) && FAST_DMA
    ddi_dma_handle_t alt_dvmaxh; /* dvma xmit handle */
    bool dvmatxAlloc;		/* whether they got allocated, for clean up */
#endif /* defined(__sun) && FAST_DMA */

    /* ptr to recv ring */
    struct recv_bd (*recvStdRing)[RECV_STANDARD_RING_ENTRIES]; 
    struct alt_dma_info recvStdRingDmaInfo;
    caddr_t recvStdRingAlloc;

    /* ptr to recv ring */
    struct JUMBO_RECV_BD (*recvJumboRing)[RECV_JUMBO_RING_ENTRIES]; 
    struct alt_dma_info recvJumboRingDmaInfo;
    caddr_t recvJumboRingAlloc;

#if USE_MINI_RING
    /* ptr to recv ring */
    struct recv_bd (*recvMiniRing)[RECV_MINI_RING_ENTRIES]; 
    struct alt_dma_info recvMiniRingDmaInfo;
    caddr_t recvMiniRingAlloc;
#endif /* USE_MINI_RING */

    /* ptr to recv ring */
    struct recv_bd (*recvReturnRing)[RECV_RETURN_RING_ENTRIES]; 
    struct alt_dma_info recvReturnRingDmaInfo;
    caddr_t recvReturnRingAlloc;

    /* track recv MSGs */
    struct alt_dma_info rstdmsgs[RECV_STANDARD_RING_ENTRIES]; 
    struct alt_dma_info rjumbomsgs[RECV_JUMBO_RING_ENTRIES];
#if USE_MINI_RING
    struct alt_dma_info rminimsgs[RECV_MINI_RING_ENTRIES];
#endif /* USE_MINI_RING */
#if ALT_RECV_EXT
    struct alt_dma_info *rstdmsgPtrs[RECV_STANDARD_RING_ENTRIES];
    struct alt_dma_info *rjumbomsgPtrs[RECV_JUMBO_RING_ENTRIES];
#if USE_MINI_RING
    struct alt_dma_info *rminimsgPtrs[RECV_MINI_RING_ENTRIES];
#endif /* USE_MINI_RING */
#if defined(__sun)
    RECV_PKT *curRecvPkt;	/* current pkt being built by receiver */
#define SPILL_HIGH_WATER ((ALT_RECV_LOW_HIGH - ALT_RECV_LOW_LOW) * 2)
    /* use the spill queue (LIFO) for when the receive ring has overflow
     * buffers in it, but we're receiving standard buffers back from the
     * stack and there's not room to put them on the ring */
    struct alt_dma_info *recvRingSpill;	
    struct alt_dma_info *freeOverFlowBufs;
#endif /* defined(__sun) */
#endif /* ALT_RECV_EXT */

#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE recvMutex;	/* general recv mutex */
    LOCK_TYPE recvRingMutex;	/* recv ring mutex for enqueing messages */
    LOCK_TYPE recvBufsMutex;	/* mutex for protecting bufsInRecvRing cntr */
#endif /* ALT_LOCKS_SUPPORTED */
    U16 recv_std_producer;	/* standard recv producer indices */
    U16 recv_std_consumer;	/* standard recv consumer indices */
    U16 recv_jumbo_producer;	/* jumbo recv producer indices */
    U16 recv_jumbo_consumer;	/* jubmo recv consumer indices */
#if USE_MINI_RING
    U16 recv_mini_producer;	/* mini recv producer indices */
    U16 recv_mini_consumer;	/* mini recv consumer indices */
#endif /* USE_MINI_RING */
    U16 recv_return_producer;	/* return recv producer indices */
    U16 recv_return_consumer;	/* return recv consumer indices */
    U32 *recvReturnProdPtr;	/* pointer to return recv producer */
    U32 recvStdBufCount;	/* number of std bufs actually allocated */
    U32 recvJumboBufCount;	/* number of jumbo bufs actually allocated */
#if USE_MINI_RING
    U32 recvMiniBufCount;	/* number of mini bufs actually allocated */
#endif /* USE_MINI_RING */
#if ALT_RECV_EXT
    caddr_t recvStdData;	/* place for external data malloc */
    caddr_t recvJumboData;	/* place for external data malloc (jumbo) */
#if USE_MINI_RING
    caddr_t recvMiniData;	/* place for external data malloc (mini) */
#endif /* USE_MINI_RING */
#if USE_DMA_MEMALLOC
    struct alt_dma_info recvStdDataDmaInfo; /* infop for recvStdData alloc */
    struct alt_dma_info recvJumboDataDmaInfo; /* for recvJumboData alloc */
#if USE_MINI_RING
    struct alt_dma_info recvMiniDataDmaInfo; /* for recvMiniData alloc */
#endif /* USE_MINI_RING */
#endif /* USE_DMA_MEMALLOC */
#endif /* ALT_RECV_EXT */

#if defined(__sun) && FAST_DMA
    ddi_dma_handle_t alt_dvmarh; /* dvma recv handle */
    bool dvmarxAlloc;		/* whether they got allocated, for clean up */
    ddi_dma_handle_t alt_jumbo_dvmarh; /* dvma jumbo recv handle */
    bool dvmaJumborxAlloc;	/* whether they got allocated, for clean up */
#if USE_MINI_RING
    ddi_dma_handle_t alt_mini_dvmarh; /* dvma mini recv handle */
    bool dvmaMinirxAlloc;	/* whether they got allocated, for clean up */
#endif /* USE_MINI_RING */
#endif /* defined(__sun) && FAST_DMA */

    struct bpduMsg curBPDU;	/* current BPDU info for redundancy */
    U16 bpduAge;		/* cur age of bpdus, for switchover delay */
    U16 stpReadyAge;		/* age when we're ok to switch */

#if defined(__sun)
    /* keep track of which nic type we're dealing with */
    int nic_type;

    /* for ndd access and initialization */
    struct alt_param tuneParms[NUM_TUNE_PARMS];	
    caddr_t alt_nd_ptr;		/* ptr to named dispatch table */

    /* for mapping the PCI bus across the Sbus bridge  */
    struct alt_dma_info textDmaInfo; 
#endif /* defined(__sun) */

#if defined(WINNT)
    struct alt_nt_tune_parms tuneParms;
#endif /* defined(WINNT) */

#if RECV_CKSUM_DEBUG
    U32 recvIPCksumFail0;
    U32 recvIPCksumFail1;
    U32 recvIPCksumFail2;
    U32 recvIPCksumFail3;
    U32 recvIPCksumFail4;
    U32 recvTCPCksumFail0;
    U32 recvTCPCksumFail1;
    U32 recvTCPCksumFail2;
    U32 recvUDPCksumFail0;
    U32 recvUDPCksumFail1;
    U32 recvUDPCksumFail2;
    U32 recvUDPCksumFail3;
    U32 recvUDPCksumFail4;
    U32 recvCksumBail0;
    U32 recvCksumBail1;
    U32 fragip_id;
    U32 fragip_offset;
    U32 frag_pseudo;
    U32 fragudp_cksum;
#if defined(PASS_IP_ID)
    U32 BadEnd; 
    U32 BadIpID;
    U32 BadSplit;
    U32 DupEnd; 
    U32 ErrCnt; 
    U32 SplitErr; 
#endif /* defined(PASS_IP_ID) */
    U16 inProgressCksum;
    U16 LastIpID;
    bool jumboSplit;
    bool splitTCP;
    bool splitFrag;
    bool EndFrag;
#endif /* RECV_CKSUM_DEBUG */
};

/* flags bits */
#define ALT_PARTIAL_SEND_Q  0x00000002	/* when prev send not fully qd */
#define ALT_LINK_READY      0x00000004	/* set when link active on board */
#define ALT_MULTI_ACTIVE    0x00000008	/* multicast update in progress */
#define ALT_RED_ACTIVE      0x00000010 /* active redundant link */
#define ALT_RED_STANDBY     0x00000020 /* standby redundant link */
#define ALT_DONT_PASS_UP    0x00000080 /* test facility don't send pkts up stack */

/*
 * defines for the flags field...
 */
#if defined(__sun)
#define ALT_SBUS_MAPPED     0x00000100 /* whether DMA prepped for Sbus */
#define ALT_SUSPENDED       0x00000200 /* Sun's Suspend/resume stuff */
#endif /* defined(__sun) */

#define ALT_JUMBO_ENABLED   0x00000400 /* jumbograms enabled (or not) */
#define ALT_RING_FILLED     0x00000800 /* did FillRecvRing() */
#define ALT_JUMBO_FILLED    0x00001000 /* did FillJumboRecvRing() */

#if defined(__NETWARE_386__) || defined(WINNT)
#define ALT_IN_DRV_DISABLE  0x00002000 /* DriverDisableInterrupt called */
#endif /* defined(__NETWARE_386__) || defined(WINNT) */

#define ALT_OVERRIDE_MAC_ADDR 0x00004000 /* override card's MAC address */
#define ALT_DO_LINK_NEGOTIATION 0x00008000 /* do link negotiation if set */
#define ALT_DO_FDR_FILTERING 0x00010000 /* do FDR Filering if set (tigon2) */
#define ALT_PROM_MODE       0x00020000 /* keep track for reset */

#if defined(__NETWARE_386__) || defined(WINNT)
#define ALT_RESET_STATS     0x00040000 /* on reset, redo stats */
#endif /* defined(__NETWARE_386__) || defined(WINNT) */

#define ALT_MULTICAST_MODE  0x00080000 /* keep track for reset */
#define ALT_RX_FLOW_CONTROL 0x00100000 /* turn on rx flow ctrl for link neg  */
#define ALT_TX_FLOW_CONTROL 0x00200000 /* turn on tx flow ctrl for link neg  */

#if defined(__NETWARE_386__)
#define ALT_STR1_PRESENT    0x00400000 /* rupt print string 1 */
#define ALT_STR2_PRESENT    0x00800000 /* rupt print string 2 */
#endif /* defined(__NETWARE_386__) */

#define ALT_USE_VLAN_ASSIST 0x01000000 /* set vlan assist in rcbs */
#define ALT_TIMER_RUNNING   0x02000000 /* RSL timer is running */
#if USE_MINI_RING
#define ALT_MINI_FILLED     0x04000000 /* did FillMiniRecvRing() */
#endif /* USE_MINI_RING */
#define ALT_SNAP            0x08000000 /* use SNAP header */
#define ALT_RESERVED_3      0x10000000 /* open */
#define ALT_RESERVED_4      0x20000000 /* open */
#define ALT_JUMBO_INITED    0x40000000 /* indicate if jumbo mem alloced */
#define ALT_RESERVED_5      0x80000000 /* open */

/*
 * defines for the init_flags field...
 */
#define ALT_INIT_REAP_AFTER_SEND 0x01000000 /* enable reap after send */
#define ALT_INIT_SEND_UPDATE_ONLY 0x02000000 /* update send cons w/o interrupting */
#define ALT_INIT_MEM_WR_INV_DISABLE 0x04000000 /* disable mem write inval */
#define ALT_INIT_DEBUG_PCI       0x08000000 /* print PCI config space at boot */
#define ALT_INIT_USE_32_BIT_MODE 0x10000000 /* force 32 bit mode */
#define ALT_INIT_450GX_FIX       0x20000000 /* always use all byte enables */
#define ALT_INIT_RESERVED_20     0x40000000 /* open */
#define ALT_INIT_SHARED_RUPT     0x80000000 /* do we share the rupt */

#if USE_CKSUM_OFFLOAD
/* flag bits for checksum offload */
#define ALT_CKSUM_DO_XMIT_IP     1     /* do transmit IP checksum offload */
#define ALT_CKSUM_DO_XMIT_TCP    2     /* do transmit TCP checksum offload */
#define ALT_CKSUM_DO_XMIT_UDP    4     /* do transmit UDP checksum offload */
#define ALT_CKSUM_DO_RECV_IP     8     /* do receive IP checksum offload */
#define ALT_CKSUM_DO_RECV_TCP    0x10  /* do receive TCP checksum offload */
#define ALT_CKSUM_DO_RECV_UDP    0x20  /* do receive UDP checksum offload */
#define ALT_CKSUM_XMIT_NO_PSEUDO 0x40  /* don't include pseudo hdr on xmit */
#define ALT_CKSUM_RECV_NO_PSEUDO 0x80  /* don't include pseudo hdr on recv */
#endif /* USE_CKSUM_OFFLOAD */

/* flash types */
#define FLASH_AMD     1		/* AMD Flash */
#define FLASH_INTEL   2		/* Intel Flash */
#define FLASH_UNKNOWN -1	/* who knows */

/* nic types */
#define ALT_SBUS_NIC        1
#define ALT_PCI_NIC         2

#define ALT_RECV_COAL       20	/* coallesce this many recv bds */
#define ALT_JUMBO_RECV_COAL 10	/* coallesce this many jumbo recv bds */
#if USE_MINI_RING
#define ALT_MINI_RECV_COAL 40	/* coallesce this many mini recv bds */
#endif /* USE_MINI_RING */

/* for Tigon 1 and 2 to do statistics we must tell it the packet type */
#define SEND_STAT_PREP(infop, flags)				\
{								\
    if (*infop->dataAddr & 0x01) { /* group bit */		\
	if (ether_cmp((struct ether_addr *)infop->dataAddr,	\
		      &broadcastAddr) == 0)			\
	    flags |= BD_FLAG_BCAST_PKT;				\
	else							\
	    flags |= BD_FLAG_MCAST_PKT;				\
    } else							\
	flags |= BD_FLAG_UCAST_PKT;				\
}

#define MAX_TRIES     100	/* tries for cmd full while loop */ 

/* macros for enqueing and dequeing rings */
#define ENQ_COMMAND_RING(cmd, prod, shadProd, cons, shadCons, ring, spl, mup) \
{									      \
    U32 next;								      \
    U32 myCons, myProd;							      \
									      \
    ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mup); /* ENTER CRITICAL SECTION */     \
    myCons = shadCons;		/* after we're in critical setion */	      \
    myProd = shadProd;							      \
    /* is there space? */						      \
    next = INCRING(myProd, COMMAND_RING_ENTRIES);			      \
    if (next == myCons)							      \
	/* only touch hardware when necessary */			      \
	shadCons = myCons = RDSHMEM(cons);				      \
    /* command ring always ignores overflow, will "never" happen */	      \
    if (next == myCons)  {    /* ring full, never catch up */		      \
	int count = 0;							      \
									      \
	nic->cmdFull++;							      \
	ALTTRACEE(cmdfull, 0xf0015, RDSHMEM(prod), RDSHMEM(cons), shadProd,   \
                                                      shadCons);	      \
	/* but if it does give the nic a little bit of time to catch up */    \
        while (next == myCons)  {					      \
	    SNOOZE(1);							      \
	    shadCons = myCons = RDSHMEM(cons);				      \
	    if (++count == MAX_TRIES) {					      \
		(void) stopFW(nic->altp);				      \
		alt_trace_snap();					      \
		break;							      \
	    }								      \
	}								      \
	if (count != MAX_TRIES) {					      \
            ALT_MEMCPY_TODEV((MEMCPY_TYPE)&(*ring)[myProd],		      \
			     (MEMCPY_TYPE)&(cmd), (MEMCPY_SZT)sizeof(cmd));   \
									      \
	    /* increment producer, assuming that space is available */	      \
	    WRSHMEM(prod, next);					      \
	    MB();							      \
	    shadProd = next;						      \
	}								      \
    } else {								      \
        /* copy element into ring, no cache worries, shmem */		      \
        ALT_MEMCPY_TODEV((MEMCPY_TYPE)&(*ring)[myProd],			      \
	    (MEMCPY_TYPE)&cmd, (MEMCPY_SZT)sizeof(cmd));		      \
									      \
        /* increment producer, assuming that space is available */	      \
        WRSHMEM(prod, next);						      \
	MB();								      \
	shadProd = next;						      \
    }									      \
    ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup); /* LEAVE CRITICAL SECTION */      \
}

#define ENQ_COMMAND_RING_NO_PROD(cmd, prod, shadProd, cons, shadCons,	    \
				 ring, spl, mup)			    \
{									    \
    U32 next;								    \
    U32 myCons, myProd;							    \
									    \
    ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mup); /* ENTER CRITICAL SECTION */   \
    myCons = shadCons;		/* after we're in critical setion */	    \
    myProd = shadProd;							    \
    /* is there space? */						    \
    next = INCRING(myProd, COMMAND_RING_ENTRIES);			    \
    if (next == myCons)							    \
	/* only touch hardware when necessary */			    \
	shadCons = myCons = RDSHMEM(cons);				    \
    /* command ring always ignores overflow, will "never" happen */	    \
    if (next == myCons)  {    /* ring full, never catch up */		    \
	int count = 0;							    \
									    \
	nic->cmdFull++;							    \
	ALTTRACEE(cmdfull, 0xf0015, RDSHMEM(prod), RDSHMEM(cons), shadProd, \
                                                      shadCons);	    \
	/* but if it does give the nic a little bit of time to catch up */  \
        while (next == myCons)  {					    \
	    SNOOZE(1);							    \
	    shadCons = myCons = RDSHMEM(cons);				    \
	    if (++count == MAX_TRIES) {					    \
		(void) stopFW(nic->altp);				    \
		alt_trace_snap();					    \
		break;							    \
	    }								    \
	}								    \
	if (count != MAX_TRIES) {					    \
            ALT_MEMCPY_TODEV((MEMCPY_TYPE)&(*ring)[myProd],		    \
			     (MEMCPY_TYPE)&(cmd), (MEMCPY_SZT)sizeof(cmd)); \
									    \
	    /* increment producer, assuming that space is available */	    \
	    shadProd = next;						    \
	}								    \
    } else {								    \
        /* copy element into ring, no cache worries, shmem */		    \
        ALT_MEMCPY_TODEV((MEMCPY_TYPE)&(*ring)[myProd],			    \
	    (MEMCPY_TYPE)&(cmd), (MEMCPY_SZT)sizeof(cmd));		    \
									    \
        /* increment producer, assuming that space is available */	    \
	shadProd = next;						    \
    }									    \
    ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup); /* LEAVE CRITICAL SECTION */    \
}

#if defined(WINNT) && !HOST_SEND_RING
/* NT uses an extended transmit memory window */
#define SET_NIC_MEM_WINDOW(next, nic)			\
{							\
    if (((next) & SEND_RING_BDS_IN_WINDOW) == 0) {	\
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE,	\
	      SEND_RING_PTR + TG_MEM_WINDOW_SIZE *	\
	      ((((next) >> TG_MEM_WINDOW_SHIFT) & 3)));	\
    }							\
}

#define SET_NIC_MEM_WINDOW_NOW(next, nic)		\
{							\
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE,	\
	  SEND_RING_PTR + TG_MEM_WINDOW_SIZE *		\
	  ((((next) >> TG_MEM_WINDOW_SHIFT) & 3)));	\
}
#else /* defined(WINNT) && !HOST_SEND_RING */
/* not used by ODI or Solaris or host based send ring */
#define SET_NIC_MEM_WINDOW(next, nic)
#define SET_NIC_MEM_WINDOW_NOW(next, nic)
#endif /* defined(WINNT) && !HOST_SEND_RING */

#if HOST_SEND_RING
#define FILL_IN_SEND_BD(ring, prod, bufdp) \
    (*ring)[prod] = *bufdp;    /* struct copy */

#else /* HOST_SEND_RING */
#if PTRS_ARE_64
#define FILL_IN_SEND_BD(ring, prod, bufdp)				 \
{									 \
    struct send_bd *windowBd = &(*ring)[prod & SEND_RING_BDS_IN_WINDOW]; \
									 \
    /* 64 bits, must do whole host_addr */				 \
    WRSHMEM(windowBd->host_addr, bufdp->host_addr);			 \
    WRSHMEM(windowBd->w0.w0, bufdp->w0.w0);				 \
    MB();								 \
}

#else /* PTRS_ARE_64 */
#define FILL_IN_SEND_BD(ring, prod, bufdp)				 \
{									 \
    struct send_bd *windowBd = &(*ring)[prod & SEND_RING_BDS_IN_WINDOW]; \
									 \
    /* assume type half is 0 */						 \
    WRSHMEM(windowBd->host_addr.addr_lo, bufdp->host_addr.addr_lo);	 \
    WRSHMEM(windowBd->w0.w0, bufdp->w0.w0);				 \
    if (bufdp->sendbd_flags & BD_FLAG_VLAN_TAG) \
        WRSHMEM(windowBd->w1, bufdp->w1); \
}
#endif /* defined(WINNT) */
#endif /* HOST_SEND_RING */

#if defined(__sun) || (defined(WINNT) && NT_USE_DEBUGGER)
#define ENQ_SEND_RING(bufdp, prod, cons, ring, spl, rc)			\
{									\
    U16 next;								\
									\
    /* is there space? */						\
    next = INCRING(prod, SEND_RING_ENTRIES);				\
    if (next == cons) { /* ring full, never catch up */			\
        prod = next;  /* update so we back out correctly */             \
	rc = FALSE;							\
    } else {								\
									\
        ALT_DISABLE_AT_DISPATCH_LEVEL(spl, &nic->memWindowMu);		\
        /* copy element into ring */					\
	FILL_IN_SEND_BD(ring, prod, bufdp);				\
        ALT_ENABLE_AT_DISPATCH_LEVEL(spl, &nic->memWindowMu);		\
									\
        /* increment producer, assuming that space is available */	\
        SET_NIC_MEM_WINDOW(next, nic);					\
        prod = next;							\
        nic->bufsOnSendRing++;                                          \
									\
        rc = TRUE;							\
    }									\
}
#define ENQ_SEND_RING_NO_MUTEX(bufdp, prod, cons, ring, spl, rc)	\
{									\
    U16 next;								\
									\
    /* is there space? */						\
    next = INCRING(prod, SEND_RING_ENTRIES);				\
    if (next == cons) { /* ring full, never catch up */			\
        prod = next;  /* update so we back out correctly */             \
	rc = FALSE;							\
    } else {								\
									\
        /* copy element into ring */					\
	FILL_IN_SEND_BD(ring, prod, bufdp);				\
									\
        /* increment producer, assuming that space is available */	\
        SET_NIC_MEM_WINDOW(next, nic);					\
        prod = next;							\
        nic->bufsOnSendRing++;                                          \
									\
        rc = TRUE;							\
    }									\
}
#else /* defined(__sun) || (defined(WINNT) && NT_USE_DEBUGGER) */
#define ENQ_SEND_RING(bufdp, prod, cons, ring, spl, rc)			\
{									\
    U16 next;								\
									\
    /* is there space? */						\
    next = INCRING(prod, SEND_RING_ENTRIES);				\
    if (next == cons) { /* ring full, never catch up */			\
        prod = next;  /* update so we back out correctly */             \
	rc = FALSE;							\
    } else {								\
									\
        /* copy element into ring */					\
	FILL_IN_SEND_BD(ring, prod, bufdp);				\
									\
        /* increment producer, assuming that space is available */	\
        SET_NIC_MEM_WINDOW(next, nic);					\
        prod = next;							\
        nic->bufsOnSendRing++;                                          \
									\
        rc = TRUE;							\
    }									\
}
#define ENQ_SEND_RING_NO_MUTEX ENQ_SEND_RING
#endif /* defined(__sun) || (defined(WINNT) && NT_USE_DEBUGGER) */

#if ALT_RECV_EXT
#define ENQ_RECV_RING(bufd, prod, cons, ring, ringSize, handle, spl, mup, rc) \
{									      \
    U16 next;								      \
									      \
    ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mup); /* ENTER CRITICAL SECTION */     \
    /* is there space? */						      \
    next = INCRING(prod, ringSize);					      \
    if (next == cons) { /* ring full, never catch up */			      \
	ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup); /* LEAVE CRITICAL SECTION */  \
	rc = FALSE;							      \
    } else {								      \
									      \
        /* copy element into ring */					      \
        (*ring)[prod] = bufd;    	/* STRUCT COPY */		      \
									      \
        MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct recv_bd),	      \
			sizeof(struct recv_bd));			      \
									      \
        /* increment producer, assuming that space is available */	      \
        prod = next;							      \
									      \
        ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup);	/* LEAVE CRITICAL SECTION */  \
									      \
        rc = TRUE;							      \
    }									      \
}

#define ENQ_RECV_RING_NO_MUTEX(bufd, prod, cons, ring, ringSize, handle, rc) \
{									     \
    U16 next;								     \
									     \
    /* is there space? */						     \
    next = INCRING(prod, ringSize);					     \
    if (next == cons) { /* ring full, never catch up */			     \
	rc = FALSE;							     \
    } else {								     \
									     \
        /* copy element into ring */					     \
        (*ring)[prod] = bufd;    	/* STRUCT COPY */		     \
									     \
        MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct recv_bd),	     \
			sizeof(struct recv_bd));			     \
									     \
        /* increment producer, assuming that space is available */	     \
        prod = next;							     \
									     \
        rc = TRUE;							     \
    }									     \
}

#define ENQ_JUMBO_RECV_RING(bufd, prod, cons, ring, ringSize, handle, 	     \
			    spl, mup, rc)				     \
{									     \
    U16 next;								     \
									     \
    ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mup); /* ENTER CRITICAL SECTION */    \
    /* is there space? */						     \
    next = INCRING(prod, ringSize);					     \
    if (next == cons) { /* ring full, never catch up */			     \
	ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup); /* LEAVE CRITICAL SECTION */ \
	rc = FALSE;							     \
    } else {								     \
									     \
        /* copy element into ring */					     \
        (*ring)[prod] = bufd;    	/* STRUCT COPY */		     \
									     \
        MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct JUMBO_RECV_BD),   \
			sizeof(struct JUMBO_RECV_BD));			     \
									     \
        /* increment producer, assuming that space is available */	     \
        prod = next;							     \
									     \
        ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mup);	/* LEAVE CRITICAL SECTION */ \
									     \
        rc = TRUE;							     \
    }									     \
}

#define ENQ_JUMBO_RECV_RING_NO_MUTEX(bufd, prod, cons, ring, 		   \
				     ringSize, handle, rc)		   \
{									   \
    U16 next;								   \
									   \
    /* is there space? */						   \
    next = INCRING(prod, ringSize);					   \
    if (next == cons) { /* ring full, never catch up */			   \
	rc = FALSE;							   \
    } else {								   \
									   \
        /* copy element into ring */					   \
        (*ring)[prod] = bufd;    	/* STRUCT COPY */		   \
									   \
        MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct JUMBO_RECV_BD), \
			sizeof(struct JUMBO_RECV_BD));			   \
									   \
        /* increment producer, assuming that space is available */	   \
        prod = next;							   \
									   \
        rc = TRUE;							   \
    }									   \
}
#else /* ALT_RECV_EXT */
#define ENQ_RECV_RING(bufd, prod, cons, ring, ringSize, handle, spl, mup, rc) \
{									      \
    U16 next;								      \
									      \
    /* is there space? */						      \
    next = INCRING(prod, ringSize);					      \
    if (next == cons)  /* ring full, never catch up */			      \
	rc = FALSE;							      \
									      \
    /* copy element into ring */					      \
    (*ring)[prod] = bufd; 	/* STRUCT COPY */			      \
									      \
    MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct recv_bd),	      \
		    sizeof(struct recv_bd));				      \
									      \
    /* increment producer, assuming that space is available */		      \
    prod = next;							      \
									      \
    rc = TRUE;								      \
}
#define ENQ_JUMBO_RECV_RING(bufd, prod, cons, ring, ringSize, 		\
			    handle, spl, mup, rc)			\
{									\
    U16 next;								\
									\
    /* is there space? */						\
    next = INCRING(prod, ringSize);					\
    if (next == cons)  /* ring full, never catch up */			\
	rc = FALSE;							\
									\
    /* copy element into ring */					\
    (*ring)[prod] = bufd; 	/* STRUCT COPY */			\
									\
    MALLOC_CACHE_WB(handle, ring, prod * sizeof(struct JUMBO_RECV_BD),	\
		    sizeof(struct JUMBO_RECV_BD));			\
									\
    /* increment producer, assuming that space is available */		\
    prod = next;							\
									\
    rc = TRUE;								\
}
#endif /* ALT_RECV_EXT */

/* macro for placing commands on the ring */
#define DO_TG_COMMAND(cmd, index, code)				\
{								\
    tg_shmem_t *locGenComm = nic->genComm;			\
    command.TG_COMMAND_w0 =					\
	cmd << TG_COMMAND_CMD_SHIFT |				\
	code <<TG_COMMAND_CODE_SHIFT |				\
	index;							\
    ENQ_COMMAND_RING(command.TG_COMMAND_w0, 			\
		     locGenComm->command_prod_index,		\
		     nic->commandProd,				\
		     locGenComm->gen_com.command_cons_index,	\
		     nic->commandCons,				\
		     nic->cmdArr, spl, &nic->cmdMutex);		\
}

/* macro for placing commands on the ring */
#define DO_TG_EXT_COMMAND2(cmd, index, code, word1, word2)		\
{									\
    tg_shmem_t *locGenComm = nic->genComm;				\
									\
    command.TG_COMMAND_w0 =						\
	cmd << TG_COMMAND_CMD_SHIFT |					\
	code <<TG_COMMAND_CODE_SHIFT |					\
	index;								\
    ENQ_COMMAND_RING_NO_PROD(command.TG_COMMAND_w0, 			\
			     locGenComm->command_prod_index,		\
			     nic->commandProd,				\
			     locGenComm->gen_com.command_cons_index,	\
			     nic->commandCons,				\
			     nic->cmdArr, spl, &nic->cmdMutex);		\
    ENQ_COMMAND_RING_NO_PROD(word1, locGenComm->command_prod_index,	\
			     nic->commandProd,				\
			     locGenComm->gen_com.command_cons_index,	\
			     nic->commandCons,				\
			     nic->cmdArr, spl, &nic->cmdMutex);		\
    ENQ_COMMAND_RING(word2, locGenComm->command_prod_index,		\
		     nic->commandProd,					\
		     locGenComm->gen_com.command_cons_index,		\
		     nic->commandCons,					\
		     nic->cmdArr, spl, &nic->cmdMutex);			\
}

/* externs for the firmware */
extern U32 tigonFwStartAddr;
extern U32 tigonFwTextAddr;
extern int tigonFwTextLen;
extern U32 tigonFwDataAddr;
extern int tigonFwDataLen;
extern U32 tigonFwRodataAddr;
extern int tigonFwRodataLen;
extern U32 tigonFwBssAddr;
extern int tigonFwBssLen;
extern U32 tigonFwSbssAddr;
extern int tigonFwSbssLen;
extern U32 tigonFwText[]; 
extern U32 tigonFwData[];
extern U32 tigonFwRodata[];

extern U32 tigon2FwStartAddr;
extern U32 tigon2FwTextAddr;
extern int tigon2FwTextLen;
extern U32 tigon2FwDataAddr;
extern int tigon2FwDataLen;
extern U32 tigon2FwRodataAddr;
extern int tigon2FwRodataLen;
extern U32 tigon2FwBssAddr;
extern int tigon2FwBssLen;
extern U32 tigon2FwSbssAddr;
extern int tigon2FwSbssLen;
extern U32 tigon2FwText[]; 
extern U32 tigon2FwData[];
extern U32 tigon2FwRodata[];
#endif /* _ALTDRV_H */

