/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altioctl.h,v $
 * Revision 1.1.6.8  1997/12/03  01:23:00  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:21  ted]
 *
 * Revision 1.1.6.7  1997/04/12  00:11:27  ted
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:07:37  ted]
 * 
 * Revision 1.1.6.6  1997/04/11  19:34:08  ted
 * 	Let TTP run on non-sparc systems
 * 	[1997/04/11  19:33:33  ted]
 * 
 * Revision 1.1.6.5  1997/04/04  22:38:26  ted
 * 	ioctls that support the TTP page
 * 	[1997/04/04  22:15:36  ted]
 * 
 * Revision 1.1.6.4  1997/03/24  08:34:39  ted
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:31  ted]
 * 
 * Revision 1.1.6.3  1997/03/17  08:47:59  skapur
 * 	Add ALT_CLEAR_STATS and ALT_CLEAR_NIC_STATS ioctls.
 * 	[1997/03/17  08:37:14  skapur]
 * 
 * Revision 1.1.6.2  1997/03/02  18:36:54  ted
 * 	redundancy support and a few bug fixes
 * 	[1997/03/02  18:25:54  ted]
 * 
 * Revision 1.1.6.1  1997/02/22  00:20:53  ted
 * 	Make debugger a special stream that doesn't effect the HostState down call
 * 	[1997/02/22  00:12:15  ted]
 * 
 * Revision 1.1.2.1  1996/11/21  18:01:39  ted
 * 	Moved from ../unix/common and merged in WINNT changes
 * 	[1996/11/21  17:43:06  ted]
 * 
 * Revision 1.4.2.1  1996/11/01  18:25:36  ted
 * 	change rr_ to tg_ and add proper copyright
 * 	[1996/10/30  22:58:42  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altioctl.h
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/./nic/host/common/altioctl.h,v $
 * $Revision: 1.1.6.8 $ $Locker:  $
 * $Date: 1997/12/03 01:23:00 $ $Author: ted $ $State: Exp $
 */

#ifndef _ALTIOCTL_H_
#define _ALTIOCTL_H_

#if defined(ALT_STREAM)
#define MAXSTATSTR 320		/* number of streams we can report on */
                        /* 320 for 5 streams-per-instance times 64 for VLANS */
struct iocstrinfo {
    int numstr;			/* number of open streams */
    struct altstr altstr[MAXSTATSTR]; /* where to put the altstr structs */
};
#endif /* defined(ALT_STREAM) */
    
struct alt_stat {
    caddr_t altinst;		/* where to put the altinst struct */
    caddr_t nic;		/* where to put the nic struct */
#if defined(ALT_STREAM)
    struct iocstrinfo *altStrInfo; /* streams info goes here */
#endif /* defined(ALT_STREAM) */
};

/* ioctl fields */
/* system oriented ioctls */
/* global oriented */
#define ALT_SETTRACE    _IO('a', 1)	 /* toggle trace value */
#define ALT_READTRACE   _IO('a', 2)	 /* read trace value */
#define ALT_GETTRACE    _IO('a', 3)      /* read trace buffer */
#define ALT_GETSTRACE   _IO('a', 4)	 /* read snap trace buffer */
#define ALT_CLRSTRACE   _IO('a', 5)	 /* clear snap trace */

/* per card oriented (with stream for stream drivers) */
#define ALT_STAT        _IOWR('a', 6, struct alt_stat) /* get driver stats */

#if defined(ALT_STREAM)
/* attach to a particular device instance */
#define ALT_ATTACH      _IO('a', 7)
#endif /* defined(ALT_STREAM) */

/* TIGON based ioctls */
struct tg_reg {
    U32     data;
    U32     addr;		/* really an offset */
};

struct tg_mem {
    U32     tgAddr;
    caddr_t userAddr;
    int     len;
};

#define ALT_READ_TG_MEM     _IOWR('a', 10, struct tg_mem)
#define ALT_WRITE_TG_MEM    _IOWR('a', 11, struct tg_mem)
#define ALT_READ_TG_REG     _IOWR('a', 12, struct tg_reg)
#define ALT_WRITE_TG_REG    _IOWR('a', 13, struct tg_reg)

/* download of firmware */
struct tg_fw {
    U32 startAddr;
    U32 textAddr;
    int textLen;
    U32 dataAddr;
    int dataLen;
    U32 rodataAddr;
    int rodataLen;
    U32 bssAddr;
    int bssLen;
    U32 sbssAddr;
    int sbssLen;
    caddr_t fwText;
    caddr_t fwData;
    caddr_t fwRodata;
};

/* get info for coredump of NIC */
struct tg_core {
    U32     pc;			/* current pc, filled in by kernel */
    caddr_t gprs;		/* pointer to where to place GPRs */
    U32     textLen;		/* length of text area, filled in by kernel */
    U32     textAddr;		/* start of text area, filled in by kernel */
    caddr_t fwText;		/* pointer to where to place text data */
    U32     dataLen;		/* length of data area, filled in by kern */
    U32     dataAddr;		/* start of data area, filled in by kernel */
    caddr_t fwData;		/* pointer to where to place data area */
    U32     rodataLen;		/* length of rodata area, filled in by kern */
    U32     rodataAddr;		/* start of rodata area, filled in by kernel */
    caddr_t fwRodata;		/* pointer to where to place rodata */
    U32     bssLen;		/* length of bss area, filled in by kern */
    U32     bssAddr;		/* start of bss area, filled in by kernel */
    caddr_t fwBss;		/* pointer to where to place bss */
    U32     sbssLen;		/* length of sbss area, filled in by kern */
    U32     sbssAddr;		/* start of sbss area, filled in by kernel */
    caddr_t fwSbss;		/* pointer to where to place sbss */
    U32     stackLen;		/* length of the stack, filled in by kern */
    caddr_t fwStack;		/* pointer to where to place stack */
};

/* done globally */
#define ALT_CACHE_FIRMWARE  _IOWR('a', 14, struct tg_fw)

/* per card */
#define ALT_LOAD_FIRMWARE   _IO('a', 15)

/* cached file must match running firmware */
#define ALT_DUMP_FIRMWARE   _IOWR('a', 17, struct tg_core)
#define ALT_GET_NIC_STATS   _IO('a', 18) /* nic stats (struct tg_stats *) */
#define ALT_GETNTRACE       _IO('a', 19) /* read nic trace buffer */

/* tuning */
#define ALT_GET_TUNEPARMS   _IO('a', 20) /* (struct tg_tune *) */
#define ALT_SET_TUNEPARMS   _IO('a', 21) /* (struct tg_tune *) */
#define ALT_CLEAR_PROFILE   _IO('a', 22)

#if defined(ALT_STREAM)
/* I'm the debugger */
#define ALT_DEBUG           _IO('a', 23)
#endif /* defined(ALT_STREAM) */

#define ALT_SET_REDUND      _IO('a', 24)
#define ALT_REMOVE_REDUND   _IO('a', 25)

#define ALT_CLEAR_STATS     _IO('a', 26) /* clear general statistics */
#define ALT_CLEAR_NIC_STATS _IO('a', 27) /* clear NIC statistics */

/* testing */
#define ALT_SEND_PKT        _IO('a', 30)
#define ALT_RECV_PKT        _IO('a', 31)

/* special for ttp */
#if defined(__sun)
#define ALT_ALLOC_PAGE      _IO('a', 40) /* (caddr_t *) */
#define ALT_FREE_PAGE       _IO('a', 41)
#define ALT_READ_PAGE       _IO('a', 42) /* caddr_t */
#define ALT_WRITE_PAGE      _IO('a', 43) /* caddr_t */
#endif /* defined(__sun) */

#endif  /* _ALTIOCTL_H_ */


