/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */

/*
 * HISTORY
 * $Log: altrdwr.c,v $
 * Revision 1.1.2.31  1999/02/01  21:09:34  ted
 * 	Remove pieces of old endian stuff for unaligned writes that isn't necessary anymore
 * 	[1999/02/01  21:09:27  ted]
 *
 * Revision 1.1.2.30  1999/01/07  23:09:17  ted
 * 	reset window to current window not start of ring
 * 	[1999/01/07  23:03:06  ted]
 * 
 * Revision 1.1.2.29  1998/08/10  18:00:38  ted
 * 	return code fix in readTGSerEeprom
 * 	[1998/08/10  18:00:09  ted]
 * 
 * Revision 1.1.2.28  1998/08/09  02:05:19  hayes
 * 	fix serial eeprom write capability
 * 	[1998/08/09  02:05:06  hayes]
 * 
 * Revision 1.1.2.27  1998/06/16  16:45:49  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:43:01  ted]
 * 
 * Revision 1.1.2.26  1998/06/12  23:47:19  ted
 * 	snapshot of merge with new Registry handling
 * 	[1998/06/10  22:15:17  ted]
 * 
 * Revision 1.1.2.25  1998/06/08  23:55:33  ted
 * 	Cleanup tracing messages and identifiers
 * 	[1998/06/08  23:39:06  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:18  ted]
 * 
 * Revision 1.1.2.24  1998/04/14  21:42:39  ted
 * 	Merge ODI 1.8 fixes, add VLAN assist support
 * 	[1998/04/14  21:26:57  ted]
 * 
 * Revision 1.1.2.23  1998/03/09  21:36:58  venkat
 * 	introduced shared memory access MACROs to let individual platforms define
 * 	them to be system specific
 * 	[1998/03/09  21:24:59  venkat]
 * 
 * Revision 1.1.2.22  1998/03/04  18:58:47  ted
 * 	Fix typo
 * 	[1998/03/04  18:56:46  ted]
 * 
 * Revision 1.1.2.21  1998/01/24  04:06:50  venkat
 * 	OUT32 changed. Fix compiler error for WINNT
 * 	[1998/01/24  04:02:17  venkat]
 * 
 * Revision 1.1.2.20  1998/01/20  04:39:31  hayes
 * 	add tigon Rev6 support
 * 	[1998/01/20  04:38:21  hayes]
 * 
 * Revision 1.1.2.19  1998/01/16  01:23:33  ted
 * 	change non-portable KM_ constructs to portable ALT_ constructs
 * 	[1998/01/16  01:23:17  ted]
 * 
 * Revision 1.1.2.18  1998/01/06  23:35:49  ted
 * 	Fix parity of bit to match hardware documentation in debug routine
 * 	[1998/01/06  23:35:39  ted]
 * 
 * Revision 1.1.2.17  1997/12/03  01:23:01  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:23  ted]
 * 
 * Revision 1.1.2.16  1997/11/30  22:06:45  hayes
 * 	fix initialization for Tigon Rev5
 * 	[1997/11/30  22:06:19  hayes]
 * 
 * Revision 1.1.2.15  1997/11/19  19:23:53  ted
 * 	Support Tigon and Tigon2 in the same driver
 * 	[1997/11/19  19:20:53  ted]
 * 
 * Revision 1.1.2.14  1997/10/28  17:11:32  hayes
 * 	New verifyFW() function
 * 	New interface for Scratchpad routines to support PROCESSOR_B (TIGON_REV5)
 * 	[1997/10/28  02:00:33  hayes]
 * 
 * Revision 1.1.2.13  1997/10/21  00:13:00  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  23:00:09  ted]
 * 
 * Revision 1.1.13.1  1997/10/16  17:30:45  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:24:45  ted]
 * 
 * Revision 1.1.2.12  1997/09/16  03:11:29  ted
 * 	Change s to spl for better coding style
 * 	[1997/09/16  03:00:34  ted]
 * 
 * Revision 1.1.2.11  1997/08/23  17:40:29  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:38:49  skapur]
 * 
 * Revision 1.1.2.10  1997/08/13  18:53:46  ted
 * 	Put window base checking/setting in determineTGFlash
 * 	[1997/08/03  21:35:10  ted]
 * 
 * Revision 1.1.2.9  1997/08/01  00:16:22  ted
 * 	Remove extraneous window changes and mutexes in serial EEPROM accesses
 * 	[1997/08/01  00:15:33  ted]
 * 
 * Revision 1.1.2.8  1997/07/28  22:40:29  ted
 * 	Merge
 * 	[1997/07/28  22:37:44  ted]
 * 
 * 	Solaris dual homing support
 * 	[1997/07/28  21:39:47  ted]
 * 
 * Revision 1.1.2.7  1997/07/25  22:59:39  ted
 * 	Merge fixes from ge_0_22 and add link negotiation support
 * 	[1997/07/25  22:33:24  ted]
 * 
 * 	Architectural performance change
 * 	[1997/07/24  21:45:51  ted]
 * 
 * Revision 1.1.2.6  1997/05/25  19:33:04  akhan
 * 	Added Novell Specific code.
 * 	[1997/05/25  19:32:50  akhan]
 * 
 * Revision 1.1.2.5  1997/05/21  03:35:56  ted
 * 	Use pub_tg.h
 * 	[1997/05/21  03:27:17  ted]
 * 
 * Revision 1.1.2.4  1997/05/03  05:00:29  ted
 * 	Fix up statistics
 * 	[1997/05/03  04:56:53  ted]
 * 
 * 	Add Intel flash support and fix x86_solaris compile problem
 * 	[1997/04/27  21:31:57  ted]
 * 
 * Revision 1.1.2.3  1997/04/26  01:08:29  akhan
 * 	WriteTgMem fucntions had a check for dowrd alignment. For the intel processor we do not need to align it
 * 	[1997/04/26  01:03:49  akhan]
 * 
 * Revision 1.1.2.2  1997/04/04  22:38:28  ted
 * 	Fix problems with statics following altdrv.c split
 * 	[1997/04/04  22:01:13  ted]
 * 
 * Revision 1.1.2.1  1997/04/04  06:22:31  akhan
 * 	This file contains functions which were part
 * 	of altdrv.c and have been moved here because these
 * 	functions do not change a lot.
 * 	[1997/04/02  18:31:53  akhan]
 * 
 * $EndLog$
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/common/altrdwr.c,v 1.1.2.31 1999/02/01 21:09:34 ted Exp $";
#endif /* __lint */

/* General include files */
#if !defined(ALT_SYSV) && !defined(WINNT)
#include <string.h>		/* get these from systm.h via conf.h below */
#endif /* !defined(ALT_SYSV) && !defined(WINNT) */

#if !defined(WINNT)
#include <sys/types.h>
#include <sys/errno.h>
#endif	/* !defined(WINNT) */

/* System V include files */
#if defined(ALT_SYSV)
#include <sys/kmem.h>
#include <sys/ethernet.h>	/* ether_addr */
#include <sys/cmn_err.h>
#include <sys/conf.h>
#endif /* defined(ALT_SYSV) */

/* streams specific files */
#if defined(ALT_STREAM)
#include <sys/stream.h>		/* queue_t, before sunddi.h or warning */
#include <sys/dlpi.h>
#endif /* defined(ALT_STREAM) */

/* BSD include files */
#if defined(ALT_BSD)
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif /* defined(ALT_BSD) */

/* system specific include files */
#if defined(__sun)
#include <sys/byteorder.h>	/* htonl... */
#include <sys/dditypes.h>	/* ddi types */
#include <sys/ddi.h>		/* min */
#include <sys/ddidmareq.h>	/* ddi dma structs */
#include <sys/sunddi.h>		/* DDI_SUCCESS... */
#include <sys/ksynch.h>		/* kconfvar_t */
#endif /* defined(__sun) */

#if defined(WINNT)
#include <ndis.h>
#endif	/* defined(WINNT) */	

#if defined(__NETWARE_386__)
#include <odi.h>
#include <odi_nbi.h>
#include <odi_nesl.h>
#include <parser.h>
#include <cmsm.h>
#include <splat.h>
#include <portable.h>
#endif /* defined(__NETWARE_386__) */

/* alteon include files */
#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#if !defined(WINNT) && !defined(__NETWARE_386__)
#include "altioctl.h"
#endif /* !defined(WINNT) && !defined(__NETWARE_386__) */
#include "eeprom.h"
#include "pub_tg.h"
#include "altos.h"
#include "althdr.h"
#include "altreg.h"		/* must follow althdr.h */
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"

/* WARNING NOTE, HP-UX compiler does not like ANSI code in the kernel */

/*
 * miscellaneous function forwards
 */
#if defined(__STDC__) || defined(ALT_PROTOS)
STATIC U32 read_flash(struct altinst *, U32);
STATIC int write_flash(struct altinst *, U32, U32);
STATIC U32 read_serial_eeprom(struct tg_nic *, U32, bool *);
STATIC bool write_serial_eeprom(struct tg_nic *, U32, U32);
#else	/* defined(__STDC__) || defined(ALT_PROTOS) */
STATIC U32 read_flash();
STATIC int write_flash();
STATIC U32 read_serial_eeprom();
STATIC bool write_serial_eeprom();
#endif	/* defined(__STDC__) || defined(ALT_PROTOS) */

/* returns TRUE if code was running prior to stopping it */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
stopFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
stopFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U32 cpuState;
    struct tg_nic *nic = altp->initNic;

    cpuState = IN32(nic->handle, nic->base, TG_CPU_CONTROL);

    /* LINTED pointer cast may result in improper alignment */
    OUT32(nic->handle, nic->base, TG_CPU_CONTROL, cpuState | TG_CPU_HALT);

    return((cpuState & TG_CPU_HALTED) == 0);
}

/* take the buffered firmware and load it */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
loadFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
loadFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;

    if ((nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS) ||
        (nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS_B)) {
	ALTTRACE(loadFW, 0x70000, tigon2FwTextAddr, tigon2FwTextLen, 
		 tigon2FwDataAddr, tigon2FwDataLen);
	
	/* make sure the processor is stopped */
	stopFW(altp);
	
	/* write out the text segment */
	if (writeTGMem(altp, tigon2FwTextAddr, (caddr_t)tigon2FwText, 
		       tigon2FwTextLen, FALSE, FALSE) < 0)
	    return(FALSE);
	
	/* write out the rodata segment */
	if (tigon2FwRodataLen)
	    if (writeTGMem(altp, tigon2FwRodataAddr, (caddr_t)tigon2FwRodata,
			   tigon2FwRodataLen, FALSE, FALSE) < 0)
		return(FALSE);
	
	/* write out the data segment */
	if (tigon2FwDataLen)
	    if (writeTGMem(altp, tigon2FwDataAddr, (caddr_t)tigon2FwData, 
			   tigon2FwDataLen, FALSE, FALSE) < 0)
		return(FALSE);
    } else {
	ALTTRACE(loadFW, 0x7000b, tigonFwTextAddr, tigonFwTextLen, 
		 tigonFwDataAddr, tigonFwDataLen);
	
	/* make sure the processor is stopped */
	stopFW(altp);
	
	/* write out the text segment */
	if (writeTGMem(altp, tigonFwTextAddr, (caddr_t)tigonFwText, 
		       tigonFwTextLen, FALSE, FALSE) < 0)
	    return(FALSE);
	
	/* write out the rodata segment */
	if (tigonFwRodataLen)
	    if (writeTGMem(altp, tigonFwRodataAddr, (caddr_t)tigonFwRodata,
			   tigonFwRodataLen, FALSE, FALSE) < 0)
		return(FALSE);
	
	/* write out the data segment */
	if (tigonFwDataLen)
	    if (writeTGMem(altp, tigonFwDataAddr, (caddr_t)tigonFwData, 
			   tigonFwDataLen, FALSE, FALSE) < 0)
		return(FALSE);
    }

    /* mark the system state */
    altp->state = ALT_FW_LOADED;
    return(TRUE);
}
    
/* take the buffered firmware and check it */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
verifyFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
verifyFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    caddr_t tmpLoc;
    struct tg_nic *nic = altp->initNic;

    if ((nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS) ||
        (nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS_B)) {
	ALTTRACE(verifyFW, 0x7000c, tigon2FwTextAddr, tigon2FwTextLen, 
		 tigon2FwDataAddr, tigon2FwDataLen);
	
	/* make sure the processor is stopped */
	stopFW(altp);
	
	/* read the text segment and compare it */
	MEMALLOC(tmpLoc, caddr_t, tigon2FwTextLen, ALT_NOSLEEP);
	if (!tmpLoc) {
	    NOTE1("Couldn't allocate memory for text segment verification\n");
	    return(TRUE);
	}
	if (readTGMem(altp, tigon2FwTextAddr, tmpLoc, 
		      tigon2FwTextLen, FALSE, FALSE) < 0)
	    return(FALSE);
	if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigon2FwText, tigon2FwTextLen)) {
	    ERROR1("Firmware text segment not verified correctly\n");
	    MEMFREE(tmpLoc, tigon2FwTextLen, ALT_NOSLEEP);
	    return(FALSE);
	}
	MEMFREE(tmpLoc, tigon2FwTextLen, ALT_NOSLEEP);
	
	/* read the rodata segment and compare it */
	if (tigon2FwRodataLen) {
	    MEMALLOC(tmpLoc, caddr_t, tigon2FwRodataLen, ALT_NOSLEEP);
	    if (!tmpLoc) {
		NOTE1("Couldn't allocate memory for read-only data segment verification\n");
		return(TRUE);
	    }
	    if (readTGMem(altp, tigon2FwRodataAddr, tmpLoc,
			  tigon2FwRodataLen, FALSE, FALSE) < 0)
		return(FALSE);
	    if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigon2FwRodata, 
		       tigon2FwRodataLen)) {
		ERROR1("Firmware read-only data segment not verified correctly\n");
		MEMFREE(tmpLoc, tigon2FwRodataLen, ALT_NOSLEEP);
		return(FALSE);
	    }
	    MEMFREE(tmpLoc, tigon2FwRodataLen, ALT_NOSLEEP);
	}
	
	/* read the data segment and compare it */
	if (tigon2FwDataLen) {
	    MEMALLOC(tmpLoc, caddr_t, tigon2FwDataLen, ALT_NOSLEEP);
	    if (!tmpLoc) {
		NOTE1("Couldn't allocate memory for data segment verification\n");
		return(TRUE);
	    }
	    if (readTGMem(altp, tigon2FwDataAddr, tmpLoc,
			  tigon2FwDataLen, FALSE, FALSE) < 0)
		return(FALSE);
	    if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigon2FwData, tigon2FwDataLen)) {
		ERROR1("Firmware data segment not verified correctly\n");
		MEMFREE(tmpLoc, tigon2FwDataLen, ALT_NOSLEEP);
		return(FALSE);
	    }
	    MEMFREE(tmpLoc, tigon2FwDataLen, ALT_NOSLEEP);
	}
    } else {
	ALTTRACE(verifyFW, 0x7000d, tigonFwTextAddr, tigonFwTextLen, 
		 tigonFwDataAddr, tigonFwDataLen);
	
	/* make sure the processor is stopped */
	stopFW(altp);
	
	/* read the text segment and compare it */
	MEMALLOC(tmpLoc, caddr_t, tigonFwTextLen, ALT_NOSLEEP);
	if (!tmpLoc) {
	    NOTE1("Couldn't allocate memory for text segment verification\n");
	    return(TRUE);
	}
	if (readTGMem(altp, tigonFwTextAddr, tmpLoc, 
		      tigonFwTextLen, FALSE, FALSE) < 0)
	    return(FALSE);
	if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigonFwText, tigonFwTextLen)) {
	    ERROR1("Firmware text segment not verified correctly\n");
	    MEMFREE(tmpLoc, tigonFwTextLen, ALT_NOSLEEP);
	    return(FALSE);
	}
	MEMFREE(tmpLoc, tigonFwTextLen, ALT_NOSLEEP);
	
	/* read the rodata segment and compare it */
	if (tigonFwRodataLen) {
	    MEMALLOC(tmpLoc, caddr_t, tigonFwRodataLen, ALT_NOSLEEP);
	    if (!tmpLoc) {
		NOTE1("Couldn't allocate memory for read-only data segment verification\n");
		return(TRUE);
	    }
	    if (readTGMem(altp, tigonFwRodataAddr, tmpLoc,
			  tigonFwRodataLen, FALSE, FALSE) < 0)
		return(FALSE);
	    if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigonFwRodata, 
		       tigonFwRodataLen)) {
		ERROR1("Firmware read-only data segment not verified correctly\n");
		MEMFREE(tmpLoc, tigonFwRodataLen, ALT_NOSLEEP);
		return(FALSE);
	    }
	    MEMFREE(tmpLoc, tigonFwRodataLen, ALT_NOSLEEP);
	}
	
	/* read the data segment and compare it */
	if (tigonFwDataLen) {
	    MEMALLOC(tmpLoc, caddr_t, tigonFwDataLen, ALT_NOSLEEP);
	    if (!tmpLoc) {
		NOTE1("Couldn't allocate memory for data segment verification\n");
		return(TRUE);
	    }
	    if (readTGMem(altp, tigonFwDataAddr, tmpLoc,
			  tigonFwDataLen, FALSE, FALSE) < 0)
		return(FALSE);
	    if (MEMCMP(tmpLoc, (MEMCPY_TYPE)tigonFwData, tigonFwDataLen)) {
		ERROR1("Firmware data segment not verified correctly\n");
		MEMFREE(tmpLoc, tigonFwDataLen, ALT_NOSLEEP);
		return(FALSE);
	    }
	    MEMFREE(tmpLoc, tigonFwDataLen, ALT_NOSLEEP);
	}
    }

    /* mark the system state */
    altp->state = ALT_FW_LOADED;
    return(TRUE);
}
    
/* NIC memory size, 1 MB, this will change someday, need to handle that */    
#define NIC_MEM_SIZE (1024*1024)    
#define NIC_REGS_SIZE TG_SHMEM_START

/* allocate space and dump the current NIC memory and regs into it */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
dumpFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
dumpFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;
    bool running;

    /* double check we somehow didn't allocate already, if so, just
     * overlay old data */
    if (!nic->nicDump)
	MEMALLOC(nic->nicDump, caddr_t, NIC_MEM_SIZE, ALT_NOSLEEP);

    if (!nic->nicDump)
	return(FALSE);

    if (!nic->nicRegs)
	MEMALLOC(nic->nicRegs, caddr_t, NIC_REGS_SIZE, ALT_NOSLEEP);

    if (!nic->nicRegs)
	return(FALSE);

    ALTTRACE(dumpFW, 0x70001, nic->nicDump, nic->nicRegs, 0, 0);

    running = stopFW(altp);

    /* dump the NIC's memory */
    readTGMem(altp, 0, nic->nicDump, NIC_MEM_SIZE, FALSE, FALSE);

    /* dump the NIC's hardware registers */
    (void) ALT_MEMCPY_TOHOST(nic->nicRegs, nic->base, NIC_REGS_SIZE);

    if (running)
	restartFW(altp);

#if 0				/* OLD WAY, MAYBE HANDY LATER */
    struct tg_nic *nic = altp->initNic;
    char *tmpBuf;
    U32 *curp;
    int i;

    /* get the GPRs */
    MEMALLOC(tmpBuf, char *, 32 * sizeof(U32), ALT_SLEEP);

    /* loop through the regs reading them in */
    curp = (U32 *)tmpBuf;
    for (i=0; i < 32; i++) 
	*curp++ = IN32(nic->handle, nic->base, TG_GPR0 + i*sizeof(U32));
    /* copy out the info */
    if (COPYOUT(tmpBuf, core->gprs, 32 * sizeof(U32)) < 0)
	return(FALSE);
    MEMFREE(tmpBuf, 32 * sizeof(U32), ALT_SLEEP);

    /* get the text area */
    if (readTGMem(altp, altFwTextAddr, core->fwText, altFwTextLen, 
		  TRUE, FALSE) < 0)
	return(FALSE);

    /* get the data area */
    if (readTGMem(altp, altFwDataAddr, core->fwData, altFwDataLen, 
		  TRUE, FALSE) < 0)
	return(FALSE);
    
    /* get the rodata area */
    if (readTGMem(altp, altFwRodataAddr, core->fwRodata, altFwRodataLen, 
		  TRUE, FALSE) < 0)
	return(FALSE);
    
    /* get the bss area */
    if (readTGMem(altp, altFwBssAddr, core->fwBss, altFwBssLen, 
		  TRUE, FALSE) < 0)
	return(FALSE);
    
    /* get the sbss area */
    if (readTGMem(altp, altFwSbssAddr, core->fwSbss, altFwSbssLen, 
		  TRUE, FALSE) < 0)
	return(FALSE);

    /* get the stack */
    if (readTGMem(altp, FW_STACK_LOC, core->fwStack, MAX_STACK_LEN, 
		  TRUE, FALSE) < 0)
	return(FALSE);
#endif /* 0 */

    return(TRUE);
}
    
void
#if defined(__STDC__) || defined(ALT_PROTOS)
freeDump(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
freeDump(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;
    
    ALTTRACE(freeDump, 0x7000a, nic->nicDump, nic->nicRegs, 0, 0);

    if (nic->nicDump) {
	MEMFREE(nic->nicDump, NIC_MEM_SIZE, ALT_NOSLEEP);
	nic->nicDump = (caddr_t) 0;
    }

    if (nic->nicRegs) {
	MEMFREE(nic->nicRegs, NIC_REGS_SIZE, ALT_NOSLEEP);
	nic->nicRegs = (caddr_t) 0;
    }
}

/* start the firmware running from the beginning */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
startFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
startFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;

    if ((nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS) ||
        (nic->tigonRev == TG_MHC_LSI_STANDARD_CELL_VERS_B)) {
	ALTTRACE(startFW, 0x70002, tigon2FwStartAddr, 0, 0, 0);

	/* stop the processor */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_CPU_CONTROL, 
	      IN32(nic->handle, nic->base, TG_CPU_CONTROL) | TG_CPU_HALT);
	
	/* set the PC to the start address */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_PROG_CTR, tigon2FwStartAddr);
	
	/* start the processor */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_CPU_CONTROL, 
	      IN32(nic->handle, nic->base, TG_CPU_CONTROL) &  
	      ~(TG_CPU_HALT | TG_CPU_SINGLE));
    } else {
	ALTTRACE(startFW, 0x7000e, tigonFwStartAddr, 0, 0, 0);

	/* stop the processor */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_CPU_CONTROL, 
	      IN32(nic->handle, nic->base, TG_CPU_CONTROL) | TG_CPU_HALT);
	
	/* set the PC to the start address */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_PROG_CTR, tigonFwStartAddr);
	
	/* start the processor */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_CPU_CONTROL, 
	      IN32(nic->handle, nic->base, TG_CPU_CONTROL) &  
	      ~(TG_CPU_HALT | TG_CPU_SINGLE));
    }

    return(TRUE);
}

/* restart the firmware running from wherever it was */
bool
#if defined(__STDC__) || defined(ALT_PROTOS)
restartFW(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
restartFW(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;

    ALTTRACE(restrtFW, 0x70003, 0, 0, 0, 0);

    /* start the processor */
    /* LINTED pointer cast may result in improper alignment */
    OUT32(nic->handle, nic->base, TG_CPU_CONTROL, 
	  IN32(nic->handle, nic->base, TG_CPU_CONTROL) &  
	  ~(TG_CPU_HALT | TG_CPU_SINGLE));
    return(TRUE);
}

bool
#if defined(__STDC__) || defined(ALT_PROTOS)
stopMAC(struct altinst *altp)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
stopMAC(altp)
struct altinst *altp;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U32 macRecvState;
    struct tg_nic *nic = altp->initNic;

    macRecvState = IN32(nic->handle, nic->base, TG_MAC_RX_STATE);

    /* LINTED pointer cast may result in improper alignment */
    OUT32(nic->handle, nic->base, TG_MAC_RX_STATE, 
	  macRecvState | TG_MAC_RX_STATE_STOP_NEXT);

    return((macRecvState & TG_MAC_RX_STATE_STOP_NEXT) == 
	   TG_MAC_RX_STATE_STOP_NEXT);
}

/* copy from Tigon memory, local or user space */

int
#if defined(__STDC__) || defined(ALT_PROTOS)
readTGMem(struct altinst *altp, U32 tgAddr, char *userAddr, int len, 
	  bool userSpace, bool swap)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
readTGMem(altp, tgAddr, userAddr, len, userSpace, swap)
struct altinst *altp;
U32 tgAddr;
char *userAddr;
int len;
bool userSpace;
bool swap;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    int tmpLen;
    U32 origTgAddr = tgAddr;
    caddr_t origUserAddr = userAddr;
    int origLen = len;
    struct tg_nic *nic = altp->initNic;
    int rc = 0;
    U32 origWindow;

    ALT_DISABLE(spl, &nic->memWindowMu);
    /* save this inside critical section */
    origWindow = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    /* copy the first unaligned section */
    if (!WINDOW_ALIGNED(tgAddr)) {
	/* set the window */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, WINDOW_PAGE(tgAddr));
	
	tmpLen = MIN(WINDOW_RESID(tgAddr), len);
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    goto readMemExit;
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (alt_copyout((char *)nic->base + TG_MEM_WINDOW + 
			    WINDOW_OFFSET(tgAddr), 
			    userAddr, tmpLen) < 0) {
		rc = -1;		/* invalid to go to user space */
		goto readMemExit;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    /* if either not on a 4 byte boundary, hack it */
	    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (tmpLen & 3)) {
		U32 *tmpBuf;

		MEMALLOC(tmpBuf, U32 *, tmpLen+8, ALT_NOSLEEP);
		if (!tmpBuf) {
		    rc = -1;
		    goto readMemExit;
		}
		ALT_SWAP_MEMCPY_TOHOST((char *)tmpBuf,
				(char *)nic->base + TG_MEM_WINDOW +
				((int)(WINDOW_OFFSET(tgAddr)) & ~3),
				tmpLen + 4 /*- ((int)tgAddr & 3) */);
		MEMCPY(userAddr, (char *)tmpBuf + ((int)tgAddr & 3), 
		       tmpLen);

		MEMFREE(tmpBuf, tmpLen+8, ALT_NOSLEEP);
	    } else {
		if (swap) 
                  {
		    ALT_SWAP_MEMCPY_TOHOST(userAddr, 
				    (char *)nic->base + TG_MEM_WINDOW + 
				    WINDOW_OFFSET(tgAddr),
				    tmpLen);
                  }
		else
                  {
		    ALT_MEMCPY_TOHOST(userAddr, 
			       (char *)nic->base + TG_MEM_WINDOW + 
			       WINDOW_OFFSET(tgAddr),
			       tmpLen);
                  }
	    }
	}
	
	/* adjust the pointers */
	userAddr += tmpLen;
	tgAddr += tmpLen;
	len -= tmpLen;
    }

    /* copy the rest */
    while (len > 0) {
	/* set the window */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, tgAddr);
	
	tmpLen = MIN(WINDOW_LEN, len);
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    goto readMemExit;
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (alt_copyout((char *)nic->base + TG_MEM_WINDOW, 
			    userAddr, tmpLen) < 0) {
		rc = -1;
		goto readMemExit;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    /* if userAddr not on a 4 byte boundary, hack it, tgAddr will be */
	    if (((int)userAddr & 3) || (tmpLen & 3)) {
		U32 *tmpBuf;

		MEMALLOC(tmpBuf, U32 *, tmpLen+8, ALT_NOSLEEP);
		if (!tmpBuf) {
		    rc = -1;
		    goto readMemExit;
		}
		ALT_SWAP_MEMCPY_TOHOST((char *)tmpBuf,
				(char *)nic->base + TG_MEM_WINDOW +
				WINDOW_OFFSET(tgAddr),
				tmpLen + 4 /*- ((int)tgAddr & 3) */);
		MEMCPY(userAddr, (char *)tmpBuf, tmpLen);
		MEMFREE(tmpBuf, tmpLen+8, ALT_NOSLEEP);
	    } else {
		if (swap) 
                  {
		    ALT_SWAP_MEMCPY_TOHOST(userAddr, 
				    (char *)nic->base + TG_MEM_WINDOW,
				    tmpLen);
                  }
		else
                  {
		    ALT_MEMCPY_TOHOST(userAddr, 
			       (char *)nic->base + TG_MEM_WINDOW,
			       tmpLen);
                  }
	    }
	}

	/* adjust the pointers */
	userAddr += tmpLen;
	tgAddr += tmpLen;
	len -= tmpLen;
    }

    /* be careful when tracing */
    if (((int)origUserAddr & 3) && !userSpace) {
	U32 *tmpBuf;

	MEMALLOC(tmpBuf, U32 *, tmpLen+8, ALT_NOSLEEP);
	if (tmpBuf) {
	    MEMCPY(userAddr, (char *)tmpBuf, tmpLen);
	    ALTTRACE2G(rdtgmem, 0x70004, origTgAddr, origUserAddr, 
		       origLen, userSpace,
		       userSpace ? "can't read user space" : (caddr_t)tmpBuf);
	    MEMFREE(tmpBuf, tmpLen+8, ALT_NOSLEEP);
	}
    } else
	ALTTRACE2G(rdtgmem, 0x70005, origTgAddr, origUserAddr, 
		   origLen, userSpace,
		   userSpace ? "can't read user space" : origUserAddr);

readMemExit:
    /* restore memory window */
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, origWindow);
    ALT_ENABLE(spl, &nic->memWindowMu);
    return(rc);
}
	
/* copy data to Tigon via memory window, user space or kernel */

int
#if defined(__STDC__) || defined(ALT_PROTOS)
writeTGMem(struct altinst *altp, U32 tgAddr, char * userAddr, int len,
	   bool userSpace, bool swap)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
writeTGMem(altp, tgAddr, userAddr, len, userSpace, swap)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
bool userSpace;
bool swap;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;
    int tmpLen;
    int rc = 0;
    U32 origWindow;

    ALTTRACE2G(wrtgmem, 0x70006, tgAddr, userAddr, len, userSpace,
	       userSpace ? "can't read user space" : userAddr);

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3))
	return(-1);
    
    ALT_DISABLE(spl, &nic->memWindowMu);
    /* save this inside critical section */
    origWindow = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    /* copy the first unaligned section */
    if (!WINDOW_ALIGNED(tgAddr)) {
	/* set the window */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, WINDOW_PAGE(tgAddr));
	
	tmpLen = MIN(WINDOW_RESID(tgAddr), len);
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    goto memWriteExit;
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (alt_copyin(userAddr, 
			   (char *)nic->base + TG_MEM_WINDOW + 
			   WINDOW_OFFSET(tgAddr), 
			   tmpLen) < 0) {
		rc = -1;
		goto memWriteExit;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    /* XXX put non-aligned fix in here */
	    if (swap)
              {
		ALT_SWAP_MEMCPY_TODEV((char *)nic->base + TG_MEM_WINDOW +
				WINDOW_OFFSET(tgAddr), 
				userAddr, tmpLen);
              }
	    else
              {
		ALT_MEMCPY_TODEV((char *)nic->base + TG_MEM_WINDOW +
			   WINDOW_OFFSET(tgAddr), 
			   userAddr, tmpLen);
              }
	}
	
	/* adjust the pointers */
	userAddr += tmpLen;
	tgAddr += tmpLen;
	len -= tmpLen;
    }
    
    /* copy the rest */
    while (len > 0) {
	/* set the window */
	/* LINTED pointer cast may result in improper alignment */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, tgAddr);
	
	tmpLen = MIN(WINDOW_LEN, len);
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    goto memWriteExit;
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (alt_copyin(userAddr, (caddr_t)nic->base + TG_MEM_WINDOW, 
			   tmpLen) < 0) {
		rc = -1;
		goto memWriteExit;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    /* XXX put non-aligned fix in here */
	    if (swap) 
              {
		ALT_SWAP_MEMCPY_TODEV((char *)nic->base + TG_MEM_WINDOW, 
				userAddr, tmpLen);
              }
	    else
              {
		ALT_MEMCPY_TODEV((char *)nic->base + TG_MEM_WINDOW, 
			   userAddr, tmpLen);
              }
	}
	
	/* adjust the pointers */
	userAddr += tmpLen;
	tgAddr += tmpLen;
	len -= tmpLen;
    }

memWriteExit:
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, origWindow);
    ALT_ENABLE(spl, &nic->memWindowMu);
    return(rc);
}

/* copy from Tigon scratchpad memory, to local or user space */
int
#if defined(__STDC__) || defined(ALT_PROTOS)
readTGScratch(struct altinst *altp, U32 tgAddr, char * userAddr, int len, 
	      bool userSpace, bool swap, int cpu)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
readTGScratch(altp, tgAddr, userAddr, len, userSpace, swap, cpu)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
bool userSpace;
bool swap;
int cpu;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->nic;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3))
	return(-1);
    
    while (len > 0) {
	/* set the window */
	OUT32(nic->handle, nic->base, CPU_REG(TG_SRAM_ADDR, cpu), tgAddr);

	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    return(-1);		/* invalid to go to user space */
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (swap) {
		if (alt_swap_copyout(((char *)nic->base + 
				      CPU_REG(TG_SRAM_DATA, cpu)), 
				     userAddr, MIN(4, len)) < 0)
		    return(-1);
	    } else {
		if (alt_copyout(((char *)nic->base + 
				 CPU_REG(TG_SRAM_DATA, cpu)), 
				userAddr, MIN(4, len)) < 0)
		    return(-1);
	    }
#endif /* !defined(ALT_BSD) */
	} else
	    /* XXX put non-aligned fix in here */
	    if (swap)
              {
		ALT_SWAP_MEMCPY_TOHOST(userAddr, 
				(char *)nic->base + 
				CPU_REG(TG_SRAM_DATA, cpu), 4);
              }
	    else
              {
		ALT_MEMCPY_TOHOST(userAddr, 
			   (char *)nic->base + CPU_REG(TG_SRAM_DATA, cpu), 4);
              }

	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }
	
    return(0);
}
	
/* copy data from Tigon scratchpad memory, to local or user space */
int
#if defined(__STDC__) || defined(ALT_PROTOS)
writeTGScratch(struct altinst *altp, U32 tgAddr, char * userAddr, int len,
	   bool userSpace, bool swap, int cpu)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
writeTGScratch(altp, tgAddr, userAddr, len, userSpace, swap, cpu)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
bool userSpace;
bool swap;
int cpu;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    struct tg_nic *nic = altp->initNic;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3))
	return(-1);
    
    while (len > 0) {
	/* set the window */
	OUT32(nic->handle, nic->base, CPU_REG(TG_SRAM_ADDR, cpu), tgAddr);

	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    return(-1);		/* invalid to go to user space */
#else /* !defined(ALT_BSD) */
	    /* XXX put non-aligned fix in here */
	    if (alt_copyin(((char *)nic->base + CPU_REG(TG_SRAM_DATA, cpu)), 
			   userAddr, MIN(4, len)) < 0)
		return(-1);
#endif /* !defined(ALT_BSD) */
	} else
	    /* XXX put non-aligned fix in here */
	    OUT32(nic->handle, nic->base, CPU_REG(TG_SRAM_DATA, cpu), 
		  *(U32 *)userAddr);

	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }

    return(0);
}

/* copy from Tigon flash, local or user space */

int
#if defined(__STDC__) || defined(ALT_PROTOS)
readTGFlash(struct altinst *altp, U32 tgAddr, char * userAddr, int len, 
	      bool userSpace)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
readTGFlash(altp, tgAddr, userAddr, len, userSpace)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
bool userSpace;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    bool running;
    struct tg_nic *nic = altp->initNic;
    int rc = 0;
    U32 origWindow;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3) ||
        nic->flash_type == FLASH_UNKNOWN)
	return(-1);
    
    /* make sure firmware is stopped */
    running = stopFW(altp);

    ALT_DISABLE(spl, &nic->memWindowMu);
    /* save this inside critical section */
    origWindow = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    /* word by word is the only way to do flash*/
    while (len > 0) {
	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    break;
#else /* !defined(ALT_BSD) */
	    U32 tmp;

	    tmp = read_flash(altp, tgAddr);
	    if (alt_copyout((caddr_t)&tmp, userAddr, 
			    MIN(4, len)) < 0) {
		rc = -1;
		break;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    *(U32 *)userAddr = read_flash(altp, tgAddr);
	}

	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }
	
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, origWindow);
    ALT_ENABLE(spl, &nic->memWindowMu);

#if FIRMWARE_DEBUGGED
    if (running)
	if (!restartFW(altp))
	    rc = -1;
#endif /* FIRMWARE_DEBUGGED */
    
    return(rc);
}
	
/* copy data to Tigon flash via memory window, user space or kernel */

int
#if defined(__STDC__) || defined(ALT_PROTOS)
writeTGFlash(struct altinst *altp, U32 tgAddr, char * userAddr, int len,
	   bool userSpace)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
writeTGFlash(altp, tgAddr, userAddr, len, userSpace)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
int userSpace;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    bool running;
    struct tg_nic *nic = altp->nic;
    int rc = 0;
    U32 origWindow;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3) ||
        nic->flash_type == FLASH_UNKNOWN)
	return(-1);
    
    /* make sure firmware is stopped */
    running = stopFW(altp);

    ALT_DISABLE(spl, &nic->memWindowMu);
    /* save this inside critical section */
    origWindow = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    while (len > 0) {
	/* set the window */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, tgAddr);

	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    break;
#else /* !defined(ALT_BSD) */
	    U32 tmp;

	    if (alt_copyin((caddr_t)&tmp, userAddr, 
			   MIN(4, len)) < 0) {
		rc = -1;
		break;
	    }
	    if (write_flash(altp, tgAddr, tmp) < 0) {
		rc = -1;
		break;
	    }
#endif /* !defined(ALT_BSD) */
	} else
	    if (write_flash(altp, tgAddr, *(U32 *)userAddr) < 0) {
		rc = -1;
		break;
	    }

	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }
	
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, origWindow);
    ALT_ENABLE(spl, &nic->memWindowMu);

#if FIRMWARE_DEBUGGED
    if (running)
	if (!restartFW(altp))
	    return(-1);
#endif /* FIRMWARE_DEBUGGED */
    
    return(rc);
}

STATIC U32
#if defined(__STDC__) || defined(ALT_PROTOS)
read_flash(struct altinst *altp, U32 addr)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
read_flash(altp, addr)
struct altinst *altp;
U32 addr;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U32 i;
    U32 result = 0;
    U32 tmp;
    struct tg_nic *nic = altp->initNic;
    U32 bits;
    
    /* munge the misc_local bits to set different 256k parts of the flash */
    /* zero the bits then put them on based on the current address */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &
	  ~(TG_MLC_SRAM_LADDR_22 | TG_MLC_SRAM_LADDR_21));
    bits = addr & 0xc0000;
    bits >>= 4;
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) | bits);
    
    /* strip the FLASH_BASE address and shift it over 3 bits for starting */
    addr = ((~TG_BEG_FLASH & addr) << 3) | TG_BEG_FLASH;
    for (i=0; i < 32; i += 8, addr += 8) {
	/* set the window for the current byte */
	OUT32(altp->nic->handle, altp->nic->base, TG_WINDOW_BASE, addr);
	/* read the current byte */
	result |= ((tmp = IN32(altp->nic->handle, altp->nic->base, 
			       TG_WINDOW_DATA)) &
		   (unsigned)EEPROM_BYTE_LOC) >> i;
    }
    return(ntohl(result));
}

/* ugly code warning, constants in line, who wrote this :-) */
/* ok, ok it was me...Ted, come beat me up later, it was late at night */
STATIC int
#if defined(__STDC__) || defined(ALT_PROTOS)
write_flash(struct altinst *altp, U32 addr, U32 data)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
write_flash(altp, addr, data)
struct altinst *altp;
U32 addr;
U32 data;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U32 i, d;
    struct tg_nic *nic = altp->initNic;
    int wait = 0;
    U32 bits;
    U32 trash;			/* for TIGON PCI access workaround */
    int rc = 0;
    U32 status;

    /* enable writing FLASH */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) |
	  TG_MLC_EEPROM_WRITE_ENA);

    /* munge the misc_local bits to set different 256k parts of the flash */
    /* zero the bits then put them on based on the current address */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &
	  ~(TG_MLC_SRAM_LADDR_22 | TG_MLC_SRAM_LADDR_21));
    bits = addr & 0xc0000;
    bits >>= 4;
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) | bits);
    
    /* strip the FLASH_BASE address and shift it over 3 bits for starting */
    addr = ((~TG_BEG_FLASH & addr) << 3) | TG_BEG_FLASH;
    
    /* swap the bytes around if necessary */
    data = htonl(data);

    for (i=0; i < 32 && rc == 0; i += 8, addr += 8) {
	switch (nic->flash_type) {
	case FLASH_INTEL:
	    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, addr);
	    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x40404040);
	    d = (data << i) & EEPROM_BYTE_LOC;
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, d);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x50505050);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x70707070);
	    while (!(status = IN32(nic->handle, nic->base, TG_WINDOW_DATA) & 
		     0x80000000))
		if ((status & 0x1a) || ++wait > 1000000) {
		    rc = -1;
		    break;	/* wait for the write, but break eventually */
		}
	    /*
	     * always leave in READ state
	     */
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xffffffff);
	    break;

	case FLASH_AMD:
	    /* enable writing flash for each byte */
	    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
	    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xaaaaaaaa);
	    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80015550);
	    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x55555555);
	    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
	    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xa0a0a0a0);
    
	    /* set the window for the current byte */
	    d = (data << i) & EEPROM_BYTE_LOC;
	    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, addr);
	    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, d);
	    while ((IN32(nic->handle, nic->base, TG_WINDOW_DATA) & 0x80000000) 
		   !=  (d & 0x80000000))
		if (++wait > 1000000) {
		    rc = -1;
		    break;	/* wait for the write, but break eventually */
		}
	    break;

	default:
	    rc = -1;
	    break;
	}
    }

    /* disable writing FLASH */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &
	  ~TG_MLC_SER_EEPROM_WRITE_ENA);

    return(rc);
}

void
determineTGFlash(struct tg_nic *nic)
{
    U32 trash;
    U32 type, type2;
    
    ALT_DISABLE(spl, &nic->memWindowMu);
    
    /* enable writing FLASH */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) |
	  TG_MLC_EEPROM_WRITE_ENA);

    /* munge the misc_local bits to set first 256k parts of the flash */
    /* zero the bits then put them on based on the current address */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &
	  ~(TG_MLC_SRAM_LADDR_22 | TG_MLC_SRAM_LADDR_21));

    /* figure out which flash part we're looking at */
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xaaaaaaaa);
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80015550);
    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x55555555);
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
    trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x90909090);
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80000000);
    type = (IN32(nic->handle, nic->base, TG_WINDOW_DATA) & EEPROM_BYTE_LOC)
	>> 24;
    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80000008);
    type2 = (IN32(nic->handle, nic->base, TG_WINDOW_DATA) & EEPROM_BYTE_LOC)
	>> 24;
    ALTTRACEV(getFlash, 0x70008, type, type2, 0, 0);
    if (type == 1 && type2 == 0xd5) {
	/* put it back in read mode */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
	trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xaaaaaaaa);
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80015550);
	trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x55555555);
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
	trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xf0f0f0f0);
	nic->flash_type = FLASH_AMD;
    } else {
	/* get the intel flash in read mode - just incase the 
	 * auto sel didn't work right the first time */
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x8002aaa8);
	trash = IN32(nic->handle, nic->base, TG_WINDOW_BASE);
	OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xffffffff);
	/* now read the identifier to confirm that it is Intel flash */
	OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0x90909090);
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80000000);
        type = (IN32(nic->handle, nic->base, TG_WINDOW_DATA) & EEPROM_BYTE_LOC)
  	    >> 24;
	OUT32(nic->handle, nic->base, TG_WINDOW_BASE, 0x80000008);
        type2 = (IN32(nic->handle, nic->base, TG_WINDOW_DATA) & EEPROM_BYTE_LOC)
  	    >> 24;
	ALTTRACEV(getFlash, 0x70009, type, type2, 0, 0);
	if (type == 0x89 && type2 == 0xa6) {
	    nic->flash_type = FLASH_INTEL;
	    /* put the intel flash into read mode */
	    OUT32(nic->handle, nic->base, TG_WINDOW_DATA, 0xffffffff);
	} else 
	    nic->flash_type = FLASH_UNKNOWN;
    }
    
    /* disable writing FLASH */
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL, 
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &
	  ~TG_MLC_SER_EEPROM_WRITE_ENA);

    OUT32(nic->handle, nic->base, TG_WINDOW_BASE, SEND_RING_PTR);
    ALT_ENABLE(spl, &nic->memWindowMu);
}

int
#if defined(__STDC__) || defined(ALT_PROTOS)
readTGSerEeprom(struct altinst *altp, U32 tgAddr, char * userAddr, int len, 
		bool userSpace)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
readTGSerEeprom(altp, tgAddr, userAddr, len, userSpace)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
bool userSpace;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    bool running, worked;
    struct tg_nic *nic = altp->initNic;
    int rc = 0;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3))
	return(-1);
    
    /* make sure firmware is stopped */
    running = stopFW(altp);

    /* do it the slow way, word by word */
    while (len > 0) {
	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    break;
#else /* !defined(ALT_BSD) */
	    U32 tmp;

	    tmp = read_serial_eeprom(nic, tgAddr, &worked);
	    if (!worked) {
		rc = -1;
		break;
	    } else if (alt_copyout((caddr_t)&tmp, userAddr, MIN(4, len)) < 0) {
		rc = -1;
		break;
	    }
#endif /* !defined(ALT_BSD) */
	} else {
	    *(U32 *)userAddr = read_serial_eeprom(nic, tgAddr, &worked);
	    if (!worked) {
		rc = -1;
		break;
	    }
	}

	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }
	
#if FIRMWARE_DEBUGGED
    if (running)
	if (!restartFW(altp))
	    return(-1);
#endif /* FIRMWARE_DEBUGGED */
    
    return(rc);
}
	
/* copy data to Tigon Serial EEPROM via memory window, user space or kernel */

int
#if defined(__STDC__) || defined(ALT_PROTOS)
writeTGSerEeprom(struct altinst *altp, U32 tgAddr, char * userAddr, int len,
	   bool userSpace)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
writeTGSerEeprom(altp, tgAddr, userAddr, len, userSpace)
struct altinst *altp;
U32 tgAddr;
char * userAddr;
int len;
int userSpace;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    bool running;
    struct tg_nic *nic = altp->initNic;
    int rc = 0;

    /* fail if non-aligned data */
    if (((int)userAddr & 3) || ((int)tgAddr & 3) || (len & 3))
	return(-1);
    
    /* make sure firmware is stopped */
    running = stopFW(altp);

    while (len > 0) {
	/* get the value */
	if (userSpace) {
#if !defined(ALT_BSD)
	    rc = -1;		/* invalid to go to user space */
	    break;
#else /* !defined(ALT_BSD) */
	    U32 tmp;

	    if (alt_copyin((caddr_t)&tmp, userAddr, 
			   MIN(4, len)) < 0) {
		rc = -1;
		break;
	    } else if (!write_serial_eeprom(nic, tgAddr, tmp)) {
		rc = -1;
		break;
	    }
#endif /* !defined(ALT_BSD) */
	} else
	    if (!write_serial_eeprom(nic, tgAddr, *(U32 *)userAddr)) {
		rc = -1;
		break;
	    }
	
	/* adjust the pointers */
	userAddr += 4;
	tgAddr += 4;
	len -= 4;
    }

#if FIRMWARE_DEBUGGED
    if (running)
	if (!restartFW(altp))
	    return(-1);
#endif /* FIRMWARE_DEBUGGED */
    
    return(rc);
}

/* function used by serial_eeprom functions */
STATIC bool ser_eeprom_chk_ack(struct tg_nic *);
STATIC int get_serial_byte(struct tg_nic *, U32);
STATIC bool put_serial_byte(struct tg_nic *, U32 addr, U8);
STATIC void ser_eeprom_prep(struct tg_nic *, U8);

STATIC U32
#if defined(__STDC__) || defined(ALT_PROTOS)
read_serial_eeprom(struct tg_nic *nic, U32 addr, bool *worked)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
read_serial_eeprom(nic, addr, worked)
struct tg_nic *nic;
U32 addr;
bool worked;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U32 rc, data;

    if ((rc = get_serial_byte(nic, addr++)) != -1) {
	data = rc << 24;
	if ((rc = get_serial_byte(nic, addr++)) != -1) {
	    data += rc << 16;
	    if ((rc = get_serial_byte(nic, addr++)) != -1) {
		data += rc << 8;
		if ((rc = get_serial_byte(nic, addr)) != -1) {
		    data += rc;
		    *worked = TRUE;
		    return(htonl(data));
		}
	    }
	}
    }

    *worked = FALSE;
    return(0);
}

STATIC bool
#if defined(__STDC__) || defined(ALT_PROTOS)
write_serial_eeprom(struct tg_nic *nic, U32 addr, U32 data)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
write_serial_eeprom(nic, addr, data)
struct tg_nic *nic;
U32 addr;
U32 data;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    /* swap the bytes around if necessary */
    data = htonl(data);

    if (put_serial_byte(nic, addr++, (U8)(data >> 24)))
	if (put_serial_byte(nic, addr++, (U8)(data >> 16)))
	    if (put_serial_byte(nic, addr++, (U8)(data >> 8)))
		if (put_serial_byte(nic, addr, (U8)data)) 
		    return(TRUE);
    return(FALSE);
}

/* delays in usec (actual delays to right) */
#define ALT_SER_EEPROM_THIGH	 1    /* > 0.6 us */
#define ALT_SER_EEPROM_H_THIGH	 1
#define ALT_SER_EEPROM_TLOW	 2    /* > 1.2 us */
#define ALT_SER_EEPROM_H_TLOW	 1
#define ALT_SER_EEPROM_START_SU  1    /* > 0.6 us (setup time) */
#define ALT_SER_EEPROM_START_HO  1    /* > 0.6 us (hold time) */
#define ALT_SER_EEPROM_STOP_SU	 1    /* > 0.6 us (setup time) */
#define ALT_SER_EEPROM_TBUF	 2    /* > 1.2 us */

#define SER_EEPROM_WRITE_SELECT    0xa0
#define SER_EEPROM_READ_SELECT     0xa1

#define MLC_SET_BIT(bit)						\
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL,		\
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) |	(bit))
#define MLC_CLEAR_BIT(bit)						\
    OUT32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL,		\
	  IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) & ~(bit))
#define IS_SER_EEPROM_BIT_SET					\
    ((IN32(nic->handle, nic->base, TG_MISC_LOCAL_CONTROL) &	\
           TG_MLC_SER_EEPROM_DATA_IN) != 0)

#define SER_EEPROM_START			\
    SNOOZE(ALT_SER_EEPROM_H_TLOW);		\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_DATA_OUT);	\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);	\
    SNOOZE(ALT_SER_EEPROM_H_TLOW);		\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);	\
    SNOOZE(ALT_SER_EEPROM_START_SU);		\
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_DATA_OUT);	\
    SNOOZE(ALT_SER_EEPROM_START_HO);		\
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);

#define SER_EEPROM_STOP				\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);	\
    SNOOZE(ALT_SER_EEPROM_H_TLOW);		\
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_DATA_OUT);	\
    SNOOZE(ALT_SER_EEPROM_H_TLOW);		\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);	\
    SNOOZE(ALT_SER_EEPROM_STOP_SU);		\
    MLC_SET_BIT(TG_MLC_SER_EEPROM_DATA_OUT);	\
    SNOOZE(ALT_SER_EEPROM_TBUF);		\
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);

/* Gets byte from serial eeprom at byte address 'eeprom_addr'.  Can only
 * access serial eeprom at chip-address 0 and this part cannot have a density
 * of more than 64k bytes.  This routine implements a "random read" function
 * as well as a "sequential read" function.  If 'random_read' is MLC_SET_BIT a random
 * read is performed, if not, a sequential read is performed.  All writes are
 * unsupported.  See Atmel AT24C32/64 product spec for more information.
 *
 *  The basic format of the sequential read is as follows...
 *
 *	S1010ddd1ADDDDDDDD1S
 *      t        c         t
 *      a        k         o
 *      r                  p
 *      t
 *
 *  The basic format of the random read is as follows...
 *
 *	S1010ddd0A***RRRRRARRRRRRRRAS1010ddd1ADDDDDDDD1S
 *      t        c        c        ct        c         t
 *      a        k        k        ka        k         o
 *      r                           r                  p
 *      t                           t
 *
 *  Where... 	d == device address
 *		R == byte address
 *		D == data read from eeprom
 */

STATIC int
#if defined(__STDC__) || defined(ALT_PROTOS)
get_serial_byte(struct tg_nic *nic, U32 addr)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
get_serial_byte(nic, addr)
struct tg_nic *nic;
U32 addr;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    int i, val = 0;
    
    SER_EEPROM_START;
    ser_eeprom_prep(nic, SER_EEPROM_WRITE_SELECT);
    if (!ser_eeprom_chk_ack(nic))
	return(-1);
    ser_eeprom_prep(nic, (U8)(addr >> 8));
    if (!ser_eeprom_chk_ack(nic))
	return(-1);
    ser_eeprom_prep(nic, (U8)(addr & 0xff));
    if (!ser_eeprom_chk_ack(nic))
	return(-1);

    SER_EEPROM_START;
    ser_eeprom_prep(nic, SER_EEPROM_READ_SELECT);
	if (!ser_eeprom_chk_ack(nic))
	    return(-1);
    for (i=0; i<8; i++) {
	MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);
	SNOOZE(ALT_SER_EEPROM_TLOW);
	MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
	SNOOZE(ALT_SER_EEPROM_H_THIGH);
	/* sample data mid high clk */
	val = (val << 1) | IS_SER_EEPROM_BIT_SET;
	SNOOZE(ALT_SER_EEPROM_H_THIGH);
	MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
	if (i == 7)
	    MLC_SET_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);
    }
    MLC_SET_BIT(TG_MLC_SER_EEPROM_DATA_OUT);
    SNOOZE(ALT_SER_EEPROM_TLOW);
    MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
    SNOOZE(ALT_SER_EEPROM_THIGH);
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
    SER_EEPROM_STOP;
    return(val);
}

/* Writes a byte to tigon serial eeprom at byte address 'addr'.  Can only
 * access serial eeprom at device-address 0 and this part can have a density
 * of at most 64k bytes.  This routine implements a "random write" function
 * See Atmel AT24C32/64 product spec for more information.
 *
 *  The format of the random write is as follows...
 *
 *      S1010ddd0A***RRRRRARRRRRRRRADDDDDDDDAS
 *      t        c        c        c        ct 
 *      a        k        k        k        ko 
 *      r                                    p 
 *      t
 *
 *  Where...    d == device address
 *              R == byte address
 *              D == data written to eeprom
 */
 
STATIC bool
#if defined(__STDC__) || defined(ALT_PROTOS)
put_serial_byte(struct tg_nic *nic, U32 addr, U8 data)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
put_serial_byte(nic, addr, data)
struct tg_nic *nic;
U32 addr;
U8 data;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    SER_EEPROM_START;
    ser_eeprom_prep(nic, SER_EEPROM_WRITE_SELECT);
    if (!ser_eeprom_chk_ack(nic))
        return(FALSE);
    ser_eeprom_prep(nic, (U8)(addr >> 8));
    if (!ser_eeprom_chk_ack(nic))
        return(FALSE);
    ser_eeprom_prep(nic, (U8)(addr & 0xff));
    if (!ser_eeprom_chk_ack(nic))
        return(FALSE);
    ser_eeprom_prep(nic, data);
    if (!ser_eeprom_chk_ack(nic))
        return(FALSE);
    SER_EEPROM_STOP;
    SNOOZE(10000);
    return(TRUE);
}

/* TRUE = acknowledge seen
 * FALSE = acknowledge not seen
 */
STATIC bool
#if defined(__STDC__) || defined(ALT_PROTOS)
ser_eeprom_chk_ack(struct tg_nic *nic)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
ser_eeprom_chk_ack(nic)
struct tg_nic *nic;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    int save;
    
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);
    SNOOZE(ALT_SER_EEPROM_TLOW);
    MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
    SNOOZE(ALT_SER_EEPROM_H_THIGH);
    /* sample data in middle of high clk */
    save = IS_SER_EEPROM_BIT_SET;
    SNOOZE(ALT_SER_EEPROM_H_THIGH);
    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
    return(!save);
}

STATIC void
#if defined(__STDC__) || defined(ALT_PROTOS)
ser_eeprom_prep(struct tg_nic *nic, U8 pattern)
#else /* defined(__STDC__) || defined(ALT_PROTOS) */
ser_eeprom_prep(nic, pattern)
struct tg_nic *nic;
U8 pattern;
#endif /* defined(__STDC__) || defined(ALT_PROTOS) */
{
    U8 i;

    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_DATA_OUT);
    MLC_SET_BIT(TG_MLC_SER_EEPROM_WRITE_ENA);
    for (i=0; i<8; i++, pattern <<= 1) {
	SNOOZE(ALT_SER_EEPROM_H_TLOW);
	if (pattern & 0x80) {
	    MLC_SET_BIT(TG_MLC_SER_EEPROM_DATA_OUT);
        } else {
	    MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_DATA_OUT);
        }
	SNOOZE(ALT_SER_EEPROM_H_TLOW);
	MLC_SET_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
	SNOOZE(ALT_SER_EEPROM_THIGH);
	MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_CLK_OUT);
	MLC_CLEAR_BIT(TG_MLC_SER_EEPROM_DATA_OUT);
    }
}
