 /*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: alttypes.h,v $
 * Revision 1.1.5.94  1999/02/11  23:18:42  leonid
 * 	MAX_BUFDS_PER_FRAME is changed from 32 to 64 to address IBM SpecWEB failure
 * 	[1999/02/11  23:18:25  leonid]
 *
 * Revision 1.1.5.93  1999/02/02  02:17:19  dfoy
 * 	Changed PREP_FLY_DMA for ODI to support virtual memory.
 * 	[1999/02/02  02:12:43  dfoy]
 * 
 * Revision 1.1.5.92  1999/01/30  02:32:19  fhwang
 * 	add struct snap_ether_header & struct tagged_snap_ether_header definition for llc support.
 * 	[1999/01/30  02:21:10  fhwang]
 * 
 * Revision 1.1.5.91  1999/01/26  21:27:40  leonid
 * 	Three memory leaks have been fixed
 * 	[1999/01/26  21:26:33  leonid]
 * 
 * Revision 1.1.5.90  1999/01/13  21:21:11  ted
 * 	Turn off NT firmware debugger support by default
 * 	[1999/01/13  21:20:05  ted]
 * 
 * Revision 1.1.5.89  1999/01/07  23:09:20  ted
 * 	Add built for debugger flag and move deserialize flags outside of ndis4 vs. 5 ifdef
 * 	[1999/01/07  23:05:55  ted]
 * 
 * Revision 1.1.5.88  1999/01/06  00:28:58  leonid
 * 	1.1.5.87 Log is corrected
 * 	[1999/01/06  00:28:29  leonid]
 * 
 * Revision 1.1.5.87  1999/01/06  00:10:12  leonid
 * 	In NT world, #ifdef (DBG) is replaced with #if (DBG)
 * 	[1999/01/06  00:09:11  leonid]
 * 
 * Revision 1.1.5.86  1998/12/08  17:04:56  leonid
 * 	Deserialization and internal queueing is enabled for the NDIS 4
 * 	[1998/12/08  17:04:35  leonid]
 * 
 * Revision 1.1.5.85  1998/12/07  23:33:33  ted
 * 	Move NT to 512 entry ring
 * 	[1998/12/07  23:29:49  ted]
 * 
 * Revision 1.1.5.84  1998/11/30  18:48:56  hayes
 * 	mini rings work for solaris/12.3 testing
 * 	[1998/11/30  18:46:43  hayes]
 * 
 * Revision 1.1.5.83  1998/11/23  22:50:28  ted
 * 	Fix ntohl, ntohs sign extension problems, fix decrement error when no bd gotten
 * 	[1998/11/23  22:49:35  ted]
 * 
 * Revision 1.1.5.82  1998/11/10  21:28:08  ted
 * 	support jumbos on x86 Solaris (test extended recv BD)
 * 	[1998/11/10  21:27:55  ted]
 * 
 * Revision 1.1.5.81  1998/11/04  19:38:08  ted
 * 	Split altdrv.c into altcfg.c, altcinit.c, and altintr.c
 * 	[1998/11/04  19:36:59  ted]
 * 
 * Revision 1.1.5.80  1998/11/04  17:42:38  hayes
 * 	Add support for Alpha systems
 * 	Fix semantics for ATOMIC_XXX macros.
 * 	[1998/11/04  17:40:28  hayes]
 * 
 * Revision 1.1.5.79  1998/11/03  23:25:26  ted
 * 	Fix up ATOMIC_XXX macros, say out of NT resources for other than standard receive ring
 * 	[1998/11/03  22:20:52  ted]
 * 
 * Revision 1.1.5.78  1998/10/31  02:12:59  ted
 * 	Fix bad ULA packets, x86 solaris
 * 	[1998/10/31  02:12:21  ted]
 * 
 * Revision 1.1.5.77  1998/10/13  03:33:33  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:04  ted]
 * 
 * Revision 1.1.5.76  1998/09/29  21:57:19  ted
 * 	Fix data corruption problem and hybrid mode crashes
 * 	[1998/09/29  21:39:22  ted]
 * 
 * Revision 1.1.5.75  1998/09/29  18:50:25  leonid
 * 	NdisMIndicateStatusComplete() is added for each NdisMIndicateStatus()
 * 	[1998/09/29  18:50:04  leonid]
 * 
 * Revision 1.1.5.74  1998/09/03  22:16:50  hayes
 * 	add bugfixes from nt_r1_9_5
 * 	[1998/09/03  22:13:17  hayes]
 * 
 * Revision 1.1.5.73  1998/09/03  20:05:00  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:09  ted]
 * 
 * Revision 1.1.5.72  1998/09/02  19:02:51  leonid
 * 	for NT, all panic() are replaced with ERROR1
 * 	[1998/09/02  19:02:35  leonid]
 * 
 * Revision 1.1.5.71  1998/09/02  15:20:27  leonid
 * 	printf() is replaced with panic()
 * 	[1998/09/02  15:20:06  leonid]
 * 
 * Revision 1.1.5.70  1998/08/27  22:52:24  leonid
 * 	panic() is redefined as #if defined (DBG) only to avoid Microsoft certification test failure on free build
 * 	[1998/08/27  22:51:58  leonid]
 * 
 * Revision 1.1.5.69  1998/08/20  01:15:14  hayes
 * 	fix sloppy fix to ether_copy macro
 * 	[1998/08/20  01:14:26  hayes]
 * 
 * 	don't touch MDL structure before calling NdisFlushBuffer
 * 	[1998/08/20  01:01:22  hayes]
 * 
 * Revision 1.1.5.68  1998/08/18  18:07:37  ted
 * 	Correct sloppy ether_cmp, ether_copy macros
 * 	[1998/08/18  18:07:24  ted]
 * 
 * Revision 1.1.5.67  1998/08/12  01:58:03  ted
 * 	Deserialize maybe working
 * 	[1998/08/12  01:55:37  ted]
 * 
 * 	bci by hand because bmerge ran out of open file descriptors, sigh
 * 	[1998/08/10  22:47:57  ted]
 * 
 * 	Refine send deserialization
 * 	[1998/08/10  21:28:01  ted]
 * 
 * 	First shot at real deserialization
 * 	[1998/08/06  18:10:56  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:01:56  ted]
 * 
 * Revision 1.1.5.66  1998/08/07  21:28:58  hayes
 * 	require NIC send ring for Tigon1 compatibility
 * 	[1998/08/07  21:28:45  hayes]
 * 
 * Revision 1.1.5.65  1998/08/06  17:49:12  hayes
 * 	fix out of PTES error on NT
 * 	[1998/08/06  17:48:57  hayes]
 * 
 * Revision 1.1.5.64  1998/08/03  22:06:29  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:11  hayes]
 * 
 * Revision 1.1.5.63  1998/08/03  17:11:28  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:42  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:58  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:26  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:47  hayes]
 * 
 * Revision 1.1.5.62  1998/07/21  21:54:08  ted
 * 	Fix possible byte read/writes to shared memory
 * 	[1998/07/21  21:53:55  ted]
 * 
 * Revision 1.1.5.66  1998/08/07  21:28:58  hayes
 * 	require NIC send ring for Tigon1 compatibility
 * 	[1998/08/07  21:28:45  hayes]
 * 
 * Revision 1.1.5.65  1998/08/06  17:49:12  hayes
 * 	fix out of PTES error on NT
 * 	[1998/08/06  17:48:57  hayes]
 * 
 * Revision 1.1.5.64  1998/08/03  22:06:29  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:11  hayes]
 * 
 * Revision 1.1.5.63  1998/08/03  17:11:28  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:42  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:58  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:26  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:47  hayes]
 * 
 * Revision 1.1.5.62  1998/07/21  21:54:08  ted
 * 	Fix possible byte read/writes to shared memory
 * 	[1998/07/21  21:53:55  ted]
 * 
 * Revision 1.1.5.66  1998/08/07  21:28:58  hayes
 * 	require NIC send ring for Tigon1 compatibility
 * 	[1998/08/07  21:28:45  hayes]
 * 
 * Revision 1.1.5.65  1998/08/06  17:49:12  hayes
 * 	fix out of PTES error on NT
 * 	[1998/08/06  17:48:57  hayes]
 * 
 * Revision 1.1.5.64  1998/08/03  22:06:29  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:11  hayes]
 * 
 * Revision 1.1.5.63  1998/08/03  17:11:28  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:42  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:15:58  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:26  hayes]
 * 
 * 	host based send ring works
 * 	[1998/07/28  03:57:47  hayes]
 * 
 * Revision 1.1.5.62  1998/07/21  21:54:08  ted
 * 	Fix possible byte read/writes to shared memory
 * 	[1998/07/21  21:53:55  ted]
 * 
 * Revision 1.1.5.61  1998/07/14  23:27:02  ted
 * 	Put back SUSER macro for Solaris
 * 	[1998/07/14  23:26:09  ted]
 * 
 * Revision 1.1.5.60  1998/07/14  19:01:23  ted
 * 	Clean up macro parm usage
 * 	[1998/07/14  18:55:57  ted]
 * 
 * Revision 1.1.5.59  1998/07/14  18:34:51  hayes
 * 	cleanup locks
 * 	[1998/07/14  18:14:51  hayes]
 * 
 * Revision 1.1.5.58  1998/07/14  02:32:40  ted
 * 	Deserialization, NDIS 5, and white space
 * 	[1998/07/14  02:13:44  ted]
 * 
 * Revision 1.1.5.57  1998/06/18  00:06:34  hayes
 * 	add caching fixes
 * 	[1998/06/17  23:57:59  hayes]
 * 
 * Revision 1.1.5.56  1998/06/16  16:45:52  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:43:03  ted]
 * 
 * Revision 1.1.5.55  1998/06/12  23:47:24  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:12  ted]
 * 
 * 	snapshot of merge with new Registry handling
 * 	[1998/06/10  22:15:20  ted]
 * 
 * Revision 1.1.5.54  1998/06/08  23:55:39  ted
 * 	Cleanup tracing messages and identifiers
 * 	[1998/06/08  23:39:09  ted]
 * 
 * 	Redo ODI generification
 * 	[1998/06/08  22:59:23  ted]
 * 
 * 	Generify RingDequeueMsgs
 * 	[1998/06/04  23:40:15  ted]
 * 
 * 	Snapshot NT generification phase 1
 * 	[1998/06/04  19:52:21  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:23  ted]
 * 
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:08:41  ted]
 * 
 * 	snapshot 12.0 NT jumbo support
 * 	[1998/05/20  20:29:29  ted]
 * 
 * Revision 1.1.5.53  1998/05/01  22:15:39  hayes
 * 	replace alt_memcpy() with NdisMoveMemory() for MEMCPY()
 * 	[1998/05/01  22:15:16  hayes]
 * 
 * Revision 1.1.5.52  1998/05/01  05:35:11  hayes
 * 	alpha NT works with transparent shmem accesses
 * 	[1998/05/01  05:34:43  hayes]
 * 
 * 	checkpoint new ALPHA NT initialization and memory handling
 * 	[1998/04/29  23:32:19  hayes]
 * 
 * Revision 1.1.5.51  1998/04/07  06:35:53  ted
 * 	bring ODI driver fixes from 1.8 release forward to ge_latest
 * 	[1998/04/07  06:34:15  ted]
 * 
 * Revision 1.1.5.50  1998/04/02  06:03:10  ted
 * 	Add checksum debugging support
 * 	[1998/04/02  05:55:42  ted]
 * 
 * Revision 1.1.5.49  1998/03/22  21:42:30  venkat
 * 	modified ALT_SWAP_MEMCPY_TODEV macro
 * 	[1998/03/22  21:38:54  venkat]
 * 
 * Revision 1.1.5.48  1998/03/09  21:37:03  venkat
 * 	introduced shared memory access MACROs to let individual platforms define
 * 	them to be system specific
 * 	[1998/03/09  21:32:51  venkat]
 * 
 * Revision 1.1.5.47  1998/03/04  18:58:49  ted
 * 	Fix typo
 * 	[1998/03/04  18:56:48  ted]
 * 
 * Revision 1.1.5.46  1998/02/21  23:20:57  venkat
 * 	remove use of incorrect constant and introduce ETHERMTU
 * 	[1998/02/21  23:19:12  venkat]
 * 
 * Revision 1.1.5.45  1998/02/05  20:56:15  venkat
 * 	new SNOOZE algorithm
 * 	[1998/02/05  20:43:28  venkat]
 * 
 * Revision 1.1.5.44  1998/02/04  00:53:24  venkat
 * 	Remove NdisStallExecution from SNOOZE()
 * 	[1998/02/04  00:49:43  venkat]
 * 
 * Revision 1.1.5.43  1998/01/29  18:40:35  ted
 * 	add address truncation macros to avoid alignment problems on randomly aligned data traces
 * 	[1998/01/29  18:33:38  ted]
 * 
 * Revision 1.1.5.42  1998/01/24  04:06:53  venkat
 * 	new defines for SEND_RING_PTR,SEND_RING_ENTRIES and SEND_RING_BDS_IN_WINDOW
 * 	[1998/01/24  04:03:42  venkat]
 * 
 * Revision 1.1.5.41  1998/01/16  23:01:57  venkat
 * 	performance opt. in critical path
 * 	[1998/01/16  23:01:15  venkat]
 * 
 * Revision 1.1.5.40  1998/01/07  18:59:12  venkat
 * 	modify PREP_DMA and UNPREP_DMA and PREP_SEND_FLY_INFOP macros for WINNT
 * 	[1998/01/07  18:58:23  venkat]
 * 
 * Revision 1.1.5.39  1998/01/06  23:56:43  venkat
 * 	optimized Tx macros
 * 	[1998/01/06  23:55:09  venkat]
 * 
 * Revision 1.1.5.38  1998/01/05  01:31:56  ted
 * 	Put DEBUG back to 0 so we don't ASSERT for default case
 * 	[1998/01/05  01:04:47  ted]
 * 
 * Revision 1.1.5.37  1998/01/03  23:35:38  venkat
 * 	added LINK_STATE_INDICATION macro
 * 	[1998/01/03  23:33:51  venkat]
 * 
 * Revision 1.1.5.36  1998/01/03  00:40:39  venkat
 * 	modified WINNT specific macros to handle NON-copydata mode
 * 	[1998/01/03  00:35:45  venkat]
 * 
 * Revision 1.1.5.35  1997/12/29  02:25:18  venkat
 * 	Merged with changes from 1.1.5.34
 * 	[1997/12/29  02:24:57  venkat]
 * 
 * 	Fixed receive processing bug in RingDequeueMsgs
 * 	[1997/12/29  02:20:40  venkat]
 * 
 * 	changed the WINNT specific macros that manipulate INFOP
 * 	[1997/12/27  20:37:08  venkat]
 * 
 * Revision 1.1.5.34  1997/12/28  06:31:44  venkat
 * 	changed the WINNT specific macros that manipulate INFOP
 * 	[1997/12/27  20:37:08  venkat]
 * 
 * Revision 1.1.5.33  1997/12/18  23:21:32  venkat
 * 	merged from ge_0_37_latest
 * 	[1997/12/18  23:18:17  venkat]
 * 
 * Revision 1.1.5.32  1997/12/03  01:23:05  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:28  ted]
 * 
 * Revision 1.1.5.31  1997/11/26  19:18:51  ted
 * 	Change Macros to work with non-copydata mode (tested with Solaris, non-fastDMA)
 * 	[1997/11/26  19:18:03  ted]
 * 
 * Revision 1.1.5.30  1997/11/19  00:14:04  ted
 * 	merged from release 2.0
 * 	[1997/11/19  00:11:21  ted]
 * 
 * Revision 1.1.5.29  1997/11/14  19:17:16  hayes
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:16  hayes]
 * 
 * Revision 1.1.5.28  1997/10/21  00:13:04  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  23:00:25  ted]
 * 
 * Revision 1.1.43.1  1997/10/16  17:30:54  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:25:02  ted]
 * 
 * Revision 1.1.5.27  1997/09/16  03:11:38  ted
 * 	First pass on regenerification for initialization and removal of driver, NT and ODI driver work is very much incomplete, chanage s to spl for better code
 * 	[1997/09/16  03:03:30  ted]
 * 
 * Revision 1.1.5.26  1997/09/11  18:06:26  ted
 * 	Mixing vlans, jumbos, and redundancy works, plus small fixes from release 1.1 and better NIC tracing
 * 	[1997/09/11  18:01:00  ted]
 * 
 * Revision 1.1.5.25  1997/08/23  17:40:32  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:38:54  skapur]
 * 
 * Revision 1.1.5.24  1997/07/28  22:40:34  ted
 * 	Merge
 * 	[1997/07/28  22:38:00  ted]
 * 
 * 	Solaris dual homing support
 * 	[1997/07/28  21:39:58  ted]
 * 
 * Revision 1.1.5.23  1997/07/25  22:59:44  ted
 * 	Merged with changes from 1.1.5.22
 * 	[1997/07/25  22:58:01  ted]
 * 
 * 	New #define for receive buffer offset (ALT_BUF_OFFSET)
 * 	New #define for send DMA type (ALT_DMA_SEND)
 * 	Claim we're a PSO style machine even though Sun doesn't make any of
 * 	  these so that when we call PREP_CACHE_<whatever> it will work on new
 * 	  architecture machines (per Denny Gentry).
 * 	Use desballoc() instead of esballoc() to reduce mutex badness and improve
 * 	  performance.
 * 	Misc reformatting for prettier code.
 * 	[1997/07/07  22:59:49  ted]
 * 
 * Revision 1.1.5.22  1997/07/09  00:24:57  akhan
 * 	POST_PROCESSING routine changed so that the packet has NDIS_STATUS_SUCCESS.
 * 	Added code to do receive collasing.
 * 	[1997/07/09  00:18:17  akhan]
 * 
 * Revision 1.1.5.21  1997/07/02  20:41:19  suresh
 * 	AIX : call DMA_MEMFREE on failure
 * 	[1997/07/02  20:40:58  suresh]
 * 
 * Revision 1.1.5.20  1997/06/17  00:36:45  ted
 * 	Make ALIGN_DMA_ADDR macro null on Network Appliance since they don't store
 * 	a special DMA address in the  alt_dma_info structure
 * 	[1997/06/16  17:45:03  ted]
 * 
 * 	Add generic macro ALIGN_DMA_ADDRESS to allow for different mechanisms under
 * 	different OS.
 * 	[1997/06/13  17:23:04  ted]
 * 
 * 	tighten up driver code for better performance
 * 	[1997/06/06  22:38:31  ted]
 * 
 * Revision 1.1.5.19  1997/05/23  00:16:17  akhan
 * 	No code changes.
 * 	[1997/05/23  00:16:01  akhan]
 * 
 * 	Added Code for NDIS3 and NDIS4 compilation of the driver.
 * 	[1997/05/22  23:12:10  akhan]
 * 
 * Revision 1.1.5.18  1997/05/01  22:36:24  akhan
 * 	Remove ALTEON_MAJOR_VERSION and ALTEON_MINOR_VERSION into a different file.
 * 	[1997/05/01  22:36:07  akhan]
 * 
 * Revision 1.1.5.16  1997/04/26  01:09:29  akhan
 * 	CopyData is set to 1.
 * 	[1997/04/26  01:06:16  akhan]
 * 
 * Revision 1.1.5.15  1997/04/12  00:11:32  ted
 * 	Merged with changes from 1.1.5.14
 * 	[1997/04/12  00:10:54  ted]
 * 
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:07:47  ted]
 * 
 * 	Working SBUS implementation prior to MP optimization
 * 	[1997/04/06  22:09:35  ted]
 * 
 * Revision 1.1.5.14  1997/04/11  23:12:23  suresh
 * 	AIX - toggle ON copydata mode on sends
 * 	[1997/04/11  23:08:17  suresh]
 * 
 * Revision 1.1.5.13  1997/04/04  23:20:38  suresh
 * 	AIX fix PREP_DMA for 2 bus iovecs
 * 	[1997/04/04  23:20:23  suresh]
 * 
 * Revision 1.1.5.12  1997/04/04  22:38:30  ted
 * 	Merged with changes from 1.1.5.11
 * 	[1997/04/04  22:37:56  ted]
 * 
 * 	Remove kernel printf that can legitimately show up on busy systems.
 * 	[1997/04/04  22:15:15  ted]
 * 
 * Revision 1.1.5.11  1997/04/04  06:22:35  akhan
 * 	code merge for NT.
 * 	[1997/04/02  18:19:57  akhan]
 * 
 * Revision 1.1.5.10  1997/03/28  01:39:53  ted
 * 	Major merge of NDIS and ODI with mainstream Unix ports
 * 	[1997/03/28  01:33:29  ted]
 * 
 * Revision 1.1.5.9  1997/03/24  08:34:40  ted
 * 	Merged with changes from 1.1.5.8
 * 	[1997/03/24  08:33:53  ted]
 * 
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:35  ted]
 * 
 * Revision 1.1.5.8  1997/03/19  06:14:04  suresh
 * 	PREP_DMA changes - AIX
 * 	[1997/03/19  06:13:51  suresh]
 * 
 * Revision 1.1.5.7  1997/03/03  05:17:15  ted
 * 	endian fixes, non-copydata mode fixes
 * 	[1997/03/03  05:16:28  ted]
 * 
 * Revision 1.1.5.6  1997/02/21  02:06:44  ted
 * 	Merged with changes from 1.1.5.5
 * 	[1997/02/21  02:06:23  ted]
 * 
 * 	Prepare data structures and code structures to support redundant operation
 * 	[1997/02/21  01:56:53  ted]
 * 
 * Revision 1.1.5.5  1997/02/20  21:43:05  suresh
 * 	FIX COPYIN and COPYOUT macros..
 * 	[1997/02/20  21:42:05  suresh]
 * 
 * Revision 1.1.5.4  1997/02/18  04:52:24  suresh
 * 	add LOCK_INIT and LOCK_DESTROY macros for AIX
 * 	[1997/02/18  04:50:32  suresh]
 * 
 * Revision 1.1.5.3  1997/02/15  04:41:18  ted
 * 	DMA_MEMALLOCED regions are now in cache-consistent memory on Solaris
 * 	systems.  Noop the cache-sync routines on this type of memory.
 * 
 * 	Add generic LOCK macros
 * 	[1997/02/15  03:06:43  ted]
 * 
 * Revision 1.1.5.2  1997/02/07  23:23:11  ted
 * 	Get AIX to compile
 * 	[1997/02/07  23:21:48  ted]
 * 
 * Revision 1.1.5.1  1997/02/07  21:58:08  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:48:04  ted]
 * 
 * Revision 1.1.2.1  1996/11/21  18:01:42  ted
 * 	Moved from ../unix/common and merged in WINNT changes
 * 	[1996/11/21  17:43:42  ted]
 * 
 * Revision 1.5.2.2  1996/11/20  20:04:45  ted
 * 	no changes
 * 	[1996/11/20  19:57:05  ted]
 * 
 * Revision 1.5.2.1  1996/11/01  18:25:38  ted
 * 	change rr_ to tg_ and add proper copyright
 * 	[1996/10/30  22:58:48  ted]
 * 
 * $EndLog$
 */
/*
 * FILE alttypes.h
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/nic/host/common/alttypes.h,v $
 * $Revision: 1.1.5.94 $ $Locker:  $
 * $Date: 1999/02/11 23:18:42 $ $Author: leonid $ $State: Exp $
 */

#ifndef _ALTTYPES_H_
#define _ALTTYPES_H_

/* common stuff, more common stuff at the end after system specific */
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif /* TRUE */

/* for pass/fail routines that aren't boolean */
#ifndef FAILURE
#define FAILURE 1
#define PASSED 0
#endif /* FAILURE */

/* configuration constants */
#define H_DEBUG 0
#if H_DEBUG
#define DEBUG 1			/* for ASSERT */
#else /* H_DEBUG */
#define DEBUG 0			/* for removing ASSERT */
#endif /* H_DEBUG */

#if !defined(WINNT) && !defined(__NETWARE_386__)
#include <sys/debug.h>
#endif /* !defined(WINNT) && !defined(__NETWARE_386__) */

/* definitive constants */
#define JUMBOBLOCKSIZE 9018	/* jumbo size, space for mac hdr */
#define MAXBLOCKSIZE 1518	/* ethernet for now, space for mac hdr */
#define MINMTUSIZE 60		/* minimum allowable ethernet MTU */
#define USECS_PER_SEC 1000000	/* microseconds in a second */
#define MAX_BUFDS_PER_FRAME 64	/* max # buffer descriptors per send frame */

/* -------------------------------- SunOS (Solaris 2) ---------------- */
#if defined(__sun)
/* transmit ring parameters */
#define COMMAND_RING_PTR        TG_COMMAND_RING
#define SEND_RING_ENTRIES	128
#define SEND_RING_PTR		0x3800
#define SEND_RING_PHYS_ADDR     SEND_RING_PTR
#define SEND_RING_BDS_IN_WINDOW ((TG_MEM_WINDOW_SIZE/(sizeof(struct send_bd)))-1)
#define HOST_SEND_RING 0	/* bus is a scarce resource on these systems */

/* configuration parms */
#if defined(__sparc)
#define FAST_DMA 0		/* use dvma_xxxx routines, should be faster */
#define MAX_HYBRID_SIZE JUMBOBLOCKSIZE

#define HYBRID_LEN JUMBOBLOCKSIZE /* cross-over size */

#define USE_EXTENDED_JUMBO 0	/* don't use extended jumbo bds */

#else /* defined(__sparc) */

#define FAST_DMA 0		/* never possible with x86 solaris */
#define MAX_HYBRID_SIZE 2048
#define HYBRID_LEN 2048		/* cross-over size */
#define USE_EXTENDED_JUMBO 1	/* use extended jumbo bds */

#endif /* defined(__sparc) */

#define ALT_RECV_EXT 1		/* receive on prelocked buffers */
#define ALT_LOCKS_SUPPORTED 1	/* MP locks available */
#define REAP_FROM_SEND 0	/* don't get send bds in send pkt routine */
#define FLOW_CONTROL_SUPPORTED 1 /* system flow control available */
#define USE_DMA_MEMALLOC 0	/* use DMA_MEMALLOC for static buffers? */
#define READ_SERVICE 0		/* use read service routine for flow cntl */
#define ALT_DESERIALIZE 0	/* some day we should XXX */
#define USE_INTERNAL_QUEUE 0	/* Streams queues work well */
#define USE_MINI_RING 0		/* just for testing for now */
#define MINIBLOCKSIZE 128	/* enough to hold an ack plus some */

#define RECVBLOCKSIZE 1536 /* enet for now, space for mac hdr, cache line */
#define SENDBLOCKSIZE 1536 /* enet for now, space for mac hdr, cache line */
#define ALT_BUF_OFFSET 10	/* offset into the preallocated recv buf */

/* special section for adding in checksum offload debugging code */
#define RECV_CKSUM_DEBUG    0
#define SEND_CKSUM_DEBUG    0
#define DONT_INIT_JUMBO     0

#if SEND_CKSUM_DEBUG || RECV_CKSUM_DEBUG
#define USE_CKSUM_OFFLOAD 1
#define DO_CKSUM_DEBUG 1	/* for debugging checksum offload */
#else /* SEND_CKSUM_DEBUG || RECV_CKSUM_DEBUG */
#define USE_CKSUM_OFFLOAD 0
#define DO_CKSUM_DEBUG 0	/* for debugging checksum offload */
#endif /* SEND_CKSUM_DEBUG || RECV_CKSUM_DEBUG */

#if SEND_CKSUM_DEBUG
#define PREP_SEND_CKSUM_PKT(flags, pkt, msg) flags = prepSendCksumPkt(msg)
#else /* SEND_CKSUM_DEBUG */
#define PREP_SEND_CKSUM_PKT(flags, pkt, msg)
#endif /* SEND_CKSUM_DEBUG */

#if RECV_CKSUM_DEBUG
#define VERIFY_RECV_CKSUM(nic, infop, bdp, pkt) \
    verifyRecvCksum(nic, infop, bdp)
#else /* RECV_CKSUM_DEBUG */
#define VERIFY_RECV_CKSUM(nic, infop, bdp, pkt)
#endif /* RECV_CKSUM_DEBUG */

#ifndef bool
#define bool unsigned char
#endif /* bool */

/* atomic operations... */
/*
 * XXX replace with more efficient equivalents...
 */
#define ATOMIC_INCREMENT(spl, lock, x) 		\
{						\
    ALT_DISABLE(spl, lock);			\
    (x)++;					\
    ALT_ENABLE(spl, lock);			\
}

#define ATOMIC_DECREMENT(spl, lock, x) 		\
{						\
    ALT_DISABLE(spl, lock);			\
    (x)--;					\
    ALT_ENABLE(spl, lock);			\
}

#define ATOMIC_ADD(spl, lock, target, value) 	\
{						\
    ALT_DISABLE(spl, lock);			\
    target += value;				\
    ALT_ENABLE(spl, lock);			\
}

#define ATOMIC_SET(spl, lock, target, value) 	\
{						\
    ALT_DISABLE(spl, lock);			\
    target = value;				\
    ALT_ENABLE(spl, lock);			\
}

#ifndef U32
#define U64 unsigned long long
#define U32 unsigned long
#define U16 unsigned short
#define U8  unsigned char
#define S64 signed long long
#define S32 signed long
#define S16 signed short
#define S8  signed char
#endif /* U32 */

#define VOLATILE volatile	/* works on Solaris */

#define PTRS_ARE_64 0

#define DIP_TYPE void		/* device information struct type */

/* little or big endian */
#if !defined(ALT_BIG_ENDIAN)	/* gdb uses this too */
#if defined(__sparc)
#define ALT_BIG_ENDIAN 1
#else /* defined(__sparc) */
#define ALT_BIG_ENDIAN 0
#endif /* defined(__sparc) */
#endif /* !defined(ALT_BIG_ENDIAN) */

/* driver function names */
#define ALTOPEN altopen
#define ALTIOCTL alt_ioctl
#define ALTCLOSE altclose
#define ALTINTR altintr
#define ALTDOINTR altdointr
#define ALTINTRWORK altdointr
#define ALTSTART altstart

/* various things that may be renamed from system to system */
#define RDSHMEM(r) r
#define WRSHMEM(w, v) (w) = (v)
#define WRSHMEM64(w, v) (w) = (v)
#define ALT_MEMCPY_TOHOST(to, from, len) alt_memcpy(to, from, len)
#define ALT_MEMCPY_TODEV(to, from, len) alt_memcpy(to, from, len)
#define ALT_SWAP_MEMCPY_TOHOST(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
#define ALT_SWAP_MEMCPY_TODEV(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define BZERO_TODEV(s1, count)			\
{						\
    U32 locCount = (count)/4;			\
    U32 *locPtr = (U32 *)(s1);			\
						\
    while (locCount--) {			\
	WRSHMEM(*locPtr, 0);			\
	locPtr++;				\
    }						\
}

#define MEMCPY_TYPE caddr_t
#define MEMCPY_SZT size_t
#define MEMCPY(to, from, count) bcopy(from, to, count)
#define MEMCMP(s1, s2, count) bcmp(s1, s2, count)
#define BZERO(s1, count) bzero(s1, count)
#define MIN min
#define MAX max

/* send reenable macro following flow control condition */
#define REENABLE_SEND(nic) altenablewq(nic->altp)

/* use these as memory flags */
#define ALT_SLEEP      KM_SLEEP
#define ALT_NOSLEEP    KM_NOSLEEP
#define ALT_PHYSCONTIG use DMA_MEMALLOC please
#define ALT_CACHEALIGN use DMA_MEMALLOC please
#define ALT_DMA_CONSISTENT  DDI_DMA_CONSISTENT
#define ALT_DMA_STREAMING   DDI_DMA_STREAMING
#define ALT_DMA_SEND				\
    ((nic->nic_type == ALT_SBUS_NIC) ?		\
     DDI_DMA_STREAMING : DDI_DMA_CONSISTENT)

#define MEMALLOC(where, type, size, flags) \
    (where) = ((type)kmem_alloc(size, flags))
#define MEMZALLOC(where, type, size, flags) \
    (where) = ((type)kmem_zalloc(size, flags))
#define MEMFREE(ptr, size, flags) kmem_free(ptr, size)

#define DMA_ADDR_TYPE caddr_t
/* use this for structures that could be used for DMA by the NIC */
#define DMA_MEMALLOC(where, type, size, flags, rw, infop)		    \
do {									    \
    caddr_t kernAddr;							    \
    uint_t realLen, ccount;						    \
    ddi_dma_cookie_t cookie;						    \
    ddi_device_acc_attr_t access_attr = {				    \
	DDI_DEVICE_ATTR_V0,						    \
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */	    \
	DDI_STRICTORDER_ACC						    \
    };									    \
    int rc;								    \
									    \
    where = (type)0;		/*  in case of failure */		    \
    if (ddi_dma_alloc_handle(altp->dip, &nic->dmaAttr, DDI_DMA_DONTWAIT,    \
			     (void *)0, &(infop)->handle) != DDI_SUCCESS) { \
	ERROR3("%s%d: Couldn't allocate a DMA handle\n",		    \
	       nic->name, altp->instance);				    \
	break;								    \
    }									    \
    if (ddi_dma_mem_alloc((infop)->handle, size, &access_attr,		    \
			  flags, DDI_DMA_DONTWAIT, (void *)0,		    \
			  &kernAddr, &realLen,				    \
			  &(infop)->accHandle) != DDI_SUCCESS) {	    \
	ddi_dma_free_handle(&(infop)->handle);				    \
	ERROR3("%s%d: Couldn't allocate DMA memory\n",			    \
	       nic->name, altp->instance);				    \
	break;								    \
    }									    \
									    \
    if ((rc = ddi_dma_addr_bind_handle((infop)->handle, (struct as *) 0,    \
				 kernAddr, realLen,			    \
				 (rw) | (flags),			    \
				 DDI_DMA_DONTWAIT, (void *) 0,		    \
				 &cookie, &ccount))			    \
	!= DDI_SUCCESS) {						    \
	ddi_dma_mem_free(&(infop)->accHandle);				    \
	ddi_dma_free_handle(&(infop)->handle);				    \
        ERROR4("%s%d: Couldn't bind DMA Handle, rc = %d\n",		    \
	       nic->name, altp->instance, rc);				    \
	break;								    \
    }									    \
    /* raise a ruckus if ccount isn't one for now XXX */		    \
    if (ccount != 1)							    \
        (void) ERROR4("%s%d: DMA_MEMALLOC error ccount isn't 1, instead is %d\n", \
		      nic->name, altp->instance, ccount);		    \
    where = (type)kernAddr;						    \
    (infop)->dataAddrDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address;	    \
} while(0);

/* free an area accessed for DMA */		\
#define DMA_MEMFREE(ptr, size, infop)		\
{						\
    ddi_dma_unbind_handle((infop)->handle);	\
    ddi_dma_mem_free(&(infop)->accHandle);	\
    ddi_dma_free_handle(&(infop)->handle);	\
}

#define CACHE_LINE_SIZE 32

#define ALIGN_ADDR(addr, align, type) \
    (type)(((int)(addr) + ((align) - 1)) & ~((align) - 1))

#define ALIGN_DMA_ADDR(addr, align) \
    (addr) = (DMA_ADDR_TYPE)(((int)(addr) + ((align) - 1)) & ~((align) - 1))

#define ADD_DMA_ADDR(outaddr, inaddr, len) \
    (outaddr) = (U32)((inaddr) + (len))

#define SET_INFOP_ADDR(infop, addr) \
    (infop)->dataAddr = (caddr_t)(addr)

#define TRUNC_ADDR(addr, align, type) (type)((int)(addr) & ~((align) - 1))

/* types for type field below */
#define ALT_FORKERNEL DDI_DMA_SYNC_FORKERNEL
#define ALT_FORDEV    DDI_DMA_SYNC_FORDEV
#define ALT_DONT_FLUSH -1	/* from hme.c, go figure */

/* getting/freeing infops, either deserialized or not */
#if ALT_DESERIALIZE
#define INIT_FREE_INFOP(nic)
#define GET_FREE_INFOP(infop, nic, spl, entry)		\
{							\
    /* ENTER CRITICAL SECTION */			\
    ALT_DISABLE(spl, &nic->infopQueueMu);		\
    infop = nic->freeSmsgs;				\
    if (infop) {					\
        nic->freeSmsgs = infop->next;			\
        nic->smsgsFree--;				\
        (infop)->next = (struct alt_dma_info *)0;	\
    }							\
    ALT_ENABLE(spl, &nic->infopQueueMu);		\
    /* LEAVE CRITICAL SECTION */			\
}
#define GET_NEXT_INFOP(infop, nic, spl, entry) \
    infop = &nic->smsgPtrs[entry];
#define FREE_INFOP(infop, nic, spl)		\
{						\
    /* ENTER CRITICAL SECTION */		\
    ALT_DISABLE(spl, &nic->infopQueueMu);	\
    (infop)->next = nic->freeSmsgs;		\
    nic->freeSmsgs = infop;			\
    nic->smsgsFree++;				\
    ALT_ENABLE(spl, &nic->infopQueueMu);	\
    /* LEAVE CRITICAL SECTION */		\
}
#else /* ALT_DESERIALIZE */
#define INIT_FREE_INFOP(nic)
#define GET_FREE_INFOP(infop, nic, spl, entry) \
    infop = &nic->smsgs[entry];
#define GET_NEXT_INFOP GET_FREE_INFOP
#define FREE_INFOP(infop, nic, spl)
#endif /* ALT_DESERIALIZE */

#if FAST_DMA
/* prepare an area of kernel memory for DMA (bind handle, ...) */
/* rc is 0 if this fails, n for each infop filled in */
#define	PREP_DMA(rc, flags, rw, infop)					\
{									\
    ddi_dma_cookie_t cookie;						\
									\
    dvma_kaddr_load((infop)->handle, (infop)->dataAddr,			\
		    (infop)->dataLen, (infop)->index, &cookie);		\
    (infop)->dataAddrDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address;	\
    rc = 1;								\
}

#define PREP_JUMBO_DMA PREP_DMA
#define PREP_MINI_DMA PREP_DMA

#define	PREP_FLY_DMA(rc, flags, rw, infop)				\
{									\
    ddi_dma_cookie_t cookie;						\
									\
    dvma_kaddr_load((infop)->handle, (infop)->dataAddrFly,		\
		    (infop)->dataLen, (infop)->index, &cookie);		\
    (infop)->dataAddrFlyDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address;	\
    rc = 1;								\
}

#define UNPREP_DMA(infop) \
    dvma_unload((infop)->handle, (infop)->index*2, ALT_DONT_FLUSH);

/* prepare an area of kernel memory for DMA (bind handle, ...) */
#define	SLOW_PREP_DMA(rc, pflags, rw, infop)				      \
{									      \
    uint_t ccount;							      \
    ddi_dma_cookie_t cookie;						      \
    ddi_device_acc_attr_t access_attr = {				      \
	DDI_DEVICE_ATTR_V0,						      \
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */	      \
	DDI_STRICTORDER_ACC						      \
    };									      \
    struct altinst *altp = (infop)->nic->altp;				      \
    int rc2;								      \
    struct alt_dma_info *infop2;					      \
									      \
    rc = 0;								      \
    if ((rc2 = ddi_dma_alloc_handle(altp->dip, &altp->nic->dmaAttr,	      \
				    DDI_DMA_DONTWAIT,			      \
				    (void *)0, &((infop)->handle)))	      \
	== DDI_SUCCESS)							      \
        if ((rc2 = ddi_dma_addr_bind_handle((infop)->handle, (struct as *) 0, \
					    (infop)->dataAddr,		      \
					    (infop)->dataLen,		      \
					    (rw) | (pflags),		      \
					    DDI_DMA_DONTWAIT, (void *) 0,     \
					    &cookie, &ccount))		      \
	    != DDI_SUCCESS) {						      \
	    ALTTRACEE(badbind, 0xf6000, rc2, (infop)->dataLen, 0, 0);	      \
	    ERROR4("%s%d: ddi_dma_addr_bind_handle failed %d\n", 	      \
		   nic->name, altp->instance, rc2);			      \
	    ddi_dma_free_handle(&((infop)->handle));			      \
        } else {							      \
	    rc = ccount;						      \
            (infop)->dataAddrDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address;    \
	    (infop)->dataLen = cookie.dmac_size;			      \
	    /* copy into the next infop */				      \
	    while (ccount > 1) {					      \
		ddi_dma_nextcookie((infop)->handle, &cookie);		      \
		infop2 = infop;						      \
		(infop2)->flags |= ALT_DMA_INFO_FLAG_CONT;		      \
		infop2 = (infop->index == infop->lastIndex) ?		      \
		    infop - infop->index : infop + 1;			      \
	        (infop2)->dataAddrDmaAddr =				      \
                    (DMA_ADDR_TYPE)cookie.dmac_address;			      \
		infop2->handleFly = (infop)->handleFly;                       \
	        (infop2)->dataLen = cookie.dmac_size;			      \
		infop2->flags = 0;					      \
		ccount--;						      \
	    }								      \
	}								      \
    else {								      \
        ALTTRACEE(nohandle, 0xf6001, 0, 0, 0, 0);			      \
        ERROR4("%s%d: ddi_dma_alloc_handle failed %d\n",		      \
	       nic->name, altp->instance, rc2);				      \
    }									      \
}

#define SLOW_UNPREP_DMA(infop)						\
{									\
    int rc;								\
									\
    if ((rc = ddi_dma_unbind_handle((infop)->handle)) != DDI_SUCCESS) {	\
        ALTTRACEE(badunbind, 0xf6002, rc, (infop)->handle, 0, 0);	\
    } else {								\
	ALTTRACEV(unbind, 0x60000, rc, (infop)->handle, 0, 0);		\
    }									\
    ddi_dma_free_handle(&((infop)->handle));				\
}
#else /* FAST_DMA */
/* prepare an area of kernel memory for DMA (bind handle, ...) */
#define	PREP_DMA(rc, pflags, rw, infop)					      \
{									      \
    uint_t ccount;							      \
    ddi_dma_cookie_t cookie;						      \
    ddi_device_acc_attr_t access_attr = {				      \
	DDI_DEVICE_ATTR_V0,						      \
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */	      \
	DDI_STRICTORDER_ACC						      \
    };									      \
    struct altinst *altp = (infop)->nic->altp;				      \
    int rc2;								      \
    struct alt_dma_info *infop2;					      \
									      \
    rc = 0;								      \
    if ((rc2 = ddi_dma_alloc_handle(altp->dip, &altp->nic->dmaAttr,	      \
				    DDI_DMA_DONTWAIT,			      \
				    (void *)0, &((infop)->handle)))	      \
	== DDI_SUCCESS)							      \
        if ((rc2 = ddi_dma_addr_bind_handle((infop)->handle, (struct as *) 0, \
					    (infop)->dataAddr,		      \
					    (infop)->dataLen,		      \
					    (rw) | (pflags),		      \
					    DDI_DMA_DONTWAIT, (void *) 0,     \
					    &cookie, &ccount))		      \
	    != DDI_SUCCESS) {						      \
	    ALTTRACEE(badbind, 0xf6000, rc2, (infop)->dataLen, 0, 0);	      \
	    ERROR4("%s%d: ddi_dma_addr_bind_handle failed %d\n", 	      \
		   nic->name, altp->instance, rc2);			      \
	    ddi_dma_free_handle(&((infop)->handle));			      \
        } else {							      \
	    rc = ccount;						      \
            (infop)->dataAddrDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address;    \
	    (infop)->dataLen = cookie.dmac_size;			      \
	    /* copy into the next infop */				      \
	    while (ccount > 1) {					      \
		ddi_dma_nextcookie((infop)->handle, &cookie);		      \
		infop2 = infop;						      \
		infop2->flags |= ALT_DMA_INFO_FLAG_CONT;		      \
		infop2 = (infop->index == (infop)->lastIndex) ?		      \
		    (infop) - (infop)->index : (infop) + 1;		      \
	        infop2->dataAddrDmaAddr =				      \
                    (DMA_ADDR_TYPE)cookie.dmac_address;			      \
		infop2->handleFly = (infop)->handleFly;                       \
	        infop2->dataLen = cookie.dmac_size;			      \
		infop2->flags = 0;					      \
		ccount--;						      \
	    }								      \
	}								      \
    else {								      \
        ALTTRACEE(nohandle, 0xf6001, 0, 0, 0, 0);			      \
        ERROR4("%s%d: ddi_dma_alloc_handle failed %d\n", 		      \
	       nic->name, altp->instance, rc2);				      \
    }									      \
}

#if USE_EXTENDED_JUMBO
#define	PREP_JUMBO_DMA(rc, pflags, rw, infop)				   \
{									   \
    uint_t ccount;							   \
    ddi_dma_cookie_t cookie;						   \
    ddi_device_acc_attr_t access_attr = {				   \
	DDI_DEVICE_ATTR_V0,						   \
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */	   \
	DDI_STRICTORDER_ACC						   \
    };									   \
    struct altinst *altp = (infop)->nic->altp;				   \
    int rc2;								   \
    DMA_ADDR_TYPE *addrp;                                                  \
    int i;                                                                 \
                                                                           \
    rc = 0;								   \
    if ((rc2 = ddi_dma_alloc_handle(altp->dip, &altp->nic->dmaAttr,	   \
				    DDI_DMA_DONTWAIT,			   \
				    (void *)0, &((infop)->handle)))	   \
	== DDI_SUCCESS) {						   \
        if ((rc2 = ddi_dma_addr_bind_handle((infop)->handle, 		   \
					    (struct as *) 0,		   \
					    (infop)->dataAddr,		   \
					    (infop)->dataLen,		   \
					    (rw) | (pflags),		   \
					    DDI_DMA_DONTWAIT, 		   \
					    (void *) 0,			   \
					    &cookie, &ccount))		   \
	    != DDI_SUCCESS) {						   \
	    ALTTRACEE(badbind, 0xf6000, rc2, (infop)->dataLen, 0, 0);	   \
	    ERROR4("%s%d: ddi_dma_addr_bind_handle failed %d\n",	   \
		   nic->name, altp->instance, rc2);			   \
	    ddi_dma_free_handle(&((infop)->handle));			   \
        } else {							   \
	    rc = ccount;						   \
            (infop)->dataAddrDmaAddr = (DMA_ADDR_TYPE)cookie.dmac_address; \
	    (infop)->dataLen = cookie.dmac_size;			   \
	    /* copy into the next infop */				   \
	    addrp = &infop->extDataAddr1DmaAddr;			   \
	    i = 0;                                                         \
	    while (ccount > 1) {					   \
		ddi_dma_nextcookie((infop)->handle, &cookie);		   \
		infop->flags |= ALT_DMA_INFO_EXT_BD;			   \
	        *addrp = (DMA_ADDR_TYPE)cookie.dmac_address;		   \
	        infop->extDataLen[i] = cookie.dmac_size;		   \
		ccount--;						   \
		addrp++;                                                   \
		i++;                                                       \
	    }								   \
	}								   \
    } else {								   \
        ALTTRACEE(nohandle, 0xf6001, 0, 0, 0, 0);			   \
        ERROR4("%s%d: ddi_dma_alloc_handle failed %d\n",		   \
	       nic->name, altp->instance, rc2);				   \
    }									   \
}
#else /* USE_EXTENDED_JUMBO */

#define PREP_JUMBO_DMA PREP_DMA
#endif /* USE_EXTENDED_JUMBO */

#define PREP_MINI_DMA PREP_DMA

/* XXX this needs redone for ALT_DESERIALIZE */
#define	PREP_FLY_DMA(rc, pflags, rw, infop)				  \
{									  \
    uint_t ccount;							  \
    ddi_dma_cookie_t cookie;						  \
    ddi_device_acc_attr_t access_attr = {				  \
	DDI_DEVICE_ATTR_V0,						  \
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */	  \
	DDI_STRICTORDER_ACC						  \
    };									  \
    struct altinst *altp = (infop)->nic->altp;				  \
    int rc2;								  \
    struct alt_dma_info *infop2, *loc_infop;				  \
									  \
    rc = 0;								  \
    if ((rc2 = ddi_dma_alloc_handle(altp->dip, &altp->nic->dmaAttr,	  \
				    DDI_DMA_DONTWAIT,			  \
				    (void *)0, &((infop)->handleFly)))	  \
	== DDI_SUCCESS)							  \
        if ((rc2 = ddi_dma_addr_bind_handle((infop)->handleFly,		  \
					    (struct as *) 0,		  \
					    (infop)->dataAddrFly,	  \
					    (infop)->dataLen,		  \
					    (rw) | (pflags),		  \
					    DDI_DMA_DONTWAIT, (void *) 0, \
					    &cookie, &ccount))		  \
	    != DDI_SUCCESS) {						  \
	    ALTTRACEE(badbind, 0xf6000, rc2, (infop)->dataLen, 0, 0);	  \
	    ERROR4("%s%d: ddi_dma_addr_bind_handle failed %d\n",	  \
		   nic->name, altp->instance, rc2);			  \
	    ddi_dma_free_handle(&((infop)->handleFly));			  \
        } else {							  \
	    rc = ccount;						  \
            (infop)->dataAddrFlyDmaAddr =				  \
                (DMA_ADDR_TYPE)cookie.dmac_address;			  \
	    (infop)->dataLen = cookie.dmac_size;			  \
	    (infop)->flags = 0;						  \
	    loc_infop = infop;						  \
	    /* copy into the next infop */				  \
	    while (ccount > 1) {					  \
		ddi_dma_nextcookie(loc_infop->handleFly, &cookie);	  \
		infop2 = (loc_infop->index == loc_infop->lastIndex) ?	  \
		    loc_infop - loc_infop->index : loc_infop + 1;	  \
		infop2->flags = ALT_DMA_INFO_FLAG_CONT;			  \
	        infop2->dataAddrFlyDmaAddr =				  \
                    (DMA_ADDR_TYPE)cookie.dmac_address;			  \
		infop2->handleFly = (infop)->handleFly;                   \
	        infop2->dataLen = cookie.dmac_size;			  \
		infop2->INFOP_SPKT = loc_infop->pkt;			  \
		ccount--;						  \
		loc_infop = infop2;					  \
	    }								  \
	}								  \
    else {								  \
        ALTTRACEE(nohandle, 0xf6001, 0, 0, 0, 0);			  \
        ERROR4("%s%d: ddi_dma_alloc_handle failed %d\n",		  \
	       nic->name, altp->instance, rc2);				  \
    }									  \
}

#define UNPREP_DMA(infop)						\
{									\
    int rc;								\
									\
    if ((rc = ddi_dma_unbind_handle((infop)->handle)) != DDI_SUCCESS) {	\
        ALTTRACEE(badunbind, 0xf6002, rc, (infop)->handle, 0, 0);	\
    } else {								\
	ALTTRACEV(unbind, 0x60000, rc, (infop)->handle, 0, 0);		\
    }									\
    ddi_dma_free_handle(&((infop)->handle));				\
}

#define UNPREP_FLY_DMA(infop)						   \
{									   \
    int rc;								   \
									   \
    if ((rc = ddi_dma_unbind_handle((infop)->handleFly)) != DDI_SUCCESS) { \
        ALTTRACEE(badunbind, 0xf6002, rc, (infop)->handleFly, 0, 0);	   \
    } else {								   \
	ALTTRACEV(unbind, 0x60003, rc, (infop)->handleFly, 0, 0);	   \
    }									   \
    ddi_dma_free_handle(&((infop)->handleFly));				   \
}
#endif /* FAST_DMA */

/* dealing with messages we have to attach an external piece of data to
 * a message for receiving */
#define ALT_ATT_EXT_MSG(mp, infop, len)				\
    mp = desballoc((unsigned char *)(infop)->dataAddr, len, 	\
		   BPRI_MED, &(infop)->freeRtn);

/* dealing with a single receive */
#define SINGLE_PACKET_PREP(nic, mpkt, msg, len, infop, pktArr, count, bdp) \
{									   \
    /* check to see if we need to free the dma_info structure */	   \
    if (!((infop)->flags & ALT_DMA_INFO_PREALLOC)) {			   \
        (mpkt) = (infop)->INFOP_RPKT;					   \
	RECVPKT2MSG(mpkt, msg);						   \
	(nic)->overFlowRecvBufs--;					   \
	/* clean up */							   \
	UNPREP_DMA(infop);						   \
	/* free the dma_info, it was gotten for overflow */		   \
	if ((nic)->numFreeOverFlowBufs > SPILL_HIGH_WATER) {		   \
	    MEMFREE(infop, sizeof(struct alt_dma_info), ALT_NOSLEEP);	   \
	} else {							   \
	    (infop)->next = (nic)->freeOverFlowBufs;			   \
	    (nic)->freeOverFlowBufs = infop;				   \
	    (nic)->numFreeOverFlowBufs++;				   \
	}								   \
    } else {								   \
        /* WARNING, Don't use the infop structure any more in this	   \
	 * routine */							   \
        ALT_ATT_EXT_MSG(mpkt, infop, len);				   \
	if (mpkt == 0) {						   \
	    ERROR4("%s%d: Out of %ss in RingDequeueMsgs\n",		   \
		   (nic)->name, altp->instance, ALT_BUFNAME);		   \
	    ALTTRACEE(nomsgs, 0xf6006, infop, 0, 0, 0);			   \
	    break;							   \
	}								   \
	RECVPKT2MSG(mpkt, msg);						   \
    }									   \
    /* expand this to other platforms some day XXX */			   \
    if ((nic)->curRecvPkt) {						   \
        APPEND_MSG_NEXT((nic)->curRecvPkt, msg);			   \
    } else								   \
        (nic)->curRecvPkt = mpkt;					   \
									   \
    /* adjust the write pointer for the actual data length */		   \
    (msg)->b_wptr += len;						   \
}

#define PASS_RECV_PKT_UP(nic)				\
{							\
    if (bdp->BD_flags & BD_FLAG_END) {			\
        altread(nic, (nic)->curRecvPkt);		\
	(nic)->inpkts++;	/* keep stats */	\
	(nic)->curRecvPkt = (ALT_MSG *)0;		\
    }							\
}

#define RECV_POST_PROCESSING(nic, count) RecvPostProcessing(nic)

#define ALT_BUFNAME "mblk"

#define SNOOZE drv_usecwait
#define IO_BASE_TYPE	caddr_t

/* mutex() won't nest.  Be careful about this */
#define LOCK_TYPE kmutex_t
#define LOCK_INIT(mu, name, type, arg) \
    mutex_init(mu, name, type, arg)
#define LOCK_DESTROY(mu) mutex_destroy(mu)
#define ALT_DISABLE(spl, mu) { if (mu) mutex_enter(mu); }
#define ALT_ENABLE(spl, mu) { if (mu) mutex_exit(mu); }
#define ALT_DISABLE_AT_DISPATCH_LEVEL ALT_DISABLE
#define ALT_ENABLE_AT_DISPATCH_LEVEL ALT_ENABLE

/* page length macros */
#define PAGE_SIZE  MMU_PAGESIZE
#define PAGE_OFFSET(addr) ((int)(addr) & (PAGE_SIZE-1))
#define PAGE_ALIGEND(addr) (PAGE_OFFSET(addr) == 0)
#define PAGE_ROUNDUP(addr) (((int)(addr) + PAGE_SIZE - 1) & ~(PAGE_SIZE-1))
#define PAGE_RESID(addr) (PAGE_SIZE - ((int)(addr) & (PAGE_SIZE-1)))

/* copyin and copyout are strictly forbidden for streams drivers */
/* don't allow compile */
#define COPYOUT(drv, user, size) this is broken
#define COPYIN(user, drv, size)  this is broken

#define SUSER(cred) (!drv_priv(cred))

/* caching macros */
#define PSO_MEMORY 1
#if defined(PSO_MEMORY)
#define MALLOC_CACHE_INVAL(handle, addr, off, length) \
    ddi_dma_sync(handle, off, length, DDI_DMA_SYNC_FORKERNEL)

#define MALLOC_CACHE_WB(handle, addr, off, length) \
    ddi_dma_sync(handle, off, length, DDI_DMA_SYNC_FORDEV)

#else /* defined(PSO_MEMORY) */
/* Suns run in TSO style memory which means that the rings are in
 * cache-consistent memory (ALT_DMA_CONSISTENT vs. ALT_DMA_STREAMING)
 * This means we don't have to sync up the cache, cool.
 */
#define MALLOC_CACHE_INVAL(handle, addr, off, length)
#define MALLOC_CACHE_WB(handle, addr, off, length) 
#endif /* defined(PSO_MEMORY) */

#if defined(PSO_MEMORY)		/* in CONSISTENT space */
#if FAST_DMA
/* caching macros for fast dma mappings */
#define PREP_CACHE_INVAL(infop, len) \
    dvma_sync((infop)->handle, (infop)->index, DDI_DMA_SYNC_FORKERNEL)

#define PREP_CACHE_WB(infop, len) \
    dvma_sync((infop)->handle, (infop)->index, DDI_DMA_SYNC_FORDEV)

#else /* FAST_DMA */

/* caching macros */
#define PREP_CACHE_INVAL(infop, len) \
    ddi_dma_sync((infop)->handle, 0, len, DDI_DMA_SYNC_FORKERNEL)

#define PREP_CACHE_WB(infop, len) \
    ddi_dma_sync((infop)->handle, 0, len, DDI_DMA_SYNC_FORDEV)

#endif /* FAST_DMA */
#else /* defined(PSO_MEMORY) */

/* in cache-consistent memory now, no cache sync */
#define PREP_CACHE_INVAL(infop, len)
#define PREP_CACHE_WB(infop, len)

#endif /* defined(PSO_MEMORY) */
#if FAST_DMA
#define ALLOC_SEND_STUFF(rc, altp, nic)					    \
{									    \
    /*									    \
     * allocate the Page Table Entries for send buffer mapping.		    \
     * This assumes max 2 ptes per buffer.  This is true because of	    \
     * the way we align the buffers on pages in copydata mode.		    \
     */									    \
    if ((((rc = dvma_reserve((altp)->dip, &(nic)->dmaLims,		    \
			     SEND_RING_ENTRIES * PAGES_PER_SEND_BUFFER,	    \
			     &(nic)->alt_dvmaxh))) != DDI_SUCCESS)) {	    \
	ERROR3("%s%d: Couldn't reserve dvma structures for send buffers\n", \
	       (nic)->name, (altp)->instance);				    \
	ALTTRACEE(nodvmas, 0xf6003, 0, 0, 0, 0);			    \
	rc = ENOMEM;							    \
    }									    \
    (nic)->dvmatxAlloc = TRUE;						    \
}

#define	FREE_SEND_STUFF(altp, nic) \
    (void) dvma_release((nic)->alt_dvmaxh);

#define ALLOC_RECV_STUFF(rc, altp, nic)					    \
{									    \
    /* allocate the Page Table Entries for recv buffer mapping. */	    \
    if ((((rc = dvma_reserve((altp)->dip, &(nic)->dmaLims,		    \
			     RECV_STANDARD_RING_ENTRIES,		    \
			     &(nic)->alt_dvmarh))) != DDI_SUCCESS)) {	    \
	ERROR3("%s%d: Couldn't reserve dvma structures for recv buffers\n", \
	       (nic)->name, (altp)->instance);				    \
	ALTTRACEE(nodvmar, 0xf6004,					    \
		  RECV_STANDARD_RING_ENTRIES,				    \
		  rc, 0, 0);						    \
	rc = ENOMEM;							    \
    }									    \
    (nic)->dvmarxAlloc = TRUE;						    \
}

#define FREE_RECV_STUFF(altp, nic) \
    (void) dvma_release((nic)->alt_dvmarh);

#define ALLOC_JUMBO_RECV_STUFF(rc, altp, nic)				    \
{									    \
    /* allocate the Page Table Entries for recv jumbo buffer mapping */	    \
    if ((((rc = dvma_reserve((altp)->dip, &(nic)->dmaLims,		    \
			     RECV_JUMBO_RING_ENTRIES *			    \
			     PAGES_PER_JUMBO_BUFFER,			    \
			     &(nic)->alt_jumbo_dvmarh))) != DDI_SUCCESS)) { \
	ERROR3("%s%d: Couldn't reserve dvma structures for recv buffers\n", \
	       (nic)->name, (altp)->instance);				    \
	ALTTRACEE(nodvmarj, 0xf6005,					    \
		  RECV_JUMBO_RING_ENTRIES * PAGES_PER_JUMBO_BUFFER,	    \
		  rc, 0, 0);						    \
	rc = ENOMEM;							    \
    }									    \
    (nic)->dvmaJumborxAlloc = TRUE;					    \
}

#define FREE_JUMBO_RECV_STUFF(altp, nic) \
    (void) dvma_release((nic)->alt_jumbo_dvmarh);

#if USE_MINI_RING
#define ALLOC_MINI_RECV_STUFF(rc, altp, nic)				    \
{									    \
    /* allocate the Page Table Entries for recv mini buffer mapping */	    \
    if ((((rc = dvma_reserve((altp)->dip, &(nic)->dmaLims,		    \
			     RECV_MINI_RING_ENTRIES,			    \
			     &(nic)->alt_mini_dvmarh))) != DDI_SUCCESS)) {  \
	ERROR3("%s%d: Couldn't reserve dvma structures for recv buffers\n", \
	       (nic)->name, (altp)->instance);				    \
	ALTTRACEE(nodvmarm, 0xf6007,					    \
		  RECV_MINI_RING_ENTRIES,				    \
		  rc, 0, 0);						    \
	rc = ENOMEM;							    \
    }									    \
    (nic)->dvmaMinirxAlloc = TRUE;					    \
}

#define FREE_MINI_RECV_STUFF(altp, nic) \
    (void) dvma_release((nic)->alt_mini_dvmarh);

#endif /* USE_MINI_RING */

#define	PREP_SEND_INFOP(infop, altp, mnic, mindex, curPtr, physPtr)	\
{									\
    (infop)->dataAddr = curPtr;						\
    (infop)->handle = (mnic)->alt_dvmaxh;				\
    (infop)->nic = mnic;						\
    (infop)->index = mindex;						\
    /* Something like the following is necessary for FAST_DMA  (XXX)	\
     *  infop->index = i * SEND_INDEX_MULTIPLIER;			\
     */									\
}
#define PREP_SEND_FLY_INFOP(infop, mnic, mindex, addr, len, msg, pkt)	\
{									\
    (infop)->pkt = pkt;							\
    (infop)->dataAddrFly = addr;					\
    (infop)->flags = 0;							\
    (infop)->dataLen = len;						\
    (infop)->handle = (mnic)->alt_dvmaxh;				\
    (infop)->nic = mnic;						\
    (infop)->index = (mnic)->send_producer;				\
}

#define PREP_RECV_INFOP(infop, mnic, index1, index2, _lastIndex, curPtr, \
			size, physPtr)					 \
{									 \
    (infop)->dataLen = size;						 \
    (infop)->freeRtn.free_func = altRecvDataRtn;			 \
    (infop)->freeRtn.free_arg = (char *)infop;				 \
    (infop)->flags = ALT_DMA_INFO_PREALLOC;				 \
    (infop)->handle = (mnic)->alt_dvmarh;				 \
    (infop)->lastIndex = _lastIndex;					 \
    (infop)->nic = mnic;						 \
    /* will have to do something like this for FAST_DMA (XXX)		 \
     * infop->index = i * RECV_INDEX_MULTIPLIER;			 \
     */									 \
    (infop)->index = index1 * RECV_BUFS_PER_PAGE + index2;		 \
    (infop)->dataAddr = curPtr + ALT_BUF_OFFSET;  /* offset for perf */	 \
}

#else /* FAST_DMA */
#define ALLOC_SEND_STUFF(rc, altp, nic) rc = 0
#define	FREE_SEND_STUFF(altp, nic)
#define ALLOC_RECV_STUFF(rc, altp, nic) rc = 0
#define FREE_RECV_STUFF(altp, nic)
#define ALLOC_JUMBO_RECV_STUFF(rc, altp, nic) rc = 0
#define FREE_JUMBO_RECV_STUFF(altp, nic)
#if USE_MINI_RING
#define ALLOC_MINI_RECV_STUFF(rc, altp, nic) rc = 0
#define FREE_MINI_RECV_STUFF(altp, nic)
#endif /* USE_MINI_RING */

#define	PREP_SEND_INFOP(infop, altp, mnic, mindex, curPtr, physPtr)	\
{									\
    (infop)->dataAddr = curPtr;						\
    (infop)->dataLen = MAX_HYBRID_SIZE;					\
    (infop)->lastIndex = SEND_RING_ENTRIES-1;				\
    (infop)->nic = mnic;						\
    (infop)->index = mindex;						\
}

#define PREP_SEND_FLY_INFOP(infop, mnic, mindex, addr, len, msg, mpkt)	\
{									\
    (infop)->pkt = mpkt;						\
    (infop)->dataAddrFly = addr;					\
    (infop)->flags = 0;							\
    (infop)->dataLen = len;						\
    (infop)->nic = mnic;						\
    (infop)->index = (mnic)->send_producer;				\
}

#define PREP_RECV_INFOP(infop, mnic, index1, index2, _lastIndex, curPtr, \
			size, physPtr)					 \
{									 \
    (infop)->dataLen = size;						 \
    (infop)->freeRtn.free_func = altRecvDataRtn;			 \
    (infop)->freeRtn.free_arg = (char *)infop;				 \
    (infop)->flags |= ALT_DMA_INFO_PREALLOC;				 \
    (infop)->lastIndex = _lastIndex;					 \
    (infop)->nic = mnic;						 \
    (infop)->index = index1 * RECV_BUFS_PER_PAGE + index2;		 \
    (infop)->dataAddr = curPtr + ALT_BUF_OFFSET;  /* offset for perf */	 \
}

#endif /* FAST_DMA */

#define PREP_RECV_JUMBO_INFOP PREP_RECV_INFOP
#define UNPREP_SEND_INFOP(infop)
#define UNPREP_SEND_FLY_INFOP(infop)
#define UNPREP_RECV_INFOP(infop)
#if USE_MINI_RING
#define PREP_RECV_MINI_INFOP PREP_RECV_INFOP
#endif /* USE_MINI_RING */

#define ALLOC_RECV_BUF(altp, size, pri, physaddr) \
    allocb(size, pri)
#define FREE_RECV_BUF(altp, pkt) \
    freemsg(pkt)
#define LINK_STATE_INDICATION(state)
#endif /* defined(__sun) */
/* -------------------------------- SunOS (Solaris 2.x) --------------- */

/* ----------------------- Microsoft Windows NT (NDIS) ---------------- */
#if defined(WINNT)
/* transmit ring parameters */
#define COMMAND_RING_PTR        cmdRingPtr
#define SEND_RING_ENTRIES	512
#define SEND_RING_PTR           0x2000
#define SEND_RING_PHYS_ADDR     sendRingPtr
#define SEND_RING_BDS_IN_WINDOW \
    ((TG_MEM_WINDOW_SIZE / (sizeof(struct send_bd))) - 1)

/* configuration parms */
#define CACHE_LINE_SIZE 32
#define MAX_HYBRID_SIZE 512	/* maximum  */
#define HYBRID_LEN 128		/* cross-over size */
#define USE_EXTENDED_JUMBO 0	/* don't use extended jumbo bds */
#define ALT_RECV_EXT 1		/* receive on prelocked buffers */
#define ALT_LOCKS_SUPPORTED 1	/* MP locks available */
#define REAP_FROM_SEND 1	/* get send bds in send pkt routine */
#define FLOW_CONTROL_SUPPORTED 1 /* system flow control available */
#define USE_DMA_MEMALLOC 1	/* use DMA_MEMALLOC for static buffers? */
#define USE_MINI_RING 0		/* just for testing for now */
#define MINIBLOCKSIZE 128	/* enough to hold an ack plus some */
#define NT_USE_DEBUGGER 0	/* may be using firmware debugger? */
#define ALT_DESERIALIZE 1	/* NT uses deserialized sends */
#define USE_INTERNAL_QUEUE 1	/* NT deserialize can't use ndis queue */

#if NDIS50_MINIPORT
#define USE_CKSUM_OFFLOAD 1	/* NDIS 5 supports checksum offload */
#define DO_CKSUM_DEBUG 0	/* for debugging checksum offload */
#else /* NDIS50_MINIPORT */
#define USE_CKSUM_OFFLOAD 0	/* NDIS 4 does not */
#define DO_CKSUM_DEBUG 0	/* for debugging checksum offload */
#endif /* NDIS50_MINIPORT */

#define HOST_SEND_RING 0	/* bus is a scarce resource on these systems */

/* special section for adding in checksum offload debugging code */
#define RECV_CKSUM_DEBUG    0
#define SEND_CKSUM_DEBUG    0
#define DONT_INIT_JUMBO     0

#if USE_CKSUM_OFFLOAD
#define PREP_SEND_CKSUM_PKT(flags, pkt, msg) flags = prepSendCksumPkt(pkt)
#define VERIFY_RECV_CKSUM(nic, infop, bdp, pkt) \
    verifyRecvCksum(nic, infop, bdp, pkt)
#else /* USE_CKSUM_OFFLOAD */
#define PREP_SEND_CKSUM_PKT(flags, pkt, msg)
#define VERIFY_RECV_CKSUM(nic, infop, bdp, pkt)
#endif /* USE_CKSUM_OFFLOAD */

#define RECVBLOCKSIZE 1536 /* ethernet for now, space for mac hdr */
#define SENDBLOCKSIZE 1536 /* ethernet for now, space for mac hdr */
#define ALT_BUF_OFFSET 10	/* offset into the preallocated recv buf */
#define PAGE_OFFSET(addr) ((int)(addr) & (PAGE_SIZE - 1))
#define PAGE_ROUNDUP(addr) (((int)(addr) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))

#if (DBG)
#define panic(string)   DbgPrint(string)
#else
#define panic(string)
#endif

/* the following two macros were derived from sparc ethernet.h */
#define ether_copy(a, b) 			\
{ 						\
    ((short *)(b))[0] = ((short *)(a))[0];	\
    ((short *)(b))[1] = ((short *)(a))[1]; 	\
    ((short *)(b))[2] = ((short *)(a))[2]; 	\
}

#define ether_cmp(a, b) 			\
    (((short *)(b))[2] != ((short *)(a))[2] ||	\
     ((short *)(b))[1] != ((short *)(a))[1] ||	\
     ((short *)(b))[0] != ((short *)(a))[0])

#ifndef bool
#define bool unsigned char
#endif /* bool */

/* atomic operations... */
#if NDIS50_MINIPORT
#define ATOMIC_INCREMENT(spl, lock, x) NdisInterlockedIncrement(&(x))
#define ATOMIC_DECREMENT(spl, lock, x) NdisInterlockedDecrement(&(x))
#define ATOMIC_ADD(spl, lock, target, value) \
    NdisInterlockedAddUlong(&(x), value, lock)
#define ATOMIC_SET(spl, lock, target, value) \
    /* _InterlockedExchange(&(target), (value)) XXX */ target = value
#else /* NDIS50_MINIPORT */

#if (0)
#define ATOMIC_INCREMENT(spl, lock, x) { \
    extern LONG InterlockedIncrement(PLONG addend);	/* ndis.h is broken */ \
    InterlockedIncrement(&(x)); \
}

#define ATOMIC_DECREMENT(spl, lock, x) InterlockedDecrement(&(x))
#define ATOMIC_ADD(spl, lock, target, value) \
    NdisInterlockedAddUlong(&(x), value, lock) /* XXX */
#else /* XXX */
#define ATOMIC_INCREMENT(spl, lock, x) (x)++
#define ATOMIC_DECREMENT(spl, lock, x) (x)--
#endif /* XXX */

#define ATOMIC_SET(spl, lock, target, value) \
    /* _InterlockedExchange(&(target), (value)) XXX */ target = value
#endif /* NDIS50_MINIPORT */

/* things that would normally be in <sys/types.h> */
typedef char *caddr_t;
typedef unsigned char u_char;
typedef unsigned int u_int;
typedef unsigned long u_long;

/* things that would normally be in <sys/errno.h> */
#define EIO     5
#define ENOMEM  12
#define EFAULT  14
#define EBUSY   16
#define ENODEV  19
#define EINVAL  22

#define ntohl(arg)							\
    ((((U32)(arg) & 0xff) << 24) | (((U32)(arg) & 0xff00) << 8) |	\
     (((U32)(arg) & 0xff0000) >> 8) | ((U32)(arg) >> 24))

/* these macros aren't defined under WINNT */

#define ntohs(arg) ((((U16)(arg) & 0xff) << 8) | ((U16)(arg) >> 8))
#define htonl ntohl
#define htons ntohs

#ifndef U32
#define U64 NDIS_PHYSICAL_ADDRESS
#define U32 unsigned long
#define U16 unsigned short
#define U8  unsigned char
#define S32 signed long
#define S16 signed short
#define S8  signed char
#endif /* U32 */

#define VOLATILE volatile	/* works on NT */

#define PTRS_ARE_64 1

/* little or big endian */
#if !defined(ALT_BIG_ENDIAN)	/* gdb uses this too */
#define ALT_BIG_ENDIAN 0
#endif /* !defined(ALT_BIG_ENDIAN) */

/* driver function names */
#define ALTOPEN 
#define ALTIOCTL 
#define ALTCLOSE 
#define ALTINTR altintr
#define ALTDOINTR(param) 
#define ALTINTRWORK altdointr
#define ALTSTART altstart

/* various things that may be renamed from system to system */
#if defined(ALPHA)

#define RDSHMEM(r) altRead32((U32 *)&(r))
#define WRSHMEM(w,v) altWrite32((U32 *)&(w), (v))
#define WRSHMEM64(w,v) { \
    /*                                                             \
     * this looks backwards, because everything is already         \
     * byteswapped here                                            \
     */                                                            \
    altWrite32((U32 *)&(w), (v.LowPart)); \
    altWrite32((U32 *)((U32)(&(w))+4), (v.HighPart)); \
}

altWrite32((U32 *)&(w), (v))
#define ALT_MEMCPY_TOHOST(to, from, len) \
    NdisMoveFromMappedMemory(to, from, len)
#define ALT_MEMCPY_TODEV(to, from, len) \
    NdisMoveToMappedMemory(to, from, len)

/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define ALT_SWAP_MEMCPY_TOHOST(toAddr, fromAddr, len)	\
{							\
    U32 locLen = (len)/4;				\
    U32 *locTo = (U32 *)(toAddr);			\
    U32 *locFrom = (U32 *)(fromAddr);			\
							\
    while (locLen--) {					\
        *locTo = htonl(altRead32(locFrom));		\
	locTo++;					\
	locFrom++;					\
    }							\
}
 
/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define ALT_SWAP_MEMCPY_TODEV(toAddr, fromAddr, len)	\
{							\
    U32 locLen = (len)/4;				\
    U32 *locTo = (U32 *)(toAddr);			\
    U32 *locFrom = (U32 *)(fromAddr);			\
							\
    while (locLen--) {					\
	altWrite32(locTo, htonl(*locFrom));		\
	locTo++;					\
	locFrom++;					\
    }							\
}

/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define BZERO_TODEV(s1, count)			\
{						\
    U32 locCount = (count)/4;			\
    U32 *locPtr = (U32 *)(s1);			\
						\
    while (locCount--) {			\
	altWrite32(locPtr, 0);			\
	locPtr++;				\
    }						\
}
 
#else /* defined(ALPHA) */
#define RDSHMEM(r) r
#define WRSHMEM(w, v) (w) = (v)
#define WRSHMEM64(w, v) (w) = (v)
#define ALT_MEMCPY_TOHOST(to, from, len) alt_memcpy(to, from, len)
#define ALT_MEMCPY_TODEV(to, from, len) alt_memcpy(to, from, len)
#define ALT_SWAP_MEMCPY_TOHOST(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
#define ALT_SWAP_MEMCPY_TODEV(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define BZERO_TODEV(s1, count)			\
{						\
    U32 locCount = (count)/4;			\
    U32 *locPtr = (U32 *)(s1);			\
						\
    while (locCount--) {			\
	WRSHMEM(*locPtr, 0);			\
	locPtr++;				\
    }						\
}
#endif /* defined(ALPHA) */

#define MEMCPY_TYPE caddr_t
#define MEMCPY_SZT ULONG

#define MEMCPY(to, from, count) NdisMoveMemory(to, from, count)

#define MEMCMP(s1, s2, count) memcmp(s1, s2, count) 
#define BZERO(s1, count) NdisZeroMemory(s1, count)
#define MIN(a,b) ((a) < (b) ? (a) : (b)) 
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* send reenable macro following flow control condition */
#if ALT_DESERIALIZE
#define REENABLE_SEND(nic) \
    SendPackets((NDIS_HANDLE)(nic), (NDIS_PACKET **)0, 0)
#else /* ALT_DESERIALIZE */
#define REENABLE_SEND(nic) NdisMSendResourcesAvailable((nic)->handle)
#endif /* ALT_DESERIALIZE */

/* use these as memory flags */
#define ALT_SLEEP      0
#define ALT_NOSLEEP    0
#define ALT_PHYSCONTIG NDIS_MEMORY_CONTIGUOUS
#define ALT_CACHEALIGN 0
#define ALT_DMA_CONSISTENT  0
#define ALT_DMA_STREAMING   0
#define ALT_DMA_SEND        0

#if NDIS50_MINIPORT
#define	ALT_NDIS_MAJOR_VERSION 5
#define	ALT_NDIS_MINOR_VERSION 0
#else /* NDIS50_MINIPORT */
#define	ALT_NDIS_MAJOR_VERSION 4
#define	ALT_NDIS_MINOR_VERSION 0
#endif /* NDIS50_MINIPORT */

/* ether_addr structure not defined */
struct ether_addr {
    U8  ether_addr_octet[6];
};

/* ether_header structure not defined */
struct ether_header {
    struct ether_addr ether_dhost;
    struct ether_addr ether_shost;
    U16 ether_type;
};

#define ALTEON_HEADER_SIZE       sizeof(struct ether_header)
#define MULTICASTLISTMAX         4000 /* XXX fix this */

#define MEMALLOC(where, type, size, flags)				\
{									\
    DMA_ADDR_TYPE maxPhys = NDIS_PHYSICAL_ADDRESS_CONST(-1, -1);	\
    NdisAllocateMemory(&(where), size, flags, maxPhys);			\
}

#define MEMZALLOC(where, type, size, flags)				\
{									\
    DMA_ADDR_TYPE maxPhys = NDIS_PHYSICAL_ADDRESS_CONST(-1, -1);	\
    NdisAllocateMemory(&(where), size, flags, maxPhys);			\
    if (where)								\
        BZERO(where, size);						\
}

#define MEMFREE(ptr, size, flags) NdisFreeMemory(ptr, size, flags)

#define DMA_ADDR_TYPE NDIS_PHYSICAL_ADDRESS

/* works in place */
#define WORD_SWAP(out, in)						\
{									\
    U32 physLow;							\
									\
    physLow = NdisGetPhysicalAddressLow(in);				\
    NdisSetPhysicalAddressLow(out, NdisGetPhysicalAddressHigh(in));	\
    NdisSetPhysicalAddressHigh(out, physLow);				\
}

/* use this for structures that could be used for DMA by the NIC */
#define DMA_MEMALLOC(where, type, size, flags, rw, infop)		 \
{									 \
    NDIS_STATUS status;							 \
    (infop)->handle = nic->handle;					 \
    NdisMAllocateSharedMemory((infop)->handle, size, FALSE,		 \
			      (void *)&(where), &(infop)->dataAddrPhys); \
    /* word swap the address because there's no 64 bit write in x86 */	 \
    WORD_SWAP((infop)->dataAddrDmaAddr, (infop)->dataAddrPhys);		 \
}
									       
/* free an area accessed for DMA */
#define DMA_MEMFREE(ptr, size, infop)				\
    NdisMFreeSharedMemory((infop)->handle, size, FALSE, ptr,	\
			  (infop)->dataAddrPhys)

#define CACHE_LINE_SIZE 32

#define ALIGN_ADDR(addr, align, type) \
(type)(((int)(addr) + ((align) - 1)) & ~((align) - 1))

/* this macro assumes that the physical address is preswapped */
#define ALIGN_DMA_ADDR(addr, align)					\
{									\
    /* Need to watch out for 64 bit addresses */			\
    U32 physLow = NdisGetPhysicalAddressHigh(addr);			\
									\
    physLow = (physLow + ((align) - 1)) & ~((align) - 1);		\
    if (physLow < (U32)NdisGetPhysicalAddressHigh(addr))		\
	NdisSetPhysicalAddressLow(addr,					\
				  NdisGetPhysicalAddressLow(addr) + 1);	\
    NdisSetPhysicalAddressHigh(addr, physLow);				\
}

/* this macros assumes that the physical address is preswapped */
#define ADD_DMA_ADDR(outaddr, inaddr, len)				  \
{									  \
    /* Need to watch out for 64 bit addresses */			  \
    ULONG physLow = NdisGetPhysicalAddressHigh(inaddr);			  \
									  \
    physLow = physLow + len;						  \
    if (physLow < (U32)NdisGetPhysicalAddressHigh(inaddr))		  \
	NdisSetPhysicalAddressLow(outaddr,				  \
				  NdisGetPhysicalAddressLow(inaddr) + 1); \
    else								  \
	NdisSetPhysicalAddressLow(outaddr,				  \
				  NdisGetPhysicalAddressLow(inaddr));	  \
    NdisSetPhysicalAddressHigh(outaddr, physLow);			  \
}

#define SET_INFOP_ADDR(infop, addr) \
    (infop)->dataAddr = (caddr_t)(addr)

#define TRUNC_ADDR(addr, align, type) (type)((int)(addr) & ~((align) - 1))

/* getting/freeing infops, either deserialized or not */
#if ALT_DESERIALIZE
#define INIT_FREE_INFOP(nic) \
    ExInitializeSListHead(&nic->freeSmsgs);
#define GET_FREE_INFOP(infop, nic, spl, entry)			\
{								\
    infop = (struct alt_dma_info *)				\
        NdisInterlockedPopEntrySList(&nic->freeSmsgs,		\
				     &nic->infopQueueMu);	\
    if (infop)							\
        (void) NdisInterlockedDecrement(&nic->smsgsFree);	\
}
#define GET_NEXT_INFOP(infop, nic, spl, entry) \
    infop = nic->smsgPtrs[entry];
#define FREE_INFOP(infop, nic, spl)				\
{								\
    (void) NdisInterlockedPushEntrySList(&(nic)->freeSmsgs,	\
					 &(infop)->slistNext,	\
					 &(nic)->infopQueueMu);	\
    (void) NdisInterlockedIncrement(&(nic)->smsgsFree);		\
}
#else /* ALT_DESERIALIZE */
#define INIT_FREE_INFOP(nic)
#define GET_FREE_INFOP(infop, nic, spl, entry) \
    infop = &nic->smsgs[entry];
#define GET_NEXT_INFOP GET_FREE_INFOP
#define FREE_INFOP(infop, nic, spl)
#endif /* ALT_DESERIALIZE */

/* prepare an area of "shared" memory for DMA, do nothing */
#define PREP_DMA(rc, pflags, rw, infop) rc = 1
#define PREP_JUMBO_DMA PREP_DMA
#define PREP_MINI_DMA PREP_DMA

/* prepare an area of kernel memory for DMA on the fly (get physical mapping 
 * and set up infops)  */
#if ALT_DESERIALIZE
#define	PREP_FLY_DMA(rc, pflags, rw, infop)				   \
{									   \
    /* max 4 - assumes pk max frame and 4k pages */			   \
    NDIS_PHYSICAL_ADDRESS_UNIT segArray[4];				   \
    struct alt_dma_info *infop2, *loc_infop;				   \
    U32 loc_rc;								   \
    U32 loc_index = 1;							   \
									   \
    NdisMStartBufferPhysicalMapping((nic)->handle, (infop)->fly_msg,	   \
				    (infop)->index, TRUE, segArray,	   \
				    &loc_rc);				   \
    WORD_SWAP((infop)->dataAddrFlyDmaAddr, segArray[0].PhysicalAddress);   \
    (infop)->dataLen = segArray[0].Length;				   \
    (infop)->flags = 0;							   \
    rc = loc_rc;							   \
    loc_infop = infop;							   \
    ALTTRACEV(prepfdma, 0x60001, (infop)->flags, (infop)->dataLen,	   \
	      (infop)->pkt, (infop)->fly_msg);				   \
    /* copy into the next infop */					   \
    while (loc_rc > 1) {						   \
        GET_FREE_INFOP(infop2, nic, spl, 0);				   \
        if (!infop2) {							   \
	    ALTTRACE(nomorbd2, 0xf6008, nic->smsgsFree, 0, 0, 0);	   \
            rc = 0;							   \
	    break;							   \
	}								   \
	loc_infop->next = infop2;					   \
	infop2->next = (struct alt_dma_info *)0;			   \
        infop2->dataAddrFly = loc_infop->dataAddrFly + loc_infop->dataLen; \
        WORD_SWAP(infop2->dataAddrFlyDmaAddr,				   \
		  segArray[loc_index].PhysicalAddress);			   \
        infop2->dataLen = segArray[loc_index].Length;			   \
        infop2->pkt = (loc_infop)->pkt;					   \
        infop2->fly_msg = 0;						   \
        infop2->flags = ALT_DMA_INFO_FLAG_CONT;				   \
	ALTTRACEV(prpfdma2, 0x60002, infop2->flags, infop2->dataLen,	   \
		 infop2->pkt, infop2->fly_msg);				   \
	loc_rc--;							   \
	loc_index++;							   \
	loc_infop = infop2;						   \
    }									   \
}
#else /* ALT_DESERIALIZE */
#define	PREP_FLY_DMA(rc, pflags, rw, infop)				   \
{									   \
    /* max 4 - assumes pk max frame and 4k pages */			   \
    NDIS_PHYSICAL_ADDRESS_UNIT segArray[4];				   \
    struct alt_dma_info *infop2, *loc_infop;				   \
    U32 loc_rc;								   \
    U32 loc_index = 1;							   \
									   \
    NdisMStartBufferPhysicalMapping((nic)->handle, (infop)->fly_msg,	   \
				    (infop)->index, TRUE, segArray,	   \
				    &loc_rc);				   \
    WORD_SWAP((infop)->dataAddrFlyDmaAddr, segArray[0].PhysicalAddress);   \
    (infop)->dataLen = segArray[0].Length;				   \
    (infop)->flags = 0;							   \
    rc = loc_rc;							   \
    loc_infop = infop;							   \
    ALTTRACEV(prepfdma, 0x60001, (infop)->flags, (infop)->index,	   \
            (infop)->pkt, (infop)->fly_msg);				   \
    /* copy into the next infop */					   \
    while (loc_rc > 1) {						   \
	infop2 = (loc_infop->index == loc_infop->lastIndex) ?		   \
		  loc_infop - loc_infop->index : loc_infop + 1;		   \
        if (infop2->index == (nic)->send_consumer) {                       \
            NdisMCompleteBufferPhysicalMapping(nic->handle,                \
                                               (infop)->fly_msg,           \
                                               (infop)->index);            \
            rc = 0;                                                        \
            break;                                                         \
        }                                                                  \
        infop2->dataAddrFly = loc_infop->dataAddrFly + loc_infop->dataLen; \
        WORD_SWAP(infop2->dataAddrFlyDmaAddr, 				   \
		  segArray[loc_index].PhysicalAddress);			   \
        infop2->dataLen = segArray[loc_index].Length;			   \
        infop2->pkt = (loc_infop)->pkt;					   \
        infop2->fly_msg = 0;                                               \
        infop2->flags = ALT_DMA_INFO_FLAG_CONT;				   \
	ALTTRACEV(prpfdma2, 0x60002, infop2->flags, infop2->index,	   \
		 infop2->pkt, infop2->fly_msg);				   \
	loc_rc--;							   \
	loc_index++;							   \
	loc_infop = infop2;						   \
    }									   \
}
#endif /* ALT_DESERIALIZE */

/* clean up after PREP_DMA, nothing to do */
#define UNPREP_DMA(infop)

/* clean up after PREP_FLY_DMA */
#define UNPREP_FLY_DMA(infop)					\
    NdisMCompleteBufferPhysicalMapping(nic->handle,		\
				       (infop)->fly_msg,	\
				       (infop)->index)

#if USE_MINI_RING
#define SINGLE_PACKET_PREP(nic, pkt, msg, len, infop, pktArr, count, bdp)  \
{									   \
    /* keep track of the count of packets to indicate */		   \
    (count)++;								   \
									   \
    /* copy the packet into the array and also the fields in the packet */ \
    pkt = (infop)->INFOP_RPKT;						   \
    *(pktArr) = pkt;							   \
    (pkt)->Private.TotalLength = len;					   \
    if (bdp->BD_flags & BD_FLAG_JUMBO_RING) {				   \
	if ((nic)->bufsInJumboRecvRing < ALT_RECV_JUMBO_LOW_LOW) {	   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_RESOURCES);		   \
	    (nic)->lowResources++;					   \
	} else {							   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_SUCCESS);		   \
	    (nic)->normRecv++;						   \
	}								   \
    } else if (bdp->BD_flags & BD_FLAG_MINI_RING) {			   \
	if ((nic)->bufsInMiniRecvRing < ALT_RECV_MINI_LOW_LOW) {	   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_RESOURCES);		   \
	    (nic)->lowResources++;					   \
	} else {							   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_SUCCESS);		   \
	    (nic)->normRecv++;						   \
	}								   \
    } else {								   \
	if ((nic)->bufsInStdRecvRing < ALT_RECV_LOW_LOW) {		   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_RESOURCES);		   \
	    (nic)->lowResources++;					   \
	} else {							   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_SUCCESS);		   \
	    (nic)->normRecv++;						   \
	}								   \
    }									   \
    (pktArr)++;								   \
    NdisAdjustBufferLength((infop)->msg, len);				   \
}
#else /* USE_MINI_RING */
#define SINGLE_PACKET_PREP(nic, pkt, msg, len, infop, pktArr, count, bdp)  \
{									   \
    /* keep track of the count of packets to indicate */		   \
    (count)++;								   \
									   \
    /* copy the packet into the array and also the fields in the packet */ \
    pkt = (infop)->INFOP_RPKT;						   \
    *(pktArr) = pkt;							   \
    (pkt)->Private.TotalLength = len;					   \
    if (bdp->BD_flags & BD_FLAG_JUMBO_RING) {				   \
	if ((nic)->bufsInJumboRecvRing < ALT_RECV_JUMBO_LOW_LOW) {	   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_RESOURCES);		   \
	    (nic)->lowResources++;					   \
	} else {							   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_SUCCESS);		   \
	    (nic)->normRecv++;						   \
	}								   \
    } else {								   \
	if ((nic)->bufsInStdRecvRing < ALT_RECV_LOW_LOW) {		   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_RESOURCES);		   \
	    (nic)->lowResources++;					   \
	} else {							   \
	    NDIS_SET_PACKET_STATUS(pkt, NDIS_STATUS_SUCCESS);		   \
	    (nic)->normRecv++;						   \
	}								   \
    }									   \
    (pktArr)++;								   \
    NdisAdjustBufferLength((infop)->msg, len);				   \
}
#endif /* USE_MINI_RING */

#define PASS_RECV_PKT_UP(nic) (nic)->inpkts++

#define RECV_POST_PROCESSING(nic, count) RecvPostProcessing(nic, count)

#define SNOOZE(delay)				\
{						\
    int delay1 = delay;                         \
                                                \
    if (delay1 <= 50) {                         \
        NdisStallExecution(delay1);             \
    } else {                                    \
        int count = delay1 / 50;                \
                                                \
        while (count--)                         \
            NdisStallExecution(50);             \
    }                                           \
}
  
#define IO_BASE_TYPE caddr_t

/* Ndis3 Miniport Driver Will acquire spin locks before calling Send
   and DPC routines */
#define LOCK_TYPE                       NDIS_SPIN_LOCK
#define LOCK_INIT(mu, name, type, arg)  NdisAllocateSpinLock(mu)
#define LOCK_DESTROY(mu)                NdisFreeSpinLock(mu)
#define ALT_DISABLE(spl, mu)            NdisAcquireSpinLock(mu)
#define ALT_ENABLE(spl, mu)             NdisReleaseSpinLock(mu)
#define ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mu) NdisDprAcquireSpinLock(mu)
#define ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mu) NdisDprReleaseSpinLock(mu)

/* copyin and copyout are strictly forbidden for WINNT drivers */
/* don't allow compile */
#define COPYOUT(drv, user, size) this is broken
#define COPYIN(user, drv, size)  this is broken

/* caching macros */
#define MALLOC_CACHE_INVAL(infop, addr, off, length)
#define MALLOC_CACHE_WB(infop, addr, off, length)

/* caching macros for fast dma mappings */
#define PREP_CACHE_INVAL(infop, len)		\
{						\
    NdisFlushBuffer((infop)->msg, FALSE);	\
}

#define PREP_CACHE_WB(infop, len)				\
{								\
    if ((infop)->flags & ALT_DMA_INFO_PREALLOC) {		\
	NdisFlushBuffer((infop)->msg, TRUE);			\
    } else {							\
	if ((infop)->fly_msg) {					\
	    NdisFlushBuffer((infop)->fly_msg, TRUE);		\
	}							\
    }								\
}

#define ALLOC_SEND_STUFF(rc, altp, nic)					

#define FREE_SEND_STUFF(altp, nic)					

/* alloc/free the buffer/packet pools */
#define ALLOC_RECV_STUFF(rc, altp, nic)					\
{									\
    NDIS_STATUS status;							\
									\
    NdisAllocateBufferPool(&status, 					\
			   &(nic)->recv_flush_buffer_pool_handle,	\
			   RECV_STANDARD_RING_ENTRIES);			\
									\
    if (status != NDIS_STATUS_SUCCESS) {				\
	ERROR1("unable to allocate buffer pool");			\
	rc = ENOMEM;							\
    }									\
    NdisAllocatePacketPool(&status,					\
			   &(nic)->rcv_packet_pool_handle,		\
			   (RECV_STANDARD_RING_ENTRIES + 1),		\
			   16);	/* it's likely this could be 0 */	\
    if (status != NDIS_STATUS_SUCCESS) {				\
        NdisFreeBufferPool((nic)->recv_flush_buffer_pool_handle);	\
	ERROR2("NdisAllocatePacketPool Error[%x]", status);		\
	rc = ENOMEM;							\
    }									\
}

#define FREE_RECV_STUFF(altp, nic)				\
{								\
    NdisFreeBufferPool((nic)->recv_flush_buffer_pool_handle);	\
    (nic)->recv_flush_buffer_pool_handle = (NDIS_HANDLE) 0;	\
    NdisFreePacketPool((nic)->rcv_packet_pool_handle);		\
    (nic)->rcv_packet_pool_handle = (NDIS_HANDLE) 0;		\
}

#define ALLOC_JUMBO_RECV_STUFF(rc, altp, nic)				\
{									\
    NDIS_STATUS status;							\
									\
    NdisAllocateBufferPool(&status,					\
			   &(nic)->recv_flush_jumbo_buffer_pool_handle,	\
			   RECV_JUMBO_RING_ENTRIES);			\
									\
    if (status != NDIS_STATUS_SUCCESS) {				\
	ERROR1("unable to allocate buffer pool");			\
	rc = ENOMEM;							\
    }									\
    NdisAllocatePacketPool(&status,					\
			   &(nic)->jumbo_rcv_packet_pool_handle,	\
			   (RECV_JUMBO_RING_ENTRIES + 1),		\
			   16);						\
    if (status != NDIS_STATUS_SUCCESS) {				\
        NdisFreeBufferPool((nic)->recv_flush_jumbo_buffer_pool_handle);	\
	ERROR2("NdisAllocatePacketPool Error[%x]", status);		\
	rc = ENOMEM;							\
    }									\
}

#define	FREE_JUMBO_RECV_STUFF(altp, nic)				\
{									\
    NdisFreeBufferPool((nic)->recv_flush_jumbo_buffer_pool_handle);	\
    (nic)->recv_flush_jumbo_buffer_pool_handle = (NDIS_HANDLE) 0;	\
    NdisFreePacketPool((nic)->jumbo_rcv_packet_pool_handle);		\
    (nic)->jumbo_rcv_packet_pool_handle = (NDIS_HANDLE) 0;		\
}

#if USE_MINI_RING
#define ALLOC_MINI_RECV_STUFF(rc, altp, nic)				\
{									\
    NDIS_STATUS status;							\
									\
    NdisAllocateBufferPool(&status,					\
			   &(nic)->recv_flush_mini_buffer_pool_handle,	\
			   RECV_MINI_RING_ENTRIES);			\
									\
    if (status != NDIS_STATUS_SUCCESS) {				\
	ERROR1("unable to allocate buffer pool");			\
	rc = ENOMEM;							\
    }									\
    NdisAllocatePacketPool(&status,					\
			   &(nic)->mini_rcv_packet_pool_handle,	        \
			   (RECV_MINI_RING_ENTRIES + 1),		\
			   16);						\
    if (status != NDIS_STATUS_SUCCESS) {				\
        NdisFreeBufferPool((nic)->recv_flush_mini_buffer_pool_handle);	\
	ERROR2("NdisAllocatePacketPool Error[%x]", status);		\
	rc = ENOMEM;							\
    }									\
}

#define	FREE_MINI_RECV_STUFF(altp, nic)					\
{									\
    NdisFreeBufferPool((nic)->recv_flush_mini_buffer_pool_handle);	\
    (nic)->recv_flush_mini_buffer_pool_handle = (NDIS_HANDLE) 0;	\
    NdisFreePacketPool((nic)->mini_rcv_packet_pool_handle);		\
    (nic)->mini_rcv_packet_pool_handle = (NDIS_HANDLE) 0;		\
}
#endif /* USE_MINI_RING */
/* set up static buffers for send infops */
#define PREP_SEND_INFOP(infop, altp, nic, mindex, curPtr, physPtr)	\
{									\
    NDIS_STATUS status;							\
									\
    (infop)->index = mindex;						\
    (infop)->handle = (nic)->handle;					\
    (infop)->lastIndex = SEND_RING_ENTRIES - 1;				\
    (infop)->dataAddr = curPtr;						\
    (infop)->dataAddrDmaAddr = physPtr;                                 \
}

/* set up the on the fly send infops */
#define PREP_SEND_FLY_INFOP(infop, nic, mindex, addr, len, msg, pkt)	\
{									\
    (infop)->fly_msg = msg;						\
    (infop)->INFOP_SPKT = pkt;						\
    (infop)->flags = 0;							\
    (infop)->dataLen = len;						\
    (infop)->dataAddrFly = addr;					\
}

#define UNPREP_SEND_INFOP(infop)
#define UNPREP_SEND_FLY_INFOP(infop)

#define PREP_RECV_INFOP(infop, nic, index1, index2, _lastIndex, curPtr,	\
			size, physPtr)					\
{									\
    NDIS_STATUS status;							\
									\
    (infop)->index = (index1) * RECV_BUFS_PER_PAGE + (index2);		\
    (infop)->handle = (nic)->handle;					\
    (infop)->lastIndex = _lastIndex;					\
    (infop)->dataAddr = curPtr;						\
    (infop)->dataLen = size;						\
    (infop)->flags = 0;							\
    (infop)->dataAddrDmaAddr = physPtr;					\
    /* this "can't" fail, we've allocated enough buffers */		\
    NdisAllocateBuffer(&status, &(infop)->msg,				\
		       (nic)->recv_flush_buffer_pool_handle, 		\
		       curPtr, size);					\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS buffer\n");			\
									\
    /* this also "can't" fail, we've allocated enough packets */	\
    NdisAllocatePacket(&status, &(infop)->INFOP_RPKT,			\
		       (nic)->rcv_packet_pool_handle);			\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS packet\n");			\
									\
    NDIS_SET_PACKET_HEADER_SIZE((infop)->INFOP_RPKT,			\
				sizeof(struct ether_header));		\
    NdisChainBufferAtFront((infop)->INFOP_RPKT,				\
			   (infop)->msg);				\
    /* chain it back in for the free routine */				\
    *(struct alt_dma_info **)						\
        ((infop)->INFOP_RPKT->MiniportReserved) = infop;		\
}

#define PREP_RECV_JUMBO_INFOP(infop, nic, index1, index2, _lastIndex,	\
			      curPtr, size, physPtr)			\
{									\
    NDIS_STATUS status;							\
									\
    (infop)->index = (index1) * RECV_BUFS_PER_PAGE + (index2);		\
    (infop)->handle = (nic)->handle;					\
    (infop)->lastIndex = _lastIndex;					\
    (infop)->dataAddr = curPtr;						\
    (infop)->dataLen = size;						\
    (infop)->flags = 0;							\
    (infop)->dataAddrDmaAddr = physPtr;					\
    /* this "can't" fail, we've allocated enough buffers */		\
    NdisAllocateBuffer(&status, &(infop)->msg,				\
		       (nic)->recv_flush_jumbo_buffer_pool_handle,	\
		       curPtr, size);					\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS buffer\n");			\
									\
    /* this also "can't" fail, we've allocated enough packets */	\
    NdisAllocatePacket(&status, &(infop)->INFOP_RPKT,			\
		       (nic)->jumbo_rcv_packet_pool_handle);		\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS packet\n");			\
									\
    NDIS_SET_PACKET_HEADER_SIZE((infop)->INFOP_RPKT,			\
				sizeof(struct ether_header));		\
    NdisChainBufferAtFront((infop)->INFOP_RPKT,				\
			   (infop)->msg);				\
    /* chain it back in for the free routine */				\
    *(struct alt_dma_info **)						\
        ((infop)->INFOP_RPKT->MiniportReserved) = infop;		\
}

#if USE_MINI_RING
#define PREP_RECV_MINI_INFOP(infop, nic, index1, index2, _lastIndex,	\
			     curPtr, size, physPtr)			\
{									\
    NDIS_STATUS status;							\
									\
    (infop)->index = (index1) * RECV_BUFS_PER_PAGE + (index2);		\
    (infop)->handle = (nic)->handle;					\
    (infop)->lastIndex = _lastIndex;					\
    (infop)->dataAddr = curPtr;						\
    (infop)->dataLen = size;						\
    (infop)->flags = 0;							\
    (infop)->dataAddrDmaAddr = physPtr;					\
    /* this "can't" fail, we've allocated enough buffers */		\
    NdisAllocateBuffer(&status, &(infop)->msg,				\
		       (nic)->recv_flush_mini_buffer_pool_handle,	\
		       curPtr, size);					\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS buffer\n");			\
									\
    /* this also "can't" fail, we've allocated enough packets */	\
    NdisAllocatePacket(&status, &(infop)->INFOP_RPKT,			\
		       (nic)->mini_rcv_packet_pool_handle);		\
    if (status != NDIS_STATUS_SUCCESS)					\
        ERROR1("Unable to allocate an NDIS packet\n");			\
									\
    NDIS_SET_PACKET_HEADER_SIZE((infop)->INFOP_RPKT,			\
				sizeof(struct ether_header));		\
    NdisChainBufferAtFront((infop)->INFOP_RPKT,				\
			   (infop)->msg);				\
    /* chain it back in for the free routine */				\
    *(struct alt_dma_info **)						\
        ((infop)->INFOP_RPKT->MiniportReserved) = infop;		\
}
#endif /* USE_MINI_RING */

#define UNPREP_RECV_INFOP(infop)		\
{						\
    if ((infop)->msg) {				\
        NdisFreeBuffer(infop->msg);		\
	(infop)->msg = (ALT_MSG *)0;		\
    }						\
    if ((infop)->INFOP_RPKT) {			\
        NdisFreePacket((infop)->INFOP_RPKT);	\
	(infop)->INFOP_RPKT = (RECV_PKT *)0;	\
    }						\
}

#define ALLOC_RECV_BUF(altp, size, pri, physaddr)
#define FREE_RECV_BUF(altp, pkt)

#define LINK_STATE_INDICATION(state)					\
{									\
   if ((state) == TG_EVENT_CODE_LINK_UP) {				\
       NdisMIndicateStatus(nic->handle, NDIS_STATUS_MEDIA_CONNECT,	\
			   (void *)0, sizeof(U32));			\
       NdisMIndicateStatusComplete(nic->handle);			\
   } else {								\
       NdisMIndicateStatus(nic->handle, NDIS_STATUS_MEDIA_DISCONNECT,	\
			   (void *)0, sizeof(U32));			\
       NdisMIndicateStatusComplete(nic->handle);			\
   }									\
}
#endif /* defined(WINNT) */
/* ----------------------- END Microsoft Windows NT (NDIS) ---------------- */

/* ----------------------- Novell ODI --------------------------------- */
#if defined(__NETWARE_386__)
/* transmit ring parameters */
#define COMMAND_RING_PTR        TG_COMMAND_RING
#define SEND_RING_ENTRIES	128
#define SEND_RING_PTR		0x3800
#define SEND_RING_PHYS_ADDR     SEND_RING_PTR
#define SEND_RING_BDS_IN_WINDOW ((TG_MEM_WINDOW_SIZE/(sizeof(struct send_bd)))-1)

#if DEBUG
#define ASSERT(x)  { if (!(x)) EnterDebugger(); }
#else /* DEBUG */
#define ASSERT(x)
#endif /* DEBUG */

#define PAGE_SIZE  4096
#define PAGE_OFFSET(addr) ((int)(addr) & (PAGE_SIZE-1))
#define PAGE_ALIGNED(addr) (PAGE_OFFSET(addr) == 0)
#define PAGE_ROUNDUP(addr) (((int)(addr) + PAGE_SIZE - 1) & ~(PAGE_SIZE-1))
#define PAGE_RESID(addr) (PAGE_SIZE - ((int)(addr) & (PAGE_SIZE-1)))

#define CACHE_LINE_SIZE 32                
/* configuration parms */
#define MAX_HYBRID_SIZE 512	/* maximum size we allow to be set */
#define HYBRID_LEN 128		/* cross-over size */
#define USE_EXTENDED_JUMBO 0	/* don't use extended jumbo bds */
#define ALT_RECV_EXT 0		/* don't receive on prelocked buffers */
#define ALT_LOCKS_SUPPORTED 0	/* no locks on ODI? XXX */
#define REAP_FROM_SEND 0	/* don't get send bds in send pkt routine */
#define FLOW_CONTROL_SUPPORTED 0 /* system flow control available */
#define USE_DMA_MEMALLOC 0	/* use DMA_MEMALLOC for static buffers? */
#define USE_MINI_RING 0		/* just for testing for now */
#define USE_INTERNAL_QUEUE 0	/* ODI doesn't need to queue due to pkt cnt */
#define MINIBLOCKSIZE 128	/* enough to hold an ack plus some */

/* special section for adding in checksum offload debugging code */
#define RECV_CKSUM_DEBUG    0
#define SEND_CKSUM_DEBUG    0
#define DONT_INIT_JUMBO     0

#define USE_CKSUM_OFFLOAD 0
#define DO_CKSUM_DEBUG 0	/* for debugging checksum offload */
#define PREP_SEND_CKSUM_PKT(flags, pkt, msg)
#define VERIFY_RECV_CKSUM_PKT(nic, infop, bdp, pkt)

 /* space here for offset and max ethernet header and cache line boundary */
#define RECVBLOCKSIZE 1536 
#define SENDBLOCKSIZE 1536
#define ALT_BUF_OFFSET 10	/* offset into the preallocated recv buf */

/* just print a fatal error message, don't die */
#define panic(string) ERROR1(string)

#ifndef bool
#define bool unsigned char
#endif /* bool */

/* atomic operations... */
/*
 * XXX replace with more efficient equivalents...
 */
#define ATOMIC_INCREMENT(spl, lock, x) (x)++
#define ATOMIC_DECREMENT(spl, lock, x) (x)--
#define ATOMIC_ADD(spl, lock, target, value) target += value
#define ATOMIC_SET(spl, lock, target, value) target = value

/* these macros aren't defined under NOVELL  */
#define ntohl VALUE_TO_HILO_UINT32
#define ntohs VALUE_TO_HILO_UINT16
#define htonl VALUE_FROM_HILO_UINT32
#define htons VALUE_FROM_HILO_UINT16

#ifndef U32
#define U64 UINT64
#define U32 unsigned int
#define U16 unsigned short
#define U8  unsigned char
#define S32 signed int
#define S16 signed short
#define S8  signed char
#define ULONG unsigned long
#endif /* U32 */

#define VOLATILE volatile	/* volatile works in watcom c */

#define PTRS_ARE_64 0

/* little or big endian */
#define ALT_BIG_ENDIAN 0

/* driver function names */
#define ALTOPEN 
#define ALTIOCTL 
#define ALTCLOSE 
#define ALTINTR altintr
#define ALTDOINTR(param) 
#define ALTINTRWORK altdointr
#define ALTSTART altstart

/* various things that may be renamed from system to system */
#define RDSHMEM(r) r
#define WRSHMEM(w, v) (w) = (v)
#define WRSHMEM64(w, v) (w) = (v)
#define ALT_MEMCPY_TOHOST(to, from, len) alt_memcpy(to, from, len)
#define ALT_MEMCPY_TODEV(to, from, len) alt_memcpy(to, from, len)
#define ALT_SWAP_MEMCPY_TOHOST(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
#define ALT_SWAP_MEMCPY_TODEV(toAddr, fromAddr, len) \
    alt_swap_memcpy(toAddr, fromAddr, len)
/* assumes count divisibility by 4 and s1 is pointer to U32 */
#define BZERO_TODEV(s1, count)			\
{						\
    U32 locCount = (count)/4;			\
    U32 *locPtr = (U32 *)(s1);			\
						\
    while (locCount--) {			\
	WRSHMEM(*locPtr, 0);			\
	locPtr++;				\
    }						\
}

#define MEMCPY_TYPE void *
#define MEMCPY_SZT size_t
#define MEMCPY memcpy
#define MEMCMP memcmp
#define BZERO(s1, count) memset(s1, '\0', count)
 
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b)) 
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif /* MIN */

/* use these as memory flags */
#define ALT_SLEEP           0
#define ALT_NOSLEEP         0
#define ALT_PHYSCONTIG      0
#define ALT_CACHEALIGN      0
#define ALT_DMA_CONSISTENT  0
#define ALT_DMA_STREAMING   0
#define ALT_DMA_SEND        0

/* not in any header files */
#define ETHERMTU            1500

/* ether_addr structure not defined */
struct ether_addr {
    U8  ether_addr_octet[6];
};

/* ether_header structure not defined */
struct ether_header {
    struct ether_addr ether_dhost;
    struct ether_addr ether_shost;
    U16 ether_type;
};

#define MEMALLOC(where, type, size, flags) \
   (where) = (type) CMSMAlloc(altp->drvData, size)

#define MEMZALLOC(where, type, size, flags)		\
{							\
    (where) = (type) CMSMAlloc(altp->drvData, size);	\
    if( where)						\
        BZERO(where, size);				\
}

#define MEMFREE(ptr, size, flags) CMSMFree(altp->drvData, ptr); 

#define DMA_ADDR_TYPE void *

/* use this for structures that could be used for DMA by the NIC */
#define DMA_MEMALLOC(where, type, size, flags, rw, infop)		   \
{									   \
   (where) = (type)CMSMAlloc(altp->drvData, size);			   \
    if (where) {							   \
	(infop)->dataAddr = (char *)where;                                 \
    }									   \
}
									       
/* free an area accessed for DMA */
#define DMA_MEMFREE(ptr, size, infop)		\
    CMSMFree(altp->drvData, ptr);

#define CACHE_LINE_SIZE 32

#define ALIGN_ADDR(addr, align, type)           \
    (type)(((int)(addr) + ((align) - 1)) & ~((align) - 1))

#define ALIGN_DMA_ADDR(addr, align)

#define ADD_DMA_ADDR(outaddr, inaddr, len) \
     outaddr = (HOSTADDR_TYPE)((char *)(inaddr) + (len))

#define SET_INFOP_ADDR(infop, addr) \
    (infop)->dataAddr = (char *)(addr)

#define TRUNC_ADDR(addr, align, type)           \
    (type)((int)(addr) & ~((align) - 1))

/* getting/freeing infops, either deserialized or not */
#define INIT_FREE_INFOP(nic)
#define GET_FREE_INFOP(infop, nic, spl, entry) \
    infop = &nic->smsgs[entry];
#define GET_NEXT_INFOP GET_FREE_INFOP
#define FREE_INFOP(infop, nic, spl)

#define	PREP_DMA(rc, pflags, rw, infop) rc = 1;
#define PREP_JUMBO_DMA PREP_DMA
#define PREP_MINI_DMA PREP_DMA
void *MSMGetLogical(void *);
#pragma aux MSMGetLogical  \
         parm [eax]  \
         value [eax]  \
         modify [ebp ebx ecx edx esi edi];
#define	PREP_FLY_DMA(rc, pflags, rw, infop)	\
{	\
    FRAGMENT_STRUCT newFrag[4];	\
    FRAGMENT_STRUCT oldFrag;	\
    UINT32 newCnt = 0;	\
    int i;	\
    struct alt_dma_info *infop2 = infop;	\
	\
    oldFrag.FragmentAddress = infop->dataAddrFly;	\
    oldFrag.FragmentLength = infop->dataLen;	\
    if (CMSMGetPhysList(1, &oldFrag, &newCnt, newFrag, nic->altp->drvData) != ODISTAT_SUCCESSFUL) {	\
	rc = 0;	\
    }	\
    else {	\
	if ((rc = newCnt) > 1) {	\
	    infop->dataLen = newFrag[0].FragmentLength;	\
	    newCnt--;	\
	    for (i = 1 ;newCnt ; newCnt--, i++) {	\
		infop2->flags |= ALT_DMA_INFO_FLAG_CONT;		  \
		infop2 = ((infop)->index == (infop)->lastIndex) ?	  \
			(infop) - (infop)->index : (infop) + 1;		  \
	\
		infop2->INFOP_SPKT = infop->pkt;	\
		infop2->dataAddrFly = MSMGetLogical(newFrag[i].FragmentAddress);	\
		infop2->dataLen = newFrag[i].FragmentLength;	\
		infop2->flags = 0;	\
	\
	    }	\
	}	\
    }	\
}
#define UNPREP_DMA(infop)
#define UNPREP_FLY_DMA(infop)

#define SNOOZE(count)				\
{						\
    int delayc = count * 2;			\
    while (--delayc)				\
        Slow();					\
}

#define IO_BASE_TYPE void *

/* mutexes aren't needed on Netware, they know better :-( */
#define LOCK_TYPE 
#define ALT_DISABLE(spl, mu)  
#define ALT_ENABLE(spl, mu)   
#define ALT_DISABLE_AT_DISPATCH_LEVEL(spl, mu)  
#define ALT_ENABLE_AT_DISPATCH_LEVEL(spl, mu)   

/* copyin and copyout are strictly forbidden for ODI drivers */
/* don't allow compile */
#define COPYOUT(drv, user, size) this is broken
#define COPYIN(user, drv, size)  this is broken

/* caching macros don't matter, hardware handles everything */
#define MALLOC_CACHE_INVAL(handle, addr, off, length)  

#define MALLOC_CACHE_WB(handle, addr, off, length)     

/* caching macros for fast dma mappings */
#define PREP_CACHE_INVAL(infop, len)
#define PREP_CACHE_WB(infop, len)

#define ALLOC_SEND_STUFF(rc, altp, nic) rc = 0
#define	FREE_SEND_STUFF(altp, nic)
#define ALLOC_RECV_STUFF(rc, altp, nic) rc = 0
#define	FREE_RECV_STUFF(altp, nic)
#define ALLOC_JUMBO_RECV_STUFF(rc, altp, nic) rc = 0
#define	FREE_JUMBO_RECV_STUFF(altp, nic)
#if USE_MINI_RING
#define ALLOC_MINI_RECV_STUFF(rc, altp, nic) rc = 0
#define	FREE_MINI_RECV_STUFF(altp, nic)
#endif /* USE_MINI_RING */
#define PREP_SEND_INFOP(infop, altp, nic, mindex, curPtr, physPtr)	\
{									\
    (infop)->index = mindex;						\
    (infop)->lastIndex = SEND_RING_ENTRIES - 1;				\
    (infop)->dataAddr = curPtr;						\
    (infop)->flags = 0;							\
}

#define PREP_SEND_FLY_INFOP(infop, nic, mindex, addr, len, msg, pkt)	\
{									\
    (infop)->INFOP_SPKT = pkt;						\
    (infop)->dataAddrFly = addr;					\
    (infop)->flags = 0;							\
    (infop)->dataLen = len;						\
}

#define PREP_RECV_INFOP(infop, nic, index1, index2, curPtr, size, msg, pkt) \
{									    \
    (infop)->INFOP_RPKT = pkt;						    \
    (infop)->dataAddr = curPtr;						    \
    (infop)->flags = 0;							    \
    (infop)->dataLen = size;						    \
}

#define PREP_RECV_JUMBO_INFOP PREP_RECV_INFOP
#if USE_MINI_RING
#define PREP_RECV_MINI_INFOP PREP_RECV_INFOP
#endif /* USE_MINI_RING */
#define UNPREP_SEND_INFOP(infop)
#define UNPREP_SEND_FLY_INFOP(infop)
#define UNPREP_RECV_INFOP(infop)
#define ALLOC_RECV_BUF(altp, size, pri, physaddr) \
    CMSMAllocateRCB((altp)->drvData, size, &physaddr)
#define FREE_RECV_BUF(altp, pkt) \
    CMSMReturnRCB((altp)->drvData, pkt)
#define LINK_STATE_INDICATION(state)
#endif /* defined(__NETWARE_386__) */
/* ----------------------- Novell ODI --------------------------------- */

#ifndef STATIC
#if H_DEBUG
#define STATIC 
#else /* H_DEBUG */
#define STATIC static
#endif /* H_DEBUG */
#endif /* STATIC */

#if ALT_DESERIALIZE
#define NUM_INFOPS (SEND_RING_ENTRIES - 1)
#else /* ALT_DESERIALIZE */
#define NUM_INFOPS SEND_RING_ENTRIES
#endif /* ALT_DESERIALIZE */

#if ALT_RECV_EXT
#define PAGES_PER_JUMBO_BUFFER (JUMBOBLOCKSIZE/PAGE_SIZE + 1)
#define RECV_BUFS_PER_PAGE (PAGE_SIZE/RECVBLOCKSIZE)
#if USE_MINI_RING
#define MINI_BUFS_PER_PAGE (PAGE_SIZE/MINIBLOCKSIZE)
#endif /* USE_MINI_RING */
#else /* ALT_RECV_EXT */
#define PAGES_PER_JUMBO_BUFFER (JUMBOBLOCKSIZE/PAGE_SIZE + 2)
#endif /* ALT_RECV_EXT */

/* tagged_ether_header structure */
struct tagged_ether_header {
    struct ether_addr ether_dhost;
    struct ether_addr ether_shost;
    U16 tag_id;
    U16 vlan_id;
    U16 ether_type;
};

/* snap_ether_header structure */
struct snap_ether_header {
      struct ether_addr ether_dhost;
      struct ether_addr ether_shost;
      U16 ether_type;
      U8 llc_dsap;
      U8 llc_ssap;
      U8 llc_cntrl;
      U8 snap_org[3];
      U16 snap_type;
};

/* tagged_snap_ether_header structure */
struct tagged_snap_ether_header {
      struct ether_addr ether_dhost;
      struct ether_addr ether_shost;
      U16 tag_id;
      U16 vlan_id;
      U16 ether_type;
      U8 llc_dsap;
      U8 llc_ssap;
      U8 llc_cntrl;
      U8 snap_org[3];
      U16 snap_type;
};
 
/* this is used throughout the world */
#if USE_EXTENDED_JUMBO
#define JUMBO_RECV_BD ext_recv_bd
#else /* USE_EXTENDED_JUMBO */
#define JUMBO_RECV_BD recv_bd
#endif /* USE_EXTENDED_JUMBO */

#endif /* _ALTTYPES_H_ */
