/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: cksum.c,v $
 * Revision 1.1.2.16  1998/12/07  23:33:37  ted
 * 	Change level of some traces for better readability
 * 	[1998/12/07  23:30:24  ted]
 *
 * Revision 1.1.2.15  1998/11/23  22:50:35  ted
 * 	Go ahead and report failed checksum error
 * 	[1998/11/23  22:50:03  ted]
 * 
 * Revision 1.1.2.14  1998/11/17  00:19:34  hayes
 * 	update cksum offload testing framework
 * 	[1998/11/17  00:17:33  hayes]
 * 
 * Revision 1.1.2.13  1998/11/03  23:25:30  ted
 * 	New NT checksum offload interface
 * 	[1998/11/03  22:21:17  ted]
 * 
 * Revision 1.1.2.12  1998/09/03  20:05:04  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:14  ted]
 * 
 * Revision 1.1.2.11  1998/08/12  01:58:07  ted
 * 	Refine send deserialization
 * 	[1998/08/10  21:28:05  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:02:00  ted]
 * 
 * Revision 1.1.2.10  1998/07/21  01:26:32  andy
 * 	Finalize cksum support for split/fragmented jumbo frames
 * 	[1998/07/21  01:24:42  andy]
 * 
 * Revision 1.1.2.9  1998/07/02  05:30:44  andy
 * 	Handle split/fragmented jumbo packets
 * 	[1998/07/02  05:22:21  andy]
 * 
 * Revision 1.1.2.8  1998/06/23  01:29:49  shuang
 * 	fixed rare case of tcp/udp cksum verification failure.
 * 	[1998/06/23  00:55:33  shuang]
 * 
 * Revision 1.1.2.7  1998/05/20  05:38:53  andy
 * 	collect detail cksum stats
 * 	[1998/05/20  02:59:24  andy]
 * 
 * Revision 1.1.2.6  1998/05/14  02:56:28  andy
 * 	checksum debug logic
 * 	[1998/05/14  02:45:06  andy]
 * 
 * Revision 1.1.2.5  1998/04/22  21:34:24  ted
 * 	Fix Non-copydata mode IP fragged test case
 * 	[1998/04/22  21:23:55  ted]
 * 
 * Revision 1.1.2.4  1998/04/17  06:56:28  andy
 * 	Various fixes
 * 	[1998/04/17  06:56:17  andy]
 * 
 * Revision 1.1.2.3  1998/04/17  03:08:22  ted
 * 	Add check for fragment receive bailout, fix fragment receive logic
 * 	[1998/04/17  03:07:50  ted]
 * 
 * Revision 1.1.2.2  1998/04/13  18:44:01  andy
 * 	Fix checksum offload testing code
 * 	[1998/04/13  18:42:04  andy]
 * 
 * Revision 1.1.2.1  1998/04/02  06:03:12  ted
 * 	Add checksum debugging support
 * 	[1998/04/02  05:55:45  ted]
 * 
 * $EndLog$
 */

/* only for Solaris for now */
#if defined(__sun)

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/dditypes.h>	/* ddi types */
#include <sys/ddidmareq.h>	/* ddi dma structs */
#include <sys/ethernet.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"
#include "altos.h"
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"

#if RECV_CKSUM_DEBUG || SEND_CKSUM_DEBUG
/* forward */
U16 ckSum_pseudo(struct ip *);
U16 ckSum_udp_pseudo(struct ip *, struct udphdr *);
#endif /* RECV_CKSUM_DEBUG || SEND_CKSUM_DEBUG */

#if SEND_CKSUM_DEBUG
/* prepare a full packet for send with appropriate cksum debugging
 * done.  Return the flags that should be set in the buffer descriptor */
U16
prepSendCksumPkt(ALT_MSG *mp) 
{
    struct ether_header *eth;
    struct ip *ipp;
    struct tcphdr *tcpp;
    struct udphdr *udpp;
    U16 flags;
    U32 temp;
    U32 len;
    ALT_MSG *mp2;

    MSGINFO(mp, eth, len, struct ether_header *);

    if (eth->ether_type == ETHERTYPE_IP) {
       /* set up the ipp, may not be in first mblk */
       /* if we've got more than the ether_header assume the whole IP header
        * and UDP header, if it's not more than the ether_header we assume
        * it's exactly the ether_header.  This might be broken someday, but
        * seems to work for now */
       if (len > sizeof(struct ether_header))
	  ipp = (struct ip *)((caddr_t)eth + sizeof(struct ether_header));
       else {
	  mp2 = MSGNEXT(mp);
	  MSGINFO(mp2, ipp, len, struct ip *);
	  ALTTRACE(NxtMblk, 0x90002, eth, len, ipp, 0);
       }
 
	if (ipp->ip_p == IPPROTO_UDP) {
	    udpp = (struct udphdr *)((caddr_t)ipp + (ipp->ip_hl << 2));

	    /* check for fragment or not */
	    if (ipp->ip_off == IP_DF || ipp->ip_off == 0) {
		/* NON FRAG CASE */
		if (do_send_ip_cksum)
		    ipp->ip_sum = 0;     /* zap ip cksum  */
		if (do_send_tcpudp_cksum) {
		    if (do_send_pseudo_cksum)
			udpp->uh_sum = 0;     /* zap udp cksum */
		    else {
			/* calculate pseudo header */
			udpp->uh_sum = ckSum_udp_pseudo(ipp, udpp); 
		    }
		}
		
	        flags = send_bdflags;
		MEMCPY( (MEMCPY_TYPE)&temp, (MEMCPY_TYPE)ipp, sizeof(temp) );
		ALTTRACE(SndUdp, 0x90003, temp, flags, udpp->uh_sum, 0);
		
            } else if (ipp->ip_off & IP_MF) {
		ALTTRACE(udpFragIn, 0x90004, ipp->ip_off, udpp->uh_sum,
			 ipp->ip_id, 0);
		/* not last of fragment train */
		if (do_send_ip_cksum)
		    ipp->ip_sum = 0;     /* zap ip cksum  */

		/* Check if it is start or middle of Frag Frame */
		if ((ipp->ip_off & 0x1fff) == 0) {
		    /* Start of frag, zap udp checksum as well */
		    if (do_send_tcpudp_cksum) {
			if (do_send_pseudo_cksum)
			    udpp->uh_sum = 0;     /* zap udp cksum */
			else {
			    /* calculate pseudo header */
			    udpp->uh_sum = ckSum_udp_pseudo(ipp, udpp); 
			}
		    }
		}
	        flags = send_bdflags;
		if (do_send_tcpudp_cksum)
		    flags |= BD_FLAG_IP_FRAG;
		ALTTRACE(udpFragOut, 0x90005, ipp->ip_off, udpp->uh_sum,
			 ipp->ip_id, flags);
	    } else {
		/* last frag of a fragment train */
		if (do_send_ip_cksum)
		    ipp->ip_sum = 0;     /* zap ip cksum  */ 
		    
	        flags = send_bdflags;
		if (do_send_tcpudp_cksum)
		    flags |= BD_FLAG_IP_FRAG_END;
		ALTTRACE(udpFragLast, 0x90006, ipp->ip_off, udpp->uh_sum,
			 ipp->ip_id, flags);
	    }
        } else 	if (ipp->ip_p == IPPROTO_TCP) {
	    /* for TCP , no frags expected */
	    tcpp = (struct tcphdr *)((caddr_t)ipp + (ipp->ip_hl << 2));

	    ALTTRACE(SndTcpIn, 0x90007, do_send_ip_cksum, ipp->ip_sum,
		     do_send_tcpudp_cksum, tcpp->th_sum);
	    if (do_send_ip_cksum)
		ipp->ip_sum = 0;     /* zap ip cksum  */ 
	    if (do_send_tcpudp_cksum) {
		if (do_send_pseudo_cksum)
                    tcpp->th_sum = 0;    /* zap tcp cksum */
		else 
		    /* calculate pseudo header */
		    tcpp->th_sum = ckSum_pseudo(ipp);   
	    }
		
	    flags = send_bdflags;
	    ALTTRACE(SndTcpOut, 0x90008, do_send_ip_cksum, ipp->ip_sum,
		     do_send_tcpudp_cksum, tcpp->th_sum);
		
	} else { /* NON UDP and TCP */
            flags = 0;
	    ALTTRACE(nonUdpTcp, 0x90009, ipp->ip_p, 0, 0, 0);
	}
    } else { /* NON IP frame */
	flags = 0;
	ALTTRACE(nonIP, 0x9000a, eth->ether_type, 0, 0, 0);
    }

    return(flags);
}

/* tcp or udp cksum with a preseeded pseudoheader checksum */
U16
ckSumIt(char *cp, S32 len)
{
    U32 sum = 0;

    if (len < 0)
        return(0xdead);

    if (len & 1) {
        *(cp + len) = 0;	/* pad */
	len += 1;
    }
    while (len > 0) {
        sum += ntohs(*(U16 *)cp);
	cp += 2;
	len -= 2;
    }

    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);

    return (U16)(~sum & 0xffff);
}
#endif /* SEND_CKSUM_DEBUG */

#if RECV_CKSUM_DEBUG

void
verifyRecvCksum(struct tg_nic *nic, struct alt_dma_info *infop,
		struct recv_bd *bdp)
{
    struct ether_header *eth = (struct ether_header *)infop->dataAddr;
    struct ip *ipp = 
	(struct ip *)(infop->dataAddr + sizeof(struct ether_header));
    struct udphdr *udpp;
    U16 pseudo;
    U32 ckSum;
    U32 trc0, trc1, trc2, trc3;
#if defined(PASS_IP_ID)
    static U16 RecvState = 0;
    U16 DumpData = FALSE;
    U16 KillTrace = FALSE;

    /* some paranoid checking logic */
    if (do_recv_tcpudp_cksum || do_recv_ip_cksum ) {
       if (eth->ether_type == ETHERTYPE_IP &&
           ipp->ip_p == IPPROTO_TCP &&
	   ipp->ip_id != bdp->BD_vlan_tag) {
   	   nic->BadIpID++;
	   ALTTRACEE(BadIpID, 0x900d0, ipp->ip_id, bdp->BD_vlan_tag,
		bdp->BD_flags, bdp->BD_len);
       } else if (nic->LastIpID != bdp->BD_vlan_tag) {
   	   ALTTRACE(NewPkt, 0x900d1, nic->LastIpID,
		bdp->BD_vlan_tag, bdp->BD_flags, bdp->BD_len);	
	   if (bdp->BD_vlan_tag && RecvState != 0 && RecvState != 3 && RecvState != 4) {
	      nic->BadEnd++;
	      ALTTRACEE(BadEnd, 0x900d2, RecvState, 0, 0, 0);
	   }
	   if (nic->jumboSplit) {
	      nic->BadSplit++;
	      ALTTRACEE(BadSplit, 0x900d3, RecvState, 0, 0, 0);
           }
	   nic->LastIpID = bdp->BD_vlan_tag;
	   if (nic->LastIpID == 0 ) {
	      RecvState = 0;			/* Idle state */
	   } else {
	      if (bdp->BD_flags & BD_FLAG_END) {
                 if (eth->ether_type == ETHERTYPE_IP &&
                    ipp->ip_p == IPPROTO_UDP &&
                    ipp->ip_off == IP_MF) {
                       RecvState = 5;       	/* Recv Frag state */
                  } else {
                       RecvState = 3;       	/* Recv End state */
                  }
	      } else {
	   	  RecvState = 1;		/* Recv Start state */
	      }
	   }
       } else {	/* still working on the same packet(probably) */
	   switch (RecvState) {
	      case 0:					/* Idle state */
							/* probably a non-IP packet */
		   break;
	      case 1:					/* Recv Start state */
		   if (bdp->BD_flags & BD_FLAG_END) {
		      RecvState = 3;		/* Recv End state */
		   } else {
		      RecvState = 2;		/* Recv Continue state */
		   }
	   	   break;
	      case 2:					/* Recv Continue state */
                   if (bdp->BD_flags & BD_FLAG_END) {
                      RecvState = 3;       	/* Recv End state */
                   }
		   if (!(nic->jumboSplit)) {
		      nic->SplitErr++;
		      ALTTRACEE(SplitErr, 0x900d4, 0,
                         ipp->ip_p, bdp->BD_vlan_tag, bdp->BD_flags);
                      RecvState = 4;       	/* Error state */
                   }
                   break;
	      case 3:					/* Recv End state */
		   if ((bdp->BD_flags & BD_FLAG_END)) {
		      nic->DupEnd++;
                      ALTTRACEE(DupEnd, 0x900d5, 0,
                            nic->LastIpID, bdp->BD_vlan_tag, bdp->BD_flags);
		      RecvState = 4;		/* Error state */
		   } else {
		      RecvState = 3;
		   }
                   break;
	      case 4:					/* Error state */
		   nic->ErrCnt++;
                   ALTTRACEE(ErrState, 0x900d6, 0,
                        nic->LastIpID, bdp->BD_vlan_tag, bdp->BD_flags);
                   break;
	      case 5:					/* Recv Frag state */
    		   if (eth->ether_type == ETHERTYPE_IP &&
		      ipp->ip_p == IPPROTO_UDP &&
	              ipp->ip_off & IP_MF) {
		      RecvState = 5;
		   } else {
		      RecvState = 3;     
		   }
                   break;
	    } /* switch on RecvState */
	} /* paranoid logic tests */
   }  /* cksum enabled */

   if (DumpData) {
	MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr +  0), sizeof(trc0) );
	MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr +  4), sizeof(trc1) );
	MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr +  8), sizeof(trc2) );
	MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 12), sizeof(trc3) );
	ALTTRACEV(Data, 0x900d7, trc0, trc1, trc2, trc3);
	MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr + 16), sizeof(trc0) );
	MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr + 20), sizeof(trc1) );
	MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr + 24), sizeof(trc2) );
	MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 28), sizeof(trc3) );
	ALTTRACEV(Data, 0x900d8, trc0, trc1, trc2, trc3);
   } /* DumpData */

   if (KillTrace) altTracing = 0;
#endif /* defined(PASS_IP_ID) */	   

    /* handle middle and end packets for split frames */
    if (nic->jumboSplit && do_recv_tcpudp_cksum) {
	/* last in frame */
	if (bdp->BD_flags & BD_FLAG_END) {
	    ckSum = nic->inProgressCksum;
	    ckSum += bdp->BD_tcp_udp_cksum;
	    ckSum = (ckSum >> 16) + (ckSum & 0xffff);
	    nic->inProgressCksum = 0;
	    nic->jumboSplit = FALSE;
	    ALTTRACE(SplitEnd, 0x90010, ipp->ip_sum,
		     0, bdp->BD_flags, bdp->BD_len);
	    ALTTRACE(SplitEnd, 0x90011,
		     bdp->BD_ip_cksum, bdp->BD_tcp_udp_cksum, 
		     ckSum, ~ckSum);
	    /* UDP Frag handling here */
	    if (nic->splitFrag) {
		nic->splitFrag = FALSE;
		ckSum += nic->fragudp_cksum;
	    	ckSum = (ckSum >> 16) + (ckSum & 0xffff);
	    	nic->fragudp_cksum = ckSum;
                if (nic->EndFrag) {
		    nic->EndFrag = FALSE;
		    ALTTRACE(EndFrag, 0x90012, nic->fragip_id,
			nic->fragip_offset, (U16)~(nic->fragudp_cksum),
			nic->frag_pseudo);
    		    if ((U16)~(nic->fragudp_cksum) != 0) {
			ALTTRACEE(BadFragCk, 0x90013, nic->fragudp_cksum,
				    nic->frag_pseudo, bdp->BD_ip_cksum,
				    bdp->BD_tcp_udp_cksum);
			nic->recvUDPCksumFail4++;
    		    }
		    nic->fragudp_cksum = 0;
		    nic->fragip_id     = 0;
		    nic->fragip_offset = 0;
		    nic->frag_pseudo   = 0;
		}
	    } else if (ckSum != 0xffff) {
		ALTTRACEE(SplitBad, 0x90014, ckSum, ~ckSum, 0, 0);
		if (nic->splitTCP)
		    nic->recvTCPCksumFail0++;
		else 
		    nic->recvUDPCksumFail0++;
	    }
	} else {
	    /* middle of frame, just keep running cksum */
	    ckSum = nic->inProgressCksum;
	    ckSum += bdp->BD_tcp_udp_cksum;
	    ckSum = (ckSum >> 16) + (ckSum & 0xffff);
	    nic->inProgressCksum = ckSum;
	    ALTTRACE(Split, 0x90015, ipp->ip_sum,
		     0, bdp->BD_flags, bdp->BD_len);
	    ALTTRACE(Split, 0x90016,
		     bdp->BD_ip_cksum, bdp->BD_tcp_udp_cksum, 
		     ckSum, ~ckSum);
	}
	return;
    } if (nic->jumboSplit) {
	/* last in frame */
	if (bdp->BD_flags & BD_FLAG_END) {
	    nic->jumboSplit = FALSE;
	    nic->splitFrag = FALSE;
	    nic->fragudp_cksum = 0;
	    nic->fragip_id     = 0;
	    nic->fragip_offset = 0;
	    nic->frag_pseudo   = 0;
	}
	return;
    } 
    
    if (eth->ether_type == ETHERTYPE_IP) {
	if (ipp->ip_p == IPPROTO_TCP) {
	    MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr +  0), sizeof(trc0) );
	    MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr +  4), sizeof(trc1) );
	    MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr +  8), sizeof(trc2) );
	    MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 12), sizeof(trc3) );
	    ALTTRACEV(TCPData, 0x90020, trc0, trc1, trc2, trc3);
	    MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr + 16), sizeof(trc0) );
	    MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr + 20), sizeof(trc1) );
	    MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr + 24), sizeof(trc2) );
	    MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 28), sizeof(trc3) );
	    ALTTRACEV(TCPData, 0x90021, trc0, trc1, trc2, trc3);
	    MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr + 32), sizeof(trc0) );
	    MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr + 36), sizeof(trc1) );
	    MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr + 40), sizeof(trc2) );
	    MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 44), sizeof(trc3) );
	    ALTTRACEV(TCPData, 0x90022, trc0, trc1, trc2, trc3);
	    MEMCPY( (MEMCPY_TYPE)&trc0, (MEMCPY_TYPE)(infop->dataAddr + 48), sizeof(trc0) );
	    MEMCPY( (MEMCPY_TYPE)&trc1, (MEMCPY_TYPE)(infop->dataAddr + 52), sizeof(trc1) );
	    MEMCPY( (MEMCPY_TYPE)&trc2, (MEMCPY_TYPE)(infop->dataAddr + 56), sizeof(trc2) );
	    MEMCPY( (MEMCPY_TYPE)&trc3, (MEMCPY_TYPE)(infop->dataAddr + 60), sizeof(trc3) );
	    ALTTRACEV(TCPData, 0x90023, trc0, trc1, trc2, trc3);
	    /* split jumbo, must be first to get here */
	    if (!(bdp->BD_flags & BD_FLAG_END)) {
		/* check the IP cksum if necessary */
		if (do_recv_ip_cksum) {
		    if ((U16)~bdp->BD_ip_cksum != 0) {
			/* may want to panic here */
			ALTTRACEE(RcvUIPBad, 0x90025, bdp->BD_len, 
				  bdp->BD_index,
				  bdp->BD_ip_cksum,
				  bdp->BD_tcp_udp_cksum);
#if 0 /* for DEBUG, stop FW */
			stopFW(altp);    
#endif /* for DEBUG, stop FW */
			
			nic->recvIPCksumFail0++;
		    }
		}
		
		if (do_recv_tcpudp_cksum) {
		    if (do_recv_pseudo_cksum) {
		    	ckSum = 0;
		    } else {
			/* calculate pseudo header */
			pseudo = ckSum_pseudo(ipp);
			ckSum  = pseudo;
		    }
		    ckSum += bdp->BD_tcp_udp_cksum;
	    	    ckSum  = (ckSum >> 16) + (ckSum & 0xffff);
	   	    nic->inProgressCksum = ckSum; 
		    ALTTRACE(SplitTCP, 0x90026, ipp->ip_sum,
			     0, bdp->BD_flags, bdp->BD_len);
		    ALTTRACE(SplitTCP, 0x90027,
			     bdp->BD_ip_cksum, bdp->BD_tcp_udp_cksum, 
			     ckSum, pseudo );
		}

		nic->jumboSplit = TRUE;
		nic->splitTCP = TRUE;
		nic->splitFrag = FALSE;
		return;
	    }

	    /*
	     * complement ip checksum computed by adapter.
	     * if 0, we have a good checksum..
	     */
	    if (do_recv_ip_cksum) {
		if ((U16)~bdp->BD_ip_cksum != 0) {
		    /* may want to panic here */
		    ALTTRACEE(RcvIPBad, 0x90028, bdp->BD_len, 
			      bdp->BD_index,
			      bdp->BD_ip_cksum,
			      bdp->BD_tcp_udp_cksum);
#if 0 /* for DEBUG, stop FW */
                    stopFW(altp);    
#endif /* for DEBUG, stop FW */

                    nic->recvIPCksumFail1++;
		}
	    }

	    /* check the tcp checksum if necessary, 
	     * adding pseudo if necessary */
	    if (do_recv_tcpudp_cksum) {
		if (do_recv_pseudo_cksum) {
		    if ((U16)~bdp->BD_tcp_udp_cksum != 0) {
			/* may want to panic here */
			ALTTRACEE(RcvTCPBad, 0x90029, bdp->BD_len, 0,
				  bdp->BD_ip_cksum, bdp->BD_tcp_udp_cksum);
			nic->recvTCPCksumFail1++;
		    }
		} else {
		    /* calculate pseudo header */
		    pseudo = ckSum_pseudo(ipp);
	    	    ckSum = bdp->BD_tcp_udp_cksum + pseudo;
	    	    ckSum = (ckSum >> 16) + (ckSum & 0xffff);
		    if ((U16)~ckSum != 0) {
			/* may want to panic here */
			ALTTRACEE(RcvPTCPBad, 0x9002a, bdp->BD_len, pseudo,
  				  bdp->BD_ip_cksum,
				  bdp->BD_tcp_udp_cksum);
			nic->recvTCPCksumFail2++;
		    }
		}
	    }
	} else if (ipp->ip_p == IPPROTO_UDP) {
	    if (ipp->ip_off == IP_DF || ipp->ip_off == 0) {
		udpp = (struct udphdr *)(infop->dataAddr + 
					 sizeof(struct ether_header) +
					 (ipp->ip_hl << 2));
		/* NON FRAG CASE */
		/* are we in a split case? must be first one in split */
		if (!(bdp->BD_flags & BD_FLAG_END)) {
		    /* check the IP cksum if necessary */
		    if (do_recv_ip_cksum) {
			if ((U16)~bdp->BD_ip_cksum != 0) {
			    /* may want to panic here */
			    ALTTRACEE(RcvUIPBad, 0x90030, bdp->BD_len, 
				      bdp->BD_index,
				      bdp->BD_ip_cksum,
				      bdp->BD_tcp_udp_cksum);
#if 0 /* for DEBUG, stop FW */
			    stopFW(altp);    
#endif /* for DEBUG, stop FW */
			    
			    nic->recvIPCksumFail2++;
			}
		    }

		    if (do_recv_tcpudp_cksum) {
			if (do_recv_pseudo_cksum) {
			    ckSum = 0;
			} else {
			    /* calculate pseudo header */
			    pseudo = ckSum_udp_pseudo(ipp, udpp);
			    ckSum  = pseudo;
			}
			ckSum += bdp->BD_tcp_udp_cksum;
	    		ckSum  = (ckSum >> 16) + (ckSum & 0xffff);
			nic->inProgressCksum = ckSum; 
			ALTTRACE(SplitUDP, 0x90031, ipp->ip_sum,
				 udpp->uh_sum, bdp->BD_flags, bdp->BD_len);
			ALTTRACE(SplitUDP, 0x90032,
				 bdp->BD_ip_cksum, bdp->BD_tcp_udp_cksum,
				 ckSum, pseudo );
		    }

		    nic->jumboSplit = TRUE;
		    nic->splitTCP = FALSE;
		    nic->splitFrag = FALSE;
		    return;
		} else {
		    /* regular non-split, non-fragged UDP packet */
		    /*
		     * complement ip checksum computed by adapter.
		     * if 0, we have a good checksum..
		     */
		    if (do_recv_ip_cksum) {
			if ((U16)~bdp->BD_ip_cksum != 0) {
			    /* may want to panic here */
			    ALTTRACEE(RcvIPBad, 0x90033, bdp->BD_len, 
				      bdp->BD_index,
				      bdp->BD_ip_cksum,
				      bdp->BD_tcp_udp_cksum);
#if 0 /* for DEBUG, stop FW */
			    stopFW(altp);    
#endif /* for DEBUG, stop FW */

			    nic->recvIPCksumFail3++;
			}
		    }

		    /* check the udp checksum if necessary, 
		     * adding pseudo if necessary */
		    if (do_recv_tcpudp_cksum) {
			if (do_recv_pseudo_cksum) {
			    if ((U16)~bdp->BD_tcp_udp_cksum != 0) {
				/* may want to panic here */
				ALTTRACEE(RcvTCPBad, 0x90034, 
					  bdp->BD_len, 0,
					  bdp->BD_ip_cksum, 
					  bdp->BD_tcp_udp_cksum);
				nic->recvUDPCksumFail1++;
			    }
			} else {
			    /* calculate pseudo header */
			    pseudo = ckSum_udp_pseudo(ipp, udpp);
	    	    	    ckSum = bdp->BD_tcp_udp_cksum + pseudo;
	    	    	    ckSum = (ckSum >> 16) + (ckSum & 0xffff);
			    if ((U16)~ckSum != 0) {
				/* may want to panic here */
				ALTTRACEE(RcvPTCPBad, 0x90035, 
					  bdp->BD_len, pseudo,
				  	  bdp->BD_ip_cksum,
					  bdp->BD_tcp_udp_cksum);
				nic->recvUDPCksumFail2++;
			    }
			}
		    }
		}
	    } else { /* frag case */
		ALTTRACE(InFrag, 0x90036, bdp->BD_ip_cksum, 
			 bdp->BD_tcp_udp_cksum, nic->fragip_id,
			 ipp->ip_sum);
		if (nic->fragip_id && ipp->ip_id != nic->fragip_id) { 
		    /*not this frag chain*/
		    ALTTRACEE(BadFRAG, 0x90037, nic->fragip_id, ipp->ip_id, 0, 0);
		    nic->splitFrag = FALSE;
		    nic->fragudp_cksum = 0;
		    nic->fragip_id     = 0;
		    nic->fragip_offset = 0;
		    nic->frag_pseudo   = 0;
		    nic->recvCksumBail0++;
		    return;
		}
		nic->fragip_id = ipp->ip_id;
		ALTTRACE(InFrag2, 0x90038, nic->fragip_id, nic->fragip_offset,
			 ipp->ip_off, (ipp->ip_off & 0x1fff) << 3);
		if (nic->fragip_offset &&
		    (U16)(nic->fragip_offset) != ((ipp->ip_off & 0x1fff) << 3)) {
		    ALTTRACEE(BadOFF, 0x90039, 0, nic->fragip_offset,
			     ipp->ip_off, (ipp->ip_off & 0x1fff) << 3);
		    nic->splitFrag = FALSE;
		    nic->fragudp_cksum = 0;
		    nic->fragip_id     = 0;
		    nic->fragip_offset = 0;
		    nic->frag_pseudo   = 0;
		    nic->recvCksumBail1++;
		    return;
		}
		
		nic->fragip_offset += ipp->ip_len - (ipp->ip_hl << 2);
		/* verify ip checksum on frag */
		if (do_recv_ip_cksum) {
		    if ((U16)~bdp->BD_ip_cksum != 0) {
			/* may want to panic here */
			ALTTRACEE(RcvIPBad, 0x9003a, bdp->BD_len, 
				  bdp->BD_index,
				  bdp->BD_ip_cksum,
				  bdp->BD_tcp_udp_cksum);
#if 0 /* for DEBUG, stop FW */
			stopFW(altp);    
#endif /* for DEBUG, stop FW */
		    
			nic->recvIPCksumFail4++;
                        nic->splitFrag = FALSE;
			return;
		    }
		}
		
		if (do_recv_tcpudp_cksum) {
		    /* add cksum cumulatively */
		    nic->fragudp_cksum += bdp->BD_tcp_udp_cksum;
		    nic->fragudp_cksum = (nic->fragudp_cksum >> 16) + 
			(nic->fragudp_cksum & 0xffff);
		}
		if (ipp->ip_off & IP_MF) {
		    if (do_recv_tcpudp_cksum) {
			/* Check if it is start or middle of Frag Frame */
			if ((ipp->ip_off & 0x1fff) == 0) {
			    /* ok, start of frag */ 
			    /* compute pseudo */
			    if (!do_recv_pseudo_cksum) {
				udpp = (struct udphdr *)(infop->dataAddr + 
				     sizeof(struct ether_header) + (ipp->ip_hl << 2));
				nic->frag_pseudo = ckSum_udp_pseudo(ipp, udpp);
			    }
			    ALTTRACE(StartFrag, 0x9003b, nic->fragip_id, 
				   ipp->ip_off, nic->fragip_offset, nic->frag_pseudo);
			}
		    }	
                    if (!(bdp->BD_flags & BD_FLAG_END)) {
                        nic->inProgressCksum = nic->fragudp_cksum;
                        nic->fragudp_cksum = 0;
                        nic->jumboSplit = TRUE;
                        nic->splitTCP  = FALSE;
                        nic->splitFrag = TRUE;
                        nic->EndFrag   = FALSE;
		    }
		    /* are we at the last frag ? */
		} else if (ipp->ip_off & 0x1fff) {
		    if (do_recv_tcpudp_cksum) {
			if (!do_recv_pseudo_cksum) {
			    nic->fragudp_cksum += nic->frag_pseudo;
			    nic->fragudp_cksum = 
				(nic->fragudp_cksum >> 16) + 
				(nic->fragudp_cksum & 0xffff);
			    nic->fragudp_cksum = 
				(nic->fragudp_cksum >> 16) + 
				(nic->fragudp_cksum & 0xffff);
			}
			
                	if (bdp->BD_flags & BD_FLAG_END) {
                            nic->splitFrag = FALSE;
			    ALTTRACE(EndFrag, 0x9003c, nic->fragip_id,
				 nic->fragip_offset, (U16)~(nic->fragudp_cksum),
				 nic->frag_pseudo);
    			    if ((U16)~(nic->fragudp_cksum) != 0) {
			        ALTTRACEE(BadFragCk, 0x9003d, nic->fragudp_cksum,
				    nic->frag_pseudo, bdp->BD_ip_cksum,
				    bdp->BD_tcp_udp_cksum);
			        nic->recvUDPCksumFail3++;
    			    }
			} else {
                            nic->inProgressCksum = nic->fragudp_cksum;
                            nic->fragudp_cksum = 0;
			}
		    }
		    if (bdp->BD_flags & BD_FLAG_END) {
		    	nic->fragudp_cksum = 0;
		    	nic->fragip_id     = 0;
		    	nic->fragip_offset = 0;
		    	nic->frag_pseudo   = 0;
		    } else {
                        nic->jumboSplit = TRUE;
                        nic->splitTCP  = FALSE;
                        nic->splitFrag = TRUE;
                        nic->EndFrag   = TRUE;
		    }
		}
 		ALTTRACE(OutFrag, 0x9003e, nic->fragip_id, ipp->ip_off,
			 nic->fragip_offset, nic->fragudp_cksum);
 		ALTTRACE(OutFrag2, 0x9003f, bdp->BD_flags, ipp->ip_len,
			 0, 0);
 	    } /* end of frag case */
	} /* UDP */
    } /* IP */
}
#endif /* RECV_CKSUM_DEBUG */

#if RECV_CKSUM_DEBUG || SEND_CKSUM_DEBUG
U16
ckSum_pseudo(struct ip *ipp)
{
    U32 pseudo = ipp->ip_len;

    pseudo -= (ipp->ip_hl << 2);  /* Adjust for the length of IP header */
    pseudo += ipp->ip_p;
    pseudo += ipp->ip_src.S_un.S_un_w.s_w1; /* short one of ip src addr */
    pseudo += ipp->ip_src.S_un.S_un_w.s_w2; /* short two of ip src addr */
    pseudo += ipp->ip_dst.S_un.S_un_w.s_w1; /* short one of ip src addr */
    pseudo += ipp->ip_dst.S_un.S_un_w.s_w2; /* short two of ip src addr */
 
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);
    return(pseudo);
}

U16
ckSum_udp_pseudo(struct ip *ipp, struct udphdr *udpp)
{
    U32 pseudo = udpp->uh_ulen;

    pseudo += ipp->ip_p;
    pseudo += ipp->ip_src.S_un.S_un_w.s_w1; /* short one of ip src addr */
    pseudo += ipp->ip_src.S_un.S_un_w.s_w2; /* short two of ip src addr */
    pseudo += ipp->ip_dst.S_un.S_un_w.s_w1; /* short one of ip src addr */
    pseudo += ipp->ip_dst.S_un.S_un_w.s_w2; /* short two of ip src addr */
 
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);
    return(pseudo);
}
#endif /* RECV_CKSUM_DEBUG || SEND_CKSUM_DEBUG */

#elif defined(WINNT) /* defined(__sun) */
#include <ndis.h>

#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"
#include "altos.h"
#include "althdr.h"
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"

#if USE_CKSUM_OFFLOAD
U16
prepSendCksumPkt(SEND_PKT *pkt) 
{
    NDIS_PACKET_EXTENSION *pktExtPtr;
    NDIS_TCP_IP_CHECKSUM_PACKET_INFO cksumPktInfo;
    U16 flags = 0;

    if (NDIS_GET_PACKET_PROTOCOL_TYPE(pkt) == NDIS_PROTOCOL_ID_TCP_IP) {
        pktExtPtr = NDIS_PACKET_EXTENSION_FROM_PACKET(pkt);
        cksumPktInfo.Value = 
	    (ULONG)pktExtPtr->NdisPacketInfo[TcpIpChecksumPacketInfo];
	if (cksumPktInfo.Transmit.NdisPacketChecksumV4) {
	    if (cksumPktInfo.Transmit.NdisPacketIpChecksum)
		flags |= BD_FLAG_IP_CKSUM;
	    if (cksumPktInfo.Transmit.NdisPacketTcpChecksum ||
		cksumPktInfo.Transmit.NdisPacketUdpChecksum)
	        flags |= BD_FLAG_TCP_UDP_CKSUM;
	}
	ALTTRACE(prepsend, 0x90000, pktExtPtr, 
		 NDIS_GET_PACKET_PROTOCOL_TYPE(pkt), 
		 cksumPktInfo.Value, flags);
    }
    return(flags);
}

void
verifyRecvCksum(struct tg_nic *nic, struct alt_dma_info *infop, 
		struct recv_bd *bdp, RECV_PKT *pkt)
{
    NDIS_PACKET_EXTENSION *pktExtPtr;
    NDIS_TCP_IP_CHECKSUM_PACKET_INFO cksumPktInfo;

    cksumPktInfo.Value = 0;
    ALTTRACE2V(vfyrecv, 0x90001, *(U16 *)(infop->dataAddr + 12), 
	       infop->dataAddr[23], nic->csumFlags, 
	       *(U16 *)(&infop->dataAddr[20]), infop->dataAddr + 14);
    /* check for an IP checksum */
    if (bdp->BD_flags & BD_FLAG_IP_CKSUM) {
      
        if (nic->csumFlags & ALT_CKSUM_DO_RECV_IP) {
	    if (bdp->BD_ip_cksum != 0xffff) {
	        cksumPktInfo.Receive.NdisPacketIpChecksumFailed = 1;
		nic->badIpCksums++;
		ALTTRACEE(badipsum, 0x9000b, bdp->BD_ip_cksum, 0, 0, 0);
	    } else
	        cksumPktInfo.Receive.NdisPacketIpChecksumSucceeded = 1;
	}
    }

    /* check for a TCP/UDP checksum */
    if (bdp->BD_flags & BD_FLAG_TCP_UDP_CKSUM) {
        /* check for unfragmented, we can't handle frags now */
        if (!((*(U16 *)(&infop->dataAddr[20])) & 0xff3f)) { 
	    if (infop->dataAddr[23] == 6) {	    /* TCP protocol */
	        if (nic->csumFlags & ALT_CKSUM_DO_RECV_TCP) {
		    if (bdp->BD_tcp_udp_cksum != 0xffff) {
		        cksumPktInfo.Receive.NdisPacketTcpChecksumFailed = 1;
			nic->badTcpCksums++;
			ALTTRACE2E(badtcpsum, 0x9000c,
				   bdp->BD_tcp_udp_cksum, pkt, 0, 0, 
				   infop->dataAddr + 14);
		    } else
		        cksumPktInfo.Receive.NdisPacketTcpChecksumSucceeded = 
			    1;
		}
	    } else if (infop->dataAddr[23] == 17) { /* UDP protocol */
	        if (nic->csumFlags & ALT_CKSUM_DO_RECV_TCP) {
		    if (bdp->BD_tcp_udp_cksum != 0xffff) {
		        cksumPktInfo.Receive.NdisPacketUdpChecksumFailed = 1;
			nic->badUdpCksums++;
			ALTTRACE2E(badudpsum, 0x9000d,
				   bdp->BD_tcp_udp_cksum, pkt, 0, 0, 
				   infop->dataAddr + 14);
		    } else
		        cksumPktInfo.Receive.NdisPacketUdpChecksumSucceeded = 
			    1;
		}
	    }
	}
    }

    pktExtPtr = NDIS_PACKET_EXTENSION_FROM_PACKET(pkt);
    ALTTRACEV(vfyCksum, 0x90002, bdp->BD_flags, cksumPktInfo.Value,
	      bdp->BD_tcp_udp_cksum, bdp->BD_ip_cksum);
    pktExtPtr->NdisPacketInfo[TcpIpChecksumPacketInfo] = 
        (void *)cksumPktInfo.Value;
}

#if DO_CKSUM_DEBUG
struct in_addr {
    union {
	struct { U8 s_b1, s_b2, s_b3, s_b4; } S_un_b;
	struct { U16 s_w1, s_w2; } S_un_w;
	U32 S_addr;
    } S_un;
#define s_addr S_un.S_addr  /* should be used for all code */
};

struct ip {
    U8 ip_hl:4;  /* header length */
    U8 ip_v:4;   /* version */
    U8 ip_tos;   /* type of service */
    S16 ip_len;   /* total length */
    U16 ip_id;   /* identification */
    S16 ip_off;   /* fragment offset field */
#define IP_DF 0x4000   /* dont fragment flag */
#define IP_MF 0x2000   /* more fragments flag */
    U8 ip_ttl;   /* time to live */
    U8 ip_p;   /* protocol */
    U16 ip_sum;   /* checksum */
    struct in_addr ip_src, ip_dst; /* source and dest address */
};

U16
ckSum_pseudo(struct ip *ipp)
{
    U32 pseudo = htons(ipp->ip_len);

    pseudo -= ipp->ip_hl << 2;	/* Adjust for the len of IP header */
    pseudo += ipp->ip_p;
    pseudo += htons(ipp->ip_src.S_un.S_un_w.s_w1);
    pseudo += htons(ipp->ip_src.S_un.S_un_w.s_w2);
    pseudo += htons(ipp->ip_dst.S_un.S_un_w.s_w1);
    pseudo += htons(ipp->ip_dst.S_un.S_un_w.s_w2);
 
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);
    pseudo = (pseudo >> 16) + (pseudo & 0xffff);

    return((U16)pseudo);
}

/* tcp or udp cksum with a preseeded pseudoheader checksum */
U16
ckSumIt(char *cp, S32 len)
{
    U32 sum = 0;

    if (len < 0)
        return(0xdead);

    if (len & 1) {
        *(cp + len) = 0;	/* pad */
	len += 1;
    }
    while (len > 0) {
        sum += ntohs(*(U16 *)cp);
	cp += 2;
	len -= 2;
    }

    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);

    return (U16)(~sum & 0xffff);
}
#endif /* DO_CKSUM_DEBUG */
#endif /* USE_CKSUM_OFFLOAD */
#endif /* defined(__sun) */
