/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997, 1998 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: altinit.c $
 * Revision api12_3_ted/2 1999/06/18 21:49:11 ted
 * 	Remove OP14's real name from source
 * 
 * Revision ge_lat_leonid/1 1999/03/23 21:10:09 leonid
 * 	removed check for subsystem IDs
 * 
 * Revision ge_l_leonid/1 1999/03/17 22:49:52 leonid
 * 	 fixed the problem with releasing same spinlock twice in de-serialized transmit code
 * 
 * Revision 1.1.2.33  1999/01/29  18:16:20  leonid
 * 	Fixed the failure to report disconnected cable when bringing up the NT.
 * 	[1999/01/29  18:16:00  leonid]
 *
 * Revision 1.1.2.32  1999/01/07  23:09:24  ted
 * 	for debugger make keepalive checks be more or less forever
 * 	[1999/01/07  23:06:51  ted]
 * 
 * Revision 1.1.2.31  1998/12/23  16:30:20  leonid
 * 	MiniportReset changed from synchronous to asynchronous operation to fix 1c_reset
 * 	failures.
 * 	[1998/12/23  16:29:48  leonid]
 * 
 * Revision 1.1.2.30  1998/12/08  16:46:19  leonid
 * 	Fixed a problem with the 1.1.2.29 bsubmit
 * 	[1998/12/08  16:45:22  leonid]
 * 
 * Revision 1.1.2.29  1998/12/01  17:52:27  leonid
 * 	NDIS_RESOURCES status is back (the conformance test has been fixed so we don't have to use "lying send" anymore
 * 	[1998/12/01  17:49:15  leonid]
 * 
 * Revision 1.1.2.28  1998/11/23  22:50:38  ted
 * 	Fix internal queuing problems
 * 	[1998/11/23  22:47:07  ted]
 * 
 * Revision 1.1.2.27  1998/11/16  23:35:06  ted
 * 	Add registry support for the READ_MAX_DMA value
 * 	[1998/11/16  23:34:42  ted]
 * 
 * Revision 1.1.2.26  1998/11/04  17:42:45  hayes
 * 	fix Alpha support
 * 	[1998/11/04  17:42:04  hayes]
 * 
 * Revision 1.1.2.25  1998/11/03  23:25:32  ted
 * 	Rewrite of the NT internal queuing
 * 	[1998/11/03  22:21:49  ted]
 * 
 * Revision 1.1.2.24  1998/11/02  21:48:48  hayes
 * 	cleanup strings for OP14
 * 	[1998/11/02  21:48:19  hayes]
 * 
 * Revision 1.1.2.23  1998/10/13  03:33:38  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:07  ted]
 * 
 * Revision 1.1.2.28  1998/11/23  22:50:38  ted
 * 	Fix internal queuing problems
 * 	[1998/11/23  22:47:07  ted]
 * 
 * Revision 1.1.2.27  1998/11/16  23:35:06  ted
 * 	Add registry support for the READ_MAX_DMA value
 * 	[1998/11/16  23:34:42  ted]
 * 
 * Revision 1.1.2.26  1998/11/04  17:42:45  hayes
 * 	fix Alpha support
 * 	[1998/11/04  17:42:04  hayes]
 * 
 * Revision 1.1.2.25  1998/11/03  23:25:32  ted
 * 	Rewrite of the NT internal queuing
 * 	[1998/11/03  22:21:49  ted]
 * 
 * Revision 1.1.2.24  1998/11/02  21:48:48  hayes
 * 	cleanup strings for OP14
 * 	[1998/11/02  21:48:19  hayes]
 * 
 * Revision 1.1.2.23  1998/10/13  03:33:38  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:07  ted]
 * 
 * Revision 1.1.2.22  1998/09/29  18:43:49  leonid
 * 	For the "out of transmit resources" case, "RESOURCES" status is replaced
 * 	with "SUCCESS" so we can pass certification. Will have to implement
 * 	internal queueing  later on.
 * 	[1998/09/29  18:43:29  leonid]
 * 
 * Revision 1.1.2.21  1998/09/03  22:16:56  hayes
 * 	add bugfixes from nt_r1_9_5
 * 	[1998/09/03  22:13:25  hayes]
 * 
 * Revision 1.1.2.20  1998/09/03  22:06:11  ted
 * 	Bring forward log messages and cache line size fix from release 1.9
 * 	[1998/09/03  22:01:33  ted]
 * 
 * Revision 1.1.2.19  1998/09/03  20:05:05  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:15  ted]
 * 
 * Revision 1.1.2.18  1998/08/20  18:39:38  leonid
 * 	"coal_now" feature is disabled for performance reasons.
 * 	nic->nic_tracing is initialized to enable the keyword parsing
 * 	[1998/08/20  18:39:09  leonid]
 * 
 * Revision 1.1.2.17  1998/08/20  01:15:18  hayes
 * 	remove printf() and KeSetImportanceDpc()
 * 	[1998/08/20  00:50:16  hayes]
 * 
 * Revision 1.1.2.16  1998/08/12  01:58:11  ted
 * 	Deserialize maybe working
 * 	[1998/08/12  01:55:41  ted]
 * 
 * 	bci by hand because bmerge ran out of open file descriptors, sigh
 * 	[1998/08/10  22:48:01  ted]
 * 
 * 	Refine send deserialization
 * 	[1998/08/10  21:28:06  ted]
 * 
 * 	First shot at real deserialization
 * 	[1998/08/06  18:11:00  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:02:04  ted]
 * 
 * Revision 1.1.2.15  1998/08/06  23:25:47  hayes
 * 	added check for link up before calling altQueueSendMsg()
 * 	[1998/08/06  23:25:37  hayes]
 * 
 * Revision 1.1.2.14  1998/08/03  22:06:34  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:16  hayes]
 * 
 * Revision 1.1.2.13  1998/08/03  17:11:35  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:51  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:06  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:30  hayes]
 * 
 * Revision 1.1.2.15  1998/08/06  23:25:47  hayes
 * 	added check for link up before calling altQueueSendMsg()
 * 	[1998/08/06  23:25:37  hayes]
 * 
 * Revision 1.1.2.14  1998/08/03  22:06:34  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:16  hayes]
 * 
 * Revision 1.1.2.13  1998/08/03  17:11:35  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:51  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:06  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:30  hayes]
 * 
 * Revision 1.1.2.15  1998/08/06  23:25:47  hayes
 * 	added check for link up before calling altQueueSendMsg()
 * 	[1998/08/06  23:25:37  hayes]
 * 
 * Revision 1.1.2.14  1998/08/03  22:06:34  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:16  hayes]
 * 
 * Revision 1.1.2.13  1998/08/03  17:11:35  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:51  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:06  hayes]
 * 
 * 	host based send ring now works off RCB flag
 * 	[1998/07/29  18:31:30  hayes]
 * 
 * Revision 1.1.2.12  1998/07/15  00:18:59  ted
 * 	Remove test printf
 * 	[1998/07/15  00:18:30  ted]
 * 
 * Revision 1.1.2.11  1998/07/14  20:30:56  ted
 * 	Fix broken locks in SendPackets, Deserialized only
 * 	[1998/07/14  20:30:36  ted]
 * 
 * Revision 1.1.2.10  1998/07/14  18:45:05  hayes
 * 	fix merging typos...
 * 	[1998/07/14  18:44:55  hayes]
 * 
 * Revision 1.1.2.9  1998/07/14  18:34:56  hayes
 * 	cleanup locks
 * 	[1998/07/14  18:14:55  hayes]
 * 
 * Revision 1.1.2.8  1998/07/14  02:32:47  ted
 * 	Deserialized send, new registry names, shutdown fix, altCheckMsgFrags fix, and NDIS 5
 * 	[1998/07/14  02:16:20  ted]
 * 
 * Revision 1.1.2.7  1998/07/10  16:57:51  hayes
 * 	fix altCheckMsgFrags()
 * 	[1998/07/10  16:57:32  hayes]
 * 
 * Revision 1.1.2.6  1998/07/08  22:26:28  hayes
 * 	add 450gx fix
 * 	[1998/07/08  22:25:35  hayes]
 * 
 * Revision 1.1.2.5  1998/06/26  21:46:01  hayes
 * 	fix build
 * 	[1998/06/26  21:45:46  hayes]
 * 
 * Revision 1.1.2.4  1998/06/18  16:06:49  hayes
 * 	Merged with changes from 1.1.2.3
 * 	[1998/06/18  16:06:33  hayes]
 * 
 * 	complete merge
 * 	[1998/06/18  15:50:37  hayes]
 * 
 * Revision 1.1.2.3  1998/06/18  15:50:52  hayes
 * 	complete merge
 * 	[1998/06/18  15:50:37  hayes]
 * 
 * Revision 1.1.2.2  1998/06/18  00:06:41  hayes
 * 	add caching fixes
 * 	[1998/06/18  00:02:12  hayes]
 * 
 * Revision 1.1.2.1  1998/06/16  16:46:02  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:44:50  ted]
 * 
 * $EndLog$
 */
/*
 *  NOTE: for further history on this file see ntinit.c and ntsend.c which
 * have since been defuncted.
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/ndis/altinit.c,v 1.1.2.33 1999/01/29 18:16:20 leonid Exp $";
#endif /* __lint */

#define ALPHA_DEBUG 0
 
#if (ALPHA_DEBUG)
#ifndef NDIS_WRAPPER
/* Allow us to pick up more definitions from ndis.h */
#define NDIS_WRAPPER
#endif /* NDIS_WRAPPER */
#endif /* ALPHA_DEBUG */

#include <ndis.h>
#include <ntddk.h>
#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"
#include "eeprom.h"
#include "altos.h"
#include "althdr.h"
#include "altreg.h"		/* must follow althdr.h */
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"
#include "drvvers.h"
#include "drvstat.h"

#include "altntmsg.h"

#if (ALPHA_DEBUG)
U32 debug_base;
#endif /* ALPHA_DEBUG */

/* local routine forwards */
STATIC BOOLEAN CheckForHang(NDIS_HANDLE);
STATIC void DisableRupt(NDIS_HANDLE);
STATIC void EnableRupt(NDIS_HANDLE);
STATIC void HaltInstance(NDIS_HANDLE);
STATIC void ShutdownInstance(NDIS_HANDLE);
STATIC NDIS_STATUS InitInstance(PNDIS_STATUS, PUINT, PNDIS_MEDIUM, UINT,
				 NDIS_HANDLE, NDIS_HANDLE);
STATIC void ALTINTR(bool *, bool *, void *);
NDIS_STATUS InitAdapter(struct altinst *, NDIS_HANDLE);
STATIC NDIS_STATUS InitialInit(struct altinst *, struct tg_nic *);
STATIC NDIS_STATUS ReadRegistryParameters(struct altinst *altp,
					  struct tg_nic *nic);
STATIC void ReleaseResources(struct altinst *);
STATIC NDIS_STATUS ReadAdapterConfiguration(struct tg_nic *nic);
STATIC NDIS_STATUS RegisterNIC(struct tg_nic *nic, tgPciConf_t *pcip);
STATIC BOOLEAN SixtyFourBitSlotProblem(struct altinst *, struct tg_nic *);
STATIC void SetAdapterResources(struct tg_nic *, PNDIS_RESOURCE_LIST);
STATIC void RedundTimer(void *, void *, void *, void *);
STATIC void LinkTimer(void *, void *, void *, void *);
STATIC NDIS_STATUS TransUnixErrnoToNdis(int);
#if !ALT_DESERIALIZE
STATIC void SendPackets(NDIS_HANDLE, PPNDIS_PACKET, UINT);
#endif /* !ALT_DESERIALIZE */

/*
 * NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
 *                      IN PUNICODE_STRING RegistryPath)
 *
 * Input
 * DriverObject is a pointer to the DRIVER_OBJECT structure. This structure is
 * defined in ndis.h.
 * RegistryPath is a pointer to UNICODE_STRING structure defined in ndis.h.
 *
 * Output
 * STATUS_SUCCESS is returned in case driver is loaded.
 * STATUS_UNSUCCESSFUL is returned in case of error. This means driver would
 * not be loaded.
 * Description
 * This is the entry point routine. DriverEntry name is made visible by the
 * linker.
 * If this name is changed, the makefile needs to be changed so that it 
 * has the same name.
 * In this function we register our MAC driver. This make the entry points 
 * in our driver accessable to the networking layers in the system.
 *
 */

NTSTATUS 
DriverEntry(PDRIVER_OBJECT driverObj,
	    PUNICODE_STRING registryPath)
{

    NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    NDIS_MINIPORT_CHARACTERISTICS altChar; /* hand info to upper layers */
    NDIS_HANDLE wrapperHandle;	   /* handle for wrapper */

    /* Initialize the wrapper. */
    NdisMInitializeWrapper(&wrapperHandle, driverObj, 
			   registryPath, NULL);
    NdisZeroMemory(&altChar, sizeof(NDIS_MINIPORT_CHARACTERISTICS));

    /*
     * Initialize the Miniport characteristics for the call to
     * NdisMRegisterMiniport
     */
    altChar.MajorNdisVersion = ALT_NDIS_MAJOR_VERSION;
    altChar.MinorNdisVersion = ALT_NDIS_MINOR_VERSION;
    altChar.Reserved = 0x00;
    altChar.CheckForHangHandler = CheckForHang;
    altChar.DisableInterruptHandler = DisableRupt;
    altChar.EnableInterruptHandler = EnableRupt;
    altChar.HaltHandler = HaltInstance;
    altChar.HandleInterruptHandler = ALTINTRWORK;
    altChar.InitializeHandler = InitInstance;
    altChar.ISRHandler = ALTINTR;
    altChar.QueryInformationHandler = altQueryInfo;
    altChar.ReconfigureHandler = NULL;
    altChar.ResetHandler = ResetInstance;
    altChar.SetInformationHandler = altSetInfo;
    altChar.TransferDataHandler = NULL;
    altChar.ReturnPacketHandler = altRecvDataRtn;
    altChar.SendPacketsHandler = SendPackets;
    altChar.AllocateCompleteHandler = NULL;

    /* call wrapper to register our driver. */
    if (NdisMRegisterMiniport(wrapperHandle, &altChar, sizeof(altChar)) != 
	NDIS_STATUS_SUCCESS) {
	return(STATUS_UNSUCCESSFUL);
    }
    return(STATUS_SUCCESS);
}

/*
 *  NDIS_STATUS InitInstance(OUT PNDIS_STATUS OpenErrorStatus,
 *                           OUT PUINT SelectedMediumIndex,
 *                           IN PNDIS_MEDIUM MediumArray,
 *                           IN UINT MediumArraySize,
 *                           IN NDIS_HANDLE MiniportAdapterHandle,
 *                           IN NDIS_HANDLE configurationHandle
 *           
 * Input
 * OpenErrorStatus         pointer to NDIS_STATUS specifying error if returned
 *                         value is  NDIS_STATUS_OPEN_ERROR.
 * SelectedMediumIndex     pointer to UINT specifying the type of media driver
 *                         supports.
 * MediumArray             pointer to NDIS_MEDIUM supported by NDIS library.
 * MediumArraySize         size of the MediumArray.
 * MiniportAdapterHandle   handle which is used as argument to the NDIS
 *                         library functions.
 * configurationHandle     handle used in calls to NdisOpenConfiguration.
 *
 * Output
 * NDIS_SUCCESS             driver loaded successfully
 *                          otherwise error code
 * Description
 * This function is called after the DriverEntry call. The driver is required
 * to return the media type it supports, allocate all the resources that it
 * needs for the adapter, like memory, irq etc.
 * In this function the adapter is made operational.
 */

STATIC NDIS_STATUS 
InitInstance(PNDIS_STATUS openErrorStatus,
	     PUINT selectedMediumIndex,
	     PNDIS_MEDIUM mediumArray,
	     UINT mediumArraySize,
	     NDIS_HANDLE miniportAdapterHandle,
	     NDIS_HANDLE configurationHandle)
{
    U32 i;
    NDIS_STATUS status;
    struct altinst *altp = NULL;
    struct tg_nic *nic;

    LOCK_INIT(&altTraceMu, 0, 0, 0);
    alt_itrace();

    /*
     * Check if we support the media. We look like 802.3 to the 
     * protocol layers. 
     */
    for (i = 0; i < mediumArraySize; i++) {
	/* We are telling upper layer that we are ethernet type */
	if (mediumArray[i] == NdisMedium802_3) {
	    break;
	}
    }

    if (i == mediumArraySize) {
	LOG0(miniportAdapterHandle, NDIS_STATUS_UNSUPPORTED_MEDIA);
	return(NDIS_STATUS_UNSUPPORTED_MEDIA);
    }
    *selectedMediumIndex = i;

    /* Now allocate memory for the adapter instance */
    MEMZALLOC(altp, struct altinst *, sizeof(struct altinst), 0);
    if (!altp) {
	LOG1(miniportAdapterHandle, ALT_LOG_ERR_MEM_ERR1, 1);
	return(NDIS_STATUS_RESOURCES);
    }

    if (alt_InitNICstruct(altp)) {
	LOG1(miniportAdapterHandle, ALT_LOG_ERR_MEM_ERR1, 2);
	MEMFREE(altp, sizeof(struct altinst), 0);
	return(NDIS_STATUS_RESOURCES);
    }
    /* set up our local pointer */
    nic = altp->initNic;

    /* make sure we have the proper name... */
    strcpy(nic->name, VER_INTERNALNAME_STR);

    /* used as argument in NDIS Calls. */
    altp->look_ahead_size = JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE;

    /* used in calls to NdisReadConfig */
    nic->configHandle = configurationHandle;

    nic->handle = miniportAdapterHandle;

    ERROR1("Calling read registry parms\n");
    if ((status = ReadRegistryParameters(altp, nic)) != NDIS_STATUS_SUCCESS) {
	/*
	 * ReadRegistryParameters does its own logging...
	 */
	alt_FreeNICstruct(altp);
	MEMFREE(altp, sizeof(struct altinst), 0);
	return(status);
    } 

    LOCK_INIT(&nic->sendReapMu, 0, 0, 0);
#if ALT_DESERIALIZE
    LOCK_INIT(&nic->infopQueueMu, 0, 0, 0);
    LOCK_INIT(&nic->sendQueueMu, 0, 0, 0);
#endif /* ALT_DESERIALIZE */

    altp->ifconfigNic = nic;	/* primary interface */
    nic->initAltp = altp;		/* who inited me... */

    if ((status = InitialInit(altp, nic)) != NDIS_STATUS_SUCCESS) {
	LOCK_DESTROY(&nic->sendReapMu);
#if ALT_DESERIALIZE
	LOCK_DESTROY(&nic->infopQueueMu);
	LOCK_DESTROY(&nic->sendQueueMu);
#endif /* ALT_DESERIALIZE */
        alt_FreeNICstruct(altp);
        MEMFREE(altp, sizeof(struct altinst), 0);
	return(status);
    }

    if ((status = InitAdapter(altp, nic->handle)) !=
	NDIS_STATUS_SUCCESS) {
	LOCK_DESTROY(&nic->sendReapMu);
#if ALT_DESERIALIZE
	LOCK_DESTROY(&nic->infopQueueMu);
	LOCK_DESTROY(&nic->sendQueueMu);
#endif /* ALT_DESERIALIZE */
        alt_FreeNICstruct(altp);
        MEMFREE(altp, sizeof(struct altinst), 0);
	return(status);
    }
    ALTTRACE(drv_regd, 0x30000, altp, altp->initNic, &altp->initNic, 
	     altp->initNic->genComm);

   if ((nic->flags & ALT_LINK_READY) == 0)  
      LOG1(nic->handle, ALT_LOG_LINK_DOWN, nic->initAltp->instance);

    return(NDIS_STATUS_SUCCESS);
}

/*
 * NDIS_STATUS InitialInit(struct altinst *altp,
 *                         struct tg_nic *nic)
 *
 * This function performs initial init of the card.
 * It reads adapter resources and acquires it.
 * It then calls alt_initph1().
 */

STATIC NDIS_STATUS 
InitialInit(struct altinst *altp, struct tg_nic *nic)
{
    U32 map_register_count;
    NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    int rc;

    NdisMSetAttributesEx(nic->handle, /* Handle supplied by the wrapper */
			 (NDIS_HANDLE)nic, /* Private Adapter Context */
#if NT_USE_DEBUGGER
			 0xffff,      /* big as we can get */
#else /* NT_USE_DEBUGGER */
			 15,	      /* Timeout, XXX tie with stat_ticks */
#endif /* NT_USE_DEBUGGER */
#if ALT_DESERIALIZE
			 NDIS_ATTRIBUTE_DESERIALIZE |
#endif /* ALT_DESERIALIZE */
			 NDIS_ATTRIBUTE_BUS_MASTER,
			 NdisInterfacePci);

    /* 
    /* 
     * Read the info from the card...  XXX Poorly named routine, this
     * also maps the card into the system
     */
    if ((status = ReadAdapterConfiguration(nic)) != NDIS_STATUS_SUCCESS) {
	/*
	 * ReadAdapterConfiguration() does its own logging
	 */
	return(status);
    } 

    /*
     * clear MBOX so that NIC can interrupt us PR #425
     */
    OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 0);

    if ((rc = alt_InitNICmem(altp))) {
	LOG1(nic->handle, ALT_LOG_ERR_INIT_MEM, 1);
	NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
	return(TransUnixErrnoToNdis(rc));
    }

    if ((rc = alt_InitNICph1(altp))) {
	LOG0(nic->handle, ALT_LOG_ERR_INIT_PH1);
	alt_FreeNICmem(altp);
	NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
	return(TransUnixErrnoToNdis(rc));
    }

    return(NDIS_STATUS_SUCCESS);
}

/*
 * NDIS_STATUS InitAdapter(struct altinst *altp, NDIS_HANDLE handle)
 *
 * This function will register interrupt and will initialize hardware.
 */

STATIC NDIS_STATUS 
InitAdapter(struct altinst *altp, NDIS_HANDLE handle)
{
    NDIS_STATUS status = NDIS_STATUS_FAILURE;
    struct tg_nic *nic = altp->nic;
    int rc;

    if (altp->flags & ALT_REDUNDANT) 
	NdisMInitializeTimer(&nic->rslTimer, nic->handle, 
			     RedundTimer, altp);

    status = NdisMRegisterInterrupt(&nic->ruptStruct,
				    handle,
				    nic->ruptLevel,
				    nic->ruptLevel,
				    TRUE,	/* call on every interrupt */
				    TRUE,	/* shared interrupts  */
				    NdisInterruptLevelSensitive);
    if (status != NDIS_STATUS_SUCCESS) {
	LOG1(nic->handle, ALT_LOG_ERR_INIT_INTERRUPT, status);
	NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
	alt_FreeNICstruct(altp);
	MEMFREE(altp, sizeof(struct altinst), 0);
	return(status);
    } 

    if (SixtyFourBitSlotProblem(altp, nic)) {
	LOG0(nic->handle, ALT_LOG_ERR_64_BIT_PROBLEM);
	ReleaseResources(altp);
	return(NDIS_STATUS_NOT_ACCEPTED);
    }

#if !defined(OP14)
    /* must precede ph2 init */
    if (nic->flags & ALT_JUMBO_ENABLED) {
	if (altInitJumbo(altp)) {
	    LOG1(nic->handle, ALT_LOG_ERR_INIT_MEM, 2);
	    ReleaseResources(altp);
	    return(NDIS_STATUS_RESOURCES);
	}
    }
#endif /* !defined(OP14) */

    if ((rc = alt_InitNICph2(altp))) {
	LOG0(nic->handle, ALT_LOG_ERR_INIT_PH2);
	ReleaseResources(altp);
	return(TransUnixErrnoToNdis(rc));
    }

    /* the redundancy timer needs to run every 1/4 of a second */
    if (altp->flags & ALT_REDUNDANT) {
	NdisMSetPeriodicTimer(&nic->rslTimer, 250);
	nic->flags |= ALT_TIMER_RUNNING;
    }

    NdisMRegisterAdapterShutdownHandler(nic->handle, nic, ShutdownInstance);

    return NDIS_STATUS_SUCCESS;
}

STATIC VOID 
ReleaseResources(struct altinst *altp)
{
    struct tg_nic *nic = altp->nic;

    NdisMDeregisterInterrupt(&nic->ruptStruct);

    /* free NIC memory */
    alt_FreeNICmem(altp);
    NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
    if (nic)
	alt_FreeNICstruct(altp);
    MEMFREE(altp, sizeof(struct altinst), 0);
}

/* for ReadAdapterConfiguration */
#define NDIS_MEM_MODE (NDIS_MEMORY_CONTIGUOUS | NDIS_MEMORY_NONCACHED)
#define MAX_PCI_SLOT 255

STATIC NDIS_STATUS 
ReadAdapterConfiguration(struct tg_nic *nic)
{
    NDIS_STATUS status;
    tgPciConf_t *pcip;
    U32 length;
    U32 i;
    bool card_search = FALSE;
    bool card_found = FALSE;

    MEMZALLOC(pcip, tgPciConf_t *, sizeof(tgPciConf_t), NDIS_MEM_MODE);
    if (!pcip)
	return(NDIS_STATUS_RESOURCES);

    if (NdisReadPciSlotInformation(nic->handle, nic->slot, 0,
				   (void *)pcip, sizeof(tgPciConf_t)) < 
	sizeof(tgPciConf_t)) {
	card_search = TRUE;
    }

    /*
     * only look at the proper cards...  Keep OEM's happy...
     */
    if (!card_search &&
	((pcip->device_id == PCI_DEV_ID) &&
	 (pcip->vendor_id == PCI_VEND_ID))
#if defined(ALTEON)
	|| 
	((pcip->device_id == OEM_UNIVERSAL_PCI_DEV_ID) &&
	 (pcip->vendor_id == OEM_UNIVERSAL_PCI_VEND_ID) &&
	 (pcip->subsystem_id == OEM_UNIVERSAL_PCI_SUBSYS_ID) &&
	 (pcip->subsystem_vendor_id == OEM_UNIVERSAL_PCI_SUBSYS_VEND_ID))
#endif /* defined(ALTEON) */
	) {
	/*
	 * now do the actual verification and registering of the NIC...
	 */
	status = RegisterNIC(nic, pcip);
	MEMFREE(pcip, sizeof(tgPciConf_t), NDIS_MEM_MODE);
	return(status);
    } else {
	card_search = TRUE;
    }
    /*
     * either there was no card in the slot that we expected, or the card
     * did not match what was in the registry...
     *
     * search all slots for a card that matches, and insure that is NOT
     * in the registry.  We are expecting that a card was moved from
     * one slot to another.  If we find no unclaimed cards, then the
     * card must have been removed...
     */
    if (card_search) {
	for (i = 0; i < MAX_PCI_SLOT; i++) {
	    if (i == nic->slot) {
		/*
		 * skip over the card we already checked...
		 */
		continue;
	    }
	    if (NdisReadPciSlotInformation(nic->handle, i, 0, 
					   (void *)pcip, 
					   sizeof(tgPciConf_t)) < 
		sizeof(tgPciConf_t)) {
		/*
		 * no pci config info, slot must be empty...
		 */
		continue;
	    }
	    /*
	     * only look at the proper cards...  Keep OEM's happy...
	     */
	    if (((pcip->device_id == PCI_DEV_ID) &&
		 (pcip->vendor_id == PCI_VEND_ID) &&
		 (pcip->subsystem_id == PCI_SUBSYS_ID) &&
		 (pcip->subsystem_vendor_id == PCI_SUBSYS_VEND_ID))
#if defined(ALTEON)
		|| 
		((pcip->device_id == OEM_UNIVERSAL_PCI_DEV_ID) &&
		 (pcip->vendor_id == OEM_UNIVERSAL_PCI_VEND_ID) &&
		 (pcip->subsystem_id == OEM_UNIVERSAL_PCI_SUBSYS_ID) &&
		 (pcip->subsystem_vendor_id == 
		  OEM_UNIVERSAL_PCI_SUBSYS_VEND_ID))
#endif /* defined(ALTEON) */
		) {
		/*
		 * this is one of our cards, check to insure that the
		 * registry doesn't already know about it...
		 */
		/* XXX unimplemented right now, just fails */
	    }
	}
	/*
	 * make sure we update the registry with the new information...
	 */
	/* XXX unimplemented right now, just fails */
    }

    MEMFREE(pcip, sizeof(tgPciConf_t), NDIS_MEM_MODE);
    return(NDIS_STATUS_FAILURE);
}

STATIC NDIS_STATUS 
RegisterNIC(struct tg_nic *nic, tgPciConf_t *pcip)
{
    NDIS_STATUS status;
    bool fix_latency_timer = FALSE;
    PNDIS_RESOURCE_LIST resourceList;
    U32 tmp_cfg;
    U32 map_register_count;

    /*
     * dump the PCI config region to the system log if requested...
     */
    if (nic->init_flags & ALT_INIT_DEBUG_PCI) {
	NdisWriteErrorLogEntry(nic->handle,
			       (NDIS_ERROR_CODE)ALT_LOG_INFO_PCI_DEBUG_PRE, 16,
			       *((U32 *)pcip + 0),
			       *((U32 *)pcip + 1),
			       *((U32 *)pcip + 2),
			       *((U32 *)pcip + 3),
			       *((U32 *)pcip + 4),
			       *((U32 *)pcip + 5),
			       *((U32 *)pcip + 6),
			       *((U32 *)pcip + 7),
			       *((U32 *)pcip + 8),
			       *((U32 *)pcip + 9),
			       *((U32 *)pcip + 10),
			       *((U32 *)pcip + 11),
			       *((U32 *)pcip + 12),
			       *((U32 *)pcip + 13),
			       *((U32 *)pcip + 14),
			       *((U32 *)pcip + 15));
    }

#if defined(ALPHA)
    /*
     * allocate map registers-
     */
    map_register_count = 0;
    status = NdisQueryMapRegisterCount(NdisInterfacePci,
                                       &map_register_count);
    /*
     * NdisQueryMapRegisterCount() does not return NDIS_STATUS_SUCCESS
     * on x86 platforms for some reason.  We are ignoring the return code
     * here. XXX
     *
     * This returns the max the system can supply globally, NOT the
     * number currently available.
     */
 
    if (status != NDIS_STATUS_SUCCESS) {
        LOG2(nic->handle, ALT_LOG_ERR_MAP_REG, 1, status);
        ALTTRACEE(mapregQ, 0x30030, map_register_count, status, 0, 0);
        NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
        return(NDIS_STATUS_RESOURCES);
    }
    if (map_register_count < SEND_RING_ENTRIES) {
        LOG2(nic->handle, ALT_LOG_ERR_MAP_REG, 2, status);
        ALTTRACEE(mapregQx, 0x30031, map_register_count, status, 0, 0);
        NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
        return(NDIS_STATUS_RESOURCES);
    }
#endif /* defined(ALPHA) */
    map_register_count = SEND_RING_ENTRIES;
 
    /*
     * Map regs are a fixed resource on alpha platforms.
     * Most systems have 2048 globally allocated map regs.
     *
     * This has been changed with the NT4.0 DEC OEM Rev D HAL.
     * 5/27/98.  This HAL version is now required to support
     * release 1.9.x and greater of ALteon Networks NT drivers
     * on alpha platforms.  The revision of the HAL is also included
     * in NT4SP4.
     */
    status = NdisMAllocateMapRegisters(nic->handle, 0, TRUE,
        map_register_count,
        (nic->flags & ALT_JUMBO_ENABLED)?JUMBOBLOCKSIZE:MAXBLOCKSIZE);
 
    /*
     * fail completely if we couldn't the necessary map regs...
     */
    if (status != NDIS_STATUS_SUCCESS) {
        LOG3(nic->handle, ALT_LOG_ERR_MAP_REG, 3, status, map_register_count);
        ALTTRACEE(mapreg_A, 0x30032, map_register_count, status,
            (nic->flags & ALT_JUMBO_ENABLED)?JUMBOBLOCKSIZE:MAXBLOCKSIZE,
            0);
        NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
        return(NDIS_STATUS_RESOURCES);
    }
 
    if ((status = NdisMPciAssignResources(nic->handle, nic->slot,
                                          &resourceList)) !=
        NDIS_STATUS_SUCCESS) {
        LOG0(nic->handle, ALT_LOG_ERR_NO_ASSIGN_PCI_RESOURCES);
        NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
        return(status);
    }
 
    SetAdapterResources(nic, resourceList);
 
    status = NdisMMapIoSpace(&nic->base, nic->handle,
                             nic->unmappedBase, nic->memLength);
 
#if (ALPHA_DEBUG)
    {
        U32 bus_number;
        U32 address_space;
        U32 debug_status;
        NDIS_PHYSICAL_ADDRESS phy_addr;
 
        bus_number = ((PNDIS_ADAPTER_BLOCK)(nic->handle))->BusNumber;
        address_space = 0x04;        /* kernel mode dense */
 
        debug_status = HalTranslateBusAddress(PCIBus,
                                bus_number,
                                nic->unmappedBase,
                                &address_space,
                                &phy_addr);
 
        if (debug_status != TRUE) {
            goto alpha_debug_done;
        }
 
        debug_base = MmMapIoSpace(phy_addr,
                                 nic->memLength,
                                 0); /* cache enable */
 
        DbgPrint("altndis: MmMapIoSpace: debug_base = 0x%08x\n", debug_base);
 
alpha_debug_done:
        ;
    }
#endif /* ALPHA_DEBUG */

    if (status == NDIS_STATUS_SUCCESS) {

 
#if !defined(ALPHA)
        /*
         * only re-read pcip on intel systems.  For alpha systems,
         * this is QVA (quasi virtual address) space and will fault
         * the system in strange and unusual ways.
         */
        pcip = (tgPciConf_t *) nic->base;
#endif /* !defined(ALPHA) */

	/*
	 * do some additional sanity checking here, to insure that we have
	 * been properly configured on the PCI bus...
	 */
        tmp_cfg = INCFG32(nic->base, nic->base, TG_STATUS);
        if ((tmp_cfg &
             (PCI_CONF_CMD_BUS_MASTER | PCI_CONF_CMD_MEMORY)) !=
            (PCI_CONF_CMD_BUS_MASTER | PCI_CONF_CMD_MEMORY)) {
	    /*
	     * must must be configured to be a bus master and have PCI
	     * memory space enabled
	     */
            LOG2(nic->handle, ALT_LOG_ERR_INVAL_PCI_CONF, 1, tmp_cfg);
            NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
            return(NDIS_STATUS_RESOURCES);
	}

        tmp_cfg = INCFG32(nic->base, nic->base, TG_BIST);
        tmp_cfg = (tmp_cfg & 0x0000ff00) >> 8;
        if (!tmp_cfg) {
            /*
             * System has failed to set up the latency timer.
             * If registry has a valid latency timer value, use that
             * one, otherwise, write a log message to the system.
             */
#if !defined(ALPHA)
            if (nic->latTimer) {
                pcip->latency_timer = nic->latTimer;
            } else
#endif /* !defined(ALPHA) */
            {
                LOG2(nic->handle, ALT_LOG_ERR_INVAL_PCI_CONF, 3,
                     tmp_cfg);
 
                NdisMUnmapIoSpace(nic->handle, nic->base, nic->memLength);
                return(NDIS_STATUS_RESOURCES);
            }
        }

	if (nic->init_flags & ALT_INIT_DEBUG_PCI) {
	    NdisWriteErrorLogEntry(nic->handle,
				   (NDIS_ERROR_CODE) ALT_LOG_INFO_PCI_DEBUG_POST, 
				   16, 
				   *((U32 *)pcip + 0),
				   *((U32 *)pcip + 1),
				   *((U32 *)pcip + 2),
				   *((U32 *)pcip + 3),
				   *((U32 *)pcip + 4),
				   *((U32 *)pcip + 5),
				   *((U32 *)pcip + 6),
				   *((U32 *)pcip + 7),
				   *((U32 *)pcip + 8),
				   *((U32 *)pcip + 9),
				   *((U32 *)pcip + 10),
				   *((U32 *)pcip + 11),
				   *((U32 *)pcip + 12),
				   *((U32 *)pcip + 13),
				   *((U32 *)pcip + 14),
				   *((U32 *)pcip + 15));
	}
    }
    return(status);
}

STATIC bool
CheckForHang(NDIS_HANDLE context)
{
    bool rc;
    struct tg_nic *nic = (struct tg_nic *)context;
    struct altinst *altp;

    altp = nic->altp;
    rc = altWatchdog(altp);
    return(rc);
}

STATIC void 
DisableRupt(IN NDIS_HANDLE context)
{

    struct tg_nic *nic = (struct tg_nic *)context;
    U32 foo;

    if (!nic || (nic->flags & ALT_IN_DRV_DISABLE))
	/* We're already disabled, just return  */
	return;

    /* either the flag is FALSE or it's not our rupt */
    /* disable rupts and reset the rupt line (may not be necessary ) */
    /* OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 1); */
    WRSHMEM(nic->genComm->gen_com.maskRupts, 1);
    /* read it back to ensure it */
    foo = RDSHMEM(nic->genComm->gen_com.maskRupts);
    nic->flags |= ALT_IN_DRV_DISABLE;
}

STATIC void 
EnableRupt(IN NDIS_HANDLE context)
{

    struct tg_nic *nic = (struct tg_nic *)context;
    U32 foo;

    /* if we were in driver disable, enable rupts and clear the flag */
    if (nic && (nic->flags & ALT_IN_DRV_DISABLE)) {
	ALTTRACEG(enable, 0x30001, nic->flags, 0, 0, 0);
	WRSHMEM(nic->genComm->gen_com.maskRupts, 0);
	/* read it back to ensure it */
	foo = RDSHMEM(nic->genComm->gen_com.maskRupts);
	nic->flags &= ~ALT_IN_DRV_DISABLE;
    }
}

STATIC void 
ShutdownInstance(NDIS_HANDLE context)
{
    struct tg_nic *nic = (struct tg_nic *)context;
    struct altinst *altp = nic->initAltp;
    bool timerCancelled;

    /* stop the card */
    altStopCard(altp);

    /* clean up any send resources out in the card XXX */
}

STATIC void 
HaltInstance(NDIS_HANDLE context)
{
    struct tg_nic *nic = (struct tg_nic *)context;
    struct altinst *altp = nic->initAltp;
    bool timerCancelled;

    /* cancel the redundancy timer */
    if (nic->flags & ALT_TIMER_RUNNING) {
	NdisMCancelTimer(&nic->rslTimer, &timerCancelled);
	nic->flags &= ~ALT_TIMER_RUNNING;
    }

    /* stop the card */
    altStopCard(altp);
    
    /* free the resources */
    NdisMDeregisterAdapterShutdownHandler(nic->handle);
    ReleaseResources(altp);	/* this releases all the resources */
}

NDIS_STATUS 
ResetInstance(bool *addrReset, NDIS_HANDLE context)
{
    struct tg_nic *nic = (struct tg_nic *)context;
    struct altinst *altp = nic->initAltp;
#if USE_INTERNAL_QUEUE
    NDIS_PACKET *packetp;
#endif /* USE_INTERNAL_QUEUE */

    ALTTRACEE(altReset, 0x30002, nic, 0, 0, 0);
    /* save off the statistics */
    MEMCPY((MEMCPY_TYPE)nic->genInfoStats2, 
	   (MEMCPY_TYPE)&nic->genInfo->stats, 
	   sizeof(struct tg_stats));

    /* clean up the stats, a few stats shouldn't carry over */
    nic->genInfoStats2->nicRecvBDs = 0;
    nic->genInfoStats2->nicJumboRecvBDs = 0;
    nic->genInfoStats2->nicMiniRecvBDs = 0;
    nic->genInfoStats2->nicTotalRecvBDs = 0;

    /* stop the card */
    altStopCard(altp);

    /* clean up the counters */
    nic->bufsInStdRecvRing = 0;
    nic->recvStdCoal = 0;
    nic->bufsInJumboRecvRing = 0;
    nic->recvJumboCoal = 0;
#if USE_MINI_RING
    nic->bufsInMiniRecvRing = 0;
    nic->recvMiniCoal = 0;
#endif /* USE_MINI_RING */

    /* don't worry about the recv return ring, we've already handled 
     * everything that could be on it */

    /*
     * Send Ring - we must cleanup all pending packets...
     *
     * For non-deserialized driver, do not need completion,
     * but we cannot touch those pkts after returning from
     * reset().
     *
     * Cleanup all outstanding sends in the NIC and then
     * all the frames we've queued internally
     *
     */
    if (nic->send_consumer != nic->send_producer)
	altFreeSendMsgs(nic, nic->send_consumer, nic->send_producer);
#if USE_INTERNAL_QUEUE
    while((packetp = nic->pktQueue)) {
        nic->pktQueue = *(NDIS_PACKET **)packetp->MiniportReserved;
	MSGFREE(packetp, MSG_SUCCESS);
    }
#endif /* USE_INTERNAL_QUEUE */    

    /* indicate to the firmware operational case that we need to
     * reset the statistics to wherever they were */
    nic->flags |= ALT_RESET_STATS;

    /* restart the card */
    if (alt_InitNICph2(altp)) {
	LOG0(nic->handle, ALT_LOG_ERR_INIT_PH2);
	return(NDIS_STATUS_FAILURE);
    }

    NdisMInitializeTimer(&nic->linkTimer, nic->handle, LinkTimer, altp);
    nic->linkTimeout = 0;
    NdisMSetPeriodicTimer(&nic->linkTimer, 100);

    return(NDIS_STATUS_PENDING);
}

 /*
 * This is the timer procedure scheduled to run 
 * every 100 milliseconds to detect link negotiation completed.
 */

STATIC void 
LinkTimer(void *systemSpecific1, void *context, void *systemSpecific2,
	    void *systemSpecific3)
{
    bool timerCancelled;	
    struct altinst *altp = (struct altinst *)context;
    struct tg_nic *nic = altp->nic;

    nic->linkTimeout ++;

    /* Wait either till the link is up or till 5 sec timeout expires */
    if ((nic->flags & ALT_LINK_READY) || (nic->linkTimeout > 50))  
    {
     NdisMCancelTimer(&nic->linkTimer, &timerCancelled);
     NdisMResetComplete((NDIS_HANDLE)nic->handle, NDIS_STATUS_SUCCESS, FALSE);
    }
}

STATIC NDIS_STATUS
ReadRegistryParameters(struct altinst *altp,
                       struct tg_nic *nic)
{
    NDIS_STATUS status;
    alt_nt_inst_parms_t *inst_p;
    U32 inst;

    /* read the tune parms out of the registry */
    if ((status = altReadTuneParms(nic, &inst)) != NDIS_STATUS_SUCCESS) {
        return(status);
    }
 
    /* common parameters */
    nic->stat_ticks = nic->tuneParms.glob.stat_ticks;
    nic->send_bd_ticks = nic->tuneParms.glob.send_coal_ticks;
    nic->recv_bd_ticks = nic->tuneParms.glob.recv_coal_ticks;
 
    nic->send_bd_coal = nic->tuneParms.glob.send_coal_max;
    nic->recv_bd_coal = nic->tuneParms.glob.recv_coal_max;

    if (nic->tuneParms.glob.pci_mem_inv)
	nic->init_flags |= ALT_INIT_MEM_WR_INV_DISABLE;
    else
	nic->init_flags &= ~ALT_INIT_MEM_WR_INV_DISABLE;

    if (nic->tuneParms.glob.sharedRupt)
        nic->init_flags |= ALT_INIT_SHARED_RUPT;
    else
        nic->init_flags &= ~ALT_INIT_SHARED_RUPT;

    nic->latTimer = (U8)nic->tuneParms.glob.pci_latency_timer;
    nic->ruptTimeout = nic->tuneParms.glob.ruptTimeout;
    nic->hybridLen = nic->tuneParms.glob.hybridLen;
    nic->minDma = nic->tuneParms.glob.minDma;

    if (nic->tuneParms.glob.i450gx_fix)
	nic->init_flags |= ALT_INIT_450GX_FIX;
    else
	nic->init_flags &= ~ALT_INIT_450GX_FIX;
 
    if (nic->tuneParms.glob.reapAfterSend)
	nic->init_flags |= ALT_INIT_REAP_AFTER_SEND;
    else
	nic->init_flags &= ~ALT_INIT_REAP_AFTER_SEND;
 
    if (nic->tuneParms.glob.sendUpdateOnly)
	nic->init_flags |= ALT_INIT_SEND_UPDATE_ONLY;
    else
	nic->init_flags &= ~ALT_INIT_SEND_UPDATE_ONLY;

    /*
     * XXX finish this...
     */
    switch (nic->tuneParms.glob.pci_read_max) {
        case 4:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_4;
            break;
 
        case 16:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_16;
            break;
 
        case 32:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_32;
            break;
 
        case 64:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_64;
            break;

        case 128:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_128;
            break;
 
        case 256:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_256;
            break;
 
        case 1024:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_1K;
            break;
 
        case 0:
        default:
            nic->pci_read_max = TG_PCI_STATE_DMA_READ_MAX_DISA;
    }
 
    /* instance parameters */
    inst_p = &(nic->tuneParms.inst[inst]);

    nic->slot = inst_p->slot;
    altp->instance = inst;
 
    /* link stuff */
 
    if (inst_p->fdr_filter) {
        nic->flags |= ALT_DO_FDR_FILTERING;
    } else {
        nic->flags &= ~ALT_DO_FDR_FILTERING;
    }
 
    if (inst_p->rx_flow_control) {
        nic->flags |= ALT_RX_FLOW_CONTROL;
    } else {
        nic->flags &= ~ALT_RX_FLOW_CONTROL;
    }
 
    if (inst_p->tx_flow_control) {
        nic->flags |= ALT_TX_FLOW_CONTROL;
    } else {
        nic->flags &= ~ALT_TX_FLOW_CONTROL;
    }
 
    if (inst_p->link_negotiation) {
        nic->flags |= ALT_DO_LINK_NEGOTIATION;
    } else {
	nic->flags &= ~ALT_DO_LINK_NEGOTIATION;
    }

    if (inst_p->jumbo_frames) {
	nic->flags |= ALT_JUMBO_ENABLED;
	LOG0(nic->handle, ALT_LOG_INFO_JUMBO_ENABLED);
    } else {
	nic->flags &= ~ALT_JUMBO_ENABLED;
    }
 
    altTracing = inst_p->host_tracing;
    nic->nic_tracing = inst_p->nic_tracing;

    if (inst_p->debug_pci)
	nic->init_flags |= ALT_INIT_DEBUG_PCI;
    else
	nic->init_flags &= ~ALT_INIT_DEBUG_PCI;
 
    return(NDIS_STATUS_SUCCESS);
}

/*
 *    BOOLEAN SixtyFourBitSlotProblem(struct altinst *altp, 
 *                                    struct tg_nic *nic)
 * 
 *    This function checks if we has a 64 bit bus problem
 *    Read TG_PCI_STATE Register. If bit 20 of this register is 1 then
 *    we are in a 32 bit slot. We do not need to do anything.
 *    If this bit is zero, then the system thinks it is in a 64 bit slot.
 *    Here the fun begins.
 *    Write a dword amount of data (5A5A5A5A) at some location
 *     read back data at location and location +4
 *    If the system wrote the same value at location and location+4
 *    then it is performing 8 byte moves which is wrong.
 *    If this is the case, try setting bit 20 in PCI_STATE register.
 *    We will be able to set it in the standard cell version.
 *    If we are able to set it return false, otherwise return true.
 */
STATIC bool 
SixtyFourBitSlotProblem(struct altinst *altp, struct tg_nic *nic)
{
    if (!(IN32(nic->handle, nic->base, TG_PCI_STATE) & 
	  TG_PCI_STATE_USE_32_BITS_ONLY)) {
	/* make sure the firmware is stopped so we won't accidentally
	 * overwrite something */
	(void) stopFW(altp);

	/* now we check things out, use 0x600 as the base offset for
	 * dealing with this.  This is just memory on the nic in the
	 * shared memory region.  It's totally safe to work here
	 */
	OUT32(nic->handle, nic->base, 0x600, 0);
	OUT32(nic->handle, nic->base, 0x604, 0);
	OUT32(nic->handle, nic->base, 0x600, 0x5555aaaa);
	if (IN32(nic->handle, nic->base, 0x604) == 0x5555aaaa) {
	    /* ok, here's the problem, check for Tigon 2, we can fix
	     * it there, else we're screwed */
	    if (nic->tigonRev == TG_MHC_LSI_GATE_VERS) {
		return(TRUE);
	    } else {
		/* fix it up by setting a flag, the mainline code will set
		 * the PCI_STATE reg when it sets it in altInitCard() */
		nic->init_flags |= ALT_INIT_USE_32_BIT_MODE;
		return(FALSE);
	    }
	}
    }
    return(FALSE);
}

/*
 * void RedundTimer(IN PVOID SystemSpecific1,
 *                  IN PVOID Context,
 *                  IN PVOID SystemSpecific2,
 *                  IN PVOID SystemSpecific3)
 *
 * This is the timer procedure scheduled to run after
 * every 250 milliseconds for dual homing switchover.
 */

STATIC void 
RedundTimer(void *systemSpecific1, void *context, void *systemSpecific2,
	    void *systemSpecific3)
{
    altRedundTimer((struct altinst *)context);
}

/*
 * void ALTINTR(PBOOLEAN InterruptRecognized, 
 *              PBOOLEAN QueueDpc,
 *              PVOID Context)
 *
 * The hardware ISR can get invoked as a result of interrupts
 * from different nics. We save the state of interrupt in
 * altp->IntSrc to be used later on by the DPC routine.
 *
 */

STATIC void 
ALTINTR(bool *ruptRecognized, bool *queueDpc, void *context)
{
    struct tg_nic *nic = (struct tg_nic *)context;
    struct altinst *altp = nic->initAltp;

    /* break out on early interrupt */
    if (altp->state == ALT_NOTREADY) {
	*queueDpc = FALSE;
	*ruptRecognized = FALSE;
	OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 0);
	return;
    }
    if (nic->init_flags & ALT_INIT_SHARED_RUPT) {
        /*
         * we share this rupt, so see if it is ours...
         */
        if (IN32(nic->handle, nic->base, TG_MISC_HOST_CONTROL) &
            TG_MHC_RUPT_STATE) {
            *ruptRecognized = TRUE;
            *queueDpc = TRUE;
 
            /* hit mailbox 0 */
            OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 1);
        } else {
            *queueDpc = FALSE;
            *ruptRecognized = FALSE;
        }
    } else {
        /*
         * private rupt- always us!!!
         */
        *ruptRecognized = TRUE;
        *queueDpc = TRUE;
 
        /* hit mailbox 0 */
        OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 1);
    }
}

/*
 * void SetAdapterResources( struct altinst *altp,
 *                           PNDIS_RESOURCE_LIST  PResourceList )  
 *
 * This function will set the values for the adapter resources in the
 * structure altp. Later on these values will be used by NdisMMapIo
 * to acquire memory region.
 */

STATIC void 
SetAdapterResources(struct tg_nic *nic, NDIS_RESOURCE_LIST *resourceList)
{
    int count, i;

    count = resourceList->Count;	/* how many resources are descrribed */

    for (i = 0; i < count; i++) {
	switch (resourceList->PartialDescriptors[i].Type) {
	case CmResourceTypeInterrupt:

	    nic->ruptLevel = 
		resourceList->PartialDescriptors[i].u.Interrupt.Level;

	    break;

	case CmResourceTypeMemory:
	    nic->unmappedBase = 
		resourceList->PartialDescriptors[i].u.Memory.Start;
	    nic->memLength =
		resourceList->PartialDescriptors[i].u.Memory.Length;
	    ASSERT(nic->memLength == 16*1024);
	    break;

	case CmResourceTypeDma:
	    break;

	default:
	    break;
	}
    }
}

/* translate a unix errno into an NDIS_STATUS failure code */
STATIC NDIS_STATUS
TransUnixErrnoToNdis(int rc)
{
    switch(rc) {
    case ENOMEM:
	return(NDIS_STATUS_RESOURCES);

    case ENODEV:
	return(NDIS_STATUS_ADAPTER_NOT_FOUND);

    case EIO:
    case EFAULT:
    case EBUSY:
    case EINVAL:
    default:	
	return(NDIS_STATUS_FAILURE);
    }
}

#if !ALT_DESERIALIZE
STATIC
#endif /* !ALT_DESERIALIZE */
void 
SendPackets(NDIS_HANDLE handle, NDIS_PACKET **packet_array,
	    UINT number_packets)
{
    struct tg_nic *nic = (struct tg_nic *)handle;
    struct altinst *altp = nic->altp;
    NDIS_PACKET *packetp;
    U16 shadSendCons;
    U16 saveSendCons;
    bool firstTime = TRUE;
    U32 coal_now = 0;
    int sentPacket;

    if ((altp->state == ALT_FW_RUNNING) &&
	(nic->flags & ALT_LINK_READY)) {
#if USE_INTERNAL_QUEUE
	/* remove from our internal queue first, then go do the 
	 * packet array */
	if (altp->flow_controlled) {
	dointqueue:
	    /* ENTER CRITICAL SECTION */
	    ALT_DISABLE(spl, &nic->sendQueueMu);
	    while ((packetp = nic->pktQueue)) {
	        nic->pktQueue = 
		    *(NDIS_PACKET **)packetp->MiniportReserved;
	        if (!altQueueSendMsg(nic, packetp, NO_VLAN_TAG, 0)) {
		    ALTTRACE(sndmsgf1, 0x30004, firstTime, 
			     nic->pktQueue, nic->pktTail, packetp);
		    /* put the packet back on the queue */
		    nic->pktQueue = packetp;

		    /* queue all the packets in the array in case we get called
		     * recursively during reap */
		    while (number_packets > 0) {
		        number_packets--; /* not in while, could be > 1 */
			packetp = *packet_array;
			packet_array++;
		        ALTTRACE(qpkts, 0x30006, number_packets, 
				 packetp, 0, 0);

			/* keep our own queue, no test necessary */
			*(NDIS_PACKET **)nic->pktTail->MiniportReserved = 
			    packetp;
			nic->pktTail = packetp;
		    }
		    
		    /* clear out the next pointer */
		    *(NDIS_PACKET **)nic->pktTail->MiniportReserved = 
			(NDIS_PACKET *)0;
		    
		    /* reap send ring directly and try again */
		    if (firstTime) {
		        firstTime = FALSE;

			ALT_ENABLE(spl, &nic->sendQueueMu);
			/* LEAVE CRITICAL SECTION */

			/* go reap bds from the send ring */
			sentPacket = altReapSendRing(nic);
			    
			/* REENTER CRITICAL SECTION */
			ALT_DISABLE(spl, &nic->sendQueueMu);
			if (sentPacket == SEND_RING_PROCESSED) {
			    nic->reapedFromSend++;
			    continue;
			}
		    }
		    break;
		}
	    }
	    /* done with everything we need to do? turn off flow control */
	    if (nic->pktQueue) {
		ALT_ENABLE(spl, &nic->sendQueueMu);
		/* LEAVE CRITICAL SECTION */
		return;		/* we're done, leave b4 doing anything else */
	    } else
	        altp->flow_controlled = FALSE;
	    ALT_ENABLE(spl, &nic->sendQueueMu);
	    /* LEAVE CRITICAL SECTION */
	}
#endif /* USE_INTERNAL_QUEUE */

	/* sent the queued up ones, now do the rest */
	while (number_packets > 0) {
	    number_packets--;	/* not in while could be > 1 */
	    packetp = *packet_array;
	    packet_array++;

	doover:
#if 0
	    /* disabled for now for performance reasons */
            if (!number_packets)
	        coal_now = 1;
#endif /* 0 */

	    if (altQueueSendMsg(nic, packetp, NO_VLAN_TAG, coal_now)) {
#if !ALT_DESERIALIZE		/* deserialized in NDIS 5, no status */
		NDIS_SET_PACKET_STATUS(packetp, NDIS_STATUS_PENDING);
#endif /* !ALT_DESERIALIZE */
	    } else {

		/* mark us as in flow control.  It might go away later */
		nic->flowControls++;
		altp->flow_controlled = TRUE;
#if USE_INTERNAL_QUEUE
		ALTTRACE(sndmsgf2, 0x30005, firstTime, number_packets, 
			 nic->pktQueue, packetp);
		/* queue all the packets in the array in case we get called
		 * recursively during reap */
		/* ENTER CRITICAL SECTION */
		ALT_DISABLE(spl, &nic->sendQueueMu);
		/* keep our own queue */
		if (nic->pktQueue)
		    *(NDIS_PACKET **)nic->pktTail->MiniportReserved = packetp;
		else 
		    nic->pktQueue = packetp;
		nic->pktTail = packetp;
		while (number_packets > 0) {
		    number_packets--; /* not in while could be > 1 */
		    packetp = *packet_array;
		    packet_array++;
		    ALTTRACE(qpkts2, 0x30007, number_packets, packetp, 0, 0);
		    /* keep our own queue, no test necessary */
		    *(NDIS_PACKET **)nic->pktTail->MiniportReserved = packetp;
		    nic->pktTail = packetp;
		}
		/* clear out the next pointer */
		*(NDIS_PACKET **)packetp->MiniportReserved = (NDIS_PACKET *)0;

		/* reap send ring directly and try again */
		if (firstTime) {
		    firstTime = FALSE;

		    ALT_ENABLE(spl, &nic->sendQueueMu);
		    /* LEAVE CRITICAL SECTION */
		    if (altReapSendRing(nic) == SEND_RING_PROCESSED) {
		        nic->reapedFromSend++;
			goto dointqueue;
		    }
		}
		else
		 ALT_ENABLE(spl, &nic->sendQueueMu);
		/* LEAVE CRITICAL SECTION */
		return;		/* we're done, leave b4 doing anything else */
#else /* USE_INTERNAL_QUEUE */
		/* reap send ring directly and try again */
		if (firstTime) {
		    firstTime = FALSE;

		    if (altReapSendRing(nic) == SEND_RING_PROCESSED) {
		        nic->reapedFromSend++;
			altp->flow_controlled = FALSE;
			goto doover;
		    }
		}

		NDIS_SET_PACKET_STATUS(packetp, NDIS_STATUS_RESOURCES);
		/* now queue the rest of the packet_array, this will drop
		 * us out of the outer for loop, we use the same exit */
		while (number_packets > 0) {
		    number_packets--; /* not in while could be > 1 */
		    packetp = *packet_array;
		    packet_array++;
		    NDIS_SET_PACKET_STATUS(packetp, NDIS_STATUS_RESOURCES);
		}
#endif /* USE_INTERNAL_QUEUE */
	    }
	}

	/* if we are set up to reap completed packets on send ring always
	 * do it here.  This can improve performance if we can get the
	 * packets back to the stack faster, maybe */
	if (nic->init_flags & ALT_INIT_REAP_AFTER_SEND) {
	    if (altReapSendRing(nic) == SEND_RING_PROCESSED)
		nic->reapedFromSend++;
	}
	/* we're here only if we sent everything */
	altp->flow_controlled = FALSE;
    } else {
	/* here we are doing lying sends */
	while (number_packets > 0) {
	    number_packets--;	/* not in while, could be > 1 */
	    packetp = *packet_array;
	    packet_array++;
#if ALT_DESERIALIZE
	    MSGFREE(packetp, MSG_SUCCESS);
#else /* ALT_DESERIALIZE */
	    NDIS_SET_PACKET_STATUS(packetp, NDIS_STATUS_FAILURE);
#endif /* ALT_DESERIALIZE */
	}
    }
}

/* check for MDL frags */
bool
altCheckMsgFrags(ALT_MSG *msg)
{
    ALT_MSG *nextMsg;

    PEEKMSGNEXT(msg, nextMsg, count, maxCount);
    if (nextMsg)
	return (FALSE);
    return (TRUE);
}

void 
altWrite32(U32 *w, U32 value)
{
    NdisWriteRegisterUlong(w, value);
}

void 
altWrite16(U16 *w, U16 value)
{
    NdisWriteRegisterUshort(w, value);
}

void 
altWrite8(U8 *w, U8 value)
{
    NdisWriteRegisterUchar(w, value);
}

U32 
altRead32(U32 *pul)
{
    U32 ulResult;

    NdisReadRegisterUlong(pul, &ulResult);
    return ulResult;
}

U16 
altRead16(U16 *pus)
{
    U16 ulResult;

    NdisReadRegisterUshort(pus, &ulResult);
    return ulResult;
}

U8 
altRead8(U8 *puc)
{
    U8 ucResult;

    NdisReadRegisterUchar(puc, &ucResult);
    return ucResult;
}
