/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997, 1998
 * All rights reserved
 *
 */
/*
 * HISTORY
 * $Log: altioctl.c,v $
 * Revision 1.1.2.32  1999/02/16  17:07:43  leonid
 * 	OID_TCP_TASK_OFFLOAD code has been changed to comply with the current version of
 * 	 the Microsoft tester.
 * 	[1999/02/16  17:07:19  leonid]
 *
 * Revision 1.1.2.31  1999/02/16  16:44:32  leonid
 * 	Added OID_DA_CHANGE_MAC_ADDRESS_R to support mac address change for the failover
 * 	 operation.
 * 	[1999/02/16  16:44:15  leonid]
 * 
 * Revision 1.1.2.30  1999/01/26  15:19:21  leonid
 * 	Stubs for Power Management OIDs are removed for now to avoid a failure on the current ndis test tool.
 * 	[1999/01/26  15:18:59  leonid]
 * 
 * Revision 1.1.2.29  1999/01/19  22:54:08  leonid
 * 	Statistic OIDs have been fixed (they were taken from the wrong 'half'
 * 	of the 64 bit counters).
 * 	[1999/01/19  22:53:40  leonid]
 * 
 * Revision 1.1.2.28  1999/01/19  16:47:00  leonid
 * 	All OID_DA_ are made generic (non-3Com specific)
 * 	so they can be used later by Alteon protocol driver.
 * 	[1999/01/19  16:46:24  leonid]
 * 
 * Revision 1.1.2.27  1999/01/19  01:00:10  leonid
 * 	Added stubs for Power Management OIDs, DA specific OIDs made generic
 * 	[1999/01/19  00:59:53  leonid]
 * 
 * Revision 1.1.2.26  1999/01/07  23:09:27  ted
 * 	bug fixes for debugger ioctls
 * 	[1999/01/07  23:07:19  ted]
 * 
 * Revision 1.1.2.25  1998/11/25  22:35:18  leonid
 * 	3Com OIDs modified to reflect their spec change
 * 	[1998/11/25  22:32:22  leonid]
 * 
 * Revision 1.1.2.24  1998/11/23  22:50:40  ted
 * 	Fix not resetting one of the stats on stats clear
 * 	[1998/11/23  22:47:35  ted]
 * 
 * Revision 1.1.2.23  1998/11/13  23:11:00  ted
 * 	Add debugger ioctls (compiles but undebugged)
 * 	[1998/11/13  23:09:32  ted]
 * 
 * Revision 1.1.2.22  1998/11/09  18:22:12  hayes
 * 	add TUNEPARMS support
 * 	[1998/11/09  18:16:42  hayes]
 * 
 * Revision 1.1.2.21  1998/11/06  21:45:56  leonid
 * 	SetMulticastAddressList fixed to add/remove the first entry in the list properly
 * 	[1998/11/06  21:45:08  leonid]
 * 
 * Revision 1.1.2.20  1998/11/03  23:25:34  ted
 * 	New NT checksum offload interface
 * 	[1998/11/03  22:22:28  ted]
 * 
 * Revision 1.1.2.17  1998/10/13  03:33:40  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:09  ted]
 * 
 * Revision 1.1.2.16  1998/10/01  01:02:09  leonid
 * 	3Com OIDs re-defined as "OP2" only.
 * 	[1998/10/01  01:01:27  leonid]
 * 
 * Revision 1.1.2.15  1998/10/01  00:15:13  ted
 * 	Fix compile errors and white space in OID_DA_...
 * 	[1998/10/01  00:14:42  ted]
 * 
 * Revision 1.1.2.14  1998/09/29  18:47:49  leonid
 * 	3Com-specific OIDs added.
 * 	[1998/09/29  18:45:19  leonid]
 * 
 * Revision 1.1.2.13  1998/09/24  18:16:54  leonid
 * 	"bytes_written" is fixed (bug 1488)
 * 	OID_GEN_MAXIMUM_SEND_PACKETS is set to SEND_RING_ENTRIES / 8
 * 	[1998/09/24  18:16:20  leonid]
 * 
 * Revision 1.1.2.12  1998/09/03  20:05:07  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:18  ted]
 * 
 * Revision 1.1.2.11  1998/08/20  18:46:35  leonid
 * 	SetMulticastAddressList() is fixed.
 * 	OID_GEN_MSXIMUM_TOTAL_SIZE is changed from 1518 to 1514 to fix Netbeui issues.
 * 	[1998/08/20  18:45:47  leonid]
 * 
 * Revision 1.1.2.10  1998/08/20  01:15:20  hayes
 * 	remove extraneous flagMu locks
 * 	[1998/08/20  00:51:23  hayes]
 * 
 * Revision 1.1.2.9  1998/08/12  01:58:13  ted
 * 	Another bmerge failure
 * 	[1998/08/10  23:02:00  ted]
 * 
 * 	bci by hand because bmerge ran out of open file descriptors, sigh
 * 	[1998/08/10  22:48:03  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:02:06  ted]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.6  1998/07/14  18:34:58  hayes
 * 	cleanup locks
 * 	[1998/07/14  18:14:57  hayes]
 * 
 * Revision 1.1.2.5  1998/07/14  02:32:49  ted
 * 	new registry names, new version number handling
 * 	[1998/07/14  02:17:06  ted]
 * 
 * Revision 1.1.2.4  1998/06/19  17:11:37  hayes
 * 	change NDIS_MAGIC to DBG
 * 	[1998/06/19  17:10:37  hayes]
 * 
 * Revision 1.1.2.3  1998/06/18  16:06:51  hayes
 * 	Merged with changes from 1.1.2.2
 * 	[1998/06/18  16:06:37  hayes]
 * 
 * 	complete merge
 * 	[1998/06/18  15:50:38  hayes]
 * 
 * Revision 1.1.2.2  1998/06/18  15:50:54  hayes
 * 	complete merge
 * 	[1998/06/18  15:50:38  hayes]
 * 
 * Revision 1.1.2.1  1998/06/16  16:46:04  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:44:51  ted]
 * 
 * Revision 1.1.2.19  1998/11/02  23:50:23  leonid
 * 	OID_DA_NIC_CLASS fixed
 * 	[1998/11/02  23:50:00  leonid]
 * 
 * Revision 1.1.2.18  1998/10/16  22:32:12  leonid
 * 	Changed the 3Com tag name from 3COM_TAG to OP2_TAG.
 * 	[1998/10/16  22:31:03  leonid]
 * 
 * Revision 1.1.2.17  1998/10/13  03:33:40  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:09  ted]
 * 
 * Revision 1.1.2.16  1998/10/01  01:02:09  leonid
 * 	3Com OIDs re-defined as "OP2" only.
 * 	[1998/10/01  01:01:27  leonid]
 * 
 * Revision 1.1.2.15  1998/10/01  00:15:13  ted
 * 	Fix compile errors and white space in OID_DA_...
 * 	[1998/10/01  00:14:42  ted]
 * 
 * Revision 1.1.2.14  1998/09/29  18:47:49  leonid
 * 	3Com-specific OIDs added.
 * 	[1998/09/29  18:45:19  leonid]
 * 
 * Revision 1.1.2.13  1998/09/24  18:16:54  leonid
 * 	"bytes_written" is fixed (bug 1488)
 * 	OID_GEN_MAXIMUM_SEND_PACKETS is set to SEND_RING_ENTRIES / 8
 * 	[1998/09/24  18:16:20  leonid]
 * 
 * Revision 1.1.2.12  1998/09/03  20:05:07  ted
 * 	NDIS 5.0 deserialized code, mostly working
 * 	[1998/09/03  19:55:18  ted]
 * 
 * Revision 1.1.2.11  1998/08/20  18:46:35  leonid
 * 	SetMulticastAddressList() is fixed.
 * 	OID_GEN_MSXIMUM_TOTAL_SIZE is changed from 1518 to 1514 to fix Netbeui issues.
 * 	[1998/08/20  18:45:47  leonid]
 * 
 * Revision 1.1.2.10  1998/08/20  01:15:20  hayes
 * 	remove extraneous flagMu locks
 * 	[1998/08/20  00:51:23  hayes]
 * 
 * Revision 1.1.2.9  1998/08/12  01:58:13  ted
 * 	Another bmerge failure
 * 	[1998/08/10  23:02:00  ted]
 * 
 * 	bci by hand because bmerge ran out of open file descriptors, sigh
 * 	[1998/08/10  22:48:03  ted]
 * 
 * 	Add NT 5.0 checksum offload code
 * 	[1998/07/24  19:02:06  ted]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.8  1998/08/03  22:06:36  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:18  hayes]
 * 
 * Revision 1.1.2.7  1998/08/03  17:11:37  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:55  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:08  hayes]
 * 
 * Revision 1.1.2.6  1998/07/14  18:34:58  hayes
 * 	cleanup locks
 * 	[1998/07/14  18:14:57  hayes]
 * 
 * Revision 1.1.2.5  1998/07/14  02:32:49  ted
 * 	new registry names, new version number handling
 * 	[1998/07/14  02:17:06  ted]
 * 
 * Revision 1.1.2.4  1998/06/19  17:11:37  hayes
 * 	change NDIS_MAGIC to DBG
 * 	[1998/06/19  17:10:37  hayes]
 * 
 * Revision 1.1.2.3  1998/06/18  16:06:51  hayes
 * 	Merged with changes from 1.1.2.2
 * 	[1998/06/18  16:06:37  hayes]
 * 
 * 	complete merge
 * 	[1998/06/18  15:50:38  hayes]
 * 
 * Revision 1.1.2.2  1998/06/18  15:50:54  hayes
 * 	complete merge
 * 	[1998/06/18  15:50:38  hayes]
 * 
 * Revision 1.1.2.1  1998/06/16  16:46:04  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:44:51  ted]
 * 
 * $EndLog$
 */
/*
 *  NOTE: for further history on this file see ntinfo.c which has
 *  since been defuncted.
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/ndis/altioctl.c,v 1.1.2.32 1999/02/16 17:07:43 leonid Exp $";
#endif /* __lint */

#include <ndis.h>

#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"
#include "eeprom.h"
#include "altos.h"
#include "althdr.h"
#include "altreg.h"		/* must follow althdr.h */
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"
#include "drvvers.h"
#include "devioctl.h"
#include "ntddndis.h"
#include "drvstat.h"
#include "link.h"

/* function forwards */
STATIC void SetMulticastAddressList(struct altinst *, struct ether_addr *,
				    U32);
STATIC void ReadNicTraceBuf(struct altinst *, U8 *);
STATIC void UpdateFromRegistry(struct altinst *, U32 *);
STATIC void ReadRegTable(alt_reg_tab_t *, U32, NDIS_HANDLE);

#define NEW_REG	0		/* enable/disable new registry interface */

#if NEW_REG
STATIC NTSTATUS QueryCallback(PWSTR, U32, void *, U32, void *, void *);
#endif /* NEW_REG */

UINT altSupportedOids[] =
{
    OID_GEN_SUPPORTED_LIST,		/* Q M List of supported OID's */
    OID_GEN_HARDWARE_STATUS,		/* Q M Hardware Status */
    OID_GEN_MEDIA_SUPPORTED,		/* Q M Media Type supported */
    OID_GEN_MEDIA_IN_USE,		/* Q M Media type in use */
    OID_GEN_MAXIMUM_LOOKAHEAD,		/* Q M Max look ahead size in bytes */
    OID_GEN_MAXIMUM_FRAME_SIZE,		/* Q M Frame size in bytes */
    OID_GEN_LINK_SPEED,			/* Q M Link speed (100 bps) 10000000 */
    OID_GEN_TRANSMIT_BUFFER_SPACE,	/* Q M tx buf space in bytes */
    OID_GEN_RECEIVE_BUFFER_SPACE,	/* Q M rx buf space in bytes */
    OID_GEN_TRANSMIT_BLOCK_SIZE,	/* Q M tx packet 1514 or 1500 */
    OID_GEN_RECEIVE_BLOCK_SIZE,		/* Q M rx packet size, same as tx */
    OID_GEN_VENDOR_ID,			/* Q M Vendor NIC Code */
    OID_GEN_VENDOR_DESCRIPTION,		/* Q M Vendor Card Description */
    OID_GEN_CURRENT_PACKET_FILTER,	/* Q S M Packet filter */
    OID_GEN_CURRENT_LOOKAHEAD,		/* Q S M lookahead size in bytes */
    OID_GEN_DRIVER_VERSION,		/* Q M NDIS version number */
    OID_GEN_MAXIMUM_TOTAL_SIZE,		/* Q M max total pkt len (bytes) */
    OID_GEN_PROTOCOL_OPTIONS,		/* Q S M optional protocol flags */
    OID_GEN_MAC_OPTIONS,		/* Q S M Optional NIC Flags */
    OID_GEN_MEDIA_CONNECT_STATUS,	/* Q M connected to the network */
    OID_GEN_MAXIMUM_SEND_PACKETS,	/* Q M max number of send packets    */
					/*     driver can accept per call to */
					/*     its altSendPackets function   */
    OID_GEN_VENDOR_DRIVER_VERSION,	/* Q M Vendor assigned driver number */

    /* general statistics */

    OID_GEN_XMIT_OK,			/* Q M frames tx w/o error */
    OID_GEN_RCV_OK,			/* Q M frame rx w/o error */
    OID_GEN_XMIT_ERROR,			/* Q M frames not tx or tx with err */
    OID_GEN_RCV_ERROR,			/* Q M frames rx with errors */
    OID_GEN_RCV_NO_BUFFER,		/* Q M no buffers available */
    OID_GEN_DIRECTED_BYTES_XMIT,	/* Q O ucast bytes tx w/o errors */
    OID_GEN_DIRECTED_FRAMES_XMIT,	/* Q O ucast frames tx w/o errors */
    OID_GEN_MULTICAST_BYTES_XMIT,	/* Q O mcast bytes tx w/o errors */
    OID_GEN_MULTICAST_FRAMES_XMIT,	/* Q O mcast frames tx w/o errors */
    OID_GEN_BROADCAST_BYTES_XMIT,	/* Q O bcast bytes tx w/o errors */
    OID_GEN_BROADCAST_FRAMES_XMIT,	/* Q O bcast frames tx w/o errors */
    OID_GEN_DIRECTED_BYTES_RCV,		/* Q O ucast bytes rx w/o errors */
    OID_GEN_DIRECTED_FRAMES_RCV,	/* Q O ucast frames rx w/o errors */
    OID_GEN_MULTICAST_BYTES_RCV,	/* Q O mcast bytes rx w/o errors */
    OID_GEN_MULTICAST_FRAMES_RCV,	/* Q O mcast frames rx w/o errors */
    OID_GEN_BROADCAST_BYTES_RCV,	/* Q O bcast bytes rx w/o errors */
    OID_GEN_BROADCAST_FRAMES_RCV,	/* Q O bcast frames rx w/o errors */
    OID_GEN_RCV_CRC_ERROR,		/* Q frames rx with CRC error */
    OID_GEN_TRANSMIT_QUEUE_LENGTH,	/* Q length of transmit queue */

    /* ethernet operation characteristics */
    OID_802_3_PERMANENT_ADDRESS,	/* Q M permanet station address */
    OID_802_3_CURRENT_ADDRESS,		/* Q M current station address */
    OID_802_3_MULTICAST_LIST,		/* Q M M current mcast address list */
    OID_802_3_MAXIMUM_LIST_SIZE,	/* Q M max size of mcast addr list */
    OID_802_3_MAC_OPTIONS,	        /* Q M NIC flags */

    /* ethernet statistics */

    OID_802_3_RCV_ERROR_ALIGNMENT,	/* Q M frames rx with alignment err */
    OID_802_3_XMIT_ONE_COLLISION,	/* Q M frames tx with one collion */
    OID_802_3_XMIT_MORE_COLLISIONS,	/* Q M frames tx with >1 collison */
    OID_802_3_XMIT_DEFERRED,		/* Q O frames tx after deferrel */
    OID_802_3_XMIT_MAX_COLLISIONS,	/* Q O frames not tx due to colls */
    OID_802_3_RCV_OVERRUN,		/* Q O frames not rx due to overrun */
    OID_802_3_XMIT_UNDERRUN,		/* Q O frames not tx due to overrun */
    OID_802_3_XMIT_HEARTBEAT_FAILURE,	/* Q O frames tx with hbeat failure */
    OID_802_3_XMIT_TIMES_CRS_LOST,	/* Q O times CS lost during tx */
    OID_802_3_XMIT_LATE_COLLISIONS,	/* Q O late collions detected */

#if USE_CKSUM_OFFLOAD
    /* TCP/IP task offloads */
    OID_TCP_TASK_OFFLOAD,		/* tell about TCP/IP cksum offload */
#endif /* USE_CKSUM_OFFLOAD */

    OID_VENDOR_GET_TIGON_REGS,  	/* get pci config space regs */
    OID_VENDOR_GET_NIC_BUFFER,		/* get nic trace buffer */
    OID_VENDOR_GET_HOST_BUFFER,		/* get host trace buffer */
    OID_VENDOR_CLEAR_NIC_STATS,		/* clear nic stats */
    OID_VENDOR_GET_NIC_STATS,		/* get nic stats */
    OID_VENDOR_GET_DRIVER_STATS,	/* get driver stats */
    OID_VENDOR_GET_TUNEPARMS,		/* get tuning paramters */
    OID_VENDOR_SET_TUNEPARMS,		/* set tuning paramters */
    OID_VENDOR_READ_TG_MEM,		/* get tigon memory */  
    OID_VENDOR_WRITE_TG_MEM,		/* set tigon memory */
    OID_VENDOR_READ_TG_REG,		/* get tigon hardware registers */
    OID_VENDOR_WRITE_TG_REG,		/* set tigon hardware registers */

    OID_DA_CHANGE_MAC_ADDRESS,		/* change MAC address */
    OID_DA_CHANGE_MAC_ADDRESS_R,	/* change MAC address */
    OID_DA_NIC_CLASS, 	 		/* get NIC class */

};

NDIS_STATUS
altSetInfo(NDIS_HANDLE handle, NDIS_OID oid, void *infoBuffer,
	   U32 bufLen, U32 *bytes_read, U32 *bytes_needed)
{
    struct tg_nic *nic = (struct tg_nic *)handle;
    struct altinst *altp = nic->initAltp;
    U32 filter, numEntries;
    NDIS_STATUS status = NDIS_STATUS_SUCCESS;	/* Status of the operation. */
#if USE_CKSUM_OFFLOAD
    NDIS_TASK_OFFLOAD_HEADER *taskOffloadHdr;
    NDIS_TASK_OFFLOAD *taskOffload;
    NDIS_TASK_TCP_IP_CHECKSUM *tcpipOffload;
    U32 oldCsumFlags;
    U32 neededLen;
    bool resetAddr;
#endif /* USE_CKSUM_OFFLOAD */
    U32 generic_ulong;

    switch (oid) {

    case OID_DA_CHANGE_MAC_ADDRESS:
	generic_ulong = ((PULONG)(infoBuffer))[0];
	if (generic_ulong == OP2_TAG) {
	    ((PULONG)(infoBuffer))[0] = OP2_MINIPORT_TAG;	
	    MEMCPY(altp->macAddr.ether_addr_octet, ((PUCHAR)infoBuffer+4),
		   sizeof(struct ether_addr));
	    nic->flags |= ALT_OVERRIDE_MAC_ADDR;
	    altSetMacaddr(nic, &altp->macAddr);
	    ((PUCHAR)(infoBuffer))[0] = 0; /* Set Success code in the buffer */
	}
	else {
	    *bytes_read = 0;
	    *bytes_needed = 0;
			status = NDIS_STATUS_FAILURE;
	}
	break;

    case OID_DA_CHANGE_MAC_ADDRESS_R:
       	    MEMCPY(altp->macAddr.ether_addr_octet, ((PUCHAR)infoBuffer),
	    sizeof(struct ether_addr));
    	    nic->flags |= ALT_OVERRIDE_MAC_ADDR;
    	    altSetMacaddr(nic, &altp->macAddr);
	    ((PUCHAR)(infoBuffer))[0] = 0; /* Set Success code in the buffer */
	    break;

    case OID_802_3_MULTICAST_LIST:
	if ((bufLen % sizeof(struct ether_addr)) != 0) { /* Verify len */
	    ERROR1("Multicast length invalid");
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    break;
	}
	numEntries = bufLen / sizeof(struct ether_addr);
	if (numEntries > MAX_MCAST_ENTRIES) {
	    ERROR1("Too many multicast addresses requested");
	    status = NDIS_STATUS_MULTICAST_FULL;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    break;
	}
	SetMulticastAddressList(altp, (struct ether_addr *)infoBuffer,
				numEntries);
	break;

    case OID_GEN_CURRENT_PACKET_FILTER:
	if (bufLen != 4) {	/* Verify length */
	    ERROR1("Set packet filter has invalid length");
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    break;
	}
	filter = *(U32 *)infoBuffer;
	if (filter & (NDIS_PACKET_TYPE_SOURCE_ROUTING |
		      NDIS_PACKET_TYPE_SMT |
		      NDIS_PACKET_TYPE_MAC_FRAME |
		      NDIS_PACKET_TYPE_FUNCTIONAL |
		      NDIS_PACKET_TYPE_ALL_FUNCTIONAL |
		      NDIS_PACKET_TYPE_GROUP
	    )) {
	    ERROR2("Filter 0x%x has unsupported fields", filter);
	    status = NDIS_STATUS_NOT_SUPPORTED;
	    *bytes_read = 4;
	    *bytes_needed = 0;
	    break;
	}
	if (filter & NDIS_PACKET_TYPE_PROMISCUOUS) {
	    altp->flags |= ALT_PROMISC;
	    nic->flags |= ALT_PROM_MODE;
	    altSetPromiscuous(altp, TRUE);
	} else {
	    altSetPromiscuous(altp, FALSE);
	    altp->flags &= ~ALT_PROMISC;
	    nic->flags &= ~ALT_PROM_MODE;
	}
	if (filter & NDIS_PACKET_TYPE_ALL_MULTICAST) {
	    nic->flags |= ALT_MULTICAST_MODE;
	    altSetMultiCastMode(altp, TRUE);
	} else {
	    altSetMultiCastMode(altp, FALSE);
	    nic->flags &= ~ALT_MULTICAST_MODE;
	}
	break;

    case OID_GEN_CURRENT_LOOKAHEAD:
	if (bufLen != 4) {	/* Verify length */
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    ERROR1("Set Lookahead has invalid length");
	    break;
	}
	altp->look_ahead_size = *(U32 *)infoBuffer;
	break;

#if USE_CKSUM_OFFLOAD
    case OID_TCP_TASK_OFFLOAD:
	/* see if we're being asked to disable task offload first */
	neededLen = sizeof(NDIS_TASK_OFFLOAD_HEADER);
        if (bufLen < neededLen) {
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    ERROR3("Set TCP_TASK_OFFLOAD has invalid length %d, expected %d", 
		   bufLen, neededLen);
	    break;
	}
	taskOffloadHdr = (NDIS_TASK_OFFLOAD_HEADER *)infoBuffer;

	if (taskOffloadHdr->Version != NDIS_TASK_OFFLOAD_VERSION) {
	    status = NDIS_STATUS_BAD_VERSION;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    break;
	}	
		
	if (!taskOffloadHdr->OffsetFirstTask) {
	    /* disable all checksum offload */
	    oldCsumFlags = nic->csumFlags;
	    nic->csumFlags = 0;
	    /* we have to stop and restart the card to get cksum options in */
	    if (oldCsumFlags != nic->csumFlags)
		ResetInstance(&resetAddr, handle);
	    break;
	}

	/* now look for setting up a task offload */
	neededLen = sizeof(NDIS_TASK_OFFLOAD_HEADER) +
	    sizeof(NDIS_TASK_OFFLOAD) + sizeof(NDIS_TASK_TCP_IP_CHECKSUM);
        if (bufLen < neededLen) {
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    ERROR3("Set TCP_TASK_OFFLOAD has invalid length %d, expected %d", 
		   bufLen, neededLen);
	    break;
	}

	/* find out what we're supposed to support */
        taskOffload = (NDIS_TASK_OFFLOAD *)((char *)infoBuffer + 
					    taskOffloadHdr->OffsetFirstTask);
	tcpipOffload = (NDIS_TASK_TCP_IP_CHECKSUM *)taskOffload->TaskBuffer;

	if((tcpipOffload->V6Transmit.IpOptionsSupported) ||
	  (tcpipOffload->V6Transmit.TcpOptionsSupported) ||
	  (tcpipOffload->V6Transmit.TcpChecksum) ||
	  (tcpipOffload->V6Transmit.UdpChecksum) ||
	  (tcpipOffload->V6Receive.IpOptionsSupported) ||
	  (tcpipOffload->V6Receive.TcpOptionsSupported) ||
	  (tcpipOffload->V6Receive.TcpChecksum) ||
	  (tcpipOffload->V6Receive.UdpChecksum))
	{
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_read = 0;
	    *bytes_needed = 0;
	    break;
	}	

	oldCsumFlags = nic->csumFlags;
	nic->csumFlags = 0;
	if (tcpipOffload->V4Transmit.IpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_XMIT_IP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_XMIT_IP;
	
	if (tcpipOffload->V4Transmit.TcpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_XMIT_TCP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_XMIT_TCP;

	if (tcpipOffload->V4Transmit.UdpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_XMIT_UDP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_XMIT_UDP;
	
	if (tcpipOffload->V4Receive.IpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_RECV_IP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_RECV_IP;
	
	if (tcpipOffload->V4Receive.TcpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_RECV_TCP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_RECV_TCP;

	if (tcpipOffload->V4Receive.UdpChecksum)
	    nic->csumFlags |= ALT_CKSUM_DO_RECV_UDP;
	else
	    nic->csumFlags &= ~ALT_CKSUM_DO_RECV_UDP;

	    nic->csumFlags |= ALT_CKSUM_XMIT_NO_PSEUDO;

	ALTTRACE(settask, 0x80001, taskOffload, tcpipOffload,
		 *(U32 *)&tcpipOffload->V4Transmit,
		 *(U32 *)&tcpipOffload->V4Receive);

	/* we have to stop and restart the card to get checksum options in */
	if (oldCsumFlags != nic->csumFlags)
	    ResetInstance(&resetAddr, handle);
	break;
#endif /* USE_CKSUM_OFFLOAD */

    default:
	status = NDIS_STATUS_INVALID_OID;
	*bytes_read = 0;
	*bytes_needed = 0;
	break;

    }
    if (status == NDIS_STATUS_SUCCESS) {
	*bytes_read = bufLen;
	*bytes_needed = 0;
    }
    return(status);
}

STATIC void
SetMulticastAddressList(struct altinst *altp, struct ether_addr *mcTable,
			U32 numEntries)
{
    U32 i, j;
    bool match;

    /* first find out what we need to add */
    altp->addmccount = 0;
    for (i = 0; i < numEntries; i++) {
	match = FALSE;
	for (j= 0; j < altp->mccount; j++) {
	    /* if we get a match it's already there */
	    if (ether_cmp(&mcTable[i], &altp->mctab[j].addr) == 0) {
		match = TRUE;
		break;
	    }
	}
	/* if we're here we need to add it */
	if (!match) {
	    ether_copy(&mcTable[i], &altp->addtab[altp->addmccount]);
	    altp->addmccount++;
	}
    }

    /* now find out what we need to delete */
    altp->delmccount = 0;
    for (i = 0; i < altp->mccount; i++) {
	match = FALSE;
	for (j= 0; j < numEntries; j++) {
	    /* if we get a match it's still there */
	    if (ether_cmp(&altp->mctab[i].addr, &mcTable[j]) == 0) {
		match = TRUE;
		break;
	    }
	}
	/* if we're here we need to delete it */
	if (!match) {
	    ether_copy(&altp->mctab[i].addr, &altp->deltab[altp->delmccount]);
	    altp->delmccount++;
	}
    }

    /* update the count and fix the table */
    altp->mccount = altp->mccount - altp->delmccount + altp->addmccount;
    for (i = 0; i < altp->mccount; i++) {
	ether_copy(&mcTable[i], &altp->mctab[i].addr);
    }

    /* and start telling the firmware about it */
    while (altp->addmccount) {
	altAddMultiCast(altp, &altp->addtab[0]);
	/* shift the entries up */
        for (i = 1; i < altp->addmccount; i++)
	    ether_copy(&altp->addtab[i], &altp->addtab[i-1]);
	altp->addmccount--;
    }
    while (altp->delmccount) {
	/* shift the entries up */
	altDelMultiCast(altp, &altp->deltab[0]);
        for (i = 1; i < altp->delmccount; i++)
	    ether_copy(&altp->deltab[i], &altp->deltab[i-1]);
	altp->delmccount--;
    }
}

NDIS_STATUS
altQueryInfo(NDIS_HANDLE handle, NDIS_OID oid, void *infoBuffer,
	     U32 bufLen, U32 *bytes_written, U32 *bytes_needed)
{
    struct tg_nic *nic = (struct tg_nic *)handle;
    struct altinst *altp = nic->initAltp;
    NDIS_STATUS status = NDIS_STATUS_SUCCESS;
    NDIS_HARDWARE_STATUS hardware_status;
    NDIS_MEDIA_STATE media_status;
    NDIS_MEDIUM medium = NdisMedium802_3;
    U32 inst;
    DRIVER_STATS driver_stats;
    U32 macspot[2];
    int len;
    U32 tgAddr;
    struct tg_reg nicReg;	/* for gdb register accesses */
    struct tg_mem nicMem;	/* for gdb memory accesses */
#if USE_CKSUM_OFFLOAD
    NDIS_TASK_OFFLOAD_HEADER *taskOffloadHdr;
    NDIS_TASK_OFFLOAD *taskOffload;
    NDIS_TASK_TCP_IP_CHECKSUM *tcpipOffload;
    U32 neededLen;
#endif /* USE_CKSUM_OFFLOAD */

    /* note that this set of variables assumes that the "usual"
     * response is a 4 byte unsigned field.  So things are initialized
     * this way and then overwritten in the cases where they AREN'T as
     * expected */
    U32 generic_ulong;
    U16 generic_ushort;
    U8 generic_array[80];
    U32 move_bytes = sizeof(U32);
    void *move_source = (void *)(&generic_ulong);
    U32 gig_link_status;
    U32 fast_link_status;

    /* Make sure that int is 4 bytes.  Else generic_ulong must change */

    ASSERT(sizeof(ULONG) == 4);

    switch (oid) {
    case OID_DA_NIC_CLASS:     /* support class value for the 3Com NIC */
        generic_ulong = ((PULONG)(infoBuffer))[0];
	if (generic_ulong == OP2_TAG) {
	    ((PULONG)(infoBuffer))[0] = OP2_MINIPORT_TAG;	
	    ((PUSHORT)(infoBuffer))[2] = 3;
	    move_source = (void *)infoBuffer;
	    move_bytes = 6;
	} else
	    status = NDIS_STATUS_FAILURE;
	break;

    case OID_VENDOR_GET_TIGON_REGS:
	move_source = (void *)nic->base;
	move_bytes = 64;	/* for now hardwire it util Ted changes it */
	break;

    case OID_VENDOR_GET_NIC_BUFFER:
	/* this code has a big mutex window on the memory register */
	/* we just ignore it because it's only for debug, but ... it's there */
	ReadNicTraceBuf(altp, (U8 *)infoBuffer);
	move_source = (void *)infoBuffer;
	move_bytes = RDSHMEM(nic->genComm->gen_com.nic_trace_len);
	break;

    case OID_VENDOR_GET_HOST_BUFFER:
	MEMCPY((MEMCPY_TYPE)alt_trtmp,
	       (MEMCPY_TYPE)(alt_curtb + ALT_TRACE_ELEM_SIZE),
	       sizeof(u_long)*(alt_endtb - alt_curtb));
	MEMCPY((MEMCPY_TYPE)(alt_trtmp +
			   (alt_endtb - (alt_curtb + ALT_TRACE_ELEM_SIZE))),
	       (MEMCPY_TYPE) alt_tbuf,
	       sizeof(u_long) * (alt_curtb - alt_tbuf));
	move_source = (void *)alt_trtmp;
	move_bytes = ALT_TRACE_SIZE * ALT_TRACE_ELEM_SIZE;
	break;

    case OID_VENDOR_CLEAR_NIC_STATS:
	altClearNicStats(altp);
	nic->inpkts = 0;
	nic->outpkts = 0;
	nic->outDrops = 0;
	nic->outDropsDmaResources = 0;
	nic->outDropsNoSpace = 0;
	nic->flowControls = 0;
	nic->reapedFromSend = 0;
	nic->xmit_acked = 0;
	nic->cmdFull = 0;
	nic->dataRtnPkts = 0;
#if USE_CKSUM_OFFLOAD
	nic->badIpCksums = 0;
	nic->badTcpCksums = 0;
	nic->badUdpCksums = 0;
#endif /* USE_CKSUM_OFFLOAD */
	nic->pendPkts = 0;
	nic->requeuedPkts = 0;
	nic->lowResources = 0;
	nic->normRecv = 0;
	/* don't really return anything */
	generic_ulong = 0;
	break;

    case OID_VENDOR_GET_NIC_STATS:
	move_source = (void *)&nic->genInfo->stats;
	move_bytes = sizeof(struct tg_stats);
	break;

    case OID_VENDOR_GET_DRIVER_STATS:
        driver_stats.base = nic->base;
        driver_stats.flags = nic->flags;
        driver_stats.init_flags = nic->init_flags;
        driver_stats.tigonRev = nic->tigonRev;
        driver_stats.genInfo = nic->genInfo;
        driver_stats.eventArr = nic->eventArr;
        driver_stats.eventProdPtr = nic->eventProdPtr;
        driver_stats.sendRing = nic->sendRing;
        driver_stats.sendBufCount = nic->sendBufCount;
        driver_stats.send_producer = nic->send_producer;
        driver_stats.send_consumer = nic->send_consumer;
        driver_stats.sendConsPtr = nic->sendConsPtr;
        driver_stats.recvStdRing = nic->recvStdRing;
        driver_stats.recvJumboRing = nic->recvJumboRing;
#if USE_MINI_RING
        driver_stats.recvMiniRing = nic->recvMiniRing;
#else /* USE_MINI_RING */
        driver_stats.recvMiniRing = (void *)0;
#endif /* USE_MINI_RING */
        driver_stats.recvReturnRing = nic->recvReturnRing;
        driver_stats.recvStdBufCount = nic->recvStdBufCount;
        driver_stats.recv_std_producer = nic->recv_std_producer;
        driver_stats.recv_std_consumer = nic->recv_std_consumer;
        driver_stats.recvJumboBufCount = nic->recvJumboBufCount;
        driver_stats.recv_jumbo_producer = nic->recv_jumbo_producer;
        driver_stats.recv_jumbo_consumer = nic->recv_jumbo_consumer;
#if USE_MINI_RING
        driver_stats.recvMiniBufCount = nic->recvMiniBufCount;
        driver_stats.recv_mini_producer = nic->recv_mini_producer;
        driver_stats.recv_mini_consumer = nic->recv_mini_consumer;
#else /* USE_MINI_RING */
        driver_stats.recvMiniBufCount = 0;
        driver_stats.recv_mini_producer = 0;
        driver_stats.recv_mini_consumer = 0;
#endif /* USE_MINI_RING */
        driver_stats.recv_return_producer = nic->recv_return_producer;
        driver_stats.recv_return_consumer = nic->recv_return_consumer;
        driver_stats.recvReturnProdPtr = nic->recvReturnProdPtr;
        driver_stats.inpkts = nic->inpkts;
        driver_stats.bufsInStdRecvRing = nic->bufsInStdRecvRing;
        driver_stats.bufsInJumboRecvRing = nic->bufsInJumboRecvRing;
#if USE_MINI_RING
        driver_stats.bufsInMiniRecvRing = nic->bufsInMiniRecvRing;
#else /* USE_MINI_RING */
        driver_stats.bufsInMiniRecvRing = 0;
#endif /* USE_MINI_RING */
	driver_stats.bufsOnSendRing = nic->bufsOnSendRing;
        driver_stats.bufsLowWater = nic->bufsLowWater;
        driver_stats.recvStdCoal = nic->recvStdCoal;
        driver_stats.recvJumboCoal = nic->recvJumboCoal;
#if USE_MINI_RING
        driver_stats.recvMiniCoal = nic->recvMiniCoal;
#else /* USE_MINI_RING */
        driver_stats.recvMiniCoal = 0;
#endif /* USE_MINI_RING */
        driver_stats.outpkts = nic->outpkts;
        driver_stats.outDrops = nic->outDrops;
        driver_stats.outDropsDmaResources = nic->outDropsDmaResources;
        driver_stats.outDropsNoSpace = nic->outDropsNoSpace;
        driver_stats.flowControls = nic->flowControls;
        driver_stats.reapedFromSend = nic->reapedFromSend;
        driver_stats.cmdFull = nic->cmdFull;
        driver_stats.pendPkts = nic->pendPkts;
        driver_stats.requeuedPkts = nic->requeuedPkts;
        driver_stats.dataRtnPkts = nic->dataRtnPkts;
#if USE_CKSUM_OFFLOAD
	driver_stats.badIpCksums = nic->badIpCksums;
	driver_stats.badTcpCksums = nic->badTcpCksums;
	driver_stats.badUdpCksums = nic->badUdpCksums;
#else /* USE_CKSUM_OFFLOAD */
	driver_stats.badIpCksums = 0;
	driver_stats.badTcpCksums = 0;
	driver_stats.badUdpCksums = 0;
#endif /* USE_CKSUM_OFFLOAD */
        driver_stats.lowResources = nic->lowResources;
        driver_stats.normRecv = nic->normRecv;
	driver_stats.outEarly = nic->outEarly;

	move_source = (void *)&driver_stats;
	move_bytes = sizeof(DRIVER_STATS);
	break;

    case OID_VENDOR_GET_TUNEPARMS:
	move_source = (void *)&nic->genComm->gen_com.tuneParms;
	move_bytes = sizeof(struct tg_tune);
	break;

    case OID_VENDOR_SET_TUNEPARMS:
	gig_link_status = nic->genComm->gen_com.tuneParms.link_negotiation;
	fast_link_status = 
	    nic->genComm->gen_com.tuneParms.fast_link_negotiation;

	/*
	 * it is up to the application to update the tuning parameters
	 * in the Registry if it so desires...
	 */
	move_source = (void *)&nic->genComm->gen_com.tuneParms;
	move_bytes = sizeof(struct tg_tune);
	
	MEMCPY(move_source, infoBuffer, move_bytes);
	if ((gig_link_status != 
	     nic->genComm->gen_com.tuneParms.link_negotiation) ||
	    (fast_link_status != 
	     nic->genComm->gen_com.tuneParms.fast_link_negotiation)) {
	    altDoLinkNegotiation(altp);
	}
	break;

    case OID_VENDOR_READ_TG_MEM:
	len = ((struct tg_mem *)infoBuffer)->len;
	tgAddr = ((struct tg_mem *)infoBuffer)->tgAddr;
	if (len > MAX_TG_DATA_LEN) {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}
	/* copy addrs and stuff into our local structure */
	nicMem.len = len;
	nicMem.tgAddr = tgAddr;
	move_source = (void *)&nicMem;
	/* check for the area we're trying to work with */
	if (
#if TG_BEG_SRAM		/* typically 0 */
	    tgAddr >= TG_BEG_SRAM &&
#endif /* TG_BEG_SRAM */
	    tgAddr + len <= TG_END_SRAM) {
	    if (readTGMem(altp, tgAddr, nicMem.data, 
			  len, FALSE, TRUE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);
	} else if (tgAddr >= TG_BEG_SCRATCH &&
		   tgAddr + len <= TG_END_SCRATCH) {
	    if ((IN32(altp->initNic->handle, altp->initNic->base, 
		      TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		TG_MHC_LSI_GATE_VERS) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    if (readTGScratch(altp, tgAddr, nicMem.data, 
			      len, FALSE, TRUE, TG_PROCESSOR_A) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_SCRATCH_B_DEBUG &&
		   tgAddr + len <= TG_END_SCRATCH_B_DEBUG) {
	    if ((IN32(altp->initNic->handle, altp->initNic->base, 
		      TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		TG_MHC_LSI_GATE_VERS) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    if (readTGScratch(altp, tgAddr - TG_SCRATCH_DEBUG_OFF, 
			      nicMem.data, 
			      len, FALSE, TRUE, TG_PROCESSOR_B) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_FLASH &&
		   tgAddr + len <= TG_END_FLASH) {
	    if (readTGFlash(altp, tgAddr, nicMem.data, len, FALSE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_SER_EEPROM &&
		   tgAddr + len <= TG_END_SER_EEPROM) {
	    if (readTGSerEeprom(altp, tgAddr, nicMem.data, len, FALSE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_REGS &&
		   tgAddr + len <= TG_END_REGS) {
	    ALTTRACE2G(rdMemReg, 0x50013, tgAddr, len, 0, 0,
		       altp->initNic->base + TG_DEV_ID + 
		       (tgAddr - TG_BEG_REGS));
	    ALT_SWAP_MEMCPY_TOHOST(nicMem.data, 
				   altp->initNic->base + TG_DEV_ID + 
				   (tgAddr - TG_BEG_REGS), len);
	    move_bytes = sizeof(struct tg_mem);

	} else {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}
	  
	ALTTRACEG(readMem, 0x5000b, tgAddr, len,  *(U32 *)nicMem.data,
		  *(U32 *)(&nicMem.data[4]));
	break;

    case OID_VENDOR_WRITE_TG_MEM:
	len = ((struct tg_mem *)infoBuffer)->len;
	tgAddr = ((struct tg_mem *)infoBuffer)->tgAddr;
	if (len > MAX_TG_DATA_LEN) {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}

	/* copy addrs and stuff into our local structure */
	nicMem.len = len;
	nicMem.tgAddr = tgAddr;
	MEMCPY(nicMem.data, ((struct tg_mem *)infoBuffer)->data, 
	       MAX_TG_DATA_LEN);
	move_source = (void *)&nicMem;
	  
	ALTTRACEG(writeMem, 0x5000c, tgAddr, len, *(U32 *)nicMem.data,
		  *(U32 *)(&nicMem.data[4]));

	/* copy data requested into correct area */
	if (
#if TG_BEG_SRAM		/* typically 0 and an unsigned compare */
	    tgAddr >= TG_BEG_SRAM &&
#endif /* TG_BEG_SRAM */
	    tgAddr + len <= TG_END_SRAM) {
	    if (writeTGMem(altp, tgAddr, nicMem.data, 
			   len, FALSE, TRUE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_SCRATCH &&
		   tgAddr + len <= TG_END_SCRATCH) {
	    if ((IN32(altp->initNic->handle, altp->initNic->base, 
		      TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		TG_MHC_LSI_GATE_VERS) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }

	    if (writeTGScratch(altp, tgAddr, nicMem.data, 
			       len, FALSE, TRUE, TG_PROCESSOR_A) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_SCRATCH_B_DEBUG &&
		   tgAddr + len <= TG_END_SCRATCH_B_DEBUG) {
	    if ((IN32(altp->initNic->handle, altp->initNic->base, 
		      TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		TG_MHC_LSI_GATE_VERS) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }

	    if (writeTGScratch(altp, tgAddr - TG_SCRATCH_DEBUG_OFF, 
			       nicMem.data, 
			       len, FALSE, TRUE, TG_PROCESSOR_B) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_FLASH &&
		   tgAddr + len <= TG_END_FLASH) {
	    if (writeTGFlash(altp, tgAddr, nicMem.data, len, FALSE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if (tgAddr >= TG_BEG_SER_EEPROM &&
		   tgAddr + len <= TG_END_SER_EEPROM) {
	    if (writeTGSerEeprom(altp, tgAddr, nicMem.data, len, FALSE) < 0) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
	    move_bytes = sizeof(struct tg_mem);

	} else if ((tgAddr >= TG_BEG_REGS &&
		    tgAddr + len <= TG_END_WRITE_REGS) ||
		   (tgAddr >= TG_BEG_REGS2 &&
		    tgAddr + len <= TG_END_REGS)) {
	    ALTTRACE2G(wrMemReg, 0x50012, tgAddr, len, 0, 0, 
		       nicMem.data);
	    /* special case writing the TG_WINDOW_BASE register, we
		 * need this to be untouched for accessing the send ring */
	    if ((altp->flags & ALT_HOST_UP) &&
		tgAddr >= TG_WINDOW_BASE && 
		(tgAddr + len) <= (TG_WINDOW_BASE + 4)) {
		status = NDIS_STATUS_INVALID_DATA;
		*bytes_written = 0;
		*bytes_needed = 0;
		break;
	    }
		
	    ALT_SWAP_MEMCPY_TODEV((MEMCPY_TYPE)altp->initNic->base + 
				   TG_DEV_ID + (tgAddr - TG_BEG_REGS),
				   nicMem.data, len);
	    move_bytes = sizeof(struct tg_mem);
	} else {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}
	break;

    case OID_VENDOR_READ_TG_REG:
	/* copy addrs and stuff into our local structure */
	nicReg.addr = ((struct tg_reg *)infoBuffer)->addr;
	move_source = (void *)&nicReg;

	/* validate some numbers */
	if (nicReg.addr > TG_MAX_REG) {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}

	/* LINTED pointer cast may result in improper alignment */
	nicReg.data = IN32(altp->initNic->handle, altp->initNic->base, 
			   nicReg.addr);
	ALTTRACEG(readReg, 0x5000d, nicReg.addr, nicReg.data, 0, 0);
	move_bytes = sizeof(struct tg_reg);
	break;

    case OID_VENDOR_WRITE_TG_REG:
	/* copy addrs and stuff into our local structure */
	nicReg.addr = ((struct tg_reg *)infoBuffer)->addr;
	nicReg.data = ((struct tg_reg *)infoBuffer)->data;
	move_source = (void *)&nicReg;

	ALTTRACEG(writeReg, 0x5000e, nicReg.addr, nicReg.data, 0, 0);
	  
	/* validate some numbers */
	/* special case on the WINDOW_BASE reg, we don't let the
	 * users touch this because we depend on it for the send
	 * ring */
	if (nicReg.addr > TG_MAX_REG ||
	    ((altp->flags & ALT_HOST_UP) &&
	     nicReg.addr == TG_WINDOW_BASE)) {
	    status = NDIS_STATUS_INVALID_DATA;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}
	    
	/* LINTED pointer cast may result in improper alignment */
	OUT32(altp->initNic->handle, altp->initNic->base, 
	      nicReg.addr, nicReg.data);

	move_bytes = sizeof(struct tg_reg);
	break;


    case OID_GEN_SUPPORTED_LIST:
	move_source = (void *)altSupportedOids;
	move_bytes = sizeof(altSupportedOids);
	break;

    case OID_GEN_HARDWARE_STATUS:
	/*
	 * I am seting the hardware stautus to ready for now,
	 * but would like to make sure that the receive rings have
	 * been setup properly.
	 */
	hardware_status = NdisHardwareStatusReady;
	move_source = (void *)&hardware_status;
	move_bytes = sizeof(NDIS_HARDWARE_STATUS);
	break;

    case OID_GEN_MEDIA_SUPPORTED:
    case OID_GEN_MEDIA_IN_USE:
	move_source = (void *)&medium;
	move_bytes = sizeof(NDIS_MEDIUM);
	break;

    case OID_GEN_MAXIMUM_LOOKAHEAD:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE - 4) :
		(MAXBLOCKSIZE - ALTEON_HEADER_SIZE - 4);
	break;

    case OID_GEN_MAXIMUM_FRAME_SIZE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE - 4) :
		(MAXBLOCKSIZE - ALTEON_HEADER_SIZE - 4);
	break;

    case OID_GEN_LINK_SPEED:
	generic_ulong = (ULONG)(10000000);
	break;

    case OID_GEN_TRANSMIT_BUFFER_SPACE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (SEND_RING_ENTRIES * JUMBOBLOCKSIZE) :
		(SEND_RING_ENTRIES * MAXBLOCKSIZE);
	break;

    case OID_GEN_RECEIVE_BUFFER_SPACE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (RECV_STANDARD_RING_ENTRIES * MAXBLOCKSIZE) +
	    (RECV_JUMBO_RING_ENTRIES * JUMBOBLOCKSIZE) :
		(RECV_STANDARD_RING_ENTRIES * MAXBLOCKSIZE);
	break;

    case OID_GEN_TRANSMIT_BLOCK_SIZE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE - 4) :
		(MAXBLOCKSIZE - ALTEON_HEADER_SIZE - 4);
	break;

    case OID_GEN_RECEIVE_BLOCK_SIZE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE - 4) :
		(MAXBLOCKSIZE - ALTEON_HEADER_SIZE - 4);
	break;

    case OID_GEN_VENDOR_ID:
	generic_ulong = 0;
	MEMCPY((MEMCPY_TYPE)&generic_ulong,
	       (MEMCPY_TYPE)altp->macAddr.ether_addr_octet, 3);
	break;

    case OID_GEN_VENDOR_DESCRIPTION:
	move_source = (void *)NAME_STRING;
	move_bytes = NAME_STRING_SIZE;
	break;

    case OID_GEN_VENDOR_DRIVER_VERSION:
	generic_ushort =
	  ((U16)VER_DRIVER_MAJOR << 8) | VER_DRIVER_MINOR;
	move_source = (void *)(&generic_ushort);
	move_bytes = sizeof(generic_ushort);
	break;

    case OID_GEN_CURRENT_PACKET_FILTER:
	/* build up the filtering bits from our state */
	generic_ulong =
	    ((nic->flags & ALT_PROM_MODE) ?
	     NDIS_PACKET_TYPE_PROMISCUOUS : 0) |
	    ((nic->flags & ALT_MULTICAST_MODE) ?
	     NDIS_PACKET_TYPE_ALL_MULTICAST : 0);
	break;

    case OID_GEN_CURRENT_LOOKAHEAD:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - ALTEON_HEADER_SIZE - 4) :
		(MAXBLOCKSIZE - ALTEON_HEADER_SIZE - 4);
	break;

    case OID_GEN_DRIVER_VERSION:
	generic_ushort =
	  ((U16)ALT_NDIS_MAJOR_VERSION << 8) | ALT_NDIS_MINOR_VERSION;
	move_source = (void *)&generic_ushort;
	move_bytes = sizeof(generic_ushort);
	break;

    case OID_GEN_PROTOCOL_OPTIONS:
#if NEW_REG
	/*
	 * this OID is overloaded; it will also cause the tuneParms to
	 * be refreshed from the registry.
	 */
	UpdateFromRegistry(altp, &inst);
#endif /* NEW_REG */
	generic_ulong = 0;
	break;

    case OID_GEN_MAXIMUM_TOTAL_SIZE:
	generic_ulong = (nic->flags & ALT_JUMBO_ENABLED) ?
	    (JUMBOBLOCKSIZE - 4) : (MAXBLOCKSIZE - 4 );
	break;

    case OID_GEN_MAC_OPTIONS:
	generic_ulong = NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA |
	    NDIS_MAC_OPTION_FULL_DUPLEX |
	    NDIS_MAC_OPTION_NO_LOOPBACK;
	break;

    case OID_GEN_MEDIA_CONNECT_STATUS:
	if (nic->flags & ALT_LINK_READY) {
	    media_status = NdisMediaStateConnected;
	} else {
	    media_status = NdisMediaStateDisconnected;
	}

	move_source = (void *)&media_status;
	move_bytes = sizeof(NDIS_MEDIA_STATE);
	break;

    case OID_GEN_MAXIMUM_SEND_PACKETS:
	generic_ulong = (ULONG)SEND_RING_ENTRIES / 8;
	break;

    case OID_GEN_XMIT_OK:
	/* note that this blunders if we roll over a 32 bit word */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCOutUcastPkts) +
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCOutMulticastPkts) +
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCOutBroadcastPkts);
	break;

    case OID_GEN_RCV_OK:
	/* note that this blunders if we roll over a 32 bit word */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInUcastPkts) +
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCInMulticastPkts) +
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCInBroadcastPkts);
	break;

    case OID_GEN_XMIT_ERROR:
	generic_ulong = nic->genInfo->stats.ifOutErrors;
	break;

    case OID_GEN_RCV_ERROR:
	generic_ulong = nic->genInfo->stats.ifInErrors;
	break;

    case OID_GEN_RCV_NO_BUFFER:
	generic_ulong = nic->genInfo->stats.ifInDiscards;
	break;

    case OID_GEN_DIRECTED_BYTES_XMIT:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCOutOctets);
	break;

    case OID_GEN_DIRECTED_FRAMES_XMIT:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCOutUcastPkts);
	break;

    case OID_GEN_MULTICAST_BYTES_XMIT:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCOutOctets);
	break;

    case OID_GEN_MULTICAST_FRAMES_XMIT:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCOutMulticastPkts);
	break;

    case OID_GEN_BROADCAST_BYTES_XMIT:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCOutOctets);
	break;

    case OID_GEN_BROADCAST_FRAMES_XMIT:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->
				      stats.ifHCOutBroadcastPkts);
	break;

    case OID_GEN_DIRECTED_BYTES_RCV:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInOctets);
	break;

    case OID_GEN_DIRECTED_FRAMES_RCV:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInUcastPkts);
	break;

    case OID_GEN_MULTICAST_BYTES_RCV:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInOctets);
	break;

    case OID_GEN_MULTICAST_FRAMES_RCV:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInMulticastPkts);
	break;

    case OID_GEN_BROADCAST_BYTES_RCV:
	/* XXX not correct */
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInOctets);
	break;

    case OID_GEN_BROADCAST_FRAMES_RCV:
	generic_ulong =
	    NdisGetPhysicalAddressHigh(nic->genInfo->stats.ifHCInBroadcastPkts);
	break;

    case OID_GEN_RCV_CRC_ERROR:
	generic_ulong = nic->genInfo->stats.dot3StatsFCSErrors;
	break;

    case OID_GEN_TRANSMIT_QUEUE_LENGTH:
	generic_ulong = nic->bufsOnSendRing;
	break;

    case OID_802_3_PERMANENT_ADDRESS:
	/* we've done this already, we should still be able to, just ignore
	 * the return code */
	readTGSerEeprom(altp,
			(U32)(((eeprom_layout_t *)TG_BEG_SER_EEPROM)->
			      manufact.serial_number),
			(char *)macspot, 8, FALSE);
	/* copy the MAC address */
	MEMCPY(generic_array, ((char *)macspot) + 2, 6);
	move_source = (void *)generic_array;
	move_bytes = sizeof(struct ether_addr);
	break;

    case OID_802_3_CURRENT_ADDRESS:
        MEMCPY(generic_array, altp->macAddr.ether_addr_octet,
	       sizeof(struct ether_addr));
	move_source = (void *)generic_array;
	move_bytes = sizeof(struct ether_addr);
	break;

    case OID_802_3_MAXIMUM_LIST_SIZE:
	generic_ulong = MAX_MCAST_ENTRIES;
	break;

    case OID_802_3_MAC_OPTIONS:
	generic_ulong = 0;
	break;

    case OID_802_3_RCV_ERROR_ALIGNMENT:
	generic_ulong = nic->genInfo->stats.dot3StatsFrameTooLongs;
	break;

    case OID_802_3_XMIT_ONE_COLLISION:
	generic_ulong = nic->genInfo->stats.dot3StatsSingleCollisionFrames;
	break;

    case OID_802_3_XMIT_MORE_COLLISIONS:
	generic_ulong = nic->genInfo->stats.dot3StatsMultipleCollisionFrames;
	break;

    case OID_802_3_XMIT_DEFERRED:
	generic_ulong = nic->genInfo->stats.dot3StatsDeferredTransmissions;
	break;

    case OID_802_3_XMIT_MAX_COLLISIONS:
	generic_ulong = nic->genInfo->stats.dot3StatsExcessiveCollisions;
	break;

    case OID_802_3_RCV_OVERRUN:
	generic_ulong = nic->genInfo->stats.nicMacRxTruncNoResources +
	    nic->genInfo->stats.nicMacRxDropSpace;
	break;

    case OID_802_3_XMIT_UNDERRUN:
	/* we can't do this */
	generic_ulong = 0;
	break;

    case OID_802_3_XMIT_HEARTBEAT_FAILURE:
	generic_ulong = 0;
	break;

    case OID_802_3_XMIT_TIMES_CRS_LOST:
	generic_ulong = nic->genInfo->stats.dot3StatsCarrierSenseErrors;
	break;

    case OID_802_3_XMIT_LATE_COLLISIONS:
	generic_ulong = nic->genInfo->stats.dot3StatsSingleCollisionFrames;
	break;

#if USE_CKSUM_OFFLOAD
    case OID_TCP_TASK_OFFLOAD:
	neededLen = sizeof(NDIS_TASK_OFFLOAD_HEADER) +
	    sizeof(NDIS_TASK_OFFLOAD) + sizeof(NDIS_TASK_TCP_IP_CHECKSUM);
        ALTTRACEE(tcpoffld, 0x80004, bufLen, neededLen, 0, 0);
        if (bufLen < neededLen) {
	    status = NDIS_STATUS_INVALID_LENGTH;
	    *bytes_written = 0;
	    *bytes_needed = neededLen;
	    ERROR1("Get TCP_TASK_OFFLOAD has invalid length, requesting more");
	    break;
	}
	BZERO(generic_array, 80);
	/* copy passed in offload header */
	taskOffloadHdr = (NDIS_TASK_OFFLOAD_HEADER *)generic_array;
	/* STRUCT COPY */
	*taskOffloadHdr = *(NDIS_TASK_OFFLOAD_HEADER *)infoBuffer;
	if (taskOffloadHdr->Version != NDIS_TASK_OFFLOAD_VERSION) {
	    status = NDIS_STATUS_BAD_VERSION;
	    *bytes_written = 0;
	    *bytes_needed = 0;
	    break;
	}	
		
	/* set up offset */
	taskOffloadHdr->OffsetFirstTask = sizeof(NDIS_TASK_OFFLOAD_HEADER);

        /* fill in the task offload struct in the generic_array area */
        taskOffload = (NDIS_TASK_OFFLOAD *)(generic_array + 
					    sizeof(NDIS_TASK_OFFLOAD_HEADER));
	taskOffload->Size = sizeof(NDIS_TASK_OFFLOAD);
	taskOffload->Task = TcpIpChecksumNdisTask;
	taskOffload->OffsetNextTask = 0;
	taskOffload->TaskBufferLength = sizeof(NDIS_TASK_TCP_IP_CHECKSUM);

	/* fill in the tcpip offload struct in the right place in the
	 * generic_array area */
	tcpipOffload = (NDIS_TASK_TCP_IP_CHECKSUM *)taskOffload->TaskBuffer;
	tcpipOffload->V4Transmit.IpOptionsSupported = 1;
	tcpipOffload->V4Transmit.TcpOptionsSupported = 1;
	tcpipOffload->V4Transmit.TcpChecksum = 1;
	tcpipOffload->V4Transmit.UdpChecksum = 1;
	tcpipOffload->V4Transmit.IpChecksum = 1;

	tcpipOffload->V6Transmit.IpOptionsSupported = 0;
	tcpipOffload->V6Transmit.TcpOptionsSupported = 0;
	tcpipOffload->V6Transmit.TcpChecksum = 0;
	tcpipOffload->V6Transmit.UdpChecksum = 0;

	tcpipOffload->V4Receive.IpOptionsSupported = 1;
	tcpipOffload->V4Receive.TcpOptionsSupported = 1;
	tcpipOffload->V4Receive.TcpChecksum = 1;
	tcpipOffload->V4Receive.UdpChecksum = 1;
	tcpipOffload->V4Receive.IpChecksum = 1;

	tcpipOffload->V6Receive.IpOptionsSupported = 0;
	tcpipOffload->V6Receive.TcpOptionsSupported = 0;
	tcpipOffload->V6Receive.TcpChecksum = 0;
	tcpipOffload->V6Receive.UdpChecksum = 0;

        ALTTRACEE(tcpoffld, 0x80007, taskOffload, tcpipOffload, generic_array,
		  0);
	ALTTRACEE(taskstc, 0x80008, taskOffload->Size, taskOffload->Task,
		  taskOffload->OffsetNextTask, taskOffload->TaskBufferLength);
	move_source = (void *)generic_array;
	move_bytes = neededLen;
	ALTTRACEE(tcpipstc, 0x8000a, *(U32 *)taskOffload->TaskBuffer,
		  *(U32 *)&taskOffload->TaskBuffer[4], 
		  *(U32 *)&taskOffload->TaskBuffer[8], 
		  *(U32 *)&taskOffload->TaskBuffer[12]);
	break;
#endif /* USE_CKSUM_OFFLOAD */

    default:
	status = NDIS_STATUS_INVALID_OID;
	break;
    }

    if (status == NDIS_STATUS_SUCCESS) {
	if (move_bytes > bufLen) {
	    /* Not enough room in infoBuffer. Punt */
	    *bytes_needed = move_bytes;
	    status = NDIS_STATUS_INVALID_LENGTH;
	} else {
	    /* Store result. */
	    MEMCPY(infoBuffer, move_source, move_bytes);
	    *bytes_written = move_bytes;
	}
    }

    return(status);
}

STATIC void
ReadNicTraceBuf(struct altinst *altp, U8 *buf)
{
    bool running;
    U32 curNicTb, nicTb, endNicTb;

    running = stopFW(altp);
    curNicTb = RDSHMEM(altp->initNic->genComm->gen_com.nic_trace_ptr);
    nicTb = RDSHMEM(altp->initNic->genComm->gen_com.nic_trace_start);
    endNicTb = nicTb + RDSHMEM(altp->initNic->genComm->gen_com.nic_trace_len);

    /* protect against a core dump */
    if (!curNicTb)
        return;

    if (readTGMem(altp, curNicTb + (ALT_TRACE_ELEM_SIZE * sizeof(U32)), buf,
		  endNicTb - curNicTb -(ALT_TRACE_ELEM_SIZE * sizeof(U32)),
		  FALSE, TRUE) < 0) {
	ERROR1("Could not read NIC trace buffer\n");
    }
    if (readTGMem(altp, nicTb,
		  (buf - (ALT_TRACE_ELEM_SIZE * sizeof(U32))) +
		  endNicTb - curNicTb,
		  curNicTb - nicTb, FALSE, TRUE) < 0) {
	ERROR1("Could not read NIC trace buffer\n");
    }
    if (running) {
	restartFW(altp);
    }
}

#if NEW_REG
/* what's this? XXX */
STATIC NTSTATUS
QueryCallback(PWSTR name, U32 type, void *data, U32 len,
	      void *context, void *entry_context)
{
    ALTTRACE(ntQryCB, 0x8000b, 0, 0, 0, 0);
    return(0);	/* hack */
}
#endif /* NEW_REG */

STATIC void
UpdateFromRegistry(struct altinst *altp, U32 *inst)
{
    struct tg_nic *nic;
    alt_nt_inst_parms_t *inst_p;
    NDIS_STATUS status;
    U32 updateTP = 0;
    U32 regChanged = 0;
    U32 linkChanged = 0;
    U32 rc;
    alt_reg_tab_t *rtp;
    PWSTR path;

#if NEW_REG
#if 0
    status = altReadTuneParms(nic, inst);

    if (status != NDIS_STATUS_SUCCESS) {
	ALTTRACE(ErReadTP, 0x8000c, altp, *inst, 0, 0);
        return;
    }
#else /* XXX */
    /*
     * now read instance parms
     */
    rtp = alt_inst_reg_tab;
    ReadRegTableNoHandle(altp, rtp, 2);	/* XXX only check 1st 2 entries */

#endif /* XXX */
#endif /* NEW_REG */


    /* XXX debug... */
    if (*inst != altp->instance) {
	ALTTRACE(ErrInst, 0x8000d, altp->instance, *inst, 0, 0);
        return;
    }

    nic = altp->nic;

    /* common parameters */
    if (nic->stat_ticks != nic->tuneParms.glob.stat_ticks) {
	nic->stat_ticks = nic->tuneParms.glob.stat_ticks;
	updateTP = 1;
	regChanged = 1;
    }

    if (nic->send_bd_ticks != nic->tuneParms.glob.send_coal_ticks) {
	nic->send_bd_ticks = nic->tuneParms.glob.send_coal_ticks;
	updateTP = 1;
	regChanged = 1;
    }

    if (nic->recv_bd_ticks != nic->tuneParms.glob.recv_coal_ticks) {
	nic->recv_bd_ticks = nic->tuneParms.glob.recv_coal_ticks;
	updateTP = 1;
	regChanged = 1;
    }

    if (nic->send_bd_coal != nic->tuneParms.glob.send_coal_max) {
	nic->send_bd_coal = nic->tuneParms.glob.send_coal_max;
	updateTP = 1;
	regChanged = 1;
    }

    if (nic->recv_bd_coal != nic->tuneParms.glob.recv_coal_max) {
	nic->recv_bd_coal = nic->tuneParms.glob.recv_coal_max;
	updateTP = 1;
	regChanged = 1;
    }

    /* this incredibly ugly construction says that if the state of the
     * INIT_MEM_WR_INV_DISABLE bit is set differently than the tuneparms
     * field then the init_flags needs to be updated */
    if ((U32)((nic->init_flags & ALT_INIT_MEM_WR_INV_DISABLE) ==
	      ALT_INIT_MEM_WR_INV_DISABLE) !=
	nic->tuneParms.glob.pci_mem_inv) {
	if (nic->tuneParms.glob.pci_mem_inv)
	    nic->init_flags |= ALT_INIT_MEM_WR_INV_DISABLE;
	else
	    nic->init_flags &= ~ALT_INIT_MEM_WR_INV_DISABLE;
	/* XXX do I need to do anything here?  Yes, bunches of stuff */
	regChanged = 1;
    }

    if (nic->latTimer != nic->tuneParms.glob.pci_latency_timer) {
	nic->latTimer = (U8)nic->tuneParms.glob.pci_latency_timer;
	/* XXX do I need to do anything here? Yes, set it in the PCI region */
	regChanged = 1;
    }

    /* instance parameters */
    inst_p = &(nic->tuneParms.inst[*inst]);

    /* link stuff */

    if (inst_p->fdr_filter) {
        if (nic->flags & ALT_DO_FDR_FILTERING) {
	    /* nothing */
	} else {
	    altSetFdrMode(altp, (bool)inst_p->fdr_filter);
	    updateTP = 1;
	    regChanged = 1;
	}
    } else {
        if (nic->flags & ALT_DO_FDR_FILTERING) {
	    altSetFdrMode(altp, (bool)inst_p->fdr_filter);
	    updateTP = 1;
	    regChanged = 1;
	} else {
	    /* nothing */
	}
    }

    if (inst_p->tx_flow_control) {
        if (nic->flags & ALT_TX_FLOW_CONTROL) {
	    /* nothing */
	} else {
	    nic->flags |= ALT_TX_FLOW_CONTROL;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) |
		LNK_TX_FLOW_CTL_Y));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	}
    } else {
        if (nic->flags & ALT_TX_FLOW_CONTROL) {
	    nic->flags &= ~ALT_TX_FLOW_CONTROL;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) &
		~LNK_TX_FLOW_CTL_Y));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	} else {
	    /* nothing */
	}
    }

    if (inst_p->rx_flow_control) {
        if (nic->flags & ALT_RX_FLOW_CONTROL) {
	    /* nothing */
	} else {
	    nic->flags |= ALT_RX_FLOW_CONTROL;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) |
		LNK_RX_FLOW_CTL_Y));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	}
    } else {
        if (nic->flags & ALT_RX_FLOW_CONTROL) {
	    nic->flags &= ~ALT_RX_FLOW_CONTROL;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) &
		~LNK_RX_FLOW_CTL_Y));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	} else {
	    /* nothing */
	}
    }

    if (inst_p->link_negotiation) {
        if (nic->flags & ALT_DO_LINK_NEGOTIATION) {
	    /* nothing */
	} else {
	    nic->flags |= ALT_DO_LINK_NEGOTIATION;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) |
		LNK_NEGOTIATE));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	}
    } else {
        if (nic->flags & ALT_DO_LINK_NEGOTIATION) {
	    nic->flags &= ~ALT_DO_LINK_NEGOTIATION;
	    WRSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation,
		(RDSHMEM(nic->genComm->gen_com.tuneParms.link_negotiation) &
		~LNK_NEGOTIATE));
	    updateTP = 1;
	    regChanged = 1;
	    linkChanged = 1;
	} else {
	    /* nothing */
	}
    }

    if (inst_p->jumbo_frames) {
        if (nic->flags & ALT_JUMBO_ENABLED) {
	    /* nothing */
	} else {
	    /*
	     * can't deal with enabling jumbos after host stack
	     * is already up, for now...
	     */
	    if (!(altp->flags & ALT_HOST_UP) && !altInitJumbo(altp)) {
#if 0
		altp->jumbo[alt_param_arr[ND_INSTANCE].parm_val] = TRUE;
#endif /* XXX */
		nic->flags |= ALT_JUMBO_ENABLED;
		if (altp->state == ALT_FW_RUNNING)
		    FillJumboRecvRing(altp);

#if 0
		/* and on the redundant card */
		if (altp->flags & ALT_REDUNDANT) {
		    altp2 = altp->attachNic->initAltp;
		    if (rc = altInitJumbo(altp2))
			return(rc);
		    altp2->jumbo[alt_param_arr[ND_INSTANCE].parm_val] = TRUE;
		    altp2->initNic->flags |= ALT_JUMBO_ENABLED;
		    if (altp2->state == ALT_FW_RUNNING)
			FillJumboRecvRing(altp2);
		}
#endif /* XXX */

		/* since jumbo enabled, inform nic about new mtu */
		WRSHMEM(nic->genComm->gen_com.ifMTU, JUMBOBLOCKSIZE);
		altUpdateGencomStats(altp);

		updateTP = 1;
		regChanged = 1;
	    }
	}
    } else {
        if (nic->flags & ALT_JUMBO_ENABLED) {
	    nic->flags &= ~ALT_JUMBO_ENABLED;
	    updateTP = 1;
	    regChanged = 1;
	} else {
	    /* nothing */
	}
    }

    if ((U32)altTracing != inst_p->host_tracing) {
	altTracing = inst_p->host_tracing;
	regChanged = 1;
    }

#if 0
    if (nic_trace != inst_p->nic_tracing) {
	nic_trace = inst_p->nic_tracing;
	updateTP = 1;
	regChanged = 1;
    }
#endif /* XXX */

    if (updateTP)
	altUpdateTuneParms(nic);

    if (linkChanged)
	altDoLinkNegotiation(altp);

    if (regChanged) {
	/* mark the registry as consistent */
    }
}

/*
 * altReadTuneParms() is protected protected against MP interactions
 * by the semantics of the miniportQueryInformation() routine, through
 * which all accesses are serialized.
 */

NDIS_STATUS
altReadTuneParms(struct tg_nic *nic, U32 *inst)
{
    NDIS_STATUS status;
    NDIS_HANDLE confighandle;
    alt_reg_tab_t *rtp;
    alt_nt_tune_parms_t *tpp;
    int i;

    NdisOpenConfiguration(&status, &confighandle, nic->configHandle);
    if (status != NDIS_STATUS_SUCCESS) {
	ALTTRACE(ErOpnCnf, 0x8000e, nic, status, nic->configHandle, 0);
        return(status);
    }

    /*
     * first read global parms
     */
    rtp = alt_glob_reg_tab;
    ReadRegTable(rtp, ALT_REG_KEY_GLOB_LAST_INDEX, confighandle);

    /*
     * now copy from global area into the nt_TuneParms area hanging off
     * the nic structure...
     */
    nic->tuneParms.glob = ntp.glob;    /* struct copy */

    /*
     * now read instance parms
     */
    rtp = alt_inst_reg_tab;
    ReadRegTable(rtp, ALT_REG_KEY_INST_LAST_INDEX, confighandle);

    /*
     * now copy from global area into the tuneParms area hanging off
     * the nic structure...
     */
    nic->tuneParms.inst[inst_tp.instance] = inst_tp; /* struct copy */

    /* set instance */
    *inst = inst_tp.instance;

    NdisCloseConfiguration(confighandle);

    return(NDIS_STATUS_SUCCESS);
}


/*
 * rtp points to the start of a reg_tab, last is the last index,
 * confighandle is a previously opened registry handle
 */

STATIC void
ReadRegTable(alt_reg_tab_t *rtp, U32 last, NDIS_HANDLE confighandle)
{
    U32 i;
    PNDIS_CONFIGURATION_PARAMETER retval;
    NDIS_STRING ndis_string;
    U32 tmp;

    for (i = 0; i < (last + 1); i++) {
	NdisInitializeString(&ndis_string, rtp->string);

        NdisReadConfiguration(&(rtp->ndis_status),
			      &retval, confighandle,
			      &ndis_string, rtp->ndis_type);

	switch (rtp->ndis_status) {
	    case NDIS_STATUS_SUCCESS:
		switch (rtp->ndis_type) {
		    case NdisParameterInteger:
		    case NdisParameterHexInteger:
			tmp = retval->ParameterData.IntegerData;
			if ((tmp < rtp->min) ||
			    (tmp > rtp->max)) {
			    /* outside the boundaries, set to default... */
			    *(rtp->cfg_parm) = rtp->def;
			    ALTTRACE(regObnds, 0x80002,
				     *((U32 *)&rtp->string[0]),
				     *((U32 *)&rtp->string[4]),
				     *((U32 *)&rtp->string[8]),
				     *rtp->cfg_parm);
			} else {
			    *(rtp->cfg_parm) =
				retval->ParameterData.IntegerData;
			    ALTTRACE(regFound, 0x80003,
				     *((U32 *)&rtp->string[0]),
				     *((U32 *)&rtp->string[4]),
				     *((U32 *)&rtp->string[8]),
				     *rtp->cfg_parm);
			}
			break;

		    /* XXX need to work on these... */
		    case NdisParameterString:
		    case NdisParameterMultiString:
		    default:
			NOTE2("ReadRegTable:unknown parameter type [%x]",
			    rtp->ndis_type);
		}
		break;

	    case NDIS_STATUS_FAILURE:
		/* on failure, set to default value */
		*(rtp->cfg_parm) = rtp->def;
		ALTTRACE(regNtFnd, 0x80005,
			 *((U32 *)&rtp->string[0]),
			 *((U32 *)&rtp->string[4]),
			 *((U32 *)&rtp->string[8]),
			 *rtp->cfg_parm);
		break;

	    case NDIS_STATUS_RESOURCES:
		ALTTRACE(regResrc, 0x80006, 0, 0, 0, 0);
		break;

	    default:
		ALTTRACE(regWhat?, 0x80009, 0, 0, 0, 0);

        }
	NdisFreeString(ndis_string);
        rtp++;
    }
}


#if NEW_REG
/*
 * rtp points to the start of a reg_tab, last is the last index,
 * confighandle is a previously opened registry handle
 */

void
ReadRegTableNoHandle(struct altinst *altp, alt_reg_tab_t *rtp, U32 last)
{
    U32 i;
    U32 tmp;
    U32 default_response;
    U32 direct_response;
    NDIS_STRING request;

#define QUERY_TAB_ENTRIES	2

    RTL_QUERY_REGISTRY_TABLE query_tab[QUERY_TAB_ENTRIES];

    /* XXX need to use passed instance... */
    UNICODE_STRING uni_altndis1;
    UNICODE_STRING path;

    RtlInitUnicodeString (&uni_altndis1, L"\\REGISTRY\\Machine\\System\\CurrentControlSet\\Services\\altndis1\\Parameters");

    path.MaximumLength = uni_altndis1.MaximumLength;
    path.Length = 0;

    path.Buffer = ExAllocatePool(NonPagedPool, path.MaximumLength);
    if (!path.Buffer)
	return;

    RtlZeroMemory(path.Buffer, path.MaximumLength);
    RtlCopyUnicodeString(&path, &uni_altndis1);


    for (i = 0; i < (last + 1); i++) {

	/* populate query_tab */
	BZERO(query_tab, sizeof(RTL_QUERY_REGISTRY_TABLE) * QUERY_TAB_ENTRIES);

	NdisInitializeString(&request, rtp->string);
	direct_response = 0;
	default_response = rtp->def;

	query_tab[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
	query_tab[0].Name = request.Buffer;
	query_tab[0].EntryContext = &direct_response;
	query_tab[0].DefaultType = REG_DWORD;
	query_tab[0].DefaultData = &default_response;
	query_tab[0].DefaultLength = sizeof(U32);

	rtp->ndis_status = RtlQueryRegistryValues(
	    RTL_REGISTRY_ABSOLUTE | 0,
	    path.Buffer, query_tab, 0, 0);

	switch (rtp->ndis_status) {
	    case STATUS_SUCCESS:
		/*
		 * XXX Always assumes returned type is DWORD...
		 */
		tmp = direct_response;
		if ((tmp < rtp->min) ||
		    (tmp > rtp->max)) {
		    /* outside the boundaries, set to default... */
		    *(rtp->cfg_parm) = rtp->def;
		    ALTTRACE(regObnds, 0x8000f,
			     *((U32 *)&rtp->string[0]),
			     *((U32 *)&rtp->string[4]),
			     *((U32 *)&rtp->string[8]),
			     *rtp->cfg_parm);
		} else {
		    *(rtp->cfg_parm) = direct_response;
		    ALTTRACE(regFound, 0x80010,
			     *((U32 *)&rtp->string[0]),
			     *((U32 *)&rtp->string[4]),
			     *((U32 *)&rtp->string[8]),
			     *rtp->cfg_parm);
		}
		break;

	    case STATUS_OBJECT_NAME_NOT_FOUND:
		/* on failure, set to default value */
		*(rtp->cfg_parm) = rtp->def;
		ALTTRACE(regNtFnd, 0x80011,
			 *((U32 *)&rtp->string[0]),
			 *((U32 *)&rtp->string[4]),
			 *((U32 *)&rtp->string[8]),
			 *rtp->cfg_parm);
		break;

	    case STATUS_INVALID_PARAMETER:
		ALTTRACE(regInval, 0x80012, 0, 0, 0, 0);
		break;

	    default:
		ALTTRACE(regWhat?, 0x80013, 0, 0, 0, 0);

        }
	NdisFreeString(request);
        rtp++;
    }
}
#endif /* NEW_REG */
