/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */

/*
 * HISTORY
 * $Log: drvstat.h,v $
 * Revision 1.1.2.18  1999/02/16  16:43:42  leonid
 * 	Added OID_DA_CHANGE_MAC_ADDRESS_R to support mac address change for the failover operation.
 * 	[1999/02/16  16:43:08  leonid]
 *
 * Revision 1.1.2.17  1999/01/19  00:42:04  leonid
 * 	DA specific OIDs made generic.
 * 	[1999/01/19  00:41:43  leonid]
 * 
 * Revision 1.1.2.16  1998/11/25  22:35:50  leonid
 * 	3Com OIDs modified to reflect their spec change
 * 	[1998/11/25  22:33:22  leonid]
 * 
 * Revision 1.1.2.15  1998/11/13  23:11:02  ted
 * 	Add debugger ioctls (compiles but undebugged)
 * 	[1998/11/13  23:09:34  ted]
 * 
 * Revision 1.1.2.14  1998/11/09  18:22:15  hayes
 * 	add TUNEPARMS support
 * 	[1998/11/09  18:16:55  hayes]
 * 
 * Revision 1.1.2.13  1998/11/03  23:25:37  ted
 * 	Missed in merge
 * 	[1998/11/03  22:31:58  ted]
 * 
 * 	Fix Mini ring handling, add checksum offload fail stats
 * 	[1998/11/03  22:23:51  ted]
 * 
 * Revision 1.1.2.11  1998/10/13  03:33:43  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:12  ted]
 * 
 * Revision 1.1.2.10  1998/10/01  01:03:07  leonid
 * 	3COM OIDs re-defined as "OP2" only.
 * 	[1998/10/01  01:02:48  leonid]
 * 
 * Revision 1.1.2.9  1998/09/29  18:48:10  leonid
 * 	3Com specific OIDs added.
 * 	[1998/09/29  18:46:35  leonid]
 * 
 * Revision 1.1.2.8  1998/08/20  18:44:09  leonid
 * 	Definition for custom OIDs is fixed
 * 	[1998/08/20  18:43:33  leonid]
 * 
 * Revision 1.1.2.7  1998/08/03  22:06:38  hayes
 * 	cleanup, add registry options
 * 	[1998/08/03  22:05:20  hayes]
 * 
 * Revision 1.1.2.6  1998/08/03  17:11:39  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:07:56  hayes]
 * 
 * 	checkin so merge w/ fc can happen
 * 	[1998/07/29  23:16:10  hayes]
 * 
 * Revision 1.1.2.5  1998/06/16  16:46:07  ted
 * 	NT initialization cleanup/rewrite
 * 	[1998/06/16  16:43:12  ted]
 * 
 * Revision 1.1.2.4  1998/06/12  23:47:34  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:28  ted]
 * 
 * 	snapshot of merge with new Registry handling
 * 	[1998/06/10  22:15:25  ted]
 * 
 * Revision 1.1.2.3  1998/06/08  23:55:45  ted
 * 	Snapshot NT generification phase 1
 * 	[1998/06/04  19:52:26  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:29  ted]
 * 
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:08:48  ted]
 * 
 * 	snapshot 12.0 NT jumbo support
 * 	[1998/05/20  20:29:36  ted]
 * 
 * Revision 1.1.2.2  1997/12/18  23:51:33  venkat
 * 	merged from ge_0_37_latest
 * 	[1997/12/18  23:31:25  venkat]
 * 
 * Revision 1.1.2.1  1997/06/11  21:25:35  akhan
 * 	moved to ndis
 * 	[1997/06/11  21:25:20  akhan]
 * 
 * Revision 1.1.2.2  1997/05/23  00:06:56  akhan
 * 	Rearranged stat fields.
 * 	[1997/05/23  00:06:42  akhan]
 * 
 * Revision 1.1.2.1  1997/04/26  01:52:19  akhan
 * 	Template for driver stats.
 * 	[1997/04/26  01:51:54  akhan]
 * 
 * $EndLog$
 */


#ifndef _DRVSTAT_H
#define _DRVSTAT_H

#define SAMPLE_SIZE 150
#define BIN_SIZE (1500 / SAMPLE_SIZE ) +1


typedef struct _DRIVER_STATS
{
    void *base;
    U32 flags;
    U32 init_flags;
    U32 tigonRev;		/* Tigon revision of this piece of hardware */
    void *genInfo;
    void *eventArr;
    void *eventProdPtr;
    void *sendRing;
    U32 sendBufCount;
    U32 send_producer;
    U32 send_consumer;
    void *sendConsPtr;
    void *recvStdRing;
    void *recvJumboRing;
    void *recvMiniRing;
    void *recvReturnRing;
    U32 recvStdBufCount;
    U32 recv_std_producer;
    U32 recv_std_consumer;
    U32 recvJumboBufCount;
    U32 recv_jumbo_producer;
    U32 recv_jumbo_consumer;
    U32 recvMiniBufCount;
    U32 recv_mini_producer;
    U32 recv_mini_consumer;
    U32 recv_return_producer;
    U32 recv_return_consumer;
    void *recvReturnProdPtr;
    U32 inpkts;
    U32 bufsInStdRecvRing;
    U32 bufsInJumboRecvRing;
    U32 bufsInMiniRecvRing;
    U32 bufsOnSendRing;
    U32 bufsLowWater;
    U32 recvStdCoal;
    U32 recvJumboCoal;
    U32 recvMiniCoal;
    U32 outpkts;
    U32 outDrops;
    U32 outDropsDmaResources;
    U32 outDropsNoSpace;
    U32 flowControls;
    U32 reapedFromSend;
    U32 cmdFull;
    U32 pendPkts;
    U32 requeuedPkts;
    U32 dataRtnPkts;
    U32 badIpCksums;
    U32 badTcpCksums;
    U32 badUdpCksums;
    U32 lowResources;
    U32 normRecv;
    U32 outEarly;
} DRIVER_STATS, *PDRIVER_STATS;

#define OID_VENDOR_GET_TIGON_REGS   0xFF010100
#define OID_VENDOR_GET_NIC_BUFFER   0xFF010101
#define OID_VENDOR_GET_HOST_BUFFER  0xFF010102
#define OID_VENDOR_CLEAR_NIC_STATS  0xFF010104
#define OID_VENDOR_GET_NIC_STATS    0xFF010105
#define OID_VENDOR_GET_DRIVER_STATS 0xFF010106

/* 3COM specific OIDs */
#define OID_DA_CHANGE_MAC_ADDRESS   0xFFDA0002
#define OID_DA_CHANGE_MAC_ADDRESS_R 0xFF010103
#define OID_DA_NIC_CLASS	    0xFFDA0014
#define OP2_TAG		            0xADC33CDA
#define OP2_MINIPORT_TAG 	    0x3CDAADC3

/* OIDs for modifying tuning stuff on the fly */
#define OID_VENDOR_GET_TUNEPARMS    0xff010120
#define OID_VENDOR_SET_TUNEPARMS    0xff010121

/* OIDs to support gdb on NT */
#define OID_VENDOR_READ_TG_MEM      0xff010130
#define OID_VENDOR_WRITE_TG_MEM     0xff010131
#define OID_VENDOR_READ_TG_REG      0xff010132
#define OID_VENDOR_WRITE_TG_REG     0xff010133

/* TIGON based ioctls */
struct tg_reg {
    U32     data;
    U32     addr;		/* really an offset */
};

#define MAX_TG_DATA_LEN 1024
struct tg_mem {
    U32     tgAddr;
    int     len;
    U8      data[MAX_TG_DATA_LEN];
};

#endif /* _DRVSTAT_H */

