/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997
 * All rights reserved
 *
 */
/*
 * HISTORY
 * $Log: altstat.c,v $
 * Revision 1.1.2.6  1998/11/16  23:35:08  ted
 * 	Add printing of bad checksum info
 * 	[1998/11/16  23:33:06  ted]
 *
 * Revision 1.1.2.5  1998/10/13  03:33:44  ted
 * 	Support Mini ring, compiles on Solaris
 * 	[1998/10/13  03:31:13  ted]
 * 
 * Revision 1.1.2.4  1998/07/14  16:50:46  ted
 * 	Support ndis 5 devices and -5 option
 * 	[1998/07/14  02:18:57  ted]
 * 
 * Revision 1.1.2.3  1998/06/12  23:47:51  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:43  ted]
 * 
 * Revision 1.1.2.2  1998/06/08  23:55:56  ted
 * 	Make it much more like Unix version
 * 	[1998/06/04  19:50:31  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:32  ted]
 * 
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:08:51  ted]
 * 
 * 	snapshot 12.0 NT jumbo support
 * 	[1998/05/20  20:29:39  ted]
 * 
 * Revision 1.1.4.2  1998/05/13  23:23:12  ted
 * 	checkpoint
 * 
 * Revision 1.1.2.1  1998/04/21  18:33:16  hayes
 * 	remove fundefs.h
 * 	[1998/04/21  18:33:00  hayes]
 * 
 * 	first cut at cleaned up altstat
 * 	[1998/04/21  18:24:09  hayes]
 * 
 * $EndLog$
 */

#ifndef U32
#define U64 unsigned __int64
#define U32 unsigned long
#define U16 unsigned short
#define U8  unsigned char
#define S32 signed long
#define S16 signed short
#define S8  signed char
#define bool unsigned char
#endif /* U32 */

#define VOLATILE volatile

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <winioctl.h>
#include <ntddndis.h>

#include "drvstat.h"

#include "alt.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"

/* forward declarations... */
static void PrintNicProfile(struct tg_stats *statp);
static void PrintAlt(DRIVER_STATS *nicp);
static void PrintTGStats(struct tg_stats *statp);
static bool ReadNicStats(void);
static bool ReadDriverStats(void);
static LPTSTR GetLastErrorText(LPTSTR, U64);
static void ClearNicStats(void);
static void usage(char *);

#undef printf

#define PROG "altstat"

static char *handler_names[32] =
{
    "idle_time (SW0)",
    "h_mac_10_check (SW1)",
    "scream_handler (Mac Tx Comp)",
    "h_send_data_ready (SW2)",
    "h_recv_bd_ready (SW3)",
    "scream_handler (Mac Rx Start)",
    "h_mac_rx_comp (Mac Rx Comp)",
    "scream_handler (SW4)",
    "h_send_bd_ready (SW5)",
    "h_recv_bd_ready (SW6)",
    "scream_handler (SW7)",
    "scream_handler (SW8)",
    "h_dma_wr_assist_lo (DMA WR Asst Lo Comp)",
    "h_dma_rd_assist_lo (DMA RD Asst Lo Comp)",
    "h_dma_wr_assist_hi (DMA WR Asst Hi Comp)",
    "h_dma_rd_assist_hi (DMA RD Asst Hi Comp)",
    "scream_handler (SW9)",
    "scream_handler (DMA WR Comp)",
    "scream_handler (DMA RD Comp)",
    "h_dma_wr_attn (DMA WR Attn)",
    "h_dma_rd_attn (DMA RD Attn)",
    "scream_handler (SW10)",
    "mh_command (Mailbox)",
    "h_send_replicate (SW11)",
    "h_mac_10_negotiate (SW12)",
    "h_mac_rx_attn (Mac RX Attn)",
    "h_mac_tx_attn (Mac TX Attn)",
    "scream_handler (SW13)",
    "h_timer (TIMER)",
    "scream_handler (SW14)",
    "scream_handler (External Serial Data)",
    "scream_handler (SW15)"
};

#define ntohl(arg)				\
    (((arg & 0xff) << 24) |			\
     ((arg & 0xff00) << 8) |			\
     ((arg & 0xff0000) >> 8) |			\
      (arg >> 24))

#define ntohs(arg)				\
    (((arg & 0xff) << 8) |			\
      (arg >> 8))

#define htonl ntohl
#define htons ntohs

#define SWAP_DOUBLE(x) (x << 32 | x >> 32)

HANDLE hMAC;
TCHAR szErr[256];

struct tg_stats stats;
DRIVER_STATS dstats;

#define DEVICE_PREFIX   "\\\\.\\"

static int flagc = 0;	/* display configuration */
static int flagg = 0;	/* display general statistics */
static int flagn = 0;	/* display NIC statistics */
static int flagp = 0;	/* display NIC profiling data */
static int flagC = 0;	/* clear general statistics */
static int flaga = 0;	/* display all statistics */
static int flag5 = 0;	/* NT 5.0 interface */
static char textArr[80] = "1";

int __cdecl 
main(int argc, char **argv)
{
    U8 linkName[512];
    U8 deviceName[80];
    U8 pathName[80];
    U8 szMACFileName[80];
    bool bCreatedDevice;
    unsigned long err;
    int instance = 1;
    char *textInst = textArr;
    int type;
    static char *cmdname;
    int errs = 0;

    /* save off the command name */
    cmdname = argv[0];

    /*
     * Handle command line args 
     */
    argv++; argc--;
    while (argc > 0 && (argv[0][0] == '-' || argv[0][0] == '/')) {
        switch (argv[0][1]) {
	case '5': flag5++;                                      break;
	case 'c': flagc++;		                        break;
	case 'C': flagC++;		                        break;
	case 'g': flagg++;		                        break;
	case 'n': flagn++;		                        break;
	case 'a': flaga++;		                        break;
	case 'p': flagp++;		                        break;
	default: errs++;                                        break;
	}
        argv++; argc--;
    }
    if (errs || (argc != 0 && argc != 1)) {
	usage("usage");
	exit(1);
    }

    if (argc == 1) {
	/* get the instance number */
        textInst = argv[0];
	if (!flag5)
	    instance = atoi(argv[0]);
    }

    if (flaga)
	flagg = flagc = flagp = flagn = 1;

    if (!(flagc || flagg || flagp || flagn || flagC)) {
	usage("usage");
	exit(1);
    }
    if (flagC && (flagc || flagg || flagp || flagn)) {
	(void) printf("-C must stand alone\n");
	usage("usage");
	exit(1);
    }

    /*
     * Check to see if the DOS name for the MAC driver already exists.
     * Its not created automatically in version 3.1 but may be later.
     */
    if (flag5) {
        strcpy(deviceName, textInst);
    } else {
        strcpy(deviceName, "altndis");
	strcat(deviceName, textInst);
    }

    if (QueryDosDevice(deviceName, linkName, sizeof(linkName)) == 0) {
	if ((err = GetLastError()) == ERROR_FILE_NOT_FOUND) {
	    strcpy(pathName, "\\Device\\");
	    strcat(pathName, "altndis");
	    if (!flag5)
	        strcat(pathName, textInst);

	    /*
	     * It doesn't exist so create it.
	     */
	    if (!DefineDosDevice(DDD_RAW_TARGET_PATH, deviceName, pathName)) {
		printf("DefineDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
		return (1);
	    }
	    bCreatedDevice = TRUE;
	} else {
	    printf("QueryDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
	    return (1);
	}
    } else {
	bCreatedDevice = FALSE;
    }

    /*
     * Construct a device name to pass to CreateFile
     */
    strcpy(szMACFileName, DEVICE_PREFIX);
    strcat(szMACFileName, deviceName);
    hMAC = CreateFile(szMACFileName,
		      GENERIC_READ,
		      FILE_SHARE_READ | FILE_SHARE_WRITE,
		      NULL,
		      OPEN_EXISTING,
		      0,
		      INVALID_HANDLE_VALUE);

    if (hMAC != INVALID_HANDLE_VALUE) {
        /* check for the "clear the stats" flag */
        if (flagC) {
	    ClearNicStats();
	    return(0);
	}
	/* read the host stats and maybe the nic stats */
	if (!ReadDriverStats())
	    return(0);
	PrintAlt(&dstats);
	if (flagn || flagp) {
	    if (!ReadNicStats())
	        return(0);
	    if (flagn)
  	        PrintTGStats(&stats);
	    if (flagp)
	        PrintNicProfile(&stats);
	}
    } else {
        printf("Device %s doesn't exist\n", deviceName);
	usage("usage");
    }
    if (bCreatedDevice) {
	/*
	 * The MAC driver wasn't visible in the Win32 name space so we created
	 * a link.  Now we have to delete it.
	 */
	if (!DefineDosDevice(DDD_RAW_TARGET_PATH | DDD_REMOVE_DEFINITION | 
			     DDD_EXACT_MATCH_ON_REMOVE, deviceName, 
			     pathName)) {
	    printf("DefineDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
	    return (1);
	}
    }
    return (0);
}

static bool
ReadNicStats(void)
{
    NDIS_OID oidCode;
    U32 count;

    oidCode = OID_VENDOR_GET_NIC_STATS;
    if (DeviceIoControl(hMAC, IOCTL_NDIS_QUERY_GLOBAL_STATS, &oidCode,
			sizeof(oidCode), &stats, sizeof(stats),
			&count,	NULL)) {
        return(TRUE);
    } else {
	printf("ReadNicStats:[%s]\n", GetLastErrorText(szErr, 256));
	return(FALSE);
    }
}

static void
ClearNicStats(void)
{
    NDIS_OID oidCode;
    U32 count;

    oidCode = OID_VENDOR_CLEAR_NIC_STATS;
    if (DeviceIoControl(hMAC, IOCTL_NDIS_QUERY_GLOBAL_STATS, &oidCode,
			sizeof(oidCode), &stats, sizeof(stats),
			&count,	NULL)) {
	printf("nic stats cleared \n");
    } else {
	printf("ClearNicStats:[%s]\n", GetLastErrorText(szErr, 256));
    }
}

static bool
ReadDriverStats(void)
{
    NDIS_OID oidCode;
    U32 count;

    oidCode = OID_VENDOR_GET_DRIVER_STATS;
    if (DeviceIoControl(hMAC, IOCTL_NDIS_QUERY_GLOBAL_STATS, &oidCode,
			sizeof(oidCode), &dstats, sizeof(dstats),
			&count,	NULL)) {
	if (count == sizeof(dstats)) {
	    return(TRUE);
	} else {
	    printf("DeviceIoControl returned an invalid count = %d\n", 
		   count);
	}
    } else {
	printf("ReadDriverStats:[%s]\n", GetLastErrorText(szErr, 256));
    }
    return(FALSE);
}


/*
 *  FUNCTION: GetLastErrorText
 *
 *  PURPOSE: copies error message text to string
 *
 *  PARAMETERS:
 *    lpszBuf - destination buffer
 *    dwSize - size of buffer
 *
 *  RETURN VALUE:
 *    destination buffer
 *
 *  COMMENTS:
 */
LPTSTR 
GetLastErrorText(LPTSTR lpszBuf, U64 dwSize)
{
    U64 dwRet;
    LPTSTR lpszTemp = NULL;

    dwRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			  FORMAT_MESSAGE_FROM_SYSTEM | 
			  FORMAT_MESSAGE_ARGUMENT_ARRAY,
			  NULL,
			  GetLastError(),
			  LANG_NEUTRAL,
			  (LPTSTR) & lpszTemp,
			  0,
			  NULL);

    /* supplied buffer is not long enough */
    if (!dwRet || ((long)dwSize < (long)dwRet + 14))
	lpszBuf[0] = TEXT('\0');
    else {
	/* remove cr and newline character */
	lpszTemp[lstrlen(lpszTemp) - 2] = TEXT('\0');	

	sprintf(lpszBuf, TEXT("%s (0x%x)"), lpszTemp, GetLastError());
    }

    if (lpszTemp)
	LocalFree((HLOCAL) lpszTemp);

    return lpszBuf;
}


void
PrintTGStats(struct tg_stats *statp)
{
    U32 i;

    (void) printf("\nMAC statistics:\n");
    (void) printf("   dot3StatsAlignmentErrors: %ld\n", 
		  statp->dot3StatsAlignmentErrors);
    (void) printf("   dot3StatsFCSErrors: %ld\n", statp->dot3StatsFCSErrors);
    (void) printf("   dot3StatsSingleCollisionFrames: %ld\n",
		  statp->dot3StatsSingleCollisionFrames);
    (void) printf("   dot3StatsMultipleCollisionFrames: %ld\n",
		  statp->dot3StatsMultipleCollisionFrames);
    (void) printf("   dot3StatsSQETestErrors: %ld\n",
		  statp->dot3StatsSQETestErrors);
    (void) printf("   dot3StatsDeferredTransmissions: %ld\n",
		  statp->dot3StatsDeferredTransmissions);
    (void) printf("   dot3StatsLateCollisions: %ld\n",
		  statp->dot3StatsLateCollisions);
    (void) printf("   dot3StatsExcessiveCollisions: %ld\n",
		  statp->dot3StatsExcessiveCollisions);
    (void) printf("   dot3StatsInternalMacTransmitErrors: %ld\n",
		  statp->dot3StatsInternalMacTransmitErrors);
    (void) printf("   dot3StatsCarrierSenseErrors: %ld\n",
		  statp->dot3StatsCarrierSenseErrors);
    (void) printf("   dot3StatsFrameTooLongs: %ld\n",
		  statp->dot3StatsFrameTooLongs);
    (void) printf("   dot3StatsInternalMacReceiveErrors: %ld\n",
		  statp->dot3StatsInternalMacReceiveErrors);

    (void) printf("\nInterface statistics:\n");
    (void) printf("   ifIndex: %ld\n", statp->ifIndex);

#if ALT_BIG_ENDIAN
    (void) printf("   ifDescr: %s\n", statp->ifDescr);
#else /* ALT_BIG_ENDIAN */
    {
#define IF_DESCR_LEN	32
	U8 swapped_ifDescr[IF_DESCR_LEN];
	U8 *s, *d;

	d = (U8 *)&swapped_ifDescr;

	for (i=0; i < IF_DESCR_LEN; i+=4) {
	    (volatile U8 *)s = &statp->ifDescr[i+3];
	    *d++ = *s--;
	    *d++ = *s--;
	    *d++ = *s--;
	    *d++ = *s--;
	}
	(void) printf("   ifDescr: %s\n", swapped_ifDescr);
    }
#endif /* ALT_BIG_ENDIAN */

    (void) printf("   ifType: %ld\n", statp->ifType);
    (void) printf("   ifMtu: %ld\n", statp->ifMtu);
    (void) printf("   ifSpeed: %ld\n", statp->ifSpeed);
    (void) printf("   ifPhysAddress: ");

#if ALT_BIG_ENDIAN
    for (i=0; i < 5; i++) {
	(void) printf("%x:", statp->ifPhysAddress.octet[i]);
    }
    (void) printf("%x\n", statp->ifPhysAddress.octet[i]);
#else /* ALT_BIG_ENDIAN */
    (void) printf("%x:", statp->ifPhysAddress.octet1[1]);
    (void) printf("%x:", statp->ifPhysAddress.octet1[0]);
    (void) printf("%x:", statp->ifPhysAddress.octet2[3]);
    (void) printf("%x:", statp->ifPhysAddress.octet2[2]);
    (void) printf("%x:", statp->ifPhysAddress.octet2[1]);
    (void) printf("%x\n", statp->ifPhysAddress.octet2[0]);
#endif /* ALT_BIG_ENDIAN */

    (void) printf("   ifAdminStatus: %ld\n", statp->ifAdminStatus);
    (void) printf("   ifOperStatus: %ld\n", statp->ifOperStatus);
    (void) printf("   ifLastChange: %ld\n", statp->ifLastChange);

    (void) printf("   ifInDiscards: %ld\n", statp->ifInDiscards);
    (void) printf("   ifInErrors: %ld\n", statp->ifInErrors);
    (void) printf("   ifInUnknownProtos: %ld\n", statp->ifInUnknownProtos);

    (void) printf("   ifOutDiscards: %ld\n", statp->ifOutDiscards);
    (void) printf("   ifOutErrors: %ld\n", statp->ifOutErrors);
    (void) printf("   ifOutQLen: %ld\n", statp->ifOutQLen);
    (void) printf("   ifHCInOctets: %lld\n", 
		  SWAP_DOUBLE((U64)statp->ifHCInOctets));
    (void) printf("   ifHCInUcastPkts: %lld\n", 
		  SWAP_DOUBLE((U64)statp->ifHCInUcastPkts));
    (void) printf("   ifHCInMulticastPkts: %lld\n", 
		  SWAP_DOUBLE(statp->ifHCInMulticastPkts));
    (void) printf("   ifHCInBroadcastPkts: %lld\n", 
		  SWAP_DOUBLE(statp->ifHCInBroadcastPkts));
    (void) printf("   ifHCOutOctets: %lld\n", 
		  SWAP_DOUBLE((U64)statp->ifHCOutOctets));
    (void) printf("   ifHCOutUcastPkts: %lld\n", 
		  SWAP_DOUBLE((U64)statp->ifHCOutUcastPkts));
    (void) printf("   ifHCOutMulticastPkts: %lld\n", 
		  SWAP_DOUBLE(statp->ifHCOutMulticastPkts));
    (void) printf("   ifHCOutBroadcastPkts: %lld\n", 
		  SWAP_DOUBLE(statp->ifHCOutBroadcastPkts));
    (void) printf("   ifLinkUpDownTrapEnable: %ld\n", 
		  statp->ifLinkUpDownTrapEnable);
    (void) printf("   ifHighSpeed: %ld\n", statp->ifHighSpeed);
    (void) printf("   ifPromiscuousMode: %ld\n", statp-> ifPromiscuousMode); 
    (void) printf("   ifConnectorPresent: %ld\n", statp->ifConnectorPresent);

    (void) printf("\nInternal MAC RX statistics:\n");
    (void) printf("   nicMacRxLateColls: %ld\n", statp->nicMacRxLateColls);
    (void) printf("   nicMacRxLinkLostDuringPkt: %ld\n", 
		  statp->nicMacRxLinkLostDuringPkt);
    (void) printf("   nicMacRxPhyDecodeErr: %ld\n", 
		  statp->nicMacRxPhyDecodeErr);
    (void) printf("   nicMacRxMacAbort: %ld\n", statp->nicMacRxMacAbort);
    (void) printf("   nicMacRxTruncNoResources: %ld\n", 
		  statp->nicMacRxTruncNoResources);
    (void) printf("   nicMacRxDropUla: %ld\n", statp->nicMacRxDropUla);
    (void) printf("   nicMacRxDropMcast: %ld\n", statp->nicMacRxDropMcast);
    (void) printf("   nicMacRxFlowControl: %ld\n", statp->nicMacRxFlowControl);
    (void) printf("   nicMacRxDropSpace: %ld\n", statp->nicMacRxDropSpace);
    (void) printf("   nicMacRxColls: %ld\n", statp->nicMacRxColls);
    (void) printf("   nicMacRxTotalAttentions: %ld\n", 
		  statp->nicMacRxTotalAttns);
    (void) printf("   nicMacRxLinkAttentions: %ld\n", 
		  statp->nicMacRxLinkAttns);
    (void) printf("   nicMacRxSyncAttentions: %ld\n", 
		  statp->nicMacRxSyncAttns);
    (void) printf("   nicMacRxConfigAttentions: %ld\n", 
		  statp->nicMacRxConfigAttns);
    (void) printf("   nicMacReset: %ld\n", statp->nicMacReset);
    (void) printf("   nicMacRxBufferAttentions: %ld\n", 
		  statp->nicMacRxBufAttns);
    (void) printf("   nicMacRxZeroFrameCleanup: %ld\n", 
		  statp->nicMacRxZeroFrameCleanup);
    (void) printf("   nicMacRxOneFrameCleanup: %ld\n", 
		  statp->nicMacRxOneFrameCleanup);
    (void) printf("   nicMacRxMultipleFrameCleanup: %ld\n", 
		  statp->nicMacRxMultipleFrameCleanup);
    (void) printf("   nicMacRxTimerCleanup: %ld\n", 
		  statp->nicMacRxTimerCleanup);

    (void) printf("\nInternal MAC TX statistics:\n");
    (void) printf("   nicMacTxTotalAttentions: %ld\n", 
		  statp->nicMacTxTotalAttns);
    (void) printf("\n   MAC TX Collision Histogram :\n");
    for (i= 0; i <15; i++) {
	(void) printf("   nicMacTxCollision[%ld]: %ld\n",i+1,  
		      statp->nicMacTxCollisionHistogram[i]);
    }

    (void) printf("\nLocal statistics:\n");
    (void) printf("\nHost Commands:\n");
    (void) printf("   nicCmdsHostState: %ld\n", statp->nicCmdsHostState);
    (void) printf("   nicCmdsFDRFiltering: %ld\n", statp->nicCmdsFDRFiltering);
    (void) printf("   nicCmdsSetRecvProdIndex: %ld\n", 
		  statp->nicCmdsSetRecvProdIndex);
    (void) printf("   nicCmdsUpdateGencommStats: %ld\n", 
		  statp->nicCmdsUpdateGencommStats);
    (void) printf("   nicCmdsResetJumboRing: %ld\n", 
		  statp->nicCmdsResetJumboRing);
    (void) printf("   nicCmdsAddMCastAddr: %ld\n", statp->nicCmdsAddMCastAddr);
    (void) printf("   nicCmdsDelMCastAddr: %ld\n", statp->nicCmdsDelMCastAddr);
    (void) printf("   nicCmdsSetPromiscMode: %ld\n", 
		  statp->nicCmdsSetPromiscMode);
    (void) printf("   nicCmdsLinkNegotiate: %ld\n", 
		  statp->nicCmdsLinkNegotiate);
    (void) printf("   nicCmdsSetMACAddr: %ld\n", statp->nicCmdsSetMACAddr);
    (void) printf("   nicCmdsClearProfile: %ld\n", statp->nicCmdsClearProfile);
    (void) printf("   nicCmdsSetMulticastMode: %ld\n", 
		  statp->nicCmdsSetMulticastMode);
    (void) printf("   nicCmdsClearStats: %ld\n", statp->nicCmdsClearStats);
    (void) printf("   nicCmdsSetRecvJumboProdIndex: %ld\n", 
		  statp->nicCmdsSetRecvJumboProdIndex);
    (void) printf("   nicCmdsSetRecvMiniProdIndex: %ld\n", 
		  statp->nicCmdsSetRecvMiniProdIndex);
    (void) printf("   nicCmdsRefreshStats: %ld\n", statp->nicCmdsRefreshStats);
    (void) printf("   nicCmdsUnknown: %ld\n", statp->nicCmdsUnknown);

    (void) printf("\nNIC Events:\n");
    (void) printf("   nicEventsNICFirmwareOperational: %ld\n",
		  statp->nicEventsNICFirmwareOperational);
    (void) printf("   nicEventsStatsUpdated: %ld\n", 
		  statp->nicEventsStatsUpdated);
    (void) printf("   nicEventsLinkStateChanged: %ld\n", 
		  statp->nicEventsLinkStateChanged);
    (void) printf("   nicEventsError: %ld\n", statp->nicEventsError);
    (void) printf("   nicEventsMCastListUpdated: %ld\n", 
		  statp->nicEventsMCastListUpdated);
    (void) printf("   nicEventsResetJumboRing: %ld\n", 
		  statp->nicEventsResetJumboRing);
    (void) printf("   nicRingSetSendProdIndex: %ld\n", 
		  statp->nicRingSetSendProdIndex);
    (void) printf("   nicRingSetSendConsIndex: %ld\n", 
		  statp->nicRingSetSendConsIndex);
    (void) printf("   nicRingSetRecvReturnProdIndex: %ld\n", 
		  statp->nicRingSetRecvReturnProdIndex);
    (void) printf("   nicEventThresholdHit: %ld\n", 
		  statp->nicEventThresholdHit);
    (void) printf("   nicSendThresholdHit: %ld\n", statp->nicSendThresholdHit);
    (void) printf("   nicRecvThresholdHit: %ld\n", statp->nicRecvThresholdHit);

    (void) printf("\nDMA Attentions:\n");
    (void) printf("   nicDmaRdOverrun: %ld\n", statp->nicDmaRdOverrun);
    (void) printf("   nicDmaRdUnderrun: %ld\n", statp->nicDmaRdUnderrun);
    (void) printf("   nicDmaWrOverrun: %ld\n", statp->nicDmaWrOverrun);
    (void) printf("   nicDmaWrUnderrun: %ld\n", statp->nicDmaWrUnderrun);
    (void) printf("   nicDmaRdMasterAborts: %ld\n", 
		  statp->nicDmaRdMasterAborts);
    (void) printf("   nicDmaWrMasterAborts: %ld\n", 
		  statp->nicDmaWrMasterAborts);
    
    (void) printf("\nRing/Misc. Resources:\n");
    (void) printf("   nicDmaWriteRingFull: %ld\n", statp->nicDmaWriteRingFull);
    (void) printf("   nicDmaReadRingFull: %ld\n", statp->nicDmaReadRingFull);
    (void) printf("   nicEventRingFull: %ld\n", statp->nicEventRingFull);
    (void) printf("   nicEventProducerRingFull: %ld\n", 
		  statp->nicEventProducerRingFull);
    (void) printf("   nicTxMacDescrRingFull: %ld\n", 
		  statp->nicTxMacDescrRingFull);
    (void) printf("   nicTotalRecvBDs: %ld\n", statp->nicTotalRecvBDs);
    (void) printf("   nicJumboRecvBDs: %ld\n", statp->nicJumboRecvBDs);
    (void) printf("   nicMiniRecvBDs: %ld\n", statp->nicMiniRecvBDs);
    (void) printf("   nicStdRecvBDs: %ld\n", statp->nicRecvBDs);
    (void) printf("   nicNoMoreRxBDs: %ld\n", statp->nicNoMoreRxBDs);
    (void) printf("   nicJumboSpillOver: %ld\n", statp->nicJumboSpillOver);
    (void) printf("   nicSbusHangCleared: %ld\n", statp->nicSbusHangCleared);
    (void) printf("   nicNoMoreWrDMADescriptors: %ld\n", 
		  statp->nicNoMoreWrDMADescriptors);
    (void) printf("   nicOutOfTxBufSpaceFrameRetry: %ld\n", 
		  statp->nicOutOfTxBufSpaceFrameRetry);
    (void) printf("   nicNoSpaceInReturnRing: %ld\n", 
		  statp->nicNoSpaceInReturnRing);

    (void) printf("\nHost Interrupts:\n");
    (void) printf("   nicInterrupts: %ld\n", statp->nicInterrupts);
    (void) printf("   nicAvoidedInterrupts: %ld\n", 
		  statp->nicAvoidedInterrupts);

}

void
PrintNicProfile(struct tg_stats *statp)
{
    U32 ticks = 0;
    int i;

    /* add up the total counts for ease of use */
    for (i=0; i<32; i++)
	ticks += statp->nicProfile[i];

    (void) printf("\nEvent Profiling:\n");
    for (i=31; i >= 0; i--)
	(void) printf("   %s: %lu (%2.2f%%)\n", 
		      handler_names[i],
		      statp->nicProfile[i], 
		      (statp->nicProfile[i] * 100.)/ticks);

    (void) printf("\nTotal Ticks: %lu\n", ticks);
}

static void
usage(char *msg)
{
    (void) printf("%s : ", msg);
    (void) printf("%s [-c] [-g] [-n] [-p] [-C] [-a] [<instance>]\n", PROG);
    (void) printf("\t-c Display configuration\n");
    (void) printf("\t-g Display general statistics\n");
    (void) printf("\t-n Display NIC statistics\n");
    (void) printf("\t-p Display NIC profiling data\n");
    (void) printf("\t-C Clear general statistics\n");
    (void) printf("\t-a Display all statistics\n");

    (void) printf("\tinstance is the device instance number\n");
}

static void
PrintAlt(DRIVER_STATS *nicp)
{
    int i, j;

    if (flagc) {
	(void) printf("Statistics for Alteon PCI Gigabit Ethernet card, Rev %ld\n",
		      nicp->tigonRev >> 28);
	(void) printf("   card base is located at 0x%x\n",
		      (int)(nicp->base));
	(void) printf("   Nic flags are 0x%x\n", (int)nicp->flags);

	(void) printf("   General Communicates area is at %x\n", 
		  (int)nicp->genInfo);
	(void) printf("   General Info is at %x\n", (int)nicp->genInfo);
	(void) printf("   Event ring is at %x\n", (int)nicp->eventArr);
	(void) printf("   Event producer is at %x\n", (int)nicp->eventProdPtr);
	(void) printf("   Send ring is at %x\n", (int)nicp->sendRing);
	(void) printf("   Send buffer count is %ld\n", nicp->sendBufCount);
	(void) printf("   Send producer is %ld\n", nicp->send_producer);
	(void) printf("   Send consumer is %ld\n", nicp->send_consumer);
	(void) printf("   Send consumer pointer is at %x\n", 
		      (int)nicp->sendConsPtr);
	(void) printf("   Standard Receive ring is at %x\n", 
		      (int)nicp->recvStdRing);
	(void) printf("   Jumbo Receive ring is at %x\n", 
		      (int)nicp->recvJumboRing);
#if USE_MINI_RING 
	(void) printf("   Mini Receive ring is at %x\n", 
		      (int)nicp->recvMiniRing);
#endif /* USE_MINI_RING */
	(void) printf("   Receive Return ring is at %x\n", 
		      (int)nicp->recvReturnRing);
	(void) printf("   Standard receive buffer count is %ld\n", 
		      nicp->recvStdBufCount);
	(void) printf("   Standard receive producer is %ld\n", 
		      nicp->recv_std_producer);
	(void) printf("   Standard receive consumer is %ld\n", 
		      nicp->recv_std_consumer);
	(void) printf("   Jumbo receive buffer count is %ld\n", 
		      nicp->recvJumboBufCount);
	(void) printf("   Jumbo receive producer is %ld\n", 
		      nicp->recv_jumbo_producer);
	(void) printf("   Jumbo receive consumer is %ld\n", 
		      nicp->recv_jumbo_consumer);
#if USE_MINI_RING
	(void) printf("   Mini receive buffer count is %ld\n", 
		      nicp->recvMiniBufCount);
	(void) printf("   Mini receive producer is %ld\n", 
		      nicp->recv_mini_producer);
	(void) printf("   Mini receive consumer is %ld\n", 
		      nicp->recv_mini_consumer);
#endif /* USE_MINI_RING */
	(void) printf("   Return receive producer is %ld\n", 
		      nicp->recv_return_producer);
	(void) printf("   Return receive consumer is %ld\n", 
		      nicp->recv_return_consumer);
	(void) printf("   Return Receive producer pointer is at %x\n", 
		      (int)nicp->recvReturnProdPtr);
    }

    if (flagg) {
      (void) printf("   inpkts: %ld\n", nicp->inpkts);
      (void) printf("   bufsInStdRecvRing: %d\n", 
		    (int)nicp->bufsInStdRecvRing);
      (void) printf("   bufsInJumboRecvRing: %d\n", 
		    (int)nicp->bufsInJumboRecvRing);
#if USE_MINI_RING
      (void) printf("   bufsInMiniRecvRing: %d\n", 
		    (int)nicp->bufsInMiniRecvRing);
#endif /* USE_MINI_RING */
      (void) printf("   bufsOnSendRing: %d\n", (int)nicp->bufsOnSendRing);
      (void) printf("   bufsLowWater: %d\n", (int)nicp->bufsLowWater);
      (void) printf("   recvStdCoal: %d\n", (int)nicp->recvStdCoal);
      (void) printf("   recvJumboCoal: %d\n", (int)nicp->recvJumboCoal);
#if USE_MINI_RING
      (void) printf("   recvMiniCoal: %d\n", (int)nicp->recvMiniCoal);
#endif /* USE_MINI_RING */
      (void) printf("   outpkts: %ld\n", nicp->outpkts);
      (void) printf("   outDrops: %ld\n", nicp->outDrops);
      (void) printf("   outDropsDmaResources: %ld\n", 
		    nicp->outDropsDmaResources);
      (void) printf("   outDropsNoSpace: %ld\n", nicp->outDropsNoSpace);
      (void) printf("   cmdFull: %ld\n", nicp->cmdFull);
      (void) printf("   Flow Controls: %ld\n", nicp->flowControls);
      (void) printf("   Reaped from send: %ld\n", nicp->reapedFromSend);
      (void) printf("   Pending Pkts: %ld\n", nicp->pendPkts);
      (void) printf("   Requeued Pkts: %ld\n", nicp->requeuedPkts);
      (void) printf("   Data Return Pkts: %ld\n", nicp->dataRtnPkts);
      (void) printf("   Normal Indicate Pkts: %ld\n", nicp->normRecv);
      (void) printf("   Low Resources Pkts: %ld\n", nicp->lowResources); 
      (void) printf("   Bad IP Checksum Pkts: %ld\n", nicp->badIpCksums);
      (void) printf("   Bad TCP Checksum Pkts: %ld\n", nicp->badTcpCksums);
      (void) printf("   Bad UDP Checksum Pkts: %ld\n", nicp->badUdpCksums);
    }
}
