/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997, 1998 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: alttune.c,v $
 * Revision 1.1.2.1  1998/11/09  18:20:28  hayes
 * 	initial checkin
 * 	[1998/11/09  18:19:56  hayes]
 *
 * Revision 1.1.2.3  1998/07/14  16:50:48  ted
 * 	Support ndis 5 devices and -5 option, remove dead code, improve boolean handling
 * 	[1998/07/14  02:20:04  ted]
 * 
 * Revision 1.1.2.2  1998/06/12  23:47:53  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:46  ted]
 * 
 * Revision 1.1.2.1  1998/06/08  23:56:00  ted
 * 	Port Unix trace buffer display utility
 * 	[1998/06/04  19:50:59  ted]
 * 
 * $EndLog$
 */

#ifndef U32
#define U64 unsigned __int64
#define U32 unsigned long
#define U16 unsigned short
#define U8  unsigned char
#define S32 signed long
#define S16 signed short
#define S8  signed char
#define bool unsigned char
#endif /* U32 */

#define VOLATILE volatile

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <winioctl.h>
#include <ntddndis.h>

#include "drvstat.h"

#include "alt.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "pub_tg.h"

/* forward declarations... */
static void do_tuneparms(int instance);
void update_parm(char *parm_title, U32 *parm_addr);
static void getline(char *s);
static LPTSTR GetLastErrorText(LPTSTR, U64);
static void usage(char *);

#undef printf

#define ntohl(arg)				\
    (((arg & 0xff) << 24) |			\
     ((arg & 0xff00) << 8) |			\
     ((arg & 0xff0000) >> 8) |			\
      (arg >> 24))

#define ntohs(arg)				\
    (((arg & 0xff) << 8) |			\
      (arg >> 8))

#define htonl ntohl
#define htons ntohs

#define SWAP_DOUBLE(x) (x << 32 | x >> 32)

HANDLE hMAC;
TCHAR szErr[256];

#define DEVICE_PREFIX   "\\\\.\\"

static char instArr[80] = "1";

int __cdecl 
main(int argc, char **argv)
{
    U8 linkName[512];
    U8 deviceName[80];
    U8 pathName[80];
    U8 szMACFileName[80];
    bool bCreatedDevice;
    unsigned long err;
    int instance = 1;
    char *textInst = instArr;
    bool flag5 = FALSE;
    int errs = 0;
    static char *cmdname;

    /* save off the command name */
    cmdname = argv[0];

    /*
     * Handle command line args 
     */
    argv++; argc--;
    while (argc > 0 && (argv[0][0] == '-' || argv[0][0] == '/')) {
        switch (argv[0][1]) {
	case '5':
	    flag5 = TRUE;
	    break;

	default:
	    errs++;
	    break;
	}
        argv++; argc--;
    }

    if (argc == 1) {
	/* get the instance number/string */
        textInst = argv[0];
	if (!flag5)
	  instance = atoi(argv[0]);
    }

    /*
     * Check to see if the DOS name for the MAC driver already exists.
     * Its not created automatically in version 3.1 but may be later.
     */

    if (flag5) {
        strcpy(deviceName, textInst);
    } else {
        strcpy(deviceName, "altndis");
	strcat(deviceName, textInst);
    }

    if (QueryDosDevice(deviceName, linkName, sizeof(linkName)) == 0) {
	if ((err = GetLastError()) == ERROR_FILE_NOT_FOUND) {
	    strcpy(pathName, "\\Device\\");
	    strcat(pathName, "altndis");
	    if (!flag5)
	        strcat(pathName, textInst);

	    /*
	     * It doesn't exist so create it.
	     */
	    if (!DefineDosDevice(DDD_RAW_TARGET_PATH, deviceName, pathName)) {
		printf("DefineDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
		return (1);
	    }
	    bCreatedDevice = TRUE;
	} else {
	    printf("QueryDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
	    return (1);
	}
    } else {
	bCreatedDevice = FALSE;
    }

    /*
     * Construct a device name to pass to CreateFile
     */
    strcpy(szMACFileName, DEVICE_PREFIX);
    strcat(szMACFileName, deviceName);
    hMAC = CreateFile(szMACFileName,
		      GENERIC_READ,
		      FILE_SHARE_READ | FILE_SHARE_WRITE,
		      NULL,
		      OPEN_EXISTING,
		      0,
		      INVALID_HANDLE_VALUE);

    if (hMAC != INVALID_HANDLE_VALUE) {
	do_tuneparms(instance);
    } else {
        printf("Device %s doesn't exist\n", deviceName);
	usage(cmdname);
    }
    if (bCreatedDevice) {
	/*
	 * The MAC driver wasn't visible in the Win32 name space so we created
	 * a link.  Now we have to delete it.
	 */
	if (!DefineDosDevice(DDD_RAW_TARGET_PATH | DDD_REMOVE_DEFINITION | 
			     DDD_EXACT_MATCH_ON_REMOVE, deviceName, 
			     pathName)) {
	    printf("DefineDosDevice:[%s]\n", GetLastErrorText(szErr, 256));
	    return (1);
	}
    }
    return (0);
}

static void 
do_tuneparms(int instance)
{
    NDIS_OID oidCode;
    DWORD count;
    struct tg_tune tuneParms;

    oidCode = OID_VENDOR_GET_TUNEPARMS;
    if (DeviceIoControl(hMAC, IOCTL_NDIS_QUERY_GLOBAL_STATS,
			&oidCode, sizeof(oidCode), &tuneParms,
			sizeof(struct tg_tune), &count, NULL)) {
	if (count == sizeof(struct tg_tune)) {
	    
	    update_parm("recv_coal_ticks", (U32 *)&tuneParms.recv_coal_ticks);
	    update_parm("send_coal_ticks", (U32 *)&tuneParms.send_coal_ticks);
	    update_parm("stat_ticks", (U32 *)&tuneParms.stat_ticks);
	    update_parm("send_max_coalesced_bds", (U32 *)&tuneParms.send_max_coalesced_bds);
	    update_parm("recv_max_coalesced_bds", (U32 *)&tuneParms.recv_max_coalesced_bds);
	    update_parm("nic_tracing", (U32 *)&tuneParms.tracing);
	    update_parm("link_negotiation", (U32 *)&tuneParms.link_negotiation);
	    update_parm("fast_link_negotiation", (U32 *)&tuneParms.fast_link_negotiation);

	    oidCode = OID_VENDOR_SET_TUNEPARMS;
	    if (DeviceIoControl(hMAC, IOCTL_NDIS_QUERY_GLOBAL_STATS,
				&oidCode, sizeof(oidCode), &tuneParms,
				sizeof(struct tg_tune), &count, NULL)) {
		if (count != sizeof(struct tg_tune)) {
		    printf("DeviceIoControl returned an invalid count = %d: SET_TUNEPARMS failed\n", 
			   count);
		}
	    }
	} else {
	    printf("DeviceIoControl returned an invalid count = %d: GET_TUNEPARMS failed\n", 
		   count);
	}
    } else {
	printf("do_tuneparms [%s]\n", GetLastErrorText(szErr, 256));
    }
}

void
update_parm(char *parm_title, U32 *parm_addr) 
{
    char string[80];
    U32 val;

    printf("%s value is: %i (0x%08x). new value: ", 
	parm_title, *parm_addr, *parm_addr);
    getline(string);
    if (string[0] != '\n') {
	val = strtoul(string, (char **)NULL, 0);
	*parm_addr = val;
	printf ("---> new value: %i (0x%08x).\n", *parm_addr, *parm_addr);
    }
}

static void
getline(char *s)
{
    while((*s++ = getchar()) != '\n');
    *s = '\0';
}

/*
 *  FUNCTION: GetLastErrorText
 *
 *  PURPOSE: copies error message text to string
 *
 *  PARAMETERS:
 *    lpszBuf - destination buffer
 *    dwSize - size of buffer
 *
 *  RETURN VALUE:
 *    destination buffer
 *
 *  COMMENTS:
 */
static LPTSTR 
GetLastErrorText(LPTSTR lpszBuf, U64 dwSize)
{
    U64 dwRet;
    LPTSTR lpszTemp = NULL;

    dwRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			  FORMAT_MESSAGE_FROM_SYSTEM | 
			  FORMAT_MESSAGE_ARGUMENT_ARRAY,
			  NULL,
			  GetLastError(),
			  LANG_NEUTRAL,
			  (LPTSTR) & lpszTemp,
			  0,
			  NULL);

    /* supplied buffer is not long enough */
    if (!dwRet || ((long)dwSize < (long)dwRet + 14))
	lpszBuf[0] = TEXT('\0');
    else {
	/* remove cr and newline character */
	lpszTemp[lstrlen(lpszTemp) - 2] = TEXT('\0');	

	sprintf(lpszBuf, TEXT("%s (0x%x)"), lpszTemp, GetLastError());
    }

    if (lpszTemp)
	LocalFree((HLOCAL) lpszTemp);

    return lpszBuf;
}

static void
usage(char *str)
{
    (void) printf("%s\n", str);
    (void) printf("usage : alttune [-5] [<instance>]\n");
    exit(1);
}

