/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997, 1998 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: netcard.cpp $
 * Revision api_leonid/2 1999/03/27 01:45:20 leonid
 * 	back to nt1_10
 * 
 * Revision 1.1.4.2  1998/09/02  22:51:02  ted
 * 	Move forward 1.9 dll fixes to ge_latest
 * 	[1998/09/02  22:50:20  ted]
 *
 * Revision 1.1.2.2  1998/09/02  22:43:28  ted
 * 	Allow updating driver by building registry names if not there
 * 	[1998/09/02  22:36:02  ted]
 * 
 * Revision 1.1.2.1  1998/08/22  21:34:14  ted
 * 	Add jumbo, rx/tx flow control to dlls and put under source control
 * 	[1998/08/22  21:24:42  ted]
 * 
 * $EndLog$
 */
#include "stdafx.h"
#include "netcard.h"

extern char *lpSoftware;

RegManger::RegManger()
{
    m_pCards = NULL;
    m_hkeyNetworkCard = NULL;
    m_NumCards = 0;
}

BOOL RegManger::Init()
{
    LONG ret;
    char buf[256];
    DWORD dwSubKeys;
    char regname[256];
    LONG size;
    int i;
    HKEY h;
    DWORD dwType;

    ret = ::RegOpenKey(HKEY_LOCAL_MACHINE, NETWORKCARD, &m_hkeyNetworkCard);
    if (ret != ERROR_SUCCESS) 
	return FALSE;
    ret = ::RegQueryInfoKey(m_hkeyNetworkCard, NULL, NULL, NULL, &dwSubKeys,
			    NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (ret != ERROR_SUCCESS) 
	return FALSE;
    m_NumCards = (int)dwSubKeys;
    m_pCards = new NETCARD[m_NumCards];

    i = 0;
    for (i = 0; i < (int)dwSubKeys; i++) {
	ret = ::RegEnumKey(m_hkeyNetworkCard, i, buf, 256);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	ret = ::RegOpenKey(m_hkeyNetworkCard, buf, &h);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	size  = 256;
	ret = ::RegQueryValueEx(h, (LPSTR)"Description", NULL, &dwType, 
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	buf[size] = 0;
	strcpy(regname, lpSoftware);
	if (strcmp(buf, regname)) {
	    ret = ::RegCloseKey(h);
	    if (ret != ERROR_SUCCESS) 
		return FALSE;
	    m_NumCards--;
	    continue;
	}
	ret = ::RegQueryValueEx(h, (LPSTR)"ServiceName", NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	buf[size] = 0;
	ret = ::RegCloseKey(h);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	strcpy(m_pCards[i + m_NumCards - dwSubKeys].m_ServiceName, buf);
    }
    for (i = 0; i < m_NumCards; i++) {
	//Adapter type
	sprintf(regname, "%s\\%s\\Parameters", SERVICES, 
		m_pCards[i].m_ServiceName);
	// Now get the service HKEY, Adaptertype
	ret = ::RegOpenKey(HKEY_LOCAL_MACHINE, regname, &h);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	m_pCards[i].m_HServiceKey = h;

	// m_BusNumber
	size = 256;
	ret = ::RegQueryValueEx(h, BUSNUMBER, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_BusNumber, buf, 4);
	// m_Instance
	size = 256;
	ret = ::RegQueryValueEx(h, INSTANCE, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_Instance, buf, 4);
	// m_link
	size = 256;
	ret = ::RegQueryValueEx(h, LINK, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	// on upgrade we have to create it
	if (ret == ERROR_FILE_NOT_FOUND) {
	    dwType = 1;		// default is link negotiation enabled
	    memcpy(buf, &dwType, 4); // skip requerying, why do it?
	    ret = ::RegSetValueEx(h, LINK, NULL, REG_DWORD, 
				  (unsigned char *)&dwType, 4);
	    if (ret != ERROR_SUCCESS) {
		return(FALSE);
	    }
	} else if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_link, buf, 4);
	// m_jumbos
	size = 256;
	ret = ::RegQueryValueEx(h, JUMBOS, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	// on upgrade we have to create it
	if (ret == ERROR_FILE_NOT_FOUND) {
	    dwType = 0;		// default is jumbos disabled
	    memcpy(buf, &dwType, 4); // skip requerying, why do it?
	    ret = ::RegSetValueEx(h, JUMBOS, NULL, REG_DWORD,
				  (unsigned char *)&dwType, 4);
	    if (ret != ERROR_SUCCESS) {
		return(FALSE);
	    }
	} else if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_jumbo, buf, 4);
	// m_rx_flow
	size = 256;
	ret = ::RegQueryValueEx(h, RX_FLOW, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	// on upgrade we have to create it
	if (ret == ERROR_FILE_NOT_FOUND) {
	    dwType = 1;		// default is rx flow control enabled
	    memcpy(buf, &dwType, 4); // skip requerying, why do it?
	    ret = ::RegSetValueEx(h, RX_FLOW, NULL, REG_DWORD,
				  (unsigned char *)&dwType, 4);
	    if (ret != ERROR_SUCCESS) {
		return(FALSE);
	    }
	} else if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_rx_flow, buf, 4);
	// m_tx_flow
	size = 256;
	ret = ::RegQueryValueEx(h, TX_FLOW, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	// on upgrade we have to create it
	if (ret == ERROR_FILE_NOT_FOUND) {
	    dwType = 0;		// default is tx flow control disabled
	    memcpy(buf, &dwType, 4); // skip requerying, why do it?
	    ret = ::RegSetValueEx(h, TX_FLOW, NULL, REG_DWORD,
				  (unsigned char *)&dwType, 4);
	    if (ret != ERROR_SUCCESS) {
		return(FALSE);
	    }
	} else if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_tx_flow, buf, 4);
	// m_SlotNumber
	size = 256;
	ret = ::RegQueryValueEx(h, SLOTNUMBER, NULL, &dwType,
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	memcpy(&m_pCards[i].m_SlotNumber, buf, 4);

    }

    return TRUE;
}

BOOL RegManger::WriteBack()
{
    int i;
    DWORD ret;
	
    for (i = 0; i < m_NumCards; i++) {
	HKEY h = m_pCards[i].m_HServiceKey;

	// m_BusNumber
	ret = ::RegSetValueEx(h, BUSNUMBER, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_BusNumber,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_Instance
	ret = ::RegSetValueEx(h, INSTANCE, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_Instance,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_link
	ret = ::RegSetValueEx(h, LINK, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_link,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_jumbos
	ret = ::RegSetValueEx(h, JUMBOS, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_jumbo,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_rx_flow
	ret = ::RegSetValueEx(h, RX_FLOW, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_rx_flow,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_tx_flow
	ret = ::RegSetValueEx(h, TX_FLOW, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_tx_flow,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	// m_SlotNumber
	ret = ::RegSetValueEx(h, SLOTNUMBER, NULL, REG_DWORD,
			      (unsigned char *)&m_pCards[i].m_SlotNumber,
			      sizeof(DWORD));
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	RegCloseKey(h);
    }
    return TRUE;
}

BOOL RegManger::WriteDisplayName()
{
    LONG ret;
    char buf[256];
    DWORD dwSubKeys;
    LONG size;
    int i, j;
    HKEY h, hkeyNetworkCard;
    DWORD dwType;
	
    ret = ::RegOpenKey(HKEY_LOCAL_MACHINE, NETWORKCARD, &hkeyNetworkCard);
    if (ret != ERROR_SUCCESS) 
	return FALSE;
    ret = ::RegQueryInfoKey(hkeyNetworkCard, NULL, NULL, NULL, &dwSubKeys,
			    NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (ret != ERROR_SUCCESS) 
	return FALSE;

    i = 0;
    for (i = 0; i < (int)dwSubKeys; i++) {
	ret = ::RegEnumKey(hkeyNetworkCard, i, buf, 256);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	ret = ::RegOpenKey(hkeyNetworkCard, buf, &h);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	size  = 256;
	ret = ::RegQueryValueEx(h, (LPSTR)"ServiceName", NULL, &dwType, 
				(unsigned char *)buf, (unsigned long *)&size);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
	buf[size]=0;
	for (j = 0; j < m_NumCards; j++) {
	    if (!strcmp(buf, m_pCards[j].m_ServiceName))
		break;
	}
	if (j < m_NumCards) {   // so I found a suitable card to do rename
	    sprintf(buf, TITLENAME, m_pCards[j].m_Instance, STANDALONE);
	    ret = ::RegSetValueEx(h, TITLE, NULL, REG_SZ, 
				  (unsigned char *)buf, strlen(buf));
	}
	ret = ::RegCloseKey(h);
	if (ret != ERROR_SUCCESS) 
	    return FALSE;
    }
    ::RegCloseKey(hkeyNetworkCard);
    return TRUE;
}

BOOL RegManger::Quit()
{
    LONG ret;
    int i;
	
    for (i = 0; i < m_NumCards; i++) 
	::RegCloseKey(m_pCards[i].m_HServiceKey);
    delete m_pCards;
    ret = ::RegCloseKey(m_hkeyNetworkCard);
    if (ret == ERROR_SUCCESS) 
	return TRUE;
    return FALSE;
}

