/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996, 1997 
 * All rights reserved
 * 
 */
/*
 * HISTORY
 * $Log: alt_nd.c,v $
 * Revision 1.1.2.3  1997/09/11  18:06:33  ted
 * 	Mixing vlans, jumbos, and redundancy works, plus small fixes from release 1.1 and better NIC tracing
 * 	[1997/09/11  18:01:21  ted]
 *
 * Revision 1.1.2.2  1997/07/27  22:03:53  ted
 * 	Fix off by one bug in ndd ? support
 * 	[1997/07/27  22:01:14  ted]
 * 
 * Revision 1.1.2.1  1997/06/17  00:36:49  ted
 * 	Cleanse ndd code with new, Alteon developed code
 * 	[1997/06/13  02:28:36  ted]
 * 
 * $EndLog$
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/streams/SunOS/alt_nd.c,v 1.1.2.3 1997/09/11 18:06:33 ted Exp $";
#endif /* __lint */

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/kmem.h>
#include <sys/errno.h>
#include <stdio.h>

#include <sys/ddi.h>

/* would normally include <inet/common.h> but there are conflicting macros */
typedef	int (*pfi_t)();
/* would normally include <inet/led.h>, but there are conflicting macros */
typedef	struct msgb *MBLKP;

#include <inet/nd.h>
#include <sys/stream.h>
#include <sys/ethernet.h>

#include "alt.h"
#include "alttypes.h"

#include "alt_nd.h"

/* forward declarations */
static int nd_get_names(queue_t *q, MBLKP mp, caddr_t nd_param);
static int nd_get_default(queue_t *q, MBLKP mp, caddr_t data);
static int nd_set_default(queue_t *q, MBLKP mp, char *value, caddr_t);

boolean_t
nd_load(caddr_t *nd_pparam, char *name, pfi_t get_pfi, 
	pfi_t set_pfi, caddr_t data)
{
    alt_ndpp_t *ndp;

    if (!*nd_pparam) {
	/* allocate a new nd_pparam structure and add default entry '?' */
	MEMALLOC(*nd_pparam, caddr_t, sizeof(alt_ndpp_t), 0);
	if (!nd_pparam) {
	    return(FALSE);
	} else {
	    /* add default entry */
	    ndp = (alt_ndpp_t *)*nd_pparam;
	    strncpy(ndp->name, "?", MAX_NDPP_NAMELEN);
	    ndp->get = nd_get_names;
	    ndp->set = nd_set_default;
	    ndp->data = (caddr_t)ndp;
	    ndp->next = 0;
	}
    }
    /* walk chain and allocate new entry */
    ndp = (alt_ndpp_t *)*nd_pparam;
    while (ndp->next)
	ndp = ndp->next;

    MEMALLOC(ndp->next, alt_ndpp_t *, sizeof(alt_ndpp_t), 0);
    if (!ndp->next) {
	return(FALSE);
    } else {
	/* add new entry */
	ndp = ndp->next;
	strncpy(ndp->name, name, MAX_NDPP_NAMELEN);
	ndp->get = get_pfi;
	ndp->set = set_pfi;
	ndp->data = data;
	ndp->next = 0;
    }
    return(TRUE);
}

void
nd_free(caddr_t *nd_pparam)
{
    alt_ndpp_t *ndp;
    alt_ndpp_t *f_ndp;

    if (!nd_pparam)
	return;
    
    ndp = (alt_ndpp_t *)*nd_pparam;
    while (ndp->next) {
	f_ndp = ndp;
	ndp = ndp->next;
	MEMFREE(f_ndp, sizeof(alt_ndpp_t), 0);
    }
    nd_pparam = 0;
}

int
nd_getset(queue_t *q, caddr_t nd_param, MBLKP mp)
{
    struct iocblk *ip;
    alt_ndpp_t *ndp;
    char *name;

    if (!nd_param || !mp) {
	return(FALSE);
    }

    ndp = (alt_ndpp_t *)nd_param;
    name = (char *)mp->b_cont->b_rptr;
    
    /* free any remaining message blocks */
    if (mp->b_cont->b_cont) {
	freemsg(mp->b_cont->b_cont);
	mp->b_cont->b_cont = 0;
    }

    /* get to proper ndp entry */
    do {
	if (strncmp(name, ndp->name, MAX_NDPP_NAMELEN) == 0)
	    break;
	ndp = ndp->next;
    } while (ndp);
    if (!ndp) {
	return(FALSE);
    }

    ip = (struct iocblk *)mp->b_rptr;
    switch(ip->ioc_cmd) {
	caddr_t value;
	case ND_GET:
	    /* prep b_wptr */
	    mp->b_cont->b_wptr = mp->b_cont->b_rptr;
	    ip->ioc_error = (ndp->get)(q, mp->b_cont, ndp->data);
	    mp->b_datap->db_type = M_IOCACK;
	    *mp->b_cont->b_wptr = 0; /* null terminate */
	    mp->b_cont->b_wptr++;
	    ip->ioc_count = mp->b_cont->b_wptr - mp->b_cont->b_rptr;
	    break;

	case ND_SET:
	    /* value is immediately after the null terminated name */
	    value = (char *)mp->b_cont->b_rptr + 
		strlen((char *)mp->b_cont->b_rptr) + 1;
	    ip->ioc_error = (ndp->set)(q, mp->b_cont, value, 
				       ndp->data);
	    mp->b_datap->db_type = M_IOCACK;
	    ip->ioc_count = mp->b_cont->b_wptr - mp->b_cont->b_rptr;
	    break;

	default:
	    ip->ioc_error = EINVAL;
	    mp->b_datap->db_type = M_IOCNAK;
	    ip->ioc_count = 0;
	    return(FALSE);
    }
    return(TRUE);
}

static int
nd_get_names(queue_t *q, MBLKP mp, caddr_t data)
{
    alt_ndpp_t *ndp;
    int i;
#define NAME_ALIGN_WIDTH	25

    if (!data || !mp)
	return(EINVAL);

    ndp = (alt_ndpp_t *)data;
    mp->b_wptr = mp->b_rptr;

    do {
	sprintf((char *)mp->b_wptr, "%s ", ndp->name);
	i = NAME_ALIGN_WIDTH - strlen((char *)mp->b_wptr);
	mp->b_wptr += strlen((char *)mp->b_wptr);
	while (i-- > 0) {
	    *mp->b_wptr = ' ';
	    mp->b_wptr++;
	}

	if (ndp->get == nd_get_default) {
	    if (ndp->set == nd_set_default) { /* no get, no set */
		sprintf((char *)mp->b_wptr, "(no read or write)");
	    } else {                          /* no get, set ok */
		sprintf((char *)mp->b_wptr, "(write only)");
	    }
	} else {
	    if (ndp->set == nd_set_default) { /* get ok, no set */
		sprintf((char *)mp->b_wptr, "(read only)");
	    } else {                          /* get ok, set ok */
		sprintf((char *)mp->b_wptr, "(read and write)");
	    }
	}
	mp->b_wptr += strlen((char *)mp->b_wptr) + 1;
	ndp = ndp->next;
    } while (ndp);

    return(0);
}

static int
nd_get_default(queue_t *q, MBLKP mp, caddr_t data)
{
    return(EACCES);
}

static int
nd_set_default(queue_t *q, MBLKP mp, char *value, caddr_t data)
{
    return(EACCES);
}

