/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altinit.c,v $
 * Revision 1.4.7.54  1999/01/30  02:32:24  fhwang
 * 	add llc support, the routine alt_param_snap_set is added.
 * 	[1999/01/30  00:44:38  fhwang]
 *
 * Revision 1.4.7.53  1998/11/09  19:33:51  ted
 * 	support jumbos on x86 Solaris (test extended recv BD)
 * 	[1998/11/09  19:33:22  ted]
 * 
 * Revision 1.4.7.52  1998/10/01  20:27:55  ted
 * 	Add new NIC stats into Solaris and ODI local stats
 * 	[1998/10/01  20:27:30  ted]
 * 
 * Revision 1.4.7.51  1998/09/29  21:57:23  ted
 * 	Add support for a non-shared interrupt
 * 	[1998/09/29  21:35:54  ted]
 * 
 * Revision 1.4.7.50  1998/09/02  22:45:31  ted
 * 	remove jumbo frame support for x86 Solaris
 * 	[1998/09/02  22:45:14  ted]
 * 
 * Revision 1.4.7.49  1998/07/14  16:50:49  ted
 * 	Fix missing shutdown cleanup in error cases
 * 	[1998/07/14  02:21:00  ted]
 * 
 * Revision 1.4.7.48  1998/06/08  23:56:35  ted
 * 	Cleanup tracing messages and identifiers
 * 	[1998/06/08  23:39:27  ted]
 * 
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:40  ted]
 * 
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:09:00  ted]
 * 
 * Revision 1.4.7.47  1998/05/08  21:04:40  ted
 * 	Fix suspend bug
 * 	[1998/05/08  20:25:37  ted]
 * 
 * Revision 1.4.7.46  1998/04/14  21:42:48  ted
 * 	Add 12.0 stats support and rework interrupt handler for better performance on PCI boards
 * 	[1998/04/14  21:37:29  ted]
 * 
 * Revision 1.4.7.45  1998/03/05  03:03:19  suresh
 * 	Fix for PR 928 (host panics with jumbo configs)
 * 	- Also an unrelated but reqd correction in altstream.c to make
 * 	  sure that we set the max_sdu to the tagged header in the VLAN case
 * 	  and the untagged in the non-vlan case. This was reversed..
 * 	[1998/03/05  03:02:54  suresh]
 * 
 * Revision 1.4.7.44  1998/02/23  00:59:45  ted
 * 	Add new reset jumbo rings stats
 * 	[1998/02/23  00:58:49  ted]
 * 
 * Revision 1.4.7.43  1998/01/30  06:45:24  hayes
 * 	add forgotten ~
 * 	[1998/01/30  06:45:10  hayes]
 * 
 * Revision 1.4.7.42  1998/01/30  03:40:28  ted
 * 	New ndd functions for flow control and link negotiation
 * 	[1998/01/30  03:38:12  ted]
 * 
 * Revision 1.4.7.41  1997/12/23  00:00:22  ted
 * 	Add check for NIC busy to detach routine, remove 'alt' references in error messages
 * 	[1997/12/22  23:59:18  ted]
 * 
 * Revision 1.4.7.40  1997/12/03  01:23:14  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:40  ted]
 * 
 * Revision 1.4.7.39  1997/11/30  23:01:08  ted
 * 	Make Suspend/Resume work
 * 	[1997/11/30  23:01:01  ted]
 * 
 * Revision 1.4.7.38  1997/11/27  02:53:19  ted
 * 	Get Suspend/Resume to not crash
 * 	[1997/11/27  02:52:51  ted]
 * 
 * Revision 1.4.7.37  1997/11/25  22:07:33  ted
 * 	Fix up statistics
 * 	[1997/11/25  22:02:25  ted]
 * 
 * Revision 1.4.7.36  1997/11/20  04:56:50  ted
 * 	Get the iblock_cookie with ddi_get_iblock_cookie instead of with ddi_add_intr so we have the cookie when we initialize our locks in alt_InitNICmem
 * 	[1997/11/20  04:54:34  ted]
 * 
 * Revision 1.4.7.35  1997/11/19  00:50:11  ted
 * 	Pick up latest fixes from release 2.0
 * 	[1997/11/19  00:48:53  ted]
 * 
 * Revision 1.4.7.34  1997/11/14  19:17:22  hayes
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:14:46  hayes]
 * 
 * Revision 1.4.7.33  1997/10/21  00:13:29  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  23:01:15  ted]
 * 
 * Revision 1.4.23.1  1997/10/16  17:32:08  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:28:37  ted]
 * 
 * Revision 1.4.7.32  1997/09/16  21:47:08  ted
 * 	Change from MIB-II V1 (RFC 1213) to MIB-II V2 (RFC 1573)
 * 	[1997/09/16  21:43:16  ted]
 * 
 * Revision 1.4.7.31  1997/09/16  03:12:02  ted
 * 	Per vlan statistics and feature bug fixes, change s to spl for better coding style
 * 	[1997/09/16  03:07:13  ted]
 * 
 * Revision 1.4.7.30  1997/09/11  18:06:35  ted
 * 	Mixing vlans, jumbos, and redundancy works, plus small fixes from release 1.1 and better NIC tracing
 * 	[1997/09/11  18:01:25  ted]
 * 
 * Revision 1.4.7.29  1997/08/23  18:18:08  ted
 * 	Merged with changes from 1.4.7.28
 * 	[1997/08/23  18:17:48  ted]
 * 
 * 	Allocate nic structure memory separately from other memory
 * 	[1997/08/23  18:13:34  ted]
 * 
 * Revision 1.4.7.28  1997/08/23  17:40:35  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:39:00  skapur]
 * 
 * Revision 1.4.7.27  1997/08/01  21:07:33  ted
 * 	Change ALT_OP1 to OEM_OP1 so we get the correct release identifier
 * 	[1997/08/01  21:05:43  ted]
 * 
 * Revision 1.4.7.26  1997/07/29  03:41:16  ted
 * 	Add OP1 LED support
 * 	[1997/07/29  03:40:26  ted]
 * 
 * Revision 1.4.7.25  1997/07/28  22:40:40  ted
 * 	Merge
 * 	[1997/07/28  22:38:27  ted]
 * 
 * 	Solaris dual homing support
 * 	[1997/07/28  21:40:30  ted]
 * 
 * Revision 1.4.7.24  1997/07/27  22:03:55  ted
 * 	Remove extraneous printf
 * 	[1997/07/27  22:01:30  ted]
 * 
 * Revision 1.4.7.23  1997/07/25  22:59:50  ted
 * 	Merged with changes from 1.4.7.22
 * 	[1997/07/25  22:58:09  ted]
 * 
 * 	Merge fixes from ge_0_22 and add link negotiation support
 * 	[1997/07/25  22:33:30  ted]
 * 
 * 	Support flow control for STREAMS
 * 	[1997/07/19  20:37:42  ted]
 * 
 * 	Better error messages for PR 317
 * 	[1997/07/07  23:00:16  ted]
 * 
 * Revision 1.4.7.22  1997/07/05  21:17:49  ted
 * 	Fix naming issues as reported in bug 317
 * 	[1997/07/05  21:15:27  ted]
 * 
 * Revision 1.4.7.22  1997/07/05  21:17:49  ted
 * 	Fix naming issues as reported in bug 317
 * 	[1997/07/05  21:15:27  ted]
 * 
 * Revision 1.4.7.21  1997/06/17  17:56:40  ted
 * 	Put back in include of <inet/nd.h> to get prototypes.
 * 	[1997/06/17  17:56:28  ted]
 * 
 * Revision 1.4.7.20  1997/06/17  00:36:52  ted
 * 	Merged with changes from 1.4.7.19
 * 	[1997/06/17  00:36:19  ted]
 * 
 * 	Cleanse ndd code with new, Alteon developed code
 * 	[1997/06/13  02:28:44  ted]
 * 
 * 	tighten up driver code for better performance
 * 	[1997/06/06  22:38:37  ted]
 * 
 * Revision 1.4.7.19  1997/06/03  18:55:15  ted
 * 	Changes to support Sun naming of drivers (vge vs. alt)
 * 	[1997/06/03  18:54:17  ted]
 * 
 * Revision 1.4.7.18  1997/05/23  21:31:52  ted
 * 	Remove references to nicCmdsGetStats, which no longer exists.
 * 	[1997/05/23  21:30:58  ted]
 * 
 * Revision 1.4.7.17  1997/05/21  22:59:55  ted
 * 	Minor statistics tweaks
 * 	[1997/05/21  22:59:21  ted]
 * 
 * Revision 1.4.7.16  1997/05/21  03:42:45  ted
 * 	Fix compile warning
 * 	[1997/05/21  03:42:32  ted]
 * 
 * Revision 1.4.7.15  1997/05/21  03:36:14  ted
 * 	Use pub_tg.h.
 * 	Fix bug 218 (kstats not all initialized properly).
 * 	[1997/05/21  03:29:48  ted]
 * 
 * Revision 1.4.7.14  1997/05/12  21:20:29  ted
 * 	Change ierrors statistic to not include discarded packets, reducing the
 * 	number of errors shown in netstat -i to a reasonable number.
 * 	[1997/05/12  21:20:20  ted]
 * 
 * Revision 1.4.7.13  1997/05/03  05:00:33  ted
 * 	Fix up statistics
 * 	[1997/05/03  04:57:02  ted]
 * 
 * Revision 1.4.7.12  1997/04/25  18:49:05  ted
 * 	Support single driver for Sbus and PCi on Solaris
 * 	[1997/04/25  18:46:06  ted]
 * 
 * Revision 1.4.7.11  1997/04/12  00:11:49  ted
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:08:14  ted]
 * 
 * Revision 1.4.7.10  1997/03/24  08:34:42  ted
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:39  ted]
 * 
 * Revision 1.4.7.9  1997/03/17  08:48:00  skapur
 * 	Increase watchdog timeout to avoid snapping while firmware in sync loops.
 * 	Move interrupt lock to minimize probability of spurious interrupts.
 * 	Disable shared interrupts to eliminiate spurious interrupts message (this
 * 	will need correction with a coordinated change with the firmware).
 * 	[1997/03/17  08:40:19  skapur]
 * 
 * Revision 1.4.7.8  1997/03/03  05:17:17  ted
 * 	endian fixes, non-copydata mode fixes
 * 	[1997/03/03  05:16:32  ted]
 * 
 * Revision 1.4.7.7  1997/03/02  18:36:57  ted
 * 	Merged with changes from 1.4.7.6
 * 	[1997/03/02  18:36:20  ted]
 * 
 * 	redundancy support and a few bug fixes
 * 	[1997/03/02  18:26:00  ted]
 * 
 * Revision 1.4.7.6  1997/03/01  01:59:23  ted
 * 	Put the base address for the config space in the nic structure for the
 * 	temporary time we have it mapped so that config space accesses work in
 * 	the ph1 routine.
 * 	[1997/03/01  01:58:22  ted]
 * 
 * Revision 1.4.7.5  1997/02/21  02:06:55  ted
 * 	Prepare data structures and code structures to support redundant operation
 * 	[1997/02/21  01:57:17  ted]
 * 
 * Revision 1.4.7.4  1997/02/15  04:41:20  ted
 * 	Remove power management cruft (per Nara)
 * 	[1997/02/15  04:36:50  ted]
 * 
 * Revision 1.4.7.3  1997/02/12  00:28:08  ted
 * 	Implement ndd support using code borrowed from Sun
 * 	[1997/02/12  00:27:42  ted]
 * 
 * Revision 1.4.7.2  1997/02/11  04:08:15  ted
 * 	Add non-ddi/dki routines for support of nd_xxx and mi_xxx
 * 	[1997/02/11  04:04:07  ted]
 * 
 * Revision 1.4.7.1  1997/02/07  21:58:36  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:50:14  ted]
 * 
 * Revision 1.4.2.3  1996/11/25  20:05:56  ted
 * 	Look for Alteon PCI device info instead of Essential
 * 	[1996/11/25  20:05:42  ted]
 * 
 * Revision 1.4.2.2  1996/11/20  20:04:48  ted
 * 	Redo include file structuring.
 * 	[1996/11/20  19:57:51  ted]
 * 
 * Revision 1.4.2.1  1996/11/01  18:25:42  ted
 * 	change rr_ to tg_ and add propery copyright
 * 	[1996/10/30  23:01:09  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altinit.c
 *
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/streams/SunOS/altinit.c,v $
 * $Revision: 1.4.7.54 $ $Locker:  $
 * $Date: 1999/01/30 02:32:24 $ $Author: fhwang $ $State: Exp $
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/streams/SunOS/altinit.c,v 1.4.7.54 1999/01/30 02:32:24 fhwang Exp $";
#endif /* __lint */

/*
 * This file contains initialization functions for Solaris 2.x systems.
 */

/* general include files */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/kmem.h>
#include <sys/errno.h>
#include <ctype.h>

/* system specific include files */
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/kstat.h>		/* kernel statistics stuff */
#include <sys/ddi.h>
#include <sys/ddidmareq.h>	/* ddi dma structs */
#include <sys/sunddi.h>
#include <sys/ethernet.h>
/* would normally include <inet/common.h> but there are conflicting macros */
typedef	int	(*pfi_t)();
/* would normally include <inet/led.h>, but there are conflicting macros */
typedef	struct msgb	*MBLKP;
#include <inet/nd.h>

/* alteon include files */
#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "altstr.h"
#include "altioctl.h"
#include "pub_tg.h"
#include "altos.h"
#include "althdr.h"
#include "altreg.h"
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"
#include "release.h"
#include "link.h"

/* forwards for the local functions */
STATIC int altidentify(dev_info_t *);
STATIC int altprobe(dev_info_t *);
STATIC int altattach(dev_info_t *, ddi_attach_cmd_t);
STATIC int altdetach(dev_info_t *, ddi_detach_cmd_t);
STATIC int altgetinfo(dev_info_t *, ddi_info_cmd_t, void *, void **);
STATIC int altstat_kstat_update(kstat_t *, int);
STATIC void altstatinit(struct altinst *, int);
STATIC bool alt_param_register(struct tg_nic *);

/* definition of which address space to get */
#define CONFIG_REGION 0
#define SHARED_REGION 1

/* since we build on 2.5.1 systems we need to add this because it doesn't
 * show up until 2.6 systems.  If we start building on 2.6 systems this
 * has to go away.  The C compiler claims it defines a variable 
 * __`uname -s`_`uname -r` which would be __SunOS_5.6 for Solaris 2.6, but
 * this doesn't work for some reason */
#define DDI_PM_RESUME 2

/*
 * Definitions for module_info.
 */
#define	ALTIDNUM  (200)		/* module ID number */
#define	ALTMINPSZ (0)		/* min packet size */
#define	ALTMAXPSZ JUMBOBLOCKSIZE /* max packet size */
#define	ALTHIWAT  (128 * 1024)	/* hi-water mark */
#define	ALTLOWAT  (1)		/* lo-water mark */

/* structures for loadable driver */

static	struct module_info altminfo = {
    ALTIDNUM,			/* mi_idnum */
    ALTNAME,			/* mi_idname */
    ALTMINPSZ,			/* mi_minpsz */
    ALTMAXPSZ,			/* mi_maxpsz */
    ALTHIWAT,			/* mi_hiwat */
    ALTLOWAT			/* mi_lowat */
};

static	struct qinit altrinit = {
    (int (*)()) NULL,		/* qi_putp */
#if READ_SERVICE
    altrsrv,			/* qi_srvp */
#else /* READ_SERVICE */
    (int (*)()) NULL,		/* qi_srvp */
#endif /* READ_SERVICE */
    ALTOPEN,			/* qi_qopen */
    ALTCLOSE,			/* qi_qclose */
    (int (*)()) NULL,		/* qi_qadmin */
    &altminfo,			/* qi_minfo */
    NULL			/* qi_mstat */
};

static	struct qinit altwinit = {
    altwput,			/* qi_putp */
    altwsrv,			/* qi_srvp */
    (int (*)()) NULL,		/* qi_qopen */
    (int (*)()) NULL,		/* qi_qclose */
    (int (*)()) NULL,		/* qi_qadmin */
    &altminfo,			/* qi_minfo */
    NULL			/* qi_mstat */
};

static struct streamtab	alt_info = {
    &altrinit,			/* st_rdinit */
    &altwinit,			/* st_wrinit */
    NULL,			/* st_muxrinit */
    NULL			/* st_muxwrinit */
};

STATIC struct cb_ops alt_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &alt_info,			/* cb_stream */
    (D_MP | D_NEW)		/* no perimeters, let's see how this goes */
};
    

STATIC struct dev_ops alt_ops = {
    DEVO_REV,			/* devo_rev */
    0,				/* devo_refcnt */
    altgetinfo,			/* devo_getinfo */
    altidentify,		/* devo_identify */
    altprobe,			/* devo_probe */
    altattach,			/* devo_attach */
    altdetach,			/* devo_detach */
    nodev,			/* devo_reset */
    &alt_cb_ops,		/* devo_cb_ops */
    NULL,			/* devo_bus_ops */
    ddi_power			/* devo_power */
};

static struct modldrv modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    RELEASE_PRODUCT_NAME,	/* Name and version number */
    &alt_ops			/* driver ops */
};

static struct modlinkage modlinkage = {
    MODREV_1, &modldrv, NULL
};


int 
_init(void)
{
    int rc;

    ddi_soft_state_init(&altStatep, sizeof(struct altinst), NALT);
    mutex_init(&altTraceMu, "alt trace lock", MUTEX_SPIN, NULL);
    alt_itrace();
    if ((rc = mod_install(&modlinkage))) {
	mutex_destroy(&altTraceMu);
	ddi_soft_state_fini(&altStatep);
    }
    return(rc);
}

int
_fini(void)
{
    int rc;

    if (!(rc = mod_remove(&modlinkage))) {
	mutex_destroy(&altTraceMu);
	rw_destroy(&altstruplock);
	ddi_soft_state_fini(&altStatep);
    }
    return(rc);
}

int
_info(struct modinfo *modinfop)
{
    return(mod_info(&modlinkage, modinfop));
}

STATIC int
altidentify(dev_info_t *dip)
{
    if (strcmp(ddi_get_name(dip), ALTNAME) == 0)
	return(DDI_IDENTIFIED);
    else if (strcmp(ddi_get_name(dip), "SUNW,vge") == 0)
	return(DDI_IDENTIFIED);
    else if (strcmp(ddi_get_name(dip), "pci12ae,1") == 0)
	return(DDI_IDENTIFIED);
    else {
	printf("Unidentified card in altidentify, %s\n", ddi_get_name(dip));
	return(DDI_NOT_IDENTIFIED);
    }
}

STATIC int
altprobe(dev_info_t *dip)
{
    return(DDI_PROBE_SUCCESS);
}

STATIC int
altattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    struct altinst *altp, *altp2;
    struct tg_nic *nic;
    int instance;
    int rc;
    int i;
    static bool once = TRUE;

    /* define device attributes */
    ddi_device_acc_attr_t access_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_NEVERSWAP_ACC,	/* we let the card take care of this */
	DDI_STRICTORDER_ACC
    };

    switch (cmd) {
      case DDI_ATTACH:
	/* get assigned instance number */
	instance = ddi_get_instance(dip);

	/* allocate the altinst structure */
	if (ddi_soft_state_zalloc(altStatep, instance))
	    return(DDI_FAILURE);
	altp = ddi_get_soft_state(altStatep, instance);

	ALTTRACE(altattch, 0x20000, dip, instance, altp, altStatep);

	/* save off the dip */
	ddi_set_driver_private(dip, (caddr_t)altp);
	altp->dip = dip;
	altp->instance = instance;

	/* allocate the NIC structure and link it in */
	if ((rc = alt_InitNICstruct(altp))) {
	    ddi_soft_state_free(altStatep, instance);
	    ERROR2("Allocation of the NIC structure failed %d\n", rc);
	    ALTTRACEE(inicmemf, 0xf2002, rc, 0, 0, 0);
	    return(DDI_FAILURE);
	}

	/* initialize state info */
	altp->state = ALT_NOTREADY;
	nic = altp->initNic;
	strcpy(nic->name, ALTNAME); /* put in product name */
#if defined(OEM_OP1)	
	altp->flags |= ALT_DRIVER_IS_OP1; /* so other modules can know */
#endif /* defined(OEM_OP1) */

	/* figure out whether it's a PCI bus or an SBus NIC */
	if (strcmp(ddi_get_name(dip), ALTNAME) == 0 ||
	    strcmp(ddi_get_name(dip), "SUNW,vge") == 0) {
	    NOTE4("%s%d: %s SBus NIC found", nic->name, altp->instance,
		  ALTIDENT);
	    nic->nic_type = ALT_SBUS_NIC;
	} else if (strcmp(ddi_get_name(dip), "pci12ae,1") == 0) {
	    NOTE4("%s%d: %s PCI NIC found", nic->name, altp->instance,
		  ALTIDENT);
	    nic->nic_type = ALT_PCI_NIC;
	} else {
	    ERROR4("%s%d: Unknown driver name %s\n", 
		   nic->name, altp->instance, ddi_get_name(dip));
	    ALTTRACEE(unkdrnam, 0xf200b, rc, 0, 0, 0);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}
	    
	/* mark as the primary interface */
	altp->ifconfigNic = nic;
	nic->initAltp = altp;

	/* start with shared rupts, may be turned off later */
	nic->init_flags = ALT_INIT_SHARED_RUPT;

	/* get the iblock cookie so we can initialize the locks correctly */
	if ((rc = ddi_get_iblock_cookie(dip, 0, &nic->iblock_cookie)) !=
	    DDI_SUCCESS) {
	    ddi_soft_state_free(altStatep, instance);
	    ERROR2("ddi_get_iblock_cookie failed %d\n", rc);
	    ALTTRACEE(ERROR2, 0xf200c, rc, 0, 0, 0);
	    alt_FreeNICstruct(altp);
	    return(DDI_FAILURE);
	}

	/* initialize memory for the NIC */
	if ((rc = alt_InitNICmem(altp))) {
	    ddi_soft_state_free(altStatep, instance);
	    ERROR2("Initialization of the NIC structure memory failed %d\n", 
		   rc);
	    ALTTRACEE(inicmemfr, 0xf200f, rc, 0, 0, 0);
	    alt_FreeNICstruct(altp);
	    return(DDI_FAILURE);
	}

	/* map device registers, config space only */
	if ((rc = ddi_regs_map_setup(dip, CONFIG_REGION, &nic->configBase, 
				     0, 0, &access_attr, 
				     &nic->configHandle)) != DDI_SUCCESS) {
	    ERROR4("%s%d: ddi_regs_map_setup failed %d\n", 
		   nic->name, altp->instance, rc);
	    ALTTRACEE(noddireg, 0xf2001, rc, 0, 0, 0);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* initialize the hardware (config space and stuff) */
	if ((rc = alt_InitNICph1(altp))) {
	    ERROR4("%s%d: Phase 1 Hardware Initialization failed %d\n", 
		   nic->name, altp->instance, rc);
	    ALTTRACEE(altph1fl, 0xf2003, rc, 0, 0, 0);
	    ddi_regs_map_free(&nic->configHandle);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* setup shared memory access */
	if ((rc = ddi_regs_map_setup(dip, SHARED_REGION, &nic->base, 0, 0,
			       &access_attr, &nic->handle)) != DDI_SUCCESS) {
	    ERROR4("%s%d: ddi_regs_map_setup space 1 failed %d\n",
		   nic->name, altp->instance, rc);
	    ALTTRACEE(noddirgs, 0xf2004, rc, 0, 0, 0);
	    ddi_regs_map_free(&nic->configHandle);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* check for high level rupts, fail for now */
	if (ddi_intr_hilevel(dip, 0)) {
	    ERROR3("%s%d: Rupts are high level, not supported\n",
		   nic->name, altp->instance);
	    ALTTRACEE(2hirupt, 0xf2005, 0, 0, 0, 0);
 	    ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* set up the interrupt handler */
	if ((rc = ddi_add_intr(dip, 0, (ddi_iblock_cookie_t *) 0,
			       &nic->idev_cookie, 
			       ALTINTR, (char *)altp->nic) != DDI_SUCCESS)) {
	    ERROR4("%s%d: ddi_add_intr failed %d\n",
		   nic->name, altp->instance, rc);
	    ALTTRACEE(addintrf, 0xf2006, rc, 0, 0, 0);
 	    ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

        /* init the trace buffer mutex, if necessary */
        if (!altTraceMutexInited) {
            mutex_init(&altTraceMu, "alt trace lock", MUTEX_DRIVER,
                       (void *)nic->iblock_cookie);
            altTraceMutexInited = TRUE;
        }

	/* init the per instance locks */
	mutex_init(&nic->intrMu, "alt mutex", MUTEX_DRIVER, 
		   (void *)nic->iblock_cookie);

	/* use the trace mutex for exclusion on initing altstruplock */
	ALT_DISABLE(spl, &altTraceMu);
	if (once) {
	    once = FALSE;
	    rw_init(&altstruplock, "alt streams linked list lock",
		    RW_DRIVER, (void *)nic->iblock_cookie);
	}
	ALT_ENABLE(spl, &altTraceMu);

	/* Create the filesystem device node. */
	if ((rc = ddi_create_minor_node(dip, ALTNAME, S_IFCHR, instance, 
					DDI_NT_NET,
					CLONE_DEV) != DDI_SUCCESS)) {
	    ERROR4("%s%d: ddi_create_minor_node failed %d\n",
		   nic->name, altp->instance, rc);
	    ALTTRACEE(cremnf, 0xf2007, rc, 0, 0, 0);
	    /*	    ddi_remove_intr(dip, 0, nic->iblock_cookie); */
	    ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    mutex_destroy(&nic->intrMu);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* make sure we've got a valid instance to start with or we'll
	 * never get any ndd variables through */
	for (i=0; i<NALT; i++) {
	    if ((altp2 = (struct altinst *)
		 ddi_get_soft_state(altStatep, i)) != 0)
		altp2->nic->tuneParms[ND_INSTANCE].parm_val = instance;
	}
	alt_param_arr[ND_INSTANCE].parm_val = instance;

	/* register the nd variables */
	for (i = 0; i < NUM_TUNE_PARMS; i++) {
	    nic->tuneParms[i] = alt_param_arr[i];
	    nic->tuneParms[i].altp = altp;
	}
	if (!nic->alt_nd_ptr && 
	    !alt_param_register(nic)) {
	    ERROR3("%s%d: Ndd parameter registration failed error\n",
		   nic->name, altp->instance);
	    ALTTRACEE(parmregf, 0xf200a, nic->alt_nd_ptr, 0, 0, 0);
	    ddi_remove_minor_node(dip, 0);
	    /* 	    ddi_remove_intr(dip, 0, nic->iblock_cookie); */
	    ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    mutex_destroy(&nic->intrMu);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* Final NIC initialization (includes firmware loading) */
	if ((rc = alt_InitNICph2(altp))) {
	    ERROR4("%s%d: Phase 2 Hardware Initialization failed %d\n", 
		   nic->name, altp->instance, rc);
	    ALTTRACEE(nicph2f, 0xf2008, rc, 0, 0, 0);
	    ddi_remove_minor_node(dip, 0);
	    ddi_remove_intr(dip, 0, nic->iblock_cookie);
	    ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    mutex_destroy(&nic->intrMu);
	    alt_FreeNICmem(altp);
	    alt_FreeNICstruct(altp);
	    ddi_soft_state_free(altStatep, instance);
	    return(DDI_FAILURE);
	}

	/* initialize the reverse translation vlan table */
	for (i=0; i < MAX_VLAN_IDS; i++)
	    altp->vlan_translate[i] = -1;

	/* initialize statistics area (per vlan) */
	for (i=0; i < MAX_VLANS; i++)
	    altstatinit(altp, i);

	/* start the watchdog timer running at 1/3 freq of stats */
	/* XXX 100X for now, due to mac sync delays */
	altp->timeoutId = 
	    timeout(altWatchdog, (char *)altp, 
		    drv_usectohz(100 * 
				 (nic->genComm->gen_com.tuneParms.stat_ticks *
				  (USECS_PER_SEC / TG_TICKS_PER_SEC))));

	/* start the redundancy timer running at 1/4 second intervals */
	altp->redTimeoutId = 
	    timeout(altRedundTimer, (char *)altp, drv_usectohz(250000));

	ddi_report_dev(dip);
	return(DDI_SUCCESS);

    case DDI_RESUME:
    case DDI_PM_RESUME:
        instance = ddi_get_instance(dip);
        altp = ddi_get_soft_state(altStatep, instance);
        /* if no soft state, we must not really be up, cleaned up in attach */
        if (!altp)
            return(DDI_FAILURE);
 
        nic = altp->initNic;
        nic->flags &= ~ALT_SUSPENDED;
 
        /* restart everything, but don't allocate anything */
        /* initialize the hardware (config space and stuff) */
        if ((rc = alt_InitNICph1(altp))) {
	    ERROR4("%s%d: Phase 1 Hardware Initialization failed %d\n", 
                   nic->name, altp->instance, rc);
            ALTTRACEE(ph1flres, 0xf200d, rc, 0, 0, 0);
            ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
	    alt_FreeNICmem(altp);
            alt_FreeNICstruct(altp);
            ddi_soft_state_free(altStatep, instance);
            return(DDI_FAILURE);
        }
 
        /* Final NIC initialization (includes firmware loading) */
        if ((rc = alt_InitNICph2(altp))) {
	    ERROR4("%s%d: Phase 2 Hardware Initialization failed %d\n", 
                   nic->name, altp->instance, rc);
            ALTTRACEE(nicph2fres, 0xf200e, rc, 0, 0, 0);
            ddi_remove_minor_node(dip, 0);
            ddi_remove_intr(dip, 0, nic->iblock_cookie);
            ddi_regs_map_free(&nic->handle);
	    ddi_regs_map_free(&nic->configHandle);
            mutex_destroy(&nic->intrMu);
	    alt_FreeNICmem(altp);
            alt_FreeNICstruct(altp);
            ddi_soft_state_free(altStatep, instance);
            return(DDI_FAILURE);
        }

	/* set up the host state so we can start sending and receiving pkts */
	if (altp->flags & ALT_HOST_UP) {
	    altHostState(altp, TRUE);
	    if (altp->flags & ALT_REDUNDANT)
		altHostState(altp->attachNic->initAltp, TRUE);
	}
        return(DDI_SUCCESS);

      default:
	return(DDI_FAILURE);
    }
}

STATIC int
altdetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    struct altinst *altp;
    struct tg_nic *nic;
    int instance;
    int i;

    switch (cmd) {
    case DDI_DETACH:
        instance = ddi_get_instance(dip);
        altp = ddi_get_soft_state(altStatep, instance);
        /* if no soft state, we must not really be up, cleaned up in attach */
        if (!altp)
            return(DDI_FAILURE);
 
        nic = altp->initNic;
 
	/* check to see if we're busy */
	if ((altp->flags & ALT_HOST_UP) || (nic->flags & ALT_SUSPENDED)) {
	    ERROR3("%s%d: NIC is BUSY, can't detach\n",
		   nic->name, altp->instance);
	    return(DDI_FAILURE);
	}

        /* stop the device */
        if (altStopCard(altp)) {
	    ERROR3("%s%d: NIC hardware termination failed\n",
                   nic->name, altp->instance);
            return(DDI_FAILURE);
        }
 
        /* clean up the memory */
        if (alt_FreeNICmem(altp)) {
	    ERROR3("%s%d: Unable to free NIC memory\n",
                   nic->name, altp->instance);
            return(DDI_FAILURE);
        }
 
        /* remove timeout rupts */
        untimeout(altp->timeoutId);
        untimeout(altp->redTimeoutId);
 
        /* remove statistics area */
        for (i=0; i < MAX_VLANS; i++)
            kstat_delete(altp->alt_ksp[i]);
 
        /* clean everything up */
        ddi_remove_minor_node(dip, 0);
        ddi_remove_intr(dip, 0, nic->iblock_cookie);
        ddi_regs_map_free(&nic->handle);
	ddi_regs_map_free(&nic->configHandle);
        mutex_destroy(&nic->intrMu);
        alt_FreeNICstruct(altp);
        ddi_soft_state_free(altStatep, instance);
        return(DDI_SUCCESS);
 
    case DDI_SUSPEND:
    case DDI_PM_SUSPEND:
        instance = ddi_get_instance(dip);
        altp = ddi_get_soft_state(altStatep, instance);
        /* if no soft state, we must not really be up, cleaned up in attach */
        if (!altp)
            return(DDI_FAILURE);
 
        nic = altp->initNic;
        nic->flags |= ALT_SUSPENDED;
 
        /* just stop the device */
        if (altStopCard(altp)) {
	    ERROR3("%s%d: NIC hardware termination failed\n",
                   nic->name, altp->instance);
            return(DDI_FAILURE);
        }
 
        /* remove timeout rupts XXX, needs work */
        untimeout(altp->timeoutId);
        untimeout(altp->redTimeoutId);
 
        return(DDI_SUCCESS);
 
    default:
        return(DDI_FAILURE);

    }
}

STATIC int
altgetinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg, void **result)
{
    int rc = DDI_FAILURE;
    struct altstr *slp;
    struct altinst *altp;

    switch (cmd) {
      case DDI_INFO_DEVT2INSTANCE:
	*result = (void *) getminor((dev_t)arg);
	rc = DDI_SUCCESS;
	break;

      case DDI_INFO_DEVT2DEVINFO:
	rw_enter(&altstruplock, RW_READER);
	dip = NULL;
	for (slp = altstrup; slp; slp = slp->sl_nextp)
	    if (slp->sl_minor == getminor((dev_t)arg))
		break;
	if (slp && slp->sl_altp)
	    dip = slp->sl_altp->dip;
	rw_exit(&altstruplock);
	
	if (dip) {
	    *result = (void *) dip;
	    rc = DDI_SUCCESS;
	}
	break;

      default:
	rc = DDI_FAILURE;
	break;
    }

    return(rc);
}

STATIC int
altstat_kstat_update(kstat_t *ksp, int rw)
{
    struct kstatPrivStruct *kpriv = (struct kstatPrivStruct *)ksp->ks_private;
    struct altinst *altp = kpriv->altp;
    struct altstat *altsp = (struct altstat *)ksp->ks_data;
    struct tg_nic *nic = altp->initNic;
    struct tg_stats *tgsp;    
    int i;
    int vlan_id = kpriv->vlan;

    /* XXX gotta get these out to the firmware, write a new COMMAND for
     * this */
    if (rw == KSTAT_WRITE) {
	/* write the persistent statistics into best fit, we've lost details */
	tgsp = nic->genInfoStats2;
        altp->inPackets[vlan_id] = altsp->ipackets.value.ul;
	tgsp->ifHCInUcastPkts = altsp->ipackets.value.ul;
	altp->inPackets[vlan_id] = altsp->ipackets.value.ul;
	altp->outPackets[vlan_id] = altsp->opackets.value.ul;
	tgsp->dot3StatsAlignmentErrors = 
	    altsp->dot3StatsAlignmentErrors.value.ul;
	tgsp->dot3StatsFCSErrors =
	    altsp->dot3StatsFCSErrors.value.ul;
	tgsp->dot3StatsSingleCollisionFrames =
	    altsp->dot3StatsSingleCollisionFrames.value.ul;
	tgsp->dot3StatsMultipleCollisionFrames =
	    altsp->dot3StatsMultipleCollisionFrames.value.ul;
	tgsp->dot3StatsSQETestErrors =
	    altsp->dot3StatsSQETestErrors.value.ul;
	tgsp->dot3StatsDeferredTransmissions =
	    altsp->dot3StatsDeferredTransmissions.value.ul;
	tgsp->dot3StatsLateCollisions =
	    altsp->dot3StatsLateCollisions.value.ul;
	tgsp->dot3StatsExcessiveCollisions =
	    altsp->dot3StatsExcessiveCollisions.value.ul;
	tgsp->dot3StatsInternalMacTransmitErrors =
	    altsp->dot3StatsInternalMacTransmitErrors.value.ul;
	tgsp->dot3StatsCarrierSenseErrors =
	    altsp->dot3StatsCarrierSenseErrors.value.ul;
	tgsp->dot3StatsFrameTooLongs =
	    altsp->dot3StatsFrameTooLongs.value.ul;
	tgsp->dot3StatsInternalMacReceiveErrors =
	    altsp->dot3StatsInternalMacReceiveErrors.value.ul;
	tgsp->ifIndex =
	    altsp->ifIndex.value.ul;
	tgsp->ifType =
	    altsp->ifType.value.ul;
	tgsp->ifMtu =
	    altsp->ifMtu.value.ul;
	tgsp->ifSpeed =
	    altsp->ifSpeed.value.ul;
	tgsp->ifAdminStatus =
	    altsp->ifAdminStatus.value.ul;
	tgsp->ifOperStatus =
	    altsp->ifOperStatus.value.ul;
	tgsp->ifLastChange =
	    altsp->ifLastChange.value.ul;
	tgsp->ifHCInOctets =
	    altsp->ifHCInOctets.value.ul;
	tgsp->ifHCInUcastPkts =
	    altsp->ifHCInUcastPkts.value.ul;
	tgsp->ifHCInMulticastPkts =
	    altsp->ifHCInMulticastPkts.value.ul;
	tgsp->ifHCInBroadcastPkts =
	    altsp->ifHCInBroadcastPkts.value.ul;
	tgsp->ifInDiscards =
	    altsp->ifInDiscards.value.ul;
	tgsp->ifInErrors =
	    altsp->ifInErrors.value.ul;
	tgsp->ifInUnknownProtos =
	    altsp->ifInUnknownProtos.value.ul;
	tgsp->ifHCOutOctets =
	    altsp->ifHCOutOctets.value.ul;
	tgsp->ifHCOutUcastPkts =
	    altsp->ifHCOutUcastPkts.value.ul;
	tgsp->ifHCOutMulticastPkts =
	    altsp->ifHCOutMulticastPkts.value.ul;
	tgsp->ifHCOutBroadcastPkts =
	    altsp->ifHCOutBroadcastPkts.value.ul;
	tgsp->ifOutDiscards =
	    altsp->ifOutDiscards.value.ul;
	tgsp->ifOutErrors =
	    altsp->ifOutErrors.value.ul;
	tgsp->ifOutQLen =
	    altsp->ifOutQLen.value.ul;
	tgsp->ifLinkUpDownTrapEnable =
	    altsp->ifLinkUpDownTrapEnable.value.ul;
	tgsp->ifHighSpeed =
	    altsp->ifHighSpeed.value.ul;
	tgsp->ifPromiscuousMode =
	    altsp->ifPromiscuousMode.value.ul;
	tgsp->ifConnectorPresent =
	    altsp->ifConnectorPresent.value.ul;
	MEMCPY((char *)&tgsp->ifPhysAddress + 2,
	       altsp->ifPhysAddress.value.c, 6);
        MEMCPY((MEMCPY_TYPE)tgsp->ifDescr, altsp->ifDescr[0].value.c, 16);
        MEMCPY((MEMCPY_TYPE)&tgsp->ifDescr[16], altsp->ifDescr[1].value.c, 16);
	tgsp->nicCmdsHostState =
	    altsp->nicCmdsHostState.value.ul;
	tgsp->nicCmdsFDRFiltering =
	    altsp->nicCmdsFdrFiltering.value.ul;
	tgsp->nicCmdsSetRecvProdIndex =
	    altsp->nicCmdsSetRecvProdIndex.value.ul;
	tgsp->nicCmdsUpdateGencommStats =
	    altsp->nicCmdsUpdateGencommStats.value.ul;
	tgsp->nicCmdsResetJumboRing =
	    altsp->nicCmdsResetJumboRing.value.ul;
	tgsp->nicCmdsAddMCastAddr =
	    altsp->nicCmdsAddMCastAddr.value.ul;
	tgsp->nicCmdsDelMCastAddr =
	    altsp->nicCmdsDelMCastAddr.value.ul;
	tgsp->nicCmdsSetPromiscMode =
	    altsp->nicCmdsSetPromiscMode.value.ul;
	tgsp->nicCmdsLinkNegotiate =
	    altsp->nicCmdsLinkNegotiate.value.ul;
	tgsp->nicCmdsSetMACAddr =
	    altsp->nicCmdsSetMACAddr.value.ul;
	tgsp->nicCmdsClearProfile =
	    altsp->nicCmdsClearProfile.value.ul;
	tgsp->nicCmdsSetMulticastMode =
	    altsp->nicCmdsSetMulticastMode.value.ul;
	tgsp->nicCmdsClearStats =
	    altsp->nicCmdsClearStats.value.ul;
	tgsp->nicCmdsSetRecvJumboProdIndex =
	    altsp->nicCmdsSetRecvJumboProdIndex.value.ul;
	tgsp->nicCmdsSetRecvMiniProdIndex =
	    altsp->nicCmdsSetRecvMiniProdIndex.value.ul;
	tgsp->nicCmdsRefreshStats =
	    altsp->nicCmdsRefreshStats.value.ul;
	tgsp->nicCmdsUnknown =
	    altsp->nicCmdsUnknown.value.ul;
	tgsp->nicEventsNICFirmwareOperational =
	    altsp->nicEventsNICFirmwareOperational.value.ul;
	tgsp->nicEventsStatsUpdated =
	    altsp->nicEventsStatsUpdated.value.ul;
	tgsp->nicEventsLinkStateChanged =
	    altsp->nicEventsLinkStateChanged.value.ul;
	tgsp->nicEventsError =
	    altsp->nicEventsError.value.ul;
	tgsp->nicEventsMCastListUpdated =
	    altsp->nicEventsMCastListUpdated.value.ul;
	tgsp->nicEventsResetJumboRing =
	    altsp->nicEventsResetJumboRing.value.ul;
	tgsp->nicRingSetSendProdIndex =
	    altsp->nicRingSetSendProdIndex.value.ul;
	tgsp->nicRingSetSendConsIndex =
	    altsp->nicRingSetSendConsIndex.value.ul;
	tgsp->nicRingSetRecvReturnProdIndex =
	    altsp->nicRingSetRecvReturnProdIndex.value.ul;
	tgsp->nicInterrupts =
	    altsp->nicInterrupts.value.ul;
	tgsp->nicAvoidedInterrupts =
	    altsp->nicAvoidedInterrupts.value.ul;
	tgsp->nicEventThresholdHit =
	    altsp->nicEventThresholdHit.value.ul;
	tgsp->nicSendThresholdHit =
	    altsp->nicSendThresholdHit.value.ul;
	tgsp->nicRecvThresholdHit =
	    altsp->nicRecvThresholdHit.value.ul;
	tgsp->nicDmaRdOverrun =
	    altsp->nicDmaRdOverrun.value.ul;
	tgsp->nicDmaRdUnderrun =
	    altsp->nicDmaRdUnderrun.value.ul;
	tgsp->nicDmaWrOverrun =
	    altsp->nicDmaWrOverrun.value.ul;
	tgsp->nicDmaWrUnderrun =
	    altsp->nicDmaWrUnderrun.value.ul;
	tgsp->nicDmaWrMasterAborts =
	    altsp->nicDmaWrMasterAborts.value.ul;
	tgsp->nicDmaRdMasterAborts =
	    altsp->nicDmaRdMasterAborts.value.ul;
	tgsp->nicDmaWriteRingFull = 
	    altsp->nicDmaWriteRingFull.value.ul;
	tgsp->nicDmaReadRingFull = 
	    altsp->nicDmaReadRingFull.value.ul;
	tgsp->nicEventRingFull = 
	    altsp->nicEventRingFull.value.ul;
	tgsp->nicEventProducerRingFull = 
	    altsp->nicEventProducerRingFull.value.ul;
	tgsp->nicTxMacDescrRingFull = 
	    altsp->nicTxMacDescrRingFull.value.ul;
	tgsp->nicOutOfTxBufSpaceFrameRetry =
	    altsp->nicOutOfTxBufSpaceFrameRetry.value.ul;
	tgsp->nicNoMoreWrDMADescriptors = 
	    altsp->nicNoMoreWrDMADescriptors.value.ul;
	tgsp->nicNoMoreRxBDs = 
	    altsp->nicNoMoreRxBDs.value.ul;
	tgsp->nicNoSpaceInReturnRing = 
	    altsp->nicNoSpaceInReturnRing.value.ul;
	tgsp->nicRecvBDs = 
	    altsp->nicRecvBDs.value.ul;
	tgsp->nicSendBDs = 
	    altsp->nicSendBDs.value.ul;
	tgsp->nicJumboRecvBDs = 
	    altsp->nicJumboRecvBDs.value.ul;
	tgsp->nicMiniRecvBDs = 
	    altsp->nicMiniRecvBDs.value.ul;
	tgsp->nicTotalRecvBDs = 
	    altsp->nicTotalRecvBDs.value.ul;
	tgsp->nicTotalSendBDs = 
	    altsp->nicTotalSendBDs.value.ul;
        tgsp->nicJumboSpillOver =
            altsp->nicJumboSpillOver.value.ul;
        tgsp->nicSbusHangCleared =
            altsp->nicSbusHangCleared.value.ul;
        tgsp->nicEnqEventDelayed =
            altsp->nicEnqEventDelayed.value.ul;
	tgsp->nicMacRxLateColls = 
	    altsp->nicMacRxLateColls.value.ul;
	tgsp->nicMacRxLinkLostDuringPkt = 
	    altsp->nicMacRxLinkLostDuringPkt.value.ul;
	tgsp->nicMacRxPhyDecodeErr = 
	    altsp->nicMacRxPhyDecodeErr.value.ul;
	tgsp->nicMacRxMacAbort = 
	    altsp->nicMacRxMacAbort.value.ul;
	tgsp->nicMacRxTruncNoResources = 
	    altsp->nicMacRxTruncNoResources.value.ul;
	tgsp->nicMacRxDropUla = 
	    altsp->nicMacRxDropUla.value.ul;
	tgsp->nicMacRxDropMcast = 
	    altsp->nicMacRxDropMcast.value.ul;
	tgsp->nicMacRxFlowControl = 
	    altsp->nicMacRxFlowControl.value.ul;
	tgsp->nicMacRxDropSpace = 
	    altsp->nicMacRxDropSpace.value.ul;
	tgsp->nicMacRxColls = 
	    altsp->nicMacRxColls.value.ul;
	tgsp->nicMacRxTotalAttns =
	    altsp->nicMacRxTotalAttns.value.ul;
	tgsp->nicMacRxLinkAttns =
	    altsp->nicMacRxLinkAttns.value.ul;
	tgsp->nicMacRxSyncAttns =
	    altsp->nicMacRxSyncAttns.value.ul;
	tgsp->nicMacRxConfigAttns =
	    altsp->nicMacRxConfigAttns.value.ul;
	tgsp->nicMacReset =
	    altsp->nicMacReset.value.ul;
        tgsp->nicMacRxBufDescrAttns =
            altsp->nicMacRxBufDescrAttns.value.ul;
	tgsp->nicMacRxBufAttns =
	    altsp->nicMacRxBufAttns.value.ul;
	tgsp->nicMacRxZeroFrameCleanup =
	    altsp->nicMacRxZeroFrameCleanup.value.ul;
	tgsp->nicMacRxOneFrameCleanup =
	    altsp->nicMacRxOneFrameCleanup.value.ul;
	tgsp->nicMacRxMultipleFrameCleanup =
	    altsp->nicMacRxMultipleFrameCleanup.value.ul;
	tgsp->nicMacRxTimerCleanup =
	    altsp->nicMacRxTimerCleanup.value.ul;
        tgsp->nicMacRxDmaCleanup =
           altsp->nicMacRxDmaCleanup.value.ul;
	for (i=0; i<15; i++) {
	    tgsp->nicMacTxCollisionHistogram[0] =
		altsp->nicMacTxCollisionHistogram[0].value.ul;
	}
	tgsp->nicMacTxTotalAttns =
	    altsp->nicMacTxTotalAttns.value.ul;

	/* fix up cache stuff */
        MALLOC_CACHE_WB(nic->genInfoDmaInfo.handle, tgsp, 0,
                        sizeof(struct tg_stats));

        /* tell the card */
        altRefreshNicStats(altp);

    } else {
	tgsp = &nic->genInfo->stats;
	/* fix up cache stuff */
	MALLOC_CACHE_INVAL(nic->genInfoDmaInfo.handle, tgsp, 
			   0, sizeof(struct tg_stats));
	altsp->ipackets.value.ul = altp->inPackets[vlan_id];
	altsp->ierrors.value.ul = tgsp->ifInErrors;
	altsp->opackets.value.ul = altp->outPackets[vlan_id];
	altsp->oerrors.value.ul = tgsp->ifOutDiscards + tgsp->ifOutErrors;
	altsp->collisions.value.ul = tgsp->dot3StatsSingleCollisionFrames +
	    tgsp->dot3StatsMultipleCollisionFrames;
	altsp->defers.value.ul = tgsp->dot3StatsDeferredTransmissions;
	altsp->dot3StatsAlignmentErrors.value.ul = 
	    tgsp->dot3StatsAlignmentErrors;
	altsp->dot3StatsFCSErrors.value.ul = 
	    tgsp->dot3StatsFCSErrors;
	altsp->dot3StatsSingleCollisionFrames.value.ul = 
	    tgsp->dot3StatsSingleCollisionFrames;
	altsp->dot3StatsMultipleCollisionFrames.value.ul = 
	    tgsp->dot3StatsMultipleCollisionFrames;
	altsp->dot3StatsSQETestErrors.value.ul = 
	    tgsp->dot3StatsSQETestErrors;
	altsp->dot3StatsDeferredTransmissions.value.ul = 
	    tgsp->dot3StatsDeferredTransmissions;
	altsp->dot3StatsLateCollisions.value.ul = 
	    tgsp->dot3StatsLateCollisions;
	altsp->dot3StatsExcessiveCollisions.value.ul = 
	    tgsp->dot3StatsExcessiveCollisions;
	altsp->dot3StatsInternalMacTransmitErrors.value.ul = 
	    tgsp->dot3StatsInternalMacTransmitErrors;
	altsp->dot3StatsCarrierSenseErrors.value.ul = 
	    tgsp->dot3StatsCarrierSenseErrors;
	altsp->dot3StatsFrameTooLongs.value.ul = 
	    tgsp->dot3StatsFrameTooLongs;
	altsp->dot3StatsInternalMacReceiveErrors.value.ul = 
	    tgsp->dot3StatsInternalMacReceiveErrors;
	altsp->ifIndex.value.ul = 
	    tgsp->ifIndex;
	altsp->ifType.value.ul = 
	    tgsp->ifType;
	altsp->ifMtu.value.ul = 
	    tgsp->ifMtu;
	altsp->ifSpeed.value.ul = 
	    tgsp->ifSpeed;
	altsp->ifAdminStatus.value.ul = 
	    tgsp->ifAdminStatus;
	altsp->ifOperStatus.value.ul = 
	    tgsp->ifOperStatus;
	altsp->ifLastChange.value.ul = 
	    tgsp->ifLastChange;
	altsp->ifHCInOctets.value.ul = 
	    tgsp->ifHCInOctets;
	altsp->ifHCInUcastPkts.value.ul = 
	    tgsp->ifHCInUcastPkts;
	altsp->ifHCInMulticastPkts.value.ul = 
	    tgsp->ifHCInMulticastPkts;
	altsp->ifHCInBroadcastPkts.value.ul = 
	    tgsp->ifHCInBroadcastPkts;
	altsp->ifInDiscards.value.ul = 
	    tgsp->ifInDiscards;
	altsp->ifInErrors.value.ul = 
	    tgsp->ifInErrors;
	altsp->ifInUnknownProtos.value.ul = 
	    tgsp->ifInUnknownProtos;
	altsp->ifHCOutOctets.value.ul = 
	    tgsp->ifHCOutOctets;
	altsp->ifHCOutUcastPkts.value.ul = 
	    tgsp->ifHCOutUcastPkts;
	altsp->ifHCOutMulticastPkts.value.ul = 
	    tgsp->ifHCOutMulticastPkts;
	altsp->ifHCOutBroadcastPkts.value.ul = 
	    tgsp->ifHCOutBroadcastPkts;
	altsp->ifOutDiscards.value.ul = 
	    tgsp->ifOutDiscards;
	altsp->ifOutErrors.value.ul = 
	    tgsp->ifOutErrors;
	altsp->ifOutQLen.value.ul = 
	    tgsp->ifOutQLen;
	altsp->ifLinkUpDownTrapEnable.value.ul = 
	    tgsp->ifLinkUpDownTrapEnable;
	altsp->ifHighSpeed.value.ul = 
	    tgsp->ifHighSpeed;
	altsp->ifPromiscuousMode.value.ul = 
	    tgsp->ifPromiscuousMode;
	altsp->ifConnectorPresent.value.ul = 
	    tgsp->ifConnectorPresent;
	MEMCPY(altsp->ifPhysAddress.value.c, 
	       (char *)&tgsp->ifPhysAddress + 2, 16);
	MEMCPY(altsp->ifDescr[0].value.c, (MEMCPY_TYPE)tgsp->ifDescr, 16);
	MEMCPY(altsp->ifDescr[1].value.c, (MEMCPY_TYPE)&tgsp->ifDescr[16], 16);
	altsp->nicCmdsHostState.value.ul = 
	    tgsp->nicCmdsHostState;
	altsp->nicCmdsFdrFiltering.value.ul = 
	    tgsp->nicCmdsFDRFiltering;
	altsp->nicCmdsSetRecvProdIndex.value.ul = 
	    tgsp->nicCmdsSetRecvProdIndex;
	altsp->nicCmdsUpdateGencommStats.value.ul = 
	    tgsp->nicCmdsUpdateGencommStats;
	altsp->nicCmdsResetJumboRing.value.ul = 
	    tgsp->nicCmdsResetJumboRing;
	altsp->nicCmdsAddMCastAddr.value.ul = 
	    tgsp->nicCmdsAddMCastAddr;
	altsp->nicCmdsDelMCastAddr.value.ul = 
	    tgsp->nicCmdsDelMCastAddr;
	altsp->nicCmdsSetPromiscMode.value.ul = 
	    tgsp->nicCmdsSetPromiscMode;
	altsp->nicCmdsLinkNegotiate.value.ul = 
	    tgsp->nicCmdsLinkNegotiate;
	altsp->nicCmdsSetMACAddr.value.ul = 
	    tgsp->nicCmdsSetMACAddr;
	altsp->nicCmdsClearProfile.value.ul = 
	    tgsp->nicCmdsClearProfile;
	altsp->nicCmdsSetMulticastMode.value.ul = 
	    tgsp->nicCmdsSetMulticastMode;
	altsp->nicCmdsClearStats.value.ul = 
	    tgsp->nicCmdsClearStats;
	altsp->nicCmdsSetRecvJumboProdIndex.value.ul = 
	    tgsp->nicCmdsSetRecvJumboProdIndex;
	altsp->nicCmdsSetRecvMiniProdIndex.value.ul = 
	    tgsp->nicCmdsSetRecvMiniProdIndex;
	altsp->nicCmdsRefreshStats.value.ul = 
	    tgsp->nicCmdsRefreshStats;
	altsp->nicCmdsUnknown.value.ul = 
	    tgsp->nicCmdsUnknown;
	altsp->nicEventsNICFirmwareOperational.value.ul = 
	    tgsp->nicEventsNICFirmwareOperational;
	altsp->nicEventsStatsUpdated.value.ul = 
	    tgsp->nicEventsStatsUpdated;
	altsp->nicEventsLinkStateChanged.value.ul = 
	    tgsp->nicEventsLinkStateChanged;
	altsp->nicEventsError.value.ul = 
	    tgsp->nicEventsError;
	altsp->nicEventsMCastListUpdated.value.ul = 
	    tgsp->nicEventsMCastListUpdated;
	altsp->nicEventsResetJumboRing.value.ul = 
	    tgsp->nicEventsResetJumboRing;
	altsp->nicRingSetSendProdIndex.value.ul = 
	    tgsp->nicRingSetSendProdIndex;
	altsp->nicRingSetSendConsIndex.value.ul = 
	    tgsp->nicRingSetSendConsIndex;
	altsp->nicRingSetRecvReturnProdIndex.value.ul = 
	    tgsp->nicRingSetRecvReturnProdIndex;
	altsp->nicInterrupts.value.ul = 
	    tgsp->nicInterrupts;
	altsp->nicAvoidedInterrupts.value.ul = 
	    tgsp->nicAvoidedInterrupts;
	altsp->nicEventThresholdHit.value.ul = 
	    tgsp->nicEventThresholdHit;
	altsp->nicSendThresholdHit.value.ul = 
	    tgsp->nicSendThresholdHit;
	altsp->nicRecvThresholdHit.value.ul = 
	    tgsp->nicRecvThresholdHit;
	altsp->nicDmaRdOverrun.value.ul = 
	    tgsp->nicDmaRdOverrun;
	altsp->nicDmaRdUnderrun.value.ul = 
	    tgsp->nicDmaRdUnderrun;
	altsp->nicDmaWrOverrun.value.ul = 
	    tgsp->nicDmaWrOverrun;
	altsp->nicDmaWrUnderrun.value.ul = 
	    tgsp->nicDmaWrUnderrun;
	altsp->nicDmaWrMasterAborts.value.ul =
	    tgsp->nicDmaWrMasterAborts;
	altsp->nicDmaRdMasterAborts.value.ul =
	    tgsp->nicDmaRdMasterAborts;
	altsp->nicDmaWriteRingFull.value.ul =
	    tgsp->nicDmaWriteRingFull;
	altsp->nicDmaReadRingFull.value.ul =
	    tgsp->nicDmaReadRingFull;
	altsp->nicEventRingFull.value.ul =
	    tgsp->nicEventRingFull;
	altsp->nicEventProducerRingFull.value.ul =
	    tgsp->nicEventProducerRingFull;
	altsp->nicTxMacDescrRingFull.value.ul =
	    tgsp->nicTxMacDescrRingFull;
	altsp->nicOutOfTxBufSpaceFrameRetry.value.ul =
	    tgsp->nicOutOfTxBufSpaceFrameRetry;
	altsp->nicNoMoreWrDMADescriptors.value.ul =
	    tgsp->nicNoMoreWrDMADescriptors;
	altsp->nicNoMoreRxBDs.value.ul =
	    tgsp->nicNoMoreRxBDs;
	altsp->nicNoSpaceInReturnRing.value.ul =
	    tgsp->nicNoSpaceInReturnRing;
	altsp->nicSendBDs.value.ul =
	    tgsp->nicSendBDs;
	altsp->nicRecvBDs.value.ul =
	    tgsp->nicRecvBDs;
	altsp->nicJumboRecvBDs.value.ul =
	    tgsp->nicJumboRecvBDs;
	altsp->nicMiniRecvBDs.value.ul =
	    tgsp->nicMiniRecvBDs;
	altsp->nicTotalRecvBDs.value.ul =
	    tgsp->nicTotalRecvBDs;
	altsp->nicTotalSendBDs.value.ul =
	    tgsp->nicTotalSendBDs;
        altsp->nicJumboSpillOver.value.ul =
            tgsp->nicJumboSpillOver;
        altsp->nicSbusHangCleared.value.ul =
            tgsp->nicSbusHangCleared;
        altsp->nicEnqEventDelayed.value.ul =
            tgsp->nicEnqEventDelayed;
	altsp->nicMacRxLateColls.value.ul =
	    tgsp->nicMacRxLateColls;
	altsp->nicMacRxLinkLostDuringPkt.value.ul =
	    tgsp->nicMacRxLinkLostDuringPkt;
	altsp->nicMacRxPhyDecodeErr.value.ul =
	    tgsp->nicMacRxPhyDecodeErr;
	altsp->nicMacRxMacAbort.value.ul =
	    tgsp->nicMacRxMacAbort;
	altsp->nicMacRxTruncNoResources.value.ul =
	    tgsp->nicMacRxTruncNoResources;
	altsp->nicMacRxDropUla.value.ul =
	    tgsp->nicMacRxDropUla;
	altsp->nicMacRxDropMcast.value.ul =
	    tgsp->nicMacRxDropMcast;
	altsp->nicMacRxFlowControl.value.ul =
	    tgsp->nicMacRxFlowControl;
	altsp->nicMacRxDropSpace.value.ul =
	    tgsp->nicMacRxDropSpace;
	altsp->nicMacRxColls.value.ul =
	    tgsp->nicMacRxColls;
	altsp->nicMacRxColls.value.ul =
	    tgsp->nicMacRxColls;
	altsp->nicMacRxTotalAttns.value.ul =
	    tgsp->nicMacRxTotalAttns;
	altsp->nicMacRxLinkAttns.value.ul =
	    tgsp->nicMacRxLinkAttns;
	altsp->nicMacRxSyncAttns.value.ul =
	    tgsp->nicMacRxSyncAttns;
	altsp->nicMacRxConfigAttns.value.ul =
	    tgsp->nicMacRxConfigAttns;
	altsp->nicMacReset.value.ul =
	    tgsp->nicMacReset;
        altsp->nicMacRxBufDescrAttns.value.ul =
            tgsp->nicMacRxBufDescrAttns;
	altsp->nicMacRxBufAttns.value.ul =
	    tgsp->nicMacRxBufAttns;
	altsp->nicMacRxZeroFrameCleanup.value.ul =
	    tgsp->nicMacRxZeroFrameCleanup;
	altsp->nicMacRxOneFrameCleanup.value.ul =
	    tgsp->nicMacRxOneFrameCleanup;
	altsp->nicMacRxMultipleFrameCleanup.value.ul =
	    tgsp->nicMacRxMultipleFrameCleanup;
	altsp->nicMacRxTimerCleanup.value.ul =
	    tgsp->nicMacRxTimerCleanup;
	for (i=0; i<15; i++) {
	    altsp->nicMacTxCollisionHistogram[0].value.ul = 
		tgsp->nicMacTxCollisionHistogram[0];
	}
	altsp->nicMacTxTotalAttns.value.ul =
	    tgsp->nicMacTxTotalAttns;
    }
    return (0);
}

STATIC void
altstatinit(struct altinst *altp, int vlan_id)
{
    kstat_t *ksp;
    struct altstat *altsp;
    int i;
    char str[80];		/* for array init */
    
    if ((ksp = kstat_create(ALTNAME, vlan_id * NALT + altp->instance, 
			    NULL, "net", KSTAT_TYPE_NAMED,
			    sizeof(struct altstat) / sizeof(kstat_named_t),
#if KSTAT_PERSISTENT
			    KSTAT_FLAG_PERSISTENT
#else /* KSTAT_PERSISTENT */
			    0
#endif /* KSTAT_PERSISTENT */
			    )) == NULL) {
	ERROR3("%s%d: kstat_create failed", 
	       altp->initNic->name, altp->instance);
	ALTTRACEE(ckstatf, 0xf2009, altp->instance, 0, 0, 0);
	return;
    }
    altp->alt_ksp[vlan_id] = ksp;
    altsp = (struct altstat *)(ksp->ks_data);
    kstat_named_init(&altsp->ipackets, "ipackets", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ierrors, "ierrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->opackets, "opackets", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->oerrors, "oerrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->collisions, "collisions", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->defers, "defer", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsAlignmentErrors, 
		     "dot3StatsAlignmentErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsFCSErrors,
		     "dot3StatsFCSErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsSingleCollisionFrames,
		     "dot3StatsSingleCollisionFrames", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsMultipleCollisionFrames,
		     "dot3StatsMultipleCollisionFrames", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsSQETestErrors,
		     "dot3StatsSQETestErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsDeferredTransmissions,
		     "dot3StatsDeferredTransmissions", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsLateCollisions,
		     "dot3StatsLateCollisions", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsExcessiveCollisions,
		     "dot3StatsExcessiveCollisions", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsInternalMacTransmitErrors,
		     "dot3StatsInternalMacTransmitErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsCarrierSenseErrors,
		     "dot3StatsCarrierSenseErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsFrameTooLongs,
		     "dot3StatsFrameTooLongs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->dot3StatsInternalMacReceiveErrors,
		     "dot3StatsInternalMacReceiveErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifIndex, "ifIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifType, "ifType", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifMtu, "ifMtu", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifSpeed, "ifSpeed", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifAdminStatus,
		     "ifAdminStatus", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifOperStatus,
		     "ifOperStatus", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifLastChange,
		     "ifLastChange", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifHCInOctets,
		     "ifHCInOctets", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCInUcastPkts,
		     "ifHCInUcastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCInMulticastPkts,
		     "ifHCInMulticastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCInBroadcastPkts,
		     "ifHCInBroadcastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifInDiscards,
		     "ifInDiscards", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifInErrors,
		     "ifInErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifInUnknownProtos,
		     "ifInUnknownProtos", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifHCOutOctets,
		     "ifHCOutOctets", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCOutUcastPkts,
		     "ifHCOutUcastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCOutMulticastPkts,
		     "ifHCOutMulticastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifHCOutBroadcastPkts,
		     "ifHCOutBroadcastPkts", KSTAT_DATA_ULONGLONG);
    kstat_named_init(&altsp->ifOutDiscards,
		     "ifOutDiscards", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifOutErrors,
		     "ifOutErrors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifOutQLen,
		     "ifOutQLen", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifPhysAddress,
		     "ifPhysAddress", KSTAT_DATA_CHAR);
    kstat_named_init(&altsp->ifDescr[0],
		     "ifDescr (part 1)", KSTAT_DATA_CHAR);
    kstat_named_init(&altsp->ifDescr[1],
		     "ifDescr (part 2)", KSTAT_DATA_CHAR);
    kstat_named_init(&altsp->ifLinkUpDownTrapEnable,
		     "ifLinkUpDownTrapEnable", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifHighSpeed,
		     "ifHighSpeed", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifPromiscuousMode,
		     "ifPromiscuousMode", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->ifConnectorPresent,
		     "ifConnectorPresent", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsHostState,
		     "nicCmdsHostState", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsFdrFiltering,
		     "nicCmdsFdrFiltering", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetRecvProdIndex,
		     "nicCmdsSetRecvProdIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsUpdateGencommStats,
		     "nicCmdsUpdateGencommStats", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsResetJumboRing,
		     "nicCmdsResetJumboRing", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsAddMCastAddr,
		     "nicCmdsAddMCastAddr", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsDelMCastAddr,
		     "nicCmdsDelMCastAddr", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetPromiscMode,
		     "nicCmdsSetPromiscMode", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsLinkNegotiate,
		     "nicCmdsLinkNegotiate", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetMACAddr,
		     "nicCmdsSetMACAddr", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsClearProfile,
		     "nicCmdsClearProfile", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetMulticastMode,
		     "nicCmdsSetMulticastMode", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsClearStats,
		     "nicCmdsClearStats", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetRecvJumboProdIndex,
		     "nicCmdsSetRecvJumboProdIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsSetRecvMiniProdIndex,
		     "nicCmdsSetRecvMiniProdIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsRefreshStats,
		     "nicCmdsRefreshStats", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicCmdsUnknown,
		     "nicCmdsUnknown", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsNICFirmwareOperational,
		     "nicEventsNICFirmwareOperational", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsStatsUpdated,
		     "nicEventsStatsUpdated", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsLinkStateChanged,
		     "nicEventsLinkStateChanged", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsError,
		     "nicEventsError", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsMCastListUpdated,
		     "nicEventsMCastListUpdated", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventsResetJumboRing,
		     "nicEventsResetJumboRing", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicRingSetSendProdIndex,
		     "nicRingSetSendProdIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicRingSetSendConsIndex,
		     "nicRingSetSendConsIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicRingSetRecvReturnProdIndex,
		     "nicRingSetRecvReturnProdIndex", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicInterrupts,
		     "nicInterrupts", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicAvoidedInterrupts,
		     "nicAvoidedInterrupts", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventThresholdHit,
		     "nicEventThresholdHit", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicSendThresholdHit,
		     "nicSendThresholdHit", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicRecvThresholdHit,
		     "nicRecvThresholdHit", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaRdOverrun,
		     "nicDmaRdOverrun", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaRdUnderrun,
		     "nicDmaRdUnderrun", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaWrOverrun,
		     "nicDmaWrOverrun", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaWrUnderrun,
		     "nicDmaWrUnderrun", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaWrMasterAborts,
		     "nicDmaWrMasterAborts", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaRdMasterAborts,
		     "nicDmaRdMasterAborts", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaWriteRingFull,
		     "nicDmaWriteRingFull ", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicDmaReadRingFull,
		     "nicDmaReadRingFull", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventRingFull,
		     "nicEventRingFull", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEventProducerRingFull,
		     "nicEventProducerRingFull", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicTxMacDescrRingFull,
		     "nicTxMacDescrRingFull", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicOutOfTxBufSpaceFrameRetry,
		     "nicOutOfTxBufSpaceFrameRetry", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicNoMoreWrDMADescriptors,
		     "nicNoMoreWrDMADescriptors", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicNoMoreRxBDs,
		     "nicNoMoreRxBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicNoSpaceInReturnRing,
		     "nicNoSpaceInReturnRing", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicRecvBDs,
		     "nicRecvBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicSendBDs,
		     "nicSendBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicJumboRecvBDs,
		     "nicJumboRecvBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMiniRecvBDs,
		     "nicMiniRecvBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicTotalRecvBDs,
		     "nicTotalRecvBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicTotalSendBDs,
		     "nicTotalSendBDs", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicJumboSpillOver,
                     "nicJumboSpillOver", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicSbusHangCleared,
                     "nicSbusHangCleared", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicEnqEventDelayed,
                     "nicEnqEventDelayed", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxLateColls,
		     "nicMacRxLateColls", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxLinkLostDuringPkt,
		     "nicMacRxLinkLostDuringPkt", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxPhyDecodeErr,
		     "nicMacRxPhyDecodeErr", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxMacAbort,
		     "nicMacRxMacAbort", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxTruncNoResources,
		     "nicMacRxTruncNoResources", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxDropUla,
		     "nicMacRxDropUla", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxDropMcast,
		     "nicMacRxDropMcast", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxFlowControl,
		     "nicMacRxFlowControl", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxDropSpace,
		     "nicMacRxDropSpace", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxColls,
		     "nicMacRxColls", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxTotalAttns,
		     "nicMacRxTotalAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxLinkAttns,
		     "nicMacRxLinkAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxSyncAttns,
		     "nicMacRxSyncAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxConfigAttns,
		     "nicMacRxConfigAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacReset,
		     "nicMacReset", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxBufDescrAttns,
                     "nicMacRxBufDescrAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxBufAttns,
		     "nicMacRxBufAttns", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxZeroFrameCleanup,
		     "nicMacRxZeroFrameCleanup", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxOneFrameCleanup,
		     "nicMacRxOneFrameCleanup", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxMultipleFrameCleanup,
		     "nicMacRxMultipleFrameCleanup", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxTimerCleanup,
		     "nicMacRxTimerCleanup", KSTAT_DATA_ULONG);
    kstat_named_init(&altsp->nicMacRxDmaCleanup,
                         "nicMacRxDmaCleanup", KSTAT_DATA_ULONG);

    for (i=0; i < 15; i++) {
	sprintf(str, "nicMacTxCollisionHistogram[%d]", i);
	kstat_named_init(&altsp->nicMacTxCollisionHistogram[i],
			 str, KSTAT_DATA_ULONG);
    }
    kstat_named_init(&altsp->nicMacTxTotalAttns,
		     "nicMacTxTotalAttns", KSTAT_DATA_ULONG);
    ksp->ks_update = altstat_kstat_update;
    altp->kstatPrivate[vlan_id].altp = altp;
    altp->kstatPrivate[vlan_id].vlan = vlan_id;
    ksp->ks_private = (void *)&altp->kstatPrivate[vlan_id];
    kstat_install(ksp);
}

/*
 * Handle board interrupts.  
 */
u_int
ALTINTR(caddr_t arg)
{
    struct tg_nic *nic = (struct tg_nic *)arg;
    int rc;
    U32 tmp;

    /* break out on early interrupt */
    if (nic->altp->state == ALT_NOTREADY) {
	ALTTRACE(rupterly, 0x20009, 0, 0, 0, 0);

	/* hit mailbox 0 to clear rupt, in case it's ours */
	OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 0);

	/* but let others claim it, may get error message on console */
	return(DDI_INTR_UNCLAIMED); 
    }
	
    /* keep the interrupt thread from being reentered, 
     * interesting concept */
    ALT_DISABLE(spl, &nic->intrMu);
    
    /* if we've got our own rupt then we can skip a PIO read */
    if (nic->init_flags & ALT_INIT_SHARED_RUPT) {
	if (IN32(nic->handle, nic->base, TG_MISC_HOST_CONTROL) &
	    TG_MHC_RUPT_STATE) {

	    /* either signal a process or do the actual work */
	    ALTDOINTR(nic);
	    
	    /* hit mailbox 0 to clear rupt */
	    OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 0);
	    
	    rc = DDI_INTR_CLAIMED;
	} else 
	    rc = DDI_INTR_UNCLAIMED;
    } else {
	/* either signal a process or do the actual work */
	ALTDOINTR(nic);
	
	/* hit mailbox 0 to clear rupt */
	OUT32(nic->handle, nic->base, TG_MBOX0_LOW, 0);
	
	rc = DDI_INTR_CLAIMED;
    }

    /* we have to make sure a posted write doesn't hang around too long */
    tmp = IN32(nic->handle, nic->base, TG_MBOX0_LOW);

    /* okay, we can take another interrupt */
    ALT_ENABLE(spl, &nic->intrMu);

    return(rc);
}

/* one wonders why this isn't available in the kernel */

/*
 * Convert a string to an unsigned long integer.
 *
 * Ignores `locale' stuff.  Assumes that the upper and lower case
 * alphabets and digits are each contiguous.
 */
unsigned long
strtoul(const char *nptr, char **endptr, int base)
{
    const char *s = nptr;
    unsigned long acc;
    int c;
    unsigned long cutoff;
    int neg = 0, any, cutlim;

    do {
	c = *s++;
    } while (isspace(c));
    if (c == '-') {
	neg = 1;
	c = *s++;
    } else if (c == '+')
	c = *s++;
    if ((base == 0 || base == 16) &&
	c == '0' && (*s == 'x' || *s == 'X')) {
	c = s[1];
	s += 2;
	base = 16;
    }
    if (base == 0)
	base = c == '0' ? 8 : 10;
    cutoff = (unsigned long)ULONG_MAX / (unsigned long)base;
    cutlim = (unsigned long)ULONG_MAX % (unsigned long)base;
    for (acc = 0, any = 0;; c = *s++) {
	if (isdigit(c))
	    c -= '0';
	else if (isalpha(c))
	    c -= isupper(c) ? 'A' - 10 : 'a' - 10;
	else
	    break;
	if (c >= base)
	    break;
	if (any < 0 || acc > cutoff || acc == cutoff && c > cutlim)
	    any = -1;
	else {
	    any = 1;
	    acc *= base;
	    acc += c;
	}
    }
    if (any < 0) {
	acc = ULONG_MAX;
    } else if (neg)
	acc = -acc;
    if (endptr != 0)
	*endptr = (char *)(any ? s - 1 : nptr);
    return (acc);
}

/* this too, sigh */
unsigned char __ctype[] = 
{0x00, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x28, 0x28, 0x28, 
 0x28, 0x28, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x48, 0x10, 0x10, 0x10, 0x10, 0x10,
 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x84, 0x84, 0x84,
 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
 0x10, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
 0x01, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82,
 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02,
 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x10, 0x10, 0x10, 0x10, 0x20, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

/*
 * Extracts the value from the alt parameter array and prints the
 * parameter value. cp points to the required parameter.
 */
int
alt_param_get(queue_t *q, mblk_t *mp, caddr_t cp)
{
    struct alt_param *altpa = (struct alt_param *)cp;

    sprintf((char *)mp->b_wptr, "%ld", altpa->parm_val);
    mp->b_wptr += strlen((char *)mp->b_wptr) + 1;
    return (0);
}

/*
 * Extracts the value from the alt parameter array and prints the
 * parameter value. cp points to the required parameter.
 * For vlan_tag we do special stuff because of addressability issues
 */
int
alt_param_vlan_tag_get(queue_t *q, mblk_t *mp, caddr_t cp)
{
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    U32 vlanParm;

    vlanParm = altp->vlan_id[alt_param_arr[ND_INSTANCE].parm_val/NALT];
    sprintf((char *)mp->b_wptr, "0x%lx", vlanParm);
    mp->b_wptr += strlen((char *)mp->b_wptr) + 1;
    return (0);
}

/*
 * Extracts the value from the global.  Kind of broken, but it eases
 * this whole construction
 */
int
alt_param_tag_id_get(queue_t *q, mblk_t *mp, caddr_t cp)
{
    struct alt_param *altpa = (struct alt_param *)cp;

    sprintf((char *)mp->b_wptr, "0x%lx", altVlanTagId);
    mp->b_wptr += strlen((char *)mp->b_wptr) + 1;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 */
int
alt_param_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    char *ptr;
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    altpa->parm_val = new_value;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * special checks for jumbograms
 */
int
alt_param_jumbo_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp, *altp2;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    ALTTRACE(jumboset, 0x20001, altp, nic, altp->flags, nic->flags);
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    /* don't allow changes if the host stack is active, 
     * we can't deal with those situations */
    if (altp->flags & ALT_HOST_UP)
	return(EINVAL);		/* already enabled, can't disable */


#if USE_EXTENDED_JUMBO
    /* can't allow this on Tigon 1 */
    if (new_value) {
	if (nic->tigonRev == TG_MHC_LSI_GATE_VERS)
	    return(EINVAL);
    }
#endif /* USE_EXTENDED_JUMBO */
    /* check if ALT_JUMBO_ENABLED is set. If yes, then it implies
     * that the user is setting jumbo on a logical interface 
     */
    if (new_value && (nic->flags & ALT_JUMBO_ENABLED)) {
       altp->jumbo[alt_param_arr[ND_INSTANCE].parm_val/NALT] = TRUE;
       altpa->parm_val = new_value;
       return(0);
    }

    if (new_value) {
	if (rc = altInitJumbo(altp))
	    return(rc);
	altp->jumbo[alt_param_arr[ND_INSTANCE].parm_val/NALT] = TRUE;
	nic->flags |= ALT_JUMBO_ENABLED;
	if (altp->state == ALT_FW_RUNNING)
	    FillJumboRecvRing(altp);
	/* and on the redundant card */
	if (altp->flags & ALT_REDUNDANT) {
	    altp2 = altp->attachNic->initAltp;
	    if (rc = altInitJumbo(altp2))
		return(rc);
	    altp2->jumbo[alt_param_arr[ND_INSTANCE].parm_val/NALT] = TRUE;
	    altp2->initNic->flags |= ALT_JUMBO_ENABLED;
	    if (altp2->state == ALT_FW_RUNNING)
		FillJumboRecvRing(altp2);
	}
        /* since jumbo enabled, inform nic about new mtu */
        nic->genComm->gen_com.ifMTU = JUMBOBLOCKSIZE;
        altUpdateGencomStats(altp);
    }
    altpa->parm_val = new_value;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * special checks for VLAN configuration
 */
int
alt_param_vlan_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    ALTTRACE(vlanset, 0x20002, altp, nic, altp->flags, 0);
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    /* don't allow changes if vlans already enabled or the host
     * stack is active, we can't deal with those situations */
    if (new_value == 0 && 
	((nic->flags & ALT_VLANS_ACTIVE) || 
	 (altp->flags & ALT_HOST_UP)))
	return(EINVAL);		/* already enabled, can't disable */

    if (new_value) {
	altp->flags |= ALT_VLANS_ACTIVE;
    }
    altpa->parm_val = new_value;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * special checks for VLAN configuration
 * funky tagging mechanism, VLAN number in high order 16 bits, VLAN_ID
 * in lower 16 bits.  
 */
int
alt_param_vlan_tag_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    ALTTRACE(vlantset, 0x20003, altp, nic, altp->flags, 0);
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    /* don't allow changes if vlans aren't enabled */
    if (!(altp->flags & ALT_VLANS_ACTIVE))
	return(EINVAL);		/* not enabled, can't set this */

    altp->vlan_id[alt_param_arr[ND_INSTANCE].parm_val/NALT] = new_value;
    altp->vlan_translate[new_value] = alt_param_arr[ND_INSTANCE].parm_val/NALT;
    altpa->parm_val = new_value;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * Sets redundancy configuration on or off
 */
int
alt_param_redund_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp, *altp2;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    ALTTRACE(redset, 0x20004, altp, nic, altp->flags, 0);
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    /* if already on or off, leavve it be */
    if (new_value == altpa->parm_val)
	return(0);

    /* check to ensure that the instance being
     * set as the redundant interface is not in operation
     * as an individual non-redundant interface i.e.
     *  already config'ed
     */
    if (altp->flags & ALT_HOST_UP)
        return(EINVAL);

    /* do set up */
    if (new_value) {
	/* get the instance to attach to */
	if (!(altp2 = (struct altinst *)
	      ddi_get_soft_state(altStatep, 
				 altp->initNic->
				 tuneParms[REDUND_VAL].parm_val)))
	    return(EINVAL);

	/* if this one is already set up, bail */
	if (altp2->flags & ALT_REDUNDANT)
	    return(EBUSY);

	/* is the primary instance up ? */
        if (!(altp2->flags & ALT_HOST_UP))
           return(EINVAL);
 
	altAddRedund(altp2, altp);
    } else {
	return(EINVAL);
#if 0				/* can't remove redundancy XXX */
	/* check that we've got the right one */
	if ((altp->flags & (ALT_REDUNDANT | ALT_RED_ATTACH)) !=
	    (ALT_REDUNDANT | ALT_RED_ATTACH) ||
	    altp->attachNic != altp->initNic) {
	    return(EINVAL);
	altRemRedund(altp);
#endif /* 0 */
    }

    altpa->parm_val = new_value;
    return (0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * need to send a command if this one changes
 */
int
alt_param_fdr_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);

    ALTTRACE(fdrset, 0x20005, altp, new_value, 0, 0);

    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }
    /* don't do anything if it's the same value */
    if (new_value == altpa->parm_val)
	return(0);

    /* ok, update the data base and send the command to the NIC */
    altpa->parm_val = new_value;

    altSetFdrMode(altp, new_value ? TRUE : FALSE);
    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * this also sets a global for fast and easy access throughout the code
 */
int
alt_param_tag_id_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(tagidset, 0x20006, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    /* ok, update the data base and the global */
    altpa->parm_val = new_value;
    altVlanTagId = new_value;

    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * this also manipulates the link_negotiation value and renegotiates the link
 */
int
alt_param_link_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(linknset, 0x20007, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    /* don't do anything if nothing has changed */
    if (new_value == altpa->parm_val)
	return(0);

    /* ok, update the data base */
    altpa->parm_val = new_value;

    /* now redo link negotiation */
    if (new_value)
	altp->nic->genComm->gen_com.tuneParms.link_negotiation |= 
	    LNK_NEGOTIATE;
    else
	altp->nic->genComm->gen_com.tuneParms.link_negotiation &= 
	    ~LNK_NEGOTIATE;

    /* tell nic */
    altDoLinkNegotiation(altp);

    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * this also manipulates the link_negotiation value and renegotiates the link
 */
int
alt_param_rx_flow_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(rxfloset, 0x2000a, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    /* don't do anything if nothing has changed */
    if (new_value == altpa->parm_val)
	return(0);

    /* ok, update the data base */
    altpa->parm_val = new_value;

    /* now redo link negotiation */
    if (new_value)
	altp->nic->genComm->gen_com.tuneParms.link_negotiation |= 
	    LNK_RX_FLOW_CTL_Y;
    else
	altp->nic->genComm->gen_com.tuneParms.link_negotiation &= 
	    ~LNK_RX_FLOW_CTL_Y;

    /* tell nic */
    altDoLinkNegotiation(altp);

    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * this also manipulates the link_negotiation value and renegotiates the link
 */
int
alt_param_tx_flow_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->nic;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(txfloset, 0x2000b, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    /* don't do anything if nothing has changed */
    if (new_value == altpa->parm_val)
	return(0);

    /* now redo link negotiation */
    if (new_value) {
	/* not legal on a Rev 4 Tigon */
	if ((INCFG32(nic->configHandle, nic->configBase, 
		     TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) ==
	    TG_MHC_LSI_GATE_VERS) 
	    return(EINVAL);

	nic->genComm->gen_com.tuneParms.link_negotiation |=  LNK_TX_FLOW_CTL_Y;
    } else
	nic->genComm->gen_com.tuneParms.link_negotiation &=  ~LNK_TX_FLOW_CTL_Y;

    /* ok, update the data base */
    altpa->parm_val = new_value;

    /* tell nic */
    altDoLinkNegotiation(altp);

    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * this also manipulates the nic structure for hybrid operational changes
 */
int
alt_param_hybrid_len_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->nic;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(hybridlen, 0x20008, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    /* don't do anything if nothing has changed */
    if (new_value == altpa->parm_val)
	return(0);

    /* set the value */
    nic->hybridLen = new_value;

    /* ok, update the data base */
    altpa->parm_val = new_value;

    return(0);
}

/*
 * Sets the alt parameter to the value in the alt_param
 * special checks for shared interrupt
 */
int
alt_param_share_rupt_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    ALTTRACE(ruptshar, 0x2000c, altp, nic, nic->init_flags, 0);
    
    new_value = strtoul(value, &ptr, 0);
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    if (new_value) {
	nic->init_flags |= ALT_INIT_SHARED_RUPT;
    } else {
	nic->init_flags &= ~ALT_INIT_SHARED_RUPT;
    }
    altpa->parm_val = new_value;
    return (0);
}

STATIC bool
alt_param_register(struct tg_nic *nic)
{
    int i;
    int rc;
    
    /* all elements are read and write */
    for (i = 0; i < NUM_TUNE_PARMS; i++) {
	if (!nd_load(&nic->alt_nd_ptr, nic->tuneParms[i].parm_name,
		     nic->tuneParms[i].get_func, nic->tuneParms[i].set_func,
		     (caddr_t)&nic->tuneParms[i])) {
	    (void) nd_free(&nic->alt_nd_ptr);
	    return(FALSE);
	}
    }
    return(TRUE);
}

int
alt_param_snap_set(queue_t *q, mblk_t *mp, char *value, caddr_t cp)
{
    long new_value;
    struct alt_param *altpa = (struct alt_param *)cp;
    struct altinst *altp = altpa->altp;
    struct tg_nic *nic = altp->initNic;
    int rc;
    char *ptr;

    new_value = strtoul(value, &ptr, 0);
    ALTTRACE(snapset, 0x20006, altp, new_value, 0, 0);
    
    if (new_value < altpa->parm_min ||
	new_value > altpa->parm_max ||
	ptr == value) {
	return (EINVAL);
    }

    if (new_value) {
	nic->flags |= ALT_SNAP;
    } else {
	nic->flags &= ~ALT_SNAP;
    }
    /* ok, update the data base */
    altpa->parm_val = new_value;

    return(0);
}


