/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: althdr.h,v $
 * Revision 1.4.6.31  1998/10/01  20:27:58  ted
 * 	Add new NIC stats into Solaris and ODI local stats
 * 	[1998/10/01  20:27:31  ted]
 *
 * Revision 1.4.6.30  1998/09/09  02:48:35  hayes
 * 	fix host based transmit flow control
 * 	[1998/09/09  02:47:29  hayes]
 * 
 * Revision 1.4.6.29  1998/09/03  23:25:41  hayes
 * 	change FLOW_CONTROLLED flag to flow_controlled bool
 * 	[1998/09/03  23:25:28  hayes]
 * 
 * Revision 1.4.6.28  1998/06/12  23:48:02  ted
 * 	Rework NT data structures and peripheral changes
 * 	[1998/06/12  23:42:53  ted]
 * 
 * Revision 1.4.6.27  1998/04/14  21:42:50  ted
 * 	New 12.0 stat
 * 	[1998/04/14  21:37:54  ted]
 * 
 * Revision 1.4.6.26  1998/02/23  00:59:46  ted
 * 	Add new reset jumbo rings stats
 * 	[1998/02/23  00:58:59  ted]
 * 
 * Revision 1.4.6.25  1997/12/03  01:23:19  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:43  ted]
 * 
 * Revision 1.4.6.24  1997/11/25  22:07:36  ted
 * 	Fix up statistics
 * 	[1997/11/25  22:02:50  ted]
 * 
 * Revision 1.4.6.23  1997/11/19  19:24:00  ted
 * 	Support Tigon and Tigon2 in the same driver
 * 	[1997/11/19  19:21:13  ted]
 * 
 * Revision 1.4.6.22  1997/11/14  19:17:27  hayes
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:42  hayes]
 * 
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:17  hayes]
 * 
 * Revision 1.4.6.21  1997/10/21  00:13:32  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  23:01:25  ted]
 * 
 * Revision 1.4.20.1  1997/10/16  17:32:13  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:28:46  ted]
 * 
 * Revision 1.4.6.20  1997/09/16  21:47:13  ted
 * 	Change from MIB-II V1 (RFC 1213) to MIB-II V2 (RFC 1573)
 * 	[1997/09/16  21:43:26  ted]
 * 
 * Revision 1.4.6.19  1997/09/16  03:12:07  ted
 * 	Per VLAN statistics
 * 	[1997/09/16  03:08:14  ted]
 * 
 * Revision 1.4.6.18  1997/09/11  18:06:40  ted
 * 	Mixing vlans, jumbos, and redundancy works, plus small fixes from release 1.1 and better NIC tracing
 * 	[1997/09/11  18:01:36  ted]
 * 
 * Revision 1.4.6.17  1997/08/23  17:40:38  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:39:09  skapur]
 * 
 * Revision 1.4.6.16  1997/07/29  03:41:19  ted
 * 	Add OP1 LED support
 * 	[1997/07/29  03:40:31  ted]
 * 
 * Revision 1.4.6.15  1997/07/29  01:13:50  ted
 * 	Merge by hand due to bsubmit core dump, sigh
 * 	[1997/07/29  01:13:14  ted]
 * 
 * Revision 1.4.6.14  1997/07/28  22:40:44  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:40:41  ted]
 * 
 * Revision 1.4.6.13  1997/07/25  22:59:57  ted
 * 	Architectural performance change
 * 	[1997/07/24  21:45:58  ted]
 * 
 * 	Support flow control for STREAMS
 * 	[1997/07/19  20:37:47  ted]
 * 
 * Revision 1.4.6.12  1997/05/23  21:31:30  ted
 * 	Remove references to nicCmdsGetStats, which no longer exists.
 * 	[1997/05/23  21:31:17  ted]
 * 
 * Revision 1.4.6.11  1997/05/21  22:59:57  ted
 * 	Minor statistics tweaks
 * 	[1997/05/21  22:59:27  ted]
 * 
 * Revision 1.4.6.10  1997/05/21  03:36:17  ted
 * 	Fix bug 218 (kstats not all initialized properly)
 * 	[1997/05/21  03:30:19  ted]
 * 
 * Revision 1.4.6.9  1997/05/03  05:00:37  ted
 * 	Fix up statistics
 * 	[1997/05/03  04:57:10  ted]
 * 
 * Revision 1.4.6.8  1997/03/28  01:41:17  ted
 * 	Major merge of NDIS and ODI with mainstream Unix ports
 * 	[1997/03/28  01:38:25  ted]
 * 
 * Revision 1.4.6.7  1997/03/17  08:48:02  skapur
 * 	Turn the FIRMWARE_DEBUGGED flag on so Sun doesn't have to run gdb.
 * 	[1997/03/17  08:40:53  skapur]
 * 
 * Revision 1.4.6.6  1997/03/03  05:17:19  ted
 * 	endian fixes, non-copydata mode fixes
 * 	[1997/03/03  05:16:38  ted]
 * 
 * Revision 1.4.6.5  1997/03/02  18:36:59  ted
 * 	redundancy support and a few bug fixes
 * 	[1997/03/02  18:26:05  ted]
 * 
 * Revision 1.4.6.4  1997/02/21  02:06:57  ted
 * 	Prepare data structures and code structures to support redundant operation
 * 	[1997/02/21  01:57:22  ted]
 * 
 * Revision 1.4.6.3  1997/02/15  04:41:22  ted
 * 	Generify lock
 * 	[1997/02/15  04:37:26  ted]
 * 
 * Revision 1.4.6.2  1997/02/11  04:08:17  ted
 * 	Firmware isn't debugged for now.
 * 	[1997/02/11  04:04:47  ted]
 * 
 * Revision 1.4.6.1  1997/02/07  21:58:40  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:50:26  ted]
 * 
 * Revision 1.4.2.1  1996/11/01  18:25:44  ted
 * 	change rr_ to tg_ and put in proper copyright
 * 	[1996/10/30  23:04:17  ted]
 * 
 * $EndLog$
 */
/*
 * FILE althdr.h
 *
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/streams/common/althdr.h,v $
 * $Revision: 1.4.6.31 $ $Locker:  $
 * $Date: 1998/10/01 20:27:58 $ $Author: ted $ $State: Exp $
 */

#ifndef _ALTHDR_H_
#define _ALTHDR_H_

/* configuration constants */
#define FIRMWARE_DEBUGGED 1	/* set to 1 for production */
#define TG_WINDOW_WORKS 1	/* possibly not on some systems */
#define KSTAT_PERSISTENT 0	/* for persistent kernel stats */

extern void *altStatep;		/* for soft state routines */

/* struct to hold all the data about an Acteon driver instance */
struct altinst {
    dev_info_t *dip;		/* device information structure */
    int instance;		/* map to instance number */
    int state;			/* state of the device */
    int flags;			/* misc flags */
    bool flow_controlled;	/* don't need to touch other bits... */
    U32 align0;			/* make things nicely aligned... */
    struct ether_addr macAddr;	/* mac address */

    queue_t *ipq[MAX_VLANS];	/* fastpath queue, usually IP, one per VLAN */
    U16 vlan_id[MAX_VLANS];	/* VLAN ID */
    bool jumbo[MAX_VLANS];	/* per VLAN jumbo indication */
    U32 inPackets[MAX_VLANS];	/* per VLAN input statistics */
    U32 outPackets[MAX_VLANS];	/* per VLAN output statistics */
#if defined(__sun)
    kstat_t *alt_ksp[MAX_VLANS]; /* kstat pointer */
    struct kstatPrivStruct {
	struct altinst *altp;
	int vlan;
    } kstatPrivate[MAX_VLANS]; /* to keep track of kstat info per VLAN */
#endif /* defined(__sun) */
    int timeoutId;		/* timeout id for cancelling it */
    int redTimeoutId;		/* timeout id for cancelling it */

    /* 
     * note on redundancy:
     *
     * The "nic" entry points to the "active" device.  All data coming
     * down the stack will be sent out through the active device.  The
     * ifconfigNic points to the device that was ifconfig'd up.  The attachNic
     * struct points to the device that was configured as a backup slot.
     * Both nic structures point back to THIS altinst, so that when
     * data arrives due to an interrupt it will be sent upstream to the
     * stack through the right instance (the ifconfig'd instance).  The
     * other instance is marked "unavailable" for ifconfiging once it's
     * activated as a backup.  This is through the instance flag bits.
     * BPDUs are picked up on any interface and passed through the "other"
     * interface.  So if a BPDU comes in on ifconfigNic, it goes out 
     * on attachNic, and vice versa.
     *
     * BPDUs are not modified as they pass through, but they are understood.
     * In particular, the system must make the decision to switch or not
     * based on the "best" BPDU.  This is done by keeping track of the last
     * BPDU seen on an interface and also aging out BPDUs based on the 
     * maxAge value and a 1/4 second timer that advances the current age.
     * When the current age exceeds the maxAge the BPDU saved for an 
     * interface returns to the "worst case" BPDU.
     *
     * The main key to the whole thing is keeping nic pointing to the
     * right piece of hardware.  Pretty much everything just works itself
     * out.  This is done by noting the state of the BPDU frames and when
     * one link goes out of blocking into forwarding state we mark that
     * as the active link.
     *
     * When/if the interface is ifconfig'd down the altNic structure is
     * removed from this instance and the other instance reverts to an
     * unifconfig, non-redundant state.
     */
    struct tg_nic *nic;		/* pointer to nic specific structures */
    struct tg_nic *initNic;	/* pointer to initialized nic */
    struct tg_nic *ifconfigNic;	/* pointer to ifonfiged nic specific data */
    struct tg_nic *attachNic;	/* ptr to "configured redund" nic spec data */
#if ALT_LOCKS_SUPPORTED
    LOCK_TYPE nicMu;		/* flag bits mutex */
#endif /* ALT_LOCKS_SUPPORTED */

    /* multicast information */
    u_int mccount;		/* # enabled multicast addrs */
    struct inst_multicast {
	struct ether_addr addr;	/* multicast address */
	int count;		/* count of streams enabling it */
    } mctab[MAX_MCAST_ENTRIES];	/* table of multicast addrs */

    /* local statistics */
    U32 allocbfail;		/* couldn't allocate a msg */
    U32 nocanput;		/* couldn't pass a message along */
    U32 promiscDrops;		/* dropped promisc packets upstream */
    U32 infast;			/* # pkts handed up fast path */
    U32 inslow;			/* # pkts handed up normal path */
#if READ_SERVICE
    U32 flowControl;		/* # pkts flow controlled in service routine */
#endif /* READ_SERVICE */
    U32 badpkts;		/* dropped due to header mismatch */
    S8 vlan_translate[MAX_VLAN_IDS]; /* table for reverse trans vlan to ipq */
};

/* states (struct altinst) */
#define ALT_NOTREADY	0
#define ALT_FW_LOADED   1
#define ALT_FW_RUNNING  2

/* flag bits (struct altinst) */
#define ALT_HOST_UP      1	/* for when the host stack is enabled */
#define ALT_PROMISC      2	/* set when in physical promisc mode */
#define ALT_REDUNDANT    4	/* involved in a redundant setup */
#define ALT_RED_IFCONFIG 8	/* nic was ifconfiged for redundancy */
#define ALT_RED_ATTACH   0x10	/* nic was "configured redundant" */
#define ALT_DRIVER_IS_OP1 0x40	/* special flag for vge driver */
#define ALT_VLANS_ACTIVE 0x80	/* if VLANS are on on this interface */

/* state buffer for multiaction ioctls */
struct ioc_state {
    int state;
    caddr_t passedAddr;		/* addr of ioctl structure in user space */
    union {
	struct alt_stat stat;	/* for ALT_STAT ioctl */
	struct tg_mem tgMem;	/* for ALT_{READ,WRITE}_TG_MEM */
	struct tg_reg tgReg;	/* for ALT_{READ,WRITE}_TG_REG */
	struct tg_fw tgFw;	/* for ALT_CACHE_FIRMWARE */
	struct tg_core tgCore;	/* for ALT_DUMP_FIRMWARE */
	struct tg_stats tgStats; /* for ALT_GET_NIC_STATS */
    } un;
};
	    
/* states for ioctl messages */
#define STAT_GETSTRUCT  1	/* getting the alt_stat structure */
#define STAT_SENDINST   2	/* sending the altinst structure */
#define STAT_SENDNIC    3	/* sending the tg_nic structure */
#define STAT_SENDSTR    4	/* sending the altstr structure */

#define TG_GETSTRUCT    1	/* getting the tg_{mem,reg} structure */
#define TG_DONE         2	/* sent back/requested necessary info */

#define FW_GETSTRUCT    1	/* getting the tg_fw structure */
#define FW_TEXT         2	/* getting the text segment */
#define FW_DATA         3	/* getting the data segment */
#define FW_RODATA       4	/* getting the rodata segment */

/* Acteon dl address includes dl_sap */
struct alt_dladdr {
    struct ether_addr dl_phys;
    U16 dl_sap;
};

#define ALTADDRL (sizeof(struct alt_dladdr))
#define ALTHEADROOM 34		/* from AEHEADROOM, analyze this XXX */

/* minimal for normal kstats mechanism, get the rest from ioctl */
struct altstat {
    struct kstat_named ipackets;
    struct kstat_named ierrors;
    struct kstat_named opackets;
    struct kstat_named oerrors;
    struct kstat_named collisions;
    struct kstat_named defers;
    struct kstat_named dot3StatsAlignmentErrors;
    struct kstat_named dot3StatsFCSErrors;
    struct kstat_named dot3StatsSingleCollisionFrames;
    struct kstat_named dot3StatsMultipleCollisionFrames;
    struct kstat_named dot3StatsSQETestErrors;
    struct kstat_named dot3StatsDeferredTransmissions;
    struct kstat_named dot3StatsLateCollisions;
    struct kstat_named dot3StatsExcessiveCollisions;
    struct kstat_named dot3StatsInternalMacTransmitErrors;
    struct kstat_named dot3StatsCarrierSenseErrors;
    struct kstat_named dot3StatsFrameTooLongs;
    struct kstat_named dot3StatsInternalMacReceiveErrors;
    struct kstat_named ifIndex;
    struct kstat_named ifType;
    struct kstat_named ifMtu;
    struct kstat_named ifSpeed;
    struct kstat_named ifAdminStatus;
    struct kstat_named ifOperStatus;
    struct kstat_named ifLastChange;
    struct kstat_named ifHCInOctets;
    struct kstat_named ifHCInUcastPkts;
    struct kstat_named ifHCInMulticastPkts;
    struct kstat_named ifHCInBroadcastPkts;
    struct kstat_named ifInDiscards;
    struct kstat_named ifInErrors;
    struct kstat_named ifInUnknownProtos;
    struct kstat_named ifHCOutOctets;
    struct kstat_named ifHCOutUcastPkts;
    struct kstat_named ifHCOutMulticastPkts;
    struct kstat_named ifHCOutBroadcastPkts;
    struct kstat_named ifOutDiscards;
    struct kstat_named ifOutErrors;
    struct kstat_named ifOutQLen;
    struct kstat_named ifPhysAddress;
    struct kstat_named ifDescr[2];
    struct kstat_named ifLinkUpDownTrapEnable;
    struct kstat_named ifHighSpeed;
    struct kstat_named ifPromiscuousMode;
    struct kstat_named ifConnectorPresent;
    struct kstat_named nicCmdsHostState;
    struct kstat_named nicCmdsFdrFiltering;
    struct kstat_named nicCmdsSetRecvProdIndex;
    struct kstat_named nicCmdsUpdateGencommStats;
    struct kstat_named nicCmdsResetJumboRing;
    struct kstat_named nicCmdsAddMCastAddr;
    struct kstat_named nicCmdsDelMCastAddr;
    struct kstat_named nicCmdsSetPromiscMode;
    struct kstat_named nicCmdsLinkNegotiate;
    struct kstat_named nicCmdsSetMACAddr;
    struct kstat_named nicCmdsClearProfile;
    struct kstat_named nicCmdsSetMulticastMode;
    struct kstat_named nicCmdsClearStats;
    struct kstat_named nicCmdsSetRecvJumboProdIndex;
    struct kstat_named nicCmdsSetRecvMiniProdIndex;
    struct kstat_named nicCmdsRefreshStats;
    struct kstat_named nicCmdsUnknown;
    struct kstat_named nicEventsNICFirmwareOperational;
    struct kstat_named nicEventsStatsUpdated;
    struct kstat_named nicEventsLinkStateChanged;
    struct kstat_named nicEventsError;
    struct kstat_named nicEventsMCastListUpdated;
    struct kstat_named nicEventsResetJumboRing;
    struct kstat_named nicRingSetSendProdIndex;
    struct kstat_named nicRingSetSendConsIndex;
    struct kstat_named nicRingSetRecvReturnProdIndex;
    struct kstat_named nicInterrupts;
    struct kstat_named nicAvoidedInterrupts;
    struct kstat_named nicEventThresholdHit;
    struct kstat_named nicSendThresholdHit;
    struct kstat_named nicRecvThresholdHit;
    struct kstat_named nicDmaRdOverrun;
    struct kstat_named nicDmaRdUnderrun;
    struct kstat_named nicDmaWrOverrun;
    struct kstat_named nicDmaWrUnderrun;
    struct kstat_named nicDmaWrMasterAborts;
    struct kstat_named nicDmaRdMasterAborts;
    struct kstat_named nicDmaWriteRingFull;
    struct kstat_named nicDmaReadRingFull;
    struct kstat_named nicEventRingFull;
    struct kstat_named nicEventProducerRingFull;
    struct kstat_named nicTxMacDescrRingFull;
    struct kstat_named nicOutOfTxBufSpaceFrameRetry;
    struct kstat_named nicNoMoreWrDMADescriptors;
    struct kstat_named nicNoMoreRxBDs;
    struct kstat_named nicNoSpaceInReturnRing;
    struct kstat_named nicSendBDs;
    struct kstat_named nicRecvBDs;
    struct kstat_named nicJumboRecvBDs;
    struct kstat_named nicMiniRecvBDs;
    struct kstat_named nicTotalRecvBDs;
    struct kstat_named nicTotalSendBDs;
    struct kstat_named nicJumboSpillOver;
    struct kstat_named nicSbusHangCleared;
    struct kstat_named nicEnqEventDelayed;
    struct kstat_named nicMacRxLateColls;
    struct kstat_named nicMacRxLinkLostDuringPkt;
    struct kstat_named nicMacRxPhyDecodeErr;
    struct kstat_named nicMacRxMacAbort;
    struct kstat_named nicMacRxTruncNoResources;
    struct kstat_named nicMacRxDropUla;
    struct kstat_named nicMacRxDropMcast;
    struct kstat_named nicMacRxFlowControl;
    struct kstat_named nicMacRxDropSpace;
    struct kstat_named nicMacRxColls;
    struct kstat_named nicMacRxTotalAttns;
    struct kstat_named nicMacRxLinkAttns;
    struct kstat_named nicMacRxSyncAttns;
    struct kstat_named nicMacRxConfigAttns;
    struct kstat_named nicMacReset;
    struct kstat_named nicMacRxBufDescrAttns;
    struct kstat_named nicMacRxBufAttns;
    struct kstat_named nicMacRxZeroFrameCleanup;
    struct kstat_named nicMacRxOneFrameCleanup;
    struct kstat_named nicMacRxMultipleFrameCleanup;
    struct kstat_named nicMacRxTimerCleanup;
    struct kstat_named nicMacRxDmaCleanup;
    struct kstat_named nicMacTxCollisionHistogram[15];
    struct kstat_named nicMacTxTotalAttns;
};

/* prototypes */
extern int ALTOPEN(queue_t *, dev_t *, int, int, cred_t *);
extern int ALTCLOSE(queue_t *);
extern u_int ALTINTR(caddr_t);
extern int altwput(queue_t *, mblk_t *);
extern int altwsrv(queue_t *);
extern void altenablewq(struct altinst *);
#if READ_SERVICE
extern int altrsrv(queue_t *);
#endif /* READ_SERVICE */

#endif /* _ALTHDR_H_ */

