/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altdb.c,v $
 * Revision 1.2.5.8  1998/06/08  23:56:44  ted
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:47  ted]
 *
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:09:07  ted]
 *
 * Revision 1.2.5.7  1997/07/28  22:40:55  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:41:03  ted]
 * 
 * Revision 1.2.5.6  1997/06/03  18:55:19  ted
 * 	Changes to support Sun naming of drivers (vge vs. alt)
 * 	[1997/06/03  18:54:25  ted]
 * 
 * Revision 1.2.5.5  1997/05/06  00:07:51  suresh
 * 	AIX: change default dev to /dev/ent1 from /dev/alt0
 * 	[1997/05/06  00:03:33  suresh]
 * 
 * Revision 1.2.5.4  1997/02/21  19:09:38  suresh
 * 	include file changes to compile on AIX
 * 	[1997/02/21  19:08:54  suresh]
 * 
 * Revision 1.2.5.3  1997/02/11  00:02:32  ted
 * 	Finish merging
 * 	[1997/02/11  00:02:19  ted]
 * 
 * Revision 1.2.5.2  1997/02/10  22:11:05  suresh
 * 	Merged with changes from 1.2.5.1
 * 	[1997/02/10  22:10:52  suresh]
 * 
 * 	added multiple adapter support for completeness
 * 	[1997/02/10  22:07:40  suresh]
 * 
 * Revision 1.2.5.1  1997/02/07  21:58:48  ted
 * 	     Updates for AIX/NT/new Solaris options (fast DMA)
 * 	     [1997/02/07  21:50:48  ted]
 * 
 * Revision 1.2.5.1  1997/02/07  21:58:48  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:50:48  ted]
 * 
 * Revision 1.2.2.1  1996/11/01  18:25:56  ted
 * 	changed rr_ to tg_ and put in proper copyrights
 * 	[1996/10/30  23:07:52  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altdb.c
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altdb.c,v $
 * $Revision: 1.2.5.8 $ $Locker:  $
 * $Date: 1998/06/08 23:56:44 $ $Author: ted $ $State: Exp $
 */
 
#ifndef __lint
static const char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altdb.c,v 1.2.5.8 1998/06/08 23:56:44 ted Exp $";
#endif /* __lint */
 
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/errno.h>

/* streams specific include files */
#if defined(ALT_STREAM)
#include <sys/stream.h>
#endif /* defined(ALT_STREAM) */

/* system specific files */
#if defined(__sun)
#include <sys/ioccom.h>
typedef void *ddi_acc_handle_t; /* to keep prototypes from bitching */
#include <sys/dditypes.h>
#include <sys/ethernet.h>
#endif /* defined(__sun) */

#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#include "altioctl.h"

static void usage(void);

/* Stuff for getopt()   */
extern char *optarg;
extern int optind, opterr;

#if defined(ALT_STREAM)
#define ALTDEV "/dev/alt"
#else /* defined(ALT_STREAM) */
#define ALTDEV "/dev/alt0"
#endif /* defined(ALT_STREAM) */

int
main(int argc, char *argv[])
{
    int fd;
    int rc;
    int set = 0;
    int trace;
    char device[80] = ALTDEV;
    char errstr[80];
    int c;
    int errs = 0;

    opterr = 0 ;  /* keep it quiet */
    trace = -1;

    while ((c = getopt(argc, argv, "d:")) != EOF) {
	switch (c) {
	case 'd': 
	    if (strlen(optarg) > 80)
		errs++;
	    else
		strcpy(device, optarg);
	    break;

	default: 
	    (void) printf("Got unknown command line argument %c\n", c);
	    errs++;
	    break;
	}
    }
    if (errs) {
	usage();
	exit(1);
    }

    /* if we have the trace arg */
    if (argc == optind+1) {
	/* cursory glance */
	if (!isdigit(argv[optind][0])) {
	    (void) printf("trace argument is not a valid integer\n");
	    usage();
 	    exit(1);
	}
	trace = atoi(argv[optind]);
	set = 1;
    }

    if (trace > 4) {
	(void) printf("invalid tracing level %d\n", trace);
	usage();
	exit(1);
    }

    fd = open(device, O_RDWR);
    if (fd < 0) {
	sprintf(errstr, "open failed for device %s", device);
	perror(errstr);
	exit(1);
    }

    if (set) {
	rc = ioctl(fd, ALT_SETTRACE, trace);
	if (rc < 0) {
	    perror("ioctl failed");
	    exit(1);
	}
    }

    rc = ioctl(fd, ALT_READTRACE, &trace);
    if (rc < 0) {
	perror("ioctl failed");
	exit(1);
    }
    (void) printf("tracing %d\n", trace);

    rc = close(fd);
    if (rc < 0) {
	perror("close failed");
	exit(1);
    }
    return(0);
}


static void
usage( void )
{
    (void) printf("altdb [-d <device_name>] [<trace-level>]\n");
}








