/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altfw.c,v $
 * Revision 1.2.4.5  1998/06/08  23:56:45  ted
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:48  ted]
 *
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:09:08  ted]
 *
 * Revision 1.2.4.4  1997/07/28  22:40:56  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:41:07  ted]
 * 
 * Revision 1.2.4.3  1997/05/21  17:24:14  suresh
 * 	fix build-depend problems on AIX
 * 	[1997/05/21  17:23:12  suresh]
 * 
 * Revision 1.2.4.2  1997/02/10  22:11:07  suresh
 * 	added multiple adapter support for completeness
 * 	[1997/02/10  22:07:36  suresh]
 * 
 * Revision 1.2.4.1  1997/02/07  21:58:49  ted
 * 	     Updates for AIX/NT/new Solaris options (fast DMA)
 * 	     [1997/02/07  21:50:51  ted]
 * 
 * Revision 1.2.2.2  1996/11/09  02:32:19  ted
 * 	     clean up compiler warnings
 * 	     [1996/11/09  02:30:47  ted]
 * 
 * Revision 1.2.2.1  1996/11/01  18:25:57  ted
 * 	     changed rr_ to tg_ and put in proper copyrights
 * 	     [1996/10/30  23:07:55  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altfw.c
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altfw.c,v $
 * $Revision: 1.2.4.5 $ $Locker:  $
 * $Date: 1998/06/08 23:56:45 $ $Author: ted $ $State: Exp $
 */
 
#ifndef __lint
static const char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altfw.c,v 1.2.4.5 1998/06/08 23:56:45 ted Exp $";
#endif /* __lint */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <elf.h>
#include <sys/byteorder.h>
#include <sys/errno.h>
#if defined(__sun)
#include <sys/core.h>
#endif /* defined(__sun) */
#if defined(__sgi) 
#include <core.out.h>
#endif /* defined(__sgi) */

/* streams specific include files */
#if defined(ALT_STREAM)
#include <sys/stream.h>
#endif /* defined(ALT_STREAM) */

/* system specific files */
#if defined(__sun)
#include <sys/ioccom.h>
typedef void *ddi_acc_handle_t; /* to keep prototypes from bitching */
#include <sys/dditypes.h>
#include <sys/ethernet.h>
#endif /* defined(__sun) */

#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "altftns.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#include "altioctl.h"
#include "altreg.h"

static char text[MAX_TEXT_LEN], data[MAX_DATA_LEN], rodata[MAX_RODATA_LEN];
#if defined(ALT_STREAM)
static char usage[] = "%s [-l] [-c <elf_file>] [-x <core_file>] [-i <alt_instance>]\n";
static char altDefaultDev[] = "/dev/alt";
#define OPTARGS  "lc:x:i:"
#else /* defined(ALT_STREAM) */
static char usage[] = "%s [-l] [-c <elf_file>] [-x <core_file>] [-d <alt_device>]\n";
static char altDefaultDev[] = "/dev/alt0";
#define OPTARGS "lc:x:d:"
#endif /* defined(ALT_STREAM) */

/* Stuff for getopt()   */
extern char *   optarg;
extern int      optint, opterr;

int
main(int argc, char *argv[])
{
    FILE *elfFile; 
    Elf32_Ehdr elfhdr;
    Elf32_Shdr sectHdr, strHdr;
    caddr_t startAddr, bssAddr = 0, dataAddr = 0, textAddr = 0;
    caddr_t rodataAddr = 0, sbssAddr = 0;
    int bssLen = 0, dataLen = 0, textLen = 0;
    int rodataLen = 0, sbssLen = 0;
    int i, curOff, fd, strOff;
    char name[20], *cp;
    int load = 0, errs = 0;
    char *elfName = (char *)0;
    char *coreName = (char *)0;
    char *altName = (char *)0;
    struct tg_fw tgFw;
#if 0				/* core dump */
    FILE *coreFile;
    caddr_t bp;
    struct tg_core tgCore;
#if defined(__sgi)
    struct coreout coreout;
    struct vmap vmap;
#endif /* defined(__sgi) */
#if defined(__sun)
    struct core coreout;
#endif /* defined(__sun) */
    char *tmpBuf;
#endif /* 0 */
    int c;
#if defined(ALT_STREAM)
    int instance = -1;
#endif /* defined(ALT_STREAM) */

    opterr = 0;   /* keep getopt quiet */
    while ((c = getopt(argc, argv, OPTARGS)) != EOF)
	switch (c) {
	  case 'l': load++;                                   break;
	  case 'c': elfName = optarg;                         break;
	  case 'x': coreName = optarg;                        break;
#if defined(ALT_STREAM)
	  case 'i': instance = atoi(optarg);                  break;
#else /* defined(ALT_STREAM) */
	  case 'd': altName = optarg;                         break;
#endif /* defined(ALT_STREAM) */
	  default: errs++;                                    break;
	}

    /* check for parm errors */
    if (errs) {
	(void) printf(usage, argv[0]);
	exit(1);
    }

    /* check for at least one of -c and -l and -x */
    if (!load && !elfName && !coreName) {
	(void) printf("Must specify either -c, -l, or -x\n");
	(void) printf(usage, argv[0]);
	exit(1);
    }

    /* don't do -c or -l with -x */
    if (coreName && (load || elfName)) {
	(void) printf("May not specify -x with -c or -l\n");
	(void) printf(usage, argv[0]);
	exit(1);
    }

    /* check if instance or device entered, else use default */
#if defined(ALT_STREAM)
    if (instance == -1) {
        instance = 0;
        altName = altDefaultDev;
    }
#else /* defined(ALT_STREAM) */
    if (!altName)
        altName = altDefaultDev;
#endif /* defined(ALT_STREAM) */

    /* open the driver device */
    if ((fd = open(altName, O_RDWR)) < 0) {
	perror("open driver device failed");
	exit(1);
    }

#if defined(ALT_STREAM)
    /* for streams we've got to specify which hardware device */
    if (ioctl(fd, ALT_ATTACH, instance) < 0) {
	perror("ioctl ALT_ATTACH failed");
	exit(1);
    }
#endif /* defined(ALT_STREAM) */

    /* if we're putting the data into the driver cache */
    if (elfName) {
	if (!(elfFile = fopen(elfName, "r"))) {
	    (void) printf("Couldn't open file %s\n", elfName);
	    (void) printf(usage, argv[0]);
	    exit(1);
	}
	
	/* if we're supposed to do the cache part read in the file info */
	/* get the elf header */
	if (fread(&elfhdr, sizeof(Elf32_Ehdr), 1, elfFile) != 1) {
	    (void) printf("Couldn't read elf file's Ehdr\n");
	    exit(1);
	}
	
	/* save the start address */
	startAddr = (caddr_t)htonl(elfhdr.e_entry);
	
	/* check the magic number and then find out where the 
	 * program header is 
	 */
	if (elfhdr.e_ident[0] != ELFMAG0 || elfhdr.e_ident[1] != ELFMAG1 ||
	    elfhdr.e_ident[2] != ELFMAG2 || elfhdr.e_ident[3] != ELFMAG3) {
	    (void) printf("File %s is not an ELF file\n", argv[1]);
	    exit(1);
	}
	
	/* find the section header that points to the section header strings */
	if (fseek(elfFile, 
		  htonl(elfhdr.e_shoff) + 
		  (htons(elfhdr.e_shstrndx) * htons(elfhdr.e_shentsize)), 
		  SEEK_SET)) {
	    (void) printf("Couldn't seek to section header string index\n");
	    exit(1);
	}
	if (fread(&strHdr, sizeof(Elf32_Shdr), 1, elfFile) != 1) {
	    (void) printf("Couldn't read section header string index\n");
	    exit(1);
	}
	
	/* offset into the file of the section header strings */
	strOff = htonl(strHdr.sh_offset);
	
	/* search for the text, data and bss section headers */
	if (fseek(elfFile, htonl(elfhdr.e_shoff), SEEK_SET)) {
	    (void) printf("Couldn't fssek to first section header\n");
	    exit(1);
	}
	
	for (i=0; i < htons(elfhdr.e_shnum); i++) {
	    if (fread(&sectHdr, sizeof(Elf32_Shdr), 1, elfFile) != 1) {
		(void) printf("Couldn't read elf file's Shdr\n");
		exit(1);
	    }

	    /* find the section header name */
	    curOff = ftell(elfFile);
	    if (fseek(elfFile, strOff + htonl(sectHdr.sh_name), SEEK_SET)) {
		(void) printf("Couldn't seek to section header file name\n");
		exit(1);
	    }
	    cp = name;
	    while ((*cp++ = fgetc(elfFile)));
	    
	    if (strcmp(name, ".text") == 0) {
		/* read in the text section */
		if (htonl(sectHdr.sh_size) > MAX_TEXT_LEN) {
		    (void) printf("Text segment exceeds maximum size\n");
		    exit(1);
		}
		if (fseek(elfFile, htonl(sectHdr.sh_offset), SEEK_SET)) {
		    (void) printf("Couldn't seek to data section\n");
		    exit(1);
		}
		if (fread(text, sizeof(char), htonl(sectHdr.sh_size), 
			  elfFile) != 
		    htonl(sectHdr.sh_size)) {
		    (void) printf("Couldn't read text section\n");
		    exit(1);
		}
		textAddr = (caddr_t)htonl(sectHdr.sh_addr);
		textLen = htonl(sectHdr.sh_size);
		
	    } else if (strcmp(name, ".data") == 0) {
		/* read in the data section */
		if (htonl(sectHdr.sh_size) > MAX_DATA_LEN) {
		    (void) printf("Data segment exceeds maximum size\n");
		    exit(1);
		}
		if (fseek(elfFile, htonl(sectHdr.sh_offset), SEEK_SET)) {
		    (void) printf("Couldn't seek to data section\n");
		    exit(1);
		}
		if (fread(data, sizeof(char), htonl(sectHdr.sh_size), 
			  elfFile) !=
		    htonl(sectHdr.sh_size)) {
		    (void) printf("Couldn't read data section\n");
		    exit(1);
		}
		dataAddr = (caddr_t)htonl(sectHdr.sh_addr);
		dataLen = htonl(sectHdr.sh_size);
		
	    } else if (strcmp(name, ".bss") == 0) {
		if (htonl(sectHdr.sh_size) > MAX_BSS_LEN) {
		    (void) printf("BSS segment exceeds maximum size\n");
		    exit(1);
		}
		bssAddr = (caddr_t)htonl(sectHdr.sh_addr);
		bssLen = htonl(sectHdr.sh_size);
	    } else if (strcmp(name, ".rodata") == 0) {
		/* read in the rodata section */
		if (htonl(sectHdr.sh_size) > MAX_RODATA_LEN) {
		    (void) printf("Read only data segment exceeds maximum size\n");
		    exit(1);
		}
		if (fseek(elfFile, htonl(sectHdr.sh_offset), SEEK_SET)) {
		    (void) printf("Couldn't seek to data section\n");
		    exit(1);
		}
		if (fread(rodata, sizeof(char), htonl(sectHdr.sh_size), 
			  elfFile) !=
		    htonl(sectHdr.sh_size)) {
		    (void) printf("Couldn't read data section\n");
		    exit(1);
		}
		rodataAddr = (caddr_t)htonl(sectHdr.sh_addr);
		rodataLen = htonl(sectHdr.sh_size);
		
	    } else if (strcmp(name, ".sbss") == 0) {
		if (htonl(sectHdr.sh_size) > MAX_SBSS_LEN) {
		    (void) printf("SBSS segment exceeds maximum size\n");
		    exit(1);
		}
		sbssAddr = (caddr_t)htonl(sectHdr.sh_addr);
		sbssLen = htonl(sectHdr.sh_size);
	    }
	    fseek(elfFile, curOff, SEEK_SET);
	}

	/* download the file, to the kernel cache */
	tgFw.startAddr = (U32)startAddr;
	tgFw.textAddr = (U32)textAddr;
	tgFw.textLen = textLen;
	tgFw.dataAddr = (U32)dataAddr;
	tgFw.dataLen = dataLen;
	tgFw.rodataAddr = (U32)rodataAddr;
	tgFw.rodataLen = rodataLen;
	tgFw.bssAddr = (U32)bssAddr;
	tgFw.bssLen = bssLen;
	tgFw.sbssAddr = (U32)sbssAddr;
	tgFw.sbssLen = sbssLen;
	tgFw.fwText = text;
	tgFw.fwData = data;
	tgFw.fwRodata = rodata;

	if (ioctl(fd, ALT_CACHE_FIRMWARE, &tgFw) < 0) {
	    perror("ALT_CACHE_FIRMWARE ioctl failed");
	    exit(1);
	}
    }

    /* should we load it into the board ? */
    if (load) {
	if (ioctl(fd, ALT_LOAD_FIRMWARE) < 0) {
	    perror("ALT_LOAD_FIRMWARE ioctl failed");
	    exit(1);
	}
    }

    /* should we generate a core file ? */
    if (coreName) {
#if 0
	if (!(coreFile = fopen(coreName, "w"))) {
	    (void) printf("Couldn't create file %s\n", coreName);
	    (void) printf(usage, argv[0]);
	    exit(1);
	}

	/* init the tg_core struct */
	tgCore.gprs = (caddr_t)malloc(sizeof(U32)*32);
	tgCore.fwText = (caddr_t)malloc(MAX_TEXT_LEN);
	tgCore.fwData = (caddr_t)malloc(MAX_DATA_LEN);
	tgCore.fwRodata = (caddr_t)malloc(MAX_RODATA_LEN);
	tgCore.fwBss = (caddr_t)malloc(MAX_BSS_LEN);
	tgCore.fwSbss = (caddr_t)malloc(MAX_SBSS_LEN);
	tgCore.fwStack = (caddr_t)malloc(MAX_STACK_LEN);
	
	/* get the dump info */
	if (ioctl(fd, ALT_DUMP_FIRMWARE, &tgCore) < 0) {
	    perror("ALT_DUMP_FIRMWARE ioctl failed");
	    exit(1);
	}
	/* format the core file */
	coreout.c_magic = CORE_MAGIC;
#if defined(__sgi) 
	coreout.c_version = CORE_VERSION1;
#endif /* defined(__sgi) */
	coreout.c_nvmap = 6;
	memset(coreout.c_name, 0, CORE_NAMESIZE);
	strcpy(coreout.c_name, "fw");
	memset(coreout.c_args, 0, CORE_ARGSIZE);
	coreout.c_sigcause = SIGHUP;
	memset(coreout.c_idesc, 0, sizeof(struct idesc) * CORE_NIDESC);
	coreout.c_idesc[0].i_offset = sizeof(struct coreout);
	coreout.c_idesc[0].i_len = sizeof(U32) * 32;
	coreout.c_idesc[0].i_flags = IVALID;
	coreout.c_idesc[1].i_offset = coreout.c_idesc[0].i_offset +
	    coreout.c_idesc[0].i_len;
	coreout.c_idesc[1].i_len = sizeof(U32) * 32;
	coreout.c_idesc[2].i_flags = IVALID;
	coreout.c_idesc[2].i_offset = coreout.c_idesc[1].i_offset +
	    coreout.c_idesc[1].i_len;
	coreout.c_idesc[2].i_len = sizeof(U32) * 7;
	coreout.c_idesc[2].i_flags = IVALID;
	coreout.c_vmapoffset = coreout.c_idesc[2].i_offset +
	    coreout.c_idesc[2].i_len;

	/* write the coreout structure at start of corefile */
	if (fwrite(&coreout, sizeof(struct coreout), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's coreout structure\n");
	    exit(1);
	}

	/* put in the idesc information */
	if (fwrite(tgCore.gprs, sizeof(char), sizeof(U32) * 32, coreFile) != 
	    sizeof(U32) * 32) {
	    (void) printf("Couldn't write core file's GPRs\n");
	    exit(1);
	}

	/* zeroes for the floating point regs */
	tmpBuf = malloc(sizeof(U32) * 32);
	memset(tmpBuf, 0, sizeof(U32) * 32);
	if (fwrite(tmpBuf, sizeof(char), sizeof(U32) * 32, coreFile) != 
	    sizeof(U32) * 32) {
	    (void) printf("Couldn't write core file's special registers\n");
	    exit(1);
	}

	/* the PC is the only special register we care about */
	if (fwrite(&tgCore.pc, sizeof(char), sizeof(U32), coreFile) != 
	    sizeof(U32)) {
	    (void) printf("Couldn't write core file's PC\n");
	    exit(1);
	}
	
	/* zeroes for the other 6 special registers */
	if (fwrite(tmpBuf, sizeof(char), sizeof(U32) * 6, coreFile) != 
	    sizeof(U32) * 6) {
	    (void) printf("Couldn't write core file's special registers\n");
	    exit(1);
	}
	
	/* put in the vmaps, STACK, TEXT, DATA, RODATA, BSS, SBSS */
	/* STACK */
	vmap.v_vaddr = FW_STACK_LOC;
	vmap.v_len = tgCore.stackLen;
	vmap.v_offset = coreout.c_vmapoffset + sizeof(struct vmap) * 6;
	curOff = vmap.v_offset + vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VSTACK;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's stack vmap structure\n");
	    exit(1);
	}

	/* TEXT */
	vmap.v_vaddr = tgCore.textAddr;
	vmap.v_len = tgCore.textLen;
	vmap.v_offset = curOff;
	curOff += vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VTEXT;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's text vmap structure\n");
	    exit(1);
	}

	/* DATA */
	vmap.v_vaddr = tgCore.dataAddr;
	vmap.v_len = tgCore.dataLen;
	vmap.v_offset = curOff;
	curOff += vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VDATA;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's data vmap structure\n");
	    exit(1);
	}

	/* RODATA */
	vmap.v_vaddr = tgCore.rodataAddr;
	vmap.v_len = tgCore.rodataLen;
	vmap.v_offset = curOff;
	curOff += vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VDATA;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's rodata vmap structure\n");
	    exit(1);
	}

	/* BSS */
	vmap.v_vaddr = tgCore.bssAddr;
	vmap.v_len = tgCore.bssLen;
	vmap.v_offset = curOff;
	curOff += vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VDATA;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's bss vmap structure\n");
	    exit(1);
	}

	/* SBSS */
	vmap.v_vaddr = tgCore.sbssAddr;
	vmap.v_len = tgCore.sbssLen;
	vmap.v_offset = curOff;
	curOff += vmap.v_len;
	vmap.v_flags = VDUMPED;
	vmap.v_type = VDATA;
	
	if (fwrite(&vmap, sizeof(struct vmap), 1, coreFile) != 1) {
	    (void) printf("Couldn't write core file's sbss vmap structure\n");
	    exit(1);
	}

	/* finally write the actual data segments */
	/* STACK */
	if (fwrite(tgCore.fwStack, sizeof(char), tgCore.stackLen, coreFile) != 
	    tgCore.stackLen) {
	    (void) printf("Couldn't write core file's stack data\n");
	    exit(1);
	}

	/* TEXT */
	if (fwrite(tgCore.fwText, sizeof(char), tgCore.textLen, coreFile) != 
	    tgCore.textLen) {
	    (void) printf("Couldn't write core file's text data\n");
	    exit(1);
	}

	/* DATA */
	if (fwrite(tgCore.fwData, sizeof(char), tgCore.dataLen, coreFile) != 
	    tgCore.dataLen) {
	    (void) printf("Couldn't write core file's data area\n");
	    exit(1);
	}

	/* RODATA */
	if (fwrite(tgCore.fwRodata, sizeof(char), tgCore.rodataLen, 
		   coreFile) != tgCore.rodataLen) {
	    (void) printf("Couldn't write core file's rodata data\n");
	    exit(1);
	}

	/* BSS */
	if (fwrite(tgCore.fwBss, sizeof(char), tgCore.bssLen, coreFile) != 
	    tgCore.bssLen) {
	    (void) printf("Couldn't write core file's bss data\n");
	    exit(1);
	}

	/* SBSS */
	if (fwrite(tgCore.fwSbss, sizeof(char), tgCore.sbssLen, coreFile) != 
	    tgCore.sbssLen) {
	    (void) printf("Couldn't write core file's sbss data\n");
	    exit(1);
	}

	if (fclose(coreFile) != 0) {
	    (void) printf("Couldn't close core file\n");
	    exit(1);
	}
#endif /* 0 */
	(void) printf("core dump facility not supported yet\n");
    }
    return(0);
}
	
