/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altioc.c,v $
 * Revision 1.2.4.6  1998/06/08  23:56:46  ted
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:34:50  ted]
 *
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:09:10  ted]
 *
 * Revision 1.2.4.5  1997/08/23  17:40:44  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:39:21  skapur]
 * 
 * Revision 1.2.4.4  1997/07/28  22:40:58  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:41:11  ted]
 * 
 * Revision 1.2.4.3  1997/05/21  03:36:23  ted
 * 	Use pub_tg.h
 * 	[1997/05/21  03:32:24  ted]
 * 
 * Revision 1.2.4.2  1997/02/21  19:09:40  suresh
 * 	include file changes to compile on AIX
 * 	[1997/02/21  19:08:58  suresh]
 * 
 * Revision 1.2.4.1  1997/02/07  21:58:51  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:50:55  ted]
 * 
 * Revision 1.2.2.3  1996/11/20  20:04:54  ted
 * 	Redo include file structuring.
 * 	[1996/11/20  19:59:05  ted]
 * 
 * Revision 1.2.2.2  1996/11/09  02:32:20  ted
 * 	clean up compiler warnings
 * 	[1996/11/09  02:31:01  ted]
 * 
 * Revision 1.2.2.1  1996/11/01  18:25:59  ted
 * 	changed rr_ to tg_ and put in proper copyrights
 * 	[1996/10/30  23:07:59  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altioc.c
 *
 * COPYRIGHT (c) Essential Communication Corp. 1995
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altioc.c,v $
 * $Revision: 1.2.4.6 $ $Locker:  $
 * $Date: 1998/06/08 23:56:46 $ $Author: ted $ $State: Exp $
 */

#ifndef __lint
static const char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/utils/common/altioc.c,v 1.2.4.6 1998/06/08 23:56:46 ted Exp $";
#endif /* __lint */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/errno.h>

/* streams specific include files */
#if defined(ALT_STREAM)
#include <sys/stream.h>
#endif /* defined(ALT_STREAM) */

/* system specific files */
#if defined(__sun)
typedef void cond_t;		/* to keep prototypes from bitching */
typedef void *ddi_acc_handle_t;	/* to keep prototypes from bitching */
#include <sys/dditypes.h>	/* ddi types */
#include <sys/ksynch.h>		/* kconfvar_t */
#include <sys/stropts.h>	/* must precede ddi.h (which defines min) */
#include <sys/ddi.h>		/* getminor */
#include <sys/dlpi.h>		/* DL_UNATTACHED ... */
#include <sys/ioccom.h>
#include <sys/ethernet.h>
#endif /* defined(__sun) */

#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#include "altioctl.h"
#include "pub_tg.h"

#if defined(ALT_STREAM)
#define ALTDEV "/dev/alt"
#define MINARGS 2
#else /* defined(ALT_STREAM) */
#define ALTDEV "/dev/alt0"
#define MINARGS 1
#endif /* defined(ALT_STREAM) */

int
main(int argc, char **argv)
{
    int fd;
    int rc;
    struct tg_reg readData, writeData;
    struct tg_mem memData;
    U32 data;
    char sdata[20], sdata2[20];
    U32 *datap;
    int i;

    fd = open(ALTDEV, O_RDWR);
    if (fd < 0) {
	perror("open failed");
	exit(1);
    }

#if defined(ALT_STREAM)
    /* for streams we've got to specify which hardware device */
    rc = ioctl(fd, ALT_ATTACH, 0);
    if (rc < 0) {
	perror("ioctl ALT_ATTACH failed");
	exit(1);
    }
#endif /* defined(ALT_STREAM) */

    /* first make sure the endian bits are set properly */
    writeData.addr = TG_MISC_HOST_CONTROL;
#if ALT_BIG_ENDIAN
    writeData.data = 
	(TG_MHC_BYTE_SWAP | TG_MHC_WORD_SWAP | TG_MHC_CLEAR_RUPT) |
	    ((TG_MHC_BYTE_SWAP | TG_MHC_WORD_SWAP | TG_MHC_CLEAR_RUPT) << 24);
#else /* ALT_BIG_ENDIAN */
    writeData.data = TG_MHC_CLEAR_RUPT;
#endif /* ALT_BIG_ENDIAN */
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    /* try reading the cpu state/control as a register */
    readData.addr = TG_CPU_CONTROL;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("Reading cpu state/control as register yields %x\n", 
	   (u_int)readData.data);

    /* try reading it as memory */
    memData.tgAddr = TG_BEG_REGS + TG_CPU_CONTROL;
    memData.userAddr = (caddr_t)&data;
    memData.len = 4;
    if (ioctl(fd, ALT_READ_TG_MEM, &memData) < 0) {
	perror("couldn't do ALT_READ_TG_MEM");
	exit(1);
    }

    printf("Reading cpu state/control as memory yields %x\n", (u_int)data);

    /* try writing the cpu state/control as reg (to stop things) */
    writeData.addr = TG_CPU_CONTROL;
    writeData.data |= TG_CPU_HALT;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    /* read it back */
    readData.addr = TG_CPU_CONTROL;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("cpu state/control is now %x\n", (u_int)readData.data);

    /* try writing the prog counter as a reg */
    writeData.addr = TG_PROG_CTR;
    writeData.data = 0x1000;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    /* read it back */
    readData.addr = TG_PROG_CTR;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("prog counter is now %x, expected 0x1000\n", (u_int)readData.data);

    /* try writing prog counter as memory */
    memData.tgAddr = TG_BEG_REGS + TG_PROG_CTR;
    data = 0x1000;
    memData.userAddr = (caddr_t)&data;
    memData.len = 4;
    if (ioctl(fd, ALT_WRITE_TG_MEM, &memData) < 0) {
	perror("couldn't do ALT_WRITE_TG_MEM");
	exit(1);
    }

    /* read it back */
    readData.addr = TG_PROG_CTR;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("prog counter is now %x, expected 0x1000\n", (u_int)readData.data);

    /* try writing some data at 0x1000 */
    memData.tgAddr = 0x1000;
    for (i=0; i < 20; i++)
	sdata[i] = i;
    memData.userAddr = sdata;
    memData.len = 20;
    if (ioctl(fd, ALT_WRITE_TG_MEM, &memData) < 0) {
	perror("couldn't do ALT_WRITE_TG_MEM");
	exit(1);
    }

    /* check the window reg */
    readData.addr = TG_WINDOW_BASE;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("window base is %x\n", (u_int)readData.data);

    /* read data at window base */
    readData.addr = TG_WINDOW_DATA;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("window data is %x\n", (u_int)readData.data);

    /* then read it back */
    memData.tgAddr = 0x1000;
    memData.userAddr = sdata2;
    memData.len = 20;
    if (ioctl(fd, ALT_READ_TG_MEM, &memData) < 0) {
	perror("couldn't do ALT_READ_TG_MEM");
	exit(1);
    }

    for (i=0; i < 20; i++)
	if (sdata[i] != sdata2[i])
	    printf("mismatched data at %d, got %x expected %x\n", i,
		   sdata2[i], sdata[i]);

    /* try to execute some instructions */

    /* put three nops in a row */
    datap = (U32 *)sdata;
    for (i=0; i<3; i++)
	*datap++ = 0x00000000;
    memData.tgAddr = 0x1000;
    memData.userAddr = sdata;
    memData.len = 12;
    if (ioctl(fd, ALT_WRITE_TG_MEM, &memData) < 0) {
	perror("couldn't do ALT_WRITE_TG_MEM");
	exit(1);
    }

    /* bounce the pc to 0 then back to 0x1000 to clear cache */
    writeData.addr = TG_PROG_CTR;
    writeData.data = 0x0;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    writeData.addr = TG_PROG_CTR;
    writeData.data = 0x1000;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    /* execute the code */
    writeData.addr = TG_CPU_CONTROL;
    writeData.data = TG_CPU_HALT | TG_CPU_SINGLE;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    readData.addr = TG_CPU_CONTROL;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("cpu state/control is now %x\n", (u_int)readData.data);

    writeData.addr = TG_CPU_CONTROL;
    writeData.data = TG_CPU_HALT | TG_CPU_SINGLE;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    readData.addr = TG_CPU_CONTROL;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("cpu state/control is now %x\n", (u_int)readData.data);

    writeData.addr = TG_CPU_CONTROL;
    writeData.data = TG_CPU_HALT | TG_CPU_SINGLE;
    if (ioctl(fd, ALT_WRITE_TG_REG, &writeData) < 0) {
	perror("couldn't do ALT_WRITE_TG_REG");
	exit(1);
    }

    readData.addr = TG_CPU_CONTROL;
    if (ioctl(fd, ALT_READ_TG_REG, &readData) < 0) {
	perror("couldn't do ALT_READ_TG_REG");
	exit(1);
    }

    printf("cpu state/control is now %x\n", (u_int)readData.data);
    return(0);
}

