/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: tsock.c,v $
 * Revision 1.6.4.9  1998/06/08  23:57:01  ted
 * 	Excise AIX stuff from the source base
 * 	[1998/05/21  22:35:08  ted]
 *
 * 	checkpoint Solaris hybrid working
 * 	[1998/05/21  22:09:28  ted]
 *
 * Revision 1.6.4.8  1997/09/16  03:12:26  ted
 * 	remove bogus use of __alpha for checking for DEC OSF
 * 	[1997/09/16  03:10:24  ted]
 * 
 * Revision 1.6.4.7  1997/08/23  17:40:50  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:39:37  skapur]
 * 
 * Revision 1.6.4.6  1997/07/28  22:41:11  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:41:46  ted]
 * 
 * Revision 1.6.4.5  1997/05/21  17:24:24  suresh
 * 	fix build-depend problems on AIX
 * 	[1997/05/21  17:23:27  suresh]
 * 
 * Revision 1.6.4.4  1997/04/12  00:12:07  ted
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:08:35  ted]
 * 
 * Revision 1.6.4.3  1997/03/24  08:34:49  ted
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:53  ted]
 * 
 * Revision 1.6.4.2  1997/02/21  00:38:30  ted
 * 	Add -A flag (support for a frame count in the pattern).
 * 	[1997/02/21  00:38:19  ted]
 * 
 * Revision 1.6.4.1  1997/02/07  21:58:58  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:51:14  ted]
 * 
 * Revision 1.6.2.3  1996/11/20  20:04:58  ted
 * 	Redo include file structuring.
 * 	[1996/11/20  19:59:39  ted]
 * 
 * Revision 1.6.2.2  1996/11/09  02:32:23  ted
 * 	clean up compiler warnings
 * 	[1996/11/09  02:31:38  ted]
 * 
 * Revision 1.6.2.1  1996/11/01  18:26:06  ted
 * 	changed rr_ to tg_ and put in proper copyrights
 * 	[1996/10/30  23:08:18  ted]
 * 
 * $EndLog$
 */
/*
 * FILE tsock.c
 *
 * Module Name: tsock.c  --  UltraNet Socket-based Network Exerciser
 *
 * Copyright (c) 1988 Ultra Network Technologies, Inc.
 *  All rights reserved.
 *
 * Original Author: Wayne Hathaway
 *
 * $Id: tsock.c,v 1.6.4.9 1998/06/08 23:57:01 ted Exp $
 *
 */

#if !defined(__lint)
char copyright[]    = "Copyright (c) 1988 Ultra Network Technologies, Inc.";
static const char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/utils/common/tsock.c,v 1.6.4.9 1998/06/08 23:57:01 ted Exp $";
#endif /* !defined(__lint) */

/*===========================================================================*/

/*
 *    This program is based on the "ttcp" program originally developed by
 *    T.C. Slattery, USNA, and improved by Mike Muuss and Terry Slattery.
 *    Due to the large number of changes, however (asynchio, data checking,
 *    random length and offset, fake-data option, ...), it has effectively
 *    been completely rewritten.
 */

/*
 *    Processing logic is as follows:
 *
 *                     fetch and validate parameters
 *                     miscellaneous initialization
 *                     define socket
 *                     rest depends on options
 *                    /                       \
 *            datagram               connection-oriented
 *            /      \                 open connection
 *          trans   rcvr              /               \
 *                          asynch i/o                 normal i/o
 *                          /        \                 /        \
 *                     sinkmode     filemode       sinkmode     filemode
 *                      /    \       /    \         /    \       /    \
 *                   trans  rcvr  trans  rcvr    trans  rcvr  trans  rcvr
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#if defined(__sgi)
#include <bstring.h>
#endif /* defined(__sgi) */
#if defined(__sun)
#include <strings.h>
#endif /* defined(__sun) */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/param.h>                  /* for NBPW [used in btow()] */
                                        /*  and for MAXPATHLEN */

#include <sys/resource.h>
#define RANDOM  rand                  /* account for Sys V / BSD diffs ... */
#define SRANDOM srand
#include <errno.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/buf.h>                    /* for unetadt (unetioctl includes!) */
#include <sys/file.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#if !defined(_AIX)
#include <sys/fcntl.h>
#endif /* !defined(_AIX) */
/* streams specific include files */
#if defined(ALT_STREAM)
#include <sys/stream.h>
#endif /* defined(ALT_STREAM) */

/* system specific files */
#if defined(__sun)
#include <sys/ioccom.h>
#include <sys/ethernet.h>
#endif /* defined(__sun) */

#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#include "altftns.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#include "altioctl.h"

#if defined(mips) && !defined(sgi)
#define _undef_                 /* This is a real _undef_ system */
#endif /* defined(mips) && !defined(sgi) */

#if !defined(NULL)
#define NULL 0
#endif /* !defined(NULL) */

#ifndef BCOPY
#define BCOPY bcopy
#endif
#ifndef BZERO
#define BZERO bzero
#endif
#ifndef BCMP
#define BCMP bcmp
#endif 

#ifdef __hpux
#define SEL_TYPE int
#endif /* __hpux */

#if defined(__sgi) || defined(__sun) || defined(_AIX)
#define SEL_TYPE fd_set
#endif /* defined(__sgi) || defined(__sun) || defined(_AIX) */

#define ONE_GIG (1024 * 1024 * 1024)

/* prototypes for function calls */
static int checkwcnt(int, int);
static void dumphex(u_char *, int);
static void catch_alarm(int);
static void usage(void);
static void err(char *);
static void badarg(char *);
static void mes(char *);
static void mes2(char *, char *);
static int gethex(int);
static void cvtpatstring(char *);
static void cvtpatfile(char *);
static void pattern(u_char *, int);
static int checkpat(u_char *, int);
static void do_forks(int);
static void dorandom(void);
static void prep_timer(void);
static void read_timer(char *, int);
static void domark(int);
static void maybe_do_lseek(int);
static int Nrecv(int, char *, int);
static int Nread(int, char *, int);
static int Nsend(int, char *, int);
static int Nwrite(int, char *, int);
static void delay(int);
static int mread(int, char *, unsigned);
static int cmppat(u_char *, u_char *, int);
#ifdef __hpux
static char *valloc(size_t);  
#endif /* __hpux */

#define DEFAULT_BFRLEN 1024
#define DEFAULT_NBFR   1024
#define DEFAULT_PORT   2000
#define MARK_CHAR      "X"
                                        /* PARAMETER FLAGS AND/OR VALUES ... */
static int   sockbufsize;               /* -b value for SO_{SND|RCV}BUF */
static int   bfrlen = DEFAULT_BFRLEN;   /* length of buffer */
static int   nbfr   = DEFAULT_NBFR;     /* num buffers to send in sinkmode */
static int   nbfr_specified = 0;        /* 1="-n" specified (ie, not default) */
static u_short port = DEFAULT_PORT;     /* port number */
static int   datagram;                  /* 0=COTS, 1=CLTS, -1=CLTS w/out end */
static int   trans = -1;                /* 0=receive, 1=transmit mode */
static int   sinkmode;                  /* 0=normal I/O, 1=sink/source mode */
static int   asynchio;                  /* 0=synch (UNIX) I/O, 1=asynch I/O */
static int   asynchbfrs;                /* number of asynch I/O buffers */
static int   bfroffset;                 /* offset from valloc-ed bfr (eg, to */
                                        /*  test with NON-ALIGNED buffer) */
static int   ranlen;                    /* 0=fixed length bufs, 1=random */
static int   ranstart;                  /* 0=normal, 8=use random start loc */
                                        /*  (8 is extra bfr bytes needed) */
static int   rannoseed;                 /* 0=normal, 1=do not do random seed */
                                        /* with random seed (ie time or pid) */
static u_int ranseedval;                /* 0=noseed, nonzero = seed value */
static int   randata = 0;               /* 0=normal, 1=random data in xmit buf*/
static int   ranregenno = 0;            /* 0=once only, num = xmit num buffs */
                                        /* then create new random data again */
static int   checkdata;                 /* 0=ignore recved data, n=check it */
static int   checkcnt;                  /*  (every nth message is checked) */
static int   checkextra = 1;            /* 1=check for extra received data */
                                        /*  (starts at 1 because not used */
                                        /*    without -c; set to 0 by -X) */
static int   markfreq;                  /* 0=no progress marking; n=print */
                                        /*  MARK_CHAR every n bytes */
static int   seek_cntr = ONE_GIG;       /* Counter for doing lseek() */
static int   nodelay = 0;	        /* set TCP_NODELAY socket option */
static int   b_flag = 0;                /* 0=use read(), 1=use mread() */
static int   seltimeout = -1;           /* -1=no select(), else timeout */
                                        /*  0=timeout is poll */
static int   fork_cnt = 0;              /* number of times to fork */
static int   zigzag = 0;                /* 0=normal, 1=zigzag when forking */
static u_char minchar = 0x20;           /* pattern min and max chars */
static u_char maxchar = 0x7e;
static u_char patbuf = 0;
static int   altpattern = 0;            /*  0 = normal pattern, */
                                        /*  1 = default alternate pattern, */
                                        /* -1 = pattern from string "altpat" */
                                        /* -2 = pattern from file "apfile" */
static char *altpat;                    /* alternate pattern string */
static char *apfile;                    /* name of pattern file */
static u_char apat[1000];               /* binary alternate pattern */
static int   aplen;                     /* length of binary pattern */
static char *host;                      /* ptr to name of host (loc or rem) */
static char *Host;                      /* ptr to local hostname (-t only) */
static struct sockaddr_in sinme;
static struct sockaddr_in sinhim;
static struct sockaddr_in frominet;

static int fromlen;
static int fd;                          /* fd of network socket */
static int fd1;                         /* extra fd returned by accept */
static int xlen;                        /* actual length used for transfer */
static int xoff;                        /* offset into buffer to start xfer */
static int ranerror = 0;                /* nonzero means bad '?' parameter */
static int marklast = 0;                /* remembers bytes at last mark */
static int markline = 0;                /* counts marks per line */
static int timelimit = 0;               /* seconds to run (used to limit -n) */
                                        /*  (set to -1 once timer goes off) */
static int clearProf = 0;	        /* clear Alteon NIC perf profile */

static U32  zero = 0;                  /* for improved datagram robustness */
static char *zp = (char *) &zero;       /* special "end signal" datagrams */

static char *bfr;                       /* ptr to dynamic buffer */
static char *cmpbfr;                    /* ptr to compare buf (data checking)*/
static char  cmd[MAXPATHLEN+32];        /* either "tsock-r" or "tsock-t", */
                                        /* with optional port in brackets */
static char *cmdname;                   /* command name (argv[0] on entry) */

static struct hostent *addr;
static int so_debug = 0;                /* flag (0=NO SO_DEBUG on socket */
                                        /* 1 = SO_DEBUG on socket */
extern int errno;

static char Usage1[] = "\
Usage: %s [-options] [host] <in >out\n\
        -r      operate as receiver (passive initiation, receive data)\n\
        -t      operate as transmitter (active initiation, transmit data)\n\
        -s      source/sink mode (generate and discard a data pattern)\n\
        -d[x]   use datagrams (with -s only) [x omits ending signals]\n\
        -l##    length of tsock send/recv buffer (default %d)\n\
        -n##    number of messages to transmit (with -s only, default %d)\n\
        -T##    maximum seconds to transmit (with -s only)\n\
";
static char Usage2[] = "\
        -m##    mark progress with '%s' for every ## bytes transferred\n\
        -?l     use random length messages <= -l## value (without -d only)\n\
        -?s     use random starting location (i.e., random byte boundary)\n\
        -?r[##] use ## as seed (-?r alone does NOT seed random num generator)\n\
        -?R     fill transmit buffer with random data\n\
        -?N##   regenerate random data every ## buffer (-?N implies -?R)\n\
        -p##    port number for connection or datagrams (default %d)\n\
        -b##    socket buffer size (SO_SNDBUF/SO_RCVBUF)\n\
        -h      set SO_DEBUG on socket (used to Help debug connection)\n\
";
static char Usage3[] = "\
        -o##    offset from 'natural buffer boundary' (eg, page boundary)\n\
        -B      only transmit full blocks, as specified in -l## (for TAR)\n\
        -a##    use asynch network I/O (eg, UNICOS reada) with ## buffers\n\
        -c##    check every ##th received message (receiver with -s only)\n\
        -X      do NOT check for extra received bytes\n\
        -x##    run ## processes in parallel (via fork)\n\
";

static char Usage4[] = "\
        -z##    same as -x## but fork both receivers and transmitters\n\
        -Pxpat  use alternate pattern hex 'xpat' (default 0x00-0xff)\n\
        -Ffile  use alternate pattern from file\n\
        -S##    use select with ## second timeout for accept()\n\
        -C      clear Alteon NIC host profiling counts\n\
	-D	don't buffer TCP writes (sets TCP_NODELAY socket option)\n\
        host    host name (optional local for -r, required remote for -t)\n\
";

static char stats[128];
#if defined(U64)
static U64 nbytes;		        /* bytes on net, real big */
#else /* defined(U64) */
static U32 nbytes;		        /* bytes on net */
                                        /* (made u_long to allow > 2 Gbytes) */
#endif /* defined(U64) */

static double cput; 
static clock_t realt;              /* user, real time (seconds) */
static S32  clk;                        /* for "ctime" */

/*---------------------------------------------------------------------------*/

/*
 *    Miscellaneous support routines (using global variables WAY too much!)
 */

static void			/* routine to pick new random values */
dorandom()
{
    if (ranlen)
        xlen = RANDOM() % bfrlen + 1;
    else
        xlen = bfrlen;
    if (ranstart)
        xoff = RANDOM() % 8;
    else
        xoff = 0;
}

static void			/* routine to do progress marking */
domark(int cnt)
{
    marklast += cnt;
    while (marklast >= markfreq) {
        if (++markline > 50) {
            (void) fprintf(stderr, "\n%s: ", cmd);
            markline = 1;
        }
        (void) fprintf(stderr, "%s", MARK_CHAR);
        (void) fflush(stderr);
        marklast -= markfreq;
    }
}


static void
maybe_do_lseek(int cnt)
{

/*
 *  Some systems have a problem reading/writing more than 2 or 4 Gbyte
 *  of data on one connection.  The problem is caused by the the OS
 *  maintaining a "current position" pointer for the Ultra connection,
 *  which to the OS is a "character special" device.  The position
 *  pointer is usually in a 32 bit word, and when it overflows (at 2GB
 *  if it's signed, 4 GB if unsigned), the OS gets confused and may return
 *  an error to the next read/write.  Doing an "lseek(fd, SEEK_SET, 0)"
 *  resets the pointer, thus avoiding the problem.  We do the lseek
 *  every 1 GB, which is often enough to be sure we don't hit a problem,
 *  but infrequent enough to avoid adding noticeable overhead from extra
 *  system calls.
 *
 *  There's really no point in maintaining a position pointer for a
 *  character special device, but several OS's are known to do it.
 *  The code here isn't ifdef'd for the OS known to have the problem,
 *  as doing the lseek can't hurt, and this way we're covered on all
 *  XSsystems without worrying about which ones need it.
 */

    if ((seek_cntr -= cnt) < 0) {
        /*
         * puposely not checking returned status:  we know it
         * will fail on real sockets, which doesn't hurt anything.
         */
        (void)lseek(fd, (off_t)0, SEEK_SET);
        seek_cntr = ONE_GIG;
    }
}

static int                              /* routine to check for xmit errors */
checkwcnt(int wcnt, int error)          /* returns "corrected" wcnt */
{
    if ((wcnt > 0) && (nbfr != -1) && (timelimit != -1))
        (void) fprintf(stderr,"%s: ABORT! wrote %d of %d bytes, %d buf(s) to go\n",
                       cmd, wcnt, xlen, nbfr + 1);
    else if ((wcnt == 0) && (nbfr != -1) && (timelimit != -1))
        (void) fprintf(stderr,"%s: ABORT! early EOF, %d buf(s) to go\n",
                       cmd, nbfr + 1);
/* now handle System V, which returns EINTR when SIGALRM interrupts write! */
    else if ((wcnt < 0) && (error == EINTR) && (timelimit == -1))
        wcnt = 0;
    return(wcnt);
}


static void
usage()
{
    (void) fprintf(stderr, Usage1, cmdname, DEFAULT_BFRLEN, DEFAULT_NBFR);
    (void) fprintf(stderr, Usage2, MARK_CHAR, DEFAULT_PORT);
    (void) fprintf(stderr, Usage3);
    (void) fprintf(stderr, Usage4);
}


static void
badarg(char *s)
{
    (void) fprintf(stderr, "%s: invalid parameter '%s'\n", cmdname, s);
    usage();
    exit(1);
}

static void			/* routine to create n copies */
do_forks(int n)			/*  (possibly alternating xmit/recv) */
{
    int rc;
    while (n--) {
        rc = fork();
        if (rc == 0)
            return;
        if (rc < 0)
            err("fork");
        if (zigzag)
            if (trans) {
                trans = 0;
                cmd[strlen(cmdname)+1] = 'r';
            } else {
                trans = 1;
                cmd[strlen(cmdname)+1] = 't';
            }
        (void) strcat(cmd, "+");
    }
}

#define ALTDEV "/dev/alt"

/*
 *    DUMPHEX  --  internal routine to dump a string of chars in hex
 *
 *    Note:  Does NOT put a '\n' at the end of the printed hex characters.
 */

static void
dumphex(u_char *c, int n)
{
    static char *hex[] = { "0", "1", "2", "3", "4", "5", "6", "7",
                           "8", "9", "a", "b", "c", "d", "e", "f" };
    while (n--) {
        (void) fprintf(stderr, "%s%s",
                       hex[(u_char)(*c & 0xf0) >> 4], 
		       hex[(*c & 0x0f)]);
        c++;
    }
}



/*
 *    AADD_ALARM  --  catch timeout (SIGALRM) on "aadd" operation
 */

/*ARGSUSED*/
static void
catch_alarm(int sig)
{
    nbfr = 1;                           /* simply say "one more buffer" and */
    timelimit = -1;                     /*  set flag to avoid "ABORT!" msg */
}

/*---------------------------------------------------------------------------*/

int
main(argc,argv)
    int argc;
    char **argv;
{
    register int rcnt = 0;
    register int wcnt = 0;
    register int save_errno = 0;
    register int t;                     /* temp */

    cmdname = argv[0];
    (void) strncpy(cmd, cmdname, sizeof(cmd));     /* in case of early error */
    if (argc < 2) {
        usage();
        exit(1);
    }
    argv++; argc--;
    while (argc>0 && argv[0][0] == '-')  {
        switch (argv[0][1]) {

        case 'A':
            patbuf = 1;
	    printf("marking first byte of each word with the buffer number\n");
            break;
        case 'B':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            b_flag = 1;
            break;
        case 't':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            if (trans == 0) {
                (void) fprintf(stderr,"%s: can't have both -r and -t\n",
                               cmdname);
                usage();
                exit(1);
            }
            trans = 1;
            break;
        case 'r':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            if (trans == 1) {
                (void) fprintf(stderr, "%s: can't have both -r and -t\n",
                               cmdname);
                usage();
                exit(1);
            }
            trans = 0;
            break;
        case 'c':
            checkdata = atoi(&argv[0][2]);
            if (checkdata == 0)
                checkdata = 1;
            break;
        case 'C':
            clearProf = 1;
            break;
        case 'X':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            checkextra = 0;
            break;
        case 'n':
            nbfr = atoi(&argv[0][2]);
            nbfr_specified = 1;
            if (nbfr <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'l':
            bfrlen = atoi(&argv[0][2]);
            if (bfrlen <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            t = argv[0][strlen(argv[0])-1];   /* farble: allow shorthand ... */
            if (t == 'k' || t == 'K')
                bfrlen *= 1024;                        /*  .. K => kilobytes */
            else if (t == 'm' || t == 'M')
                bfrlen *= (1024 * 1024);               /*  .. M => megabytes */
            else if (t == 'g' || t == 'G')
                bfrlen *= (1024 * 1024 * 1024);        /*  .. G => gigabytes */
            break;
        case 'b':
            sockbufsize = atoi(&argv[0][2]);
            if (sockbufsize <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            t = argv[0][strlen(argv[0])-1];   /* farble: allow shorthand ... */
            if (t == 'k' || t == 'K')
                sockbufsize *= 1024;                   /*  .. K => kilobytes */
            else if (t == 'm' || t == 'M')
                sockbufsize *= (1024 * 1024);          /*  .. M => megabytes */
            else if (t == 'g' || t == 'G')
                sockbufsize *= (1024 * 1024 * 1024);   /*  .. G => gigabytes */
            break;
        case 'm':
            markfreq = atoi(&argv[0][2]);
            if (markfreq <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            t = argv[0][strlen(argv[0])-1];   /* farble: allow shorthand ... */
            if (t == 'k' || t == 'K')
                markfreq *= 1024;                      /*  .. K => kilobytes */
            else if (t == 'm' || t == 'M')
                markfreq *= (1024 * 1024);             /*  .. M => megabytes */
            else if (t == 'g' || t == 'G')
                markfreq *= (1024 * 1024 * 1024);      /*  .. G => gigabytes */
            if (markfreq == 0)
                markfreq = 10 * DEFAULT_BFRLEN;
            break;
        case 'a':
            asynchio = 1;
            asynchbfrs = atoi(&argv[0][2]);
            if (asynchbfrs <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case '?':
        case '/':                       /* undocumented, to avoid escaping ? */
            switch (argv[0][2]) {
            case 'l':
                ranlen = 1;
                break;
            case 's':
                ranstart = 8;           /* "random start loc" is 0 through 7 */
                break;
            case 'r':
                rannoseed = 1;          /* 1 = do not pick a random seed to */
                                        /* put into SRANDOM, user has just  */
                                        /* given it to us or use random num */
                                        /* generator default  (do no seed)  */
                ranseedval = atoi(&argv[0][3]);
                break;
            case 'R':                   /* fill xmit buffer with random data */
                randata = 1;
                break;
            case 'N':
                                        /* number of buffers to xmit ran data */
                                        /* before refilling with new ran data */
                ranregenno = atoi(&argv[0][3]);
                if (ranregenno == 0) {
                    badarg(argv[0]);
                }
                randata = 1;
                break;
            default:
                ranerror = 1;
                break;
            }
            break;
        case 's':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            sinkmode = 1;               /* source OR sink, really */
            break;
        case 'p':
            port = atoi(&argv[0][2]);
            if (port == 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'd':
            if (argv[0][2] == 'x')      /* handle flag saying "do not send */
                datagram = -1;          /*  ending signal "zp" datagrams" */
            else if (argv[0][2] == '\0')
                datagram = 1;
            else
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'z':
            zigzag = 1;
            /*FALLTHROUGH*/
        case 'x':
            if (fork_cnt && zigzag) {
                (void) fprintf(stderr,"%s: can't have both -x and -z\n",
                               cmdname);
                usage();
                exit(1);
            }
            fork_cnt = atoi(&argv[0][2]) - 1;
            if (fork_cnt <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'P':
            if (strlen(argv[0]) == 2) {       /* alternate pattern 0x00-0xff */
                altpattern = 1;
                minchar = 0;
                maxchar = 0xff;
            } else {                          /* alternate pattern string */
                altpattern = -1;
                altpat = (char *)&argv[0][2];
            }
            break;
        case 'F':
            if (strlen(argv[0]) == 2) {       /* alternate pattern 0x00-0xff */
                (void) fprintf(stderr,"%s: '-F' must have file name\n",
                               cmdname);
                usage();
                exit(1);
            } else {                          /* alternate pattern from file */
                altpattern = -2;
                apfile = (char *)&argv[0][2];
            }
            break;
        case 'S':
            seltimeout = atoi(&argv[0][2]);
            if (seltimeout <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'o':
            bfroffset = atoi(&argv[0][2]);
            if (bfroffset <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
        case 'H':
            if (argv[0][2] == '\0') {
                (void) fprintf(stderr, "%s: -H: missing host name (must follow '-H' with no spaces)\n", cmdname);
                usage();
                exit(1);
            }
            Host = &argv[0][2];
            break;
        case 'h':
            if (argv[0][2] != '\0')
                badarg(argv[0]);                    /* NOTE: does not return */
            so_debug = 1;
            break;
        case 'T':
            timelimit = atoi(&argv[0][2]);
            if (timelimit <= 0)
                badarg(argv[0]);                    /* NOTE: does not return */
            break;
	case 'D':
	    nodelay = 1;
	    break;
        default:
            badarg(argv[0]);                        /* NOTE: does not return */
        }
        argv++; argc--;
    }
    if (argc > 1) {                     /* flag possible argument screwup */
        mes("host must be last parameter (or something else wrong?)");
        usage();
        exit(1);
    }
    host = argv[0];
    if (trans < 0) {                    /* must have either -r or -t ... */
        (void) fprintf(stderr,"%s: must have either -r or -t\n", cmdname);
        usage();
        exit(1);
    }
    if ((port < 1024) && geteuid()) {
        (void) fprintf(stderr,"%s: port must be >= 1024\n", cmdname);
        exit(1);
    }
                                        /* set up command name for output */
    if (port == DEFAULT_PORT)
        (void) sprintf(cmd, "%s%s", cmdname, trans ? "-t" : "-r");
    else
        (void) sprintf(cmd, "%s%s[%d]", cmdname, trans ? "-t" : "-r", port);
                                        /* now convert alt pattern (if any) */
    if (altpattern == -1)
        cvtpatstring(altpat);
    else if (altpattern == -2)
        cvtpatfile(apfile);
                                        /* now a little sanity checking ... */

    if (argc > 1) {               /* (leftover args after first without "-") */
        mes("host must be last parameter (or something else wrong?)");
        usage();
        exit(1);
    }
    if (ranerror) {
        mes("unrecognizable '?' parameter; it was ignored");
    }
    if (datagram && !sinkmode) {
        mes("-d option [datagrams] only applicable with -s; -d ignored");
        datagram = 0;
    }
    if (datagram && zigzag) {
        mes("-z (zigzag mode) doesn't work for datagrams; terminated");
        exit(1);
    }
    if (datagram && (seltimeout != -1)) {
        mes("-S option [select] not applicable with -d; -S ignored");
        seltimeout = -1;
    }
    if ((bfrlen < 4) && (datagram > 0) && !trans) {
        mes("buflen must be at least 4 for datagrams; buflen set to 4\n");
        bfrlen = 4;
    }
    if (checkdata && !sinkmode) {
        mes("-c option [check data] only applicable with -s; -c ignored");
        checkdata = 0;
    }
    if (checkdata && trans
        && !zigzag
    ) {
        mes("-c option [check data] for receiver only; -c ignored");
        checkdata = 0;
    }
    if (altpattern && !sinkmode) {
        mes("alternate pattern meaningless without -s; ignored");
        altpattern = 0;
    }
    if (!trans
        && !zigzag
        && altpattern && !checkdata) {
        mes("alternate pattern meaningless on receiver without -c; ignored");
        altpattern = 0;
    }
    if (!trans
        && !zigzag
        && (altpattern == -2) && (ranlen || ranstart)) {
        mes("pattern file won't check correctly with randomness; aborted");
        exit(1);
    }
    if (b_flag && trans) {
        mes("-B option [full blocks] for receiver only; -b ignored");
        b_flag = 0;
    }
    if (nodelay && !trans) {
        mes("-B option [full blocks] for transmitter only; -D ignored");
        b_flag = 0;
    }
    if (
        zigzag &&
        ranlen) {
        mes("warning: -z and -?l together probably won't terminate!");
    }
    if (nbfr_specified && !trans
        && !zigzag
        ) {
        mes("-n option [nbufs] for transmitter only; -n ignored");
        nbfr = DEFAULT_NBFR;
        nbfr_specified = 0;
    }
    if (timelimit && !trans
        && !zigzag
        ) {
        mes("-T option [time limit] for transmitter only; -T ignored");
        timelimit = 0;
    }
    if (Host && !trans) {
        mes("-H option [local hostname] for transmitter only; -H ignored");
        Host = (char *) 0;
    }
    if (nbfr_specified && !sinkmode) {
        mes("-n option [nbufs] only applicable with -s; -n ignored");
        nbfr = DEFAULT_NBFR;
        nbfr_specified = 0;
    }
    if (timelimit && !sinkmode) {
        mes("-T option [time limit] only applicable with -s; -T ignored");
        timelimit = 0;
    }
    if (timelimit && !nbfr_specified)   /* -T without -n means stop ONLY */
        nbfr = -1;                      /*  after given number of seconds */

    if (asynchio) {
        mes("-a option [asynch I/O] not supported on this system; -a ignored");
        asynchio = 0;
    }

    if (trans)  {
                                                    /* setup for TRANSMITTER */
        if (argc != 1) {
            mes("remote host name required for transmitter");
            usage();
            exit(1);
        }
        BZERO((char *)&sinhim, sizeof(sinhim));
        if (
            isascii((int)*host) &&
            isdigit((int)*host))  {
                                        /* Numeric */
            sinhim.sin_family = AF_INET;
            sinhim.sin_addr.s_addr = inet_addr(host);
        } else {
            if ((addr=gethostbyname(host)) == NULL) {
                mes2("bad hostname", host);
                exit(1);
            }
            sinhim.sin_family = addr->h_addrtype;
            BCOPY(addr->h_addr,(char*)&sinhim.sin_addr, addr->h_length);
        }
        sinhim.sin_port = htons(port);

        if (Host) {                     /* have local hostname for bind() */
            if (
                isascii((int)*Host) &&
                isdigit((int)*Host))  {
                                        /* Numeric */
                sinme.sin_family = AF_INET;
                sinme.sin_addr.s_addr = inet_addr(Host);
            } else {
                if ((addr = gethostbyname(Host)) == NULL) {
                    mes2("bad local hostname", Host);
                    exit(1);
                }
                sinme.sin_family = addr->h_addrtype;
                BCOPY(addr->h_addr, (char *) &sinme.sin_addr.s_addr,
                      addr->h_length);
            }
        }                               /* no local hostname, use wildcard */

        sinme.sin_port = 0;             /* free choice */

    } else {                                           /* setup for RECEIVER */

        if (!zigzag) {                    /* set to run forever unless zigzag */
            nbfr = -1;
        }

        if (host) {                     /* have local hostname specified */
            if (
                isascii((int)*host) &&
                isdigit((int)*host))  {
                                        /* Numeric */
                sinme.sin_family = AF_INET;
                sinme.sin_addr.s_addr = inet_addr(host);
            } else {
                if ((addr = gethostbyname(host)) == NULL) {
                    mes2("bad hostname", host);
                    exit(1);
                }
                sinme.sin_family = addr->h_addrtype;
                BCOPY(addr->h_addr, (char *) &sinme.sin_addr.s_addr,
                      addr->h_length);
            }
        }                               /* no local hostname, use wildcard */
                                        /* (note bad technique assuming that */
                                        /*  wildcard (INADDR_ANY) is zero!)*/
        sinme.sin_port =  htons(port);
    }

    sinme.sin_family = AF_INET;

    if (trans
        || zigzag
        ) {                             /* tell user chosen param settings */
        if (sinkmode) {
            if (timelimit) {
                if (nbfr_specified)
                    (void) fprintf(stderr,
                                   "%s: nbuf=%d (or %d secs), buflen=%d%s",
                                   cmd, nbfr, timelimit, bfrlen,
                                   ranlen?"[max]":"");
                else
                    (void) fprintf(stderr,"%s: %d seconds, buflen=%d%s",
                                   cmd, timelimit, bfrlen, ranlen?"[max]":"");
            } else
                (void) fprintf(stderr,"%s: nbuf=%d, buflen=%d%s",
                               cmd, nbfr, bfrlen, ranlen?"[max]":"");
        } else
            (void) fprintf(stderr,"%s: buflen=%d%s",
                           cmd, bfrlen, ranlen?"[max]":"");
    } else {
        (void) fprintf(stderr,"%s: buflen=%d%s",
                       cmd, bfrlen, ranlen?"[max]":"");
    }
    if (sockbufsize) {
        (void) fprintf(stderr, ", sockbufsize = %d", sockbufsize);
    }
    if (asynchio) {
        (void) fprintf(stderr,", asynch I/O (%d buffers)", asynchbfrs);
    }
    if (host) {
        (void) fprintf(stderr,", host '%s'", host);
    }
    if (fork_cnt) {
        (void) fprintf(stderr,", %d parallel processes", fork_cnt+1);
    }
    (void) fprintf(stderr,"\n");
    (void) fflush(stderr);

    if ((fd = socket(AF_INET, datagram ? SOCK_DGRAM : SOCK_STREAM, 0)) < 0) {
        err("socket");
    }
    mes("socket");

    if (so_debug){
        if (setsockopt(fd, SOL_SOCKET, SO_DEBUG, (void *)&so_debug,
			sizeof(so_debug)) < 0) {
            err("setsockopt SO_DEBUG");
        } else {
            mes("SO_DEBUG");
        }
    }

    if (sockbufsize) {                  /* set socket buf size if requested */
        if (trans
            || zigzag
	    ) {
            if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF,
                           (void *)&sockbufsize, sizeof(sockbufsize)) < 0) {
                err("setsockopt SO_SNDBUF");
            }
        }
        if (!trans
            || zigzag
	    ) {
            if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF,
                           (void *)&sockbufsize, sizeof(sockbufsize)) < 0) {
                err("setsockopt SO_RCVBUF");
            }
        }
    }

#if 0
    {
	int getsendbufsize, getrecvbufsize;
	int optlen;

	optlen = sizeof(int);
	if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF,
		       (void *)&getsendbufsize, &optlen) < 0) {
	    err("getsockopt SO_SNDBUF");
	}
	if (optlen != sizeof(int))
	    err("getsockopt SO_SNDBUF return unexpected optlen");
	optlen = sizeof(int);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF,
		       (void *)&getrecvbufsize, &optlen) < 0) {
	    err("getsockopt SO_SNDBUF");
	}
	if (optlen != sizeof(int))
	    err("getsockopt SO_SNDBUF return unexpected optlen");
	(void) fprintf(stderr,"%s: sockbuflen=%d:%d\n", 
		       cmd, getsendbufsize, getrecvbufsize);
    }
#endif /* 0 */

    if (bind(fd, (void *) &sinme, sizeof(sinme)) < 0) {
        err("bind");
    }

    errno = 0;

    if (!asynchio) {                    /* set up normal and datagram buffer */
        if ((bfr =
             valloc((size_t)(bfrlen + bfroffset + ranstart + checkextra)))
                                                   == (char *) 0) {
            err("valloc");
        }
        if (checkdata) {
            if ((cmpbfr =
                 valloc((size_t)(bfrlen + bfroffset + ranstart)))==(char *)0){
                err("valloc");
            }
        }

        bfr += bfroffset;               /* deliberately MISALIGN if -o used */

        if ((trans
	     || zigzag
	     ) && sinkmode) {
            pattern((u_char *)bfr, bfrlen + ranstart);
	    if (patbuf) {
		patbuf++;
		if (patbuf == 0)
		    patbuf = 1;
	    }
	}
         /* generate random data for checking and put into cmpbfr */
        if (!trans && randata && checkdata) {
            pattern((u_char *)cmpbfr, bfrlen + ranstart);
        }
    }


    (void) signal(SIGPIPE, SIG_IGN);    /* SIGPIPE just means receiver */
                                        /*  closed, so treat same as EOF */

    if (datagram) {                     /* now do "option dependent" stuff */

/*============================    DATAGRAM    ===============================*/

        do_forks(fork_cnt);             /* make multiple copies if requested */

        if (timelimit) {
            (void) signal(SIGALRM, catch_alarm);   /* timer driven for easier */
            (void) alarm((u_int)timelimit);        /*  use in scripts etc. */
        }

        if (ranlen || ranstart) {       /* seed random numbers */
            if (rannoseed) {            /* if 'don't pick random seed flag' */
                                        /* set then use seed user gave to us */
                                        /* or don't seed generator at all */
                if (ranseedval) {
                   SRANDOM(ranseedval);
                }
	    } else {
		SRANDOM((unsigned int)getpid());
            }
        }
        if (trans)  {                                     /* TRANSMITTER ... */
            int i;
            int j = ranregenno;
            dorandom();                 /* initialize possible randomizing */
            prep_timer();               /* then start clock (and print msg) */

            while (nbfr-- && (wcnt = Nsend(fd, bfr + xoff, xlen)) == xlen) {

                if (ranregenno) {
                    if (j-- <= 1) {
                        pattern((u_char *)bfr, bfrlen + ranstart);
                        j = ranregenno;
                    }
                }
		if (patbuf) {
		    pattern((u_char *)bfr, bfrlen + ranstart);
		    patbuf++;
		    if (patbuf == 0)
			patbuf = 1;
		}
		    
                nbytes += wcnt;
                dorandom();
                if (markfreq)
                    domark(wcnt);
            }
            save_errno = errno;                          /* just in case ... */
            if (datagram > 0) {              /* unless we were told not to, */
                for (i = 0; i < 10; i++)     /* signal "end" to receiver */
                    (void) Nsend(fd, zp, 4); /* (multiple for possible forks) */
	    }
            wcnt = checkwcnt(wcnt, save_errno);   /* then possible error msg */

        } else {                                          /* RECEIVER ... */
            int j = ranregenno;

            if (checkdata) {
                BZERO(bfr, bfrlen + ranstart + checkextra);
                checkcnt = 0;
            }
            dorandom();
            while (nbfr-- && (rcnt = Nrecv(fd, bfr + xoff, xlen)) > 0)  {
                static int going = 0;
                if (!going) {           /* first datagram in starts timing */
                    going = 1;
                    prep_timer();
                }
                if ((rcnt == 4) && !BCMP(bfr + xoff, zp, 4))
                    break;                  /* "EOF" */
                if (rcnt > xlen) {
                    (void) fprintf(stderr,
                         "%s: RECEIVE ERROR: asked for %d bytes, received %d",
                                 cmd, xlen, rcnt);
                    break;
                }
                if (checkdata) {
                    if (randata) {
                        if (ranregenno) {
                            if (j-- <= 1) {
                                pattern((u_char *)cmpbfr, bfrlen + ranstart);
                                j = ranregenno;
                            }
                        }
                        if ((++checkcnt == checkdata) &&
                            cmppat((u_char *)(bfr + xoff),
                                   (u_char *)(cmpbfr + xoff), rcnt)) {
                            break;
                        }
                        else if (checkcnt == checkdata-1) {
                            BZERO(bfr, bfrlen + ranstart + checkextra);
                        }
                    } else {
                        if ((++checkcnt == checkdata) &&
                            checkpat((u_char *)(bfr + xoff), rcnt))
                            break;
                        else if (checkcnt == checkdata-1)
                            BZERO(bfr, bfrlen + ranstart + checkextra);
                    }
                }
                nbytes += rcnt;
                dorandom();
                if (markfreq)
                    domark(rcnt);
            }
            save_errno = errno;                          /* just in case ... */
        }

    } else {

/*======================    CONNECTION-ORIENTED   ===========================*/

                                        /* first open the connection ... */
        if (trans) {                    /*  .. we are client if transmitting */
	    if (nodelay) {
		struct protoent *p;
		p = getprotobyname("tcp");
		if( p && setsockopt(fd, p->p_proto, TCP_NODELAY, 
				    (void *)&nodelay, sizeof(nodelay)) < 0)
		    err("setsockopt: nodelay");
		mes("nodelay");
	    }
            if (connect(fd, (void *) &sinhim, sizeof(sinhim)) < 0)
                err("connect");
            mes("connect");
        } else {                        /*  .. otherwise, we are server */
            (void) listen(fd,0);
            fromlen = sizeof(frominet);
            if (seltimeout >= 0) {
                struct timeval timeval;
                int nfds = 0;
                fd_set readfds;

                if (seltimeout == 0)
                    (void) fprintf(stderr, "select polling\n");
                while (nfds == 0) {
                    if (seltimeout) {
                        timeval.tv_sec = seltimeout;
                        timeval.tv_usec = 0;
                    }
		    FD_SET(fd, &readfds);
                    nfds = select(32, (SEL_TYPE *)&readfds, (SEL_TYPE *)0, 
				  (SEL_TYPE *)0, &timeval);
                    if (nfds == 0 && seltimeout != 0)
                        (void) fprintf(stderr,
                                       "select timeout %d\n", seltimeout);
                }
            }
            if ((fd1=accept(fd, (void *) &frominet, &fromlen)) < 0)
                err("accept");
            (void) close(fd);           /* minimize UNet device use ... */
            fd = fd1;
            mes("accept");
        }

        do_forks(fork_cnt);             /* make multiple copies if requested */

        if (timelimit) {
            (void) signal(SIGALRM, catch_alarm);   /* timer driven for easier */
            (void) alarm((u_int)timelimit);        /*  use in scripts etc. */
        }

        if (ranlen || ranstart) {       /* seed random numbers */
            if (rannoseed) {            /* if 'don't pick random seed flag' */
                                        /* set then use seed user gave to us */
                                        /* or don't seed generator at all */
                if (ranseedval) {
                    SRANDOM(ranseedval);
                }
            } else {
                SRANDOM((unsigned int)getpid());
            }
        }
        prep_timer();                   /* now start timer (and print msg) */

        if (asynchio) {

/*==============================    ASYNCH    ===============================*/
	    mes("asynchio not supported on this machine");


        } else {

/*=============================     NORMAL     ==============================*/

            if (sinkmode) {

/*==========================    NORMAL SINKMODE   ===========================*/

                if (trans)  {                             /* TRANSMITTER ... */
                    int j = ranregenno;

                    dorandom();
                    while (nbfr-- &&
                           (wcnt = Nwrite(fd, bfr + xoff, xlen)) > 0) {

                        if (ranregenno) {
                            if (j-- <= 1) {
                                pattern((u_char *)bfr, bfrlen + ranstart);
                                j = ranregenno;
                            }
                        }
			if (patbuf) {
			    pattern((u_char *)bfr, bfrlen + ranstart);
			    patbuf++;
			    if (patbuf == 0)
				patbuf = 1;
			}
                        nbytes += wcnt;
                        maybe_do_lseek(wcnt);
                        dorandom();
                        if (markfreq)
                            domark(wcnt);
                    }
                    save_errno = errno;                  /* just in case ... */
                    (void) close(fd);             /* signal "end" to receiver */
                    wcnt = checkwcnt(wcnt, save_errno); /* & possible err msg */

                } else {                                  /* RECEIVER ... */
                    int j = ranregenno;

                    if (checkdata) {
                        BZERO(bfr, bfrlen + ranstart + checkextra);
                        checkcnt = 0;
                    }
                    dorandom();
                    while (nbfr-- &&
                           (rcnt = Nread(fd, bfr + xoff, xlen)) > 0)  {
                        if (rcnt > xlen) {
                            (void) fprintf(stderr,
                          "%s: RECEIVE ERROR: asked for %d bytes, received %d",
                                               cmd, xlen, rcnt);
                            break;
                        }
                        if (checkdata) {
                            if (randata) {
                                if (ranregenno) {
                                    if (j-- <= 1) {
                                        pattern((u_char *)cmpbfr,
                                            bfrlen + ranstart);
                                        j = ranregenno;
                                    }
                                }
                                if ((++checkcnt == checkdata) &&
                                    cmppat((u_char *)(bfr + xoff),
                                           (u_char *)(cmpbfr + xoff), rcnt)) {
                                    break;
                                }
                                else if (checkcnt == checkdata-1) {
                                    BZERO(bfr, bfrlen + ranstart + checkextra);
                                }

                            } else {
                                if ((++checkcnt == checkdata) &&
                                    checkpat((u_char *)(bfr+ xoff), rcnt))
                                    break;
                                else if (checkcnt == checkdata-1)
                                    BZERO(bfr, bfrlen + ranstart + checkextra);
                            }
                        }
                        nbytes += rcnt;
                        maybe_do_lseek(rcnt);
                        dorandom();
                        if (markfreq)
                            domark(rcnt);
                    }
                    save_errno = errno;                  /* just in case ... */
                }

            } else {

/*==========================    NORMAL FILEMODE   ===========================*/

                if (trans)  {                             /* TRANSMITTER ... */

                    while ((rcnt=read(0, (void *)bfr, (size_t)bfrlen)) > 0 &&
                           (wcnt=Nwrite(fd, bfr, rcnt)) == rcnt) {
                        nbytes += rcnt;
                        maybe_do_lseek(rcnt);
                        if (markfreq)
                            domark(rcnt);
                    }

                }  else  {                                /* RECEIVER ... */

                    while ((rcnt=Nread(fd, bfr, bfrlen)) > 0 &&
                           (wcnt=write(1, (void *)bfr, (size_t)rcnt)) == rcnt) {
                        nbytes += rcnt;
                        maybe_do_lseek(rcnt);
                        if (markfreq)
                            domark(rcnt);
                    }
                }
            }

/*===========================================================================*/

        }                               /* end of ASYNCH vs NORMAL */

    }                                   /* end of DATAGRAM vs CONNECTION */

    if (markfreq)
        (void) fprintf(stderr, "\n");
    if (rcnt < 0 || wcnt < 0) {         /* if had error, msg and quit */
        if (rcnt < 0)
            (void) fprintf(stderr,"%s: read returned %d\n", cmd, rcnt);
        if (wcnt < 0)
            (void) fprintf(stderr,"%s: write returned %d\n", cmd, wcnt);
        errno = save_errno;
        err("IO");
    }
    read_timer(stats, sizeof(stats));     /*  otherwise print stats */
    clk = time((S32 *)0);
    (void) fprintf(stderr,"%s: transfer end:   %s", cmd, ctime((S32 *) &clk));
    if (strlen(stats))                  /* stats may be null on fsome UNIXes */
        (void) fprintf(stderr,"%s: %s\n", cmd, stats);
    if (cput  <= 0.0)  cput  = 0.001;
#if defined(U64)
    (void) fprintf(stderr,"%s: %llu bytes in %f CPU seconds = %f MB/cpu sec\n",
                   cmd, nbytes, cput, 
		   ((double)nbytes)/cput/1000000);
    (void) fprintf(stderr,"%s: %llu bytes in %f real seconds = %f MB/sec\n",
                   cmd, nbytes,
		   (double)realt/sysconf(_SC_CLK_TCK),
                   ((double)nbytes)/((double)realt/sysconf(_SC_CLK_TCK))/1000000);
#else /* defined(U64) */
    (void) fprintf(stderr,"%s: %lu bytes in %f CPU seconds = %f MB/cpu sec\n",
                   cmd, nbytes, cput, 
		   ((double)nbytes)/cput/1000000);
    (void) fprintf(stderr,"%s: %lu bytes in %f real seconds = %f MB/sec\n",
                   cmd, nbytes, (double)realt/sysconf(_SC_CLK_TCK),
                   ((double)nbytes)/((double)realt/sysconf(_SC_CLK_TCK))/1000000);
#endif /* defined(U64) */
    (void) fflush(stderr);

    if ((datagram > 0) && trans) {
        (void) Nsend(fd, zp, 4);        /* signal extra "ends" to receiver */
        (void) sleep((u_int) 1);        /* in datagram case for robustness */
        (void) Nsend(fd, zp, 4);
        (void) sleep((u_int) 1);
        (void) Nsend(fd, zp, 4);
        (void) sleep((u_int) 1);
        (void) Nsend(fd, zp, 4);
    } else if (!trans)                  /* for receiver, close net conn */
        (void) close(fd);

    return(0);

/*---------------------------------------------------------------------------*/

}


/*===========================================================================*/
/*===========================================================================*/

static void
err(char *s)
{
    S32 c = time((S32 *) 0);
    (void) fprintf(stderr,"%s: errno=%d: ", cmd, errno);
    perror(s);
    (void) fprintf(stderr,"%s: aborted %s", cmd, ctime(&c));
    /* ensure it's back to one */
    exit(1);
}

static void
mes(char *s)
{
    (void) fprintf(stderr,"%s: %s\n", cmd, s);
}

static void
mes2(char *s1, char *s2)
{
    (void) fprintf(stderr,"%s: %s '%s'\n", cmd, s1, s2);
}

static int			/* convert a hex character to an int */
gethex(int c)			/*  (can you say "brute force"?) */
{
    switch (c) {
    case '0':
        return(0);
    case '1':
        return(1);
    case '2':
        return(2);
    case '3':
        return(3);
    case '4':
        return(4);
    case '5':
        return(5);
    case '6':
        return(6);
    case '7':
        return(7);
    case '8':
        return(8);
    case '9':
        return(9);
    case 'a':
    case 'A':
        return(10);
    case 'b':
    case 'B':
        return(11);
    case 'c':
    case 'C':
        return(12);
    case 'd':
    case 'D':
        return(13);
    case 'e':
    case 'E':
        return(14);
    case 'f':
    case 'F':
        return(15);
    default:
        return(-1);
    }
}

static void                             /* read in and convert pattern file */
cvtpatfile(char *fnam)
{                                       /* returns globals apat[] and aplen */
    FILE *f;                            /* file descriptor */
    int   hlen;                         /* len of pattern file */
    u_char *h;                          /* hex (input) char scanner */
    u_char *b;                          /* binary (output) char scanner */
    int   leftnibble;                   /* 1=left nibble, 0=right nibbble */
    int   nibble;                       /* converted nibble */
    int   byte = 0;		        /* converted byte */

    if ((f = fopen(fnam, "r")) == (FILE *)NULL)
        err(fnam);
    if ((hlen = fread((void *)apat, sizeof(char), sizeof(apat), f)) < 0)
        err(fnam);
    if (fclose(f) < 0)
        err(fnam);
    h = apat;                           /* conversion done in place in apat */
    b = apat;                           /*  since always 2 chars -> 1 byte */
    leftnibble = 1;
    while (hlen-- > 0)
        if ((nibble = gethex((int)(*h++))) >= 0)
            if (leftnibble) {
                byte = nibble << 4;
                leftnibble = 0;
            } else {
                *b++ = byte + nibble;
                leftnibble = 1;
            }
    if (!leftnibble) {
        mes2("aborted: odd number of hex digits in pattern file", fnam);
        exit(1);
    }
    if ((aplen = b - apat) <= 0) {
        mes2("aborted: empty pattern file", fnam);
        exit(1);
    }
}

static void                             /*convert alternate pattern string */
cvtpatstring(char *apstr)
{                                       /* returns globals apat[] and aplen */
    int   hlen;                         /* len of alternate pattern string */
    u_char *h;                          /* hex (input) char scanner */
    u_char *b;                          /* binary (output) char scanner */
    int   leftnibble;                   /* 1=left nibble, 0=right nibbble */
    int   nibble;                       /* converted nibble */
    int   byte = 0;                         /* converted byte */

    (void) strncpy((char *)apat, apstr, sizeof(apat));
    hlen = strlen((char *)apat);

    h = apat;                           /* conversion done in place in apat */
    b = apat;                           /*  since always 2 chars -> 1 byte */
    leftnibble = 1;
    while (hlen-- > 0)
        if ((nibble = gethex((int)(*h++))) >= 0)
            if (leftnibble) {
                byte = nibble << 4;
                leftnibble = 0;
            } else {
                *b++ = byte + nibble;
                leftnibble = 1;
            }
    if (!leftnibble) {
        mes("'-P' string must be an even number of hex digits; aborted");
        exit(1);
    }
    if ((aplen = b - apat) <= 0) {
        mes2("aborted: error in alternate pattern string", apstr);
        exit(1);
    }
}

static void
pattern(u_char *cp, int cnt)
{                                       /* uses globals altpattern, apat, */
                                        /*  aplen, and (maybe) fakedata */
    register u_char c;

    if (!randata) {
        if (altpattern < 0) {           /* specified alt pattern, replicate */
            register int i = aplen;     /*  (only once, so brute force ok) */
            register u_char *pat = apat;
            while (cnt-- > 0) {
                *cp++ = *pat++;
                if (--i <= 0) {
                    i = aplen;
                    pat = apat;
                }
            }
        } else {                        /* simple byte incrementing */
            c = minchar-1;
            while (cnt-- > 0)  {
                if (++c > maxchar) c = minchar;
		if (patbuf && (!((U32)cp & 3))) {
		    *cp++ = patbuf;
		} else {
		    *cp++ = c;
		}
            }
        }
    } else {
        while(cnt-- > 0) {
            *cp++ = RANDOM() & 0xff;
        }
    }
}

static int
checkpat(u_char *cp, int cnt)
                                        /* also sets global checkcnt to zero */
{
    u_char c;
    int  rc = 0;
    int  savecnt = cnt;
    u_char *savecp = cp;

    if (altpattern < 0) {               /* specified alt pattern, compare it */

        while (cnt > 0) {
            if (cnt >= aplen) {
                if (BCMP((char *) cp, (char *) apat, aplen))
                    rc = 2;
            } else
                if (BCMP((char *) cp, (char *) apat, cnt))
                    rc = 2;
            if (rc == 0) {
                cnt -= aplen;
                cp  += aplen;
            } else {                    /* had error, find bad byte position */
                register u_char *pat = apat;
                while (*cp++ == *pat++)        /* "know" this will terminate */
                    cnt--;                     /*   since bcmp() said so ... */
                cp--;                   /* fix up for error message below */
                c = *(pat - 1);
                cnt--;
                break;
            }
        }

    } else {                            /* simple byte incrementing */
        c = *cp;
        if ((c > maxchar) || (c < minchar)) {
            (void) fprintf(stderr,"%s: RECEIVE ERROR: bad first byte; data=0x", cmd);
            dumphex(cp, (cnt < 10 ? cnt : 10));
            if (cnt > 10)
                (void) fprintf(stderr,"...");
            (void) fprintf(stderr,"\n");
            (void) fflush(stderr);
            rc = 1;
        } else {                        /* got starting char, adjust things */
            cnt--;                      /*  and then compare remaining ones */
            while (cnt-- > 0)  {
                ++cp;
                if (++c > maxchar) c = minchar;
                if (*cp != c) {
                                        /* farble for real sockets: allow */
                                        /* restarting sequence at any point */
                                        /*  (kernel buffering hides EOM!) */
                    if (*cp >= 0x20 && *cp <= 0x27) { /* (range 20-27 is for */
                        c = *cp;                       /* sender using -?s) */
                    } else {
                        rc = 2;
                        break;
                    }
                }
            }
        }
    }
    if (rc == 2) {                      /* if had bad byte tell user */
        int i;
        (void) fprintf(stderr,"%s: RECEIVE ERROR: expected 0x", cmd);
        dumphex((u_char *) &c, 1);
        (void) fprintf(stderr,", received 0x");
        dumphex(cp, 1);
        (void) fprintf(stderr,"; error context:\n");
        (void) fprintf(stderr,"%s:   0x", cmd);
        i = (cp - savecp < 16) ? (cp - savecp) : 16;
        dumphex(cp - i, i);
        (void) fprintf(stderr,"[");
        dumphex(cp, 1);
        (void) fprintf(stderr,"]");
        dumphex(cp + 1, (cnt < 16 ? cnt : 16));
        (void) fprintf(stderr,"\n");
        (void) fflush(stderr);
    }
                                        /* if still ok check 1st unread byte */
    if ((rc == 0) && checkextra && *(++cp)) {   /* (but not if skipping ...) */
        int n = 0;                      /*  .. oops, nonzero, report error */
        u_char *last = cp - 1;
        while (*cp++)
            n++;
        (void) fprintf(stderr,"%s: RECEIVE ERROR: %d extra byte(s) read, transition=", cmd, n);
        dumphex(last-3, 4);
        (void) fprintf(stderr,":");
        dumphex(last+1, 4);
        (void) fprintf(stderr,"\n");
        (void) fflush(stderr);
        rc = 3;
    }
    if (rc) {
        if (ranlen)
            (void) fprintf(stderr,"%s: last read was for %d bytes, returned %d, error on byte %d\n",
                           cmd, xlen, savecnt, savecnt - cnt);
        else
            (void) fprintf(stderr,"%s: last read returned %d bytes, error on byte %d\n",
                           cmd, savecnt, savecnt - cnt);
    }
    checkcnt = 0;                       /* set up for next checking interval */
    return(rc);
}


/*===========================================================================*/

/*
 *                      P R E P _ T I M E R
 */
static void
prep_timer()
{
    struct tms myTimes;
    int instance = 0;
    int rc;
    int fd;

    clk = time((S32 *)0);
    (void) fprintf(stderr,"%s: transfer begin: %s", cmd, ctime((S32 *)&clk));
    if (markfreq)
        (void) fprintf(stderr,"%s: ", cmd);
    (void) fflush(stderr);
    cput  = (double) clock() * 0.000001;
    realt = times(&myTimes);

    /* clear the Alteon NIC profiling */
    if (clearProf) {
	fd = open(ALTDEV, O_RDWR);
	if (fd < 0) {
	    perror("open failed");
	    exit(1);
	}
	
	/* for streams we've got to specify which hardware device */
#if defined(ALT_STREAM)
	rc = ioctl(fd, ALT_ATTACH, instance);
	if (rc < 0) {
	    perror("ioctl ALT_ATTACH failed");
	    exit(1);
	}
#endif /* defined(ALT_STREAM) */
	
	if (ioctl(fd, ALT_CLEAR_PROFILE) < 0) {
	    perror("couldn't do ALT_CLEAR_PROFILE");
	    exit(1);
	}

	(void) close(fd);
    }
}

/*
 *                      R E A D _ T I M E R
 *
 */
/*ARGSUSED*/
static void
read_timer(char *str, int len)
{
    struct tms myTimes;

    *str = '\0';                        /* no "stats" available on System V */
    cput  = (double) clock() * 0.000001 - cput;
    realt = times(&myTimes) - realt;
}

/*===========================================================================*/

static int
Nrecv(int s, char *bp, int count)
{
    struct sockaddr_in from;
    int len = sizeof(from);
    return (recvfrom(s, (void *)bp, count, 0, (void *) &from, &len));
}

/*---------------------------------------------------------------------------*/

static int
Nread(int s, char *bp, int count)
{
    register int cnt;
    if (b_flag)
        cnt = mread(s, bp,(unsigned) count);     /* fill buf */
    else
        cnt = read(s, (void *)bp, (size_t)count);
    return (cnt);
}


/*---------------------------------------------------------------------------*/

static int
Nsend(int s, char *cp, int count)
{
    register int cnt;

    while (((cnt = sendto(s, (void *)cp, count, 0, (void *) &sinhim,
                          sizeof(sinhim))) < 0) && (errno == ENOBUFS)) {
        delay(18000);
        errno = 0;
    }
    return (cnt);
}

/*---------------------------------------------------------------------------*/

static int
Nwrite(int s, char *cp, int count)
{
    int   rc;
    rc = write(s, (void *)cp, (size_t)count);
    return(rc);
}


/*---------------------------------------------------------------------------*/
/*
 *    DELAY  --  delay some number of microseconds
 */

static void
delay(int us)
{
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = us;
    (void) select(1, (SEL_TYPE *)0, (SEL_TYPE *)0, (SEL_TYPE *)0, 
		  (struct timeval *) &tv);
}

/*---------------------------------------------------------------------------*/

/*
 *                      M R E A D
 *
 * This function performs the function of a read(II) but will
 * call read(II) multiple times in order to get the requested
 * number of characters.  This can be necessary because
 * network connections don't deliver data with the same
 * grouping as it is written with.  Written by Robert S. Miles, BRL.
 */
static int
mread(int s, char *bfrp, unsigned n)
{
    register unsigned   count = 0;
    register int                nread;

    do {
        nread = read(s, (void *)bfrp, (size_t) (n - count));
        if (nread < 0)  {
            perror("tsock_mread");
            return (-1);
        }
        if (nread == 0)
            return ((int) count);
        count += (unsigned) nread;
        bfrp += nread;
    } while (count < n);

    return ((int) count);
}



/*---------------------------------------------------------------------------*/
/*
 *    CMPPAT  --  compare two buffers; return 0 if bufs equal, non-0 otherwise
 */

static int
cmppat(u_char *buf, u_char *ref_buf, int count)
{
    return(memcmp((void *)buf, (void *)ref_buf, (size_t)count));
}

/*---------------------------------------------------------------------------*/
#ifdef __hpux
/*
 *   VALLOC  -- allocate memory on page boundaries
 */
static char *valloc(size_t size)
{
    char *mem;

    mem = (char *)malloc(size+4096);
    if (mem)
	return((void *)(((int)mem + 4095) & ~4095));
    else
	return((char *)0);
}
#endif /* __hpux */





