/*
 * COPYRIGHT NOTICE
 * Copyright (c) Alteon Networks, Inc. 1996
 * All rights reserved
 */
/*
 * HISTORY
 * $Log: altstream.c,v $
 * Revision 1.7.6.53  1999/02/03  01:05:12  fhwang
 * 	to correct some trace message content, and put comment in inforeq routine for
 * 	future revisit on MTU size, and jumbo frame ether_type should not be 0.
 * 	[1999/02/03  00:51:06  fhwang]
 *
 * Revision 1.7.6.52  1999/01/30  02:32:29  fhwang
 * 	add llc support
 * 	[1999/01/30  00:42:03  fhwang]
 * 
 * Revision 1.7.6.51  1998/11/09  19:33:56  ted
 * 	support jumbos on x86 Solaris (test extended recv BD)
 * 	[1998/11/09  19:33:26  ted]
 * 
 * Revision 1.7.6.50  1998/10/16  21:31:42  ted
 * 	Fix Sun 16/32 bit bug in htons/ntohs
 * 	[1998/10/16  21:31:19  ted]
 * 
 * Revision 1.7.6.49  1998/10/15  20:59:57  ted
 * 	Fix endian problem for VLANs
 * 	[1998/10/15  20:57:57  ted]
 * 
 * Revision 1.7.6.48  1998/10/01  23:21:40  ted
 * 	Goose timer, Tigon 2 times out on x86 Solaris
 * 	[1998/10/01  23:18:39  ted]
 * 
 * Revision 1.7.6.47  1998/09/03  23:25:44  hayes
 * 	change FLOW_CONTROLLED flag to flow_controlled bool
 * 	[1998/09/03  23:25:31  hayes]
 * 
 * Revision 1.7.6.46  1998/09/02  23:51:04  ted
 * 	Fix 802.3 appletalk bug from Sun
 * 	[1998/09/02  23:50:50  ted]
 * 
 * Revision 1.7.6.45  1998/08/03  17:11:47  hayes
 * 	finish update for API changes
 * 	[1998/08/03  17:08:04  hayes]
 * 
 * Revision 1.7.6.44  1998/06/08  23:56:39  ted
 * 	Fix tracing numbering
 * 	[1998/06/08  22:48:17  ted]
 * 
 * 	Snapshot NT generification phase 1
 * 	[1998/06/04  19:52:34  ted]
 * 
 * Revision 1.7.6.43  1998/04/14  21:42:51  ted
 * 	Support VLAN assist interface
 * 	[1998/04/14  21:38:20  ted]
 * 
 * Revision 1.7.6.42  1998/03/28  21:19:54  ted
 * 	Fixes for dual homing problems (PR 980)
 * 	[1998/03/28  21:18:38  ted]
 * 
 * Revision 1.7.6.41  1998/03/05  03:03:21  suresh
 * 	Fix for PR 928 (host panics with jumbo configs)
 * 	- Also an unrelated but reqd correction in altstream.c to make
 * 	  sure that we set the max_sdu to the tagged header in the VLAN case
 * 	  and the untagged in the non-vlan case. This was reversed..
 * 	[1998/03/05  03:03:06  suresh]
 * 
 * Revision 1.7.6.40  1998/03/04  00:01:56  ted
 * 	plug potential buffer leak
 * 	[1998/03/04  00:00:57  ted]
 * 
 * Revision 1.7.6.39  1998/03/03  22:23:58  ted
 * 	Implement faster dual homing code
 * 	[1998/03/03  22:22:56  ted]
 * 
 * Revision 1.7.6.38  1998/01/30  03:40:31  ted
 * 	remove old stupid way of doing link negotiation command to NIC
 * 	[1998/01/30  03:39:32  ted]
 * 
 * Revision 1.7.6.37  1998/01/20  04:39:35  hayes
 * 	add tigon Rev6 support
 * 	[1998/01/20  04:38:42  hayes]
 * 
 * Revision 1.7.6.36  1998/01/06  02:46:15  ted
 * 	Fix many multicast problems
 * 	[1998/01/06  02:45:25  ted]
 * 
 * Revision 1.7.6.35  1997/12/06  21:05:48  ted
 * 	Fix PR 615
 * 	[1997/12/06  21:05:16  ted]
 * 
 * Revision 1.7.6.34  1997/12/03  01:23:20  ted
 * 	Bring forward ODI fixes from release 2
 * 	[1997/12/03  01:14:45  ted]
 * 
 * Revision 1.7.6.33  1997/11/19  19:24:01  ted
 * 	Support Tigon and Tigon2 in the same driver
 * 	[1997/11/19  19:21:17  ted]
 * 
 * Revision 1.7.6.32  1997/11/19  00:50:14  ted
 * 	Pick up latest fixes from release 2.0
 * 	[1997/11/19  00:49:00  ted]
 * 
 * Revision 1.7.6.31  1997/11/14  19:17:29  hayes
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:48  hayes]
 * 
 * 	first checkin after mongo merge...
 * 	[1997/11/14  19:13:24  hayes]
 * 
 * Revision 1.7.6.30  1997/10/28  17:13:07  hayes
 * 	Add support for Processor B scratch pad access via gdb
 * 	[1997/10/28  01:57:24  hayes]
 * 
 * Revision 1.7.6.29  1997/10/21  00:13:33  ted
 * 	Bring up to release 2.0 functionality
 * 	[1997/10/20  23:01:30  ted]
 * 
 * Revision 1.7.25.1  1997/10/16  17:32:18  ted
 * 	Generification, initial ODI driver, some bug fixes, preparation for use of MEM WRITE INVALIDATE
 * 	[1997/10/16  17:28:50  ted]
 * 
 * Revision 1.7.6.28  1997/09/16  03:12:10  ted
 * 	Per VLAN statistics, change s to spl for better coding style
 * 	[1997/09/16  03:08:53  ted]
 * 
 * Revision 1.7.6.27  1997/09/11  18:06:45  ted
 * 	Mixing vlans, jumbos, and redundancy works, plus small fixes from release 1.1 and better NIC tracing
 * 	[1997/09/11  18:01:49  ted]
 * 
 * Revision 1.7.6.26  1997/08/23  18:11:40  ted
 * 	Merged with changes from 1.7.6.25
 * 	[1997/08/23  18:11:27  ted]
 * 
 * 	Fix up locking issues
 * 	[1997/08/23  18:08:22  ted]
 * 
 * 	Reduce tracing level for "normal" error case
 * 	[1997/08/03  21:36:32  ted]
 * 
 * Revision 1.7.6.25  1997/08/23  17:40:40  skapur
 * 	Solaris jumbogram support
 * 	[1997/08/23  17:39:14  skapur]
 * 
 * Revision 1.7.6.24  1997/08/01  21:07:40  ted
 * 	Add dual homing configuration support prior to ifconfiging interface
 * 	[1997/08/01  21:06:23  ted]
 * 
 * Revision 1.7.6.23  1997/07/29  01:13:52  ted
 * 	Merge by hand due to bsubmit core dump, sigh
 * 	[1997/07/29  01:13:18  ted]
 * 
 * Revision 1.7.6.22  1997/07/28  22:40:48  ted
 * 	Solaris dual homing support
 * 	[1997/07/28  21:40:51  ted]
 * 
 * Revision 1.7.6.21  1997/07/27  22:04:00  ted
 * 	Remove telltale comment
 * 	[1997/07/27  22:03:01  ted]
 * 
 * Revision 1.7.6.20  1997/07/25  23:00:00  ted
 * 	Merge fixes from ge_0_22 and add link negotiation support
 * 	[1997/07/25  22:33:49  ted]
 * 
 * 	Architectural performance change
 * 	[1997/07/24  21:46:02  ted]
 * 
 * 	Support flow control for STREAMS
 * 	[1997/07/19  20:37:50  ted]
 * 
 * Revision 1.7.6.19  1997/05/21  03:36:19  ted
 * 	Use pub_tg.h.
 * 	Minor change in tracing for IP_SANITY_CHECK
 * 	[1997/05/21  03:31:53  ted]
 * 
 * Revision 1.7.6.18  1997/05/03  05:00:39  ted
 * 	Fix up statistics
 * 	[1997/05/03  04:57:13  ted]
 * 
 * Revision 1.7.6.17  1997/04/12  00:11:56  ted
 * 	Merged Sbus code, some performance enhancements
 * 	[1997/04/12  00:08:24  ted]
 * 
 * Revision 1.7.6.16  1997/04/04  22:38:36  ted
 * 	Support TTP page.
 * 	[1997/04/04  22:17:29  ted]
 * 
 * Revision 1.7.6.15  1997/03/28  01:41:20  ted
 * 	Major merge of NDIS and ODI with mainstream Unix ports
 * 	[1997/03/28  01:38:32  ted]
 * 
 * Revision 1.7.6.14  1997/03/25  23:51:48  ted
 * 	Merged with changes from 1.7.6.13
 * 	[1997/03/25  23:51:20  ted]
 * 
 * 	Support in ioctls for reading and writing scratch pad areas
 * 	[1997/03/25  23:22:32  ted]
 * 
 * Revision 1.7.6.13  1997/03/24  08:34:44  ted
 * 	New event ring handling, producer lives in host memory, misc. other more modest changes
 * 	[1997/03/24  08:28:44  ted]
 * 
 * Revision 1.7.6.12  1997/03/17  08:48:03  skapur
 * 	Fixes to IP_SANITY length check (ifdefed out now).
 * 	Add ALT_CLEAR_STATS and ALT_CLEAR_NIC_STATS ioctl support.
 * 	[1997/03/17  08:41:53  skapur]
 * 
 * Revision 1.7.6.11  1997/03/03  05:17:20  ted
 * 	endian fixes, non-copydata mode fixes
 * 	[1997/03/03  05:16:42  ted]
 * 
 * Revision 1.7.6.10  1997/03/02  18:37:00  ted
 * 	Merged with changes from 1.7.6.9
 * 	[1997/03/02  18:36:25  ted]
 * 
 * 	redundancy support and a few bug fixes
 * 	[1997/03/02  18:26:08  ted]
 * 
 * Revision 1.7.6.9  1997/02/28  23:02:17  ted
 * 	Change parallel EEPROM accesses to flash accesses
 * 	[1997/02/28  23:02:05  ted]
 * 
 * Revision 1.7.6.8  1997/02/26  03:59:46  ted
 * 	Correct addressing problems (off by one) in TG_MEM_READ/TG_MEM_WRITE.
 * 	Fix last segment problem in firmware cache ioctl.
 * 	[1997/02/26  03:59:30  ted]
 * 
 * Revision 1.7.6.7  1997/02/24  20:05:32  ted
 * 	Remove debugging snap traces when receiving unexpected packets.
 * 	[1997/02/24  20:05:19  ted]
 * 
 * Revision 1.7.6.6  1997/02/22  00:20:56  ted
 * 	Make debugger a special stream that doesn't effect the HostState down call
 * 	[1997/02/22  00:12:22  ted]
 * 
 * Revision 1.7.6.5  1997/02/21  02:06:58  ted
 * 	Prepare data structures and code structures to support redundant operation
 * 	[1997/02/21  01:57:26  ted]
 * 
 * Revision 1.7.6.4  1997/02/15  04:41:24  ted
 * 	Generify locks
 * 	Remove power management cruft (per Nara)
 * 	[1997/02/15  04:38:33  ted]
 * 
 * Revision 1.7.6.3  1997/02/12  00:28:11  ted
 * 	Implement ndd support using code borrowed from Sun
 * 	[1997/02/12  00:27:50  ted]
 * 
 * Revision 1.7.6.2  1997/02/11  04:08:18  ted
 * 	- Memory fault protection on nic trace buffer if nic isn't up enough to have
 * 	set the values yet.
 * 	- Fix memory offsets so gdb works properly
 * 	[1997/02/11  04:06:06  ted]
 * 
 * Revision 1.7.6.1  1997/02/07  21:58:42  ted
 * 	Updates for AIX/NT/new Solaris options (fast DMA)
 * 	[1997/02/07  21:50:33  ted]
 * 
 * Revision 1.7.2.5  1996/11/24  04:02:40  ted
 * 	Update MEMFREE to latest macro shape
 * 	[1996/11/24  04:01:58  ted]
 * 
 * Revision 1.7.2.4  1996/11/20  20:04:50  ted
 * 	Redo include file structuring.
 * 	Deal with new gen_com level of structure.
 * 	[1996/11/20  19:58:30  ted]
 * 
 * Revision 1.7.2.3  1996/11/09  02:32:16  ted
 * 	Add calls to read and write the serial eeprom
 * 	[1996/11/09  02:30:25  ted]
 * 
 * Revision 1.7.2.2  1996/11/05  00:04:25  ted
 * 	Fix up for new 8 byte mailboxes
 * 	[1996/11/05  00:00:53  ted]
 * 
 * Revision 1.7.2.1  1996/11/01  18:25:47  ted
 * 	change rr_ to tg_ and put in proper copyright
 * 	[1996/10/30  23:05:59  ted]
 * 
 * $EndLog$
 */
/*
 * FILE altstream.c
 *
 * $Source: /projects/rcs/sw/ge/./nic/host/unix/streams/common/altstream.c,v $
 * $Revision: 1.7.6.53 $ $Locker:  $
 * $Date: 1999/02/03 01:05:12 $ $Author: fhwang $ $State: Exp $
 */

#ifndef __lint
static char rcsid[] = "$Header: /projects/rcs/sw/ge/./nic/host/unix/streams/common/altstream.c,v 1.7.6.53 1999/02/03 01:05:12 fhwang Exp $";
#endif /* __lint */

/* some debugging defines */
#define IP_SANITY_CHECK 0

/* 
 * streams driver interface file 
 */

/* general include files */
#include <sys/types.h>		/* generic types */
#include <sys/errno.h>		/* EXXXXX errors */
#include <sys/kmem.h>		/* KM_SLEEP, kmem_alloc ... */
#include <sys/cred.h>
#include <sys/dditypes.h>	/* ddi types */
#include <sys/ksynch.h>		/* kconfvar_t */
#include <sys/stropts.h>	/* must precede ddi.h (which defines min) */
#include <sys/ddi.h>		/* getminor */
#include <sys/ddidmareq.h>	/* ddi dma structs */
#include <sys/dlpi.h>		/* DL_UNATTACHED ... */
#include <sys/ioccom.h>		/* _IO, ... */
#include <sys/ethernet.h>	/* ETHERMTU... */
#include <sys/byteorder.h>	/* htons */

/* system specific include files */
#if defined(__sun)
#include <sys/strsun.h>		/* DB_TYPE, ... */
#include <sys/sunddi.h>		/* ddi_get_soft_state */
/* would normally include <inet/common.h> but there are conflicting macros */
typedef	int	(*pfi_t)();
/* would normally include <inet/led.h>, but there are conflicting macros */
typedef	struct msgb	*MBLKP;
#include <inet/nd.h>
#if IP_SANITY_CHECK
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#endif /* IP_SANITY_CHECK */

#endif /* defined(__sun) */

/* alteon include files */
#include "alt.h"
#include "alttypes.h"
#include "nic_conf.h"
#include "nic_api.h"
#if defined(ALT_STREAM)
#include "altstr.h"
#endif /* defined(ALT_STREAM) */
#include "altioctl.h"
#include "pub_tg.h"
#include "altos.h"
#include "althdr.h"
#include "altreg.h"		/* must follow althdr.h */
#include "altftns.h"
#include "altbpdu.h"
#include "altdrv.h"

/* forwards */
STATIC void altproto(queue_t *, mblk_t *);
STATIC void altinforeq(queue_t *, mblk_t *);
STATIC void altattreq(queue_t *, mblk_t *);
STATIC void altdetreq(queue_t *, mblk_t *);
STATIC void altbindreq(queue_t *, mblk_t *);
STATIC void altunbindreq(queue_t *, mblk_t *);
STATIC void altemreq(queue_t *, mblk_t *);
STATIC void altdmreq(queue_t *, mblk_t *);
STATIC void altpareq(queue_t *, mblk_t *);
STATIC void altspareq(queue_t *, mblk_t *);
STATIC void altudreq(queue_t *, mblk_t *);
STATIC void altponreq(queue_t *, mblk_t *);
STATIC void altpoffreq(queue_t *, mblk_t *);
STATIC void altsetipq(struct altinst *, ulong);
STATIC void altsendup(struct altinst *, mblk_t *, 
		      struct altstr *(*acceptfunc)(struct altstr *,
						   struct altinst *,
						   unsigned short, 
						   struct ether_addr *, 
						   U16));
STATIC mblk_t *altaddudind(struct altinst *, mblk_t *, struct ether_addr *,
			   struct ether_addr *, int, ulong);
STATIC struct altstr *altaccept(struct altstr *, struct altinst *, 
				unsigned short, struct ether_addr *, U16);
STATIC struct altstr *altpaccept(struct altstr *, struct altinst *, 
				 unsigned short, struct ether_addr *, U16);
STATIC altmcmatch(struct altstr *slp, struct ether_addr *addrp);
STATIC void altioctl(queue_t *, mblk_t *);
STATIC void altioc(queue_t *, mblk_t *);
STATIC void altdliochdrinfo(queue_t *, mblk_t *);
STATIC void altdlack(queue_t *, mblk_t *, ulong);
STATIC void altdlnak(queue_t *, mblk_t *, ulong, ulong, ulong);
STATIC void altdlbindack(queue_t *, mblk_t *, ulong, caddr_t, int, int, ulong);
STATIC void altdlpaack(queue_t *, mblk_t *, caddr_t, int);
STATIC void altdludbadind(queue_t *, mblk_t *, u_char *, 
			  u_long, u_long, u_long);
STATIC mblk_t *altmexchange(queue_t *, mblk_t *, int, unsigned char, long);
STATIC void altmerror(queue_t *, mblk_t *, int);
STATIC void altiocnak(queue_t *, mblk_t *, int, int);
STATIC void altiocack(queue_t *, mblk_t *, int, int);
STATIC void msgcopyout(queue_t *, mblk_t *, mblk_t *, int, caddr_t, caddr_t, 
		       bool);
STATIC void msgcopyout_swap(queue_t *, mblk_t *, mblk_t *, int, 
			    caddr_t, caddr_t, bool);
STATIC void msgcopyin(queue_t *, mblk_t *, mblk_t *, int, caddr_t);
STATIC void altdodetach(struct altstr *);
STATIC int altmsgsize(mblk_t *);

#if IP_SANITY_CHECK > 1
STATIC U16 udpcksum(struct ip *pip);
STATIC U16 tcpcksum(struct ip *pip);
#endif /* IP_SANITY_CHECK > 1 */


/*
 * Our DL_INFO_ACK template.
 */
static	dl_info_ack_t altinfoack = {
    DL_INFO_ACK,		/* dl_primitive */
    ETHERMTU,			/* dl_max_sdu, ETHER to start adjust later */
    0,				/* dl_min_sdu */
    ALTADDRL,			/* dl_addr_length */
    DL_ETHER,			/* dl_mac_type */
    0,				/* dl_reserved */
    0,				/* dl_current_state */
    -2,				/* dl_sap_length */
    DL_CLDLS,			/* dl_service_mode */
    0,				/* dl_qos_length */
    0,				/* dl_qos_offset */
    0,				/* dl_range_length */
    0,				/* dl_range_offset */
    DL_STYLE2,			/* dl_provider_style */
    sizeof (dl_info_ack_t),	/* dl_addr_offset */
    DL_VERSION_2,		/* dl_version */
    ETHERADDRL,			/* dl_brdcst_addr_length */
    sizeof (dl_info_ack_t) + ALTADDRL, /* dl_brdcst_addr_offset */
    0				/* dl_growth */
};

/*
 * Ethernet broadcast address definition.
 */
static struct ether_addr etherbroadcastaddr = {
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

#define	ALTSAPMATCH(sap, type, flags) ((sap == type) ? TRUE : \
	((flags & SLALLSAP) ? TRUE : \
	((sap <= ETHERMTU) && (sap > 0) && (type <= ETHERMTU)) ? TRUE : 0))

/* macros to clean up READ_SERVICE #ifdefs */
#if READ_SERVICE
#define CANPUTNEXT canput
#define PUTNEXT    putq
#else /* READ_SERVICE */
#define CANPUTNEXT canputnext
#define PUTNEXT    putnext
#endif /* READ_SERVICE */

#define SNAPMATCH(p) \
    ((((struct snap_ether_header *)(p))->llc_dsap == 0xaa) && \
        (((struct snap_ether_header *)(p))->llc_ssap == 0xaa) && \
        (((struct snap_ether_header *)(p))->llc_cntrl == 0x03) && \
        (((struct snap_ether_header *)(p))->snap_org[0] == 0) && \
        (((struct snap_ether_header *)(p))->snap_org[1] == 0) && \
        (((struct snap_ether_header *)(p))->snap_org[2] == 0))

#define TAGGEDSNAPMATCH(p) \
    ((((struct tagged_snap_ether_header *)(p))->llc_dsap == 0xaa) && \
        (((struct tagged_snap_ether_header *)(p))->llc_ssap == 0xaa) && \
        (((struct tagged_snap_ether_header *)(p))->llc_cntrl == 0x03) && \
        (((struct tagged_snap_ether_header *)(p))->snap_org[0] == 0) && \
        (((struct tagged_snap_ether_header *)(p))->snap_org[1] == 0) && \
        (((struct tagged_snap_ether_header *)(p))->snap_org[2] == 0))

/*
 * ALTOPEN
 *	STREAMS device driver open routine.
 *	Returns 0 for success otherwise an error from errno.h
 */
int
ALTOPEN(queue_t *rqp, dev_t *devp, int oflag, int sflag, cred_t *crp)
{
    int minordev;
    struct altinst *altp;
    struct altstr *slp, **prevslp;
    int rc = 0;

    ALTTRACE(alt_open, 0x50001, rqp, devp, oflag, sflag);
    
    /*
     * Serialize all driver open and closes.
     */
    rw_enter(&altstruplock, RW_WRITER);

    /*
     * Determine minor device number.
     */
    prevslp = &altstrup;
    if (sflag == CLONEOPEN) {
	minordev = 0;
	for (; (slp = *prevslp); prevslp = &slp->sl_nextp) {
	    if (minordev < slp->sl_minor)
		break;
	    minordev++;
	}
	*devp = makedevice(getmajor(*devp), minordev);
    } else 
	minordev = getminor(*devp);
    
    /* are we already open, we're done */
    if (!rqp->q_ptr) {
	MEMZALLOC(slp, struct altstr *, sizeof(struct altstr), ALT_SLEEP);
	if (slp) {
	    slp->sl_minor = minordev;
	    slp->sl_rq = rqp;
	    slp->sl_wq = WR(rqp);
	    slp->sl_state = DL_UNATTACHED;
	    slp->sl_sap = 0;
	    slp->sl_flags = 0;
	    slp->sl_altp = (struct altinst *)0;
	    slp->sl_mccount = 0;
	    LOCK_INIT(&slp->sl_lock, "alteon stream lock", 
		       MUTEX_DRIVER, (void *)0);
	    
	    /*
	     * Link new entry into the list of active entries.
	     */
	    slp->sl_nextp = *prevslp;
	    *prevslp = slp;
	    
	    rqp->q_ptr = WR(rqp)->q_ptr = (void *)slp;
	    /*
	     * Disable automatic enabling of our write service procedure.
	     * We control this explicitly.
	     */
	    noenable(WR(rqp));
	} else
	    rc = ENOMEM;
    }
	
    /* turn on the queues */
    rw_exit(&altstruplock);
    qprocson(rqp);

    return(rc);
}


/*
 * ALTCLOSE
 *	Character device driver close routine.
 *	Returns 0 for success otherwise an error from errno.h
 */
int
ALTCLOSE(queue_t *rqp)
{
    struct altstr *slp, **prevslp, *newslp;
    bool promisc = FALSE;
    struct altinst *altp;
    
    qprocsoff(rqp);
    
    slp = (struct altstr *)rqp->q_ptr;
    altp = slp->sl_altp;
    
    ALTTRACE(altclose, 0x50002, rqp, rqp->q_ptr, slp->sl_minor, altp);

    /*
     * Implicit detach Stream from interface.
     */
    if (slp->sl_altp)
	altdodetach(slp);
    
    rw_enter(&altstruplock, RW_WRITER);

    /*
     * Unlink the per-Stream entry from the active list and free it.
     */
    for (prevslp = &altstrup; (slp = *prevslp);
	 prevslp = &slp->sl_nextp)
	if (slp == (struct altstr *)rqp->q_ptr)
	    break;
    ASSERT(slp);
    *prevslp = slp->sl_nextp;
    
    LOCK_DESTROY(&slp->sl_lock);
    MEMFREE((char *)slp, sizeof (struct altstr), ALT_SLEEP);
    
    rqp->q_ptr = WR(rqp)->q_ptr = (void *)0;

    /*
     * Determine if promiscuous mode. If firmware is up.
     */
    if (altp && altp->state == ALT_FW_RUNNING) {
	for (newslp = altstrup; newslp; newslp = newslp->sl_nextp)
	    if (newslp->sl_altp == altp && (newslp->sl_flags & SLALLPHYS))
		promisc = TRUE;

	/* now set the bit and let the firmware know */
	if (promisc) {
	    altp->flags |= ALT_PROMISC;
	    altSetPromiscuous(altp, TRUE);
	    /* if we're redundant do same thing to other interface */
	    if (altp->flags & ALT_REDUNDANT) {
		altp->attachNic->initAltp->flags |= ALT_PROMISC;
		altSetPromiscuous(altp->attachNic->initAltp, TRUE);
	    }
	} else {
	    if (altp->flags & ALT_PROMISC)
		altSetPromiscuous(altp, FALSE);
	    altp->flags &= ~ALT_PROMISC;
	    /* if we're redundant do same thing to other interface */
	    if (altp->flags & ALT_REDUNDANT) {
		if (altp->flags & ALT_PROMISC)
		    altSetPromiscuous(altp->attachNic->initAltp, FALSE);
		altp->attachNic->initAltp->flags &= ~ALT_PROMISC;
	    }
	}
    }

    rw_exit(&altstruplock);
    return(0);
}

/*
 * altwput
 *      Handle any messages arriving on the write queue
 */
int
altwput(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp = (struct altstr *)wq->q_ptr;
    struct altinst *altp = slp->sl_altp;
    int len;
    caddr_t dp;
    mblk_t *curMp = mp;

    ALTTRACEG(altwput, 0x50035, wq, slp, altp, DB_TYPE(mp));
    ALTTRACEG(altwpt2, 0x50036, mp, mp->b_cont, 0, 0);
    switch (DB_TYPE(mp)) {
    case M_DATA: 
	if (altTraceWholePacket)
	    while (curMp) {
		len = curMp->b_wptr - curMp->b_rptr;
		ALTTRACEV(altwpda, 0x50003, altp, curMp, curMp->b_rptr, len);
		dp = (caddr_t)((int)curMp->b_rptr & ~3);
		while (len >= 40) {
		    ALTTRACE2V(altwpda2, 0x50003, *(U32 *)dp, *(U32 *)(dp+4), 
			       *(U32 *)(dp + 8), *(U32 *)(dp + 12), dp+16);
		    len -= 40;
		    dp += 40;
		}
		/* handle last trace buffer(s) carefully, could run off end */
		if (len > 0) {
		    if (len <= 4) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 0, 0, 0);
		    } else if (len <= 8) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 0, 0);
		    } else if (len <= 12) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), *(U32 *)(dp+8), 0);
		    } else if (len <= 16) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), *(U32 *)(dp+8), 
				  *(U32 *)(dp+12));
		    } else if (len <= 20) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 0, 0, 0);
		    } else if (len <= 24) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 
				  *(U32 *)(dp+20), 
				  0, 0);
		    } else if (len <= 28) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 
				  *(U32 *)(dp+20), 
				  *(U32 *)(dp+24), 0);
		    } else if (len <= 32) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 
				  *(U32 *)(dp+20), 
				  *(U32 *)(dp+24), *(U32 *)(dp+28));
		    } else if (len <= 36) {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 
				  *(U32 *)(dp+20), 
				  *(U32 *)(dp+24), *(U32 *)(dp+28));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+32), 0, 0, 0);
		    } else {
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)dp, 
				  *(U32 *)(dp+4), 
				  *(U32 *)(dp+8), *(U32 *)(dp+12));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+16), 
				  *(U32 *)(dp+20), 
				  *(U32 *)(dp+24), *(U32 *)(dp+28));
			ALTTRACEV(altwpda2, 0x50003, *(U32 *)(dp+32), 
				  *(U32 *)(dp+36),
				  0, 0);
		    }
		}
		curMp = curMp->b_cont;
	    }
	else {
	    ALTTRACE2(altwpda, 0x50003, altp, mp, mp->b_rptr, 
		      mp->b_wptr - mp->b_rptr, mp->b_rptr);
	}

	ALTTRACEE(llchdr, 0x50042, 0, mp->b_wptr - mp->b_rptr, altmsgsize(mp),
	    ((struct ether_header *)mp->b_rptr)->ether_type);

	if (altp->flags & ALT_VLANS_ACTIVE) {
	    /* hack an ethertype of 0x800 in there, we know it should
               be, XXX */
	    if (!(slp->sl_flags & SLRAW))
		if (altp->initNic->flags & ALT_SNAP) {
		    ((struct tagged_snap_ether_header *)mp->b_rptr)->snap_type
		      =	htons(slp->sl_sap);
		    /* fill in length - including SNAP header */
		    ((struct tagged_snap_ether_header *)mp->b_rptr)->ether_type
		      =	altmsgsize(mp) - sizeof(struct tagged_ether_header); 
		} else
		    ((struct tagged_ether_header *)mp->b_rptr)->ether_type =
			htons(slp->sl_sap);
	} else {
	    /* hack an ethertype of 0x800 in there, we know it should
               be, XXX */
	    if (!(slp->sl_flags & SLRAW))
		if (altp->initNic->flags & ALT_SNAP) {
		    ((struct snap_ether_header *)mp->b_rptr)->snap_type = 
			htons(slp->sl_sap);
		    /* fill in length - including SNAP header */
		    ((struct snap_ether_header *)mp->b_rptr)->ether_type =
			altmsgsize(mp) - sizeof(struct ether_header);
		} else
		    ((struct ether_header *)mp->b_rptr)->ether_type = 
			htons(slp->sl_sap);
	}
	ALTTRACEE(llchdr, 0x50042, 0, mp->b_wptr - mp->b_rptr, altmsgsize(mp),
	    ((struct ether_header *)mp->b_rptr)->ether_type);

	if (!(slp->sl_flags & (SLFAST | SLRAW)) ||
	    slp->sl_state != DL_IDLE ||
	    !altp) {
	    altmerror(wq, mp, EPROTO);
	} else if (wq->q_first || (altp->flags & ALT_PROMISC)) {
	    /*
	     * If any msgs already enqueued or the interface will
	     * loop back up the message (due to ALT_PROMISC), then
	     * enqueue the msg.  Otherwise just xmit it directly.
	     */
	    ALTTRACE(mdatputq, 0x50039, altp, mp, wq, 0);
	    (void) putq(wq, mp);
	    qenable(wq);
	} else if (!ALTSTART(altp, mp)) {
	    ALTTRACE(mdatpqns, 0x50037, altp, mp, wq, 0);
	    altp->nic->flowControls++;
	    (void) putq(wq, mp);
	    qenable(wq);
	}
	break;
	
    case M_PROTO:
    case M_PCPROTO:
	altproto(wq, mp);
	break;
    
    case M_IOCTL:
	altioctl(wq, mp);
	break;

    case M_IOCDATA:
	altioc(wq, mp);
	break;

    case M_FLUSH:
	ALTTRACE(altwpfl, 0x50005, mp, altp, DB_TYPE(mp), (int)(*mp->b_rptr));
	if (*mp->b_rptr & FLUSHW) {
	    flushq(wq, FLUSHALL);
	    *mp->b_rptr &= ~FLUSHW;
	}
	if (*mp->b_rptr & FLUSHR)
	    qreply(wq, mp);
	else
	    freemsg(mp);
	break;
	
    default:
	ALTTRACE(altwpunk, 0x50006, wq, mp, altp, DB_TYPE(mp));
	freemsg(mp);
	break;
    }

    return (0);
}

/* if we're on this queue it's because we ran out of space, 
 * try to empty it as quickly as possible, no messages should get
 * here except M_DATA messages and M_PROTO or M_PCPROTO messages*/
int
altwsrv(queue_t *wq)
{
    struct altstr *slp = (struct altstr *)wq->q_ptr;
    struct altinst *altp = slp->sl_altp;
    mblk_t *mp;

    /* debug_enter("in altwsrv"); */
    ALTTRACE(altwsrv1, 0x50038, wq, altp, slp, 0);
    while (mp = getq(wq)) {
	ALTTRACE(altwsrv, 0x50033, wq, mp, altp, DB_TYPE(mp));
	switch (DB_TYPE(mp)) {
	case M_DATA:
	    if (altp) {
		if (!ALTSTART(altp, mp)) {
		    (void) putbq(wq, mp);
		    altp->flow_controlled = TRUE;
		    return(0);
		}
	    } else
		freemsg(mp);
	    break;

	case M_PROTO:
	case M_PCPROTO:
	    altproto(wq, mp);
	    break;
	    
	default:
	    ASSERT(0);
	    break;
	}
    }
    return (0);
}

/*
 * Start xmit on any msgs previously enqueued on any write queues.
 */
void
altenablewq(struct altinst *altp)
{
    struct altstr *slp;
    queue_t *wq;
    
    do {
	altp->flow_controlled = FALSE;
	for (slp = altstrup; slp; slp = slp->sl_nextp)
	    if (slp->sl_wq->q_first)
		qenable(slp->sl_wq);
    } while (altp->flow_controlled);
}

#if READ_SERVICE
int
altrsrv(queue_t *rq)
{
    mblk_t *mp;

    ALTTRACE(altrsrv, 0x50024, rq, 0, 0, 0);
    /* pluck all the stuff off the queue and discard it for now */
    while (mp = getq(rq)) {
	if (canputnext(rq))
	    (void) putnext(rq, mp);
	else {
	    /* couldn't push it up, put it back on our queue */
	    putbq(rq, mp);
	    ALTTRACE(flowrsrv, 0xf5004, 0, 0, 0, 0);
	    ((struct altstr *)rq->q_ptr)->sl_altp->flowControl++;
	    break;		/* can't do any more for now */
	}
    }
    return(0);
}
#endif /* READ_SERVICE */

STATIC void
altproto(queue_t *wq, mblk_t *mp)
{
    union DL_primitives *dlp;
    struct altstr *slp;
    ulong prim;

    slp = (struct altstr *)wq->q_ptr;
    dlp = (union DL_primitives *)mp->b_rptr;
    prim = dlp->dl_primitive;

    ALTTRACE(altwppt, 0x50004, mp, slp, dlp, prim);

    ALT_DISABLE(spl, &slp->sl_lock); /* ENTER CRITICAL SECTION */

    switch (prim) {
      case DL_INFO_REQ:
	altinforeq(wq, mp);
	break;

      case DL_ATTACH_REQ:
	altattreq(wq, mp);
	break;

      case DL_DETACH_REQ:
	altdetreq(wq, mp);
	break;

      case DL_PHYS_ADDR_REQ:
	altpareq(wq, mp);
	break;

      case DL_SET_PHYS_ADDR_REQ:
	altspareq(wq, mp);
	break;

      case DL_BIND_REQ:
	altbindreq(wq, mp);
	break;

      case DL_UNBIND_REQ:
	altunbindreq(wq, mp);
	break;

      case DL_UNITDATA_REQ:
	altudreq(wq, mp);
	break;

      case DL_ENABMULTI_REQ:
	altemreq(wq, mp);
	break;

      case DL_DISABMULTI_REQ:
	altdmreq(wq, mp);
	break;

      case DL_PROMISCON_REQ:
	altponreq(wq, mp);
	break;

      case DL_PROMISCOFF_REQ:
	altpoffreq(wq, mp);
	break;

      default:
	altdlnak(wq, mp, prim, DL_UNSUPPORTED, 0);
	break;
    }
    
    ALT_ENABLE(spl, &slp->sl_lock); /* LEAVE CRITICAL SECTION */
}

STATIC void
altinforeq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    dl_info_ack_t *dlip;
    struct alt_dladdr *dlap;
    struct ether_addr *ep;
    int size;
    struct altinst *altp;

    slp = (struct altstr *)wq->q_ptr;
    altp = slp->sl_altp;
    
    if (MBLKL(mp) < DL_INFO_REQ_SIZE) {
	altdlnak(wq, mp, DL_INFO_REQ, DL_BADPRIM, 0);
	return;
    }

    ALTTRACE(inforeq, 0x5003d, slp, altp, altp ? altp->flags : 0,
	     altp ? altp->initNic->flags : 0);

    /*
     * Exchange current msg for a DL_INFO_ACK.
     */
    size = sizeof(dl_info_ack_t) + ALTADDRL + ETHERADDRL;
    if ((mp = altmexchange(wq, mp, size, M_PCPROTO, DL_INFO_ACK)) == NULL)
	return;
    
    /*
     * Fill in the DL_INFO_ACK fields and reply.
     */
    dlip = (dl_info_ack_t *)mp->b_rptr;
    *dlip = altinfoack;		/* STRUCT COPY */
    dlip->dl_current_state = slp->sl_state;
    dlap = (struct alt_dladdr *)(mp->b_rptr + dlip->dl_addr_offset);
    dlap->dl_sap = slp->sl_sap;
    /*
     * The following calculation of mtu size need to be revisited
     */
    if (altp) {
	if (altp->initNic->flags & ALT_JUMBO_ENABLED) {
	    if (altp->flags & ALT_VLANS_ACTIVE) {
		if (altp->jumbo[slp->sl_instance/NALT])
		    if (altp->initNic->flags & ALT_SNAP) 
			dlip->dl_max_sdu = 9000 - 8; /* XXX */
		    else
			dlip->dl_max_sdu = 9000; /* XXX */
	    } else {
		if (altp->initNic->flags & ALT_SNAP) 
		    dlip->dl_max_sdu = 9000 - 8; /* XXX */
		else
		    dlip->dl_max_sdu = 9000; /* XXX */
	    }
	} else {
	    if (altp->flags & ALT_VLANS_ACTIVE) {
		if (altp->jumbo[slp->sl_instance/NALT])
		    if (altp->initNic->flags & ALT_SNAP) 
			dlip->dl_max_sdu = ETHERMTU - 8;
	    } else {
		if (altp->initNic->flags & ALT_SNAP) 
		    dlip->dl_max_sdu = ETHERMTU - 8;
	    }
	}
	ether_copy(&slp->sl_altp->macAddr, &dlap->dl_phys);
    } else {
	BZERO((caddr_t)&dlap->dl_phys, ETHERADDRL);
    }
    ALTTRACE(inforeq, 0x5013d, dlip->dl_max_sdu, 
	     sizeof(struct snap_ether_header), 0, 0);
    ep = (struct ether_addr *)(mp->b_rptr + dlip->dl_brdcst_addr_offset);
    ether_copy(&etherbroadcastaddr, ep);
    
    qreply(wq, mp);
}

STATIC void
geninfoack(queue_t *wq)
{
    struct altstr *slp;
    dl_info_ack_t *dlip;
    struct alt_dladdr *dlap;
    struct ether_addr *ep;
    int size;
    struct altinst *altp;
    mblk_t *mp;
    
    slp = (struct altstr *)wq->q_ptr;
    altp = slp->sl_altp;
    
    ALTTRACE(geninfoa, 0x5003c, slp, altp, altp ? altp->flags : 0,
	     altp ? altp->initNic->flags : 0);

    /*
     * Ge msg for a DL_INFO_ACK.
     */
    size = sizeof(dl_info_ack_t) + ALTADDRL + ETHERADDRL;
    if ((mp = allocb(size, BPRI_MED)) == NULL) {
	if (mp = allocb(1, BPRI_MED))
	    altmerror(wq, mp, ENOSR);
	return;
    }

    /* set it up */
    mp->b_datap->db_type = M_PCPROTO;
    mp->b_rptr = mp->b_datap->db_base;
    mp->b_wptr = mp->b_rptr + size;
    *(long *)mp->b_rptr = DL_INFO_ACK;
    
    /*
     * Fill in the DL_INFO_ACK fields and reply.
     */
    dlip = (dl_info_ack_t *)mp->b_rptr;
    *dlip = altinfoack;		/* STRUCT COPY */
    dlip->dl_current_state = slp->sl_state;
    dlap = (struct alt_dladdr *)(mp->b_rptr + dlip->dl_addr_offset);
    dlap->dl_sap = slp->sl_sap;
    if (altp) {
	if (altp->initNic->flags & ALT_JUMBO_ENABLED) {
	    if (altp->flags & ALT_VLANS_ACTIVE) {
		if (altp->jumbo[slp->sl_instance/NALT])
		    if (altp->initNic->flags & ALT_SNAP) 
			dlip->dl_max_sdu = 9000 - 8; /* XXX */
		    else
			dlip->dl_max_sdu = 9000; /* XXX */
	    } else {
		if (altp->initNic->flags & ALT_SNAP) 
		    dlip->dl_max_sdu = 9000 - 8; /* XXX */
		else
		    dlip->dl_max_sdu = 9000; /* XXX */
	    }
	} else {
	    if (altp->flags & ALT_VLANS_ACTIVE) {
		if (altp->jumbo[slp->sl_instance/NALT])
		    if (altp->initNic->flags & ALT_SNAP) 
			dlip->dl_max_sdu = ETHERMTU - 8;
	    } else {
		if (altp->initNic->flags & ALT_SNAP) 
		    dlip->dl_max_sdu = ETHERMTU - 8;
	    }
	}
	ether_copy(&slp->sl_altp->macAddr, &dlap->dl_phys);
    } else {
	BZERO((caddr_t)&dlap->dl_phys, ETHERADDRL);
    }
    ALTTRACE(geninfoa, 0x5013c, dlip->dl_max_sdu, 
	     sizeof(struct snap_ether_header), 0, 0);
    ep = (struct ether_addr *)(mp->b_rptr + dlip->dl_brdcst_addr_offset);
    ether_copy(&etherbroadcastaddr, ep);
    
    qreply(wq, mp);
}

STATIC void
altattreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    union DL_primitives *dlp;
    struct altinst *altp;
    int ppa;
    int i;
    ulong inst;

    slp = (struct altstr *)wq->q_ptr;
    dlp = (union DL_primitives *)mp->b_rptr;
    inst = dlp->attach_req.dl_ppa;

    ALTTRACE(altattrq, 0x5002b, dlp, inst, slp, 0);

    if (MBLKL(mp) < DL_ATTACH_REQ_SIZE) {
	altdlnak(wq, mp, DL_ATTACH_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state != DL_UNATTACHED) {
	altdlnak(wq, mp, DL_ATTACH_REQ, DL_OUTSTATE, 0);
	return;
    }

    /* if device instance is out of range fail */
    if (inst >= NALT * MAX_VLANS) {
	altdlnak(wq, mp, DL_ATTACH_REQ, DL_BADPPA, 0);
	return;
    }

    /* get the device instance from the ppa field */
    if (!(altp = (struct altinst *)
	  ddi_get_soft_state(altStatep, inst % NALT))) {
	altdlnak(wq, mp, DL_ATTACH_REQ, DL_BADPPA, 0);
	return;
    }

    /* is device running? fail if not */
    if (altp->state != ALT_FW_RUNNING) {
	for (i=0; i<10; i++) {
	    SNOOZE(1000000);
	    if (altp->state == ALT_FW_RUNNING) 
		break;
	}
	if (altp->state != ALT_FW_RUNNING) {
	    altdlnak(wq, mp, dlp->dl_primitive, DL_INITFAILED, 0);
	    return;
	}
    }

    /*
     * Set link to device and update our state.
     */
    slp->sl_instance = inst;
    slp->sl_altp = altp;
    slp->sl_state = DL_UNBOUND;

    /* reply to the attach request */
    altdlack(wq, mp, DL_ATTACH_REQ);

    /* send up an info ack in case we need to change the max_dl_sdu */
    geninfoack(wq);
}

STATIC void
altdetreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    
    slp = (struct altstr *)wq->q_ptr;
    
    if (MBLKL(mp) < DL_DETACH_REQ_SIZE) {
	altdlnak(wq, mp, DL_DETACH_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state != DL_UNBOUND) {
	altdlnak(wq, mp, DL_DETACH_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    altdodetach(slp);
    altdlack(wq, mp, DL_DETACH_REQ);
}

STATIC void
altbindreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    union DL_primitives *dlp;
    struct altinst *altp;
    struct alt_dladdr altaddr;
    u_long sap;
    int xidtest;
    
    slp = (struct altstr *)wq->q_ptr;
    
    if (MBLKL(mp) < DL_BIND_REQ_SIZE) {
	altdlnak(wq, mp, DL_BIND_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state != DL_UNBOUND) {
	altdlnak(wq, mp, DL_BIND_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    dlp = (union DL_primitives *)mp->b_rptr;
    altp = slp->sl_altp;
    sap = dlp->bind_req.dl_sap;
    xidtest = dlp->bind_req.dl_xidtest_flg;
    
    ASSERT(altp);
    
    if (xidtest) {
	altdlnak(wq, mp, DL_BIND_REQ, DL_NOAUTO, 0);
	return;
    }

    if (sap > ETHERTYPE_MAX) {
	altdlnak(wq, mp, dlp->dl_primitive, DL_BADSAP, 0);
	return;
    }

    /*
     * Has device been told to start processing packets?
     */
    if (!(altp->flags & ALT_HOST_UP)) {
	altHostState(altp, TRUE);
	altp->flags |= ALT_HOST_UP;
	if (altp->flags & ALT_REDUNDANT) {
	    altHostState(altp->attachNic->initAltp, TRUE);
	    altp->attachNic->initAltp->flags |= ALT_HOST_UP;
	}
    }
    
    /*
     * Save SAP value for this Stream and change state.
     */
    slp->sl_sap = sap;
    slp->sl_state = DL_IDLE;
    
    altaddr.dl_sap = sap;
    ether_copy(&altp->macAddr, &altaddr.dl_phys);
    altdlbindack(wq, mp, sap, (caddr_t) &altaddr, ALTADDRL, 0, 0);

    ALTTRACE(dlbind, 0x5001b, altp, sap, xidtest, 0);

    altsetipq(altp, slp->sl_instance);
}

STATIC void
altunbindreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    
    slp = (struct altstr *)wq->q_ptr;
    
    if (MBLKL(mp) < DL_UNBIND_REQ_SIZE) {
	altdlnak(wq, mp, DL_UNBIND_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state != DL_IDLE) {
	altdlnak(wq, mp, DL_UNBIND_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    slp->sl_state = DL_UNBOUND;
    slp->sl_sap = 0;
    
    (void) putnextctl1(RD(wq), M_FLUSH, FLUSHRW);
    altdlack(wq, mp, DL_UNBIND_REQ);
    
    altsetipq(slp->sl_altp, slp->sl_instance);
}

STATIC void
altemreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    struct altinst *altp;
    struct tg_nic *nic;
    union DL_primitives *dlp;
    struct ether_addr *addrp;
    int off;
    int len, i;
    U32 macspot[2];		/* space for the padded MAC address */
    
    slp = (struct altstr *)wq->q_ptr;
    altp = slp->sl_altp;
    nic = altp->nic;
    
    if (MBLKL(mp) < DL_ENABMULTI_REQ_SIZE) {
	altdlnak(wq, mp, DL_ENABMULTI_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state == DL_UNATTACHED) {
	altdlnak(wq, mp, DL_ENABMULTI_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    dlp = (union DL_primitives *)mp->b_rptr;
    len = dlp->enabmulti_req.dl_addr_length;
    off = dlp->enabmulti_req.dl_addr_offset;
    addrp = (struct ether_addr *)(mp->b_rptr + off);
    
    if ((len != ETHERADDRL) ||
	!MBLKIN(mp, off, len) ||
	((addrp->ether_addr_octet[0] & 01) == 0)) {
	altdlnak(wq, mp, DL_ENABMULTI_REQ, DL_BADADDR, 0);
	return;
    }
    
    if ((slp->sl_mccount + 1) >= MAX_MCAST_ENTRIES) {
	altdlnak(wq, mp, DL_ENABMULTI_REQ, DL_TOOMANY, 0);
	return;
    }

    /* stick the address where we can trace it */
    macspot[0] = 0;
    MEMCPY(((char *)macspot)+2, (MEMCPY_TYPE)addrp, 6);
    ALTTRACE(emreq, 0x5003e, slp->sl_mccount, macspot[0], macspot[1], 0);

    /*
     * Check to see if the address is already in the table.
     */
    for (i = 0; i < slp->sl_mccount; i++) {
	if (ether_cmp(&slp->sl_mctab[i], addrp) == 0) {
	    altdlack(wq, mp, DL_ENABMULTI_REQ);
	    return;
	}
    }
    
    slp->sl_mctab[slp->sl_mccount++] = *addrp;

    /* keep track of instance structure too */
    for (i = 0; i < altp->mccount; i++) {
	if (ether_cmp(&altp->mctab[i].addr, addrp) == 0) {
	    altp->mctab[i].count++;       /* a second stream asked */
	    altdlack(wq, mp, DL_ENABMULTI_REQ);
	    return;
	}
    }
    
    /* set new mcast address's addr and count in the instance table */
    altp->mctab[altp->mccount++].addr = *addrp;
    altp->mctab[i].count = 1;

    /* and tell the firmware about it */
    altAddMultiCast(altp, addrp);
    if (altp->flags & ALT_REDUNDANT)
	altAddMultiCast(altp->attachNic->initAltp, addrp);
    
    altdlack(wq, mp, DL_ENABMULTI_REQ);
}

STATIC void
altdmreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    struct altinst *altp;
    union DL_primitives *dlp;
    struct ether_addr *addrp;
    int off;
    int len;
    int i;
    bool foundIt = FALSE;
    U32 macspot[2];		/* space for the padded MAC address */
    
    slp = (struct altstr *)wq->q_ptr;
    altp = slp->sl_altp;
    
    if (MBLKL(mp) < DL_DISABMULTI_REQ_SIZE) {
	altdlnak(wq, mp, DL_DISABMULTI_REQ, DL_BADPRIM, 0);
	return;
    }
    
    if (slp->sl_state == DL_UNATTACHED) {
	altdlnak(wq, mp, DL_DISABMULTI_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    dlp = (union DL_primitives *)mp->b_rptr;
    len = dlp->disabmulti_req.dl_addr_length;
    off = dlp->disabmulti_req.dl_addr_offset;
    addrp = (struct ether_addr *)(mp->b_rptr + off);
    
    if ((len != ETHERADDRL) || !MBLKIN(mp, off, len)) {
	altdlnak(wq, mp, DL_DISABMULTI_REQ, DL_BADADDR, 0);
	return;
    }
    
    /* stick the address where we can trace it */
    macspot[0] = 0;
    MEMCPY(((char *)macspot)+2, (MEMCPY_TYPE)addrp, 6);
    ALTTRACE(dmreq, 0x5003f, slp->sl_mccount, macspot[0], macspot[1], 0);

    /*
     * Find the address in the multicast table for this Stream
     * and delete it by shifting all subsequent multicast
     * table entries over one.
     */
    for (i = 0; i < slp->sl_mccount; i++)
	if (ether_cmp(addrp, &slp->sl_mctab[i]) == 0) {
	    MEMCPY((MEMCPY_TYPE)&slp->sl_mctab[i],
		   (MEMCPY_TYPE)&slp->sl_mctab[i+1],
		   ((slp->sl_mccount - i) *
		    sizeof (struct ether_addr)));
	    slp->sl_mccount--;
	    foundIt = TRUE;
	    break;
	}
	    
    if (!foundIt)
	altdlnak(wq, mp, DL_DISABMULTI_REQ, DL_NOTENAB, 0);

    /* found it in the streams table, find it in the instance table
     * and remove it, but only if reference count goes to 0 */
    for (i = 0; i < altp->mccount; i++)
	if (ether_cmp(addrp, &altp->mctab[i].addr) == 0) {
	    if (--altp->mctab[i].count == 0) {
		altp->mccount--;
		MEMCPY((MEMCPY_TYPE)&altp->mctab[i],
		       (MEMCPY_TYPE)&altp->mctab[i+1],
		       ((altp->mccount - i) *
			sizeof (struct inst_multicast)));
		/* and tell the firmware about it */
		altDelMultiCast(altp, addrp);
		if (altp->flags & ALT_REDUNDANT)
		    altDelMultiCast(altp->attachNic->initAltp, addrp);
	    }
	    break;
	}

    altdlack(wq, mp, DL_DISABMULTI_REQ);
}

STATIC void
altpareq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    union DL_primitives *dlp;
    int type;
    struct altinst *altp;
    struct ether_addr addr;
    
    slp = (struct altstr *)wq->q_ptr;
    
    if (MBLKL(mp) < DL_PHYS_ADDR_REQ_SIZE) {
	altdlnak(wq, mp, DL_PHYS_ADDR_REQ, DL_BADPRIM, 0);
	return;
    }
    
    dlp = (union DL_primitives *)mp->b_rptr;
    type = dlp->physaddr_req.dl_addr_type;
    altp = slp->sl_altp;
    
    if (!altp) {
	altdlnak(wq, mp, DL_PHYS_ADDR_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    switch (type) {
      case DL_FACT_PHYS_ADDR:
      case DL_CURR_PHYS_ADDR:	/* we have to decide if we can set this */
	ether_copy(&altp->macAddr, &addr);
	break;
	
      default:
	altdlnak(wq, mp, DL_PHYS_ADDR_REQ, DL_NOTSUPPORTED, 0);
	return;
    }

    altdlpaack(wq, mp, (caddr_t) &addr, ETHERADDRL);
}

STATIC void
altspareq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    union DL_primitives *dlp;
    int off;
    int len;
    struct ether_addr *addrp;
    struct altinst *altp;

    slp = (struct altstr *)wq->q_ptr;

    if (MBLKL(mp) < DL_SET_PHYS_ADDR_REQ_SIZE) {
	altdlnak(wq, mp, DL_SET_PHYS_ADDR_REQ, DL_BADPRIM, 0);
	return;
    }

    dlp = (union DL_primitives *)mp->b_rptr;
    len = dlp->set_physaddr_req.dl_addr_length;
    off = dlp->set_physaddr_req.dl_addr_offset;

    if (!MBLKIN(mp, off, len)) {
	altdlnak(wq, mp, DL_SET_PHYS_ADDR_REQ, DL_BADPRIM, 0);
	return;
    }
    
    addrp = (struct ether_addr *)(mp->b_rptr + off);
    
    /*
     * Error if length of address isn't right or the address
     * specified is a multicast or broadcast address.
     */
    if ((len != ETHERADDRL) ||
	((addrp->ether_addr_octet[0] & 01) == 1) ||
	(ether_cmp(addrp, &etherbroadcastaddr) == 0)) {
	altdlnak(wq, mp, DL_SET_PHYS_ADDR_REQ, DL_BADADDR, 0);
	return;
    }
    
    /*
     * Error if this stream is not attached to a device.
     */
    if ((altp = slp->sl_altp) == NULL) {
	altdlnak(wq, mp, DL_SET_PHYS_ADDR_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    /*
     * Set new interface local address and tell firmware.
     * This is destructive to any other streams attached
     * to this device.
     */
    ether_copy(addrp, &altp->macAddr);
    /* set both interfaces for redundant case */
    if (altp->flags & ALT_REDUNDANT) {
	altSetMacaddr(altp->ifconfigNic, addrp);
	altSetMacaddr(altp->attachNic, addrp);
    } else {
	altSetMacaddr(altp->nic, addrp);
    }
    
    altdlack(wq, mp, DL_SET_PHYS_ADDR_REQ);
}

STATIC void
altudreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    struct altinst *altp;
    dl_unitdata_req_t *dludp;
    mblk_t *nmp;
    struct alt_dladdr *dlap;
    struct ether_header *headerp;
    struct snap_ether_header *sheaderp;
    struct tagged_ether_header *theaderp;
    struct tagged_snap_ether_header *tsheaderp;
    ulong off, len;
    ulong sap;
    int rc;
    mblk_t *curMp = mp->b_cont;
    caddr_t dp;
    int mpLen;
    int size;
    bool roomInFrame;

    /* there needs to be data to send else, nak it */
    if (!curMp) {
	altdlnak(wq, mp, DL_UNITDATA_REQ, DL_BADDATA, 0);
	return;
    }
    if (altTraceWholePacket)
	while (curMp) {
	    mpLen = curMp->b_wptr - curMp->b_rptr;
	    ALTTRACEV(altudreq, 0x5001e, wq, curMp, curMp->b_rptr, mpLen);
	    dp = (caddr_t)((int)curMp->b_rptr & ~3);
	    while (mpLen >= 40) {
		ALTTRACE2V(altudrq2, 0x5001e, *(U32 *)dp, *(U32 *)(dp+4), 
			   *(U32 *)(dp + 8), *(U32 *)(dp + 12), dp+16);
		mpLen -= 40;
		dp += 40;
	    }
	    /* handle last trace buffer(s) carefully, could run off end */
	    if (mpLen > 0) {
		if (mpLen <= 4) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 0, 0, 0);
		} else if (mpLen <= 8) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 0, 0);
		} else if (mpLen <= 12) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), *(U32 *)(dp+8), 0);
		} else if (mpLen <= 16) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), *(U32 *)(dp+8), 
			      *(U32 *)(dp+12));
		} else if (mpLen <= 20) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 0, 0, 0);
		} else if (mpLen <= 24) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 
			      *(U32 *)(dp+20), 
			      0, 0);
		} else if (mpLen <= 28) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 
			      *(U32 *)(dp+20), 
			      *(U32 *)(dp+24), 0);
		} else if (mpLen <= 32) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 
			      *(U32 *)(dp+20), 
			      *(U32 *)(dp+24), *(U32 *)(dp+28));
		} else if (mpLen <= 36) {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 
			      *(U32 *)(dp+20), 
			      *(U32 *)(dp+24), *(U32 *)(dp+28));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+32), 0, 0, 0);
		} else {
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)dp, 
			      *(U32 *)(dp+4), 
			      *(U32 *)(dp+8), *(U32 *)(dp+12));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+16), 
			      *(U32 *)(dp+20), 
			      *(U32 *)(dp+24), *(U32 *)(dp+28));
		    ALTTRACEV(altudrq2, 0x5001e, *(U32 *)(dp+32), 
			      *(U32 *)(dp+36),
			      0, 0);
		}
	    }
	    curMp = curMp->b_cont;
	}
    else
	ALTTRACE2V(altudreq, 0x5001e, wq, curMp, curMp->b_rptr, 
		   curMp->b_wptr - curMp->b_rptr, curMp->b_rptr);

    slp = (struct altstr *)wq->q_ptr;
    altp = slp->sl_altp;
    
    if (slp->sl_state != DL_IDLE) {
	altdlnak(wq, mp, DL_UNITDATA_REQ, DL_OUTSTATE, 0);
	return;
    }
    
    dludp = (dl_unitdata_req_t *)mp->b_rptr;
    
    off = dludp->dl_dest_addr_offset;
    len = dludp->dl_dest_addr_length;
    
    /*
     * Validate destination address format.
     */
    if (!MBLKIN(mp, off, len) || (len != ALTADDRL)) {
	altdlnak(wq, mp, DL_UNITDATA_REQ, DL_BADADDR, 0);
	return;
    }
    
    /*
     * Error if no M_DATA follows.
     */
    nmp = mp->b_cont;
    if (nmp == NULL) {
	altdlnak(wq, mp, DL_UNITDATA_REQ, DL_BADDATA, 0);
	return;
    }
    
    dlap = (struct alt_dladdr *)(mp->b_rptr + off);
    
    /* find out the size we're using */
    if (altp->flags & ALT_VLANS_ACTIVE)
	if (altp->initNic->flags & ALT_SNAP) 
	    size = sizeof(struct tagged_snap_ether_header);
	else
	    size = sizeof(struct tagged_ether_header);
    else 
	if (altp->initNic->flags & ALT_SNAP) 
	    size = sizeof(struct snap_ether_header);
	else
	    size = sizeof(struct ether_header);

    /*
     * Create ethernet header by either prepending it onto the
     * next mblk if possible, or reusing the M_PROTO block if not.
     */
    ALTTRACEE(llchdr0, 0x50042, size, mp->b_wptr - mp->b_rptr, 
	      altmsgsize(mp), 0);
    if ((DB_REF(nmp) == 1) &&
	(MBLKHEAD(nmp) >= size) &&
	(((ulong)nmp->b_rptr & 0x1) == 0)) {
	nmp->b_rptr -= size;
	roomInFrame = TRUE;
    } else {
	DB_TYPE(mp) = M_DATA;
	mp->b_wptr = mp->b_rptr + size;
	roomInFrame = FALSE;
	nmp = mp;
    }
    ALTTRACEE(llchdr1, 0x50042, size, nmp->b_wptr - nmp->b_rptr, 
	      altmsgsize(nmp), roomInFrame);

    sap = dlap->dl_sap;		/* get SAP from the udreq block */
    if (altp->flags & ALT_VLANS_ACTIVE) {
	if (altp->initNic->flags & ALT_SNAP) {
	    tsheaderp = (struct tagged_snap_ether_header *)nmp->b_rptr;

	    ether_copy(&dlap->dl_phys, &tsheaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &tsheaderp->ether_shost);

	    /* fill in the VLAN fields */
	    tsheaderp->tag_id = htons(altVlanTagId);
	    tsheaderp->vlan_id = htons(altp->vlan_id[slp->sl_instance/NALT]);

	    /* LLC fields */
	    tsheaderp->ether_type = altmsgsize(nmp) - 
	      sizeof(struct tagged_ether_header);
	    /*
	     * make sure the length field is valid, for jumbos, use 0
	     */
	    if (tsheaderp->ether_type > ETHERMTU)
		tsheaderp->ether_type = 0;

	    tsheaderp->llc_dsap = 0xaa;
	    tsheaderp->llc_ssap = 0xaa;
	    tsheaderp->llc_cntrl = 0x03;

	    /* SNAP fields */
	    tsheaderp->snap_org[0] = 0;
	    tsheaderp->snap_org[1] = 0;
	    tsheaderp->snap_org[2] = 0;
	    /*
	     * sap value in the range [0-1500] are treated equivalent
	     * and represent a desire to use 802.3 mode. Therefore compute
	     * the length and put it in the type field in header.  But assume
	     * the one passed is correct unless it's 0
	     */
	    if (sap == 0) {
		/* XXX */
		sap = altmsgsize(mp) - sizeof(struct tagged_ether_header);
		tsheaderp->snap_type = htons(sap);
	    } else 
		tsheaderp->snap_type = htons(sap);
	} else {
	    theaderp = (struct tagged_ether_header *)nmp->b_rptr;

	    ether_copy(&dlap->dl_phys, &theaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &theaderp->ether_shost);
	    /* fill in the VLAN fields */
	    theaderp->tag_id = htons(altVlanTagId);
	    theaderp->vlan_id = htons(altp->vlan_id[slp->sl_instance/NALT]);
	    /*
	     * sap value in the range [0-1500] are treated equivalent
	     * and represent a desire to use 802.3 mode. Therefore compute
	     * the length and put it in the type field in header.  But assume
	     * the one passed is correct unless it's 0
	     */
	    if (sap == 0) {
		sap = altmsgsize(mp) - sizeof(struct tagged_ether_header);
		theaderp->ether_type = htons(sap);
	    } else 
		theaderp->ether_type = htons(sap);
	}
    } else {
	if (altp->initNic->flags & ALT_SNAP) {
	    sheaderp = (struct snap_ether_header *)nmp->b_rptr;

	    ether_copy(&dlap->dl_phys, &sheaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &sheaderp->ether_shost);

	    sheaderp->ether_type = altmsgsize(nmp) - 
	      sizeof(struct ether_header);
	    /*
	     * make sure the length field is valid, for jumbos, use 0
	     */
	    if (sheaderp->ether_type > ETHERMTU)
		sheaderp->ether_type = 0;

	    /* LLC fields */
	    sheaderp->llc_dsap = 0xaa;
	    sheaderp->llc_ssap = 0xaa;
	    sheaderp->llc_cntrl = 0x03;

	    /* SNAP fields */
	    sheaderp->snap_org[0] = 0;
	    sheaderp->snap_org[1] = 0;
	    sheaderp->snap_org[2] = 0;
	    /*
	     * sap value in the range [0-1500] are treated equivalent
	     * and represent a desire to use 802.3 mode. Therefore compute
	     * the length and put it in the type field in header.  But assume
	     * the one passed is correct unless it's 0
	     */
	    if (sap == 0) {
		/* XXX */
		sap = altmsgsize(mp) - sizeof(struct ether_header);
		sheaderp->snap_type = htons(sap);
	    } else 
		sheaderp->snap_type = htons(sap);
	} else {
	    headerp = (struct ether_header *)nmp->b_rptr;

	    ether_copy(&dlap->dl_phys, &headerp->ether_dhost);
	    ether_copy(&altp->macAddr, &headerp->ether_shost);
	    /*
	     * sap value in the range [0-1500] are treated equivalent
	     * and represent a desire to use 802.3 mode. Therefore compute
	     * the length and put it in the type field in header.  But assume
	     * the one passed is correct unless it's 0
	     */
	    if (sap == 0) {
		sap = altmsgsize(mp) - sizeof(struct ether_header);
		headerp->ether_type = htons(sap);
	    } else 
		headerp->ether_type = htons(sap);
	}
    }

    /* we had room in the same mblk, have to free up M_PROTO */
    if (roomInFrame) {
	freeb(mp);
	mp = nmp;
    }

    /* if couldn't start put it on the write queue for later */
    if (!ALTSTART(altp, mp)) {
	ALTTRACE(udreqptq, 0x5003b, altp, mp, wq, 0);
	(void) putq(wq, mp);
	qenable(wq);
    }
}

STATIC void
altponreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;

    slp = (struct altstr *)wq->q_ptr;

    if (MBLKL(mp) < DL_PROMISCON_REQ_SIZE) {
	altdlnak(wq, mp, DL_PROMISCON_REQ, DL_BADPRIM, 0);
	return;
    }

    ALTTRACE(promon, 0x50023, wq, mp, 
	     ((dl_promiscon_req_t *)mp->b_rptr)->dl_level, 0);

    switch (((dl_promiscon_req_t *)mp->b_rptr)->dl_level) {
      case DL_PROMISC_PHYS:
	slp->sl_flags |= SLALLPHYS;
	break;

      case DL_PROMISC_SAP:
	slp->sl_flags |= SLALLSAP;
	break;
	
      case DL_PROMISC_MULTI:
	slp->sl_flags |= SLALLMULTI;
	break;
	
      default:
	altdlnak(wq, mp, DL_PROMISCON_REQ, DL_NOTSUPPORTED, 0);
	return;
    }

    /* tell the hardware, for physical promiscuous mode */
    if (slp->sl_altp && (slp->sl_flags & SLALLPHYS)) {
	slp->sl_altp->flags |= ALT_PROMISC;
	altSetPromiscuous(slp->sl_altp, TRUE);
    }

    /* reset the fast path, off */
    if (slp->sl_altp)
	altsetipq(slp->sl_altp, slp->sl_instance);
    
    altdlack(wq, mp, DL_PROMISCON_REQ);
}

STATIC void
altpoffreq(queue_t *wq, mblk_t *mp)
{
    struct altstr *slp;
    int flag;

    slp = (struct altstr *)wq->q_ptr;

    if (MBLKL(mp) < DL_PROMISCOFF_REQ_SIZE) {
	altdlnak(wq, mp, DL_PROMISCOFF_REQ, DL_BADPRIM, 0);
	return;
    }

    ALTTRACE(promoff, 0x50028, wq, mp, 
	     ((dl_promiscon_req_t *)mp->b_rptr)->dl_level, 0);

    switch (((dl_promiscoff_req_t *)mp->b_rptr)->dl_level) {
      case DL_PROMISC_PHYS:
	flag = SLALLPHYS;
	break;
	
      case DL_PROMISC_SAP:
	flag = SLALLSAP;
	break;
	
      case DL_PROMISC_MULTI:
	flag = SLALLMULTI;
	break;
	
      default:
	altdlnak(wq, mp, DL_PROMISCOFF_REQ, DL_NOTSUPPORTED, 0);
	return;
    }

    if ((slp->sl_flags & flag) == 0) {
	altdlnak(wq, mp, DL_PROMISCOFF_REQ, DL_NOTENAB, 0);
	return;
    }
    
    /* let the firmware know, if we're turning it off */
    slp->sl_flags &= ~flag;
    if (slp->sl_altp && !(slp->sl_flags & SLALLPHYS)) {
	slp->sl_altp->flags &= ~ALT_PROMISC;
	altSetPromiscuous(slp->sl_altp, FALSE);
    }

    /* reset the fast path queue */
    if (slp->sl_altp)
	altsetipq(slp->sl_altp, slp->sl_instance);
    
    altdlack(wq, mp, DL_PROMISCOFF_REQ);
}

/*
 * Set or clear the device ipq pointer.
 * Assumes IP is SLFAST.
 */
STATIC void
altsetipq(struct altinst *altp, ulong instance)
{
    struct altstr *slp;
    bool ok = TRUE;
    queue_t *ipq = NULL;
    
    rw_enter(&altstruplock, RW_READER);

    for (slp = altstrup; slp; slp = slp->sl_nextp)
	if (slp->sl_altp == altp && slp->sl_instance == instance) {
	    if (slp->sl_flags & (SLALLPHYS | SLALLSAP))
		ok = FALSE;
	    if (slp->sl_sap == ETHERTYPE_IP)
		if (ipq == NULL)
		    ipq = slp->sl_rq;
		else
		    ok = FALSE;
	}

    rw_exit(&altstruplock);

    if (ok)
	altp->ipq[instance/NALT] = ipq;
    else
	altp->ipq[instance/NALT] = NULL;
}

/*
 * ALTSTART
 *	Called by streams driver to start a transfer. The mblk_t is 
 *      passed.
 */
bool
ALTSTART(struct altinst *altp, mblk_t *mp)
{
    mblk_t *nmp;		/* for promiscuous copy */
    int rc = 0;
    U16 vlan_id;
    int ipq_inst;
    LOCK_TYPE *mutexp;

    ALTTRACE(altstart, 0x50022, altp, mp, altp->state, altp->nic->flags); 
    
    /* drop the message if firmware isn't running */
    if (altp->state != ALT_FW_RUNNING || 
	!(altp->nic->flags & ALT_LINK_READY)) {
	freemsg(mp);
	return(TRUE);
    }
    
    /* check for promiscuous mode, if so dup message and send it up */
    if (altp->flags & ALT_PROMISC) {
	if ((nmp = copymsg(mp)))
	    altsendup(altp, nmp, altpaccept);
	else 
	    altp->promiscDrops++;
    }
	    
    if (altp->flags & ALT_VLANS_ACTIVE)
	if (altp->initNic->flags & ALT_SNAP)
	    vlan_id = 
	      ntohs(((struct tagged_snap_ether_header *)mp->b_rptr)->vlan_id);
	else
	    vlan_id = 
	      ntohs(((struct tagged_ether_header *)mp->b_rptr)->vlan_id);

    /* save this away in case the redunancy code changes it while we're
     * in here */
    mutexp = &altp->nic->sendMutex;
    ALT_DISABLE(spl, mutexp);
    /* queue the message for sending */
    rc = altQueueSendMsg(altp->nic, mp, NO_VLAN_TAG, 0);
    ALT_ENABLE(spl, mutexp);
    if (rc) {
	/* keep track of per VLAN stats */
	if (altp->flags & ALT_VLANS_ACTIVE) {
	    /* keep track of per VLAN statistics */
	    if ((ipq_inst = altp->vlan_translate[vlan_id]) != -1)
		altp->outPackets[ipq_inst]++;
	    else
		/* unmapped VLANs get counted toward base interface */
		altp->outPackets[0]++;
	} else {
	    altp->outPackets[0]++;
	}
    }
    return(rc);
}

void
altread(struct tg_nic *nic, mblk_t *mp)
{
    struct altinst *altp = nic->altp;
    struct ether_header *headerp;
    struct snap_ether_header *sheaderp;
    struct tagged_ether_header *theaderp;
    struct tagged_snap_ether_header *tsheaderp;
    int len = mp->b_wptr - mp->b_rptr;
    caddr_t dp;
    mblk_t *nmp;		/* for promiscuous copy */
    queue_t *ipq;
    int size;
    int ipq_inst;
#if IP_SANITY_CHECK > 1
    U16 udpsum, tcpsum;
#endif /* IP_SANITY_CHECK */

    /* figure out the ipq, null implies we don't use it */
    if (altp->flags & ALT_VLANS_ACTIVE) {
	if (TAGGEDSNAPMATCH(mp->b_rptr)) {
	    headerp = (struct ether_header *)mp->b_rptr;
	    tsheaderp = (struct tagged_snap_ether_header *)mp->b_rptr;
	    /* do per VLAN statistics for inbound packets, count untagged
	     * packets to interface 0 (the base interface) */
	    if ((ipq_inst = altp->
		vlan_translate[(unsigned short)htons(tsheaderp->vlan_id)]) 
		!= -1)
		altp->inPackets[ipq_inst]++;
	    else
		altp->inPackets[0]++;
		
	    if (((unsigned short)ntohs(tsheaderp->snap_type) == 
		 (unsigned short)ETHERTYPE_IP) && 
		((tsheaderp->ether_dhost.ether_addr_octet[0] & 01) == 0) &&
		ipq_inst != -1) 
		ipq = altp->ipq[ipq_inst];
	    else
		ipq = (queue_t *)0;
	    size = sizeof(struct tagged_snap_ether_header);
	} else {
	    theaderp = (struct tagged_ether_header *)mp->b_rptr;
	    /* for BPDU check */
	    headerp = (struct ether_header *)mp->b_rptr;
	    /* if we're expected tagged frames and we get an untagged one
	     * discard it, unless it's a bpdu */
	    if ((unsigned short)ntohs(theaderp->tag_id) != 
		(unsigned short)altVlanTagId &&
		ether_cmp(headerp, &etherBPDUaddr)) {
		freemsg(mp);
		return;
	    }
	    /* do per VLAN statistics for inbound packets, count untagged
	     * packets to interface 0 (the base interface) */
	    if ((ipq_inst = altp->
		vlan_translate[(unsigned short)htons(theaderp->vlan_id)]) 
		!= -1)
		altp->inPackets[ipq_inst]++;
	    else
		altp->inPackets[0]++;
		
	    if (((unsigned short)ntohs(theaderp->ether_type) == 
		 (unsigned short)ETHERTYPE_IP) && 
		((theaderp->ether_dhost.ether_addr_octet[0] & 01) == 0) &&
		ipq_inst != -1) 
		ipq = altp->ipq[ipq_inst];
	    else
		ipq = (queue_t *)0;
	    size = sizeof(struct tagged_ether_header);
	}
    } else {
	if (SNAPMATCH(mp->b_rptr)) {
	    headerp = (struct ether_header *)mp->b_rptr;
	    sheaderp = (struct snap_ether_header *)mp->b_rptr;
	    if (((unsigned short)ntohs(sheaderp->snap_type) == 
		 (unsigned short)ETHERTYPE_IP) && 
		((sheaderp->ether_dhost.ether_addr_octet[0] & 01) == 0))
		ipq = altp->ipq[0];
	    else
		ipq = (queue_t *)0;
	    altp->inPackets[0]++;
	    size = sizeof(struct snap_ether_header);
	} else {
	    headerp = (struct ether_header *)mp->b_rptr;
	    if (((unsigned short)ntohs(headerp->ether_type) == 
		 (unsigned short)ETHERTYPE_IP) && 
		((headerp->ether_dhost.ether_addr_octet[0] & 01) == 0))
		ipq = altp->ipq[0];
	    else
		ipq = (queue_t *)0;
	    altp->inPackets[0]++;
	    size = sizeof(struct ether_header);
	}
    }

    if (altTraceWholePacket) {
	ALTTRACE(altread, 0x50019, nic, mp, ipq, len);
	dp = (caddr_t)mp->b_rptr;
	/* round data pointer to a word boundary or we can crash */
	dp = (caddr_t)((int)dp & ~3);
	while (len >= 40) {
	    ALTTRACE2V(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
		       *(U32 *)(dp + 8), *(U32 *)(dp + 12), dp+16);
	    len -= 40;
	    dp += 40;
	}
	/* handle last trace buffer(s) carefully, could run off end */
	if (len > 0) {
	    if (len <= 4) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, 0, 0, 0);
	    } else if (len <= 8) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 0, 0);
	    } else if (len <= 12) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), 0);
	    } else if (len <= 16) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
	    } else if (len <= 20) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), 0, 0, 0);
	    } else if (len <= 24) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), *(U32 *)(dp+20), 
			  0, 0);
	    } else if (len <= 28) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), *(U32 *)(dp+20), 
			  *(U32 *)(dp+24), 0);
	    } else if (len <= 32) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), *(U32 *)(dp+20), 
			  *(U32 *)(dp+24), *(U32 *)(dp+28));
	    } else if (len <= 36) {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), *(U32 *)(dp+20), 
			  *(U32 *)(dp+24), *(U32 *)(dp+28));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+32), 0, 0, 0);
	    } else {
		ALTTRACEV(altread2, 0x50019, *(U32 *)dp, *(U32 *)(dp+4), 
			  *(U32 *)(dp+8), *(U32 *)(dp+12));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+16), *(U32 *)(dp+20), 
			  *(U32 *)(dp+24), *(U32 *)(dp+28));
		ALTTRACEV(altread2, 0x50019, *(U32 *)(dp+32), *(U32 *)(dp+36),
			  0, 0);
	    }
	}
    } else
	ALTTRACE2(altread, 0x50019, nic, mp, ipq, 
		   mp->b_wptr - mp->b_rptr, headerp);

    /* throw away frames for the redundant interface except BPDUs */
    if ((altp->flags & ALT_REDUNDANT) && nic != altp->nic) {
	if (!ether_cmp(headerp, &etherBPDUaddr))
	    altCheckBPDU(nic, mp);
	else
	    freemsg(mp);
	ALTTRACE(reddrop, 0x50041, altp->nic, 0, 0, 0);
	return;
    }

    /*
     * IP shortcut
     *
     * NOTE the (unsigned short) cast is necessary for suns because for
     * whatever reason they decide to return junk in the top half of the
     * return from ntohs.  Go figure.
     */
    if (ipq &&
	CANPUTNEXT(ipq)) {
	mp->b_rptr += size;
	altp->infast++;
#if IP_SANITY_CHECK		/* doesn't work for b_cont recvs XXX */
	{
	    U16 tmp = (U16)ntohs(((struct ip *)(mp->b_rptr))->ip_len);

	    /* check the IP length */
	    if ((U16)(mp->b_wptr - mp->b_rptr) < tmp) {
		(void) stopFW(altp);
		(void) stopMAC(altp);
		ALTTRACEE(BADIPPKT, 0xf5008, mp->b_wptr - mp->b_rptr, 
			  ntohs(((struct ip *)(mp->b_rptr))->ip_len), 
			  nic->recv_std_consumer, nic->recv_std_producer);
		altp->state = ALT_NOTREADY;
		alt_trace_snap();
	    }
	}

#if IP_SANITY_CHECK > 1
        /* if it's UDP check the cksum, on unfragmented packets */
        if (((struct ip *)(mp->b_rptr))->ip_p == IPPROTO_UDP) {
            if (!(ntohs(((struct ip *)(mp->b_rptr))->ip_off) & 0x2fff)) {
                udpsum = udpcksum((struct ip *)(mp->b_rptr));
                if (udpsum != 0 && udpsum != 0xffff) {
                    ALTTRACE2E(BADUCSUM, 0xf5006, udpsum, mp->b_rptr,
                               (U16)ntohs(((struct udphdr *)
                                           ((caddr_t)mp->b_rptr +
                                            sizeof(struct ip)))->uh_sum), 0,
                               mp->b_rptr + sizeof(struct ip));
                    alt_trace_snap();
                    (void) stopFW(altp);
		    (void) stopMAC(altp);
                }
#if 0
            } else if (!(((struct ip *)(mp->b_rptr))->ip_off & 0xfff)) {
                /* for fragmented packets, with the first skip past the
                 * UDP and NFS headers, for the others don't skip.  Then
                 * look for non-zero data */
            } else {
#endif /* 0 */
            } 
	    /* same deal for tcp, we expect to not see fragmented packets */
        } else if (((struct ip *)(mp->b_rptr))->ip_p == IPPROTO_TCP) {
            if (!(ntohs(((struct ip *)(mp->b_rptr))->ip_off) & 0x2fff)) {
                tcpsum = tcpcksum((struct ip *)(mp->b_rptr));
                if (tcpsum != 0 && tcpsum != 0xffff) {
                    ALTTRACE2E(BADTCSUM, 0xf5002, tcpsum, mp->b_rptr,
                               (U16)ntohs(((struct tcphdr *)
                                           ((caddr_t)mp->b_rptr +
                                            sizeof(struct ip)))->th_sum), 0,
                               mp->b_rptr + sizeof(struct ip));
                    alt_trace_snap();
                    (void) stopFW(altp);
		    (void) stopMAC(altp);
		    panic("BADTCSUM");
                }
#if 0
            } else if (!(((struct ip *)(mp->b_rptr))->ip_off & 0xfff)) {
                /* for fragmented packets, with the first skip past the
                 * UDP and NFS headers, for the others don't skip.  Then
                 * look for non-zero data */
            } else {
#endif /* 0 */
	    } else {
		ALTTRACE(fragpkt, 0xf500a, 
			 ((struct ip *)(mp->b_rptr))->ip_off, 0, 0, 0);
            } 
	} else {
	    ALTTRACE(unkproto, 0xf5009, ((struct ip *)(mp->b_rptr))->ip_p,
		     0, 0, 0);
	}
#endif /* IP_SANITY_CHECK > 1 */
#endif /* IP_SANITY_CHECK */

	(void) PUTNEXT(ipq, mp);
    } else {
	if (altp->flags & ALT_VLANS_ACTIVE) {
	    /*
	     * Strip the PADs for 802.3 
	     * but not if len == 0
	     */
	    if ((U16)ntohs(theaderp->ether_type) &&
	        (((U16)ntohs(theaderp->ether_type) +
		size) < ETHERMIN))
		mp->b_wptr = mp->b_rptr + size +
		    (U16)ntohs(theaderp->ether_type);
	} else {
	    /*
	     * Strip the PADs for 802.3 
	     * but not if len == 0
	     */
	    if ((U16)ntohs(headerp->ether_type) &&
	        (((U16)ntohs(headerp->ether_type) +
		size) < ETHERMIN))
		mp->b_wptr = mp->b_rptr + size +
		    (U16)ntohs(headerp->ether_type);
	}

	/* check for BPDUs */
	if ((altp->flags & ALT_REDUNDANT) && 
	    !ether_cmp(headerp, &etherBPDUaddr)) {
	    /* if in promiscuous mode send it up too */
	    if (altp->flags & ALT_PROMISC) {
		if ((nmp = copymsg(mp)))
		    altsendup(altp, nmp, altpaccept);
		else 
		    altp->promiscDrops++;
	    }
	    altCheckBPDU(nic, mp);
	    freemsg(mp);
	} else
	    altsendup(altp, mp, altaccept);
    }
}

#if IP_SANITY_CHECK > 1
#define UDP_ALEN 4
 
STATIC U16
udpcksum(struct ip *pip)
{
    struct udphdr *pudp = (struct udphdr *)((caddr_t)pip + sizeof(struct ip));
    U16 *psh;
    U32 sum;
    int len = (U16)ntohs(pudp->uh_ulen);
    int i;
 
    sum = 0;
 
    /* notice these are shorts so we get both addresses */
    psh = (U16 *) &pip->ip_src.s_addr;
    for (i=0; i < UDP_ALEN; ++i)
        sum += (U16)ntohs(*psh++);
 
    psh = (U16 *)pudp;
    sum += pip->ip_p + len;
    if (len & 1) {
        ((U8 *)pudp)[len] = 0;  /* pad */
        len += 1;               /*  for the following division */
    }
    len /= 2;                   /* convert to length in U16s */
 
    for (i=0; i < len; ++i)
        sum += (U16)ntohs(*psh++);
    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);
 
    return (U16)(~sum & 0xffff);
}

STATIC U16
tcpcksum(struct ip *pip)
{
    struct tcphdr *ptcp = (struct tcphdr *)((caddr_t)pip + sizeof(struct ip));
    U16 *psh;
    U32 sum;
    int len;
    int i;
 
    sum = 0;
 
    /* notice these are shorts so we get both addresses */
    psh = (U16 *) &pip->ip_src.s_addr;
    for (i=0; i < UDP_ALEN; ++i)
        sum += (U16)ntohs(*psh++);
 
    psh = (U16 *)ptcp;
    len = (U16)ntohs(pip->ip_len) - (pip->ip_hl << 2);
    sum += pip->ip_p + len;
    if (len & 1) {
        ((U8 *)ptcp)[len] = 0;  /* pad */
        len += 1;               /*  for the following division */
    }
    len /= 2;                   /* convert to length in U16s */
 
    for (i=0; i < len; ++i)
        sum += (U16)ntohs(*psh++);
    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);
 
    return (U16)(~sum & 0xffff);
}
#endif /* IP_SANITY_CHECK > 1 */

/*
 * Send packet upstream.
 * Assume mp->b_rptr points to ether_header.
 */
STATIC void
altsendup(struct altinst *altp, mblk_t *mp, 
	  struct altstr *(*acceptfunc)(struct altstr *, struct altinst *,
				       unsigned short, struct ether_addr *,
				       U16))
{
    struct ether_addr *dhostp, *shostp;
    struct altstr *slp, *nslp;
    mblk_t *nmp;
    ulong isgroupaddr;
    unsigned short type;
    int macspot[2];
    U16 vlan_id;
    int size;
    
    if (altp->flags & ALT_VLANS_ACTIVE) {
	if (TAGGEDSNAPMATCH(mp->b_rptr)) {
	    ALTTRACE(altsndup, 0x50b1c, altp, mp->b_rptr, 
		 &((struct tagged_snap_ether_header *)mp->b_rptr)->snap_type, 
		 ntohs(((struct tagged_snap_ether_header *)mp->b_rptr)
			   ->snap_type));

	    dhostp = &((struct tagged_snap_ether_header *)mp->b_rptr)->ether_dhost;
	    shostp = &((struct tagged_snap_ether_header *)mp->b_rptr)->ether_shost;
	    type = ntohs(((struct tagged_snap_ether_header *)mp->b_rptr)->snap_type);
	    vlan_id = ntohs(((struct tagged_snap_ether_header *)mp->b_rptr)->vlan_id);
	    size = sizeof(struct tagged_snap_ether_header);
	} else {

	    ALTTRACE(altsndup, 0x5001c, altp, mp->b_rptr, 
		     &((struct tagged_ether_header *)mp->b_rptr)->ether_type, 
		     ntohs(((struct tagged_ether_header *)mp->b_rptr)
			   ->ether_type));

	    dhostp = &((struct tagged_ether_header *)mp->b_rptr)->ether_dhost;
	    shostp = &((struct tagged_ether_header *)mp->b_rptr)->ether_shost;
	    type = 
		ntohs(((struct tagged_ether_header *)mp->b_rptr)->ether_type);
	    vlan_id = 
		ntohs(((struct tagged_ether_header *)mp->b_rptr)->vlan_id);
	    size = sizeof(struct tagged_ether_header);
	}
    } else {
	if (SNAPMATCH(mp->b_rptr)) {
	    ALTTRACE(sndupv, 0x5003a, altp, mp->b_rptr, 
		 &((struct snap_ether_header *)mp->b_rptr)->snap_type, 
		 ntohs(((struct snap_ether_header *)mp->b_rptr)->snap_type));

	    dhostp = &((struct snap_ether_header *)mp->b_rptr)->ether_dhost;
	    shostp = &((struct snap_ether_header *)mp->b_rptr)->ether_shost;
	    type = ntohs(((struct snap_ether_header *)mp->b_rptr)->snap_type);
	    vlan_id = 0;
	    size = sizeof(struct snap_ether_header);
	} else {
	    ALTTRACE(sndupv, 0x5003a, altp, mp->b_rptr, 
		     &((struct ether_header *)mp->b_rptr)->ether_type, 
		     ntohs(((struct ether_header *)mp->b_rptr)->ether_type));

	    dhostp = &((struct ether_header *)mp->b_rptr)->ether_dhost;
	    shostp = &((struct ether_header *)mp->b_rptr)->ether_shost;
	    type = ntohs(((struct ether_header *)mp->b_rptr)->ether_type);
	    vlan_id = 0;
	    size = sizeof(struct ether_header);
	}
    }
    
    isgroupaddr = dhostp->ether_addr_octet[0] & 01;
    altp->inslow++;
    
    /* we use the slp queues here, protect from a writer */
    rw_enter(&altstruplock, RW_READER);

    /*
     * attempt to match the address criteria for each stream
     * and pass up the raw M_DATA ("fastpath") or a DL_UNITDATA_IND.
     */
    
    if (!(slp = (*acceptfunc)(altstrup, altp, type, dhostp, vlan_id))) {
	MEMCPY((MEMCPY_TYPE)macspot, (MEMCPY_TYPE)dhostp, 6);
	ALTTRACE2(sndupbad, 0x5001d, type, macspot[0], mp, mp->b_rptr,
		  mp->b_rptr);
#if 0				/* DEBUGGING */
	(void) stopFW(altp);
	(void) stopMAC(altp);
	alt_trace_snap();
#endif /* 0 */
	freemsg(mp);
	altp->badpkts++;
	/* give back the lock */
	rw_exit(&altstruplock);
	return;
    }
    
    /*
     * Loop on matching open streams until (*acceptfunc)() returns NULL.
     */
    for (; nslp = (*acceptfunc)(slp->sl_nextp, altp, type, dhostp, vlan_id); 
	 slp = nslp)
	if (CANPUTNEXT(slp->sl_rq)) {
	    if (nmp = dupmsg(mp)) {
		if ((slp->sl_flags & SLFAST) && !isgroupaddr) {
		    nmp->b_rptr += size;
		    (void) PUTNEXT(slp->sl_rq, nmp);
		} else if (slp->sl_flags & SLRAW)
		    (void) PUTNEXT(slp->sl_rq, nmp);
		else if ((nmp = altaddudind(altp, nmp, shostp,
					    dhostp, type, isgroupaddr))) {
		    (void) PUTNEXT(slp->sl_rq, nmp);
		}
	    }
	    else {
		ALTTRACEE(dupmsgf, 0xf5000, 0, 0, 0, 0);
		altp->allocbfail++;
	    }
	} else {
	    ALTTRACEE(canputnf, 0xf5001, 0, 0, 0, 0);
	    altp->nocanput++;
	}

    /*
     * Do the last one.
     */
    if (CANPUTNEXT(slp->sl_rq)) {
	if ((slp->sl_flags & SLFAST) && !isgroupaddr) {
	    mp->b_rptr += size;
	    (void) putnext(slp->sl_rq, mp);
	} else if (slp->sl_flags & SLRAW)
	    (void) putnext(slp->sl_rq, mp);
	else if ((mp = altaddudind(altp, mp, shostp, dhostp,
				   type, isgroupaddr))) {
	    (void) PUTNEXT(slp->sl_rq, mp);
	}
    } else {
	freemsg(mp);
	ALTTRACEE(canputnf, 0xf5003, 0, 0, 0, 0);
	altp->nocanput++;
    }

    /* give back the lock */
    rw_exit(&altstruplock);
}

/*
 * Prefix msg with a DL_UNITDATA_IND mblk and return the new msg.
 */
STATIC mblk_t *
altaddudind(struct altinst *altp, mblk_t *mp, struct ether_addr *shostp, 
	    struct ether_addr *dhostp, int type, ulong isgroupaddr)
{
    dl_unitdata_ind_t *dludindp;
    struct alt_dladdr *dlap;
    mblk_t *nmp;
    int size, hsize;
    
    ALTTRACE(altaddud, 0x50029, shostp, dhostp, type, isgroupaddr);
    
    if (altp->flags & ALT_VLANS_ACTIVE) {
	if (TAGGEDSNAPMATCH(mp->b_rptr)) {
	    hsize = sizeof (struct tagged_snap_ether_header);
	} else {
	    hsize = sizeof (struct tagged_ether_header);
	}
    } else {
	if (SNAPMATCH(mp->b_rptr)) {
	    hsize = sizeof (struct snap_ether_header);
	} else {
	    hsize = sizeof (struct ether_header);
	}
    }
    mp->b_rptr += hsize;
    
    /*
     * Allocate an M_PROTO mblk for the DL_UNITDATA_IND.
     */
    size = sizeof (dl_unitdata_ind_t) + ALTADDRL + ALTADDRL;
    nmp = allocb((ALTHEADROOM + size), BPRI_MED);
    if (nmp == NULL) {
	altp->allocbfail++;
	freemsg(mp);
	return (NULL);
    }
    DB_TYPE(nmp) = (unsigned char)M_PROTO;
    nmp->b_wptr = nmp->b_datap->db_lim;
    nmp->b_rptr = nmp->b_wptr - size;
    
    /*
     * Construct a DL_UNITDATA_IND primitive.
     */
    dludindp = (dl_unitdata_ind_t *)nmp->b_rptr;
    dludindp->dl_primitive = DL_UNITDATA_IND;
    dludindp->dl_dest_addr_length = ALTADDRL;
    dludindp->dl_dest_addr_offset = sizeof (dl_unitdata_ind_t);
    dludindp->dl_src_addr_length = ALTADDRL;
    dludindp->dl_src_addr_offset = sizeof (dl_unitdata_ind_t) + ALTADDRL;
    dludindp->dl_group_address = isgroupaddr;
    
    dlap = (struct alt_dladdr *)(nmp->b_rptr + sizeof (dl_unitdata_ind_t));
    ether_copy(dhostp, &dlap->dl_phys);
    dlap->dl_sap = (u_short)type;
    
    /*
     * Link the M_PROTO and M_DATA together.
     */
    nmp->b_cont = mp;
    return (nmp);
}

/*
 * Test upstream destination sap and address match.
 */
STATIC struct altstr *
altaccept(struct altstr *slp, struct altinst *altp, unsigned short type, 
	 struct ether_addr *addrp, U16 vlan_id)
{
    int sap;
    int flags;
    
    for (; slp; slp = slp->sl_nextp) {
	sap = slp->sl_sap;
	flags = slp->sl_flags;
	
	if ((slp->sl_altp == altp) && ALTSAPMATCH(sap, type, flags) &&
	    (!(altp->flags & ALT_VLANS_ACTIVE) || 
	     altp->vlan_id[slp->sl_instance/NALT] == vlan_id))
	    if (!(ether_cmp(addrp, &altp->macAddr)) ||
		!(ether_cmp(addrp, &etherbroadcastaddr)) ||
		(flags & SLALLPHYS) ||
		altmcmatch(slp, addrp)) {
		return (slp);
	    }
    }
    return (NULL);
}

/*
 * Test upstream destination sap and address match for SLALLPHYS only.
 */
STATIC struct altstr *
altpaccept(struct altstr *slp, struct altinst *altp, unsigned short type, 
	   struct ether_addr *addrp, U16 vlan_id)
{
    int sap;
    int flags;
    
    for (; slp; slp = slp->sl_nextp) {
	sap = slp->sl_sap;
	flags = slp->sl_flags;
	
	if ((slp->sl_altp == altp) && 
	    ALTSAPMATCH(sap, type, flags) &&
	    (!(altp->flags & ALT_VLANS_ACTIVE) ||
	     altp->vlan_id[slp->sl_instance/NALT] == vlan_id) &&
	    (flags & SLALLPHYS)) {
	    return(slp);
	}
    }
    return (NULL);
}

/*
 * Return TRUE if the given multicast address is one
 * of those that this particular Stream is interested in.
 */
STATIC
altmcmatch(struct altstr *slp, struct ether_addr *addrp)
{
    register struct ether_addr *mctab;
    register int mccount;
    register int i;
    
    /*
     * Return FALSE if not a multicast address.
     */
    if (!(addrp->ether_addr_octet[0] & 01))
	return(FALSE);

    /*
     * Check if all multicasts have been enabled for this Stream
     */
    if (slp->sl_flags & SLALLMULTI)
	return(TRUE);
    
    /*
     * Return FALSE if no multicast addresses enabled for this Stream.
     */
    if (slp->sl_mccount == 0)
	return(FALSE);
    
    /*
     * Otherwise, find it in the table.
     */
    
    mccount = slp->sl_mccount;
    mctab = slp->sl_mctab;
    
    for (i = 0; i < mccount; i++)
	if (!ether_cmp(addrp, &mctab[i]))
	    return(TRUE);
    
    return(FALSE);
}

STATIC void
altioctl(queue_t *wq, mblk_t *mp)
{
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    struct altstr *slp = (struct altstr *)wq->q_ptr;
    struct altinst *altp = (struct altinst *)slp->sl_altp;
    mblk_t *tmp;
    struct ioc_state *stp;
    int rc = 0;
    U32 nicTb, curNicTb, endNicTb;
    bool running;
    int instance;
    struct altinst *altp2;
    struct tg_nic *nic;
    U32 oldDump;
    int i;

    ALTTRACEG(altioctl, 0x50007, mp, slp, altp, iocp->ioc_cmd);

    /* all our ioctls are TRANSPARENT, should this ever change we'll have
     * to check here */
    /* we also support the two sun DLPI ioctls and ND ioctls */
    switch ((U32)iocp->ioc_cmd) {
	/* two sun DLPI ioctls */
    case DLIOCRAW:
	slp->sl_flags |= SLRAW;
	altiocack(wq, mp, 0, 0);
	break;

    case DL_IOC_HDR_INFO:	/* M_DATA "fastpath" info request */
	altdliochdrinfo(wq, mp);
	break;

    case ND_GET:
	/* set the altp based on the current setting of instance */
	if (!(altp = (struct altinst *)
	      ddi_get_soft_state(altStatep, 
				 alt_param_arr[ND_INSTANCE].parm_val % 
				 NALT))) {
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}
	if (!nd_getset(wq, altp->initNic->alt_nd_ptr, mp)) {
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}
	qreply(wq, mp);
	break;
	
    case ND_SET:
	/* set the altp based on the current setting of instance */
	if (!(altp = (struct altinst *)
	      ddi_get_soft_state(altStatep, 
				 alt_param_arr[ND_INSTANCE].parm_val % 
				 NALT))) {
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}
	nic = altp->initNic;

	/* stash old nic_dump status */
	oldDump = altp->nic->tuneParms[DUMP_NIC].parm_val;

	if (!nd_getset(wq, nic->alt_nd_ptr, mp)) {
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	/* check to see if the instance pointer has been updated */
	if (alt_param_arr[ND_INSTANCE].parm_val != 
	    nic->tuneParms[ND_INSTANCE].parm_val) {
	    int i;

	    /* check to see if it's been put out of range */
	    if (!(altp2 = (struct altinst *)
		  ddi_get_soft_state(altStatep, 
			     nic->tuneParms[ND_INSTANCE].parm_val % NALT))) {
		/* it was, put it back */
		nic->tuneParms[ND_INSTANCE].parm_val = 
		    alt_param_arr[ND_INSTANCE].parm_val;
		altiocnak(wq, mp, 0, EINVAL);
		return;
	    }

	    /* keep the global instance pointer updated */
	    alt_param_arr[ND_INSTANCE].parm_val = 
		nic->tuneParms[ND_INSTANCE].parm_val;

	    /* fix up all the active instances so they all read the same */
	    for (i=0; i<NALT; i++) {
		if ((altp2 = (struct altinst *)
		     ddi_get_soft_state(altStatep, i)) != 0)
		    altp2->nic->tuneParms[ND_INSTANCE].parm_val =
			alt_param_arr[ND_INSTANCE].parm_val;
	    }
	}

	/* make sure the data is up to date in the shared memory region */
	altUpdateTuneParms(nic);

	/* if nic dump status changes we need to handle it */
	if (oldDump != nic->tuneParms[DUMP_NIC].parm_val) {
	    /* if 1, do the dump, if 0, free the data structures */
	    if (nic->tuneParms[DUMP_NIC].parm_val) {
		if (!dumpFW(altp)) {
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, ENOMEM);
		    return;
		}
	    } else
		freeDump(altp);
	}
	qreply(wq, mp);
	break;
	
    case ALT_SETTRACE:
	/* only allow superuser to get to this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	altTracing = *(int *)mp->b_cont->b_rptr;
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;
	
    case ALT_READTRACE:
	msgcopyout(wq, mp, (mblk_t *)0, sizeof(int), 
		   *(caddr_t *)mp->b_cont->b_rptr, 
		   (caddr_t)&altTracing, FALSE);
	break;
	
    case ALT_GETTRACE:
	/* snapshot trace buffer */
	MEMCPY((MEMCPY_TYPE)alt_trtmp, 
	       (MEMCPY_TYPE)(alt_curtb + ALT_TRACE_ELEM_SIZE),
	       sizeof(u_long) * (alt_endtb - alt_curtb));
	MEMCPY((MEMCPY_TYPE)(alt_trtmp + 
			(alt_endtb - (alt_curtb + ALT_TRACE_ELEM_SIZE))),
	       (MEMCPY_TYPE)alt_tbuf,
	       sizeof(u_long) * (alt_curtb - alt_tbuf));
	msgcopyout(wq, mp, (mblk_t *)0, 
		   (ALT_TRACE_SIZE+1)*ALT_TRACE_ELEM_SIZE, 
		   *(caddr_t *)mp->b_cont->b_rptr, (caddr_t)alt_trtmp, FALSE);
	break;

    case ALT_GETSTRACE:
	if (!altTraceSnap) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, ENOENT);
	    return;
	}
	/* snapshot snap trace buffer */
	MEMCPY((MEMCPY_TYPE)alt_trtmp, 
	       (MEMCPY_TYPE)(alt_curstb + ALT_TRACE_ELEM_SIZE),
	       sizeof(u_long) * (alt_endstb - alt_curstb));
	MEMCPY((MEMCPY_TYPE)(alt_trtmp + 
			(alt_endstb - (alt_curstb + ALT_TRACE_ELEM_SIZE))),
	       (MEMCPY_TYPE)alt_stbuf,
	       sizeof(u_long) * (alt_curstb - alt_stbuf));
	msgcopyout(wq, mp, (mblk_t *)0, 
		   (ALT_TRACE_SIZE+1)*ALT_TRACE_ELEM_SIZE, 
		   *(caddr_t *)mp->b_cont->b_rptr, (caddr_t)alt_trtmp, FALSE);
	break;

	/* snapshot NIC trace buffer */
    case ALT_GETNTRACE:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* stop the firmware */
	running = stopFW(altp);
	
	/* nic_trace_ptr points into a 4k page, aligned, get pieces */
	curNicTb = altp->initNic->genComm->gen_com.nic_trace_ptr;
	nicTb = altp->initNic->genComm->gen_com.nic_trace_start;
	endNicTb = nicTb + altp->initNic->genComm->gen_com.nic_trace_len;
	
	ALTTRACE(nictrace, 0x5001a, curNicTb, nicTb, endNicTb, 0);
	/* protect ourselves against a core dump if the NIC didn't work */
	if (!curNicTb) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *) 0;
	    altiocnak(wq, mp, 0, EIO);
	    break;
	}
	if (readTGMem(altp, curNicTb + (ALT_TRACE_ELEM_SIZE * sizeof(U32)), 
		      (caddr_t)alt_trtmp, endNicTb - curNicTb -
		      (ALT_TRACE_ELEM_SIZE * sizeof(U32)), FALSE, TRUE) < 0) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	
	if (readTGMem(altp, nicTb, 
		      ((caddr_t)(alt_trtmp - ALT_TRACE_ELEM_SIZE)) + 
		      endNicTb - curNicTb,
		      curNicTb - nicTb, FALSE, TRUE) < 0) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	
	/* restart the firmware */
#if FIRMWARE_DEBUGGED
	if (running)
	    if (!restartFW(altp)) {
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		break;
	    }
#endif /* FIRMWARE_DEBUGGED */
		    
	msgcopyout(wq, mp, (mblk_t *)0, 
		   (ALT_TRACE_SIZE+1)*ALT_TRACE_ELEM_SIZE, 
		   *(caddr_t *)mp->b_cont->b_rptr, (caddr_t)alt_trtmp, FALSE);
	break;
		
    case ALT_CLRSTRACE:
	/* get rid of trailing address */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;

	/* only allow superuser to get to this */
	if (!SUSER(iocp->ioc_cr)) {
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	altTraceSnap = FALSE;
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_CLEAR_STATS:
	/* get rid of trailing address */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;

	/* only allow superuser to get to this */
	if (!SUSER(iocp->ioc_cr)) {
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	nic = altp->initNic;
	nic->dmaPrepsActive = 0;
	altp->allocbfail = 0;
	altp->nocanput = 0;
	altp->promiscDrops = 0;
	nic->inpkts = 0;
	altp->infast = 0;
	altp->inslow = 0;
	altp->badpkts = 0;
	nic->bufsLowWater = nic->bufsInStdRecvRing;
#if READ_SERVICE
	altp->flowControl = 0;
#endif /* READ_SERVICE */
	nic->outpkts = 0;
	nic->outDrops = 0;
	nic->outDropsDmaResources = 0;
	nic->outDropsNoSpace = 0;
	nic->cmdFull = 0;
#if defined(__sun)
	nic->overFlowRecvBufs = 0;
	nic->totOverFlowRecvBufs = 0;
	nic->totSpillBufs = nic->spillBufs; /* kind of right */
	nic->flowControls = 0;
#endif /* defined(__sun) */
	nic->dataRtnPkts = 0;
#if USE_CKSUM_OFFLOAD
	nic->badIpCksums = 0;
	nic->badTcpCksums = 0;
	nic->badUdpCksums = 0;
#endif /* USE_CKSUM_OFFLOAD */

	altiocack(wq, mp, 0, 0);
	break;

    case ALT_CLEAR_NIC_STATS:
	/* get rid of trailing address */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;

	/* only allow superuser to get to this */
	if (!SUSER(iocp->ioc_cr)) {
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}
	
	altClearNicStats(altp);
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_ATTACH:
	instance = *(int *)mp->b_cont->b_rptr;
	/* attach to a particular instance */
	ALTTRACE(ioattach, 0x50018, instance, 0, 0, 0);
	/* fail if it is out of range */
	if (instance >= NALT * MAX_VLANS) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	if (!(slp->sl_altp = (struct altinst *)
	      ddi_get_soft_state(altStatep, instance % NALT))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}
	/* return ok */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_STAT:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}
	/* allocate a state buffer */
	if (!(tmp = allocb(sizeof(struct ioc_state), BPRI_MED))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EAGAIN);
	    return;
	}
	tmp->b_wptr += sizeof(struct ioc_state);
	stp = (struct ioc_state *)tmp->b_rptr;
	stp->state = STAT_GETSTRUCT;

	/* copy in the stat structure then copyout the requested structs */
	msgcopyin(wq, mp, tmp, sizeof(struct alt_stat), 
		  *(caddr_t *)mp->b_cont->b_rptr);
	break;

    case ALT_READ_TG_MEM:
    case ALT_WRITE_TG_MEM:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}
	/* allocate a state buffer */
	if (!(tmp = allocb(sizeof(struct ioc_state), BPRI_MED))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EAGAIN);
	    return;
	}
	tmp->b_wptr += sizeof(struct ioc_state);
	stp = (struct ioc_state *)tmp->b_rptr;
	stp->state = TG_GETSTRUCT;

	/* copy in the rr_mem structure then copyout the request data */
	msgcopyin(wq, mp, tmp, sizeof(struct tg_mem), 
		  *(caddr_t *)mp->b_cont->b_rptr);
	break;

    case ALT_READ_TG_REG:
    case ALT_WRITE_TG_REG:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}
	/* allocate a state buffer */
	if (!(tmp = allocb(sizeof(struct ioc_state), BPRI_MED))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EAGAIN);
	    return;
	}
	tmp->b_wptr += sizeof(struct ioc_state);
	stp = (struct ioc_state *)tmp->b_rptr;
	stp->state = TG_GETSTRUCT;
	stp->passedAddr = *(caddr_t *)mp->b_cont->b_rptr;

	/* copy in the tg_reg structure then copyout the request data */
	msgcopyin(wq, mp, tmp, sizeof(struct tg_reg), stp->passedAddr);
	break;

    case ALT_CACHE_FIRMWARE:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* only allow superuser to do this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	/* allocate a state buffer */
	if (!(tmp = allocb(sizeof(struct ioc_state), BPRI_MED))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EAGAIN);
	    return;
	}
	tmp->b_wptr += sizeof(struct ioc_state);
	stp = (struct ioc_state *)tmp->b_rptr;
	stp->state = FW_GETSTRUCT;

	/* copy in the tg_fw structure then copyin the various pieces */
	msgcopyin(wq, mp, tmp, sizeof(struct tg_fw), 
		  *(caddr_t *)mp->b_cont->b_rptr);
	break;

    case ALT_LOAD_FIRMWARE:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* only allow superuser to do this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	ALTTRACE(ioLoadFw, 0x5002e, 0, 0, 0, 0);
	
	/* make sure the firmware is stopped */
	altp->state = ALT_NOTREADY;

	/* stop the queues for a second */
	altp->initNic->flags = 
	    altp->initNic->flags & (ALT_RED_ACTIVE | ALT_RED_STANDBY);
	(void) stopFW(altp);

	/* link up the ring structures */
	initRings(altp);

	/* load and start the firmware */
	if (!loadFW(altp)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}

#if FIRMWARE_DEBUGGED
	if (!startFW(altp)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}
#endif /* FIRMWARE_DEBUGGED */

	/* and again for an attached NIC */
	if (altp->flags & ALT_REDUNDANT) {
	    /* stop the queues for a second */
	    altp->attachNic->flags = 
		altp->attachNic->flags & (ALT_RED_ACTIVE | ALT_RED_STANDBY);
	    (void) stopFW(altp->attachNic->initAltp);

	    /* link up the ring structures */
	    initRings(altp->attachNic->initAltp);

	    /* load and start the firmware */
	    if (!loadFW(altp->attachNic->initAltp)) {
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		return;
	    }

#if FIRMWARE_DEBUGGED
	    if (!startFW(altp->attachNic->initAltp)) {
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		return;
	    }
#endif /* FIRMWARE_DEBUGGED */
	}

	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;

	/* snapshot NIC statistics */
    case ALT_GET_NIC_STATS:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	ALTTRACE(nicstats, 0x5001f, 0, 0, 0, 0);
	
	msgcopyout(wq, mp, (mblk_t *)0, sizeof(struct tg_stats),
		   *(caddr_t *)mp->b_cont->b_rptr, 
		   (caddr_t)&altp->initNic->genInfo->stats, FALSE);
	break;
		
	/* snapshot NIC tuning parms */
    case ALT_GET_TUNEPARMS:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	ALTTRACE(gettune, 0x50020, 0, 0, 0, 0);
	
	msgcopyout(wq, mp, (mblk_t *)0, sizeof(struct tg_tune),
		   *(caddr_t *)mp->b_cont->b_rptr, 
		   (caddr_t)&altp->initNic->genComm->gen_com.tuneParms, TRUE);
	break;
		
    case ALT_SET_TUNEPARMS:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* only allow superuser to do this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	ALTTRACE(settune, 0x50021, 0, 0, 0, 0);

	/* copy in the tg_tune structure */
	msgcopyin(wq, mp, (mblk_t *)0, sizeof(struct tg_tune), 
		  *(caddr_t *)mp->b_cont->b_rptr);
	break;

    case ALT_CLEAR_PROFILE:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	ALTTRACE(clrprof, 0x50025, 0, 0, 0, 0);

	altClearProfile(altp);
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_SEND_PKT:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	ALTTRACEV(sendpkt, 0x50026, 0, 0, 0, 0);

	altSendPkt(altp);
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_RECV_PKT:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	ALTTRACEV(recvpkt, 0x50027, 0, 0, 0, 0);

	altRecvPkt(altp);
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_DEBUG:
	slp->sl_flags |= SLDEBUG;
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_SET_REDUND:
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* only allow superuser to do this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	/* error if the firmware isn't up */
	if (altp->state != ALT_FW_RUNNING) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EIO);
	    return;
	}

	/* which instance are we supposed to link onto this one */
	instance = *(int *)mp->b_cont->b_rptr;

	ALTTRACE(setred, 0x5002c, instance, 0, 0, 0);

	/* get the device instance from the ppa field */
	if (!(altp2 = (struct altinst *)
	      ddi_get_soft_state(altStatep, instance))) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    ALTTRACEE(badinst, 0xf5005, instance, 0, 0, 0);
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	/* check that attached isn't already busy */
	if (altp2->flags & (ALT_HOST_UP | ALT_REDUNDANT)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    ALTTRACEE(badstate, 0xf5006, altp2, altp2->flags, 
		      altp, altp->flags);
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	if (rc = altAddRedund(altp, altp2)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    ALTTRACEE(badstate, 0xf5007, altp2, altp2->flags, 
		      altp, altp->flags);
	    altiocnak(wq, mp, 0, rc);
	    return;
	}

	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_REMOVE_REDUND:
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocnak(wq, mp, 0, EINVAL);
#if 0				/* can't remove redundancy now XXX */
	/* don't continue if we're not attached */
	if (!slp->sl_altp) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EUNATCH);
	    return;
	}

	/* only allow superuser to do this */
	if (!SUSER(iocp->ioc_cr)) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EPERM);
	    return;
	}

	ALTTRACE(remred, 0x5002d, 0, 0, 0, 0);
	
	/* check that this is the "configured redundant" nic */
	if ((altp->flags & (ALT_REDUNDANT | ALT_RED_ATTACH)) !=
	    (ALT_REDUNDANT | ALT_RED_ATTACH) ||
	    altp->attachNic != altp->initNic) {
	    ALT_ENABLE(spl, &altp->nicMu);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	altRemRedund(altp);

	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
#endif /* 0 */
	break;

    case ALT_ALLOC_PAGE: {
	struct tg_nic *nic = altp->initNic;
	caddr_t dmaAddr;

	if (nic->ttpAlloc) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EBUSY);
	    return;
	}

	DMA_MEMALLOC(nic->ttpAlloc, caddr_t, PAGE_SIZE*2,
		     ALT_DMA_CONSISTENT, DDI_DMA_RDWR, 
		     &nic->ttpPageDmaInfo);
	if (!nic->ttpAlloc) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, ENOMEM);
	    return;
	}

	/* align it and the dma Address reported */
	nic->ttpPage = (caddr_t)PAGE_ROUNDUP(nic->ttpAlloc);
	dmaAddr = (DMA_ADDR_TYPE)
	    PAGE_ROUNDUP(MAP2BUS(nic->ttpPageDmaInfo.dataAddr));

	ALTTRACE(allocPg, 0x5002f, nic->ttpPage, nic->ttpAlloc, 
		 dmaAddr, PAGE_SIZE);

	/* copy out the data */
	msgcopyout(wq, mp, (mblk_t *)0, sizeof(caddr_t),
		   *(caddr_t *)mp->b_cont->b_rptr, 
		   (caddr_t)&dmaAddr, TRUE);
	break;
    }
	
    case ALT_FREE_PAGE:
	if (!altp->initNic->ttpAlloc) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	ALTTRACE(freePg, 0x50030, altp->initNic->ttpPage, 
		 altp->initNic->ttpAlloc, 0, 0);

	DMA_MEMFREE((caddr_t)altp->initNic->ttpAlloc, PAGE_SIZE*2,
		    &altp->initNic->ttpPageDmaInfo);
	altp->initNic->ttpAlloc = (caddr_t)0;
	altp->initNic->ttpPage = (caddr_t)0;
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;
	
    case ALT_WRITE_PAGE:
	if (!altp->initNic->ttpAlloc) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	ALTTRACE(writePg, 0x50031, altp->initNic->ttpPage, 
		 altp->initNic->ttpAlloc, 
		 *(caddr_t *)mp->b_cont->b_rptr, 0);

	/* copy in the data */
	msgcopyin(wq, mp, (mblk_t *)0, PAGE_SIZE,
		  *(caddr_t *)mp->b_cont->b_rptr);
	break;
	
    case ALT_READ_PAGE:
	if (!altp->initNic->ttpAlloc) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EINVAL);
	    return;
	}

	ALTTRACE(readPg, 0x50032, altp->initNic->ttpPage, 
		 altp->initNic->ttpAlloc, 
		 *(caddr_t *)mp->b_cont->b_rptr, 0);

	/* copy out the data */
	msgcopyout(wq, mp, (mblk_t *)0, PAGE_SIZE,
		   *(caddr_t *)mp->b_cont->b_rptr, 
		   altp->initNic->ttpPage, TRUE);
	break;
	
    case ALT_DUMP_FIRMWARE:
    default:
	ALTTRACE(unkioctl, 0x50008, iocp->ioc_cmd, 0, 0, 0);
	altiocnak(wq, mp, 0, ENOTTY);
	break;
    }
}

/*
 * finish or continue ioctl following message receipt for copyin/copyout
 */
STATIC void
altioc(queue_t *wq, mblk_t *mp)
{
    struct copyresp *cprspp = (struct copyresp *)mp->b_rptr;
    struct altstr *slp = (struct altstr *)wq->q_ptr;
    struct altstr *tslp;
    struct altinst *altp = (struct altinst *)slp->sl_altp;
    struct ioc_state *stp;
    struct iocstrinfo *strInfo;
    caddr_t tmpSpace, userAddr;
    int len;
    U32 tgAddr;
    U32 oldStatus;

    ALTTRACEG(altioc, 0x50009, mp, slp, altp, cprspp->cp_cmd);
    /* find out what we were doing */
    switch ((U32)cprspp->cp_cmd) {
    case ALT_READTRACE:
    case ALT_GETTRACE:
    case ALT_GETSTRACE:
    case ALT_GETNTRACE:
    case ALT_GET_NIC_STATS:
    case ALT_GET_TUNEPARMS:
    case ALT_READ_PAGE:
    case ALT_ALLOC_PAGE:
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	if (cprspp->cp_rval)
	    altiocnak(wq, mp, 0, 0);
	else
	    altiocack(wq, mp, 0, 0);
	break;

    case ALT_STAT:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case STAT_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.stat, (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct alt_stat));

	    /* now send out the altinst struct */
	    stp->state = STAT_SENDINST;
	    msgcopyout(wq, mp, cprspp->cp_private, sizeof(struct altinst),
		       stp->un.stat.altinst, (caddr_t)altp, FALSE);
	    break;

	case STAT_SENDINST:
	    /* next, send out the tg_nic struct */
	    stp->state = STAT_SENDNIC;
	    msgcopyout(wq, mp, cprspp->cp_private, sizeof(struct tg_nic),
		       stp->un.stat.nic, (caddr_t)altp->initNic, FALSE);
	    break;

	case STAT_SENDNIC:
	    /* next, send out the altstr structs */
	    /* allocate enough space for the data, then copy it out */
	    MEMALLOC(strInfo, struct iocstrinfo *, sizeof(struct iocstrinfo),
		     ALT_NOSLEEP);
	    if (!strInfo) {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, ENOMEM);
		break;
	    }
	    strInfo->numstr = 0;
	    rw_enter(&altstruplock, RW_READER);
	    for (tslp = altstrup; tslp; tslp = tslp->sl_nextp) {
		if ((tslp->sl_altp == slp->sl_altp) && 
                    (strInfo->numstr < MAXSTATSTR)) { 
		    MEMCPY((MEMCPY_TYPE)&strInfo->altstr[strInfo->numstr],
			   (MEMCPY_TYPE)tslp, sizeof(struct altstr));
		    strInfo->numstr++;
		}
	    }
	    stp->state = STAT_SENDSTR;
	    msgcopyout(wq, mp, cprspp->cp_private, sizeof(struct iocstrinfo),
		       (caddr_t)stp->un.stat.altStrInfo, (caddr_t)strInfo, 
		       FALSE);
	    MEMFREE(strInfo, sizeof(struct iocstrinfo), ALT_NOSLEEP);
	    rw_exit(&altstruplock);
	    break;

	case STAT_SENDSTR:
	    /* we're done, free up everything and ack the ioctl */
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocack(wq, mp, 0, 0);
	    break;

	default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;

    case ALT_READ_TG_MEM:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case TG_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.tgMem,
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_mem));

	    len = stp->un.tgMem.len;
	    userAddr = stp->un.tgMem.userAddr;
	    tgAddr = stp->un.tgMem.tgAddr;
	    stp->state = TG_DONE;
	    /* check for the area we're trying to work with */
	    if (
#if TG_BEG_SRAM		/* typically 0 */
		tgAddr >= TG_BEG_SRAM &&
#endif /* TG_BEG_SRAM */
		tgAddr + len <= TG_END_SRAM) {
		/* allocate space in the kernel for the data */
		MEMALLOC(tmpSpace, caddr_t, len, ALT_NOSLEEP);
		if (readTGMem(altp, tgAddr, tmpSpace, len, FALSE, TRUE) < 0) {
		    MEMFREE(tmpSpace, len, ALT_NOSLEEP);
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		msgcopyout(wq, mp, cprspp->cp_private, len, 
			   userAddr, tmpSpace, FALSE);
		MEMFREE(tmpSpace, len, ALT_NOSLEEP);

	    } else if (tgAddr >= TG_BEG_SCRATCH &&
		       tgAddr + len <= TG_END_SCRATCH) {
		if ((IN32(altp->initNic->handle, altp->initNic->base, 
			  TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		    TG_MHC_LSI_GATE_VERS) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EINVAL);
		    break;
		}
		/* allocate space in the kernel for the data */
		MEMALLOC(tmpSpace, caddr_t, len, ALT_NOSLEEP);
		if (readTGScratch(altp, tgAddr, tmpSpace, len, FALSE, TRUE, 
				  TG_PROCESSOR_A) 
		    < 0) {
		    MEMFREE(tmpSpace, len, ALT_NOSLEEP);
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		msgcopyout(wq, mp, cprspp->cp_private, len, 
			   userAddr, tmpSpace, FALSE);
		MEMFREE(tmpSpace, len, ALT_NOSLEEP);

	    } else if (tgAddr >= TG_BEG_SCRATCH_B_DEBUG &&
		       tgAddr + len <= TG_END_SCRATCH_B_DEBUG) {
		if ((IN32(altp->initNic->handle, altp->initNic->base, 
			  TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		    TG_MHC_LSI_GATE_VERS) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EINVAL);
		    break;
		}
		/* allocate space in the kernel for the data */
		MEMALLOC(tmpSpace, caddr_t, len, ALT_NOSLEEP);
		if (readTGScratch(altp, tgAddr - TG_SCRATCH_DEBUG_OFF, 
				  tmpSpace, len, FALSE, TRUE, TG_PROCESSOR_B) 
		    < 0) {
		    MEMFREE(tmpSpace, len, ALT_NOSLEEP);
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		msgcopyout(wq, mp, cprspp->cp_private, len, 
			   userAddr, tmpSpace, FALSE);
		MEMFREE(tmpSpace, len, ALT_NOSLEEP);

	    } else if (tgAddr >= TG_BEG_FLASH &&
		       tgAddr + len <= TG_END_FLASH) {
		/* allocate space in the kernel for the data */
		MEMALLOC(tmpSpace, caddr_t, len, ALT_NOSLEEP);
		if (readTGFlash(altp, tgAddr, tmpSpace, len, FALSE) < 0) {
		    MEMFREE(tmpSpace, len, ALT_NOSLEEP);
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		msgcopyout(wq, mp, cprspp->cp_private, len, 
			   userAddr, tmpSpace, FALSE);
		MEMFREE(tmpSpace, len, ALT_NOSLEEP);

	    } else if (tgAddr >= TG_BEG_SER_EEPROM &&
		       tgAddr + len <= TG_END_SER_EEPROM) {
		/* allocate space in the kernel for the data */
		MEMALLOC(tmpSpace, caddr_t, len, ALT_NOSLEEP);
		if (readTGSerEeprom(altp, tgAddr, tmpSpace, len, FALSE) < 0) {
		    MEMFREE(tmpSpace, len, ALT_NOSLEEP);
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		msgcopyout(wq, mp, cprspp->cp_private, len, 
			   userAddr, tmpSpace, FALSE);
		MEMFREE(tmpSpace, len, ALT_NOSLEEP);

	    } else if (tgAddr >= TG_BEG_REGS &&
		       tgAddr + len <= TG_END_REGS) {
		ALTTRACE2G(rdMemReg, 0x50013, tgAddr, userAddr, len, 0,
			   altp->initNic->base + TG_DEV_ID + 
			   (tgAddr - TG_BEG_REGS));
		msgcopyout_swap(wq, mp, cprspp->cp_private, len, userAddr,
				altp->initNic->base + TG_DEV_ID + 
				(tgAddr - TG_BEG_REGS),
				TRUE);

	    } else {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		break;
	    }
	  
	    ALTTRACEG(readMem, 0x5000b, tgAddr, userAddr, len, 0);
	    break;

	case TG_DONE:
	    /* we're done, free up everything and ack the ioctl */
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocack(wq, mp, 0, 0);
	    break;

	default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;

    case ALT_WRITE_TG_MEM:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case TG_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.tgMem, 
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_mem));
	    stp->state = TG_DONE;
	    msgcopyin(wq, mp, cprspp->cp_private, stp->un.tgMem.len,
		      stp->un.tgMem.userAddr);
	    break;

	case TG_DONE:
	    len = stp->un.tgMem.len;
	    userAddr = stp->un.tgMem.userAddr;
	    tgAddr = stp->un.tgMem.tgAddr;
	    ALTTRACEG(writeMem, 0x5000c, tgAddr, userAddr, len, 0);
	  
	    /* copy data requested into correct area */
	    if (
#if TG_BEG_SRAM		/* typically 0 and an unsigned compare */
		tgAddr >= TG_BEG_SRAM &&
#endif /* TG_BEG_SRAM */
		tgAddr + len <= TG_END_SRAM) {
		if (writeTGMem(altp, tgAddr, (caddr_t)mp->b_cont->b_rptr, len, 
			       FALSE, TRUE) < 0) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
	    } else if (tgAddr >= TG_BEG_SCRATCH &&
		       tgAddr + len <= TG_END_SCRATCH) {
		if ((IN32(altp->initNic->handle, altp->initNic->base, 
			  TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		    TG_MHC_LSI_GATE_VERS) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EINVAL);
		    break;
		}

		if (writeTGScratch(altp, tgAddr, (caddr_t)mp->b_cont->b_rptr, 
				   len, FALSE, TRUE, TG_PROCESSOR_A) < 0) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
	    } else if (tgAddr >= TG_BEG_SCRATCH_B_DEBUG &&
		       tgAddr + len <= TG_END_SCRATCH_B_DEBUG) {
		if ((IN32(altp->initNic->handle, altp->initNic->base, 
			  TG_MISC_HOST_CONTROL) & TG_MHC_TIGON_VERSION_MASK) <
		    TG_MHC_LSI_GATE_VERS) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EINVAL);
		    break;
		}

		if (writeTGScratch(altp, tgAddr - TG_SCRATCH_DEBUG_OFF, 
				   (caddr_t)mp->b_cont->b_rptr, 
				   len, FALSE, TRUE, TG_PROCESSOR_B) < 0) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
	    } else if (tgAddr >= TG_BEG_FLASH &&
		       tgAddr + len <= TG_END_FLASH) {
		if (writeTGFlash(altp, tgAddr, (caddr_t)mp->b_cont->b_rptr, 
				 len, FALSE) < 0) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
	    } else if (tgAddr >= TG_BEG_SER_EEPROM &&
		       tgAddr + len <= TG_END_SER_EEPROM) {
		if (writeTGSerEeprom(altp, tgAddr, 
				     (caddr_t)mp->b_cont->b_rptr, 
				     len, FALSE) < 0) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
	    } else if ((tgAddr >= TG_BEG_REGS &&
			tgAddr + len <= TG_END_WRITE_REGS) ||
		       (tgAddr >= TG_BEG_REGS2 &&
			tgAddr + len <= TG_END_REGS)) {
		ALTTRACE2G(wrMemReg, 0x50012, tgAddr, len, 0, 0, 
			   mp->b_cont->b_rptr);
		/* special case writing the TG_WINDOW_BASE register, we
		 * need this to be untouched for accessing the send ring */
		if ((altp->flags & ALT_HOST_UP) &&
		    tgAddr >= TG_WINDOW_BASE && 
		    (tgAddr + len) <= (TG_WINDOW_BASE + 4)) {
		    freemsg(cprspp->cp_private);
		    freemsg(mp->b_cont);
		    mp->b_cont = (mblk_t *)0;
		    altiocnak(wq, mp, 0, EFAULT);
		    break;
		}
		
		alt_swap_memcpy((MEMCPY_TYPE)altp->initNic->base + TG_DEV_ID + 
				(tgAddr - TG_BEG_REGS),
				(MEMCPY_TYPE)(mp->b_cont->b_rptr), len);
	    } else {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EINVAL);
		break;
	    }

	    /* we're done, free up everything and ack the ioctl */
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocack(wq, mp, 0, 0);
	    break;

	default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;

    case ALT_READ_TG_REG:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case TG_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.tgReg, 
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_reg));

	    /* validate some numbers */
	    if (stp->un.tgReg.addr > TG_MAX_REG) {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EINVAL);
		break;
	    }
	    /* LINTED pointer cast may result in improper alignment */
	    stp->un.tgReg.data = IN32(altp->initNic->handle, 
				      altp->initNic->base, 
				      stp->un.tgReg.addr);
	    stp->state = TG_DONE;
	    ALTTRACEG(readReg, 0x5000d, stp->un.tgReg.addr,
		      stp->un.tgReg.data, 0, 0);
	    /* now copy out the local structure */
	    msgcopyout(wq, mp, cprspp->cp_private, sizeof(struct tg_reg), 
		       stp->passedAddr, (caddr_t)&stp->un, FALSE);
	    break;

	case TG_DONE:
	    /* we're done, free up everything and ack the ioctl */
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocack(wq, mp, 0, 0);
	    break;

	default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;
	    
    case ALT_WRITE_TG_REG:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case TG_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.tgReg, 
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_reg));

	    ALTTRACEG(writeReg, 0x5000e, stp->un.tgReg.addr, 
		      stp->un.tgReg.data, 0, 0);
	  
	    /* validate some numbers */
	    /* special case on the WINDOW_BASE reg, we don't let the
	     * users touch this because we depend on it for the send
	     * ring */
	    if (stp->un.tgReg.addr > TG_MAX_REG ||
		((altp->flags & ALT_HOST_UP) &&
		 stp->un.tgReg.addr == TG_WINDOW_BASE)) {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		break;
	    }
	    
	    /* LINTED pointer cast may result in improper alignment */
	    OUT32(altp->initNic->handle, altp->initNic->base, 
		  stp->un.tgReg.addr, stp->un.tgReg.data);

	    /* we're done, free up everything and ack the ioctl */
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocack(wq, mp, 0, 0);
	    break;

	default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;

    case ALT_CACHE_FIRMWARE:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	stp = (struct ioc_state *)cprspp->cp_private->b_rptr;
	switch(stp->state) {
	case FW_GETSTRUCT:
	    /* copy the data into the local structure */
	    MEMCPY((MEMCPY_TYPE)&stp->un.tgFw, 
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_fw));

	    ALTTRACE(cacheFw, 0x5000f, stp->un.tgFw.startAddr, 
		     stp->un.tgFw.textAddr, 
		     stp->un.tgFw.dataAddr, stp->un.tgFw.bssAddr);
	    ALTTRACE(cacheFw, 0x50010, stp->un.tgFw.textLen, 
		     stp->un.tgFw.dataLen, stp->un.tgFw.bssLen, 0);
	    ALTTRACE(cacheFw, 0x50011, stp->un.tgFw.rodataAddr, 
		     stp->un.tgFw.rodataLen, 
		     stp->un.tgFw.sbssAddr, stp->un.tgFw.sbssLen);
	  
	    /* check to see if it fits */
	    if (stp->un.tgFw.textLen > MAX_TEXT_LEN ||
		stp->un.tgFw.dataLen > MAX_DATA_LEN ||
		stp->un.tgFw.rodataLen > MAX_RODATA_LEN) {
		freemsg(cprspp->cp_private);
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, ENOMEM);
		break;
	    }

	    stp->state = FW_TEXT;
	    msgcopyin(wq, mp, cprspp->cp_private, stp->un.tgFw.textLen,
		      stp->un.tgFw.fwText);
	    break;
	    
	case FW_TEXT: {
	      int curLen = stp->un.tgFw.textLen;
	      struct tg_nic *nic = altp->initNic;
	      MEMCPY_TYPE curAddr;
	      mblk_t *curmp = mp->b_cont;
	     
	      switch (nic->tigonRev) {			
		case TG_MHC_LSI_STANDARD_CELL_VERS:
		case TG_MHC_LSI_STANDARD_CELL_VERS_B:
		    curAddr = (MEMCPY_TYPE) tigon2FwText;
		    break;

		case TG_MHC_LSI_GATE_VERS:
		    curAddr = (MEMCPY_TYPE) tigonFwText;
		    break;

		default:
		    ASSERT(0);	/* XXX error! */
	      }

	      ALTTRACE(newText, 0x50034, curLen, 0, 0, 0);
	      /* copy the data into the local area */
	      /* it might not all fit in a single mblk, loop through */
	      /* for little endian hosts this data is byte swapped since
	       * the elf file is generated in big-endian format.  Take care
	       * of this
	       */
	      while (curLen > 0) {
		  alt_swap_memcpy(curAddr,
				  (MEMCPY_TYPE)curmp->b_rptr,
				  MIN(curLen, curmp->b_wptr - curmp->b_rptr));
		  curAddr += MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  curLen -= MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  if (curLen > 0 && !(curmp = curmp->b_cont)) {
		      freemsg(mp->b_cont);
		      mp->b_cont = (mblk_t *)0;
		      altiocnak(wq, mp, 0, EFAULT);
		      break;
		  }
			   
	      }
	      stp->state = FW_DATA;
	      msgcopyin(wq, mp, cprspp->cp_private, stp->un.tgFw.dataLen,
			stp->un.tgFw.fwData);
	      break;
	  }
	    
	case FW_DATA: {
	      int curLen = stp->un.tgFw.dataLen;
	      struct tg_nic *nic = altp->initNic;
	      MEMCPY_TYPE curAddr;
	      mblk_t *curmp = mp->b_cont;

	      switch (nic->tigonRev) {			
		case TG_MHC_LSI_STANDARD_CELL_VERS:
		case TG_MHC_LSI_STANDARD_CELL_VERS_B:
		    curAddr = (MEMCPY_TYPE) tigon2FwData;
		    break;

		case TG_MHC_LSI_GATE_VERS:
		    curAddr = (MEMCPY_TYPE) tigonFwData;
		    break;

		default:
		    ASSERT(0);	/* XXX error! */
	      }

	      /* copy the data into the local area */
	      /* it might not all fit in a single mblk, loop through */
	      /* for little endian hosts this data is byte swapped since
	       * the elf file is generated in big-endian format.  Take care
	       * of this
	       */
	      while (curLen > 0) {
		  alt_swap_memcpy(curAddr,
				  (MEMCPY_TYPE)curmp->b_rptr,
				  MIN(curLen, curmp->b_wptr - curmp->b_rptr));
		  curAddr += MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  curLen -= MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  if (curLen > 0 && !(curmp = curmp->b_cont)) {
		      freemsg(mp->b_cont);
		      mp->b_cont = (mblk_t *)0;
		      altiocnak(wq, mp, 0, EFAULT);
		      break;
		  }
	      }
	      stp->state = FW_RODATA;
	      msgcopyin(wq, mp, cprspp->cp_private, stp->un.tgFw.rodataLen,
			stp->un.tgFw.fwRodata);
	      break;
	  }
	  
	case FW_RODATA: {
	      int curLen = stp->un.tgFw.rodataLen;
	      struct tg_nic *nic = altp->initNic;
	      MEMCPY_TYPE curAddr;
	      mblk_t *curmp = mp->b_cont;

	      switch (nic->tigonRev) {			
		case TG_MHC_LSI_STANDARD_CELL_VERS:
		case TG_MHC_LSI_STANDARD_CELL_VERS_B:
		    curAddr = (MEMCPY_TYPE) tigon2FwRodata;
		    break;

		case TG_MHC_LSI_GATE_VERS:
		    curAddr = (MEMCPY_TYPE) tigonFwRodata;
		    break;

		default:
		    ASSERT(0);	/* XXX error! */
	      }

	      /* copy the data into the local area */
	      /* it might not all fit in a single mblk, loop through */
	      /* for little endian hosts this data is byte swapped since
	       * the elf file is generated in big-endian format.  Take care
	       * of this
	       */
	      while (curLen > 0) {
		  alt_swap_memcpy(curAddr,
				  (MEMCPY_TYPE)curmp->b_rptr,
				  MIN(curLen, curmp->b_wptr - curmp->b_rptr));
		  curAddr += MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  curLen -= MIN(curLen, curmp->b_wptr - curmp->b_rptr);
		  if (curLen > 0 && !(curmp = curmp->b_cont)) {
		      freemsg(mp->b_cont);
		      mp->b_cont = (mblk_t *)0;
		      altiocnak(wq, mp, 0, EFAULT);
		      break;
		  }
	      }

	      /* copy in the rest of the data */
	      switch (nic->tigonRev) {			
		case TG_MHC_LSI_STANDARD_CELL_VERS:
		case TG_MHC_LSI_STANDARD_CELL_VERS_B:
		    tigon2FwStartAddr = stp->un.tgFw.startAddr;
		    tigon2FwTextAddr = stp->un.tgFw.textAddr;
		    tigon2FwTextLen = stp->un.tgFw.textLen;
		    tigon2FwDataAddr = stp->un.tgFw.dataAddr;
		    tigon2FwDataLen = stp->un.tgFw.dataLen;
		    tigon2FwRodataAddr = stp->un.tgFw.rodataAddr;
		    tigon2FwRodataLen = stp->un.tgFw.rodataLen;
		    tigon2FwBssAddr = stp->un.tgFw.bssAddr;
		    tigon2FwBssLen = stp->un.tgFw.bssLen;
		    tigon2FwSbssAddr = stp->un.tgFw.sbssAddr;
		    tigon2FwSbssLen = stp->un.tgFw.sbssLen;
		    break;

		case TG_MHC_LSI_GATE_VERS:
		    tigonFwStartAddr = stp->un.tgFw.startAddr;
		    tigonFwTextAddr = stp->un.tgFw.textAddr;
		    tigonFwTextLen = stp->un.tgFw.textLen;
		    tigonFwDataAddr = stp->un.tgFw.dataAddr;
		    tigonFwDataLen = stp->un.tgFw.dataLen;
		    tigonFwRodataAddr = stp->un.tgFw.rodataAddr;
		    tigonFwRodataLen = stp->un.tgFw.rodataLen;
		    tigonFwBssAddr = stp->un.tgFw.bssAddr;
		    tigonFwBssLen = stp->un.tgFw.bssLen;
		    tigonFwSbssAddr = stp->un.tgFw.sbssAddr;
		    tigonFwSbssLen = stp->un.tgFw.sbssLen;
		    break;

		default:
		    ASSERT(0);	/* XXX error! */
	      }
		  
	      /* we're done, free up everything and ack the ioctl */
	      freemsg(cprspp->cp_private);
	      freemsg(mp->b_cont);
	      mp->b_cont = (mblk_t *)0;
	      altiocack(wq, mp, 0, 0);
	      break;
	  }

	  default:
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	break;

    case ALT_SET_TUNEPARMS:
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}
	/* stash old link_negotiation status */
	oldStatus = altp->nic->tuneParms[LINK_NEGOTIATE].parm_val;
	
	/* copy the data into the tg_tune area of the genComm area */
	alt_memcpy((MEMCPY_TYPE)&altp->initNic->genComm->gen_com.tuneParms, 
		   (MEMCPY_TYPE)mp->b_cont->b_rptr,
		   sizeof(struct tg_tune));
	
#if defined(__sun)
	/* put the tuned parms into the ndd location as well */
	altUpdateNddParms(altp->initNic);
#endif /* defined(__sun) */

	/* if link negotation status changes we need to tell NIC */
	if (oldStatus != altp->nic->tuneParms[LINK_NEGOTIATE].parm_val)
	    altDoLinkNegotiation(altp);

	/* we're done, free up everything and ack the ioctl */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;

    case ALT_WRITE_PAGE: {
	int curLen = PAGE_SIZE;
	MEMCPY_TYPE curAddr = (MEMCPY_TYPE) altp->initNic->ttpPage;
	mblk_t *curmp = mp->b_cont;
	
	/* if it failed clean up and nak */
	if (cprspp->cp_rval) {
	    freemsg(cprspp->cp_private);
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    break;
	}

	/* copy the data into the page */

	while (curLen > 0) {
	    MEMCPY((MEMCPY_TYPE)curAddr,
		   (MEMCPY_TYPE)mp->b_cont->b_rptr, 
		   MIN(curLen, curmp->b_wptr - curmp->b_rptr));
	    curAddr += MIN(curLen, curmp->b_wptr - curmp->b_rptr);
	    curLen -= MIN(curLen, curmp->b_wptr - curmp->b_rptr);
	    if (curLen > 0 && !(curmp = curmp->b_cont)) {
		freemsg(mp->b_cont);
		mp->b_cont = (mblk_t *)0;
		altiocnak(wq, mp, 0, EFAULT);
		break;
	    }
	    
	}
	/* we're done, free up everything and ack the ioctl */
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocack(wq, mp, 0, 0);
	break;
    }

    case ALT_DUMP_FIRMWARE:
    default:
	ALTTRACE(unkioc, 0x5000a, cprspp->cp_cmd, cprspp->cp_rval, 0, 0);
	freemsg(mp);
	break;
    }
}

/*
 * M_DATA "fastpath" info request.
 * Following the M_IOCTL mblk should come a DL_UNITDATA_REQ mblk.
 * We ack with an M_IOCACK pointing to the original DL_UNITDATA_REQ mblk
 * followed by an mblk containing the raw ethernet header corresponding
 * to the destination address.  Subsequently, we may receive M_DATA
 * msgs which start with this header and may send up
 * up M_DATA msgs with b_rptr pointing to a (ulong) group address
 * indicator followed by the network-layer data (IP packet header).
 * This is all selectable on a per-Stream basis.
 */
STATIC void
altdliochdrinfo(queue_t *wq, mblk_t *mp)
{
    mblk_t *nmp;
    struct altstr *slp;
    struct alt_dladdr *dlap;
    dl_unitdata_req_t *dludp;
    struct ether_header *headerp;
    struct snap_ether_header *sheaderp;
    struct tagged_ether_header *theaderp;
    struct tagged_snap_ether_header *tsheaderp;
    struct altinst *altp;
    int off, len;
    int minsize;
    int size;
    
    slp = (struct altstr *)wq->q_ptr;
    minsize = sizeof (dl_unitdata_req_t) + ALTADDRL;
    
    /*
     * Sanity check the request.
     */
    if (!mp->b_cont ||
	(MBLKL(mp->b_cont) < minsize) ||
	(*((u_long *)mp->b_cont->b_rptr) != DL_UNITDATA_REQ) ||
	!(altp = slp->sl_altp)) {
	altiocnak(wq, mp, 0, EINVAL);
	return;
    }
    
    /*
     * Sanity check the DL_UNITDATA_REQ destination address
     * offset and length values.
     */
    dludp = (dl_unitdata_req_t *)mp->b_cont->b_rptr;
    off = dludp->dl_dest_addr_offset;
    len = dludp->dl_dest_addr_length;
    if (!MBLKIN(mp->b_cont, off, len) || (len != ALTADDRL)) {
	altiocnak(wq, mp, 0, EINVAL);
	return;
    }
    
    dlap = (struct alt_dladdr *)(mp->b_cont->b_rptr + off);

    if (altp->flags & ALT_VLANS_ACTIVE)
	if (altp->initNic->flags & ALT_SNAP)
	    size = sizeof(struct tagged_snap_ether_header);
	else
	    size = sizeof(struct tagged_ether_header);
    else
	if (altp->initNic->flags & ALT_SNAP)
	    size = sizeof(struct snap_ether_header);
	else
	    size = sizeof(struct ether_header);
    
    /*
     * Allocate a new mblk to hold the ether header.
     */
    if ((nmp = allocb(size, BPRI_MED)) == NULL) {
	altiocnak(wq, mp, 0, ENOMEM);
	return;
    }
    nmp->b_wptr += size;
    
    /*
     * Fill in the ether header.
     */
    if (altp->flags & ALT_VLANS_ACTIVE) {
	if (altp->initNic->flags & ALT_SNAP) {
	    tsheaderp = (struct tagged_snap_ether_header *)nmp->b_rptr;
	    ether_copy(&dlap->dl_phys, &tsheaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &tsheaderp->ether_shost);

	    /* fill in the VLAN fields */
	    tsheaderp->tag_id = htons(altVlanTagId);
	    tsheaderp->vlan_id = htons(altp->vlan_id[slp->sl_instance/NALT]);
	    tsheaderp->ether_type = 0; /* unknown length */

	    /* LLC header */
	    tsheaderp->llc_dsap = 0xaa;
	    tsheaderp->llc_ssap = 0xaa;
	    tsheaderp->llc_cntrl = 0x03;

	    /* SNAP header */
	    tsheaderp->snap_org[0] = 0;
	    tsheaderp->snap_org[1] = 0;
	    tsheaderp->snap_org[2] = 0;
	    tsheaderp->snap_type = htons(slp->sl_sap);
	} else {
	    theaderp = (struct tagged_ether_header *)nmp->b_rptr;
	    ether_copy(&dlap->dl_phys, &theaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &theaderp->ether_shost);

	    /* fill in the VLAN fields */
	    theaderp->tag_id = htons(altVlanTagId);
	    theaderp->vlan_id = htons(altp->vlan_id[slp->sl_instance/NALT]);
	    theaderp->ether_type = htons(slp->sl_sap);
	}
    } else {
	if (altp->initNic->flags & ALT_SNAP) {
	    sheaderp = (struct snap_ether_header *)nmp->b_rptr;
	    ether_copy(&dlap->dl_phys, &sheaderp->ether_dhost);
	    ether_copy(&altp->macAddr, &sheaderp->ether_shost);
	    sheaderp->ether_type = 0; /* unknown length */

	    /* LLC header */
	    sheaderp->llc_dsap = 0xaa;
	    sheaderp->llc_ssap = 0xaa;
	    sheaderp->llc_cntrl = 0x03;

	    /* SNAP header */
	    sheaderp->snap_org[0] = 0;
	    sheaderp->snap_org[1] = 0;
	    sheaderp->snap_org[2] = 0;
	    sheaderp->snap_type = htons(slp->sl_sap);
	} else {
	    headerp = (struct ether_header *)nmp->b_rptr;
	    ether_copy(&dlap->dl_phys, &headerp->ether_dhost);
	    ether_copy(&altp->macAddr, &headerp->ether_shost);
	    headerp->ether_type = htons(slp->sl_sap);
	}
    }
    
    /*
     * Link new mblk in after the "request" mblks.
     */
    linkb(mp, nmp);
    
    slp->sl_flags |= SLFAST;
    
    altiocack(wq, mp, altmsgsize(mp->b_cont), 0);
}

/*
 * Reply OK to a DL primitive
 */
STATIC void
altdlack(queue_t *wq, mblk_t *mp, ulong prim)
{
    union DL_primitives *dlp;
    
    ALTTRACEV(altdlack, 0x50015, mp, prim, 0, 0);

    if (!(mp = altmexchange(wq, mp, sizeof(dl_ok_ack_t), M_PCPROTO,
			   DL_OK_ACK)))
	return;

    dlp = (union DL_primitives *)mp->b_rptr;
    dlp->ok_ack.dl_correct_primitive = prim;
    qreply(wq, mp);
}

/*
 * NAK a DL primitive
 */
STATIC void
altdlnak(queue_t *wq, mblk_t *mp, ulong prim, ulong errno, ulong unix_errno)
{
    union DL_primitives *dlp;
    
    ALTTRACEV(altdlnak, 0x50014, mp, prim, errno, unix_errno);

    if (!(mp = altmexchange(wq, mp, sizeof(dl_error_ack_t), M_PCPROTO,
			   DL_ERROR_ACK)))
	return;

    dlp = (union DL_primitives *)mp->b_rptr;
    dlp->error_ack.dl_error_primitive = prim;
    dlp->error_ack.dl_errno = errno;
    dlp->error_ack.dl_unix_errno = unix_errno;
    qreply(wq, mp);
}

/*
 * ACK a DL_BIND_REQ primitive
 */
STATIC void
altdlbindack(queue_t *wq, mblk_t *mp, ulong sap, caddr_t addrp, int addrlen,
	     int maxconind, ulong xidtest)
{
    union DL_primitives *dlp;
    int size;
    
    size = sizeof (dl_bind_ack_t) + addrlen;
    if (!(mp = altmexchange(wq, mp, size, M_PCPROTO, DL_BIND_ACK)))
	return;
    
    dlp = (union DL_primitives *)mp->b_rptr;
    dlp->bind_ack.dl_sap = sap;
    dlp->bind_ack.dl_addr_length = addrlen;
    dlp->bind_ack.dl_addr_offset = sizeof (dl_bind_ack_t);
    dlp->bind_ack.dl_max_conind = maxconind;
    dlp->bind_ack.dl_xidtest_flg = xidtest;
    MEMCPY((MEMCPY_TYPE)(mp->b_rptr + sizeof (dl_bind_ack_t)),
	  addrp, addrlen);
    
    qreply(wq, mp);
}

/*
 * ACK a DL_PHYS_ADDR_REQ primitive
 */
STATIC void
altdlpaack(queue_t *wq, mblk_t *mp, caddr_t addrp, int len)
{
    union DL_primitives *dlp;
    int size;
    
    size = sizeof (dl_phys_addr_ack_t) + len;
    if (!(mp = altmexchange(wq, mp, size, M_PCPROTO, DL_PHYS_ADDR_ACK)))
	return;
    dlp = (union DL_primitives *)mp->b_rptr;
    dlp->physaddr_ack.dl_addr_length = len;
    dlp->physaddr_ack.dl_addr_offset = sizeof (dl_phys_addr_ack_t);
    MEMCPY((caddr_t)(mp->b_rptr + sizeof (dl_phys_addr_ack_t)), addrp, len);
    qreply(wq, mp);
}

#define DLADDRL (80)

STATIC void
altdludbadind(queue_t *wq, mblk_t *mp, u_char *addrp,
u_long addrlen, u_long errno, u_long unix_errno)
{
    union DL_primitives *dlp;
    char buf[DLADDRL];
    int size;
    
    if (addrlen > DLADDRL)
	addrlen = DLADDRL;
    
    MEMCPY(buf, (MEMCPY_TYPE)addrp, addrlen);
    
    size = sizeof (dl_uderror_ind_t) + addrlen;
    
    if (!(mp = altmexchange(wq, mp, size, M_PCPROTO, DL_UDERROR_IND)))
	return;
    
    dlp = (union DL_primitives *)mp->b_rptr;
    dlp->uderror_ind.dl_dest_addr_length = addrlen;
    dlp->uderror_ind.dl_dest_addr_offset = sizeof (dl_uderror_ind_t);
    dlp->uderror_ind.dl_unix_errno = unix_errno;
    dlp->uderror_ind.dl_errno = errno;
    MEMCPY((MEMCPY_TYPE)(mp->b_rptr + sizeof (dl_uderror_ind_t)), 
	   buf, addrlen);
    qreply(wq, mp);
}

/*
 * Exchange one msg for another.  Free old msg and allocate
 * a new one if either (1) mp is NULL, (2), requested size
 * is larger than current size, or (3) reference count of
 * the current msg is greater than one.
 * Set db_type and b_rptr/b_wptr appropriately.
 * Set the first longword of the msg to 'primtype' if
 * 'primtype' is not -1.
 *
 * On allocb() failure, return NULL after sending an
 * M_ERROR msg w/ENOSR error upstream.
 */

STATIC mblk_t *
altmexchange(queue_t *wq, mblk_t *mp, int size,
unsigned char type, long primtype)
{
    if (!mp || (MBLKSIZE(mp) < size) || (DB_REF(mp) > 1)) {
 	freemsg(mp);
	if ((mp = allocb(size, BPRI_MED)) == NULL) {
	    if (mp = allocb(1, BPRI_MED))
		altmerror(wq, mp, ENOSR);
	    return (NULL);
	}
    }
    mp->b_datap->db_type = type;
    mp->b_rptr = mp->b_datap->db_base;
    mp->b_wptr = mp->b_rptr + size;
    if (primtype >= 0)
	*(long *)mp->b_rptr = primtype;
    return (mp);
}

STATIC void
altmerror(queue_t *wq, mblk_t *mp, int error)
{
    ALTTRACE(altmerr, 0x5002a, wq, mp, error, 0);

    if ((mp = (mblk_t *)altmexchange(wq, mp, 1, M_ERROR, -1)) == NULL)
	return;
    *mp->b_rptr = (unsigned char) error;
    qreply(wq, mp);
}

/*
 * Convert an M_IOCTL into an M_IOCACK.
 * Assumption:  mp points to an M_IOCTL msg.
 */
STATIC void
altiocack(queue_t *wq, mblk_t *mp, int count, int error)
{
    struct iocblk *iocp;
    
    ALTTRACEG(altioack, 0x50016, mp, count, error, 0);

    mp->b_datap->db_type = M_IOCACK;
    iocp = (struct iocblk *)mp->b_rptr;
    iocp->ioc_count = count;
    iocp->ioc_error = error;
    iocp->ioc_rval = 0;
    qreply(wq, mp);
}

/*
 * Convert an M_IOCTL into an M_IOCNAK.
 * Assumption:  mp points to an M_IOCTL msg.
 */
STATIC void
altiocnak(queue_t *wq, mblk_t *mp, int count, int error)
{
    struct iocblk *iocp;
    
    ALTTRACEG(altionak, 0x50017, mp, count, error, 0);

    mp->b_datap->db_type = M_IOCNAK;
    iocp = (struct iocblk *)mp->b_rptr;
    iocp->ioc_count = count;
    iocp->ioc_error = error;
    iocp->ioc_rval = 0;
    qreply(wq, mp);
}

STATIC void
msgcopyout(queue_t *wq, mblk_t *mp, mblk_t *state, int size, caddr_t userAddr,
	   caddr_t kernAddr, bool wordCopy)
{
    struct copyreq *cprqp = (struct copyreq *)mp->b_rptr;
    mblk_t *tmp;
    U32 *wordp1, *wordp2;
    int i;

    /* build the copyout message */
    cprqp->cq_private = state;
    cprqp->cq_size = size;
    cprqp->cq_flag = 0;
    cprqp->cq_addr = userAddr;
    mp->b_datap->db_type = M_COPYOUT;
    mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
    /* get rid of the old small msg piece, too complicated to try to use it */
    freemsg(mp->b_cont);
    /* get a new message and copy the data into it */
    if (!(tmp = allocb(size, BPRI_MED))) {
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocnak(wq, mp, 0, EAGAIN);
	return;
    }
    tmp->b_wptr += size;
    if (wordCopy) {
	if ((int)kernAddr & 3) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}
	if ((int)tmp->b_rptr & 3) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}

	wordp1 = (U32 *)kernAddr;
	wordp2 = (U32 *)tmp->b_rptr;
	for (i=0; i < size/4; i++)
	    *wordp2++ = *wordp1++;
    } else
	MEMCPY((MEMCPY_TYPE)tmp->b_rptr, kernAddr, size);
    mp->b_cont = tmp;
    qreply(wq, mp);
}

STATIC void
msgcopyout_swap(queue_t *wq, mblk_t *mp, mblk_t *state, int size, 
		caddr_t userAddr, caddr_t kernAddr, bool wordCopy)
{
    struct copyreq *cprqp = (struct copyreq *)mp->b_rptr;
    mblk_t *tmp;
    U32 *wordp1, *wordp2;
    int i;

    /* build the copyout message */
    cprqp->cq_private = state;
    cprqp->cq_size = size;
    cprqp->cq_flag = 0;
    cprqp->cq_addr = userAddr;
    mp->b_datap->db_type = M_COPYOUT;
    mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
    /* get rid of the old small msg piece, too complicated to try to use it */
    freemsg(mp->b_cont);
    /* get a new message and copy the data into it */
    if (!(tmp = allocb(size, BPRI_MED))) {
	freemsg(mp->b_cont);
	mp->b_cont = (mblk_t *)0;
	altiocnak(wq, mp, 0, EAGAIN);
	return;
    }
    tmp->b_wptr += size;
    if (wordCopy) {
	if ((int)kernAddr & 3) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}
	if ((int)tmp->b_rptr & 3) {
	    freemsg(mp->b_cont);
	    mp->b_cont = (mblk_t *)0;
	    altiocnak(wq, mp, 0, EFAULT);
	    return;
	}

	wordp1 = (U32 *)kernAddr;
	wordp2 = (U32 *)tmp->b_rptr;
	for (i=0; i < size/4; i++)
	    *wordp2++ = htonl(*wordp1++);
    } else
	MEMCPY((MEMCPY_TYPE)tmp->b_rptr, kernAddr, size);
    mp->b_cont = tmp;
    qreply(wq, mp);
}

STATIC void
msgcopyin(queue_t *wq, mblk_t *mp, mblk_t *state, int size, caddr_t userAddr)
{
    struct copyreq *cprqp = (struct copyreq *)mp->b_rptr;

    /* build the copyout message */
    cprqp->cq_private = state;
    cprqp->cq_size = size;
    cprqp->cq_flag = 0;
    cprqp->cq_addr = userAddr;
    mp->b_datap->db_type = M_COPYIN;
    mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
    freemsg(mp->b_cont);
    mp->b_cont = (mblk_t *)0;
    qreply(wq, mp);
}

STATIC void
altdodetach(struct altstr *slp)
{
    struct altinst *altp, *altp2;
    struct altstr *tslp;
    struct ether_addr *ethp;
    int i;
    
    ASSERT(slp->sl_altp);

    ALTTRACE(dodetach, 0x50040, slp, slp->sl_flags, slp->sl_mccount,
	     slp->sl_mctab);

    altp = slp->sl_altp;
    slp->sl_altp = (struct altinst *)0;
    
    /*
     * Disable promiscuous mode if on.
     */
    if (slp->sl_flags & SLALLPHYS)
	slp->sl_flags &= ~SLALLPHYS;
    /*
     * Disable ALLSAP mode if on.
     */
    if (slp->sl_flags & SLALLSAP)
	slp->sl_flags &= ~SLALLSAP;
    
    /*
     * Disable ALLMULTI mode if on.
     */
    if (slp->sl_flags & SLALLMULTI)
	slp->sl_flags &= ~SLALLMULTI;
    
    /*
     * Disable any Multicast Addresses, but only if this is the last
     * one on the instance
     */
    /* tell the hardware */
    ethp = slp->sl_mctab;
    while (slp->sl_mccount) {
	/* found it in the streams table, find it in the instance table
	 * and remove it, but only if reference count goes to 0 */
	for (i = 0; i < altp->mccount; i++) {
	    if (ether_cmp(ethp, &altp->mctab[i].addr) == 0) {
		if (--altp->mctab[i].count == 0) {
		    altp->mccount--;
		    MEMCPY((MEMCPY_TYPE)&altp->mctab[i],
			   (MEMCPY_TYPE)&altp->mctab[i+1],
			   ((altp->mccount - i) *
			    sizeof (struct inst_multicast)));
		    /* and tell the firmware about it */
		    altDelMultiCast(altp, ethp);
		    if (altp->flags & ALT_REDUNDANT)
			altDelMultiCast(altp->attachNic->initAltp, ethp);
		}
		break;
	    }
	}
	ethp++;
	slp->sl_mccount--;
    }

    /*
     * Detach from device structure.
     * Let firmware know the host stack is down 
     * when no other streams are attached to it.
     */
    rw_enter(&altstruplock, RW_READER);
    for (tslp = altstrup; tslp; tslp = tslp->sl_nextp)
	if (tslp->sl_altp == altp && !(tslp->sl_flags & SLDEBUG))
	    break;
    rw_exit(&altstruplock);
#if 0				/* can't remove redundancy now XXX */
    if (tslp == NULL && (altp->flags & ALT_HOST_UP)) {
	/* if this happens, we need to check the redundancy configured and
	 * remove any that's active */
	if ((altp->flags & (ALT_REDUNDANT | ALT_RED_IFCONFIG)) ==
	     (ALT_REDUNDANT | ALT_RED_IFCONFIG)) {
	    ALT_DISABLE(spl, &altp->nicMu);

	    /* unlink it */
	    altp2 = altp->attachNic->initAltp;

	    /* get back the attached's original macAddr (from Serial EEPROM) */
	    altGetMacAddr(altp2, &altp2->macAddr);
	    altSetMacaddr(altp2->attachNic, &altp2->macAddr);
	    
	    /* unlink it */
	    altp->attachNic = (struct tg_nic *)0;
	    altp->nic = altp->initNic;   /* put them back to orig state */
	    altp2->nic = altp2->initNic; /* put them back to orig state */
	    altp->flags &= 
		~(ALT_REDUNDANT |  ALT_RED_IFCONFIG | ALT_RED_ATTACH);
	    altp2->flags &= 
		~(ALT_REDUNDANT | ALT_RED_IFCONFIG | ALT_RED_ATTACH);
	    altp->nic->flags &= ~(ALT_RED_ACTIVE | ALT_RED_STANDBY);
	    altp2->nic->flags &= ~(ALT_RED_ACTIVE | ALT_RED_STANDBY);
	    altp->nic->altp = altp;
	    altp->ifconfigNic = altp->initNic;
	    altp2->attachNic = (struct tg_nic *)0;
	    for (i=0; i < MAX_VLANS; i++)
		altp2->ipq[i] = (queue_t *)0; /* null it out */
	    ALT_ENABLE(spl, &altp->nicMu);
	    
	    /* make the MAC stop accepting BPDU packets */
	    altDelMultiCast(altp, &etherBPDUaddr);
	    altDelMultiCast(altp2, &etherBPDUaddr);

	    /* delete all other multicast addresses from the list */
	    for (i = 0; i < altp->mccount; i++) {
		altDelMultiCast(altp2, &altp->mctab[i].addr);
	    }


	    /* if promiscuity was on, turn it off, so we're back to normal */
	    if (altp2->flags & ALT_PROMISC) {
		altSetPromiscuous(altp, FALSE);
		altp2->flags &= ~ALT_PROMISC;
	    }

	    altHostState(altp, FALSE);
	    altp->flags &= ~ALT_HOST_UP;
	    altHostState(altp2, FALSE);
	    altp2->flags &= ~ALT_HOST_UP;

	    /* if wasn't the ifconfiged side of a redundant link, check if
	     * it's the attached side.  If so, we don't tell the NIC the
	     * host stack is down because it's still up even if we have
	     * no other streams attached because they're attached to the
	     * ifconfiged side of redundant pair */
	} else if (!(altp->flags & ALT_REDUNDANT)) {
	    altHostState(altp, FALSE);
	    altp->flags &= ~ALT_HOST_UP;
	}
	/* turn off receiver on redundant nic too */
	if (altp->flags & ALT_REDUNDANT) {
	    altHostState(altp->attachNic->initAltp, FALSE);
	    altp->attachNic->initAltp->flags &= ~ALT_HOST_UP;
	}
	altHostState(altp, FALSE);
	altp->flags &= ~ALT_HOST_UP;
    }
#else /* 0 */
    /* old way, just turn off flag, if it's not the attached side */
    /* the attached side typically won't have any streams attached, but
     * it's still gonna be up, so leave it be */
    if (tslp == NULL && (altp->flags & ALT_HOST_UP) && 
	!(altp->flags & ALT_RED_ATTACH)) {
	/* turn off receiver on redundant nic too */
	if (altp->flags & ALT_REDUNDANT) {
	    altHostState(altp->attachNic->initAltp, FALSE);
	    altp->attachNic->initAltp->flags &= ~ALT_HOST_UP;
	}
	altHostState(altp, FALSE);
	altp->flags &= ~ALT_HOST_UP;
    }
#endif /* 0 */
    
    slp->sl_state = DL_UNATTACHED;
    slp->sl_sap = 0;

    altsetipq(altp, slp->sl_instance);
}

STATIC int
altmsgsize(mblk_t *mp)
{
    register int n = 0;
    
    for (; mp; mp = mp->b_cont)
	n += MBLKL(mp);
    
    return(n);
}
